#ifndef YSMACOSXJOYSTICK_IS_INCLUDED
#define YSMACOSXJOYSTICK_IS_INCLUDED

#include <IOKit/hid/IOHIDLib.h>

const int YsJoyReaderMaxNumAxis      = 6;
const int YsJoyReaderMaxNumButton    = 10;
const int YsJoyReaderMaxNumHatSwitch = 2;

class YsJoyReaderElement
{
public:
	int exist;
	IOHIDElementRef elem;
	int value;

	YsJoyReaderElement();
};

class YsJoyReaderAxis : public YsJoyReaderElement
{
public:
	int min,max;
	int scaledMin,scaledMax;
	int calibCenter,calibMin,calibMax;

	YsJoyReaderAxis();
	double GetCalibratedValue();

private:
	void UpdateMinMaxCentre();
};

class YsJoyReaderButton : public YsJoyReaderElement
{
public:
	YsJoyReaderButton();
};

class YsJoyReaderHatSwitch : public YsJoyReaderElement
{
public:
	YsJoyReaderHatSwitch();
	int valueNeutral;
	int value0Deg;
	int value90Deg;
	int value180Deg;
	int value270Deg;
	int GetDiscreteValue() const;
};

class YsJoyReader
{
public:
	YsJoyReaderAxis      axis[YsJoyReaderMaxNumAxis];
	YsJoyReaderButton    button[YsJoyReaderMaxNumButton];
	YsJoyReaderHatSwitch hatSwitch[YsJoyReaderMaxNumHatSwitch];

	YsJoyReader();
	~YsJoyReader();
	int SetUpInterface(int joyId, IOHIDDeviceRef hidDev);
	void Read();

protected:
	static IOHIDManagerRef hidManager;
	static CFMutableArrayRef devArray;

	int joyId;
	IOHIDDeviceRef hidDev;

	void AddAxis(int axisId, IOHIDElementRef elem, int min, int max, int scaledMin, int scaledMax);

public:
	static int SetUpJoystick(YsJoyReader joystick[], int maxNumJoystick);
};

#endif
