
part of web_play;


class C_Bullet extends C_Object
{
  static const int BULLETLIFETIME = 8000;
  static C_Image s_oBulletImg = null;
  double _m_dTotAlive;
  bool _m_bNeverUpdated;
  C_TileMap _m_oMap;

  C_Bullet(C_TileMap i_oMap, double i_dPosX, double i_dPosY, double i_dVelX, double i_dVelY, double i_dDirection, int i_iOwnerID)
  {
    m_iOwnerID = i_iOwnerID;

    _m_oMap = i_oMap;
    _m_stScreenArea = _m_oMap.GetScreenRect();

    m_dPosX = i_dPosX; m_dPosY = i_dPosY;
    _m_dVelX = i_dVelX; _m_dVelY = i_dVelY;
    _m_dDirection = i_dDirection;

    _m_oImage = s_oBulletImg;
    _m_stImageRect = null;
    _m_dTotAlive = 0.0;

    _m_bNeverUpdated = true; //used to skip 1st pos update
  }

  bool Update(double i_dTime)
  {
    if(!_m_bNeverUpdated) {
      _m_dTotAlive += i_dTime;
      if(_m_dTotAlive > BULLETLIFETIME) return false;
      OUpdate(i_dTime);
    }
    _m_bNeverUpdated = false;

    return !_m_oMap.CheckCollision(this);
  }

  void Draw()
  {
    int px = _m_oMap.GetPosX();
    int py = _m_oMap.GetPosY();
    _m_stScreenArea = _m_oMap.GetScreenRect();
    ODraw(s_oBulletImg, _m_stImageRect/*null*/, px, py);
  }

  int _m_iBulletUniqeVal; //to identify this exact bullet
  void SetUId(int i_iBulletUniqeVal) {_m_iBulletUniqeVal = i_iBulletUniqeVal;}
  int GetUId() {return _m_iBulletUniqeVal;}
}

/////////////////////////////////////////////////////////////////////////////////////////
class C_BulletList
{
  List<C_Bullet> _m_oList;
  int _m_iGetPos = 0;

  C_BulletList()
  {
    _m_oList = new List<C_Bullet>();
  }

  C_Bullet GetNext(bool i_bReset)
  {
    if(i_bReset) _m_iGetPos = 0;
    if(_m_iGetPos<_m_oList.length) return _m_oList[_m_iGetPos++];
    else return null;
  }

  void Add(C_Bullet i_oBullet)
  {
    _m_oList.add(i_oBullet);
  }

  void Remove(int i_iUID)
  {
    C_Bullet oBullet;
    int i, iCount = _m_oList.length;

    for(i=0; i<iCount; i++) {
      oBullet = _m_oList[i];

      if(oBullet.GetUId() == i_iUID) {
        if(i<_m_iGetPos) _m_iGetPos--;
        _m_oList.removeAt(i);
        break;
      }
    }
  }

  void Update(double i_dTime)
  {
    C_Bullet oBullet;
    int i, iCount = _m_oList.length;

    for(i=0; i<iCount; i++) {
      oBullet = _m_oList[i];

      if( !oBullet.Update(i_dTime) ) {
        //collision with map or full lifetime reached
        _m_oList.removeAt(i);
        iCount--; i--;
      }
    }
  }

  void Draw()
  {
    int i, iCount = _m_oList.length;

    for(i=0; i<iCount; i++) {
      _m_oList[i].Draw();
    }
  }

  bool CheckCollision(C_Object i_oObject)
  {
    C_Bullet oBullet;
    int i, iCount = _m_oList.length;
    bool bResult = false;
    int iBulletOwner;
    int iObjId = i_oObject.m_iOwnerID;

    if(iObjId>C_Object.COMPUTER_ENEMY_BASEID) iObjId = C_Object.COMPUTER_ENEMY_BASEID; //special case for computer enemies

    for(i=0; i<iCount; i++) {
      oBullet = _m_oList[i];

      iBulletOwner = oBullet.m_iOwnerID;
      if(iBulletOwner>C_Object.COMPUTER_ENEMY_BASEID) iBulletOwner = C_Object.COMPUTER_ENEMY_BASEID; //owner is a computer enemy
      if(iBulletOwner != iObjId) {
        if(i_oObject.OCheckCollision2(oBullet)) {
          _m_oList.removeAt(i);
          bResult = true;
          break; //first found bullet hit wins
        }
      }
    }
    return bResult;
  }
}
