
part of web_play;

class S_Rect
{
  S_Rect(int x, int y, int width, int height)
  {
    this.x = x;
    this.y = y;
    this.width = width;
    this.height = height;
  }

  Set(int x, int y, int width, int height)
  {
    this.x = x;
    this.y = y;
    this.width = width;
    this.height = height;
  }

  Set2(S_Rect src)
  {
    this.x = src.x;
    this.y = src.y;
    this.width = src.width;
    this.height = src.height;
  }

  int x, y, width, height;
}

class C_Image
{
  C_Image() {
  }

  //not null when the texture is loaded
  webgl.Texture texture = null;

  ImageElement image = null;
  int width, height;

  //example (using C_Graph)
  //_graph.loadTextureFromImageSrc(_image, "./files/ship_w.png");
  //_image.isLoaded needs to be checked before the texture is used in drawing
  //image valid, w,h valid (when isLoaded)

  //example (using C_Graph)
  //_graph.loadTextureFromDataRGBA(_image, w, h, rgba_pixels);
  //_image.isLoaded is true directly
  //image is null, w,h valid

  bool isLoaded = false;

  //mask data to be able to do pixel collision detection
  //this is generated with
  //_graph.generateImageMask(_ti);
  List<List<bool>> mask = null;
}

class C_Graph
{
	bool m_bInited = false;

  webgl.RenderingContext _m_oGL;
  webgl.Program _m_oShaderProgram;
  webgl.Program _m_oShaderProgram_T;
  webgl.Program _m_oShaderProgram_TB;
  int m_iViewportWidth;
  int m_iViewportHeight;

  Matrix4 _m_mxIdentity = new Matrix4.identity();
  Matrix4 _m_mxProjection = new Matrix4.identity();
  Matrix4 _m_mxView = new Matrix4.identity();
  Queue<Matrix4> _m_oMatrixStack;

  //shader associations
  webgl.UniformLocation _uPMatrix;
  webgl.UniformLocation _uMVMatrix;
  webgl.UniformLocation _colorUniform;
  webgl.UniformLocation _uPMatrix_T;
  webgl.UniformLocation _uMVMatrix_T;
  webgl.UniformLocation _samplerUniform_T;
  webgl.UniformLocation _cropUniform_T;
  webgl.UniformLocation _uPMatrix_TB;
  webgl.UniformLocation _uMVMatrix_TB;
  webgl.UniformLocation _samplerUniform_TB;
  webgl.UniformLocation _cropUniform_TB;
  webgl.UniformLocation _colorUniform_TB;
  int _aVertexPosition;
  int _aVertexPosition_T;
  int _aTextureCoord_T;
  int _aVertexPosition_TB;
  int _aTextureCoord_TB;

  //databuffers used in drawing
  Float32List _m_lMatrixValues = new Float32List(16);
  Vector3 _m_tmpVector3 = new Vector3(0.0, 0.0, 0.0);
  static final Vector4 _s_WhiteColor = new Vector4(1.0, 1.0, 1.0, 1.0);
  
  //unit rect
  webgl.Buffer _m_UnitRectVertexPosBuf;
  webgl.Buffer _m_UnitRectTextureCoordBuf;
  S_Rect _m_tmpRect1 = new S_Rect(0,0,0,0);
  S_Rect _m_tmpRect2 = new S_Rect(0,0,0,0);

  //initing

  C_Graph()
  {
  }

  void _initShaders()
  {
    //only color:
    // vertex shader source code.
    String vsSource = """
    attribute vec3 aVertexPosition;
  
    uniform mat4 uMVMatrix;
    uniform mat4 uPMatrix;
  
    void main(void) {
      gl_Position = uPMatrix * uMVMatrix * vec4(aVertexPosition, 1.0);
    }
    """;

    // fragment (pixel) shader source code.
    String fsSource = """
    precision mediump float;

    uniform vec4 uColorSource;

    void main(void) {
      gl_FragColor = uColorSource;
    }
    """;

    //only texture:
    String vsSourceT = """
    attribute vec3 aVertexPosition;
    attribute vec2 aTextureCoord;
  
    uniform mat4 uMVMatrix;
    uniform mat4 uPMatrix;
  
    varying vec2 vTextureCoord;
  
    void main(void) {
      gl_Position = uPMatrix * uMVMatrix * vec4(aVertexPosition, 1.0);
      vTextureCoord = aTextureCoord;
    }
    """;

    String fsSourceT = """
    precision mediump float;

    varying vec2 vTextureCoord;

    uniform sampler2D uSampler;
    uniform vec4 uCropSource;

    void main(void) {
      gl_FragColor = texture2D(uSampler, vec2(vTextureCoord.s, vTextureCoord.t));
    }
    """;

    //texture and blend color:
    String vsSourceTB = """
    attribute vec3 aVertexPosition;
    attribute vec2 aTextureCoord;
  
    uniform mat4 uMVMatrix;
    uniform mat4 uPMatrix;

    varying vec2 vTextureCoord;
  
    void main(void) {
      gl_Position = uPMatrix * uMVMatrix * vec4(aVertexPosition, 1.0);
      vTextureCoord = aTextureCoord;
    }
    """;

    String fsSourceTB = """
    precision mediump float;

    varying vec2 vTextureCoord;

    uniform sampler2D uSampler;
    uniform vec4 uCropSource;
    uniform vec4 uColorSource;

    void main(void) {
      gl_FragColor = uColorSource*texture2D(uSampler, vec2(vTextureCoord.x * uCropSource.z, vTextureCoord.y * uCropSource.w) + uCropSource.xy);
    }
    """;

    // vertex shader compilation
    webgl.Shader vs = _m_oGL.createShader(webgl.RenderingContext.VERTEX_SHADER);
    _m_oGL.shaderSource(vs, vsSource);
    _m_oGL.compileShader(vs);

    // fragment shader compilation
    webgl.Shader fs = _m_oGL.createShader(webgl.RenderingContext.FRAGMENT_SHADER);
    _m_oGL.shaderSource(fs, fsSource);
    _m_oGL.compileShader(fs);

    // vertex shader compilation
    webgl.Shader vs_t = _m_oGL.createShader(webgl.RenderingContext.VERTEX_SHADER);
    _m_oGL.shaderSource(vs_t, vsSourceT);
    _m_oGL.compileShader(vs_t);

    // fragment shader compilation
    webgl.Shader fs_t = _m_oGL.createShader(webgl.RenderingContext.FRAGMENT_SHADER);
    _m_oGL.shaderSource(fs_t, fsSourceT);
    _m_oGL.compileShader(fs_t);

    // vertex shader compilation
    webgl.Shader vs_tb = _m_oGL.createShader(webgl.RenderingContext.VERTEX_SHADER);
    _m_oGL.shaderSource(vs_tb, vsSourceTB);
    _m_oGL.compileShader(vs_tb);

    // fragment shader compilation
    webgl.Shader fs_tb = _m_oGL.createShader(webgl.RenderingContext.FRAGMENT_SHADER);
    _m_oGL.shaderSource(fs_tb, fsSourceTB);
    _m_oGL.compileShader(fs_tb);

    // attach shaders to a WebGL program
    _m_oShaderProgram = _m_oGL.createProgram();
    _m_oGL.attachShader(_m_oShaderProgram, vs);
    _m_oGL.attachShader(_m_oShaderProgram, fs);
    _m_oGL.linkProgram(_m_oShaderProgram);
    _m_oGL.useProgram(_m_oShaderProgram);

    _m_oShaderProgram_T = _m_oGL.createProgram();
    _m_oGL.attachShader(_m_oShaderProgram_T, vs_t);
    _m_oGL.attachShader(_m_oShaderProgram_T, fs_t);
    _m_oGL.linkProgram(_m_oShaderProgram_T);
    _m_oGL.useProgram(_m_oShaderProgram_T);

    _m_oShaderProgram_TB = _m_oGL.createProgram();
    _m_oGL.attachShader(_m_oShaderProgram_TB, vs_tb);
    _m_oGL.attachShader(_m_oShaderProgram_TB, fs_tb);
    _m_oGL.linkProgram(_m_oShaderProgram_TB);
    _m_oGL.useProgram(_m_oShaderProgram_TB);

    /**
     * Check if shaders were compiled properly. This is probably the most painful part
     * since there's no way to "debug" shader compilation
     */
    if (!_m_oGL.getShaderParameter(vs, webgl.RenderingContext.COMPILE_STATUS)) {
      print(_m_oGL.getShaderInfoLog(vs));
    }
    if (!_m_oGL.getShaderParameter(fs, webgl.RenderingContext.COMPILE_STATUS)) {
      print(_m_oGL.getShaderInfoLog(fs));
    }
    if (!_m_oGL.getProgramParameter(_m_oShaderProgram, webgl.RenderingContext.LINK_STATUS)) {
      print(_m_oGL.getProgramInfoLog(_m_oShaderProgram));
    }

    if (!_m_oGL.getShaderParameter(vs_t, webgl.RenderingContext.COMPILE_STATUS)) {
      print(_m_oGL.getShaderInfoLog(vs_t));
    }
    if (!_m_oGL.getShaderParameter(fs_t, webgl.RenderingContext.COMPILE_STATUS)) {
      print(_m_oGL.getShaderInfoLog(fs_t));
    }
    if (!_m_oGL.getProgramParameter(_m_oShaderProgram_T, webgl.RenderingContext.LINK_STATUS)) {
      print(_m_oGL.getProgramInfoLog(_m_oShaderProgram_T));
    }

    if (!_m_oGL.getShaderParameter(vs_tb, webgl.RenderingContext.COMPILE_STATUS)) {
      print(_m_oGL.getShaderInfoLog(vs_tb));
    }
    if (!_m_oGL.getShaderParameter(fs_tb, webgl.RenderingContext.COMPILE_STATUS)) {
      print(_m_oGL.getShaderInfoLog(fs_tb));
    }
    if (!_m_oGL.getProgramParameter(_m_oShaderProgram_TB, webgl.RenderingContext.LINK_STATUS)) {
      print(_m_oGL.getProgramInfoLog(_m_oShaderProgram_TB));
    }

    _aVertexPosition = _m_oGL.getAttribLocation(_m_oShaderProgram, "aVertexPosition");
    _uPMatrix = _m_oGL.getUniformLocation(_m_oShaderProgram, "uPMatrix");
    _uMVMatrix = _m_oGL.getUniformLocation(_m_oShaderProgram, "uMVMatrix");
    _colorUniform = _m_oGL.getUniformLocation(_m_oShaderProgram, "uColorSource");

    _aVertexPosition_T = _m_oGL.getAttribLocation(_m_oShaderProgram_T, "aVertexPosition");
    _aTextureCoord_T = _m_oGL.getAttribLocation(_m_oShaderProgram_T, "aTextureCoord");
    _uPMatrix_T = _m_oGL.getUniformLocation(_m_oShaderProgram_T, "uPMatrix");
    _uMVMatrix_T = _m_oGL.getUniformLocation(_m_oShaderProgram_T, "uMVMatrix");
    _samplerUniform_T = _m_oGL.getUniformLocation(_m_oShaderProgram_T, "uSampler");
    _cropUniform_T = _m_oGL.getUniformLocation(_m_oShaderProgram_T, "uCropSource");

    _aVertexPosition_TB = _m_oGL.getAttribLocation(_m_oShaderProgram_TB, "aVertexPosition");
    _aTextureCoord_TB = _m_oGL.getAttribLocation(_m_oShaderProgram_TB, "aTextureCoord");
    _uPMatrix_TB = _m_oGL.getUniformLocation(_m_oShaderProgram_TB, "uPMatrix");
    _uMVMatrix_TB = _m_oGL.getUniformLocation(_m_oShaderProgram_TB, "uMVMatrix");
    _samplerUniform_TB = _m_oGL.getUniformLocation(_m_oShaderProgram_TB, "uSampler");
    _cropUniform_TB = _m_oGL.getUniformLocation(_m_oShaderProgram_TB, "uCropSource");
    _colorUniform_TB = _m_oGL.getUniformLocation(_m_oShaderProgram_TB, "uColorSource");
  }

	bool Init(CanvasElement i_oCanvas)
	{
    m_iViewportWidth = i_oCanvas.width;
    m_iViewportHeight = i_oCanvas.height;
    _m_oGL = i_oCanvas.getContext("webgl");
    if(_m_oGL==null) _m_oGL = i_oCanvas.getContext("experimental-webgl");
    if(_m_oGL==null) {
      m_bInited = false;
      return m_bInited;
    }

    _m_oMatrixStack = new Queue();

    _initShaders();

    //set some gl operation state
    _m_oGL.clearColor(0.0, 0.0, 0.0, 1.0);
    _m_oGL.viewport(0, 0, m_iViewportWidth, m_iViewportHeight);
    _m_oGL.disable(webgl.RenderingContext.DEPTH_TEST);
    _m_oGL.enable(webgl.RenderingContext.BLEND);
    //_m_oGL.blendFunc(webgl.RenderingContext.SRC_ALPHA, webgl.RenderingContext.ONE_MINUS_SRC_ALPHA); //webgl.RenderingContext.ONE
    _m_oGL.blendFuncSeparate(webgl.RenderingContext.SRC_ALPHA, webgl.RenderingContext.ONE_MINUS_SRC_ALPHA,
      webgl.RenderingContext.ONE, webgl.RenderingContext.ONE_MINUS_SRC_ALPHA);


    //create a unit rectangle with texture coords
    _m_UnitRectVertexPosBuf = _m_oGL.createBuffer();
    _m_UnitRectTextureCoordBuf = _m_oGL.createBuffer();

    Float32List lTmpFloats = new Float32List(4*4);
    
    _m_oGL.bindBuffer(webgl.RenderingContext.ARRAY_BUFFER, _m_UnitRectVertexPosBuf);
    // [0]---[1]
    //  |     |
    //  |     |
    // [2]---[3]
    double xd = 0.0; double yd = 0.0;
    double wd = 1.0; double hd = 1.0;
    lTmpFloats[0] = xd;    lTmpFloats[1] = yd;     lTmpFloats[2] = 0.0;
    lTmpFloats[3] = xd+wd; lTmpFloats[4] = yd;     lTmpFloats[5] = 0.0;
    lTmpFloats[6] = xd;    lTmpFloats[7] = yd+hd;  lTmpFloats[8] = 0.0;
    lTmpFloats[9] = xd+wd; lTmpFloats[10] = yd+hd; lTmpFloats[11] = 0.0;
    _m_oGL.bufferData(webgl.RenderingContext.ARRAY_BUFFER, lTmpFloats, webgl.RenderingContext.STATIC_DRAW);

    //set texture coordinates
    _m_oGL.bindBuffer(webgl.RenderingContext.ARRAY_BUFFER, _m_UnitRectTextureCoordBuf);
    double sxd = 0.0; double syd = 0.0;
    wd = 1.0; hd = 1.0;
    lTmpFloats[0] = sxd;    lTmpFloats[1] = syd;
    lTmpFloats[2] = sxd+wd; lTmpFloats[3] = syd;
    lTmpFloats[4] = sxd;    lTmpFloats[5] = syd+hd;
    lTmpFloats[6] = sxd+wd; lTmpFloats[7] = syd+hd;
    _m_oGL.bufferData(webgl.RenderingContext.ARRAY_BUFFER, lTmpFloats, webgl.RenderingContext.STATIC_DRAW);


    m_bInited = true;
    return m_bInited;
	}

  webgl.RenderingContext GetGL()
  {
    if(m_bInited) return _m_oGL;
    else return null;
  }

  void _onTextureLoaded(C_Image oImg)
  {
    oImg.texture = _m_oGL.createTexture();
    //_m_oGL.pixelStorei(webgl.RenderingContext.UNPACK_FLIP_Y_WEBGL, 1);

    _m_oGL.bindTexture(webgl.RenderingContext.TEXTURE_2D, oImg.texture);
    _m_oGL.texImage2D(webgl.RenderingContext.TEXTURE_2D, 0, webgl.RenderingContext.RGBA, webgl.RenderingContext.RGBA, webgl.RenderingContext.UNSIGNED_BYTE, oImg.image);

    //will work on textures with a dimensions that is not a power of 2
    _m_oGL.texParameteri(webgl.RenderingContext.TEXTURE_2D, webgl.RenderingContext.TEXTURE_WRAP_S, webgl.RenderingContext.CLAMP_TO_EDGE);
    _m_oGL.texParameteri(webgl.RenderingContext.TEXTURE_2D, webgl.RenderingContext.TEXTURE_WRAP_T, webgl.RenderingContext.CLAMP_TO_EDGE);
    //_m_oGL.texParameteri(webgl.RenderingContext.TEXTURE_2D, webgl.RenderingContext.TEXTURE_MAG_FILTER, webgl.RenderingContext.LINEAR);
    _m_oGL.texParameteri(webgl.RenderingContext.TEXTURE_2D, webgl.RenderingContext.TEXTURE_MIN_FILTER, webgl.RenderingContext.LINEAR);

    _m_oGL.bindTexture(webgl.RenderingContext.TEXTURE_2D, null);

    oImg.width = oImg.image.width;
    oImg.height = oImg.image.height;
    oImg.isLoaded = true;
  }

  void LoadTextureFromImageSrc(C_Image i_oImg, String i_szFileSrc)
  {
    i_oImg.image = new Element.tag('img');
    i_oImg.image.onLoad.listen((Event e) {
      _onTextureLoaded(i_oImg);
    });
    i_oImg.image.src = i_szFileSrc;

    //i_oImg.isLoaded = true is set later when the texture has been loaded
  }

  void LoadTextureFromDataRGBA(C_Image i_oImg, int w, int h, Uint8List i_lRGBAData)
  {
    i_oImg.texture = _m_oGL.createTexture();
    //_m_oGL.pixelStorei(webgl.RenderingContext.UNPACK_FLIP_Y_WEBGL, 0);

    _m_oGL.bindTexture(webgl.RenderingContext.TEXTURE_2D, i_oImg.texture);
    _m_oGL.texImage2D(
      webgl.RenderingContext.TEXTURE_2D, 0,
      webgl.RenderingContext.RGBA,
      w, h, 0,
      webgl.RenderingContext.RGBA,
      webgl.RenderingContext.UNSIGNED_BYTE,
      i_lRGBAData);

    //will work on textures with dimensions that is not a power of 2
    _m_oGL.texParameteri(webgl.RenderingContext.TEXTURE_2D, webgl.RenderingContext.TEXTURE_WRAP_S, webgl.RenderingContext.CLAMP_TO_EDGE);
    _m_oGL.texParameteri(webgl.RenderingContext.TEXTURE_2D, webgl.RenderingContext.TEXTURE_WRAP_T, webgl.RenderingContext.CLAMP_TO_EDGE);
    //_m_oGL.texParameteri(webgl.RenderingContext.TEXTURE_2D, webgl.RenderingContext.TEXTURE_MAG_FILTER, webgl.RenderingContext.LINEAR);
    //_m_oGL.texParameteri(webgl.RenderingContext.TEXTURE_2D, webgl.RenderingContext.TEXTURE_MIN_FILTER, webgl.RenderingContext.LINEAR);
    _m_oGL.texParameteri(webgl.RenderingContext.TEXTURE_2D, webgl.RenderingContext.TEXTURE_MIN_FILTER, webgl.RenderingContext.NONE);
    
    _m_oGL.bindTexture(webgl.RenderingContext.TEXTURE_2D, null);

    i_oImg.width = w;
    i_oImg.height = h;
    //i_oImg.isLoaded = true is set here directly
    i_oImg.isLoaded = true;
  }

  //drawing

  void _mvPushMatrix()
  {
    _m_oMatrixStack.addFirst(_m_mxView.clone());
  }

  void _mvPopMatrix()
  {
    if (0 == _m_oMatrixStack.length) {
      throw new Exception("Invalid popMatrix!");
    }
    _m_mxView = _m_oMatrixStack.removeFirst();
  }

  void _setMatrixUniforms(int i_iShaderNumber)
  {
    _m_mxProjection.copyIntoArray(_m_lMatrixValues);
    if(i_iShaderNumber==0) _m_oGL.uniformMatrix4fv(_uPMatrix, false, _m_lMatrixValues);
    else if(i_iShaderNumber==1) _m_oGL.uniformMatrix4fv(_uPMatrix_T, false, _m_lMatrixValues);
    else if(i_iShaderNumber==2) _m_oGL.uniformMatrix4fv(_uPMatrix_TB, false, _m_lMatrixValues);

    _m_mxView.copyIntoArray(_m_lMatrixValues);
    if(i_iShaderNumber==0) _m_oGL.uniformMatrix4fv(_uMVMatrix, false, _m_lMatrixValues);
    else if(i_iShaderNumber==1) _m_oGL.uniformMatrix4fv(_uMVMatrix_T, false, _m_lMatrixValues);
    else if(i_iShaderNumber==2) _m_oGL.uniformMatrix4fv(_uMVMatrix_TB, false, _m_lMatrixValues);
  }

  void BeginScene()
  {
    _m_oGL.clear(webgl.RenderingContext.COLOR_BUFFER_BIT /*| webgl.RenderingContext.DEPTH_BUFFER_BIT*/);
  }

  void EndScene()
  {
    _m_oGL.flush();
  }

  void Set2D()
  {
    //set ortho projection matrix
    _m_mxProjection = makeOrthographicMatrix(0.0, m_iViewportWidth.toDouble(), m_iViewportHeight.toDouble(), 0.0, -1.0, 1.0);
  }

  void Set3D()
  {
    //field of view is 45°, width-to-height ratio, hide things closer than 0.1 or further than 100
    _m_mxProjection = makePerspectiveMatrix(radians(45.0), m_iViewportWidth / m_iViewportHeight, 0.1, 100.0);
  }

  int _iCurrentProgram = -1;
  void _SetProgram(int i_iProgram)
  {
    if(i_iProgram != _iCurrentProgram) {
      switch(_iCurrentProgram) {
        case 0:
          _m_oGL.disableVertexAttribArray(_aVertexPosition);
          break;
        case 1:
          _m_oGL.disableVertexAttribArray(_aVertexPosition_T);
          _m_oGL.disableVertexAttribArray(_aTextureCoord_T);
          break;
        case 2:
          _m_oGL.disableVertexAttribArray(_aVertexPosition_TB);
          _m_oGL.disableVertexAttribArray(_aTextureCoord_TB);
          break;
      }
      switch(i_iProgram) {
        case 0:
          _m_oGL.useProgram(_m_oShaderProgram);
          _m_oGL.enableVertexAttribArray(_aVertexPosition);
          break;
        case 1:
          _m_oGL.useProgram(_m_oShaderProgram_T);
          _m_oGL.enableVertexAttribArray(_aVertexPosition_T);
          _m_oGL.enableVertexAttribArray(_aTextureCoord_T);
          break;
        case 2:
          _m_oGL.useProgram(_m_oShaderProgram_TB);
          _m_oGL.enableVertexAttribArray(_aVertexPosition_TB);
          _m_oGL.enableVertexAttribArray(_aTextureCoord_TB);
          break;
      }
      _iCurrentProgram = i_iProgram;
    }
  }

  void DrawImage2D(C_Image i_oImg, S_Rect i_oSrcRect, int x, int y)
  {
    DrawImage2D_Blend(i_oImg, i_oSrcRect, x, y, _s_WhiteColor);
  }

  void DrawImage2D_Blend(C_Image oImg, S_Rect i_oSrcRect, int x, int y, Vector4 i_oBlendColor)
  {
    int sxd = 0;
    int syd = 0;
    int swd = oImg.width;
    int shd = oImg.height;
    if(i_oSrcRect!=null) {
      sxd = i_oSrcRect.x;
      syd = i_oSrcRect.y;
      swd = i_oSrcRect.width;
      shd = i_oSrcRect.height;
    }

    _m_tmpRect1.x = sxd;
    _m_tmpRect1.y = syd;
    _m_tmpRect1.width = swd;
    _m_tmpRect1.height = shd;
    _m_tmpRect2.x = x;
    _m_tmpRect2.y = y;
    _m_tmpRect2.width = swd;
    _m_tmpRect2.height = shd;

    DrawImage2D_Blend2(oImg, _m_tmpRect1, _m_tmpRect2, i_oBlendColor);
  }

  void DrawImage2D_Blend2(C_Image i_oImg, S_Rect i_oSrcRect, S_Rect i_oDstRect, Vector4 i_oBlendColor)
  {
    double xd = i_oDstRect.x.toDouble();
    double yd = i_oDstRect.y.toDouble();
    double sxd = 0.0;
    double syd = 0.0;
    double wimg = i_oImg.width.toDouble();
    double himg = i_oImg.height.toDouble();
    double wd = wimg;
    double hd = himg;
    if(i_oSrcRect!=null) {
      sxd = i_oSrcRect.x.toDouble();
      syd = i_oSrcRect.y.toDouble();
      wd = i_oSrcRect.width.toDouble();
      hd = i_oSrcRect.height.toDouble();
    }

    _SetProgram(2);

    //set blend color
    _m_oGL.uniform4f(_colorUniform_TB, i_oBlendColor.r, i_oBlendColor.g, i_oBlendColor.b, i_oBlendColor.a);
    //set texture position
_m_oGL.uniform4f(_cropUniform_TB, (sxd)/wimg, (syd)/himg, wd/wimg, hd/himg);
//_m_oGL.uniform4f(_cropUniform_TB, (sxd+0.5)/wimg, (syd-0.5)/himg, wd/wimg, hd/himg); //+-0.5 fix to hit centre of texel

    //draw it

    _m_mxIdentity.copyInto(_m_mxView); //_mvMatrixIdentity -> _mvMatrix
_m_tmpVector3.x = xd; _m_tmpVector3.y = yd; _m_tmpVector3.z = 0.0;
//_m_tmpVector3.x = xd+0.5; _m_tmpVector3.y = yd+0.5; _m_tmpVector3.z = 0.0; //add 0.5 to hit centre of pixel
    _m_mxView.translate(_m_tmpVector3);
    _m_tmpVector3.x = i_oDstRect.width.toDouble(); _m_tmpVector3.y = i_oDstRect.height.toDouble(); _m_tmpVector3.z = 0.0;
    _m_mxView.scale(_m_tmpVector3);
    _setMatrixUniforms(2);

    //verticies
    _m_oGL.bindBuffer(webgl.RenderingContext.ARRAY_BUFFER, _m_UnitRectVertexPosBuf);
    _m_oGL.vertexAttribPointer(_aVertexPosition_TB, 3, webgl.RenderingContext.FLOAT, false, 0, 0);
    //texture
    _m_oGL.bindBuffer(webgl.RenderingContext.ARRAY_BUFFER, _m_UnitRectTextureCoordBuf);
    _m_oGL.vertexAttribPointer(_aTextureCoord_TB, 2, webgl.RenderingContext.FLOAT, false, 0, 0);
    _m_oGL.activeTexture(webgl.RenderingContext.TEXTURE0);
    _m_oGL.bindTexture(webgl.RenderingContext.TEXTURE_2D, i_oImg.texture);
    _m_oGL.uniform1i(_samplerUniform_TB, 0);

    _m_oGL.drawArrays(webgl.RenderingContext.TRIANGLE_STRIP, 0, 4); // triangle strip, start at 0, total 4

    _m_oGL.bindTexture(webgl.RenderingContext.TEXTURE_2D, null);
  }

  void DrawBuffer(C_Image i_oImg, S_Rect r, int x, int y, webgl.Buffer vb, webgl.Buffer tb, int iNumTriangles)
  {
    _SetProgram(1);

    //draw it

    _m_mxIdentity.copyInto(_m_mxView); //_mvMatrixIdentity -> _mvMatrix
    _m_tmpVector3.x = -x.toDouble()+r.x; _m_tmpVector3.y = -y.toDouble()+r.y; _m_tmpVector3.z = 0.0;
    _m_mxView.translate(_m_tmpVector3);
    _setMatrixUniforms(1);

    //verticies
    _m_oGL.bindBuffer(webgl.RenderingContext.ARRAY_BUFFER, vb);
    _m_oGL.vertexAttribPointer(_aVertexPosition_T, 3, webgl.RenderingContext.FLOAT, false, 0, 0);
    //texture
    _m_oGL.bindBuffer(webgl.RenderingContext.ARRAY_BUFFER, tb);
    _m_oGL.vertexAttribPointer(_aTextureCoord_T, 2, webgl.RenderingContext.FLOAT, false, 0, 0);
    _m_oGL.activeTexture(webgl.RenderingContext.TEXTURE0);
    _m_oGL.bindTexture(webgl.RenderingContext.TEXTURE_2D, i_oImg.texture);
    _m_oGL.uniform1i(_samplerUniform_T, 0);

    _m_oGL.drawArrays(webgl.RenderingContext.TRIANGLES, 0, iNumTriangles);

    _m_oGL.bindTexture(webgl.RenderingContext.TEXTURE_2D, null);
  }

  void DrawRectangle2D(S_Rect i_oDstRect, Vector4 i_oBlendColor)
  {
    _SetProgram(0);

    //set blend color
    _m_oGL.uniform4f(_colorUniform, i_oBlendColor.r, i_oBlendColor.g, i_oBlendColor.b, i_oBlendColor.a);

    //draw it

    _m_mxIdentity.copyInto(_m_mxView); //_mvMatrixIdentity -> _mvMatrix
    _m_tmpVector3.x = i_oDstRect.x.toDouble(); _m_tmpVector3.y = i_oDstRect.y.toDouble(); _m_tmpVector3.z = 0.0;
    _m_mxView.translate(_m_tmpVector3);
    _m_tmpVector3.x = i_oDstRect.width.toDouble(); _m_tmpVector3.y = i_oDstRect.height.toDouble(); _m_tmpVector3.z = 0.0;
    _m_mxView.scale(_m_tmpVector3);
    _setMatrixUniforms(0);

    //verticies
    _m_oGL.bindBuffer(webgl.RenderingContext.ARRAY_BUFFER, _m_UnitRectVertexPosBuf);
    _m_oGL.vertexAttribPointer(_aVertexPosition, 3, webgl.RenderingContext.FLOAT, false, 0, 0);

    _m_oGL.drawArrays(webgl.RenderingContext.TRIANGLE_STRIP, 0, 4); // triangle strip, start at 0, total 4

    _m_oGL.bindTexture(webgl.RenderingContext.TEXTURE_2D, null);
  }

  void GenerateImageMask(C_Image i_oImg)
  {
    //read from texture
    Uint8List pixels = new Uint8List(4 * i_oImg.width * i_oImg.height);
    webgl.Framebuffer fb = _m_oGL.createFramebuffer();
    _m_oGL.bindFramebuffer(webgl.RenderingContext.FRAMEBUFFER, fb);
    _m_oGL.framebufferTexture2D(webgl.RenderingContext.FRAMEBUFFER, webgl.RenderingContext.COLOR_ATTACHMENT0,
      webgl.RenderingContext.TEXTURE_2D, i_oImg.texture, 0);
    _m_oGL.readPixels(0, 0, i_oImg.width, i_oImg.height, webgl.RenderingContext.RGBA, webgl.RenderingContext.UNSIGNED_BYTE, pixels);
    _m_oGL.bindFramebuffer(webgl.RenderingContext.FRAMEBUFFER, null); //back to screen
    _m_oGL.deleteFramebuffer(fb);

    //create mask pixel for pixel
    int i, j;
    int pos = 0;
    i_oImg.mask = new List<List<bool>>(i_oImg.height);
    for(i=0; i<i_oImg.height; i++) {
      i_oImg.mask[i] = new List<bool>(i_oImg.width);
      for(j=0; j<i_oImg.width; j++) {
        i_oImg.mask[i][j] = ((pixels[pos+0]+pixels[pos+1]+pixels[pos+2])!=0);
        pos+=4;
      }
    }
  }

}
