
part of web_play;


class C_Object
{
  //ID constants
  static final int COMPUTER_ENEMY_BASEID = 8;
  static final int DOOR_BASEID = 112;
  static final Random rand = new Random();

  //common object info about position and position change (pixels)
  double m_dPosX;           //object global position in x-direction
  double m_dPosY;           //object global position in y-direction
  double _m_dVelX;           //object speed in x-direction
  double _m_dVelY;           //object speed in y-direction
  double _m_dAcceleration;   //object acceleration in the heading direction
  double _m_dDirection;      //object direction in degrees

  //acceleration modifiers
  double _m_dResistanceX, _m_dResistanceY; //velocity dependent object (air)resistance (0 for no resistance)
  double _m_dGravityX, _m_dGravityY;       //the gravity the object will feel (can be negative) (0 for no gravity) (pixel/sec2)

  //approximate direction
  int    _m_iDirStep;        //number of degrees between each step/image (10=default)
  int    _m_iRealDir = 0;    //(int)m_dDirection/m_iDirStep (the direction that the object can have)

  //object identifiers
  static int _s_iIDCounter = 0;
  int         m_iOwnerID;    //id of the object that owns this object (or the main objects id)

  //graphics for the object (for drawing and collision)
  C_Image    _m_oImage = null;
  S_Rect     _m_stImageRect = null;
  bool       _m_bCanCollide; //this must be false when the object is displayed but it cannot collide. ex. during explosion maybe
  //drawing variables
  S_Rect _m_stScreenArea;    //the area of the map that is visible on screen (from map object)
  static C_Graph s_oGraph;   //must be set externally before drawing any objects

  S_Rect     _m_stTmpRect = new S_Rect(0,0,0,0);

  //statics
  static double PointDistance(int i_iX1, int i_iY1, int i_iX2, int i_iY2)
  {
    int iDistX = i_iX1 - i_iX2;
    int iDistY = i_iY1 - i_iY2;
    return sqrt(iDistX*iDistX + iDistY*iDistY);
  }

  //constructor
  C_Object()
  {
    _m_bCanCollide = true;
    _m_iDirStep = 10;
    _m_stScreenArea = null;

    m_iOwnerID = -1;

    _m_dAcceleration = 0.0;
    _m_dDirection = 0.0;
    _m_dResistanceX = _m_dResistanceY = 0.0;
    _m_dGravityX = _m_dGravityY = 0.0;
  }

  /* calculates new position and speed, depending on the elapsed time
   * i_dTime    time in milliseconds since last call
   * uses m_dDirection, m_dAcceleration as input; they must be set to the newest value
   */
  void OUpdate(double i_dTime)
  {
    double dTime = (i_dTime*0.001); //easier to work with seconds

    //calc velocity
    _m_iRealDir = (_m_dDirection ~/_m_iDirStep);

    double ax = cos(_m_iRealDir*(_m_iDirStep*(PI/180.0))) * _m_dAcceleration;
    double ay = sin(_m_iRealDir*(_m_iDirStep*(PI/180.0))) * _m_dAcceleration;

    ax += _m_dGravityX;
    ay += _m_dGravityY;
    ax -= _m_dVelX*_m_dResistanceX;
    ay -= _m_dVelY*_m_dResistanceY;

    //calc position
    m_dPosX += (_m_dVelX * dTime) + (0.5*ax*dTime*dTime); //??
    m_dPosY += (_m_dVelY * dTime) + (0.5*ay*dTime*dTime); //??

    //calc vel
    _m_dVelX += ax * dTime;
    _m_dVelY += ay * dTime;
  }


  /* transparent draw of the provided image onto screen if visible
   * (clipping preformed if partly visible)
   * i_oImg     image
   * i_stRect   object rectangle within i_Img, NULL = entire img
   * i_iX       x-cordinate of upper left corner of screen in the global space
   * i_iY       y-cordinate of upper left corner of screen in the global space
   *screenposition is (m_dPosX-i_x, m_dPosY-i_Y)  */
  void ODraw(C_Image i_oImg, S_Rect i_stRect, int i_iX, int i_iY, [Vector4 i_oColor])
  {
    if(i_oImg==null) return;

    bool bDraw = true;
    int iPosX = m_dPosX.toInt() - i_iX;
    int iPosY = m_dPosY.toInt() - i_iY;
    int iWidth, iHeight;
    //S_Rect stRect;

    //get src dimensions
    if(i_stRect!=null) {
      _m_stTmpRect.Set2(i_stRect);
    } else {
      _m_stTmpRect.Set(0, 0, i_oImg.width, i_oImg.height);
    }
    iWidth = _m_stTmpRect.width;
    iHeight = _m_stTmpRect.height;

    //clipping x
    if(iPosX < 0) {
      _m_stTmpRect.x -= iPosX;
      _m_stTmpRect.width += iPosX;
      iPosX = 0;
    } else if(iPosX+iWidth > _m_stScreenArea.width) {
      _m_stTmpRect.width -= (iPosX+iWidth) - _m_stScreenArea.width;
    }
    if(_m_stTmpRect.width <= 0) bDraw = false;
    else { //clipping y
      if(iPosY < 0) {
        _m_stTmpRect.y -= iPosY;
        _m_stTmpRect.height += iPosY;
        iPosY = 0;
      } else if(iPosY+iHeight > _m_stScreenArea.height) {
        _m_stTmpRect.height -= (iPosY+iHeight) - _m_stScreenArea.height;
      }
      if(_m_stTmpRect.height <= 0) bDraw = false;
    }

    //draw?
    if(bDraw) {
      int iScreenX = iPosX+_m_stScreenArea.x;
      int iScreenY = iPosY+_m_stScreenArea.y;
      if(i_oColor!=null) s_oGraph.DrawImage2D_Blend(i_oImg, _m_stTmpRect, iScreenX, iScreenY, i_oColor);
      else s_oGraph.DrawImage2D(i_oImg, _m_stTmpRect, iScreenX, iScreenY);
    }
  }

  //returns x=width and y=height
  int GetSizeX()
  {
    int w;

    if(_m_stImageRect!=null) {
      w = _m_stImageRect.width;
    } else if(_m_oImage!=null) {
      w = _m_oImage.width;
    } else w = 0;

    return w;
  }
  int GetSizeY()
  { //returns x=width and y=height
    int h;

    if(_m_stImageRect!=null) {
      h = _m_stImageRect.height;
    } else if(_m_oImage!=null) {
      h = _m_oImage.height;
    } else h = 0;

    return h;
  }

	int GetCenterPosX()
  {
    int x = GetSizeX();
    return (m_dPosX + x*0.5).toInt();
  }
  int GetCenterPosY()
  {
    int y = GetSizeY();
    return (m_dPosY + y*0.5).toInt();
  }
  
  double GetDirection()
  {
    return _m_dDirection;
  }

  double GetDistanceTo(C_Object i_oObj)
  {
    double dDistX, dDistY, dDist;

    int x1 = GetCenterPosX();
    int y1 = GetCenterPosY();
    int x2 = i_oObj.GetCenterPosX();
    int y2 = i_oObj.GetCenterPosY();

    dDistX = (x1 - x2).toDouble();
    dDistY = (y1 - y2).toDouble();
    dDist = sqrt(dDistX*dDistX + dDistY*dDistY);

    return dDist;
  }

  /* tests if two images have collided, first check if they overlap,
   * and if they do, check their common overlapping rectangles pixel by pixel,
   * and if two pixels on the same position is != 0 there is a collision (regardless of alphachannel)
   * i_OtherImg other object image
   * i_iOtherX  x-cordinate of upper left corner of OtherImg in the global space
   * i_iOtherY  y-cordinate of upper left corner of OtherImg in the global space
   */
  bool OCheckCollision(C_Image i_oOtherImg, int i_iOtherX, int i_iOtherY, S_Rect i_stOtherRect)
  {
    if(!_m_bCanCollide || _m_oImage==null) return false; //no collision

    return C_Collision.Compare(_m_oImage, m_dPosX.toInt(), m_dPosY.toInt(), i_oOtherImg,
      i_iOtherX, i_iOtherY, _m_stImageRect, i_stOtherRect);
  }

  bool OCheckCollision2(C_Object i_oObject)
  {
    if(!_m_bCanCollide || _m_oImage==null || !i_oObject._m_bCanCollide || i_oObject._m_oImage==null) return false; //no collision

    return C_Collision.Compare(_m_oImage, m_dPosX.toInt(), m_dPosY.toInt(), i_oObject._m_oImage,
      i_oObject.m_dPosX.toInt(), i_oObject.m_dPosY.toInt(), _m_stImageRect, i_oObject._m_stImageRect);
  }
}
