
part of web_play;

class C_SoundFx
{
  C_SoundFx(int i_iId)
  {
    this.iId = i_iId;
  }

  int iId;
  AudioBuffer oBuffer = null;

  bool bIsLoaded = false;
}

class C_SoundFxPlayingHandle
{
  AudioBufferSourceNode oNode;
  GainNode oVolume;
  PannerNode oPanner;
  
  double dTime = 0.0; //fade out timer
}

class C_Sound
{
  AudioContext _m_oAudioContext;
  List<C_SoundFx> _m_oSoundList;

  static C_Sound s_oSound = null;

  C_Sound()
  {
    try {
      _m_oAudioContext = new AudioContext();
    } catch(e) {
      _m_oAudioContext = null;
    }
    _m_oSoundList = new List<C_SoundFx>();
    s_oSound = this;
  }

  void LoadSound(String i_szUrl, int i_iId)
  {
    //load .fnt file
    C_SoundFx oSfx = new C_SoundFx(i_iId);
    _m_oSoundList.add(oSfx);
    HttpRequest.request(i_szUrl, method: 'GET', responseType: 'arraybuffer')
      .then((xhr) {   // .success
        if(_m_oAudioContext!=null) {
          _m_oAudioContext.decodeAudioData(xhr.response)
            .then((buffer) {
              // successful decode
              oSfx.oBuffer = buffer;
              oSfx.bIsLoaded = true;
            });
        } else {
          //fake ok
          oSfx.bIsLoaded = true;
        }
      })
      .catchError((Error error) {
        print(error.toString());
      });
  }

  bool IsLoaded(int iId)
  {
    int i;
    for(i=0; i<_m_oSoundList.length; i++) {
      if(_m_oSoundList[i].iId == iId) {
        return _m_oSoundList[i].bIsLoaded;
      }
    }
    return false;
  }
  
  bool m_bOn = true;
  void ToggleOnOff()
  {
    m_bOn = !m_bOn;
  }

  C_SoundFxPlayingHandle PlaySound(int iId, double dVol, double dPan, bool bLoop)
  {
    if(_m_oAudioContext==null || !m_bOn) return null;

    int i;
    for(i=0; i<_m_oSoundList.length; i++) {
      if(_m_oSoundList[i].iId == iId) {
        C_SoundFxPlayingHandle ph = new C_SoundFxPlayingHandle();
        ph.oVolume = _m_oAudioContext.createGain();
        ph.oPanner = _m_oAudioContext.createPanner();
        ph.oVolume.gain.value = dVol;
        ph.oPanner.panningModel = "equalpower";
        ph.oPanner.distanceModel = "linear";
        ph.oPanner.maxDistance = 1.1;
        ph.oPanner.setPosition(dPan, 0.0, 1.0 - dPan.abs());
        
        ph.oNode = _m_oAudioContext.createBufferSource();
        ph.oNode.buffer = _m_oSoundList[i].oBuffer;
        
        ph.oNode.connectNode(ph.oVolume);
        ph.oVolume.connectNode(ph.oPanner);
        ph.oPanner.connectNode(_m_oAudioContext.destination);

        ph.oNode.loop = bLoop;
        ph.oNode.start(0);
        
        return ph;
      }
    }
    return null;
  }
}
