
part of web_play;


class S_MapHeader
{
  String szGridfile;
  String szTileset;
  int iTileSize;
  int iWidth, iHeight;
}


class C_TileMap
{
  final int MAP_MISSION = 1;
  final int MAP_DOGFIGHT = 2;
  final int MAP_RACE = 3;
  final int MAP_MISSION_COOP = 4;

  //map grid info
  List<List<int>> _m_grid;
  int _m_iWidth, _m_iHeight;

  //file info
  S_MapHeader _m_stHeader;
  C_Viewport _m_oExtended;

  //draw info
  S_Rect _m_stDrawarea;
  int _m_iDrawareaWidthSave, _m_iDrawareaHeightSave;
  int _m_iPixMaxX, _m_iPixMaxY;
  int _m_iScrollMaxX, _m_iScrollMaxY;
  int _m_iLastX, _m_iLastY;
  bool _m_bWrapX, _m_bWrapY;
  S_Rect _m_stTmpRect = new S_Rect(0,0,0,0);

  //tileset info
  int _m_iTileSize; //pixelsize of
  int _m_iNumTiles;
  C_Image _m_oTilesImg = new C_Image();
  List<S_Rect> _m_astTileRects;

  //gl info
  webgl.Buffer _m_glMapVertexPositionBuffer;
  webgl.Buffer _m_glMapVertexTextureCoordBuffer;
  int _m_iMapNumVert = 0;

  int _m_iMapType = 1;
  int _m_iMapMaxPlayers;

  //graph and loading
  C_Graph _m_oGraph;

  bool _m_bHdrLoaded = false;
  bool _m_bGridLoaded = false;
  bool _m_bRectInited = false;

  int GetType()
  {
    return _m_iMapType;
  }

  // constructor to create a map from an existing file
  C_TileMap(String i_szFilename, C_Graph i_oGraph, S_Rect i_stDrawarea)
  {
    //copy params
    _m_oGraph = i_oGraph;
    _m_stDrawarea = i_stDrawarea;
    String szFilename = "files/"+i_szFilename;

    //allocate memory for map, and set defaults
    _m_grid = null;
    _m_iLastX = _m_iLastY = 0;

    //load
    LoadHeader(szFilename); //get info about the map/level (a textfile)
    //this starts a chain that loads
    // grid, the mapfile (binary)
    // tileset, png image with tiles is loaded and prepared
    //check IsLoaded() to see if all is done

    // save the area entered by the user if wrap around mode should change later
    _m_iDrawareaWidthSave = _m_stDrawarea.width;
    _m_iDrawareaHeightSave = _m_stDrawarea.height;
  }

  //must be run successfully before using the rest of the functions
  bool IsLoaded()
  {
    if(_m_oTilesImg.isLoaded && _m_bGridLoaded && !_m_bRectInited) {
      int i, j;

      //analyse tileset (get rectangles)
      int iWidth, iHeight, iNumX, iNumY, iTileNum;
      iWidth = this._m_oTilesImg.width;
      iHeight = _m_oTilesImg.height;
      iNumX = iWidth ~/ _m_iTileSize;
      iNumY = iHeight ~/ _m_iTileSize;
      _m_iNumTiles = iNumX * iNumY;
      _m_astTileRects = new List<S_Rect>(_m_iNumTiles);

      iTileNum = 0;
      for(j=0; j<iNumY; j++) {
        for(i=0; i<iNumX; i++) {
          _m_astTileRects[iTileNum] = new S_Rect(i*_m_iTileSize, j*_m_iTileSize, _m_iTileSize, _m_iTileSize);
          iTileNum++;
        }
      }
      _m_bRectInited = true;


      List<double> vert = new List<double>(), texc = new List<double>();
      double wd, xd = 0.0;
      double hd, yd = 0.0;
      double wimg = _m_oTilesImg.width.toDouble();
      double himg = _m_oTilesImg.height.toDouble();

      webgl.RenderingContext gl = _m_oGraph.GetGL();
      _m_glMapVertexPositionBuffer = gl.createBuffer();
      _m_glMapVertexTextureCoordBuffer = gl.createBuffer();
      _m_iMapNumVert = 0;

      //create quad
      gl.bindBuffer(webgl.RenderingContext.ARRAY_BUFFER, _m_glMapVertexPositionBuffer);
      // [0]---[1/5]
      //  |     |
      //  |     |
      // [2/4]-[3]
      for(j=0; j<_m_grid.length; j++) {
        for(i=0; i<_m_grid[j].length; i++) {
xd = i*32.0;
yd = j*32.0;
//xd = i*32.0+0.5; //add 0.5 to hit centre of pixel
//yd = j*32.0+0.5; //add 0.5 to hit centre of pixel
          vert.add(xd);      vert.add(yd);      vert.add(0.0);
          vert.add(xd+32.0); vert.add(yd);      vert.add(0.0);
          vert.add(xd);      vert.add(yd+32.0); vert.add(0.0);

          vert.add(xd+32.0); vert.add(yd+32.0); vert.add(0.0);
          vert.add(xd);      vert.add(yd+32.0); vert.add(0.0);
          vert.add(xd+32.0); vert.add(yd);      vert.add(0.0);
          _m_iMapNumVert += 6;
        }
      }
      gl.bufferData(webgl.RenderingContext.ARRAY_BUFFER, new Float32List.fromList(vert), webgl.RenderingContext.STATIC_DRAW);

      //set texture coordinates
      gl.bindBuffer(webgl.RenderingContext.ARRAY_BUFFER, _m_glMapVertexTextureCoordBuffer);
      double half_px_x = 0.5/wimg; //0.5 to hit centre of texel (+)
      double half_px_y = 0.5/himg; //0.5 to hit centre of texel (-)
      for(j=0; j<_m_grid.length; j++) {
        for(i=0; i<_m_grid[j].length; i++) {
          S_Rect stRectRef = _m_astTileRects[_m_grid[j][i]];
xd = (stRectRef.x.toDouble() / wimg);
yd = (stRectRef.y.toDouble() / himg);
//xd = (stRectRef.x.toDouble() / wimg)+half_px_x;
//yd = (stRectRef.y.toDouble() / himg)-half_px_y;
          wd = (stRectRef.width.toDouble() / wimg);
          hd = (stRectRef.height.toDouble() / himg);
          texc.add(xd);    texc.add(yd);
          texc.add(xd+wd); texc.add(yd);
          texc.add(xd);    texc.add(yd+hd);

          texc.add(xd+wd); texc.add(yd+hd);
          texc.add(xd);    texc.add(yd+hd);
          texc.add(xd+wd); texc.add(yd);
        }
      }
      gl.bufferData(webgl.RenderingContext.ARRAY_BUFFER, new Float32List.fromList(texc), webgl.RenderingContext.STATIC_DRAW);

      _m_oGraph.GenerateImageMask(_m_oTilesImg);
    }

    return (_m_bHdrLoaded && _m_bGridLoaded && _m_oTilesImg.isLoaded && _m_bRectInited);
  }

  //set the draw area
  void SetScreenRect(S_Rect i_pstDrawarea)
  {
    _m_stDrawarea = i_pstDrawarea;
    _m_iDrawareaWidthSave = _m_stDrawarea.width;
    _m_iDrawareaHeightSave = _m_stDrawarea.height;
    SetMapWrapping(_m_bWrapX, _m_bWrapY);
  }

  // set if the map shall stop or wrap around when scrolling past the borders
  // must be done once after all is loaded
  void SetMapWrapping(bool i_bHorizontal, bool i_bVertical)
  {
    _m_bWrapX = i_bHorizontal;
    _m_bWrapY = i_bVertical;

    _m_stDrawarea.width = _m_iDrawareaWidthSave;
    _m_stDrawarea.height = _m_iDrawareaHeightSave;
    // if the show area is larger than the map, and wrapparound
    //  shall not be used, we need to limit the showarea
    if(!_m_bWrapX && _m_iWidth*_m_iTileSize<_m_iDrawareaWidthSave) _m_stDrawarea.width = _m_iWidth*_m_iTileSize;
    if(!_m_bWrapY && _m_iHeight*_m_iTileSize<_m_iDrawareaHeightSave) _m_stDrawarea.height = _m_iHeight*_m_iTileSize;

    // calc maximum upper left coordinate we can scroll to
    //  without getting outside the map (used if no wrapparound)
    _m_iPixMaxX = _m_stDrawarea.x + _m_stDrawarea.width;
    _m_iPixMaxY = _m_stDrawarea.y + _m_stDrawarea.height;
    _m_iScrollMaxX = _m_iWidth*_m_iTileSize - _m_stDrawarea.width;
    _m_iScrollMaxY = _m_iHeight*_m_iTileSize - _m_stDrawarea.height;
    if(_m_iScrollMaxX < 0) _m_iScrollMaxX = 0;
    if(_m_iScrollMaxY < 0) _m_iScrollMaxY = 0;
  }

  void LoadHeader(String i_szMapHeader)
  {
//    _state = 1;

    //load .des file
    HttpRequest.getString(i_szMapHeader)
      .then((String fileContents) {   // .success
        _onHeaderLoaded(fileContents);
//        _state = 2;
      })
      .catchError((Error error) {
        print(error.toString());
//        _state = 3;
      });
  }

  void _LoadGrid(String i_szMapGrid)
  {
//    _state = 1;

    //load .fnt file
    HttpRequest.request(i_szMapGrid, method: 'GET', responseType: 'arraybuffer')
      .then((xhr) {   // .success
        _onGridLoaded(xhr.response);
      })
      .catchError((Error error) {
        print(error.toString());
//        _state = 3;
      });
  }

  void _LoadTileSet(String i_szTileSet)
  {
    //load .png file
    _m_oGraph.LoadTextureFromImageSrc(_m_oTilesImg, i_szTileSet);
  }

  List<String> _GetTokens(String i_szString)
  {
    int iEnd = 0;
    List<String> oTokens = new List<String>();
    String szWorkString = i_szString.trim();
    while(iEnd!=-1) {
      iEnd = szWorkString.indexOf(' ', 0);
      if(iEnd!=-1) {
        String szToken = szWorkString.substring(0, iEnd);
        szWorkString = szWorkString.substring(iEnd).trim();
        oTokens.add(szToken);
      } else {
        String szToken = szWorkString.trim();
        oTokens.add(szToken);
      }
    }
    return oTokens;
  }


  // loads the map info from the file contents
  void _onHeaderLoaded(String i_szMapHeaderData)
  {
    int i, j;
    int iEnemyCount = 0;
    int iZoneCounter = 0;
    int iDoorCounter = 0;
    int iStartPosCounter = 0;

    _m_stHeader = new S_MapHeader();
    _m_oExtended = new C_Viewport();
    _m_iMapType = 1;
    C_Player.SHIPGRAVITYBASEX = C_Player.DEFAULT_SHIPGRAVITYBASEX;
    C_Player.SHIPGRAVITYBASEY = C_Player.DEFAULT_SHIPGRAVITYBASEY;
    C_Player.SHIPRESISTANCEX = C_Player.DEFAULT_SHIPRESISTANCEX;
    C_Player.SHIPRESISTANCEY = C_Player.DEFAULT_SHIPRESISTANCEY;

    List<String> oLines = i_szMapHeaderData.split('\n');

    for(i=0; i<oLines.length; i++) {
      String strLine = oLines[i].trim();
    	if(strLine.length == 0) continue;

    	int iCmdEnd = strLine.indexOf(' ', 0);
      String strCommand = strLine.substring(0, iCmdEnd);
      String strData = strLine.substring(iCmdEnd).trim();
      //List<String> oTokens = strData.split(" "); //does not work ok, we want to ignore any number of white space between tokens
      List<String> oTokens = _GetTokens(strData);

      if(strCommand.compareTo("*MAPTYPE") == 0) {
        String szType = oTokens[0];
        if(szType.compareTo("RACE") == 0) _m_iMapType = 3;
        if(szType.compareTo("DOGFIGHT") == 0) _m_iMapType = 2;
        if(szType.compareTo("MISSION") == 0) _m_iMapType = 1;
        if(szType.compareTo("MISSION_COOP") == 0) _m_iMapType = 4;
        if(_m_iMapType == 3) _m_oExtended.m_iNumLapsToRace = int.parse(oTokens[1]);
      } else if(strCommand.compareTo("*MAPSIZE") == 0) {
        _m_stHeader.iWidth = int.parse(oTokens[0]);
        _m_stHeader.iHeight = int.parse(oTokens[1]);
      } else if(strCommand.compareTo("*MAPFILE") == 0) {
        _m_stHeader.szGridfile = strData;;
      } else if(strCommand.compareTo("*TILESIZE") == 0) {
        _m_stHeader.iTileSize = int.parse(oTokens[0]);
      } else if(strCommand.compareTo("*TILESET") == 0) {
        _m_stHeader.szTileset = strData.replaceFirst(".tga", ".png");
      } else if(strCommand.compareTo("*GRAVITY") == 0) {
        C_Player.SHIPGRAVITYBASEX = double.parse(oTokens[0]);
        C_Player.SHIPGRAVITYBASEY = double.parse(oTokens[1]);
      } else if(strCommand.compareTo("*RESISTANCE") == 0) {
        C_Player.SHIPRESISTANCEX = double.parse(oTokens[0]);
        C_Player.SHIPRESISTANCEY = double.parse(oTokens[1]);
      } else if(strCommand.compareTo("*PLAYERSTARTPOS") == 0) {
        _m_oExtended.m_aiPlayerStartPosX[iStartPosCounter] = int.parse(oTokens[0]);
        _m_oExtended.m_aiPlayerStartPosY[iStartPosCounter] = int.parse(oTokens[1]);
        iStartPosCounter++;
      } else if(strCommand.compareTo("*MAXPLAYERS") == 0) {
        _m_iMapMaxPlayers = int.parse(oTokens[0]);
      } else if(strCommand.compareTo("*LANDINGZONE") == 0) {
        int x = int.parse(oTokens[0]);
        int y = int.parse(oTokens[1]);
        int w = int.parse(oTokens[2]);
        bool bHomeBase = (oTokens[3].compareTo("HOMEBASE") == 0);
        bool bExtraLife = (oTokens[4].compareTo("EXTRALIFE") == 0);
        bool bAntenna = (oTokens[5].compareTo("ANTENNA") == 0);
        bool bHouse = (oTokens[6].compareTo("WAREHOUSE") == 0);
        int iNumCargo;
        List<int> aiCargo = null;
        if(!bHomeBase) {
          iNumCargo = int.parse(oTokens[7]);
          if(iNumCargo>0) {
            aiCargo = new List<int>(iNumCargo);
            for(j=0; j<iNumCargo; j++) aiCargo[j] = int.parse(oTokens[8+j]);
          }
        }
        C_LandingZone oZone = new C_LandingZone(this, iZoneCounter, x, y, w, bAntenna, bHouse, aiCargo, bExtraLife);
        _m_oExtended.m_oLandingZList.add(oZone);
        iZoneCounter++;
      } else if(strCommand.compareTo("*ENEMY") == 0) {
        S_EnemyInfo stParams = new S_EnemyInfo();
        stParams.iEnemyType = int.parse(oTokens[0]);

        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        stParams.iAngle = int.parse(oTokens[1]);
        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        stParams.iFireInterval = int.parse(oTokens[1]);
        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        stParams.iSpeed = int.parse(oTokens[1]);
        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        int iNumWP = int.parse(oTokens[1]);
        stParams.astWayPoints = new List<Point>(iNumWP); //init number of waypoints
        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        for(j=0; j<iNumWP; j++) stParams.astWayPoints[j] = new Point(int.parse(oTokens[j+1]), 0);
        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        for(j=0; j<iNumWP; j++) stParams.astWayPoints[j] = new Point(stParams.astWayPoints[j].x, int.parse(oTokens[j+1]));
        //add enemy to list
        C_ComputerEnemy m_oCompEnemy = new C_ComputerEnemy(this, _m_oExtended.m_oBulletList, stParams, C_Object.COMPUTER_ENEMY_BASEID+iEnemyCount);
        _m_oExtended.m_oCompEnemyList.add(m_oCompEnemy);
        iEnemyCount++;
      } else if(strCommand.compareTo("*CHECKPOINTS") == 0) {
        int iNumCP = int.parse(oTokens[0]);
        //init number of checkpoints and read them
        for(j=0; j<iNumCP; j++) {
          strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
          C_CheckPoint oCP = new C_CheckPoint(this, new Point(int.parse(oTokens[1]), int.parse(oTokens[2])), new Point(int.parse(oTokens[3]), int.parse(oTokens[4])), j);
          _m_oExtended.m_oCheckPList.add(oCP);
          if(j==0) oCP.SetBlinkState(true); //first CP is the one to get first
        }
      } else if(strCommand.compareTo("*DOOR") == 0) {
        int x, y, iAngle;
        S_DoorInfo stParams = new S_DoorInfo();
        x = int.parse(oTokens[0]);
        y = int.parse(oTokens[1]);
        stParams.iSize = int.parse(oTokens[2]);

        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        iAngle = int.parse(oTokens[1]);
        stParams.bHorizontal = ((iAngle~/90)%2)!=0;
        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        stParams.iOpenedForTime = int.parse(oTokens[1]);
        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        stParams.iClosedForTime = int.parse(oTokens[1]);
        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        stParams.iOpeningSpeed = int.parse(oTokens[1]);
        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        stParams.iClosingSpeed = int.parse(oTokens[1]);

        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        stParams.iNumButtons = int.parse(oTokens[1]);

        stParams.stButtonPos = new List<Point>(stParams.iNumButtons);
        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        for(j=0; j<stParams.iNumButtons; j++) stParams.stButtonPos[j] = new Point(int.parse(oTokens[j+1]), 0);
        strLine = oLines[++i].trim(); oTokens = _GetTokens(strLine);
        for(j=0; j<stParams.iNumButtons; j++) stParams.stButtonPos[j] = new Point(stParams.stButtonPos[j].x, int.parse(oTokens[j+1]));

        //add door to list
        C_Door oDoor = new C_Door(this, x, y, stParams, C_Object.DOOR_BASEID+iDoorCounter);
        _m_oExtended.m_oDoorList.add(oDoor);
        iDoorCounter++;
      }
    }

    _m_iWidth = _m_stHeader.iWidth;
    _m_iHeight = _m_stHeader.iHeight;
    _m_iTileSize = _m_stHeader.iTileSize;

    _m_grid = new List<List<int>>(_m_iHeight); //alloc rows and columns
    for(i=0; i<_m_iHeight; i++) _m_grid[i] = new List<int>(_m_iWidth);

    _m_bHdrLoaded = true;

    //begin loading grid and tileset
    _LoadGrid("files/"+_m_stHeader.szGridfile);
    _LoadTileSet("files/"+_m_stHeader.szTileset);
  }

  // loads the map grid data from the file mentioned in the "header"
  void _onGridLoaded(ByteBuffer i_pMem)
  {
    int i;
    for(i=0; i<_m_iHeight; i++) {
      _m_grid[i] = i_pMem.asUint8List(i *_m_iWidth, _m_iWidth);
    }
    _m_bGridLoaded = true;
  }

  // draws all the currently visible tiles in the map on screen
  void Draw()
  {
/*    int iTempX, iTempY;
    int iBaseTileX, iBaseOffsetX;
    int iTileX, iTileY;
    int iOffsetX, iOffsetY;
    int iScreenX = _m_stDrawarea.x, iScreenY = _m_stDrawarea.y;
    int x, y, w, h; //calculated srcrectangle within the tileset
    S_Rect stRectRef; //to access the right tile within the tileset

    iBaseTileX = iTileX = _m_iLastX ~/ _m_iTileSize;
    iBaseOffsetX = iOffsetX = _m_iLastX % _m_iTileSize;
    iTileY = _m_iLastY ~/ _m_iTileSize;
    iOffsetY = _m_iLastY % _m_iTileSize;
    do {
      do {
        stRectRef = _m_astTileRects[_m_grid[iTileY][iTileX]];
        x = stRectRef.x + iOffsetX;
        y = stRectRef.y + iOffsetY;
        iTempX = _m_iPixMaxX - iScreenX;
        iTempY = _m_iPixMaxY - iScreenY;
        w = (_m_iTileSize - iOffsetX)<iTempX ? (_m_iTileSize - iOffsetX):iTempX;
        h = (_m_iTileSize - iOffsetY)<iTempY ? (_m_iTileSize - iOffsetY):iTempY;

        _stTmpRect.Set(x,y,w,h);
        _m_oGraph.DrawImage2D(_m_oTilesImg, _stTmpRect, iScreenX, iScreenY);

        iTileX++;
        if(_m_bWrapX) iTileX %= _m_iWidth;
        iScreenX += w;
        iOffsetX = 0;
      } while (iScreenX < _m_iPixMaxX);
      iOffsetX = iBaseOffsetX;
      iOffsetY = 0;
      iTileX = iBaseTileX;
      iTileY++;
      if(_m_bWrapY) iTileY %= _m_iHeight;

      iScreenX = _m_stDrawarea.x;
      iScreenY += h;
    } while (iScreenY < _m_iPixMaxY);*/
    _m_oGraph.DrawBuffer(_m_oTilesImg, _m_stDrawarea, _m_iLastX, _m_iLastY, _m_glMapVertexPositionBuffer, _m_glMapVertexTextureCoordBuffer, _m_iMapNumVert);
  }

  // check all of i_stGameArea of the map against the provided image
  S_Rect tmpRect = new S_Rect(0, 0, 0, 0);
  bool CheckMaskCollision(int inX, int inY, int inW, int inH, C_Object i_oTestObject)
  {
    int iTempX, iTempY, iTileNum;
    bool bCollision = false;
    int iBaseTileX, iBaseOffsetX;
    int iTileX, iTileY;
    int iOffsetX, iOffsetY;
    int x, y, w, h; //calculated srcrectangle within the tileset
    S_Rect stRectRef; //to access the right tile within the tileset

    int iScreenX = 0;
    int iScreenY = 0;
    int iScreenMaxX = inW;
    int iScreenMaxY = inH;

    iBaseTileX = iTileX = inX ~/ _m_iTileSize;
    iBaseOffsetX = iOffsetX = inX % _m_iTileSize;
    iTileY = inY ~/ _m_iTileSize;
    iOffsetY = inY % _m_iTileSize;
    do {
      do {
        iTileNum = _m_grid[iTileY][iTileX];
        iTempX = iScreenMaxX - iScreenX;
        iTempY = iScreenMaxY - iScreenY;
        w = (_m_iTileSize - iOffsetX)<iTempX ? (_m_iTileSize - iOffsetX):iTempX;
        h = (_m_iTileSize - iOffsetY)<iTempY ? (_m_iTileSize - iOffsetY):iTempY;

        if(iTileNum!=0) { //only test for collision if nonblack
          stRectRef = _m_astTileRects[iTileNum];
          x = stRectRef.x + iOffsetX;
          y = stRectRef.y + iOffsetY;

          tmpRect.Set(x, y, w, h);
          bCollision = i_oTestObject.OCheckCollision(_m_oTilesImg, iTileX*_m_iTileSize+iOffsetX, iTileY*_m_iTileSize+iOffsetY, tmpRect);
          if(bCollision) return true;
        }

        iTileX++;
        if(_m_bWrapX) iTileX %= _m_iWidth;
        iScreenX += w;
        iOffsetX = 0;
      } while (iScreenX < iScreenMaxX);
      iOffsetX = iBaseOffsetX;
      iOffsetY = 0;
      iTileX = iBaseTileX;
      iTileY++;
      if(_m_bWrapY) iTileY %= _m_iHeight;

      iScreenX = 0;
      iScreenY += h;
    } while (iScreenY < iScreenMaxY);

    return bCollision;
  }

  // tests if an object has collided with nonblack pixels in the map
  bool CheckCollision(C_Object i_oObject)
  {
    bool bResult = false;
    int iX, iY, iW, iH;
    iX = i_oObject.m_dPosX.toInt();
    iY = i_oObject.m_dPosY.toInt();
    iW=i_oObject.GetSizeX(); iH=i_oObject.GetSizeY();

    //check boundries
    if(iX<0 || iY<0 || iX+iW>_m_iWidth*_m_iTileSize || iY+iH>_m_iHeight*_m_iTileSize) bResult = true;

    if(!bResult) { //check pixels
      bResult = CheckMaskCollision(iX,iY,iW,iH, i_oObject);
    }
    return bResult;
	}

  // moves the showpoint of the map relative to the last
  // position and draws the result
  void MoveRelative(int i_iX, int i_iY)
  {
    MoveAbsolute(_m_iLastX + i_iX, _m_iLastY + i_iY);
  }

  // moves the showpoint of the map to a new position,
  // given in world pixel coordinates
  Vector2 _tmpP = new Vector2(0.0, 0.0);
  void MoveAbsolute(int i_iX, int i_iY)
  {
    int iCmpx = _m_bWrapX ? _m_iWidth*_m_iTileSize  : _m_iScrollMaxX;
    int iCmpy = _m_bWrapY ? _m_iHeight*_m_iTileSize : _m_iScrollMaxY;
    _tmpP.x = i_iX.toDouble();
    _tmpP.y = i_iY.toDouble();
    _WrappAroundTest(_tmpP, iCmpx, iCmpy);

    _m_iLastX = _tmpP.x.toInt(); _m_iLastY = _tmpP.y.toInt();
  }

  int GetPosX()
  {
    return _m_iLastX;
  }
  int GetPosY()
  {
    return _m_iLastY;
  }

  S_Rect GetScreenRect()
  {
    return _m_stDrawarea;
  }

  // checks if the values goes outside the limits, and take actions
  bool _WrappAroundTest(Vector2 io_p, int i_iWidth, int i_iHeight)
  {
    bool bResult = true;

    int x = io_p.x.toInt();
    int y = io_p.y.toInt();

    if(x>=i_iWidth) {
      if(_m_bWrapX) x %= i_iWidth;
      else {
        x = i_iWidth;
        bResult = false;
      }
    }
    if(y>=i_iHeight) {
      if(_m_bWrapY) y %= i_iHeight;
      else {
        y = i_iHeight;
        bResult = false;
      }
    }
    if(x<0) {
      if(_m_bWrapX) x = i_iWidth + (x % i_iWidth);
      else {
        x = 0;
        bResult = false;
      }
    }
    if(y<0) {
      if(_m_bWrapY) y = i_iHeight + (y % i_iHeight);
      else {
        y = 0;
        bResult = false;
      }
    }
    io_p.x = x.toDouble();
    io_p.y = y.toDouble();

    return bResult;
  }

  C_Viewport GetExtendedInfo()
  {
    return _m_oExtended;
  }

  int MapType()
  {
    return _m_iMapType;
  }

  int MaxPlayers()
  {
    return _m_iMapMaxPlayers;
  }
}
