library web_play;

import 'package:vector_math/vector_math.dart';
import 'dart:html';
import 'dart:collection';
import 'dart:web_gl' as webgl;
import 'dart:typed_data';
import 'dart:math';
import 'dart:web_audio';

part 'C_Graph.dart';
part 'C_Font.dart';
part 'C_Sound.dart';

part 'C_Object.dart';
part 'C_Viewport.dart';
part 'C_TileMap.dart';
part 'C_MapObjects.dart';
part 'C_Player.dart';
part 'C_Bullet.dart';
part 'C_Collision.dart';


class C_GameHandler
{
  C_Sound _m_oSound;
  C_Graph _m_oGraph;
  C_Font _m_oFont;
  C_Font _m_oFont2;

  double _m_dLastTime = 0.0;

  C_Image _m_oShipImg;
  C_Image _m_oExplosionImg;
  C_Image _m_oBulletImg;
  C_Image _m_oLandingZoneImg;
  C_Image _m_oCpImg;
  C_Image _m_oE0Img, _m_oE1Img, _m_oE2Img, _m_oE3Img;
  C_Image _m_oE4Img, _m_oE5Img, _m_oE6Img, _m_oE7Img;
  C_Image _m_oDoorImg;
  C_Image _m_oCargoFuel;
  C_Image _m_oLapsTime;

  C_TileMap _m_oMap;
  C_Viewport _m_oViewport;
  bool _m_bEnd;
  double _m_dTimeLeft, _m_dTimeToEnd;
  int _m_iState = 0;

  String m_szDesFile;

  CanvasElement m_oCanvas;

  bool _m_bHasWebGL = false;

  C_GameHandler(CanvasElement canvas)
  {
    m_oCanvas = canvas;
    _m_oSound = new C_Sound();
    _m_oGraph = new C_Graph();
    _m_bHasWebGL = _m_oGraph.Init(canvas); //init webgl
  }

  void _loadTextures()
  {
    _m_oShipImg = new C_Image();
    _m_oExplosionImg = new C_Image();
    _m_oBulletImg = new C_Image();
    _m_oLandingZoneImg = new C_Image();
    _m_oCpImg = new C_Image();
    _m_oE0Img = new C_Image();
    _m_oE1Img = new C_Image();
    _m_oE2Img = new C_Image();
    _m_oE3Img = new C_Image();
    _m_oE4Img = new C_Image();
    _m_oE5Img = new C_Image();
    _m_oE6Img = new C_Image();
    _m_oE7Img = new C_Image();
    _m_oDoorImg = new C_Image();
    _m_oCargoFuel = new C_Image();
    _m_oLapsTime = new C_Image();

    _m_oGraph.LoadTextureFromImageSrc(_m_oShipImg, "./files/ship_w.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oExplosionImg, "./files/explosion.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oBulletImg, "./files/bullet.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oLandingZoneImg, "./files/zone.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oCpImg, "./files/cpbk.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE0Img, "./files/enemy0.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE1Img, "./files/enemy1.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE2Img, "./files/enemy2.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE3Img, "./files/enemy3.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE4Img, "./files/enemy4.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE5Img, "./files/enemy5.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE6Img, "./files/enemy6.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE7Img, "./files/enemy7.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oDoorImg, "./files/door.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oCargoFuel, "./files/cargo_fuel.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oLapsTime, "./files/laps_time.png");

    _m_oFont = new C_Font(_m_oGraph);
    _m_oFont.Load("./files/arial-12");

    _m_oFont2 = new C_Font(_m_oGraph);
    _m_oFont2.Load("./files/arial-34");
  }

  double _m_dScale = 1.0;
  void ScreenSize()
  {
    _m_dScale = 1.0;
    if(window.innerWidth>900) _m_dScale = 0.75;
    if(window.innerWidth>1400) _m_dScale = 0.5;
    //double dAspect = window.innerWidth/window.innerHeight;
    m_oCanvas.width  = (window.innerWidth*_m_dScale).toInt();
    m_oCanvas.height = (window.innerHeight*_m_dScale).toInt();
    m_oCanvas.style.width = (window.innerWidth-0).toInt().toString()+'px';
    m_oCanvas.style.height = (window.innerHeight-0).toInt().toString()+'px';

    _m_oGraph.Init(m_oCanvas);
    if(_m_oViewport!=null) {
      _m_oViewport.SetSize(_m_oGraph.m_iViewportWidth, _m_oGraph.m_iViewportHeight);
      _m_oViewport.Draw(); //done to avoid flickering since the canvas is
                           // cleared to white on resize
    }
  }

  List<bool> m_lKeyState = new List<bool>(256);
  void HandleKeyEventDown(KeyboardEvent event)
  {
    KeyEvent keyEvent = new KeyEvent.wrap(event);
    if(keyEvent.keyCode<256) {
      m_lKeyState[keyEvent.keyCode] = true;

      if(keyEvent.keyCode==27 || keyEvent.keyCode==55) { //esc, 7
        if(_m_oViewport!=null) _m_oViewport.SoundOn(false);
        _m_iState = 0; //restart regardless of current state
      } else if(keyEvent.keyCode==78) { //n
        C_Sound.s_oSound.ToggleOnOff();
      }

    }
  }
  void HandleKeyEventUp(KeyboardEvent event)
  {
    KeyEvent keyEvent = new KeyEvent.wrap(event);
    if(keyEvent.keyCode<256) {
      m_lKeyState[keyEvent.keyCode] = false;
    }
  }

  bool POINTINRECT(xx, yy, rect) {
    return (((xx) >= (rect).x) && ((xx) <= (rect).x+(rect).width) && ((yy) >= (rect).y) && ((yy) <= (rect).y+(rect).height));
  }

  double fixAngle(double diff)
  {
    if (diff >= 360)
      return diff - 360;
    else if (diff < 0)
      return diff + 360;

    return diff;
  }

  List<bool> m_lTouch = new List<bool>(5);
  double touch_angle = -1.0;
  int touchSteerIdL = -1;
  int touchSteerIdR = -1;
  int touchThrustId = -1;
  int touchUpId = -1;

  void HandleTouch(TouchEvent event)
  {
    event.preventDefault();

    int i, id;

    //http://www.html5rocks.com/en/mobile/touch/

    if (event.type == 'touchend' ||
        event.type == 'touchcancel' ||
        event.type == 'touchleave')
    {
      for(i=0; i<event.changedTouches.length; i++) {
        id = event.changedTouches[i].identifier;

        if (touchThrustId == id) {
          m_lTouch[0] = false;
          touchThrustId = -1;
        }
        if (touchSteerIdL == id) {
          touchSteerIdL = -1;
          m_lTouch[2] = false;
          //m_lTouch[3] = false;
        }
        if (touchSteerIdR == id) {
          touchSteerIdR = -1;
          //m_lTouch[2] = false;
          m_lTouch[3] = false;
        }
        if (touchUpId == id) {
          m_lTouch[1] = false;
          touchUpId = -1;
        }
      }

      return;
    }

    double px, py;
    //bool bSteer = false;

    if (event.type == 'touchstart') {
      for(i=0; i<event.touches.length; i++) {
        id = event.touches[i].identifier;

        px = (event.touches[i].page.x - m_oCanvas.offsetLeft.toDouble());
        py = (event.touches[i].page.y - m_oCanvas.offsetTop.toDouble());

        if ((py > window.innerHeight-window.innerHeight/4.0) && touchThrustId == -1) {
          // thrust
          touchThrustId = id;
          m_lTouch[0] = true;
          continue;
        }
        if ((py < 0+window.innerHeight/5.0) && touchUpId == -1) {
          // fire, prep for landing
          touchUpId = id;
          m_lTouch[1] = true;
          continue;
        }

        if (px < (window.innerWidth / 2) && touchSteerIdL==-1 ) {
          touchSteerIdL = id;
          m_lTouch[2] = true;
        }
        if (px >= (window.innerWidth / 2) && touchSteerIdR==-1) {
          touchSteerIdR = id;
          m_lTouch[3] = true;
        }
      }
    }

    /*if (event.type == 'touchmove') {
      for(i=0; i<event.touches.length; i++) {
        id = event.touches[i].identifier;
      }
    }*/
  }

  void TouchUpdate()
  {
    /*double ship_angle = _m_oViewport.m_oPlayer.GetDirection();

    double a = 0.0;
    if(touch_angle!= -1.0) a = fixAngle(touch_angle - ship_angle);

    //print(touch_angle.toString()+ " : " + ship_angle.toString() + " : " + a.toString());
    //print("Angle:" + a.toString());

    if (a < 6 || a > 354) {
      m_lTouch[2] = false;
      m_lTouch[3] = false;
      return;
    }
    if (a <= 180) {
      m_lTouch[3] = true; //right
      m_lTouch[2] = false; //not left
    } else if (a > 180) {
      m_lTouch[2] = true; //left
      m_lTouch[3] = false; //not right
    }*/
  }

  static Vector4 oColorDG = new Vector4(0x55/255, 0x55/255, 0x55/255, 1.0);
  static Vector4 oColorLG = new Vector4(0xaa/255, 0xaa/255, 0xaa/255, 1.0);
  static Vector4 oColorR = new Vector4(0xee/255, 0x00/255, 0x00/255, 1.0);
  void DrawLoaderStatus(int progress, int total)
  {
    //_canvas.width
    int iYPos = _m_oGraph.m_iViewportHeight~/2 - 40~/2;
    S_Rect stRect = new S_Rect(0, 0, _m_oGraph.m_iViewportWidth, _m_oGraph.m_iViewportHeight);
    _m_oGraph.DrawRectangle2D(stRect, oColorDG);
    stRect.Set(0, iYPos, _m_oGraph.m_iViewportWidth, 40);
    _m_oGraph.DrawRectangle2D(stRect, oColorLG);
    int step = _m_oGraph.m_iViewportWidth~/(total+1);
    stRect.Set(0, iYPos+2, progress*step+step, 40-4);
    _m_oGraph.DrawRectangle2D(stRect, oColorR);
  }

  bool Loop(num time)
  {
    double dCurTime = time.toDouble() - _m_dLastTime;
    if(dCurTime>500.0) dCurTime = 500.0; //limit frametime to protect against huge lags
    _m_dLastTime = time.toDouble();

    _m_oGraph.Set2D();
    _m_oGraph.BeginScene();

    int iLoaderProgress = 0;

    // update and draw frame
    switch(_m_iState) {
      case 0: //wait for files to load
        if(_m_oShipImg.isLoaded) iLoaderProgress++;
        if(_m_oExplosionImg.isLoaded) iLoaderProgress++;
        if(_m_oBulletImg.isLoaded) iLoaderProgress++;
        if(_m_oLandingZoneImg.isLoaded) iLoaderProgress++;
        if(_m_oCpImg.isLoaded) iLoaderProgress++;
        if(_m_oE0Img.isLoaded) iLoaderProgress++;
        if(_m_oE1Img.isLoaded) iLoaderProgress++;
        if(_m_oE2Img.isLoaded) iLoaderProgress++;
        if(_m_oE3Img.isLoaded) iLoaderProgress++;
        if(_m_oE4Img.isLoaded) iLoaderProgress++;
        if(_m_oE5Img.isLoaded) iLoaderProgress++;
        if(_m_oE6Img.isLoaded) iLoaderProgress++;
        if(_m_oE7Img.isLoaded) iLoaderProgress++;
        if(_m_oDoorImg.isLoaded) iLoaderProgress++;
        if(_m_oCargoFuel.isLoaded) iLoaderProgress++;
        if(_m_oLapsTime.isLoaded) iLoaderProgress++;
        if(_m_oFont.IsLoaded()) iLoaderProgress++;
        if(_m_oFont2.IsLoaded()) iLoaderProgress++;

        for(int i=0; i<10; i++) {
          if(_m_oSound.IsLoaded(i)) iLoaderProgress++;
        }
        DrawLoaderStatus(iLoaderProgress, 30);

        //check loading
        bool bLoaded = true; //assume ok

        if(!_m_oShipImg.isLoaded || !_m_oExplosionImg.isLoaded || !_m_oBulletImg.isLoaded ||
          !_m_oLandingZoneImg.isLoaded || !_m_oCpImg.isLoaded || !_m_oE0Img.isLoaded ||
          !_m_oE1Img.isLoaded || !_m_oE2Img.isLoaded || !_m_oE3Img.isLoaded ||
          !_m_oE4Img.isLoaded || !_m_oE5Img.isLoaded || !_m_oE6Img.isLoaded ||
          !_m_oE7Img.isLoaded || !_m_oDoorImg.isLoaded || !_m_oCargoFuel.isLoaded ||
          !_m_oLapsTime.isLoaded) bLoaded = false;

        if(!_m_oFont.IsLoaded()) bLoaded = false;
        if(!_m_oFont2.IsLoaded()) bLoaded = false;

        for(int i=0; i<10; i++) {
          if(!_m_oSound.IsLoaded(i)) bLoaded = false;
        }

        if(bLoaded) {
          _m_oGraph.GenerateImageMask(_m_oShipImg);
          _m_oGraph.GenerateImageMask(_m_oBulletImg);
          _m_oGraph.GenerateImageMask(_m_oE0Img);
          _m_oGraph.GenerateImageMask(_m_oE1Img);
          _m_oGraph.GenerateImageMask(_m_oE2Img);
          _m_oGraph.GenerateImageMask(_m_oE3Img);
          _m_oGraph.GenerateImageMask(_m_oE4Img);
          _m_oGraph.GenerateImageMask(_m_oE5Img);
          _m_oGraph.GenerateImageMask(_m_oE6Img);
          _m_oGraph.GenerateImageMask(_m_oE7Img);
          _m_oGraph.GenerateImageMask(_m_oDoorImg);

          C_Viewport.InitImagesAndStuff(_m_oGraph, _m_oShipImg, _m_oExplosionImg, _m_oBulletImg, _m_oLandingZoneImg,
              _m_oCpImg, _m_oE0Img, _m_oE1Img, _m_oE2Img, _m_oE3Img, _m_oE4Img, _m_oE5Img, _m_oE6Img, _m_oE7Img, _m_oDoorImg,
              _m_oCargoFuel, _m_oLapsTime, _m_oFont, m_lKeyState, m_lTouch);
          C_Object.s_oGraph = _m_oGraph;


          S_Rect stDrawRect = new S_Rect(0, 0, _m_oGraph.m_iViewportWidth, _m_oGraph.m_iViewportHeight);
          _m_oMap = new C_TileMap( m_szDesFile, _m_oGraph, stDrawRect);

          _m_iState++;
        }

        break;
      case 1:  //game init
        if(_m_oMap.IsLoaded()) {
          _m_oMap.SetMapWrapping(false, false);
          _m_oViewport = _m_oMap.GetExtendedInfo();
          _m_oViewport.InitPlayersAndSetMap(_m_oMap);
          _m_oViewport.SetSize(_m_oGraph.m_iViewportWidth, _m_oGraph.m_iViewportHeight);
          _m_dTimeLeft = 0.0;
          _m_dTimeToEnd = 0.0;
          _m_bEnd = false;

          _m_iState++;
        }

        break;
      case 2:  //game loop
        _m_dTimeLeft += dCurTime;
        if(_m_bEnd) _m_dTimeToEnd += dCurTime;

        if(_m_dTimeToEnd>6500) {
          _m_oViewport.SoundOn(false);
          _m_iState++;
          _m_dTimeToEnd = 0.0;
        }

        TouchUpdate();

        while(_m_dTimeLeft>=4.0) {
          int iRet = _m_oViewport.Update(4.0);
          if(iRet<0) _m_bEnd = true;
          _m_dTimeLeft -= 4.0;
        }
        _m_oViewport.Draw();
        break;
      case 3:
        S_Rect stRect = new S_Rect(0, 0, _m_oGraph.m_iViewportWidth, _m_oGraph.m_iViewportHeight);
        _m_oGraph.DrawRectangle2D(stRect, oColorDG);
        String szScore;
        if(_m_oViewport.IsRace()) {
          szScore = _m_oViewport.GetTimeAsString(_m_oViewport.GetRaceTime().toDouble());
        } else {
          szScore = (_m_oViewport.GetScore()/1000).toString();
        }
        _m_oFont2.PutString(stRect, true, szScore);

        _m_dTimeToEnd += dCurTime;
        if(_m_dTimeToEnd>6500) {
          _m_iState = 0; //restart
        }

        break;
      case 4: //error
//        m_oDBImage.getGraphics().setColor(Color.RED);
//        m_oDBImage.getGraphics().drawString("Error: Bad replay", 5, this.getSize().height-32);
        break;
    }

    _m_oGraph.EndScene();

    // keep looping
    this._renderFrame();
    return true;
  }

  void OnResize(Event e)
  {
    ScreenSize();
  }

  void Start()
  {
    if(!_m_bHasWebGL) {
      window.alert("WebGL not supported by this browser");
      return; //exit program
    }

    for(int i=0; i<256; i++)
      m_lKeyState[i] = false;
    for(int i=0; i<5; i++)
      m_lTouch[i] = false;
    window.onKeyDown.listen(HandleKeyEventDown);
    window.onKeyUp.listen(HandleKeyEventUp);
    m_oCanvas.onTouchStart.listen(HandleTouch);
    m_oCanvas.onTouchMove.listen(HandleTouch);
    m_oCanvas.onTouchEnd.listen(HandleTouch);
    window.onResize.listen(OnResize);

    InputElement e1 = document.getElementById('Level');
    InputElement e2 = document.getElementById('Name');
    InputElement e3 = document.getElementById('debug_localhost');

    String szMap = e1.value;
    String szName = e2.value;
    //use this in the release version:
    String szURL = szMap+".des";
    //use this while debugging locally:
    if(e3!=null && int.parse(e3.value)==1) {
      //szURL = "/race02_4pl.des";
      //szURL = "/mission03.des";
      szURL = "/race00_zero.des";
    }
    m_szDesFile = szURL;

    _loadTextures(); //begin loading all static textures

    _m_oSound.LoadSound("files/bulletfire.wav", 0);
    _m_oSound.LoadSound("files/explosion.wav", 1);
    _m_oSound.LoadSound("files/engine.wav", 2);
    _m_oSound.LoadSound("files/landonzone.wav", 3);
    _m_oSound.LoadSound("files/bullethit.wav", 4);
    _m_oSound.LoadSound("files/cp.wav", 5);
    _m_oSound.LoadSound("files/cp_end.wav", 6);
    _m_oSound.LoadSound("files/bounceonzone.wav", 7);
    _m_oSound.LoadSound("files/loadcargo.wav", 8);
    _m_oSound.LoadSound("files/unloadcargo.wav", 9);

    ScreenSize();

    this._renderFrame();
  }

  void _renderFrame() {
    window.requestAnimationFrame((num time) { this.Loop(time); });
  }

}

void main()
{
  C_GameHandler oGame = new C_GameHandler(querySelector('#webplay-canvas'));
  oGame.Start();
}
