package GalaxyPK;

import java.awt.*;

class C_Vector2D
{
	public C_Vector2D()
	{
	}
	public C_Vector2D(Point p)
	{
		x = p.x;
		y = p.y;
	}
	public C_Vector2D(double i_dX, double i_dY)
	{
		x = i_dX;
		y = i_dY;
	}
	public double x;
	public double y;
}

class C_Vector2DFunc
{

	static void VectorSub(C_Vector2D a, C_Vector2D b, C_Vector2D c)
	{
		c.x = a.x - b.x;
		c.y = a.y - b.y;
	}
	static double VectorMag(C_Vector2D a)
	{
		return Math.sqrt(a.x * a.x + a.y * a.y);
	}
	static double VectorMagSquared(C_Vector2D a)
	{
		return (a.x * a.x + a.y * a.y);
	}
	static double VectorDist(C_Vector2D a, C_Vector2D b)
	{
		double x, y;
		x = b.x - a.x;
		y = b.y - a.y;
		
		return Math.sqrt(x*x + y*y);
	}
	static void VectorNorm(C_Vector2D a)
	{
		double len = Math.sqrt(a.x*a.x + a.y*a.y);
		a.x /= len;
		a.y /= len;
	}
	static double VectorDot(C_Vector2D a, C_Vector2D b)
	{
		return (a.x*b.x + a.y*b.y);
	}
	static int m_iValInRangeClosest;
	static boolean ValInRange(double i_dRVal1, double i_dRVal2, double i_dVal)
	{
		boolean bResult;
		m_iValInRangeClosest = 0;
		
		if(i_dRVal1 > i_dRVal2) {
			bResult = (i_dVal<=i_dRVal1 && i_dVal>=i_dRVal2);
			if(!bResult) {
				if(i_dVal>i_dRVal1) m_iValInRangeClosest = 1;
				else m_iValInRangeClosest = 2;
			}
		} else {
			bResult = (i_dVal<=i_dRVal2 && i_dVal>=i_dRVal1);
			if(!bResult) {
				if(i_dVal>i_dRVal2) m_iValInRangeClosest = 2;
				else m_iValInRangeClosest = 1;
			}
		}
		return bResult;
	}
	
	public static double DistPtToLineSeg(Point ip, Point il0, Point il1)
	{
		C_Vector2D p = new C_Vector2D(ip);
		C_Vector2D l0 = new C_Vector2D(il0);
		C_Vector2D l1 = new C_Vector2D(il1);
		C_Vector2D l0_p = new C_Vector2D();
		C_Vector2D l0_l1 = new C_Vector2D();
		C_Vector2D q = new C_Vector2D();
		double w, dist;
		boolean bInRangeX, bInRangeY;
		int iTestX = 0, iTestY = 0;
		
		VectorSub(p, l0, l0_p);
		VectorSub(l1, l0, l0_l1);
		
		//don't try to take Vector_norm if division by zero will result.
		//check using Vec_mag_squared instead of xpos == 0 && ypos == 0 && zpos == 0 

		if( VectorMagSquared(l0_l1)<1.0f) {
			q.x = l0.x;
			q.y = l0.y;
			
			dist = VectorDist(q, p);
		} else {
			VectorNorm(l0_l1);
			w = VectorDot(l0_p, l0_l1);
			
			q.x = (int)(l0.x + (l0_l1.x * w));
			q.y = (int)(l0.y + (l0_l1.y * w));
			
			bInRangeX = ValInRange(l0.x, l1.x, q.x);
			iTestX = m_iValInRangeClosest;
			bInRangeY = ValInRange(l0.y, l1.y, q.y);
			iTestY = m_iValInRangeClosest;
			if(bInRangeX && bInRangeY) {
				dist = VectorDist(q, p);
			} else {
				if(iTestX==1 || iTestY==1) {
					dist = VectorDist(l0, p);
				} else /*if(iTestX==2 || iTestY==2)*/ {
					dist = VectorDist(l1, p);
				}
			}
		}
		return dist;
	}

}

public class C_CheckPoint extends C_Object
{
	private static final int CP_RANGE = 11;
	private int m_iPosX1, m_iPosY1;
	private int m_iPosX2, m_iPosY2;
	private int m_iCPNum;
	private String m_szString;
	
	private C_TileMap m_oMap;
	private boolean m_bBlinkState;
	private double m_dBlinkTimer;
	private int m_iTextWidth;
	
	private static Font s_oFont;
	public static Image s_oCpImg;


	public C_CheckPoint(C_TileMap i_oMap, Point i_stP1, Point i_stP2, int i_iCPNum)
	{
		super();

		m_iPosX1 = i_stP1.x;
		m_iPosY1 = i_stP1.y;
		m_iPosX2 = i_stP2.x;
		m_iPosY2 = i_stP2.y;
		m_iCPNum = i_iCPNum;
		m_szString = String.valueOf(m_iCPNum);
		
		m_oMap = i_oMap;
		m_stScreenArea = m_oMap.GetScreenRect();
		m_dBlinkTimer = 0.0;
		m_bBlinkState = false;
		
		m_oImage = s_oCpImg;
		m_stImageRect = null;
		
		if(s_oFont==null) {
			s_oFont = new Font("Arial", Font.BOLD, 14);
		}
		//get textwidth
		//FontMetrics fm = getFontMetrics(m_oFont);
		//m_iTextWidth = fm.stringWidth(m_szString);
		m_iTextWidth = m_szString.length()==1 ? 9 : 18; //simplified
	}
	
	public void SetBlinkState(boolean i_bBlinkState)
	{
		if(i_bBlinkState && !m_bBlinkState) m_dBlinkTimer = 0.0;
		m_bBlinkState = i_bBlinkState;
	}
	
	public void Update(double i_dTime)
	{
		m_dBlinkTimer += i_dTime;
	}

	//true player is in range of checkpoint
	public boolean PlayerAtCP(C_Object i_oPlayer)
	{
		Point stCenter = i_oPlayer.GetCenterPos();
		Point stSize = this.GetSize();
		
		//calculate shortest distance from player to checkpoint line
		Point p, l0, l1;
		p = stCenter;
		l0 = new Point(m_iPosX1+(stSize.x/2), m_iPosY1+(stSize.y/2));
		l1 = new Point(m_iPosX2+(stSize.x/2), m_iPosY2+(stSize.y/2));
		double dDist = C_Vector2DFunc.DistPtToLineSeg(p, l0, l1);
		
		if(dDist<CP_RANGE) {
			SetBlinkState(false);
			return true;
		}
		return false;
	}
	
	public void Draw()
	{
		Point p = m_oMap.GetPos();
		int iOffsetX = 0, iOffsetY = 0;
		
		boolean bDrawText = !m_bBlinkState || (m_bBlinkState && ((int)m_dBlinkTimer % 650 < 400) );
		if(bDrawText) {
			s_oGraph.setFont(s_oFont);
			s_oGraph.setColor(Color.WHITE);
			//calc centeroffset
			iOffsetX = (p.x-10)+(m_iTextWidth/2);
			iOffsetY = (p.y-13);
			//^screenposition is (m_dPosX-p.x, m_dPosY-p.y)
		}
		m_dPosX = m_iPosX1;
		m_dPosY = m_iPosY1;
		ODraw(m_oImage, null, p.x, p.y);
		if(bDrawText) s_oGraph.drawString(m_szString, m_iPosX1-iOffsetX, m_iPosY1-iOffsetY);
		m_dPosX = m_iPosX2;
		m_dPosY = m_iPosY2;
		ODraw(m_oImage, null, p.x, p.y);
		if(bDrawText) s_oGraph.drawString(m_szString, m_iPosX2-iOffsetX, m_iPosY2-iOffsetY);
	}
}
