package GalaxyPK;

import java.awt.*;
import java.awt.image.*;
import java.util.*;

class C_MaskAssociation
{
	public Image   oImg;
	public boolean abImgMask[][];

	public C_MaskAssociation(Image i_oImg, boolean i_abImgMask[][])
	{
		oImg = i_oImg;
		abImgMask = i_abImgMask;
	}
}

public class C_Collision
{
	public static int s_iDetectPixelStepX = 1; //increase to make it faster and check less pixels (minimum 1)
	public static int s_iDetectPixelStepY = 1; //increase to make it faster and check less pixels (minimum 1)

	public static final int PIXELCOMPAREMAXWIDTH  = 64;
	public static final int PIXELCOMPAREMAXHEIGHT = 64;
	public static int s_Img1[] = new int[PIXELCOMPAREMAXWIDTH*PIXELCOMPAREMAXHEIGHT];
	public static int s_Img2[] = new int[PIXELCOMPAREMAXWIDTH*PIXELCOMPAREMAXHEIGHT];

	private static ArrayList<C_MaskAssociation> s_oMaskList = null;

	public static boolean Compare(Image i_Img1, int i_iX1, int i_iY1, Image i_Img2, int i_iX2, int i_iY2, Rectangle i_stRect1, Rectangle i_stRect2)
	{
		int i, j;
		Rectangle stR1, stR2;
		Rectangle r1 = new Rectangle(), r2 = new Rectangle(), rtmp;

		if(i_Img1==null && i_stRect1==null) return false;
		if(i_Img2==null && i_stRect2==null) return false;
		
		//get image info
		r1.x = i_iX1; r1.y = i_iY1;
		if(i_stRect1==null) {
			r1.width = i_Img1.getWidth(null);
			r1.height = i_Img1.getHeight(null);
		} else {
			r1.width = i_stRect1.width;
			r1.height = i_stRect1.height;
		}
		r2.x = i_iX2; r2.y = i_iY2;
		if(i_stRect2==null) {
			r2.width = i_Img2.getWidth(null);
			r2.height = i_Img2.getHeight(null);
		} else {
			r2.width = i_stRect2.width;
			r2.height = i_stRect2.height;
		}
		
		//r1, r2 filled now; make intersectrect
		rtmp = r1.intersection(r2);
		if(rtmp.width<=0 || rtmp.height<=0) return false; //no collision
		if(i_Img1==null && i_Img2==null) return true;  //rectangle collision

		//calc stR1, stR2
		stR1 = new Rectangle(rtmp);
		stR2 = new Rectangle(rtmp);
		stR1.x -= i_iX1; stR1.y -= i_iY1;
		stR2.x -= i_iX2; stR2.y -= i_iY2;
		if(i_stRect1!=null) { stR1.x += i_stRect1.x; stR1.y += i_stRect1.y; }
		if(i_stRect2!=null) { stR2.x += i_stRect2.x; stR2.y += i_stRect2.y; }

		PixelGrabber pg; 
		//grab the img1 area to the 1st pixelarray
		try {
			pg = new PixelGrabber(i_Img1, stR1.x,stR1.y,stR1.width,stR1.height, s_Img1,0,PIXELCOMPAREMAXWIDTH);
			pg.grabPixels(); 
			//grab the img2 area to the 2nd pixelarray
			pg = new PixelGrabber(i_Img2, stR2.x,stR2.y,stR2.width,stR2.height, s_Img2,0,PIXELCOMPAREMAXWIDTH);
			pg.grabPixels(); 
		} catch(InterruptedException e) {};

		int iIncY = PIXELCOMPAREMAXWIDTH*s_iDetectPixelStepY;
		int bufpos = 0; //use same position for both buffers
		int iW = stR1.width;
		int iH = stR1.height;
		
		
		if(i_Img1!=null && i_Img2!=null) {
			for(j=0; j<iH; j+=s_iDetectPixelStepY) { //y
				for(i=0; i<iW; i+=s_iDetectPixelStepX) { //x
					int pix1 = s_Img1[bufpos+i] & 0x00ffffff;
					int pix2 = s_Img2[bufpos+i] & 0x00ffffff;
					if(pix1==0 || pix2==0) continue;
					return true; //both pixels were non black
				}
				bufpos += iIncY;
			}
		} else {
			if(i_Img1!=null) {
				for(j=0; j<iH; j+=s_iDetectPixelStepY) { //y
					for(i=0; i<iW; i+=s_iDetectPixelStepX) { //x
						int pix1 = s_Img1[bufpos+i] & 0x00ffffff;
						if(pix1==0) continue;
						return true;
					}
					bufpos += iIncY;
				}
			} else {
				for(j=0; j<iH; j+=s_iDetectPixelStepY) { //y
					for(i=0; i<iW; i+=s_iDetectPixelStepX) { //x
						int pix2 = s_Img2[bufpos+i] & 0x00ffffff;
						if(pix2==0) continue;
						return true;
					}
					bufpos += iIncY;
				}
			}
		}

		return false;
	}

	public static void BufferImage(Image i_oImage)
	{
		int i, j;
		int imgWidth = i_oImage.getWidth(null);
		int imgHeight = i_oImage.getHeight(null);
		int tmpPixels[] = new int[imgHeight*imgWidth];
		boolean abPixelMask[][] = new boolean[imgHeight][imgWidth];
	
		//grab the whole image to the pixelarray
		try {
			PixelGrabber pg = new PixelGrabber(i_oImage, 0,0,imgWidth,imgHeight, tmpPixels,0,imgWidth);
			pg.grabPixels(); 
		} catch(InterruptedException e) {};
	
		for(j=0; j<imgHeight; j+=s_iDetectPixelStepY) { //y
			int iOffset = j*imgWidth;
			for(i=0; i<imgWidth; i+=s_iDetectPixelStepX) { //x
				abPixelMask[j][i] = ((tmpPixels[iOffset+i] & 0x00ffffff)!=0);
			}
		}

		//associate i_oImage with abPixelMask and put to list
		C_MaskAssociation oItem = new C_MaskAssociation(i_oImage, abPixelMask);
		if(s_oMaskList==null) s_oMaskList = new ArrayList<C_MaskAssociation>();
		s_oMaskList.add(oItem);
	}

	public static void UnbufferImage(Image i_oImage)
	{
		if(s_oMaskList==null) return;

		int i, iSize = s_oMaskList.size();
		for(i=0; i<iSize; i++) {
			C_MaskAssociation oItem = (C_MaskAssociation)s_oMaskList.get(i);
			if(oItem.oImg==i_oImage) {
				s_oMaskList.remove(i);
				break;
			}
		}
	}
	
	public static boolean[][] GetMask(Image i_oImage)
	{
		if(s_oMaskList==null) return null;

		int i, iSize = s_oMaskList.size();
		for(i=0; i<iSize; i++) {
			C_MaskAssociation oItem = (C_MaskAssociation)s_oMaskList.get(i);
			if(oItem.oImg==i_oImage) {
				return oItem.abImgMask;
			}
		}
		return null;
	}

	public static boolean CompareBuffered(Image i_Img1, int i_iX1, int i_iY1, Image i_Img2, int i_iX2, int i_iY2, Rectangle i_stRect1, Rectangle i_stRect2)
	{
		boolean abPixelMask1[][] = null;
		boolean abPixelMask2[][] = null;
		boolean bFound1 = false, bFound2 = false;
		if(i_Img1==null) bFound1 = true;
		if(i_Img2==null) bFound2 = true;

		//lookup maskbuffer of image
		int i, iSize = s_oMaskList.size();
		for(i=0; i<iSize; i++) {
			C_MaskAssociation oItem = (C_MaskAssociation)s_oMaskList.get(i);
			if(!bFound1 && oItem.oImg==i_Img1) {
				abPixelMask1 = oItem.abImgMask;
				bFound1 = true;
			}
			if(!bFound2 && oItem.oImg==i_Img2) {
				abPixelMask2 = oItem.abImgMask;
				bFound2 = true;
			}

			if(bFound1 && bFound2) break;
		}
		if(!bFound1 || !bFound2) return false;
		
		return Compare(abPixelMask1, i_iX1, i_iY1, abPixelMask2, i_iX2, i_iY2, i_stRect1, i_stRect2);
	}

	private static boolean Compare(boolean i_Img1[][], int i_iX1, int i_iY1, boolean i_Img2[][], int i_iX2, int i_iY2, Rectangle i_stRect1, Rectangle i_stRect2)
	{
		int i, j;
		Rectangle stR1, stR2;
		Rectangle r1 = new Rectangle(), r2 = new Rectangle(), rtmp;
		
		if(i_Img1==null && i_stRect1==null) return false;
		if(i_Img2==null && i_stRect2==null) return false;
		
		//'image' dimensions
		int img1Width = 0, img2Width = 0, img2Height = 0;
		if(i_Img1!=null) {
			img1Width = i_Img1[0].length;
			//img1Height = i_Img1.length;
		}
		if(i_Img2!=null) {
			img2Width = i_Img2[0].length;
			img2Height = i_Img2.length;
		}

		//get image info
		r1.x = i_iX1; r1.y = i_iY1;
		if(i_stRect1==null) {
			r1.width = img1Width;
			r1.height = img2Height;
		} else {
			r1.width = i_stRect1.width;
			r1.height = i_stRect1.height;
		}
		r2.x = i_iX2; r2.y = i_iY2;
		if(i_stRect2==null) {
			r2.width = img2Width;
			r2.height = img2Height;
		} else {
			r2.width = i_stRect2.width;
			r2.height = i_stRect2.height;
		}
		
		//r1, r2 filled now; make intersectrect
		rtmp = r1.intersection(r2);
		if(rtmp.width<=0 || rtmp.height<=0) return false; //no collision
		if(i_Img1==null && i_Img2==null) return true;  //rectangle collision

		//calc stR1, stR2
		stR1 = new Rectangle(rtmp);
		stR2 = new Rectangle(rtmp);
		stR1.x -= i_iX1; stR1.y -= i_iY1;
		stR2.x -= i_iX2; stR2.y -= i_iY2;
		if(i_stRect1!=null) { stR1.x += i_stRect1.x; stR1.y += i_stRect1.y; }
		if(i_stRect2!=null) { stR2.x += i_stRect2.x; stR2.y += i_stRect2.y; }

		int iPosX1 = stR1.x; int iPosX2 = stR2.x;
		int iPosY1 = stR1.y; int iPosY2 = stR2.y;
		int iW = stR1.width; int iH = stR1.height;
		if(i_Img1!=null && i_Img2!=null) {
			for(j=0; j<iH; j+=s_iDetectPixelStepY) { //y
				for(i=0; i<iW; i+=s_iDetectPixelStepX) { //x
					if(!i_Img1[iPosY1][iPosX1+i] || !i_Img2[iPosY2][iPosX2+i]) continue;
					return true; //both pixels were non black
				}
				iPosY1+=s_iDetectPixelStepY; iPosY2+=s_iDetectPixelStepY;
			}
		} else {
			if(i_Img1!=null) {
				for(j=0; j<iH; j+=s_iDetectPixelStepY) { //y
					for(i=0; i<iW; i+=s_iDetectPixelStepX) { //x
						if(!i_Img1[iPosY1][iPosX1+i]) continue;
						return true;
					}
					iPosY1+=s_iDetectPixelStepY;
				}
			} else {
				for(j=0; j<iH; j+=s_iDetectPixelStepY) { //y
					for(i=0; i<iW; i+=s_iDetectPixelStepX) { //x
						if(!i_Img2[iPosY2][iPosX2+i]) continue;
						return true;
					}
					iPosY2+=s_iDetectPixelStepY;
				}
			}
		}

		return false;
	}

}
