package GalaxyPK;

import java.awt.*;

class S_EnemyInfo
{
	int iEnemyType;
	int iAngle;
	int iFireInterval;
	int iSpeed;
	Point astWayPoints[];
}

public class C_ComputerEnemy extends C_PlayerBase
{
	//static
	//enemy properties
	private static final int ENEMY_HITSTOKILL[] = {4, 5, -1, -1, 8, 4, 4, 6}; //-1=immortal
	private static final int ENEMY_SCOREPOINTS[] = {10, 20, 0, 0, 20, 15, 25, 25};
	private static final int ENEMY_NUMBULLETS[] = {1, 3, 1, 3, 1, 2, 5, 8};
	private static final int ENEMY_BULLETSPEED[] = {190, 160, 190, 160, 110, 90, 120, 140};
	private static final int ENEMY_BULLETANGLE[][] = {
		{0,0,0,0,0,0,0,0}, {-50,0,50,0,0,0,0,0}, {0,0,0,0,0,0,0,0}, {-50,0,50,0,0,0,0,0},
		{0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0}, {-70,-35,0,35,70,0,0,0}, {-180,-135,-90,-45,0,45,90,135}};
	private static final int ENEMY_NUMFRAMES[] = {1, 1, 1, 1, 12, 1, 1, 12};
	
	//unanimated enemies (rectangles are 4 directions (0, 90, 180, 270))
	private static final Rectangle ENEMY6_RECT[] = {new Rectangle(0,14,14,14),new Rectangle(14,0,14,14),new Rectangle(14,14,14,14),new Rectangle(0,0,14,14)};
	private static final Point ENEMY6_FIREPOINT[] = {new Point(13,6),new Point(6,13),new Point(0,6),new Point(6,0)};
	//animated enemies, 1 direction (rectangles are each frame)
	private static final Rectangle ENEMY4_RECT[] = {
		new Rectangle(0,0,64,32),new Rectangle(64,0,64,32),new Rectangle(128,0,64,32),new Rectangle(192,0,64,32),
		new Rectangle(0,32,64,32),new Rectangle(64,32,64,32),new Rectangle(128,32,64,32),new Rectangle(192,32,64,32),
		new Rectangle(0,64,64,32),new Rectangle(64,64,64,32),new Rectangle(128,64,64,32),new Rectangle(192,64,64,32)};
	private static final Point ENEMY4_FIREPOINT = new Point(32,16);
	private static final Rectangle ENEMY7_RECT[] = {
		new Rectangle(0,0,64,64),new Rectangle(64,0,64,64),new Rectangle(128,0,64,64),new Rectangle(192,0,64,64),
		new Rectangle(0,64,64,64),new Rectangle(64,64,64,64),new Rectangle(128,64,64,64),new Rectangle(192,64,64,64),
		new Rectangle(0,128,64,64),new Rectangle(64,128,64,64),new Rectangle(128,128,64,64),new Rectangle(192,128,64,64)};
	private static final Point ENEMY7_FIREPOINT = new Point(32,32);
	//enemies with animated fireing, 4 directions
	private static final Rectangle ENEMY0_RECT[] = {
		new Rectangle(0,16,16,12),new Rectangle(24+0,16,16,12), new Rectangle(12,0,12,16),new Rectangle(24+12,0,12,16),
		new Rectangle(0,28,16,12),new Rectangle(24+0,28,16,12), new Rectangle(0,0,12,16),new Rectangle(24+0,0,12,16)};
	private static final Point ENEMY0_FIREPOINT[] = {new Point(16,5), new Point(5,16), new Point(0,5), new Point(5,0)};
	private static final Rectangle ENEMY2_RECT[] = {
		new Rectangle(0,16,16,12),new Rectangle(24+0,16,16,12), new Rectangle(12,0,12,16),new Rectangle(24+12,0,12,16),
		new Rectangle(0,28,16,12),new Rectangle(24+0,28,16,12), new Rectangle(0,0,12,16),new Rectangle(24+0,0,12,16)};
	private static final Point ENEMY2_FIREPOINT[] = {new Point(16,5), new Point(5,16), new Point(0,5), new Point(5,0)};
	private static final Rectangle ENEMY1_RECT[] = {
		new Rectangle(14,14,14,24),new Rectangle(28+14,14,14,24), new Rectangle(0,38,24,14),new Rectangle(28+0,38,24,14),
		new Rectangle(0,14,14,24),new Rectangle(28+0,14,14,24), new Rectangle(0,0,24,14),new Rectangle(28+0,0,24,14)};
	private static final Point ENEMY1_FIREPOINT[] = {new Point(14,12), new Point(12,14), new Point(0,12), new Point(12,0)};
	private static final Rectangle ENEMY3_RECT[] = {
		new Rectangle(14,14,14,24),new Rectangle(28+14,14,14,24), new Rectangle(0,38,24,14),new Rectangle(28+0,38,24,14),
		new Rectangle(0,14,14,24),new Rectangle(28+0,14,14,24), new Rectangle(0,0,24,14),new Rectangle(28+0,0,24,14)};
	private static final Point ENEMY3_FIREPOINT[] = {new Point(14,12), new Point(12,14), new Point(0,12), new Point(12,0)};
	private static final Rectangle ENEMY5_RECT[] = {
		new Rectangle(0,26,13,22),new Rectangle(0,48,13,22), new Rectangle(0,13,22,13),new Rectangle(22,13,22,13),
		new Rectangle(13,26,13,22),new Rectangle(13,48,13,22), new Rectangle(0,0,22,13),new Rectangle(22,0,22,13)};
	private static final Point ENEMY5_FIREPOINT[] = { new Point(13,4),new Point(13,18), new Point(4,13),new Point(18,13), new Point(0,4),new Point(0,18), new Point(4,0),new Point(18,0)};

	//common
	protected C_TileMap    m_oMap;
	protected C_Replay     m_oReplay;
	protected C_BulletList m_oBulletList;

	//movement
	protected int m_iNumWayPoints, m_iCurWP, m_iSpeed;
	protected Point m_astWayPointList[];
	protected double m_dWayTime;
	
	//animation
	protected double m_dFrameNr;
	protected double m_dFireAnimTime = 0.0;
	protected Rectangle m_astAnimRects[];
	
	//enemy specific
	protected int m_iEnemyType;
	protected int m_iHitsLeft;      //hits enemy can take before it gets killed
	protected int m_iFireInterval;
	protected double m_dFireTime;
	protected Point m_stFirePoint;
	protected boolean m_bAlive, m_bFinished; //explosion ready after enemy killed

	//images
	private static Image s_oEnemy0Img;
	private static Image s_oEnemy1Img;
	private static Image s_oEnemy2Img;
	private static Image s_oEnemy3Img;
	private static Image s_oEnemy4Img;
	private static Image s_oEnemy5Img;
	private static Image s_oEnemy6Img;
	private static Image s_oEnemy7Img;
	public static void InitImages(Image i_oEnemy0Img, Image i_oEnemy1Img, Image i_oEnemy2Img, Image i_oEnemy3Img, Image i_oEnemy4Img, Image i_oEnemy5Img, Image i_oEnemy6Img, Image i_oEnemy7Img)
	{
		s_oEnemy0Img = i_oEnemy0Img;
		s_oEnemy1Img = i_oEnemy1Img;
		s_oEnemy2Img = i_oEnemy2Img;
		s_oEnemy3Img = i_oEnemy3Img;
		s_oEnemy4Img = i_oEnemy4Img;
		s_oEnemy5Img = i_oEnemy5Img;
		s_oEnemy6Img = i_oEnemy6Img;
		s_oEnemy7Img = i_oEnemy7Img;
	}
	
	public C_ComputerEnemy(C_TileMap i_oMap, C_Replay i_oReplay, C_BulletList i_oBulletList, S_EnemyInfo i_stEnemyInfo, int i_iID)
	{
		super();
		
		m_iOwnerID = i_iID;
		m_oMap = i_oMap;
		m_oReplay = i_oReplay;
		m_oBulletList = i_oBulletList;
		
		m_iEnemyType = i_stEnemyInfo.iEnemyType;
		m_dDirection = i_stEnemyInfo.iAngle;
		m_iSpeed = i_stEnemyInfo.iSpeed;
		m_iNumWayPoints = i_stEnemyInfo.astWayPoints.length;
		m_astWayPointList = i_stEnemyInfo.astWayPoints;
		m_dPosX = m_astWayPointList[0].x;
		m_dPosY = m_astWayPointList[0].y;
		m_dVelX = m_dVelY = m_dAcceleration = 0.0;
		m_iCurWP = 1; //after 0
		m_dWayTime = 0.0;
		m_dFrameNr  = 0.0;
		m_dFireAnimTime = 0.0;
		
		m_stScreenArea = m_oMap.GetScreenRect();

		m_bAlive    = true;
		m_bFinished = false;
		m_iHitsLeft = ENEMY_HITSTOKILL[m_iEnemyType];
		switch(m_iEnemyType) {
			case 0:
				m_oImage = s_oEnemy0Img;
				m_astAnimRects = ENEMY0_RECT;
				m_stImageRect = ENEMY0_RECT[i_stEnemyInfo.iAngle/90];
				m_stFirePoint = ENEMY0_FIREPOINT[i_stEnemyInfo.iAngle/90];
			break;
			case 1:
				m_oImage = s_oEnemy1Img;
				m_astAnimRects = ENEMY1_RECT;
				m_stImageRect = ENEMY1_RECT[i_stEnemyInfo.iAngle/90];
				m_stFirePoint = ENEMY1_FIREPOINT[i_stEnemyInfo.iAngle/90];
			break;
			case 2:
				m_oImage = s_oEnemy2Img;
				m_astAnimRects = ENEMY2_RECT;
				m_stImageRect = ENEMY2_RECT[i_stEnemyInfo.iAngle/90];
				m_stFirePoint = ENEMY2_FIREPOINT[i_stEnemyInfo.iAngle/90];
			break;
			case 3:
				m_oImage = s_oEnemy3Img;
				m_astAnimRects = ENEMY3_RECT;
				m_stImageRect = ENEMY3_RECT[i_stEnemyInfo.iAngle/90];
				m_stFirePoint = ENEMY3_FIREPOINT[i_stEnemyInfo.iAngle/90];
			break;
			case 4:
				m_oImage = s_oEnemy4Img;
				m_astAnimRects = ENEMY4_RECT;
				m_stImageRect = ENEMY4_RECT[0];
				m_stFirePoint = ENEMY4_FIREPOINT;
				break;
			case 5:
				m_oImage = s_oEnemy5Img;
				m_astAnimRects = ENEMY5_RECT;
				m_stImageRect = ENEMY5_RECT[i_stEnemyInfo.iAngle/90];
				m_stFirePoint = ENEMY5_FIREPOINT[i_stEnemyInfo.iAngle/90];
				break;
			case 6:
				m_oImage = s_oEnemy6Img;
				m_astAnimRects = ENEMY6_RECT;
				m_stImageRect = ENEMY6_RECT[i_stEnemyInfo.iAngle/90];
				m_stFirePoint = ENEMY6_FIREPOINT[i_stEnemyInfo.iAngle/90];
				break;
			case 7:
				m_oImage = s_oEnemy7Img;
				m_astAnimRects = ENEMY7_RECT;
				m_stImageRect = ENEMY7_RECT[0];
				m_stFirePoint = ENEMY7_FIREPOINT;
				break;
		}

	}

	public int Update(double i_dTime)
	{
		int iResult = -1;
		int i, iPlaySound = -1;
		C_Bullet oBullet;
		boolean bFireBullet = false;
		
		if(!m_bAlive) m_bFinished = UpdateExplosion(i_dTime);
		if(m_bFinished) return -2;

		S_NetMsg1 stMsg1;
		S_ReplayAction stAction;
		while( (stAction = m_oReplay.Get(m_iOwnerID, 1))!=null ) {
			stMsg1 = stAction.stMsg1;
			//movement for moving enemies
			if((stMsg1.iMsg & S_NetMsg1.N_MSG1_MOVEMENT)!=0) GetCommonInfo(stMsg1);

			//bullet(s) fired
			if((stMsg1.iMsg & S_NetMsg1.N_MSG1_NEWBULLET)!=0) {
				iPlaySound = 0; //bullet sound
				int iBulletBase = stMsg1.iBulletUniqeValBase >> 8;
				for(i=0; i<ENEMY_NUMBULLETS[m_iEnemyType]; i++) {
					if(m_iEnemyType==5) m_stFirePoint = ENEMY5_FIREPOINT[2*(int)(m_dDirection/90)+i];
					oBullet = CreateBullet(i);
					int iUId = (stMsg1.iBulletUniqeValBase & 0xff) | ((iBulletBase+i) << 8);
					oBullet.SetUId(iUId);
					m_oBulletList.Add(oBullet);
				}
				bFireBullet = true;
			}

			//is hit
			if((stMsg1.iMsg & S_NetMsg1.N_MSG1_KILLED)!=0) {
				m_oBulletList.Remove(stMsg1.iCauseID);
				iPlaySound = 4; //hit sound
				m_iHitsLeft = stMsg1.iReason_Hits_Door;

				if(m_iHitsLeft <= 0 && m_iHitsLeft!=-1) {
					m_bAlive = false;
					m_dVelX = m_dVelY = 0; //make explosion be still
					m_bCanCollide = false;
					iPlaySound = 1; //explode sound
					iResult = stMsg1.iCauseID & 255;
				}
			}
		}
		Move(i_dTime);

		//animated?
		if(m_iEnemyType==4 || m_iEnemyType==7) {
			m_dFrameNr += i_dTime*0.021; //21 fps anim
			int iFrameNr = (int)m_dFrameNr;
			if(iFrameNr >= ENEMY_NUMFRAMES[m_iEnemyType]) {
				m_dFrameNr = 0.0; //restart anim
				iFrameNr = 0;
			}
			m_stImageRect = m_astAnimRects[iFrameNr];
		}
	
		//animated fireing
		if(m_iEnemyType==5 || m_iEnemyType==0 || m_iEnemyType==2 || m_iEnemyType==1 || m_iEnemyType==3) {
			if(bFireBullet) m_dFireAnimTime = 250.0; //250 ms
			m_dFireAnimTime -= i_dTime;
			int iRect = 2*(int)(m_dDirection/90);
			m_stImageRect = m_astAnimRects[iRect];
			if(m_dFireAnimTime<=0) m_dFireAnimTime = 0;
			else m_stImageRect = m_astAnimRects[iRect+1];
		}
		
		if(iPlaySound != -1) {
			CalcSoundVolPan(s_oListner);
			m_oSound.PlaySound(iPlaySound, m_dVol, m_dPan, false);
		}
	
	   return iResult;
	}

	void Move(double i_dTime)
	{
		if(m_iNumWayPoints>1) { //move enemy
			OUpdate(i_dTime);
		}
	}

	public void Draw()
	{
		Point p = m_oMap.GetPos();
		
		if(m_bAlive) ODraw(m_oImage, m_stImageRect, p.x, p.y);
		else DrawExplosion(p.x, p.y);
	}

	C_Bullet CreateBullet(int i_iIndex)
	{
		double dAngle = m_dDirection+ENEMY_BULLETANGLE[m_iEnemyType][i_iIndex];
		double dTemp = dAngle*(Math.PI/180);
		double dSin = Math.sin(dTemp);
		double dCos = Math.cos(dTemp);
		double dPosX = m_dPosX + m_stFirePoint.x;
		double dPosY = m_dPosY + m_stFirePoint.y;
		double dVelX = dCos * ENEMY_BULLETSPEED[m_iEnemyType];
		double dVelY = dSin * ENEMY_BULLETSPEED[m_iEnemyType];
		double dDirection = dAngle;
		
		return new C_Bullet(m_oMap, dPosX, dPosY, dVelX, dVelY, dDirection, m_iOwnerID);
	}

	int GetScoreValue()
	{
		return ENEMY_SCOREPOINTS[m_iEnemyType];
	}

}
