package GalaxyPK;


public class C_KeyHandler
{
	private static final int BUFFERSIZE = 1100;
	private boolean m_abKeyStatus[]; //main buffer

	private boolean m_abPressed[];   //keep track of if the keypress has been reported
	private int     m_aiRepeatCount[];
	private double  m_adRepeat[];
	private int     m_iRepeatRate, m_iRepeatDelay;
	
	public C_KeyHandler()
	{
		m_abKeyStatus   = new boolean[BUFFERSIZE];
		m_abPressed     = new boolean[BUFFERSIZE];
		m_aiRepeatCount = new int[BUFFERSIZE];
		m_adRepeat      = new double[BUFFERSIZE];
		
		m_iRepeatRate  = 45;
		m_iRepeatDelay = 150;
	}
	
	//must be called for each keydown/keyup event
	public void UpdateKeyStatus(int i_iKey, boolean i_bDown)
	{
		if(i_iKey >= BUFFERSIZE) return; //odd keys might give 65536 (like the NUMPAD5 if numlock is off and ctrl+alt+5 and so on)
		m_abKeyStatus[i_iKey] = i_bDown;
	}

	//must be called continuously to get the keyrepeat to work
	public void UpdateTimer(double i_dTime)
	{
		for(int i=0; i<BUFFERSIZE; i++) {
			if( m_abKeyStatus[i] ) {
				if(m_adRepeat[i]!=0.0) m_adRepeat[i] += i_dTime;
				else {
					m_aiRepeatCount[i] = 0;
					m_adRepeat[i] = m_iRepeatRate+0.1;
				}
			} else {
				m_adRepeat[i] = 0;
				m_abPressed[i] = false;
			}
		}
	}

	public void SetRate(int i_iMsRate, int i_iMsDelay)
	{
		m_iRepeatRate = i_iMsRate;
		m_iRepeatDelay = i_iMsDelay;
	}

	public boolean IsPressed(int i_iKey)
	{
		m_abPressed[i_iKey] = m_abKeyStatus[i_iKey];
		return m_abPressed[i_iKey];
	}

	public boolean IsPressedOnce(int i_iKey)
	{
		if(!m_abPressed[i_iKey] && m_abKeyStatus[i_iKey]) {
			m_abPressed[i_iKey] = true;
			return true;
		}
		return false;
	}

	public boolean IsPressedRepeat(int i_iKey)
	{
		int iRepeatVal = m_iRepeatRate;
		if(m_aiRepeatCount[i_iKey] == 1) iRepeatVal = m_iRepeatDelay;
		
		if(m_adRepeat[i_iKey] > iRepeatVal) {
			m_abPressed[i_iKey] = true;
			m_aiRepeatCount[i_iKey]++;
			m_adRepeat[i_iKey] = 0.1;
			return true;
		}
		return false;
	}

}
