package GalaxyPK;

import java.awt.*;

public class C_LandingZone extends C_Object
{
	//image
	private static Image s_oLandingZoneImg;
	private static final Rectangle s_oLandingRect = new Rectangle(0,44,96,4);
	private static final Rectangle s_oHouseRect = new Rectangle(70,23,24,21);
	private static final Rectangle s_oAntennaRects[] = {new Rectangle(0,15,4,29),new Rectangle(4,15,4,29),new Rectangle(8,15,4,29),new Rectangle(12,15,4,29),new Rectangle(16,15,4,29)};
	private static final Rectangle s_oCargoRects[] = {new Rectangle(54,38,14,6),new Rectangle(54,32,14,6),new Rectangle(54,26,14,6),new Rectangle(54,20,14,6)};
	private static final Rectangle s_oExtraLifeRect = new Rectangle(20,25,14,19);
	public static void InitImage(Image i_oLandingZoneImg)
	{
		s_oLandingZoneImg = i_oLandingZoneImg;
	}
	
	//other
	private C_TileMap m_oMap;
	private boolean m_bShowAntenna, m_bShowHouse;
	private double m_dAntennaFrame;
	private int m_aiCargo[]; //array of cargo weights (small = 1..5,6..10,11..15,16..20+ = huge)
	private boolean m_bExtraLife;
	private int m_iZoneSize;
	
	private static final Point LAND_BPOS = new Point(10,42);

	public C_LandingZone(C_TileMap i_oMap, int i_iZoneID, int i_iX, int i_iY, int i_iZoneSize, boolean i_bShowAntenna, boolean i_bShowHouse, int i_aiCargo[], boolean i_bExtraLife)
	{
		super();
		m_dPosX = i_iX;
		m_dPosY = i_iY;
		m_dVelX = m_dVelY = m_dAcceleration = 0.0;
		m_dDirection = 270.0;
		m_iZoneSize = i_iZoneSize;
		m_iOwnerID = i_iZoneID;
		
		m_oMap = i_oMap;
		m_stScreenArea = m_oMap.GetScreenRect();
		
		m_oImage = s_oLandingZoneImg;
		m_stImageRect = s_oLandingRect;
		
		m_bShowAntenna = i_bShowAntenna;
		m_bShowHouse = i_bShowHouse;
		m_aiCargo = i_aiCargo;
		m_dAntennaFrame = 0.0;
		
		m_bExtraLife = i_bExtraLife;
	}

	public void Update(double i_dTime)
	{
		m_dAntennaFrame += i_dTime*0.0044; //4.4 fps anim
		if(m_dAntennaFrame>=s_oAntennaRects.length) m_dAntennaFrame = 0.0;
	}
	
	public void UpdateFuel(C_Player i_oPlayer, double i_dTime)
	{
		//refuel if homebase
		if(!m_bShowHouse) return;
		Rectangle stRect = new Rectangle((int)m_dPosX, (int)m_dPosY, 32*m_iZoneSize, s_oLandingRect.height+20);
		double x = i_oPlayer.m_dPosX+LAND_BPOS.x;
		double y = i_oPlayer.m_dPosY+LAND_BPOS.y;
		boolean bPointInRect = (((x) >= stRect.x) && ((x) <= stRect.x+stRect.width) && ((y) >= stRect.y) && ((y) <= stRect.y+stRect.height));
		if( i_oPlayer.m_iRealDir==27 && bPointInRect)
		{
			i_oPlayer.AddFuel((i_dTime*0.01)*1.0); //refuel 10 times faster than fuelburn
		}
	}

	public void Draw()
	{
		Point p = m_oMap.GetPos();
		Rectangle stRect = new Rectangle(m_stImageRect);
		stRect.width = 32;
		ODraw(m_oImage, stRect, p.x, p.y); //left part
		stRect.x += 32;
		for(int i=0; i<m_iZoneSize-2; i++) {
			ODraw(m_oImage, stRect, p.x-32-i*32, p.y); //centre parts
		}
		stRect.x += 32;
		ODraw(m_oImage, stRect, p.x-(m_iZoneSize-1)*32, p.y); //right part
		int iAdjustX = -(m_iZoneSize-3)*32;
		
		if(m_bShowHouse) ODraw(m_oImage, s_oHouseRect, p.x-70+iAdjustX, p.y+21);
		if(m_bShowAntenna) ODraw(m_oImage, s_oAntennaRects[(int)m_dAntennaFrame], p.x, p.y+29);
		if(m_aiCargo!=null) {
			int i, iSize = m_aiCargo.length;
			int iBaseX = m_bShowHouse ? -50+iAdjustX : -78+iAdjustX;
			for(i=0; i<iSize; i++) {
				int iCargoType = 3-(m_aiCargo[i]-1)/5; //1..5=small container, 6..10, 11..15, 16..20+ = large container
				if(iCargoType<0) iCargoType=0;
				//if(iCargoType>3) iCargoType=3;
				ODraw(m_oImage, s_oCargoRects[iCargoType], p.x+iBaseX+((i/3)*14), p.y+6+((i%3)*6));
			}
		}
		if(m_bExtraLife) {
			int iBaseX = m_bShowAntenna ? -8 : -4;
			ODraw(m_oImage, s_oExtraLifeRect, p.x+iBaseX, p.y+s_oExtraLifeRect.height);
		}
	}
	
	void TakeExtraLife()
	{
		m_bExtraLife = false;
	}
	
	int PopCargo()
	{
		int iCargo = -1;
		int i, iPos = m_aiCargo.length-1;
		if(iPos>=0) {
			iCargo = m_aiCargo[iPos];
			int aiNewCargo[] = new int[iPos]; 
			for(i=0; i<iPos; i++) aiNewCargo[i] = m_aiCargo[i];
			m_aiCargo = aiNewCargo;
		}
		return iCargo;
	}

	void PushCargo(int i_iWeight)
	{
		int i, iPos = m_aiCargo.length;
		int aiNewCargo[] = new int[iPos+1]; 
		for(i=0; i<iPos; i++) aiNewCargo[i] = m_aiCargo[i];
		aiNewCargo[iPos] = i_iWeight;
		m_aiCargo = aiNewCargo;
	}

	public int GetTotalCargo()
	{
		int iCargo = 0;
		if(m_aiCargo!=null) {
			int i, iSize = m_aiCargo.length;
			for(i=0; i<iSize; i++) iCargo+=m_aiCargo[i];
		}
		return iCargo;
	}
}
