package GalaxyPK;

import java.awt.*;
import java.awt.event.KeyEvent;
import java.awt.image.*;
import java.util.*;

import SoundPK.*;

public class C_MapObjects
{
	private static C_KeyHandler s_oKeyHandler;
	private static BufferedImage s_oShipImg[];
	private static C_Sound s_oSound;

	public int m_aiPlayerStartPosX[] = new int[8], m_aiPlayerStartPosY[] = new int[8];
	public ArrayList<C_LandingZone> m_oLandingZList = new ArrayList<C_LandingZone>();
	public ArrayList<C_CheckPoint> m_oCheckPList = new ArrayList<C_CheckPoint>();
	public ArrayList<C_ComputerEnemy> m_oCompEnemyList = new ArrayList<C_ComputerEnemy>();
	public ArrayList<C_Door> m_oDoorList = new ArrayList<C_Door>();
	public C_BulletList m_oBulletList = new C_BulletList();
	public C_Player     m_oPlayers[];
	public C_TileMap    m_oMap;
	public C_Replay     m_oReplay;
	public int          m_iNumLapsToRace;
	
	private int m_iReplayFollowPlayer;
	private int m_iReplayOldCP;
	
	private static Font s_oFont;
	private static Graphics s_oGraph;
	
	private static Image m_oCargoFuel;
	private static Image m_oLapsTime;
	
	private static final int MAXSPACEINHOLDUNITS = 3;
	private static final int MAXSPACEINHOLDWEIGHT = 50;
	
	private boolean m_bTimeCounting[] = new boolean[8];
	private double m_dTotalTime[] = new double[8];
	private double m_dTotalTimePlayed;
	
	private static BufferedImage BlendColor(BufferedImage oSrc, Color i_stColor, int x, int y, int w, int h)
	{
		ColorModel oCM = oSrc.getColorModel();
		BufferedImage oDst = new BufferedImage(oCM, oSrc.copyData(null), oCM.isAlphaPremultiplied(), null); 

		float sa, sr, sg, sb;
		for(int i=x; i<w; i++) {
			for(int j=y; j<h; j++) {
				int iRGB = oDst.getRGB(i, j);
				sa = ((iRGB >> 24) & 0xff);
				if(sa==0) continue;
				sr = ((iRGB >> 16) & 0xff);
				sg = ((iRGB >> 8) & 0xff);
				sb = (iRGB & 0xff);
				//blend color
				sr = (sr * i_stColor.getRed())   / 255.0f;
				sg = (sg * i_stColor.getGreen()) / 255.0f;
				sb = (sb * i_stColor.getBlue())  / 255.0f;
				Color c = new Color((int)sr, (int)sg, (int)sb, (int)sa);
				oDst.setRGB(i, j, c.getRGB());
			}
		}
		return oDst;
	}
	
	public static void InitImagesAndStuff(Graphics i_oBackBuffer, C_KeyHandler i_oKeyHandler, Image i_oShipImg, Image i_oExplosionImg, Image i_oBulletImg, Image i_oLandingZoneImg, Image i_oCpImg,
			Image i_oE0, Image i_oE1, Image i_oE2, Image i_oE3, Image i_oE4, Image i_oE5, Image i_oE6, Image i_oE7, Image i_oDoorImg, Image i_oCargoFuel, Image i_oLapsTime)
	{
		s_oSound = C_Sound.GetSoundObject();
		s_oKeyHandler = i_oKeyHandler;
		s_oGraph = C_Object.s_oGraph = i_oBackBuffer;
		s_oShipImg = new BufferedImage[8];
		BufferedImage oSrc = (BufferedImage)i_oShipImg;
		for(int i=0; i<8; i++) {
			s_oShipImg[i] = BlendColor(oSrc, STATUSCOLOR[i], 0, 0, oSrc.getWidth(), oSrc.getHeight()/2); 
		}
		C_Player.InitExplosionImage(i_oExplosionImg);
		C_Bullet.s_oBulletImg = i_oBulletImg;
		C_LandingZone.InitImage(i_oLandingZoneImg);
		C_ComputerEnemy.InitImages(i_oE0, i_oE1, i_oE2, i_oE3, i_oE4, i_oE5, i_oE6, i_oE7);
		C_CheckPoint.s_oCpImg = i_oCpImg;
		C_Door.InitImages(i_oDoorImg);
		
		s_oFont = new Font("Arial", Font.BOLD, 12);
		
		m_oCargoFuel = i_oCargoFuel;
		m_oLapsTime = i_oLapsTime;
	}
	
	public void InitPlayersAndSetMap(C_TileMap i_oMap)
	{
		m_oMap = i_oMap;
		m_oPlayers = new C_Player[8];
		S_ReplayHeader stInfo = m_oReplay.GetGameInfo();
		for(int i=0; i<8; i++) {
			m_bTimeCounting[i] = false;
			m_dTotalTime[i] = 0.0;
			if(stInfo.szPlayerName[i].length()>0) {
				m_oPlayers[i] = new C_Player(i, stInfo.szPlayerName[i], s_oShipImg[i], m_oReplay, m_oMap, m_oBulletList, m_aiPlayerStartPosX[i], m_aiPlayerStartPosY[i], stInfo.iLevelType==3 ? -1 : 5);
			} else m_oPlayers[i] = null; 
		}
		m_iReplayFollowPlayer = 0;
		m_iReplayOldCP = 0;
		ReplayFollow(false);
		
		m_dTotalTimePlayed = 0.0;
	}
	
	void ReplayFollow(boolean i_bNext)
	{
		int i, iSize;
		int iOldPlayer = m_iReplayFollowPlayer;
		if(!i_bNext && m_oPlayers[m_iReplayFollowPlayer]==null) i_bNext = true;
		if(i_bNext) {
			for(i=0; i<8; i++) {
				m_iReplayFollowPlayer++;
				if(m_iReplayFollowPlayer>7) m_iReplayFollowPlayer=0;
				if(m_oPlayers[m_iReplayFollowPlayer]!=null) break;
			}
		}

		C_Player oPlayer = m_oPlayers[m_iReplayFollowPlayer];
		C_PlayerBase.SetListner((C_Object)oPlayer);
		if(oPlayer!=null) {
			Point stPos = oPlayer.GetCenterPos();
			Rectangle stRect = m_oMap.GetScreenRect();
			int iPosX = (int)stPos.x - (stRect.width/2);
			int iPosY = (int)stPos.y - (stRect.height/2);
			m_oMap.MoveAbsolute(iPosX, iPosY);
			
			iSize = m_oCheckPList.size();
			if(iSize>0) {
				for(i=0; i<iSize; i++) {
					if(i!=oPlayer.m_iCurCP) ((C_CheckPoint)m_oCheckPList.get(i)).SetBlinkState(false);
				}
				if(oPlayer.m_iCurLap < m_iNumLapsToRace)
					((C_CheckPoint)m_oCheckPList.get(oPlayer.m_iCurCP)).SetBlinkState(true);
				if(iOldPlayer == m_iReplayFollowPlayer) {
					if(m_iReplayOldCP!=oPlayer.m_iCurCP) {
						if(oPlayer.m_iCurLap < m_iNumLapsToRace) s_oSound.PlaySound(5);
						else s_oSound.PlaySound(6);
					}
				}
				m_iReplayOldCP = oPlayer.m_iCurCP;
			}
		}

	}
	
	public void SoundOn(boolean i_bSound)
	{
		for(int i=0; i<8; i++) {
			if(m_oPlayers[i]!=null) m_oPlayers[i].SoundOn(i_bSound);
		}
	}
	
	public String GetTimeAsString(double dTime)
	{
		int iSec = (int)dTime / 1000;
		int i100 = ((int)dTime/10) % 100;
		String szString = String.valueOf(iSec);
		szString += ".";
		if(i100 <10) szString += "0";
		szString += i100; 
		return szString;
	}
	
	public int Update(double i_dTime)
	{
		int i, iSize;
		int iGameStatus = 0;
		
		m_dTotalTimePlayed += i_dTime;
		
		if(s_oKeyHandler.IsPressedOnce(KeyEvent.VK_V)) ReplayFollow(true);

		//update doors
		iSize = m_oDoorList.size();
		for(i=0; i<iSize; i++) {
			C_Door oDoor = (C_Door)m_oDoorList.get(i);
			oDoor.Update(i_dTime);
		}
		
		//update bullets
		m_oBulletList.Update(i_dTime);
		C_Bullet oBullet = m_oBulletList.GetNext(true);
		while(oBullet!=null) {
			for(i=0; i<iSize; i++) {
				C_Door oDoor = (C_Door)m_oDoorList.get(i);
				if(oDoor.CheckCollission((C_Object)oBullet)) {
					m_oBulletList.Remove(oBullet.GetUId());
				}
			}
			oBullet = m_oBulletList.GetNext(false);
		}

		//update player movements
		for(i=0; i<8; i++) {
			if(m_oPlayers[i]!=null) m_oPlayers[i].Update(i_dTime);
			
			if(m_bTimeCounting[i]) m_dTotalTime[i] += i_dTime;
		}

		//update landingzones
		iSize = m_oLandingZList.size();
		for(i=0; i<iSize; i++) {
			C_LandingZone oZone = m_oLandingZList.get(i);
			oZone.Update(i_dTime);
			for(int j=0; j<8; j++) {
				if(m_oPlayers[j]!=null) oZone.UpdateFuel(m_oPlayers[j], i_dTime);
			}
		}

		//update enemies
		iSize = m_oCompEnemyList.size();
		for(i=0; i<iSize; i++) {
			C_ComputerEnemy oEnemy = m_oCompEnemyList.get(i);
			int iResult = oEnemy.Update(i_dTime);
			if(iResult==-2) { //remove enemy
				m_oCompEnemyList.remove(i);
				i--; iSize--; //compensate for remove to be able to keep on looping
			}
		}

		//update checkpoints
		iSize = m_oCheckPList.size();
		if(iSize>0) { //there are checkpoints (racemap)
			C_CheckPoint oCP;
			for(i=0; i<iSize; i++) {
				oCP = m_oCheckPList.get(i);
				oCP.Update(i_dTime);
				for(int j=0; j<8; j++) {
					if(m_oPlayers[j]!=null) if(oCP.PlayerAtCP(m_oPlayers[j])) m_bTimeCounting[j] = true; 
				}
			}
		}
		
		//set new map position
		ReplayFollow(false);
		
		return iGameStatus;
	}
	
	public void HandleCargoReply(S_NetMsg2 i_stMsg2)
	{
		int i, iSize = m_oLandingZList.size();
		for(i=0; i<iSize; i++) {
			C_LandingZone oZone = m_oLandingZList.get(i);
			if(i_stMsg2.iInt1 == oZone.m_iOwnerID) {
				/*int iCargo =*/ oZone.PopCargo();
				//TODO cargo load sound should be played but the player id does not exist in the message
				// also there is no good way of playing the unload sound
			}
		}
		//i_stMsg2.iCargoNum //should not need to be used...
	}

	public void HandleCargoRespawn(S_NetMsg2 i_stMsg2)
	{
		int i, iSize = m_oLandingZList.size();
		for(i=0; i<iSize; i++) {
			C_LandingZone oZone = m_oLandingZList.get(i);
			if(i_stMsg2.iInt1 == oZone.m_iOwnerID) {
				oZone.PushCargo(i_stMsg2.iInt3);
			}
		}
		//i_stMsg2.iCargoNum //should not need to be used...
	}
	
	public void HandleLifeRequest(S_NetMsg2 i_stMsg2)
	{
		int i, iSize = m_oLandingZList.size();
		for(i=0; i<iSize; i++) {
			C_LandingZone oZone = m_oLandingZList.get(i);
			if(i_stMsg2.iInt1 == oZone.m_iOwnerID) {
				oZone.TakeExtraLife();
				if(i_stMsg2.iInt3==1) { 
					if(m_oPlayers[i_stMsg2.iPlayerId]!=null) m_oPlayers[i_stMsg2.iPlayerId].AddLife();
				}
			}
		}
	}
	
	public void Draw()
	{
		int i, iSize;
		
		//draw map
		m_oMap.Draw();

		//draw doors
		iSize = m_oDoorList.size();
		for(i=0; i<iSize; i++) {
			C_Door oDoor = m_oDoorList.get(i);
			oDoor.Draw();
		}
		
		//draw checkpoints
		iSize = m_oCheckPList.size();
		for(i=0; i<iSize; i++) {
			C_CheckPoint oCP = m_oCheckPList.get(i);
			oCP.Draw();
		}

		//draw landingzones
		iSize = m_oLandingZList.size();
		for(i=0; i<iSize; i++) {
			C_LandingZone oZone = m_oLandingZList.get(i);
			oZone.Draw();
		}

		//draw enemies
		iSize = m_oCompEnemyList.size();
		for(i=0; i<iSize; i++) {
			C_ComputerEnemy oEnemy = m_oCompEnemyList.get(i);
			oEnemy.Draw();
		}

		//draw players
		for(i=0; i<8; i++) {
			if(m_oPlayers[i]!=null) m_oPlayers[i].Draw();
		}
		
		//draw bullets
		m_oBulletList.Draw();
		
		//draw info
		DrawAllPlayersStatus();
		
		DrawAdditionalPlayerStatus(m_iReplayFollowPlayer);
	}

	
	int GetScore(int i_iPlayerIndex) //not valid for race
	{
		int iScore = 0;
		C_Player oShip = m_oPlayers[i_iPlayerIndex];
		if(oShip!=null) {
			int iLifes = oShip.GetNumLives();
			if(iLifes<0) iLifes = 0;
			iScore = (oShip.m_iScore+iLifes*50)*1000;
			if(m_oMap.MapType()==1) { //mission
				if(m_oMap.MapType()==4) { //mission coop
					int iMaxMapPlayers = m_oMap.MaxPlayers();
					int iLifesLost = 0;
					iScore = 0;
					for(int i=0; i<8; i++) {
						C_Player oShip2 = m_oPlayers[i];
						if(oShip2!=null) {
							iLifes = oShip2.GetNumLives();
							if(iLifes<0) iLifes = 0;
							iLifesLost += (5-iLifes);
							iScore += oShip2.m_iScore;
						}
					}
					iScore += (iMaxMapPlayers-iLifesLost)*50*1000;
				}
				iScore -= (int)(m_dTotalTimePlayed/2); //long time is bad in mission
			}
		}
		return iScore;
	}
	
	private static final int PLAYERSPOS[][] = { //[numplayers][player_index] used in STATUSPOS
		{6,0,0,0,0,0,0,0},
		{6,7,0,0,0,0,0,0},
		{4,6,7,0,0,0,0,0},
		{4,5,6,7,0,0,0,0},
		{2,4,5,6,7,0,0,0},
		{2,3,4,5,6,7,0,0},
		{0,2,3,4,5,6,7,0},
		{0,1,2,3,4,5,6,7}
	};
	private static Point[] STATUSPOS = {
		new Point(120+2,    -84),
		new Point(120+2+228,-84),
		new Point(120+2,    -64),
		new Point(120+2+228,-64),
		new Point(120+2,    -44),
		new Point(120+2+228,-44),
		new Point(120+2,    -24),
		new Point(120+2+228,-24)
	}; 
	private static Color[] STATUSCOLOR = {
		new Color(0xAD/255.0f,0x16/255.0f,0x16/255.0f,1),
		new Color(0x46/255.0f,0xB3/255.0f,0x9E/255.0f,1),
		new Color(0x16/255.0f,0xAE/255.0f,0x16/255.0f,1),
		new Color(0x19/255.0f,0x4F/255.0f,0xBF/255.0f,1),
		new Color(0xC4/255.0f,0xAF/255.0f,0x46/255.0f,1),
		new Color(0x9E/255.0f,0x16/255.0f,0x98/255.0f,1),
		new Color(0xCD/255.0f,0xCD/255.0f,0xCD/255.0f,1),
		new Color(0x93/255.0f,0x9F/255.0f,0x16/255.0f,1)
	};
	private static Point MAINPLAYERSTATUSPOS = new Point(4, 600-50);
	private static Color MAINPLAYERSTATUSCOLOR1 = new Color(0x09/255.0f, 0x6f/255.0f, 0xcc/255.0f, 1); 
	private static Color MAINPLAYERSTATUSCOLOR2 = new Color(0xcc/255.0f, 0x0a/255.0f, 0x2e/255.0f, 1);

	private void DrawAllPlayersStatus()
	{
		s_oGraph.setFont(s_oFont);

		Rectangle r1 = new Rectangle(0, 0, 224, 16);
		int i, j, k = 0;

		int iNumPlayers = 0;
		for(i=0; i<8; i++) {
			if(m_oPlayers[i]!=null) iNumPlayers++;
		}

		int iCurPos = 0;
		for(i=0; i<8; i++) {
			C_Player oShip = m_oPlayers[i];
			if(oShip!=null) {
				int iNumShips = oShip.GetNumLives();
				if(i!=m_iReplayFollowPlayer) {
					for(j=0; j<iNumShips; j++) {
						s_oGraph.drawImage(s_oShipImg[i], 4+j*12, 0+k*16, 4+j*12+16, 0+k*16+16, 0, 0, 48, 48, null);
					}
					k++;
				}

				int iPos = PLAYERSPOS[iNumPlayers-1][iCurPos];
				s_oGraph.setColor(new Color(0.50f, 0.50f, 0.50f, 0.60f));
				r1.x = STATUSPOS[iPos].x;
				r1.y = STATUSPOS[iPos].y+m_oMap.GetScreenRect().height;
				if( m_oCheckPList.size()>0 ) r1.x -= 48;
				s_oGraph.fillRect(r1.x, r1.y, r1.width, r1.height);

				s_oGraph.setColor(STATUSCOLOR[i]);
				r1.y += 12;
				if( m_oCheckPList.size()>0 ) {
					//race:             name, best lap time, current lap and cp
					double dBestLap = oShip.m_dBestLapTime>999000 ? 999000 : oShip.m_dBestLapTime;
					s_oGraph.drawString(oShip.m_szName, r1.x+2, r1.y);
					s_oGraph.drawString(GetTimeAsString(dBestLap), r1.x+102, r1.y);
					s_oGraph.drawString("Lap "+String.valueOf(oShip.m_iCurLap)+"["+String.valueOf(oShip.m_iCurCP)+"]", r1.x+164, r1.y);
				} else {
					//mission/dogfight: name, score
					s_oGraph.drawString(oShip.m_szName, r1.x+2,   r1.y);
					s_oGraph.drawString(String.valueOf(GetScore(i)/1000), r1.x+102, r1.y);
				}
				iCurPos++;
			}
		}
	}
	
	private void DrawAdditionalPlayerStatus(int i_iActivePlayer)
	{
		C_Player oPlayer = m_oPlayers[i_iActivePlayer];
		if(oPlayer==null) return;

		if( m_oCheckPList.size()>0 ) {
			Rectangle stRect = new Rectangle(0, 0, m_oLapsTime.getWidth(null), m_oLapsTime.getHeight(null));
			s_oGraph.drawImage(m_oLapsTime, MAINPLAYERSTATUSPOS.x, MAINPLAYERSTATUSPOS.y, MAINPLAYERSTATUSPOS.x+stRect.width, MAINPLAYERSTATUSPOS.y+stRect.height, stRect.x, stRect.y, stRect.x+stRect.width, stRect.y+stRect.height, null);
			//draw race status (Lap)
			s_oGraph.drawString(String.valueOf(oPlayer.m_iCurLap)+"/"+String.valueOf(m_iNumLapsToRace), MAINPLAYERSTATUSPOS.x+30, MAINPLAYERSTATUSPOS.y+4+12);
			
			double dSec = m_dTotalTime[i_iActivePlayer]/1000;
			s_oGraph.drawString(String.format("%.2f", dSec), MAINPLAYERSTATUSPOS.x+30, MAINPLAYERSTATUSPOS.y+25+12);
		} else {
			if(m_oMap.MapType()==2) { //dogfight
				Rectangle stRect = new Rectangle(0,22, 112, 20);
				s_oGraph.drawImage(m_oCargoFuel, MAINPLAYERSTATUSPOS.x, MAINPLAYERSTATUSPOS.y+22, MAINPLAYERSTATUSPOS.x+stRect.width, MAINPLAYERSTATUSPOS.y+22+stRect.height, stRect.x, stRect.y, stRect.x+stRect.width, stRect.y+stRect.height, null);
			} else {
				Rectangle stRect = new Rectangle(0, 0, m_oCargoFuel.getWidth(null), m_oCargoFuel.getHeight(null));
				s_oGraph.drawImage(m_oCargoFuel, MAINPLAYERSTATUSPOS.x, MAINPLAYERSTATUSPOS.y, MAINPLAYERSTATUSPOS.x+stRect.width, MAINPLAYERSTATUSPOS.y+stRect.height, stRect.x, stRect.y, stRect.x+stRect.width, stRect.y+stRect.height, null);
				//draw cargostatus
				Color oCol = ((oPlayer.CargoNumInHold()>=MAXSPACEINHOLDUNITS) ||
					(oPlayer.CargoWeightInHold()>=MAXSPACEINHOLDWEIGHT)) ? MAINPLAYERSTATUSCOLOR2 : MAINPLAYERSTATUSCOLOR1;
				double dPercentUsed = (double)oPlayer.CargoWeightInHold() / MAXSPACEINHOLDWEIGHT;
				int iWidth = (int)(110*dPercentUsed);
				if(iWidth>0) {
					stRect = new Rectangle(MAINPLAYERSTATUSPOS.x+1, MAINPLAYERSTATUSPOS.y+1, iWidth, 18);
					s_oGraph.setColor(oCol);
					s_oGraph.fillRect(stRect.x, stRect.y, stRect.width, stRect.height);
				}
				//draw fuelstatus
				oCol = MAINPLAYERSTATUSCOLOR1;
				dPercentUsed = oPlayer.FuelPercentLeft();
				iWidth = (int)(110*dPercentUsed);
				if(iWidth>0) {
					stRect = new Rectangle(MAINPLAYERSTATUSPOS.x+1, MAINPLAYERSTATUSPOS.y+23, iWidth, 18);
					s_oGraph.setColor(oCol);
					s_oGraph.fillRect(stRect.x, stRect.y, stRect.width, stRect.height);
				}
			}
			//draw num lifes
			int i, iNumShips = oPlayer.GetNumLives();
			if(iNumShips>0) {
				int y = (MAINPLAYERSTATUSPOS.y-24)+(m_oMap.MapType()==2 ? 22:0);
				for(i=0; i<iNumShips; i++) {
					int x = MAINPLAYERSTATUSPOS.x+i*12;
					s_oGraph.drawImage(s_oShipImg[i_iActivePlayer], x, y, x+16, y+16, 0, 0, 48, 48, null);
				}
			}
		}
	}

}
