package GalaxyPK;

import java.awt.*;

public class C_Object {
	//ID constants
	public static final int COMPUTER_ENEMY_BASEID = 8;
	public static final int DOOR_BASEID = 112;
	
	//common object info about position and position change (pixels)
	public    double m_dPosX;           //object global position in x-direction
	public    double m_dPosY;           //object global position in y-direction
	protected double m_dVelX;           //object speed in x-direction
	protected double m_dVelY;           //object speed in y-direction
	protected double m_dAcceleration;   //object acceleration in the heading direction
	protected double m_dDirection;      //object direction in degrees

	//acceleration modifiers
	protected double m_dResistanceX, m_dResistanceY; //velocity dependent object (air)resistance (0 for no resistance)
	protected double m_dGravityX, m_dGravityY;       //the gravity the object will feel (can be negative) (0 for no gravity) (pixel/sec2)

	//approximate direction
	protected int    m_iDirStep;        //number of degrees between each step/image (10=default)
	protected int    m_iRealDir;        //(int)m_dDirection/m_iDirStep (the direction that the object can have)

	//object identifiers
	protected static int s_iIDCounter = 0;
	public    int    m_iOwnerID;		//id of the object that owns this object (or the main objects id)
	
	//graphics for the object (for drawing and collision)
	protected Image     m_oImage = null;
	protected Rectangle m_stImageRect = null;
	protected boolean   m_bCanCollide; //this must be false when the object is displayed but it cannot collide. ex. during explosion maybe
	//drawing variables
	protected Rectangle m_stScreenArea;  //the area of the map that is visible on screen (from map object)
	public static Graphics s_oGraph;     //must be set externally before drawing any objects

	//statics
	public static double PointDistance(int i_iX1, int i_iY1, int i_iX2, int i_iY2)
	{
		int iDistX = i_iX1 - i_iX2;
		int iDistY = i_iY1 - i_iY2; 
		return Math.sqrt(iDistX*iDistX + iDistY*iDistY);
	}

	//constructor
	public C_Object()
	{
		m_bCanCollide = true;
		m_iDirStep = 10;
		m_stScreenArea = null;
		
		m_iOwnerID = -1;
		
		m_dAcceleration = 0.0;
		m_dDirection = 0.0;
		m_dResistanceX = m_dResistanceY = 0.0;
		m_dGravityX = m_dGravityY = 0.0;
	}

	/* calculates new position and speed, depending on the elapsed time
	 * i_dTime    time in milliseconds since last call
	 * uses m_dDirection, m_dAcceleration as input; they must be set to the newest value
	 */
	void OUpdate(double i_dTime)
	{
		double dTime = (i_dTime*0.001); //easier to work with seconds
		
		//calc velocity
		m_iRealDir = (int)(m_dDirection/m_iDirStep);
		
		double ax = Math.cos(m_iRealDir*(m_iDirStep*(Math.PI/180.0))) * m_dAcceleration;
		double ay = Math.sin(m_iRealDir*(m_iDirStep*(Math.PI/180.0))) * m_dAcceleration;
		
		ax += m_dGravityX;
		ay += m_dGravityY;
		ax -= m_dVelX*m_dResistanceX;
		ay -= m_dVelY*m_dResistanceY;
		
		//calc position
		m_dPosX += (m_dVelX * dTime) + (0.5*ax*dTime*dTime); //??
		m_dPosY += (m_dVelY * dTime) + (0.5*ay*dTime*dTime); //??
		
		//calc vel
		m_dVelX += ax * dTime;
		m_dVelY += ay * dTime;
	}
	
	
	/* transparent draw of the provided image onto screen if visible
	 * (clipping preformed if partly visible)
	 * i_oImg     image
	 * i_stRect   object rectangle within i_Img, NULL = entire img
	 * i_iX       x-cordinate of upper left corner of screen in the global space
	 * i_iY       y-cordinate of upper left corner of screen in the global space
	 *screenposition is (m_dPosX-i_x, m_dPosY-i_Y)  */
	void ODraw(Image i_oImg, Rectangle i_stRect, int i_iX, int i_iY)
	{
		if(i_oImg==null) return;
		
		boolean bDraw = true;
		int iPosX = (int)m_dPosX - i_iX;
		int iPosY = (int)m_dPosY - i_iY;
		int iWidth, iHeight;
		Rectangle stRect;
		
		//get src dimensions
		if(i_stRect!=null) {
			stRect = new Rectangle(i_stRect);
		} else {
			stRect = new Rectangle(0, 0, i_oImg.getWidth(null), i_oImg.getHeight(null));
		}
		iWidth = stRect.width;
		iHeight = stRect.height;
		
		//clipping x
		if(iPosX < 0) {
			stRect.x -= iPosX;
			stRect.width += iPosX;
			iPosX = 0;
		} else if(iPosX+iWidth > m_stScreenArea.width) {
			stRect.width -= (iPosX+iWidth) - m_stScreenArea.width;
		}
		if(stRect.width <= 0) bDraw = false;
		else { //clipping y
			if(iPosY < 0) {
				stRect.y -= iPosY;
				stRect.height += iPosY;
				iPosY = 0;
			} else if(iPosY+iHeight > m_stScreenArea.height) {
				stRect.height -= (iPosY+iHeight) - m_stScreenArea.height;
			}
			if(stRect.height <= 0) bDraw = false;
		}

		//draw?
		if(bDraw) {
			int iScreenX = iPosX+m_stScreenArea.x;
			int iScreenY = iPosY+m_stScreenArea.y;
			s_oGraph.drawImage(i_oImg, iScreenX, iScreenY, iScreenX+stRect.width, iScreenY+stRect.height, stRect.x, stRect.y, stRect.x+stRect.width, stRect.y+stRect.height, null);
		}
	}

	public Point GetSize()
	{ //returns x=width and y=height
		int w, h;
		
		if(m_stImageRect!=null) {
			w = m_stImageRect.width;
			h = m_stImageRect.height;
		} else if(m_oImage!=null) {
			w = m_oImage.getWidth(null);
			h = m_oImage.getHeight(null);
		} else w = h = 0;
		
		return new Point(w, h);
	}

	public Point GetCenterPos()
	{
		Point p = GetSize();
		p.x = (int)(m_dPosX + p.x*0.5);
		p.y = (int)(m_dPosY + p.y*0.5);
		return p;
	}
	
	public double GetDistanceTo(C_Object i_oObj)
	{
		double dDistX, dDistY, dDist;
		
		Point p1 = GetCenterPos();
		Point p2 = i_oObj.GetCenterPos();
		
		dDistX = p1.x - p2.x;
		dDistY = p1.y - p2.y; 
		dDist = Math.sqrt(dDistX*dDistX + dDistY*dDistY);
		
		return dDist;
	}

	/* tests if two images have collided, first check if they overlap,
	 * and if they do, check their common overlapping rectangles pixel by pixel,
	 * and if two pixels on the same position is != 0 there is a collision (regardless of alphachannel)
	 * i_OtherImg other object image
	 * i_iOtherX  x-cordinate of upper left corner of OtherImg in the global space
	 * i_iOtherY  y-cordinate of upper left corner of OtherImg in the global space
	 */
	boolean OCheckCollision(Image i_oOtherImg, int i_iOtherX, int i_iOtherY, Rectangle i_stOtherRect)
	{
		if(!m_bCanCollide || m_oImage==null) return false; //no collision

		//return C_Collision.Compare(m_oImage, (int)m_dPosX, (int)m_dPosY, i_oOtherImg, i_iOtherX, i_iOtherY, m_stImageRect, i_stOtherRect);
		return C_Collision.CompareBuffered(m_oImage, (int)m_dPosX, (int)m_dPosY, i_oOtherImg, i_iOtherX, i_iOtherY, m_stImageRect, i_stOtherRect);
	}
	
	boolean OCheckCollision(C_Object i_oObject)
	{
		if(!m_bCanCollide || m_oImage==null || !i_oObject.m_bCanCollide || i_oObject.m_oImage==null) return false; //no collision
		
		//return C_Collision.Compare(m_oImage, (int)m_dPosX, (int)m_dPosY, i_oObject.m_oImage, (int)i_oObject.m_dPosX, (int)i_oObject.m_dPosY, m_stImageRect, i_oObject.m_stImageRect);
		return C_Collision.CompareBuffered(m_oImage, (int)m_dPosX, (int)m_dPosY, i_oObject.m_oImage, (int)i_oObject.m_dPosX, (int)i_oObject.m_dPosY, m_stImageRect, i_oObject.m_stImageRect);
	}
}
