package GalaxyPK;

import java.awt.*;

public class C_Player extends C_PlayerBase
{
	public static final double DEFAULT_SHIPGRAVITYBASEX = 0;   //pixel/second 2
	public static final double DEFAULT_SHIPGRAVITYBASEY = 70;  //pixel/second 2
	public static final double DEFAULT_SHIPRESISTANCEX = 0.68; //constant (velocity dependent)
	public static final double DEFAULT_SHIPRESISTANCEY = 0.43;

	public static double SHIPGRAVITYBASEX = 0;   //pixel/second 2
	public static double SHIPGRAVITYBASEY = 70;  //pixel/second 2
	public static double SHIPRESISTANCEX = 0.68; //constant (velocity dependent)
	public static double SHIPRESISTANCEY = 0.43; //^
	public static final int SHIPSTEARSPEED  = 235;  //degree/second
	public static final int SHIPTHRUST      = 270;  //pixel/second 2
	public static final int BULLETBASEVEL = 220;   //pixel/second
	public static final int BULLETFREETIME = 3100; //millisec to be free from bullets when just come alive

	private C_TileMap    m_oMap;
	private C_BulletList m_oBulletList;
	private C_Replay     m_oReplay;
	
	private static final int SHIPSIZE = 48;
	private Image     m_oShipImg;
	private Rectangle m_astShipRects[];
	
	private int m_iStartPosX, m_iStartPosY;

	private int m_iKeyFlag;
	private int m_iNumLifes; //-1 = no limit
	private boolean m_bAlive;
	private boolean m_bFreeFromBullets;
	private double m_dFreeFromBulletsTimer;

	private int m_iEngineSoundCh;
	private boolean m_bLanded;
	private double m_dLandTime = 0.0;
	
	//cargo variables
	public static final int MAXSPACEINHOLDWEIGHT = 50;
	public static final int MAXSPACEINHOLDUNITS = 3;
	private int m_iSpaceUsed = 0;
	private int m_iNumUsed = 0;
	private int m_aiHold[] = new int[MAXSPACEINHOLDUNITS];
	
	//score variables
	public int m_iScore = 0;
	
	//race variables
	public int m_iCurCP = 0;
	public int m_iCurLap = 0;
	public double m_dBestLapTime = 0;
	public String m_szName;

	//fuel
	private double m_dMaxFuelInTank = 60;
	private double m_dFuel = m_dMaxFuelInTank; //number of seconds one can use the thruster
	public void AddFuel(double dFuel)
	{
		if(m_dFuel<m_dMaxFuelInTank) {
			m_dFuel += dFuel; 
			if(m_dFuel>m_dMaxFuelInTank) m_dFuel=m_dMaxFuelInTank;
		}
	}
	public double FuelPercentLeft()
	{
		return m_dFuel/m_dMaxFuelInTank;
	}
	public void ConsumeFuel(double i_dTime)
	{
		if(m_oMap.MapType()!=3) {
			//consume fuel
			if(m_dAcceleration!=0) m_dFuel -= i_dTime/1000.0;
			if(m_dFuel<0) {
				m_dFuel = 0;
				//Kill(); //no thrust possible instead... it usually leads to death
			}
		}
	}

	public C_Player(int i_iId, String i_szName, Image i_oShipImg, C_Replay i_oReplay, C_TileMap i_oMap, C_BulletList i_oBulletList, int i_iStartX, int i_iStartY, int i_iNumLives)
	{
		//copy parameters and init base
		super();
		m_iOwnerID = i_iId;
		m_oMap = i_oMap;
		m_oBulletList = i_oBulletList;
		m_oReplay = i_oReplay;
		m_iKeyFlag = 0;
		m_szName = i_szName;
		
		//init ship src image info
		m_oImage = m_oShipImg = i_oShipImg;
		m_astShipRects = new Rectangle[8*9];
		int i, j;
		for(j=0; j<4; j++) {       //ship
			for(i=0; i<9; i++) {
				m_astShipRects[(j*9+i)] = new Rectangle(i*SHIPSIZE, ((j+1)%4)*SHIPSIZE, SHIPSIZE, SHIPSIZE);
			}
		}
		for(j=4; j<8; j++) {       //thrust
			for(i=0; i<9; i++) {
				m_astShipRects[(j*9+i)] = new Rectangle(i*SHIPSIZE, (4+((j+1)%4))*SHIPSIZE, SHIPSIZE, SHIPSIZE);
			}
		}
		m_stScreenArea = m_oMap.GetScreenRect();

		//init others
		m_iNumLifes = i_iNumLives;
		m_iScore = 0;
		m_iStartPosX = i_iStartX;
		m_iStartPosY = i_iStartY;
		m_dPosX = m_iStartPosX;
		m_dPosY = m_iStartPosY;
		m_dVelX = m_dVelY = m_dAcceleration = 0.0;
		m_dDirection = 270.0;

		m_dGravityX = SHIPGRAVITYBASEX;
		m_dGravityY = SHIPGRAVITYBASEY;
		m_dResistanceX = SHIPRESISTANCEX;
		m_dResistanceY = SHIPRESISTANCEY;

		m_bAlive = true;
		m_bLanded = false;
		m_bFreeFromBullets = false;
		m_dFreeFromBulletsTimer = 0.0;
	}

	public int Update(double i_dTime)
	{
		double  dSTime = (i_dTime/1000);
		int     iRval = -1;

		m_dLandTime += i_dTime;

		if(m_bFreeFromBullets) {
			m_dFreeFromBulletsTimer += i_dTime;
			if(m_dFreeFromBulletsTimer > BULLETFREETIME) {
				m_bFreeFromBullets = false;
			}
		}

		//maybe come to life again
		CalcSoundVolPan(s_oListner);
		if(!m_bAlive) {
			if( UpdateExplosion(i_dTime) ) { //set alive again
				if(m_iNumLifes != 0 ) {
					m_bAlive = true;
					m_bFreeFromBullets = true;
					m_dFreeFromBulletsTimer = 0.0;
					m_dPosX = m_iStartPosX;
					m_dPosY = m_iStartPosY;
					m_dDirection = 270.0;
					m_dGravityX  = SHIPGRAVITYBASEX;
					m_dGravityY  = SHIPGRAVITYBASEY;
					m_dFuel = m_dMaxFuelInTank;
					m_bLanded = false;
				}
			}
		}

		//always
		S_NetMsg1 stMsg1;
		S_ReplayAction stAction;
		while( (stAction=m_oReplay.Get(m_iOwnerID, 1))!=null ) {
			stMsg1 = stAction.stMsg1;
			//movement info
			if((stMsg1.iMsg & S_NetMsg1.N_MSG1_MOVEMENT)!=0) {
				if(stMsg1.fAcceleration!=m_dAcceleration) {
					if(m_dAcceleration>0) m_oSound.StopSound(m_iEngineSoundCh);
					else m_iEngineSoundCh = m_oSound.PlaySound(2, m_dVol, m_dPan, true);
				}

				GetCommonInfo(stMsg1);
				m_iSpaceUsed = stMsg1.iCargoInHold;
			}
			//keystate change
			if( (stMsg1.iMsg & S_NetMsg1.N_MSG1_KEYCHANGE)!=0 ) m_iKeyFlag = stMsg1.iKeyFlag;

			//bullet fired
			if( (stMsg1.iMsg & S_NetMsg1.N_MSG1_NEWBULLET)!=0) {
				C_Bullet oBullet = CreateBullet();

				oBullet.SetUId(stMsg1.iBulletUniqeValBase);
				m_oBulletList.Add(oBullet);
				m_oSound.PlaySound(0, m_dVol, m_dPan, false);
			}

			//player killed
			if( (stMsg1.iMsg & S_NetMsg1.N_MSG1_KILLED)!=0 ) {
				switch(stMsg1.iReason_Hits_Door) {
					case 1: //self
						Kill();
						break;
					case 2: //bullet
						Kill();
						m_oBulletList.Remove(stMsg1.iCauseID);
						iRval = (stMsg1.iCauseID & 0xff);
						break;
					case 3: //player
						iRval = stMsg1.iCauseID + 256;
						Kill();
						break;
				}
			}

			//score change
			if( (stMsg1.iMsg & S_NetMsg1.N_MSG1_SCORE)!=0 ) m_iScore = stMsg1.iScore;

			//player landed
			if( (stMsg1.iMsg & S_NetMsg1.N_MSG1_LANDED)!=0 ) LandOnZone();

			if( (stMsg1.iMsg & S_NetMsg1.N_MSG1_LAPINFO)!=0 ) {
				m_iCurLap = stMsg1.iLap;
				m_iCurCP  = stMsg1.iCP;
				m_dBestLapTime = stMsg1.fBestLapTime;
			}
		}

		//make the explosion be still even if messages lags behind
		if(!m_bAlive) {
			m_dVelX = 0.0;
			m_dVelY = 0.0;
		}

		//lifted from landingzone?
		if(m_dAcceleration!=0 && m_bLanded) {
			m_bLanded = false;
			ResetGravity();
		}
		 
		//steering
		{
			if( (m_iKeyFlag & S_NetMsg1.GF_KEY_DOWN)!=0 ) {
				if(m_dDirection!=270.0) {
					//get way to turn
					int iSign = 1, iSignAfter = 1;
					if(m_dDirection<90 || m_dDirection>270) iSign = -1;
					m_dDirection += iSign*(SHIPSTEARSPEED/3*dSTime);
					if(m_dDirection<90 || m_dDirection>270) iSignAfter = -1;
					if(iSign!=iSignAfter) m_dDirection = 270.0;
				}
			} else { //normal steering
				m_dDirection +=(m_iKeyFlag & S_NetMsg1.GF_KEY_RIGHT)!=0 ? SHIPSTEARSPEED*dSTime:0;
				m_dDirection -=(m_iKeyFlag & S_NetMsg1.GF_KEY_LEFT)!=0 ? SHIPSTEARSPEED*dSTime:0;
			}
			//clipping
			int iNumCounts = 0;
			if(m_dDirection<0) iNumCounts = (int)(m_dDirection/360.0)-1;
			else iNumCounts = (int)(m_dDirection/360.0);
			m_dDirection-=iNumCounts*360;
		}

		ConsumeFuel(i_dTime);
		OUpdate(i_dTime);

		if(m_dAcceleration!=0) { //ship is moving, set current engine volume
			m_oSound.ChangeVolAndPan(m_iEngineSoundCh, m_dVol, m_dPan);
		}

		m_stImageRect = m_astShipRects[m_iRealDir];
		return iRval;
	}
	
	public void Kill()
	{
		if(!m_bAlive) return;
		
		m_bAlive = false;
		m_bCanCollide = false;
		UpdateExplosion(0.0); //reset explosion

		//to prevent movement during explosion
		m_dGravityX = 0; 
		m_dGravityY = 0;

		//take a life if not unlimited
		if(m_iNumLifes!=-1) m_iNumLifes--;

		Stop();
		m_oSound.PlaySound(1, m_dVol, m_dPan, false); //play explosion sound
	}

	//to be able to stop the sound while pause, or when escape is pressed
	public void SoundOn(boolean i_bSound)
	{
		if(!i_bSound && m_dAcceleration!=0) m_oSound.StopSound(m_iEngineSoundCh); //off
		else if(m_dAcceleration!=0) m_iEngineSoundCh = m_oSound.PlaySound(2, m_dVol, m_dPan, true); //on
	}

	private void Stop()
	{
		m_dVelX = 0.0; m_dVelY = 0.0;
		if(m_dAcceleration!=0) m_oSound.StopSound(m_iEngineSoundCh); //stop engine sound
		m_dAcceleration = 0.0;
	}
	
	public void LandOnZone()
	{
		m_bLanded = true;
		m_dGravityX = 0;
		m_dGravityY = 0;
		m_oSound.PlaySound(3, m_dVol, m_dPan, false);
	}
	
	public void Draw()
	{
		Point p = m_oMap.GetPos();
		int iMapX = p.x;
		int iMapY = p.y;

		if(m_bAlive) {
			if(m_dAcceleration>0) ODraw(m_oShipImg, m_astShipRects[m_iRealDir+36], iMapX, iMapY);
			if(!m_bFreeFromBullets || (m_bFreeFromBullets && ((int)m_dFreeFromBulletsTimer % 300 < 200)) )
				ODraw(m_oShipImg, m_astShipRects[m_iRealDir], iMapX, iMapY);
		} else {
			DrawExplosion(iMapX, iMapY);
		}
	}

	private C_Bullet CreateBullet()
	{
		double dTemp = m_iRealDir*(m_iDirStep*(Math.PI/180));
		double dSin = Math.sin(dTemp);
		double dCos = Math.cos(dTemp);
		double dPosX = m_dPosX + SHIPSIZE/2 + dCos * (SHIPSIZE/2-2);
		double dPosY = m_dPosY + SHIPSIZE/2 + dSin * (SHIPSIZE/2-2);
		double dVelX = m_dVelX + dCos * BULLETBASEVEL;
		double dVelY = m_dVelY + dSin * BULLETBASEVEL;
		double dDirection = m_iRealDir * m_iDirStep;
		
		return new C_Bullet(m_oMap, dPosX, dPosY, dVelX, dVelY, dDirection, m_iOwnerID);
	}

	public boolean HoldEmpty()
	{
		return (m_iNumUsed==0);
	}
	public int CargoWeightInHold()
	{
		return m_iSpaceUsed;
	}
	public int CargoNumInHold()
	{
		return m_iNumUsed; //this is currently always 0 (no support in network messages)
	}
	public boolean IsLanded()
	{
		return m_bLanded;
	}
	public boolean IsLandedForTime(double i_dTime)
	{
		return (m_bLanded && (m_dLandTime>i_dTime));
	}
	public boolean IsAlive()
	{
		return m_bAlive;
	}
	public boolean LoadCargo(int iCargo)
	{
		boolean bCargoLoaded = false;
		if(m_iNumUsed<MAXSPACEINHOLDUNITS && (m_iSpaceUsed+iCargo)<=MAXSPACEINHOLDWEIGHT) {
			//space in hold, add cargo
			m_aiHold[m_iNumUsed++] = iCargo;
			m_iSpaceUsed += iCargo;
			bCargoLoaded = true;
			
			//calc new gravity is done when lifted from the zone, not here

			//play load cargo sound
			m_oSound.PlaySound(8, m_dVol, m_dPan, false);
		}
			
		return bCargoLoaded;
	}
	public void UnloadCargo()
	{
		if(m_iNumUsed>0) {
			m_iNumUsed--;
			m_iSpaceUsed -= m_aiHold[m_iNumUsed];

			//calc new gravity is done when lifted from the zone, not here
			
			//add score for the cargo moved
			m_iScore += m_aiHold[m_iNumUsed];
			
			//play unload cargo sound
			m_oSound.PlaySound(9, m_dVol, m_dPan, false);
			
			//reset landtime just to make the delay before a level ends the same every time
			m_dLandTime = 0;
		}
	}
	private void ResetGravity()
	{
		m_dGravityX = SHIPGRAVITYBASEX;
		m_dGravityY = SHIPGRAVITYBASEY + (m_iSpaceUsed*2);
	}

	public boolean FreeFromBullets()
	{
		return m_bFreeFromBullets;
	}

	public int GetNumLives()
	{
		return m_iNumLifes;
	}
	public void AddLife()
	{
		if(m_iNumLifes!=-1) m_iNumLifes++; //if mission
	}
}
