package GalaxyPK;

import java.util.*;

public class C_Replay 
{
	public static int N_MSG2_GAMEOVER     = 0x0E;
	public static int N_MSG2_SCORE        = 0x0F; //iScore
	public static int N_MSG2_NUMKILLS     = 0x10; //iNumKills(server) or iKilledBy(client)
	public static int N_MSG2_NUMLAPS      = 0x11; //iNumLaps
	public static int N_MSG2_CARGO        = 0x20; //iZoneId, iCargoNum
	public static int N_MSG2_CARGO_R      = 0x21; //iZoneId, iCargoNum, iCargoReqAccepted
	public static int N_MSG2_NEXTLEVEL    = 0x22; //
	public static int N_MSG2_CARGORESPAWN = 0x23; //iZoneId, iCargoNum, iCargoWeight
	public static int N_MSG2_LIFE_R       = 0x25; //iZoneId, iLifeReqAccepted

	private int m_iCurTimeSlot;
	private S_ReplayHeader m_stHeader;
	private ArrayList<S_ReplayAction> m_clActions;
	private int m_aiCurGetPos[] = new int[129*2];
	public boolean m_bLoaded;
	
	public C_Replay()
	{
		m_stHeader = new S_ReplayHeader();
		m_clActions = new ArrayList<S_ReplayAction>();
		Reset();
	}
	
	public void ResetBeforePlay()
	{
		for(int i=0; i<m_aiCurGetPos.length; i++) m_aiCurGetPos[i] = 0;
		m_iCurTimeSlot = 0;
	}
	
	public void Reset()
	{
		for(int i=0; i<m_aiCurGetPos.length; i++) m_aiCurGetPos[i] = 0;
		m_iCurTimeSlot = 0;
		m_clActions.clear();
		m_bLoaded = false;
	}
	
	public void IncTimeSlot()
	{
		m_iCurTimeSlot++;
	}

	public S_ReplayHeader GetGameInfo()
	{
		return m_stHeader;
	}

	public S_ReplayAction Get(int i_iId, int i_iType)
	{
		int i, iSize = m_clActions.size();
		int iIdIndex = i_iId+1+129*(i_iType-1);
		for(i=m_aiCurGetPos[iIdIndex]; i<iSize; i++) {
			S_ReplayAction stAction = m_clActions.get(i);
			if((i_iId==-1 || stAction.iId==i_iId) && stAction.iMsg==i_iType) {
				if(stAction.iTimeSlot <= m_iCurTimeSlot) { //do it as fast as possible if <
					m_aiCurGetPos[iIdIndex] = i+1;
					return stAction;
				} else break;
			}
		}
		return null;
	}

	public boolean LoadFromMem(byte i_pMem[])
	{
		boolean bResult = false;
		S_ReplayAction stAction;
		int i, j;
		Reset();
		//header
		if(i_pMem.length<160) return false;
		m_stHeader.iVersion = i_pMem[0];
		for(i=4; i<52; i++) {
			if(i_pMem[i]==0) break;
			m_stHeader.szMapFile = m_stHeader.szMapFile+new String(new byte[] {i_pMem[i]});
		}
		m_stHeader.iLevelType = i_pMem[52];
		m_stHeader.bNoShooting = i_pMem[56]!=0;
		m_stHeader.bFlyThrough = i_pMem[57]!=0;
		m_stHeader.bNoBulletsHit = i_pMem[58]!=0;
		m_stHeader.iNumPlayers = i_pMem[60];
		for(j=0; j<8; j++) {
			for(i=64+j*12; i<64+j*12+12; i++) {
				if(i_pMem[i]==0) break;
				m_stHeader.szPlayerName[j] = m_stHeader.szPlayerName[j]+new String(new byte[] {i_pMem[i]});
			}
		}
		if(VerifyHeader()) {
			int iTemp, iPos = 160;
			//elements
			while((int)(iPos+64)<=i_pMem.length) {
				bResult = false;
				stAction = new S_ReplayAction();
				stAction.iTimeSlot = GetInt(i_pMem, iPos+0, 4);
				stAction.iId = i_pMem[iPos+4]; if(stAction.iId==255) stAction.iId = -1;
				stAction.iMsg = i_pMem[iPos+5];
				if(stAction.iMsg==1) {
					//stAction.stMsg1
					iTemp = GetInt(i_pMem, iPos+6, 2); //size 58
					if(iTemp!=58) break;
					iTemp = GetInt(i_pMem, iPos+8, 2); //type 0x50
					if(iTemp!=0x50) break;
					stAction.stMsg1.iMsg = i_pMem[iPos+10];
					stAction.stMsg1.iSenderId = i_pMem[iPos+11]; if(stAction.stMsg1.iSenderId==255) stAction.stMsg1.iSenderId = -1;
					stAction.stMsg1.fPosX = GetFloat(i_pMem, iPos+12);
					stAction.stMsg1.fPosY = GetFloat(i_pMem, iPos+16);
					stAction.stMsg1.fVelX = GetFloat(i_pMem, iPos+20);
					stAction.stMsg1.fVelY = GetFloat(i_pMem, iPos+24);
					stAction.stMsg1.fAcceleration = GetFloat(i_pMem, iPos+28);
					stAction.stMsg1.fDirection = GetFloat(i_pMem, iPos+32);
					stAction.stMsg1.fGravityY = GetFloat(i_pMem, iPos+36);
					stAction.stMsg1.iCargoInHold = GetInt(i_pMem, iPos+40, 2);
					stAction.stMsg1.bCanCollide = i_pMem[iPos+42]!=0;
					stAction.stMsg1.iKeyFlag = i_pMem[iPos+43];
					stAction.stMsg1.iNumBullets = i_pMem[iPos+44];
					stAction.stMsg1.iBulletUniqeValBase = GetInt(i_pMem, iPos+45, 4);
					stAction.stMsg1.iReason_Hits_Door = i_pMem[iPos+49];
					stAction.stMsg1.iCauseID = GetInt(i_pMem, iPos+50, 4);
					stAction.stMsg1.iScore = GetInt(i_pMem, iPos+54, 4);
					stAction.stMsg1.iLap = i_pMem[iPos+58];
					stAction.stMsg1.iCP = i_pMem[iPos+59];
					stAction.stMsg1.fBestLapTime = GetFloat(i_pMem, iPos+60);
				} else if(stAction.iMsg==2) {
					//stAction.stMsg2
					iTemp = GetInt(i_pMem, iPos+6, 2); //size 48
					if(iTemp!=48) break;
					stAction.stMsg2.iType = GetInt(i_pMem, iPos+8, 2);
					stAction.stMsg2.iPlayerId = GetInt(i_pMem, iPos+10, 2);
					stAction.stMsg2.iGameId = GetInt(i_pMem, iPos+12, 2);
					stAction.stMsg2.iInt1 = GetInt(i_pMem, iPos+14, 4);
					stAction.stMsg2.iInt2 = GetInt(i_pMem, iPos+18, 4);
					stAction.stMsg2.iInt3 = GetInt(i_pMem, iPos+22, 4);
					for(i=22; i<22+24; i++) {
						if(i_pMem[i]==0) break;
						stAction.stMsg2.szLevelName = stAction.stMsg2.szLevelName+new String(new byte[] {i_pMem[i]});
					}
					for(i=22; i<22+18; i++) {
						if(i_pMem[i]==0) break;
						stAction.stMsg2.szGameName = stAction.stMsg2.szGameName+new String(new byte[] {i_pMem[i]});
					}
					for(i=40; i<40+12; i++) {
						if(i_pMem[i]==0) break;
						stAction.stMsg2.szPlayerName = stAction.stMsg2.szPlayerName+new String(new byte[] {i_pMem[i]});
					}
					stAction.stMsg2.iOptions = GetInt(i_pMem, iPos+52, 2);
				}
				if(!VerifyElement(stAction)) {
					bResult = false;
					break;
				}
				m_clActions.add(stAction);
				iPos += 64;
				bResult = true;
			}
		} else bResult = false;
		if(!bResult) Reset();
		m_bLoaded = bResult;
		return bResult;
	}
	
	int GetInt(byte i_pMem[], int i_iPos, int i_iIntSize)
	{
		/*if(i_iIntSize==4) return (((i_pMem[i_iPos+0] & 0xff) << 24) + ((i_pMem[i_iPos+1] & 0xff) << 16) + ((i_pMem[i_iPos+2] & 0xff) << 8) + (i_pMem[i_iPos+3] & 0xff));
		if(i_iIntSize==2) return (((i_pMem[i_iPos+0] & 0xff) << 8) + (i_pMem[i_iPos+1] & 0xff));
		if(i_iIntSize==1) return (i_pMem[i_iPos+0] & 0xff);*/
		if(i_iIntSize==4) return (((i_pMem[i_iPos+3] & 0xff) << 24) + ((i_pMem[i_iPos+2] & 0xff) << 16) + ((i_pMem[i_iPos+1] & 0xff) << 8) + (i_pMem[i_iPos+0] & 0xff));
		if(i_iIntSize==2) return (((i_pMem[i_iPos+1] & 0xff) << 8) + (i_pMem[i_iPos+0] & 0xff));
		if(i_iIntSize==1) return (i_pMem[i_iPos+0] & 0xff);
		return 0;
	}

	float GetFloat(byte i_pMem[], int i_iPos)
	{
		/*int value = (((i_pMem[i_iPos] & 0xff) << 24)
			+ ((i_pMem[i_iPos + 1] & 0xff) << 16)
			+ ((i_pMem[i_iPos + 2] & 0xff) << 8) + ((i_pMem[i_iPos + 3] & 0xff) << 0));
		return Float.intBitsToFloat(value);*/
		int value = (((i_pMem[i_iPos + 3] & 0xff) << 24)
			+ ((i_pMem[i_iPos + 2] & 0xff) << 16)
			+ ((i_pMem[i_iPos + 1] & 0xff) << 8) + ((i_pMem[i_iPos] & 0xff) << 0));
		return Float.intBitsToFloat(value);
	}

	public boolean VerifyString(String i_szText)
	{
		int i;
		for(i=0; i<i_szText.length(); i++) {
			if(i_szText.charAt(i)<32) {
				return false;
			}
		}
		return true;
	}

	public boolean VerifyHeader()
	{
		int i;
		if(m_stHeader.iVersion!=1) return false;
		if(m_stHeader.iNumPlayers<1 || m_stHeader.iNumPlayers>8) return false;
		if(m_stHeader.iLevelType<1 || m_stHeader.iLevelType>4) return false;

		if(!VerifyString(m_stHeader.szMapFile)) return false;
		for(i=0; i<8; i++) if(!VerifyString(m_stHeader.szPlayerName[i])) return false;

		return true;
	}

	public boolean VerifyElement(S_ReplayAction i_stAction)
	{
		if(i_stAction.iMsg!=1 && i_stAction.iMsg!=2) return false;
		if(i_stAction.iTimeSlot<0) return false;
		//if(i_stAction.iId<-1) return false;

		if(i_stAction.iMsg==1) {
			//TODO verify i_pstAction->stMsg1
		} else {
			int iType = i_stAction.stMsg2.iType;
			if(iType!=N_MSG2_GAMEOVER && iType!=N_MSG2_SCORE
				&& iType!=N_MSG2_NUMKILLS && iType!=N_MSG2_NUMLAPS
				&& iType!=N_MSG2_CARGO_R && iType!=N_MSG2_CARGORESPAWN
				&& iType!=N_MSG2_LIFE_R && iType!=N_MSG2_NEXTLEVEL)
				return false;
			//TODO further verify i_pstAction->stMsg2
		}
		return true;
	}

};
