package GalaxyPK;

import java.awt.*;
import java.io.*;
import java.util.*;

import LoaderPK.*;

class S_MapHeader
{
	public String szGridfile;
	public String szTileset;
	public int iTileSize;
	public int iWidth, iHeight;
}

public class C_TileMap
{
	//map grid info
	private byte m_grid[][];
	private int m_iWidth, m_iHeight;
	
	//file info
	private S_MapHeader m_stHeader;
	private C_MapObjects m_oExtended;
	private C_Replay m_oReplay;
	
	//draw info
	private Rectangle m_stDrawarea;
	private int m_iDrawareaWidthSave, m_iDrawareaHeightSave;
	private int m_iPixMaxX, m_iPixMaxY;
	private int m_iScrollMaxX, m_iScrollMaxY;
	private int m_iLastX, m_iLastY;
	private boolean m_bWrapX, m_bWrapY;

	//tileset info
	private int       m_iTileSize; //pixelsize of 
	private int       m_iNumTiles;
	private Image     m_oTilesImg;
	private Rectangle m_astTileRects[];
	
	//graph and loading
	private Image m_oOutput;
	private Graphics m_oGraph;
	private int m_iMapType = 1;
	private int m_iMapMaxPlayers;
	
	// constructor to create a map from an existing file
	public C_TileMap(C_Replay i_oReplay, Image i_oOutput, Rectangle i_stDrawarea)
	{
		//copy params
		m_oOutput = i_oOutput;
		m_oGraph = m_oOutput.getGraphics();
		m_stDrawarea = i_stDrawarea;
		m_oReplay = i_oReplay;
		String szFilename = "files/"+m_oReplay.GetGameInfo().szMapFile;
		
		//allocate memory for map, and set defaults
		m_grid = null;
		m_iLastX = m_iLastY = 0;
		LoadHeader(szFilename); //get info about the map/level (including the tileset) (from a textfile)
		LoadMap(); //load the mapfile (binary)
		LoadTileSet(); //an image with different tiles is loaded and prepared
		
		// save the area entered by the user if wrap around mode should change later
		m_iDrawareaWidthSave = m_stDrawarea.width;
		m_iDrawareaHeightSave = m_stDrawarea.height;
		SetMapWrapping(false, false);
		
		C_Collision.BufferImage(m_oTilesImg);
	}

	protected void finalize() throws Throwable
	{
		super.finalize();
		C_Collision.UnbufferImage(m_oTilesImg);
	}

	// set if the map shall stop or wrap around when scrolling past the borders
	void SetMapWrapping(boolean i_bHorizontal, boolean i_bVertical)
	{
		m_bWrapX = i_bHorizontal;
		m_bWrapY = i_bVertical;
		
		m_stDrawarea.width = m_iDrawareaWidthSave;
		m_stDrawarea.height = m_iDrawareaHeightSave;
		// if the show area is larger than the map, and wrapparound
		//  shall not be used, we need to limit the showarea
		if(!m_bWrapX && m_iWidth*m_iTileSize<m_iDrawareaWidthSave) m_stDrawarea.width = m_iWidth*m_iTileSize;
		if(!m_bWrapY && m_iHeight*m_iTileSize<m_iDrawareaHeightSave) m_stDrawarea.height = m_iHeight*m_iTileSize;
		
		// calc maximum upper left coordinate we can scroll to
		//  without getting outside the map (used if no wrapparound)
		m_iPixMaxX = m_stDrawarea.x + m_stDrawarea.width;
		m_iPixMaxY = m_stDrawarea.y + m_stDrawarea.height;
		m_iScrollMaxX = m_iWidth*m_iTileSize - m_stDrawarea.width;
		m_iScrollMaxY = m_iHeight*m_iTileSize - m_stDrawarea.height;
		if(m_iScrollMaxX < 0) m_iScrollMaxX = 0;
		if(m_iScrollMaxY < 0) m_iScrollMaxY = 0;
	}

	// loads the map grid data from the file associated with the map-obj
	private boolean LoadHeader(String i_szMapHeader)
	{
		int i, j;
		int iEnemyCount = 0;
		int iZoneCounter = 0;
		int iDoorCounter = 0;
		int iStartPosCounter = 0;

		m_stHeader = new S_MapHeader();
		m_oExtended = new C_MapObjects();
		m_oExtended.m_oReplay = m_oReplay;
		m_iMapType = 1;
		C_Player.SHIPGRAVITYBASEX = C_Player.DEFAULT_SHIPGRAVITYBASEX;
		C_Player.SHIPGRAVITYBASEY = C_Player.DEFAULT_SHIPGRAVITYBASEY;
		C_Player.SHIPRESISTANCEX = C_Player.DEFAULT_SHIPRESISTANCEX;
		C_Player.SHIPRESISTANCEY = C_Player.DEFAULT_SHIPRESISTANCEY;
		
		DataInputStream oLines = null;
		try {
			InputStream pStream = C_Loader.GetStream(i_szMapHeader);
			if(pStream==null) pStream = C_Loader.GetStream("/"+i_szMapHeader);
			oLines = new DataInputStream(new BufferedInputStream(pStream));
			C_LineReader stReader = new C_LineReader(oLines);
			String strLine;
			while((strLine = stReader.readLine()) != null) {
				strLine = strLine.trim();
				if(strLine.length() == 0) continue;
				
				int iCmdEnd = strLine.indexOf(' ');
				String strCommand = strLine.substring(0, iCmdEnd);
				String strData = strLine.substring(iCmdEnd).trim();
				StringTokenizer st = new StringTokenizer(strData);
	
				if(strCommand.compareTo("*MAPTYPE") == 0) {
					String szType = st.nextToken();
					if(szType.compareTo("RACE") == 0) m_iMapType = 3;
					if(szType.compareTo("DOGFIGHT") == 0) m_iMapType = 2;
					if(szType.compareTo("MISSION") == 0) m_iMapType = 1;
					if(szType.compareTo("MISSION_COOP") == 0) m_iMapType = 4;
					if(m_iMapType == 3) m_oExtended.m_iNumLapsToRace = Integer.parseInt(st.nextToken());
				} else if(strCommand.compareTo("*MAPSIZE") == 0) {
					m_stHeader.iWidth = Integer.parseInt(st.nextToken());
					m_stHeader.iHeight = Integer.parseInt(st.nextToken());
				} else if(strCommand.compareTo("*MAPFILE") == 0) {
					m_stHeader.szGridfile = new String(strData);
				} else if(strCommand.compareTo("*TILESIZE") == 0) {
					m_stHeader.iTileSize = Integer.parseInt(st.nextToken());
				} else if(strCommand.compareTo("*TILESET") == 0) {
					m_stHeader.szTileset = new String(strData);
				} else if(strCommand.compareTo("*GRAVITY") == 0) {
					C_Player.SHIPGRAVITYBASEX = Double.parseDouble(st.nextToken());
					C_Player.SHIPGRAVITYBASEY = Double.parseDouble(st.nextToken());
				} else if(strCommand.compareTo("*RESISTANCE") == 0) {
					C_Player.SHIPRESISTANCEX = Double.parseDouble(st.nextToken());
					C_Player.SHIPRESISTANCEY = Double.parseDouble(st.nextToken());
				} else if(strCommand.compareTo("*PLAYERSTARTPOS") == 0) {
					m_oExtended.m_aiPlayerStartPosX[iStartPosCounter] = Integer.parseInt(st.nextToken());
					m_oExtended.m_aiPlayerStartPosY[iStartPosCounter] = Integer.parseInt(st.nextToken());
					iStartPosCounter++;
				} else if(strCommand.compareTo("*MAXPLAYERS") == 0) {
					m_iMapMaxPlayers = Integer.parseInt(st.nextToken());
				} else if(strCommand.compareTo("*LANDINGZONE") == 0) {
					int x = Integer.parseInt(st.nextToken());
					int y = Integer.parseInt(st.nextToken());
					int w = Integer.parseInt(st.nextToken());
					boolean bHomeBase = (st.nextToken().compareTo("HOMEBASE") == 0);
					boolean bExtraLife = (st.nextToken().compareTo("EXTRALIFE") == 0);
					boolean bAntenna = (st.nextToken().compareTo("ANTENNA") == 0);
					boolean bHouse = (st.nextToken().compareTo("WAREHOUSE") == 0);
					int iNumCargo;
					int aiCargo[] = null;
					if(!bHomeBase) {
						iNumCargo = Integer.parseInt(st.nextToken());
						if(iNumCargo>0) {
							aiCargo = new int[iNumCargo];
							for(j=0; j<iNumCargo; j++) aiCargo[j] = Integer.parseInt(st.nextToken());
						}
					}
					C_LandingZone oZone = new C_LandingZone(this, iZoneCounter, x, y, w, bAntenna, bHouse, aiCargo, bExtraLife);
					m_oExtended.m_oLandingZList.add(oZone);
					iZoneCounter++;
				} else if(strCommand.compareTo("*ENEMY") == 0) {
					S_EnemyInfo stParams = new S_EnemyInfo();
					stParams.iEnemyType = Integer.parseInt(st.nextToken());
					
					StringTokenizer st1;
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*ANGLE"
					stParams.iAngle = Integer.parseInt(st1.nextToken()); //get angle
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*FIREINTERVAL"
					stParams.iFireInterval = Integer.parseInt(st1.nextToken()); //get fireinterval
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*SPEED"
					stParams.iSpeed = Integer.parseInt(st1.nextToken()); //get speed
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*NUMWAYPOINTS"
					int iNumWP = Integer.parseInt(st1.nextToken());
					stParams.astWayPoints = new Point[iNumWP]; //init number of waypoints
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*WAYPOINTSX"
					for(j=0; j<iNumWP; j++) stParams.astWayPoints[j] = new Point(Integer.parseInt(st1.nextToken()), 0);
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*WAYPOINTSY"
					for(j=0; j<iNumWP; j++) stParams.astWayPoints[j].y = Integer.parseInt(st1.nextToken());
					//add enemy to list
					C_ComputerEnemy m_oCompEnemy = new C_ComputerEnemy(this, m_oReplay, m_oExtended.m_oBulletList, stParams, C_Object.COMPUTER_ENEMY_BASEID+iEnemyCount);
					m_oExtended.m_oCompEnemyList.add(m_oCompEnemy);
					iEnemyCount++;
				} else if(strCommand.compareTo("*CHECKPOINTS") == 0) {
					int iNumCP = Integer.parseInt(st.nextToken());
					//init number of checkpoints and read them
					StringTokenizer st1;
					for(j=0; j<iNumCP; j++) {
						st1 = new StringTokenizer(stReader.readLine());
						st1.nextToken(); //skip "*CHECKPOINT"
						C_CheckPoint oCP = new C_CheckPoint(this, new Point(Integer.parseInt(st1.nextToken()), Integer.parseInt(st1.nextToken())), new Point(Integer.parseInt(st1.nextToken()), Integer.parseInt(st1.nextToken())), j);
						m_oExtended.m_oCheckPList.add(oCP);
					}
				} else if(strCommand.compareTo("*DOOR") == 0) {
					int x, y, iAngle;
					S_DoorInfo stParams = new S_DoorInfo();
					x = Integer.parseInt(st.nextToken());
					y = Integer.parseInt(st.nextToken());
					stParams.iSize = Integer.parseInt(st.nextToken());

					StringTokenizer st1;
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*ANGLE"
					iAngle = Integer.parseInt(st1.nextToken()); //get angle
					stParams.bHorizontal = ((iAngle/90)%2)!=0;
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*OPENTIME"
					stParams.iOpenedForTime = Integer.parseInt(st1.nextToken());
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*CLOSETIME"
					stParams.iClosedForTime = Integer.parseInt(st1.nextToken());
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*OPENSPEED"
					stParams.iOpeningSpeed = Integer.parseInt(st1.nextToken());
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*CLOSESPEED"
					stParams.iClosingSpeed = Integer.parseInt(st1.nextToken());

					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*NUMBUTTONS"
					stParams.iNumButtons = Integer.parseInt(st1.nextToken());

					stParams.stButtonPos = new Point[stParams.iNumButtons];
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*BUTTONSX"
					for(i=0; i<stParams.iNumButtons; i++) stParams.stButtonPos[i] = new Point(Integer.parseInt(st1.nextToken()), 0);
					st1 = new StringTokenizer(stReader.readLine());
					st1.nextToken(); //skip "*BUTTONSY"
					for(i=0; i<stParams.iNumButtons; i++) stParams.stButtonPos[i].y = Integer.parseInt(st1.nextToken());
					
					//add door to list
					C_Door oDoor = new C_Door(this, m_oReplay, x, y, stParams, C_Object.DOOR_BASEID+iDoorCounter);
					m_oExtended.m_oDoorList.add(oDoor);
					iDoorCounter++;
				}
			}
			//oFile.close() generates exception so... :(
		} catch(Exception e) {
			return false;
		}

		m_iWidth = m_stHeader.iWidth;
		m_iHeight = m_stHeader.iHeight;
		m_iTileSize = m_stHeader.iTileSize;

		m_grid = new byte[m_iHeight][m_iWidth]; //alloc rows and columns
		
		return true;
	}
	
	// loads the map grid data from the file mentioned in the "header"
	private void LoadMap()
	{
		DataInputStream oFile = null;
		try {
			InputStream pStream = C_Loader.GetStream("files/"+m_stHeader.szGridfile);
			if(pStream==null) pStream = C_Loader.GetStream("/files/"+m_stHeader.szGridfile);
			oFile = new DataInputStream(new BufferedInputStream(pStream));
			for(int i=0; i<m_iHeight; i++) oFile.readFully(m_grid[i]); //read each row
		} catch(Exception e) {
			//do nothing the level wont just work
		}
	}

	//an image with different tiles is loaded and prepared
	private void LoadTileSet()
	{
		int i, j;

		//begin loading the image
		
		m_oTilesImg = C_Loader.LoadImage("files/"+m_stHeader.szTileset.replaceFirst(".tga", ".png"));
		if(m_oTilesImg==null) m_oTilesImg = C_Loader.LoadImage("/files/"+m_stHeader.szTileset.replaceFirst(".tga", ".png"));
		
		//analyse tileset (get rectangles)
		int iWidth, iHeight, iNumX, iNumY, iTileNum;
		iWidth = m_oTilesImg.getWidth(null);
		iHeight = m_oTilesImg.getHeight(null);
		iNumX = iWidth / m_iTileSize;
		iNumY = iHeight / m_iTileSize;
		m_iNumTiles = iNumX * iNumY;
		m_astTileRects = new Rectangle[m_iNumTiles];
		
		iTileNum = 0;
		for(j=0; j<iNumY; j++) {
			for(i=0; i<iNumX; i++) {
				m_astTileRects[iTileNum] = new Rectangle(i*m_iTileSize, j*m_iTileSize, m_iTileSize, m_iTileSize);
				iTileNum++;
			}
		}
	}

	// draws all the currently visible tiles in the map on screen
	public void Draw()
	{
		int iTempX, iTempY;
		int iBaseTileX, iBaseOffsetX;
		int iTileX, iTileY;
		int iOffsetX, iOffsetY;
		int iScreenX = m_stDrawarea.x, iScreenY = m_stDrawarea.y;
		int x, y, w, h; //calculated srcrectangle within the tileset
		Rectangle stRectRef; //to access the right tile within the tileset
		
		iBaseTileX = iTileX = m_iLastX / m_iTileSize;
		iBaseOffsetX = iOffsetX = m_iLastX % m_iTileSize;
		iTileY = m_iLastY / m_iTileSize;
		iOffsetY = m_iLastY % m_iTileSize;
		do {
			do {
				stRectRef = m_astTileRects[m_grid[iTileY][iTileX]];
				x = stRectRef.x + iOffsetX;
				y = stRectRef.y + iOffsetY;
				iTempX = m_iPixMaxX - iScreenX;
				iTempY = m_iPixMaxY - iScreenY;
				w = (m_iTileSize - iOffsetX)<iTempX ? (m_iTileSize - iOffsetX):iTempX;
				h = (m_iTileSize - iOffsetY)<iTempY ? (m_iTileSize - iOffsetY):iTempY;
				
				m_oGraph.drawImage(m_oTilesImg, iScreenX, iScreenY, iScreenX+w, iScreenY+h, x, y, x+w, y+h, null);
				
				iTileX++;
				if(m_bWrapX) iTileX %= m_iWidth;
				iScreenX += w;
				iOffsetX = 0;
			} while (iScreenX < m_iPixMaxX);
			iOffsetX = iBaseOffsetX;
			iOffsetY = 0;
			iTileX = iBaseTileX;
			iTileY++;
			if(m_bWrapY) iTileY %= m_iHeight;
			
			iScreenX = m_stDrawarea.x;
			iScreenY += h;
		} while (iScreenY < m_iPixMaxY);
	}

	// draws all of i_stGameArea of the map to the provided image, it returns true if the image rendered is completely black (to save from pixelcompare later on)
	public boolean CheckMaskCollision(Rectangle i_stGameArea, C_Object i_oTestObject)
	{
		int iTempX, iTempY, iTileNum;
		boolean bCollision = false;
		int iBaseTileX, iBaseOffsetX;
		int iTileX, iTileY;
		int iOffsetX, iOffsetY;
		int x, y, w, h; //calculated srcrectangle within the tileset
		Rectangle stRectRef; //to access the right tile within the tileset
		
		int iScreenX = 0;
		int iScreenY = 0;
		int iScreenMaxX = i_stGameArea.width;
		int iScreenMaxY = i_stGameArea.height;

		iBaseTileX = iTileX = i_stGameArea.x / m_iTileSize;
		iBaseOffsetX = iOffsetX = i_stGameArea.x % m_iTileSize;
		iTileY = i_stGameArea.y / m_iTileSize;
		iOffsetY = i_stGameArea.y % m_iTileSize;
		do {
			do {
				iTileNum = m_grid[iTileY][iTileX];
				iTempX = iScreenMaxX - iScreenX;
				iTempY = iScreenMaxY - iScreenY;
				w = (m_iTileSize - iOffsetX)<iTempX ? (m_iTileSize - iOffsetX):iTempX;
				h = (m_iTileSize - iOffsetY)<iTempY ? (m_iTileSize - iOffsetY):iTempY;

				if(iTileNum!=0) { //only test for collision if nonblack
					stRectRef = m_astTileRects[iTileNum];
					x = stRectRef.x + iOffsetX;
					y = stRectRef.y + iOffsetY;

					bCollision = i_oTestObject.OCheckCollision(m_oTilesImg, iTileX*m_iTileSize+iOffsetX, iTileY*m_iTileSize+iOffsetY, new Rectangle(x, y, w, h));
					if(bCollision) return true;
				}
				
				iTileX++;
				if(m_bWrapX) iTileX %= m_iWidth;
				iScreenX += w;
				iOffsetX = 0;
			} while (iScreenX < iScreenMaxX);
			iOffsetX = iBaseOffsetX;
			iOffsetY = 0;
			iTileX = iBaseTileX;
			iTileY++;
			if(m_bWrapY) iTileY %= m_iHeight;
			
			iScreenX = 0;
			iScreenY += h;
		} while (iScreenY < iScreenMaxY);
		
		return bCollision;
	}
	
	// tests if an object has collided with nonblack pixels in the map
	public boolean CheckCollision(C_Object i_oObject)
	{
		boolean bResult = false;
		int iX, iY, iW, iH;
		iX = (int)i_oObject.m_dPosX;
		iY = (int)i_oObject.m_dPosY;
		Point p = i_oObject.GetSize();
		iW=p.x; iH=p.y;
		
		//check boundries
		if(iX<0 || iY<0 || iX+iW>m_iWidth*m_iTileSize || iY+iH>m_iHeight*m_iTileSize) bResult = true;

		if(!bResult) { //check pixels
			//boolean bAllBlack = DrawAreaToImage(new Rectangle(iX,iY,iW,iH), m_oTmpImage);
			bResult = CheckMaskCollision(new Rectangle(iX,iY,iW,iH), i_oObject);
			
		}
		return bResult;
	}

	// moves the showpoint of the map relative to the last
	// position and draws the result
	public void MoveRelative(int i_iX, int i_iY)
	{
		MoveAbsolute(m_iLastX + i_iX, m_iLastY + i_iY);
	}
	
	// moves the showpoint of the map to a new position,
	// given in world pixel coordinates
	public void MoveAbsolute(int i_iX, int i_iY)
	{
		int iCmpx = m_bWrapX ? m_iWidth*m_iTileSize  : m_iScrollMaxX;
		int iCmpy = m_bWrapY ? m_iHeight*m_iTileSize : m_iScrollMaxY;
		Point p = new Point(i_iX, i_iY);
		WrappAroundTest(p, iCmpx, iCmpy);
		
		m_iLastX = p.x; m_iLastY = p.y;
	}

	public Point GetPos()
	{
		return new Point(m_iLastX, m_iLastY);
	}
	
	public Rectangle GetScreenRect()
	{
		return m_stDrawarea;
	}

	// checks if the values goes outside the limits, and take actions
	private boolean WrappAroundTest(Point io_p, int i_iWidth, int i_iHeight)
	{
		boolean bResult = true;
		
		int x = io_p.x;
		int y = io_p.y;
		
		if(x>=i_iWidth) {
			if(m_bWrapX) x %= i_iWidth;
			else {
				x = i_iWidth;
				bResult = false;
			}
		}
		if(y>=i_iHeight) {
			if(m_bWrapY) y %= i_iHeight;
			else {
				y = i_iHeight;
				bResult = false;
			}
		}
		if(x<0) {
			if(m_bWrapX) x = i_iWidth + (x % i_iWidth);
			else {
				x = 0;
				bResult = false;
			}
		}
		if(y<0) {
			if(m_bWrapY) y = i_iHeight + (y % i_iHeight);
			else {
				y = 0;
				bResult = false;
			}
		}
		io_p.x = x;
		io_p.y = y;

		return bResult;
	}

	public C_MapObjects GetExtendedInfo()
	{
		return m_oExtended;
	}

	public int MapType()
	{
		return m_iMapType;
	}

	public int MaxPlayers()
	{
		return m_iMapMaxPlayers;
	}
}
