package SoundPK;

import java.io.*;
import javax.sound.sampled.*;

import LoaderPK.*;

public class C_Sound
{
	private static final int MAXLOADEDSOUNDS = 32;
	private static final int MAXCHANNELS = 8;

	private boolean m_bInited = false;
	//private Mixer m_oMixer; //<- not using, better off with AudioSystem.getLine()
	private AudioFormat m_oFormat;
	private double m_dFramesPerSec; //depending on m_oFormat
	private boolean m_bMute = false;
	
	//loaded sounds
	private byte m_oSampleDataList[][] = new byte[MAXLOADEDSOUNDS][];
	
	//channel info
	private SourceDataLine[] m_oChannels = new SourceDataLine[MAXCHANNELS];
	private int m_aiChPlaying[] = new int[MAXCHANNELS];
	private int m_aiChPos[] = new int[MAXCHANNELS]; //playingpos in sample
	private boolean m_aiChLoop[] = new boolean[MAXCHANNELS];
	private int m_iNumChannels = 0;

	private static C_Sound s_oSound = null;
	

	protected void finalize() throws Throwable
	{
		super.finalize();
		if(!m_bInited) return;
		
		for(int i=0; i<m_iNumChannels; i++) {
			m_oChannels[i].stop();
			m_oChannels[i].close();
		}
	}
	
	public static C_Sound GetSoundObject()
	{
		if(s_oSound==null) s_oSound = new C_Sound();
		return s_oSound;
	}
	public static void DestroySoundObject()
	{
		s_oSound = null;
	}
	
	public void Init()
	{
		int i;
		//Mixer.Info[] aMixerInfo = AudioSystem.getMixerInfo();
		//if(aMixerInfo.length==0) return;
		//m_oMixer = AudioSystem.getMixer(aMixerInfo[0]); //there is no good way to know which mixer to choose, just take mixer 0

		//the output format
		m_oFormat = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED, 44100.0F, 16, 2, 4, 44100.0F, false);
		m_dFramesPerSec = m_oFormat.getFrameRate();

		DataLine.Info oInfo = new DataLine.Info(SourceDataLine.class, m_oFormat);

		try { //obtain and open the channels.
			m_iNumChannels = 0;
			for(i=0; i<MAXCHANNELS; i++) {
				m_aiChPlaying[i] = -1; //nothing playing

				m_oChannels[i] = (SourceDataLine)AudioSystem.getLine(oInfo);
				//m_oChannels[i] = (SourceDataLine)m_oMixer.getLine(oInfo); //<- not using mixer

				m_oChannels[i].open(m_oFormat, 200000);
				m_oChannels[i].start();
				m_iNumChannels++;
			}
		} catch(LineUnavailableException e) {
			System.out.println(e.getMessage());
			if(m_iNumChannels==0) return; //return if not even a single channel could be opened
		}

		m_bInited = true;
	}

	public boolean LoadSound(String i_szFile, int i_iIndex)
	{
		if(i_iIndex>=MAXLOADEDSOUNDS || !m_bInited) return false;
		
		boolean bResult = true;
		InputStream pStream = C_Loader.GetStream(i_szFile);

		try {
			InputStream bufferedIn = new BufferedInputStream(pStream);
			AudioInputStream oStreamInput = AudioSystem.getAudioInputStream(bufferedIn);

			m_oSampleDataList[i_iIndex] = ReadWAVConverted(oStreamInput, m_oFormat);
			bResult = (m_oSampleDataList[i_iIndex]!=null);
		} catch(Exception e) {
			bResult = false;
			System.out.println(e.getMessage());
		}
		return bResult;
	}
	
	public void SoundOff(boolean i_bMute)
	{
		m_bMute = i_bMute;
	}
	
	public int PlaySound(int i_iIndex)
	{
		return PlaySound(i_iIndex, 100.0, 0, false);
	}
	
	public int PlaySound(int i_iIndex, double i_dPercentVol, double i_dPan, boolean i_bLoop)
	{
		if(m_oSampleDataList[i_iIndex]==null || !m_bInited || m_bMute) return -1;
		
		int i;
		for(i=0; i<m_iNumChannels; i++) {
			if(m_aiChPlaying[i]==-1 && m_oChannels[i].available()==m_oChannels[i].getBufferSize()) { //find a free channel and play
				SetVolAndPan(m_oChannels[i], i_dPercentVol, i_dPan);
				m_aiChPlaying[i] = i_iIndex;
				m_aiChLoop[i] = i_bLoop;
				m_aiChPos[i] = 0;
				return i;
			}
		}
		return -1;
	}
	public void StopSound(int i_iChannel)
	{
		if(i_iChannel==-1 || !m_bInited) return;
		
		m_aiChPlaying[i_iChannel] = -1;
		m_oChannels[i_iChannel].flush();
	}
	public void ChangeVolAndPan(int i_iChannel, double i_dPercentVol, double i_dPan)
	{
		if(i_iChannel==-1 || !m_bInited) return;
		
		SetVolAndPan(m_oChannels[i_iChannel], i_dPercentVol, i_dPan);
	}
	
	public void Update(double dTime)
	{
		if(!m_bInited || m_bMute) return;
		
		int i, iIndex;
		int iFrameSize = m_oFormat.getFrameSize();
		for(i=0; i<m_iNumChannels; i++) {
			iIndex = m_aiChPlaying[i];
			if(m_oChannels[i].isActive() && m_oChannels[i].available()==m_oChannels[i].getBufferSize() && iIndex==-1) {
				m_oChannels[i].flush();
			}
			if(iIndex!=-1) { //all playing channels
				int iPos = m_aiChPos[i];
				if(iPos==0) {
					dTime+=150; //render 150 ms extra 1st time
				}
				int iBytesToRender = ((int)(((dTime*1.1)*m_dFramesPerSec)*0.001))*iFrameSize;
				iBytesToRender = Math.min(m_oChannels[i].available(), iBytesToRender);
				int iNumBytes = iBytesToRender; //default
				int iBytesToEof = m_oSampleDataList[iIndex].length - iPos;
				boolean bEof = false;
				
				if(iBytesToEof <= iBytesToRender) { //reached end of sound?
					bEof = true;
					iNumBytes = iBytesToEof;
					iBytesToRender -= iBytesToEof;
				}
				if(iNumBytes>0) {
					m_oChannels[i].write(m_oSampleDataList[iIndex], iPos, iNumBytes);
					m_aiChPos[i]+=iNumBytes;
				}

				if(bEof) {
					if(m_aiChLoop[i]) { //write 1st part of looped sound
						m_oChannels[i].write(m_oSampleDataList[iIndex], 0, iBytesToRender);
						m_aiChPos[i]=iBytesToRender;
					} else m_aiChPlaying[i] = -1; //nothing playing
				}
			}
		}
	}

	private void SetVolAndPan(SourceDataLine i_oChannel, double i_dPercentVol, double i_dPan)
	{
		//get controls for volume and panning
		if(i_oChannel.isControlSupported(FloatControl.Type.MASTER_GAIN)) {
			try {
				FloatControl oCtrlGain = (FloatControl) i_oChannel.getControl(FloatControl.Type.MASTER_GAIN);
				double value = i_dPercentVol/100.0; 
				float dB = (float) (Math.log(value==0.0?0.0001:value)/Math.log(10.0)*20.0); 
				oCtrlGain.setValue(dB);
			} catch (Exception e) {
				System.out.println(e.getMessage());
			}
		}
		if(i_oChannel.isControlSupported(FloatControl.Type.PAN)) {
			try {
				FloatControl oCtrlPan  = (FloatControl) i_oChannel.getControl(FloatControl.Type.PAN);
				oCtrlPan.setValue((float)i_dPan);
			} catch (Exception e) {
				System.out.println(e.getMessage());
			}
		}
	}

	//both input and output must have the same sample type PCM_SIGNED for example
	//no support for 8 bit formats and conversions, 8bit is crap anyway
	//no support for big/little endian conversion
	//no interpolation is made in conversion between different framerates
	//no good support for stereo -> mono conversion (just takes first channel, no mixing)
	private byte[] ReadWAVConverted(AudioInputStream i_oInput, AudioFormat i_stAfOut)
	{
		//input data info
		AudioFormat stAfIn = i_oInput.getFormat();
		int iNumFramesIn = (int)i_oInput.getFrameLength();
		int iFrameSizeIn = stAfIn.getFrameSize();
		int iNumChIn = stAfIn.getChannels();
		int iNumBytesPerSampleIn = stAfIn.getSampleSizeInBits()/8;
		
		int iNumChOut = i_stAfOut.getChannels();
		int iNumBytesPerSampleOut = i_stAfOut.getSampleSizeInBits()/8;
		
		int iLengthIn = iNumFramesIn*iFrameSizeIn;
		byte aInput[] = new byte[iLengthIn];

		try {
			int iNumRead, iOffset = 0;
			while((iNumRead=i_oInput.read(aInput,iOffset,iLengthIn))>0) iOffset+=iNumRead;
			//i_oInput.read(aInput);
		} catch(Exception e) {
			System.out.println(e.getMessage());
		}

		if(i_stAfOut.matches(stAfIn)) return aInput; //same as output (no conversion)

		//no support for 8 bit at all!
		if(stAfIn.getSampleSizeInBits()!=16 || i_stAfOut.getSampleSizeInBits()!=16) return null;
		//no support for big/little endian conversion
		if(stAfIn.isBigEndian()!=i_stAfOut.isBigEndian()) return null;

		double dSampleStep = stAfIn.getFrameRate()/i_stAfOut.getFrameRate();
		int i, iLengthOut = (int)((iNumFramesIn*i_stAfOut.getFrameSize())/dSampleStep);
		byte aOutput[] = new byte[iLengthOut];
		
		//fill each byte of the output, with data from input
		double dInputFrame = 0.0;
		int iCh = 0;
		int iHiLo = 0;
		for(i=0; i<iLengthOut; i++) {
			int iSrcFrameOffset = (iCh%iNumChIn)*iNumBytesPerSampleIn+(iHiLo%iNumBytesPerSampleIn);
			int iSrcPos = (int)(dInputFrame)*iFrameSizeIn+iSrcFrameOffset;
			aOutput[i] = aInput[iSrcPos];

			//go forward in input
			iHiLo++; //next part of sample
			if(iHiLo==iNumBytesPerSampleOut) {
				iHiLo=0; //iHiLo%=iNumBytesPerSampleOut;
				iCh++; //next channel
				if(iCh==iNumChOut) {
					iCh = 0; //iCh%=iNumChOut;
					dInputFrame += dSampleStep; //next frame
					//if(dInputFrame>=iNumFramesIn) dInputFrame=iNumFramesIn; //clip due to possible rounding errors
				}
			}
		}
		return aOutput;
	}

}
