/*
 * @(#)WebReplay.java 1.0 20111011
 *
 */

import java.awt.*;

import javax.swing.*;
import java.awt.event.*;

//import java.net.*;

import LoaderPK.*;
import GalaxyPK.*;
import SoundPK.*;

public class WebReplay extends JApplet implements Runnable, KeyListener
{
	static final long serialVersionUID = 433336;
	
	private Image    m_oDBImage;
	private Graphics m_oDBGraph;
	private static volatile Thread m_oTheThread;
	
	private C_Timer  m_oTimer;
	private C_KeyHandler m_oKeyHandler;
	private double  m_dTimeLeft;
	private boolean m_bEnd = false;
	private double  m_dTimeToEnd;

	//variables used in menu and setup
	//private URL m_oBaseURL;
	
	//game variables
	private volatile int m_iState = -1;
	private boolean m_bPause = false;
	private Image m_oShipImg;
	private Image m_oExplosionImg;
	private Image m_oBulletImg;
	private Image m_oLandingZoneImg;
	private Image m_oCpImg;
	private Image m_oE0Img, m_oE1Img, m_oE2Img, m_oE3Img, m_oE4Img, m_oE5Img, m_oE6Img, m_oE7Img;
	private Image m_oDoorImg;
	private Image m_oCargoFuel;
	private Image m_oLapsTime;

	private C_Sound m_oSound;
	private boolean m_bSound = true;

	//game variables
	private C_TileMap m_oMap;
	private C_MapObjects m_oMapObjects = null;
	
	//used for replay
	private ReplayConnection m_oReplayConnection = null;
	private C_Replay         m_oReplay = null; 
	
	//used in loading phase
	private int m_iCurrentFile = 0;
	private int m_iNumFiles = 16 + (m_bSound ? 10 : 0);
	private LoaderStatus m_oLoaderStatus = null;

	private void OwnInit()
	{
		this.addKeyListener(this);
		setFocusable(true);
		requestFocusInWindow();

		m_oTimer = new C_Timer();
		m_oKeyHandler = new C_KeyHandler();
		
		//init path base url
		//m_oBaseURL = getDocumentBase();
		//init loader
		C_Loader.Init(this);
		
		//init sound
		m_oSound = SoundPK.C_Sound.GetSoundObject();
		m_oSound.SoundOff(!m_bSound);
		if(m_bSound) m_oSound.Init();

		// initialize buffer
		setBackground(Color.DARK_GRAY);
		m_oDBImage = createImage(this.getSize().width, this.getSize().height);
		m_oDBGraph = m_oDBImage.getGraphics();
		
		//init loader status bar
		m_oLoaderStatus = new LoaderStatus(this.getSize().width, this.getSize().height, m_oDBGraph, m_iNumFiles-1);
	}

	public void OwnInitAfterLoad()
	{
		//buffer some images in collision object
		C_Collision.BufferImage(m_oBulletImg);
		C_Collision.BufferImage(m_oShipImg);
		C_Collision.BufferImage(m_oLandingZoneImg);
		C_Collision.BufferImage(m_oE0Img);
		C_Collision.BufferImage(m_oE1Img);
		C_Collision.BufferImage(m_oE2Img);
		C_Collision.BufferImage(m_oE3Img);
		C_Collision.BufferImage(m_oE4Img);
		C_Collision.BufferImage(m_oE5Img);
		C_Collision.BufferImage(m_oE6Img);
		C_Collision.BufferImage(m_oE7Img);
		C_Collision.BufferImage(m_oDoorImg);

		//create replay objects
		m_oReplayConnection = new ReplayConnection();
		String szReplayMap = null;
		String szName = null;
		try {
			szReplayMap = getParameter("Level");
			szName = getParameter("Name");
		} catch(Exception e) {
			e.printStackTrace();
		}
		//for debugging:
		if(szReplayMap==null) szReplayMap = "mission00";
		if(szName==null) szName = "RH";

		//init static game info
		C_MapObjects.InitImagesAndStuff(m_oDBGraph, m_oKeyHandler, m_oShipImg, m_oExplosionImg, m_oBulletImg, m_oLandingZoneImg,
				m_oCpImg, m_oE0Img, m_oE1Img, m_oE2Img, m_oE3Img, m_oE4Img, m_oE5Img, m_oE6Img, m_oE7Img, m_oDoorImg, m_oCargoFuel, m_oLapsTime);
		
		m_oReplay = LoadReplay(szReplayMap, szName);
	}

	private C_Replay LoadReplay(String szMap, String szName)
	{
		String szURL = "http://galaxy-forces.com/hiscore_getreplay2.php?Level="+szMap+"&Name="+szName;
		String szURL2 = "http://galaxyv2.sourceforge.net/hiscore_getreplay2.php?Level="+szMap+"&Name="+szName;
		C_Replay pReplay = m_oReplayConnection.getReplay(szURL);
		if(pReplay==null) pReplay = m_oReplayConnection.getReplay(szURL2);
		//the webserver is the same but recently the galaxy-forces domain was registered.
		//apparently java-applets can only retrieve data from hosts with the same name as the applet runs from (even though it is the same sever, and both are valid).
		//so we try both.
		return pReplay;
	}
	
	public void init()
	{
		OwnInit();
	}

	public void paint(Graphics g)
	{
		//no need for anything here, the game is redrawn all the time, but -
		update(g); //must be here to run with java 1.6
	}

	//load all
	private boolean LoadAll()
	{
		switch(m_iCurrentFile) {
			//load images
			case 0: m_oBulletImg = C_Loader.LoadImage("files/bullet.png"); break;
			case 1: m_oShipImg = C_Loader.LoadImage("files/ship_w.png"); break;
			case 2: m_oExplosionImg = C_Loader.LoadImage("files/explosion.png"); break;
			case 3: m_oLandingZoneImg = C_Loader.LoadImage("files/zone.png"); break;
			case 4: m_oCpImg = C_Loader.LoadImage("files/cpbk.png"); break;
			case 5: m_oE0Img = C_Loader.LoadImage("files/enemy0.png"); break;
			case 6: m_oE1Img = C_Loader.LoadImage("files/enemy1.png"); break;
			case 7: m_oE2Img = C_Loader.LoadImage("files/enemy2.png"); break;
			case 8: m_oE3Img = C_Loader.LoadImage("files/enemy3.png"); break;
			case 9: m_oE4Img = C_Loader.LoadImage("files/enemy4.png"); break;
			case 10: m_oE5Img = C_Loader.LoadImage("files/enemy5.png"); break;
			case 11: m_oE6Img = C_Loader.LoadImage("files/enemy6.png"); break;
			case 12: m_oE7Img = C_Loader.LoadImage("files/enemy7.png"); break;
			case 13: m_oDoorImg = C_Loader.LoadImage("files/door.png"); break;
			case 14: m_oCargoFuel = C_Loader.LoadImage("files/cargo_fuel.png"); break;
			case 15: m_oLapsTime = C_Loader.LoadImage("files/laps_time.png"); break;
			
			//load sounds
			case 16: m_oSound.LoadSound("files/bulletfire.wav", 0); break;
			case 17: m_oSound.LoadSound("files/explosion.wav", 1); break;
			case 18: m_oSound.LoadSound("files/engine.wav", 2); break;
			case 19: m_oSound.LoadSound("files/landonzone.wav", 3); break;
			case 20: m_oSound.LoadSound("files/bullethit.wav", 4); break;
			case 21: m_oSound.LoadSound("files/cp.wav", 5); break;
			case 22: m_oSound.LoadSound("files/cp_end.wav", 6); break;
			case 23: m_oSound.LoadSound("files/bounceonzone.wav", 7); break;
			case 24: m_oSound.LoadSound("files/loadcargo.wav", 8); break;
			case 25: m_oSound.LoadSound("files/unloadcargo.wav", 9); break;
		}
		m_iCurrentFile++;
		m_oLoaderStatus.IncProgress();

		return (m_iCurrentFile>=m_iNumFiles);
	}
	
	public void update(Graphics g)
	{
		double dCurTime = m_oTimer.GetInterval();
		if(dCurTime>500) dCurTime = 500; //limit frametime to protect against huge lags

	 	m_oKeyHandler.UpdateTimer(dCurTime);

		// update and draw frame
		switch(m_iState) {
			case -1: //preload files
				if(LoadAll()) {
					OwnInitAfterLoad();
					m_iState++;
				}
				// draw image on the screen ("flip")
				g.drawImage(m_oDBImage, 0, 0, this);
				break;
			case 0:  //game init
				if(m_oReplay==null || !m_oReplay.m_bLoaded) {
					m_iState = 2; //error
					break;
				}
				m_bPause = false;
				Rectangle stDrawRect = new Rectangle(0, 0, this.getSize().width, this.getSize().height);
				m_oMap = new C_TileMap( m_oReplay, m_oDBImage, stDrawRect);
				if(m_oMapObjects!=null) m_oMapObjects.SoundOn(false);
				m_oMapObjects = m_oMap.GetExtendedInfo();
				m_oMapObjects.InitPlayersAndSetMap(m_oMap);
				m_dTimeLeft = 0;
				m_dTimeToEnd = 0;
				m_bEnd = false;
				m_oReplay.ResetBeforePlay();
				m_oTimer.Reset(); //it can take some time to read the map, make sure to reset timer or a huge leap will be the result the 1st update
				m_iState++;
				break;
			case 1:  //game loop
				if(!m_bPause) {
					m_dTimeLeft += dCurTime;
					if(m_bEnd) m_dTimeToEnd += dCurTime;

					while(m_dTimeLeft>=4.0) {
						S_ReplayAction stAction;
						do {
							stAction = m_oReplay.Get(-1, 2);
							if(stAction!=null) {
								S_NetMsg2 stMsg2 = stAction.stMsg2;
								switch(stMsg2.iType) {
									case S_NetMsg2.N_MSG2_NUMKILLS:
										break;
									case S_NetMsg2.N_MSG2_CARGO_R:
										m_oMapObjects.HandleCargoReply(stMsg2);
										break;
									case S_NetMsg2.N_MSG2_CARGORESPAWN:
										m_oMapObjects.HandleCargoRespawn(stMsg2);
										break;
									case S_NetMsg2.N_MSG2_GAMEOVER:
										m_bEnd = true;
										break;
								}
							}
						} while(stAction!=null);
						
						if(m_dTimeToEnd>6500) m_iState = 0; //restart
						
						m_oMapObjects.Update(4.0);
						m_dTimeLeft -= 4.0;
						m_oReplay.IncTimeSlot();
					}
					m_oMapObjects.Draw();
					// draw image on the screen ("flip")
					g.drawImage(m_oDBImage, 0, 0, this);
				}
				break;
			case 2: //error
				m_oDBImage.getGraphics().setColor(Color.RED);
				m_oDBImage.getGraphics().drawString("Error: Bad replay", 5, this.getSize().height-32);
				// draw image on the screen ("flip")
				g.drawImage(m_oDBImage, 0, 0, this);
				break;
		}
	}

	//KEYBOARD
	public void keyPressed(KeyEvent e)
	{
		int iKey = e.getKeyCode();
		m_oKeyHandler.UpdateKeyStatus(iKey, true);
		//handle event direct here
		switch(iKey) {
			case KeyEvent.VK_ESCAPE: 
				m_iState = 0;
				break; //restart regardless of current state
			case KeyEvent.VK_P:
				if(m_iState == 1) {
					m_oMapObjects.SoundOn(m_bPause);
					m_bPause = !m_bPause;
				}
				break;
			case KeyEvent.VK_S:
			case KeyEvent.VK_N:
				m_bSound = !m_bSound;
				m_oSound.SoundOff(!m_bSound);
				break;
		}
	}
	public void keyReleased(KeyEvent e)
	{
		m_oKeyHandler.UpdateKeyStatus(e.getKeyCode(), false);
	}
	public void keyTyped(KeyEvent e) { }	
	
	public void start()
	{ 
		if(m_oTheThread == null) {
			m_oTheThread = new Thread(this); 
			m_oTheThread.start();
		}
	}

	public void stop()
	{
		if(m_oTheThread != null) {
			m_oTheThread = null;
		}
	}
	
	public void run()
	{ 
		// lower ThreadPriority 
		Thread.currentThread().setPriority(Thread.MIN_PRIORITY);

		// run forever
		Thread thisThread = Thread.currentThread();
		while(m_oTheThread == thisThread) {
			// repaint the applet
			double dCurTime = m_oTimer.PeekInterval();
			if(dCurTime>500) dCurTime = 500; //limit frametime to protect against huge lags

			if(dCurTime>10.0) repaint(); //at most redraw every 10 ms
			if(dCurTime>10.0) m_oSound.Update(dCurTime);

			try { 
				// Stop thread for 5 milliseconds
				Thread.sleep(5);
			} catch(InterruptedException e) { }
			
			// set ThreadPriority to maximum value
			Thread.currentThread().setPriority(Thread.MAX_PRIORITY);
		} 
	}
}
