// C_Font

//bitmap font class (use fnt_comp to create a font).
//supports only RGBA .fnt files

part of web_replay2;

class S_CharHeader
{
  int iIndex = 0;    //ASCII value
  int iPreX = 0;     //char position modifier
  int iPostX = 0;    //char space after
  S_Rect stRect = new S_Rect(0,0,0,0); //char x,y,w,h (pixels)
}

class S_LineInfo
{
  int iNumChars = 0;
  int iWidth = 0;
}

class S_TextSizeInfo
{
  int iWidth;
  int iHeight;
  //int iActiveY;
  //int iActiveHeight;
  int iNumLines;
  List<S_LineInfo> astLineInfo;
}

class C_Font
{
  C_Graph _m_oGraph;

  //static for each font
  int       _m_iFontHeight;
  int       _m_iNumChars;
  List<S_CharHeader> _m_astChars = new List<S_CharHeader>(256); //max index is 255, so we only support ASCII

  //user changeable
  Vector4   _m_stColor; //color used when printing to screen
  int       _m_iSpaceX; //space between chars (pixels)
  int       _m_iSpaceY; //space between lines (pixels)
  List<int> _m_pTabList; //tablist, el 0 = size, el 1..size = tab positions

  int       _m_iState = 0;
  bool      _m_bFntLoaded;

  C_Image        _m_oFntImg = new C_Image();
  S_TextSizeInfo _m_tmpTextSizeInfo = new S_TextSizeInfo();

  bool IsLoaded()
  {
    return (_m_bFntLoaded && _m_oFntImg.isLoaded);
  }

  //set default data
  C_Font(C_Graph i_oGraph)
  {
    _m_oGraph = i_oGraph;
    _m_iSpaceY = 0;       //extra pixels between rows
    _m_iSpaceX = 0;       //extra pixels between chars
    _m_iFontHeight = 0;
    _m_iNumChars = 0;

    _m_bFntLoaded = false; //no font description loaded yet

    _m_pTabList = null;

    SetColor(new Vector4(1.0, 1.0, 1.0, 1.0)); //white
  }

  void _onFntLoaded(ByteBuffer i_pMem)
  {
    _m_iState = 2;

    int i;

    ByteData pMemBD = new ByteData.view(i_pMem);

    //header
    if(pMemBD.getUint8(0) != 0x4E || pMemBD.getUint8(1) != 0x46) { //magic N, F
      _m_iState = 3;
      return;
    }
    _m_iFontHeight = pMemBD.getUint16(2, Endianness.LITTLE_ENDIAN);
    _m_iNumChars   = pMemBD.getUint16(4, Endianness.LITTLE_ENDIAN);
    _m_iSpaceY     = pMemBD.getInt16(6, Endianness.LITTLE_ENDIAN);

    //read files (info about font and chars, and font image)
    for(i=0; i<_m_iNumChars; i++) {
      S_CharHeader stChar = new S_CharHeader();
      stChar.iIndex = pMemBD.getUint16(8+0+i*24, Endianness.LITTLE_ENDIAN);
      stChar.iPreX = pMemBD.getInt16(8+2+i*24, Endianness.LITTLE_ENDIAN);
      stChar.iPostX = pMemBD.getInt16(8+4+i*24, Endianness.LITTLE_ENDIAN);
      //stChar.iReserved = pMemBD.getInt16(8+6+i*24, Endianness.LITTLE_ENDIAN);
      stChar.stRect.x = pMemBD.getInt32(8+8+i*24, Endianness.LITTLE_ENDIAN);
      stChar.stRect.y = pMemBD.getInt32(8+12+i*24, Endianness.LITTLE_ENDIAN);
      stChar.stRect.width = pMemBD.getInt32(8+16+i*24, Endianness.LITTLE_ENDIAN);
      stChar.stRect.height = pMemBD.getInt32(8+20+i*24, Endianness.LITTLE_ENDIAN);
      _m_astChars[stChar.iIndex] = stChar;
    }

    _m_bFntLoaded = true;
  }

  //loads a fontfile of the format fnt
  //check bIsLoaded later
  void Load(String i_szFontNameBase)
  {
    _m_iState = 1;

    String szFullFntName = i_szFontNameBase + ".fnt";
    String szFullPngName = i_szFontNameBase + ".png";

    //load .fnt file
    HttpRequest.request(szFullFntName, method: 'GET', responseType: 'arraybuffer')
      .then((oBuffer) {   // .success
        _onFntLoaded(oBuffer.response);
      })
      .catchError((Error error) {
        print(error.toString());
        _m_iState = 3;
      });

    //load .tga file
    _m_oGraph.LoadTextureFromImageSrc(_m_oFntImg, szFullPngName);
  }

  //set tab positions, 1st int is the total list length
  void SetTablist(List<int> i_pTabList)
  {
    _m_pTabList = i_pTabList;
  }

  //set font color (only used when printing to screen)
  void SetColor(Vector4 i_stColor)
  {
    _m_stColor = i_stColor;
  }

  //get text dimensions
  S_TextSizeInfo GetTextSize(String i_szText, bool i_bLineInfo)
  {
    int iNumLines = 1, iLineLength;
    int c, i, iW = 0, iH = 0;

    if(!IsLoaded()) return null;

    //scan text for numlines
    for(i=0; i<i_szText.length; i++) {
      c = i_szText.codeUnitAt(i);
      if(c == '\t') i++;
      else if(c == '\n') iNumLines++;
    }

    //get numcharsinline[] textlinewidth[] and maxtextwidth from text
    int iBeginY = -1, iEndY = -1;
    int iPos = 0;
    List<S_LineInfo> astLine = null;
    if(i_bLineInfo) astLine = new List<S_LineInfo>(iNumLines);
    for(i=0; i<iNumLines; i++) {
      iLineLength = 0;
      int iLiNumChars = 0;
      int iLiWidth = 0;
      c = 0;
      while(iPos<i_szText.length && c!='\n') {
        c = i_szText.codeUnitAt(iPos);
        if(c == '\t') {
          iPos++;
          if(_m_pTabList!=null) iLiWidth = i_szText.codeUnitAt(iPos);
        } else {
          if(iLiWidth+_m_astChars[c].iPreX>=0) iLiWidth += _m_astChars[c].iPreX;
          iLiWidth += _m_astChars[c].iPostX+_m_astChars[c].stRect.width + _m_iSpaceX;
        }
        if(iLiWidth > iLineLength) iLineLength = iLiWidth;
        iLiNumChars++;
        iPos++;
      }
      iLiWidth = iLineLength;
      if(iLiWidth != 0) iLiWidth -= _m_iSpaceX;
      if(iLiWidth > iW) iW = iLiWidth;
      iPos++; //skip end of line (13) (or 0 on the last line)

      if(i_bLineInfo) {
        astLine[i] = new S_LineInfo();
        astLine[i].iNumChars = iLiNumChars;
        astLine[i].iWidth = iLiWidth;
      }
    }
    iH = iNumLines * (_m_iFontHeight + _m_iSpaceY) - _m_iSpaceY;
    if(_m_iSpaceY<0) iH -= _m_iSpaceY;

    _m_tmpTextSizeInfo.iWidth = iW;
    _m_tmpTextSizeInfo.iHeight = iH;
    _m_tmpTextSizeInfo.iNumLines = iNumLines;
    _m_tmpTextSizeInfo.astLineInfo = astLine;
    return _m_tmpTextSizeInfo;
  }

  //prints out a text with the loaded font to the given image
  // within given area, centred if wanted
  S_Rect _stTmpArea = new S_Rect(0, 0, 0, 0);
  int PutString(S_Rect i_pstArea, bool i_bCentered, String i_szText)
  {
    int         i;

    _stTmpArea.Set2(i_pstArea);

    //check if parameters are valid
    if(!IsLoaded()) return 0;
    S_TextSizeInfo stTextSize = GetTextSize(i_szText, true);

    if(_stTmpArea.width==0) {
      _stTmpArea.width = stTextSize.iWidth;
      _stTmpArea.height = stTextSize.iHeight;
    }
    if(stTextSize.iHeight > _stTmpArea.height || stTextSize.iWidth > _stTmpArea.width)
      return 0;

    //put lines on image
    int iPos = 0;
    int iYStart = _stTmpArea.y, iXStart = _stTmpArea.x;
    if(i_bCentered) iYStart += (_stTmpArea.height - stTextSize.iHeight)~/2;
    for(i=0; i<stTextSize.iNumLines; i++) {
      if(i_bCentered) iXStart += (_stTmpArea.width - stTextSize.astLineInfo[i].iWidth)~/2;

      //put line on image
      int c = 0;
      while(iPos<i_szText.length && c!='\n') {
        c = i_szText.codeUnitAt(iPos);
        iXStart += _m_astChars[c].iPreX; //can be negative
        if(iXStart<_stTmpArea.x) iXStart = _stTmpArea.x; //fix if negative
        _m_oGraph.DrawImage2D_Blend(_m_oFntImg, _m_astChars[c].stRect, iXStart, iYStart, _m_stColor);
        iXStart += _m_astChars[c].stRect.width+_m_astChars[c].iPostX+_m_iSpaceX;
        iPos++;
      }
      iPos++;
      iXStart = _stTmpArea.x;
      iYStart += _m_iFontHeight+_m_iSpaceY;
  	}

    return stTextSize.iWidth;
  }

}
