
part of web_replay2;


abstract class C_PlayerBase extends C_Object
{
  static final double HEARING_RANGE = 2200.0; //maxdistance sound is played at all
  static final double PANNING_RANGE = 500.0; //maxdistance sound is played in both speakers
  static final int EXPLOSIONSIZE = 56;
  //explosion gfx
  static C_Image _s_oExplosionImg;
  static List<S_Rect> _s_astExplosionRects = new List<S_Rect>(16);
  double  _m_dFrameNr;
  double _m_dVol, _m_dPan;

  static void InitExplosionImage(C_Image i_oExplosionImg)
  {
    int i, j;

    _s_oExplosionImg = i_oExplosionImg;
    for(j=0; j<2; j++) {
      for(i=0; i<8; i++) {
        _s_astExplosionRects[j*8+i] = new S_Rect(i*EXPLOSIONSIZE, j*EXPLOSIONSIZE, EXPLOSIONSIZE, EXPLOSIONSIZE);
      }
    }
  }

  C_PlayerBase()
    : super()
  {
    _m_dFrameNr = 0.0;

    _m_dVol = 1.0;
    _m_dPan = 0.0;
  }

  static C_Object _s_oListner = null;
  static void SetListner(C_Object i_oListner)
  {
    _s_oListner = i_oListner;
  }

  bool UpdateExplosion(double i_dTime)
  {
    if(i_dTime==0.0) _m_dFrameNr = 0.0; //reset
    _m_dFrameNr += i_dTime*0.018; //18 fps anim

    bool bReady = (_m_dFrameNr > 15.0);
    return bReady;
  }

  void DrawExplosion(int i_iX, int i_iY)
  {
    if(_m_dFrameNr<16.0) {
      S_Rect stRect = _s_astExplosionRects[_m_dFrameNr.toInt()];

      //center explosion
      i_iX -= (GetSizeX()~/2) - (EXPLOSIONSIZE~/2);
      i_iY -= (GetSizeY()~/2) - (EXPLOSIONSIZE~/2);

      ODraw(_s_oExplosionImg, stRect, i_iX, i_iY);
    }
  }

  void CalcSoundVolPan(C_Object i_oListner)
  {
    double dDistX = (GetCenterPosX() - i_oListner.GetCenterPosX()).toDouble();
    double dDistY = (GetCenterPosY() - i_oListner.GetCenterPosY()).toDouble();
    double dDist = sqrt(dDistX*dDistX + dDistY*dDistY);

    //volume
    _m_dVol = 1.0 - dDist/(HEARING_RANGE/1.0);
    if(_m_dVol<0) _m_dVol = 0.0;
    //panning
    _m_dPan = dDistX/(PANNING_RANGE/1.0);
  }

  void GetCommonInfo(S_NetMsg1 i_stMsg1)
  {
    _m_dAcceleration = i_stMsg1.fAcceleration;
    _m_dDirection = i_stMsg1.fDirection;
    m_dPosX = i_stMsg1.fPosX;
    m_dPosY = i_stMsg1.fPosY;
    _m_dVelX = i_stMsg1.fVelX;
    _m_dVelY = i_stMsg1.fVelY;

    _m_dGravityY = i_stMsg1.fGravityY;
    _m_bCanCollide = i_stMsg1.bCanCollide;
  }

}

/////////////////////////////////////////////////////////////////////////////////////////
class C_Player extends C_PlayerBase
{
  static final double DEFAULT_SHIPGRAVITYBASEX = 0.0;   //pixel/second 2
  static final double DEFAULT_SHIPGRAVITYBASEY = 70.0;  //pixel/second 2
  static final double DEFAULT_SHIPRESISTANCEX = 0.68; //constant (velocity dependent)
  static final double DEFAULT_SHIPRESISTANCEY = 0.43;

  static double SHIPGRAVITYBASEX = 0.0;   //pixel/second 2
  static double SHIPGRAVITYBASEY = 70.0;  //pixel/second 2
  static double SHIPRESISTANCEX = 0.68; //constant (velocity dependent)
  static double SHIPRESISTANCEY = 0.43; //^
  static final int SHIPSTEARSPEED  = 235;  //degree/second
  static final int SHIPTHRUST      = 270;  //pixel/second 2
  static final int BULLETBASEVEL = 220;   //pixel/second
  static final int BULLETFREETIME = 3100; //millisec to be free from bullets when just come alive

  //code to remove clicks/pops when stopping looping audio
  //reduce volume until samples are close to 0 before stopping
  static List<C_SoundFxPlayingHandle> _s_oFadeOutList = new List<C_SoundFxPlayingHandle>();
  static void UpdateFadeOut(double i_dTime)
  {
    int i;
    for(i=0; i<_s_oFadeOutList.length; i++) {
      _s_oFadeOutList[i].dTime += i_dTime;
      if(_s_oFadeOutList[i].dTime>20.0) { //20ms
        _s_oFadeOutList[i].dTime -= 20.0;
        if(_s_oFadeOutList[i].oVolume.gain.value>0.1) {
          _s_oFadeOutList[i].oVolume.gain.value *= 0.75;
        } else {
          _s_oFadeOutList[i].oNode.stop(0);
          _s_oFadeOutList.removeAt(i);
          i--;
        }
      }
    }
  }
  
  C_TileMap    _m_oMap;
  C_BulletList _m_oBulletList;
  C_Replay     _m_oReplay;

  static final int _SHIPSIZE = 48;
  C_Image     _m_oShipImg;
  Vector4     _m_oColor;
  List<S_Rect> _m_astShipRects;

  int _m_iStartPosX, _m_iStartPosY;

  int _m_iKeyFlag;
  int _m_iNumLifes; //-1 = no limit
  bool _m_bAlive;
  bool _m_bFreeFromBullets;
  double _m_dFreeFromBulletsTimer;

  C_SoundFxPlayingHandle _m_iEngineSoundCh;
  bool _m_bLanded;
  double _m_dLandTime = 0.0;

  //cargo variables
  static final int MAXSPACEINHOLDWEIGHT = 50;
  static final int MAXSPACEINHOLDUNITS = 3;
  int _m_iSpaceUsed = 0;
  int _m_iNumUsed = 0;
  List<int> _m_aiHold = new List<int>(MAXSPACEINHOLDUNITS);

  //score variables
  int m_iScore = 0;

  //race variables
  int m_iCurCP = 0;
  int m_iCurLap = 0;
  double m_dBestLapTime = 0.0;
  String m_szName;

  //fuel
  double _m_dMaxFuelInTank = 60.0;
  double _m_dFuel = 60.0; //number of seconds one can use the thruster
  void AddFuel(double dFuel)
  {
    if(_m_dFuel<_m_dMaxFuelInTank) {
      _m_dFuel += dFuel;
      if(_m_dFuel>_m_dMaxFuelInTank) _m_dFuel=_m_dMaxFuelInTank;
    }
  }
  double FuelPercentLeft()
  {
    return _m_dFuel/_m_dMaxFuelInTank;
  }
  void ConsumeFuel(double i_dTime)
  {
    if(_m_oMap.MapType()!=3) {
      //consume fuel
      if(_m_dAcceleration!=0) _m_dFuel -= i_dTime/1000.0;
      if(_m_dFuel<0) {
        _m_dFuel = 0.0;
        //Kill(); //no thrust possible instead... it usually leads to death
      }
    }
  }

  C_Player(int i_iId, String i_szName, C_Image i_oShipImg, Vector4 i_oColor, C_Replay i_oReplay, C_TileMap i_oMap, C_BulletList i_oBulletList, int i_iStartX, int i_iStartY, int i_iNumLives)
    : super()
  {
    //copy parameters and init base
    m_iOwnerID = i_iId;
    _m_oMap = i_oMap;
    _m_oBulletList = i_oBulletList;
    _m_oReplay = i_oReplay;
    _m_iKeyFlag = 0;
    m_szName = i_szName;

    //init ship src image info
    _m_oImage = _m_oShipImg = i_oShipImg;
    _m_oColor = i_oColor;
    _m_astShipRects = new List<S_Rect>(8*9);
    int i, j;
    for(j=0; j<4; j++) {       //ship
      for(i=0; i<9; i++) {
        _m_astShipRects[(j*9+i)] = new S_Rect(i*_SHIPSIZE, ((j+1)%4)*_SHIPSIZE, _SHIPSIZE, _SHIPSIZE);
      }
    }
    for(j=4; j<8; j++) {       //thrust
      for(i=0; i<9; i++) {
        _m_astShipRects[(j*9+i)] = new S_Rect(i*_SHIPSIZE, (4+((j+1)%4))*_SHIPSIZE, _SHIPSIZE, _SHIPSIZE);
      }
    }
    _m_stScreenArea = _m_oMap.GetScreenRect();

    //init others
    _m_iNumLifes = i_iNumLives;
    m_iScore = 0;
    _m_iStartPosX = i_iStartX;
    _m_iStartPosY = i_iStartY;
    m_dPosX = _m_iStartPosX.toDouble();
    m_dPosY = _m_iStartPosY.toDouble();
    _m_dVelX = _m_dVelY = _m_dAcceleration = 0.0;
    _m_dDirection = 270.0;

    _m_dGravityX = SHIPGRAVITYBASEX;
    _m_dGravityY = SHIPGRAVITYBASEY;
    _m_dResistanceX = SHIPRESISTANCEX;
    _m_dResistanceY = SHIPRESISTANCEY;

    _m_bAlive = true;
    _m_bLanded = false;
    _m_bFreeFromBullets = false;
    _m_dFreeFromBulletsTimer = 0.0;
  }

  int Update(double i_dTime)
  {
    double  dSTime = (i_dTime/1000);
    int     iRval = -1;

    _m_dLandTime += i_dTime;

    if(_m_bFreeFromBullets) {
      _m_dFreeFromBulletsTimer += i_dTime;
      if(_m_dFreeFromBulletsTimer > BULLETFREETIME) {
        _m_bFreeFromBullets = false;
      }
    }

    //maybe come to life again
    CalcSoundVolPan(C_PlayerBase._s_oListner);
    if(!_m_bAlive) {
      if( UpdateExplosion(i_dTime) ) { //set alive again
        if(_m_iNumLifes != 0 ) {
          _m_bAlive = true;
          _m_bFreeFromBullets = true;
          _m_dFreeFromBulletsTimer = 0.0;
          m_dPosX = _m_iStartPosX.toDouble();
          m_dPosY = _m_iStartPosY.toDouble();
          _m_dDirection = 270.0;
          _m_dGravityX  = SHIPGRAVITYBASEX;
          _m_dGravityY  = SHIPGRAVITYBASEY;
          _m_dFuel = _m_dMaxFuelInTank;
          _m_bLanded = false;
        }
      }
    }

    //always
    S_NetMsg1 stMsg1;
    S_ReplayAction stAction;
    while( (stAction=_m_oReplay.Get(m_iOwnerID, 1))!=null ) {
      stMsg1 = stAction.stMsg1;
      //movement info
      if((stMsg1.iMsg & S_NetMsg1.N_MSG1_MOVEMENT)!=0) {
        if(stMsg1.fAcceleration!=_m_dAcceleration) {
          if(_m_dAcceleration>0) {
            if(_m_iEngineSoundCh!=null) {
              //_m_iEngineSoundCh.oNode.stop(0);
              _m_iEngineSoundCh.oVolume.gain.value *= 0.7;
              _s_oFadeOutList.add(_m_iEngineSoundCh);
              _m_iEngineSoundCh = null; 
            }
          } else _m_iEngineSoundCh = C_Sound.s_oSound.PlaySound(2, _m_dVol, _m_dPan, true);
        }

        GetCommonInfo(stMsg1);
        _m_iSpaceUsed = stMsg1.iCargoInHold;
      }
      //keystate change
      if( (stMsg1.iMsg & S_NetMsg1.N_MSG1_KEYCHANGE)!=0 ) _m_iKeyFlag = stMsg1.iKeyFlag;

      //bullet fired
      if( (stMsg1.iMsg & S_NetMsg1.N_MSG1_NEWBULLET)!=0) {
        C_Bullet oBullet = _CreateBullet();

        oBullet.SetUId(stMsg1.iBulletUniqeValBase);
        _m_oBulletList.Add(oBullet);
        C_Sound.s_oSound.PlaySound(0, _m_dVol, _m_dPan, false);
      }

      //player killed
      if( (stMsg1.iMsg & S_NetMsg1.N_MSG1_KILLED)!=0 ) {
        switch(stMsg1.iReason_Hits_Door) {
          case 1: //self
            Kill();
            break;
          case 2: //bullet
            Kill();
            _m_oBulletList.Remove(stMsg1.iCauseID);
            iRval = (stMsg1.iCauseID & 0xff);
            break;
          case 3: //player
            iRval = stMsg1.iCauseID + 256;
            Kill();
            break;
        }
      }

      //score change
      if( (stMsg1.iMsg & S_NetMsg1.N_MSG1_SCORE)!=0 ) m_iScore = stMsg1.iScore;

      //player landed
      if( (stMsg1.iMsg & S_NetMsg1.N_MSG1_LANDED)!=0 ) LandOnZone();

      if( (stMsg1.iMsg & S_NetMsg1.N_MSG1_LAPINFO)!=0 ) {
        m_iCurLap = stMsg1.iLap;
        m_iCurCP  = stMsg1.iCP;
        m_dBestLapTime = stMsg1.fBestLapTime;
      }
    }

    //make the explosion be still even if messages lags behind
    if(!_m_bAlive) {
      _m_dVelX = 0.0;
      _m_dVelY = 0.0;
    }

    //lifted from landingzone?
    if(_m_dAcceleration!=0.0 && _m_bLanded) {
      _m_bLanded = false;
      ResetGravity();
    }

    //steering
    {
      if( (_m_iKeyFlag & S_NetMsg1.GF_KEY_DOWN)!=0 ) {
        if(_m_dDirection!=270.0) {
          //get way to turn
          int iSign = 1, iSignAfter = 1;
          if(_m_dDirection<90 || _m_dDirection>270) iSign = -1;
          _m_dDirection += iSign*(SHIPSTEARSPEED/3*dSTime);
          if(_m_dDirection<90 || _m_dDirection>270) iSignAfter = -1;
          if(iSign!=iSignAfter) _m_dDirection = 270.0;
        }
      } else { //normal steering
        _m_dDirection +=(_m_iKeyFlag & S_NetMsg1.GF_KEY_RIGHT)!=0 ? SHIPSTEARSPEED*dSTime:0;
        _m_dDirection -=(_m_iKeyFlag & S_NetMsg1.GF_KEY_LEFT)!=0 ? SHIPSTEARSPEED*dSTime:0;
      }
      //clipping
      int iNumCounts = 0;
      if(_m_dDirection<0) iNumCounts = (_m_dDirection~/360)-1;
      else iNumCounts = (_m_dDirection~/360);
      _m_dDirection-=iNumCounts*360;
    }

    ConsumeFuel(i_dTime);
    OUpdate(i_dTime);

    if(_m_dAcceleration!=0.0) { //ship is moving, set current engine volume
      if(_m_iEngineSoundCh!=null) {
        _m_iEngineSoundCh.oVolume.gain.value = _m_dVol;
        _m_iEngineSoundCh.oPanner.setPosition(_m_dPan, 0.0, 1.0-_m_dPan.abs());
      }
    }

    _m_stImageRect = _m_astShipRects[_m_iRealDir];
    return iRval;
  }

  void Kill()
  {
    if(!_m_bAlive) return;

    _m_bAlive = false;
    _m_bCanCollide = false;
    UpdateExplosion(0.0); //reset explosion

    //to prevent movement during explosion
    _m_dGravityX = 0.0;
    _m_dGravityY = 0.0;

    //take a life if not unlimited
    if(_m_iNumLifes!=-1) _m_iNumLifes--;

    _Stop();
    C_Sound.s_oSound.PlaySound(1, _m_dVol, _m_dPan, false); //play explosion sound
  }

  //to be able to stop the sound while pause, or when escape is pressed
  void SoundOn(bool i_bSound)
  {
    if(!i_bSound && _m_dAcceleration!=0) {
      if(_m_iEngineSoundCh!=null) _m_iEngineSoundCh.oNode.stop(0); //off
    } else if(_m_dAcceleration!=0) _m_iEngineSoundCh = C_Sound.s_oSound.PlaySound(2, _m_dVol, _m_dPan, true); //on
  }

  void _Stop()
  {
    _m_dVelX = 0.0; _m_dVelY = 0.0;
    if(_m_dAcceleration!=0.0 && _m_iEngineSoundCh!=null) _m_iEngineSoundCh.oNode.stop(0); //stop engine sound
    _m_dAcceleration = 0.0;
  }

  void LandOnZone()
  {
    _m_bLanded = true;
    _m_dGravityX = 0.0;
    _m_dGravityY = 0.0;
    C_Sound.s_oSound.PlaySound(3, _m_dVol, _m_dPan, false);
  }

  void Draw()
  {
    int iMapX = _m_oMap.GetPosX();
    int iMapY = _m_oMap.GetPosY();
    _m_stScreenArea = _m_oMap.GetScreenRect();

    if(_m_bAlive) {
      if(_m_bFreeFromBullets && (_m_dFreeFromBulletsTimer.toInt() % 300 > 200)) _m_oColor.a = 0.5; //transparent
      else _m_oColor.a = 1.0;

      //draw thrust
      if(_m_dAcceleration>0) ODraw(_m_oShipImg, _m_astShipRects[_m_iRealDir+36], iMapX, iMapY);
      //draw ship
      ODraw(_m_oShipImg, _m_astShipRects[_m_iRealDir], iMapX, iMapY, _m_oColor);
    } else {
      DrawExplosion(iMapX, iMapY);
    }
  }

  C_Bullet _CreateBullet()
  {
    double dTemp = _m_iRealDir*(_m_iDirStep*(PI/180));
    double dSin = sin(dTemp);
    double dCos = cos(dTemp);
    double dPosX = m_dPosX + _SHIPSIZE/2 + dCos * (_SHIPSIZE/2-2);
    double dPosY = m_dPosY + _SHIPSIZE/2 + dSin * (_SHIPSIZE/2-2);
    double dVelX = _m_dVelX + dCos * BULLETBASEVEL;
    double dVelY = _m_dVelY + dSin * BULLETBASEVEL;
    double dDirection = (_m_iRealDir * _m_iDirStep).toDouble();

    return new C_Bullet(_m_oMap, dPosX, dPosY, dVelX, dVelY, dDirection, m_iOwnerID);
  }

  bool HoldEmpty()
  {
    return (_m_iNumUsed==0);
  }
  int CargoWeightInHold()
  {
    return _m_iSpaceUsed;
  }
  int CargoNumInHold()
  {
    return _m_iNumUsed; //this is currently always 0 (no support in network messages)
  }
  bool IsLanded()
  {
    return _m_bLanded;
  }
  bool IsLandedForTime(double i_dTime)
  {
    return (_m_bLanded && (_m_dLandTime>i_dTime));
  }
  bool IsAlive()
  {
    return _m_bAlive;
  }
  bool LoadCargo(int iCargo)
  {
    bool bCargoLoaded = false;
    if(_m_iNumUsed<MAXSPACEINHOLDUNITS && (_m_iSpaceUsed+iCargo)<=MAXSPACEINHOLDWEIGHT) {
      //space in hold, add cargo
      _m_aiHold[_m_iNumUsed++] = iCargo;
      _m_iSpaceUsed += iCargo;
      bCargoLoaded = true;

      //calc new gravity is done when lifted from the zone, not here

      //play load cargo sound
      C_Sound.s_oSound.PlaySound(8, _m_dVol, _m_dPan, false);
    }

    return bCargoLoaded;
  }
  void UnloadCargo()
  {
    if(_m_iNumUsed>0) {
      _m_iNumUsed--;
      _m_iSpaceUsed -= _m_aiHold[_m_iNumUsed];

      //calc new gravity is done when lifted from the zone, not here

      //add score for the cargo moved
      m_iScore += _m_aiHold[_m_iNumUsed];

      //play unload cargo sound
      C_Sound.s_oSound.PlaySound(9, _m_dVol, _m_dPan, false);

      //reset landtime just to make the delay before a level ends the same every time
      _m_dLandTime = 0.0;
    }
  }
  void ResetGravity()
  {
    _m_dGravityX = SHIPGRAVITYBASEX;
    _m_dGravityY = SHIPGRAVITYBASEY + (_m_iSpaceUsed*2);
  }

  bool FreeFromBullets()
  {
    return _m_bFreeFromBullets;
  }

  int GetNumLives()
  {
    return _m_iNumLifes;
  }
  void AddLife()
  {
    if(_m_iNumLifes!=-1) _m_iNumLifes++; //if mission
  }
}
