
part of web_replay2;


class S_ReplayHeader
{
  int iVersion; //1
  String szMapFile = "";
  int iLevelType;
  bool bNoShooting, bFlyThrough, bNoBulletsHit;

  int iNumPlayers;
  List<String> szPlayerName = ["", "", "", "", "", "", "", ""];
}

class S_ReplayAction
{
  int iTimeSlot;

  int iId;   //player id (-1 for server), also in the messages below
  int iMsg; //1 or 2

  S_NetMsg1 stMsg1 = new S_NetMsg1();
  S_NetMsg2 stMsg2 = new S_NetMsg2();
}

class C_Replay
{
  static const int N_MSG2_GAMEOVER     = 0x0E;
  static const int N_MSG2_SCORE        = 0x0F; //iScore
  static const int N_MSG2_NUMKILLS     = 0x10; //iNumKills(server) or iKilledBy(client)
  static const int N_MSG2_NUMLAPS      = 0x11; //iNumLaps
  static const int N_MSG2_CARGO        = 0x20; //iZoneId, iCargoNum
  static const int N_MSG2_CARGO_R      = 0x21; //iZoneId, iCargoNum, iCargoReqAccepted
  static const int N_MSG2_NEXTLEVEL    = 0x22; //
  static const int N_MSG2_CARGORESPAWN = 0x23; //iZoneId, iCargoNum, iCargoWeight
  static const int N_MSG2_LIFE_R       = 0x25; //iZoneId, iLifeReqAccepted

  int _m_iCurTimeSlot;
  S_ReplayHeader _m_stHeader;
  List<S_ReplayAction> _m_oActions;
  List<int> _m_aiCurGetPos = new Int32List(129*2);
  bool m_bLoaded;

  C_Replay()
  {
    _m_stHeader = new S_ReplayHeader();
    _m_oActions = new List<S_ReplayAction>();
    Reset();
  }

  void ResetBeforePlay()
  {
    for(int i=0; i<_m_aiCurGetPos.length; i++) _m_aiCurGetPos[i] = 0;
    _m_iCurTimeSlot = 0;
   }

  void Reset()
  {
    for(int i=0; i<_m_aiCurGetPos.length; i++) _m_aiCurGetPos[i] = 0;
    _m_iCurTimeSlot = 0;
    _m_oActions.clear();
    m_bLoaded = false;
  }

  void IncTimeSlot()
  {
    _m_iCurTimeSlot++;
  }

  S_ReplayHeader GetGameInfo()
  {
    return _m_stHeader;
  }

  S_ReplayAction Get(int i_iId, int i_iType)
  {
    int i, iSize = _m_oActions.length;
    int iIdIndex = i_iId+1 + 129*(i_iType-1);
    for(i=_m_aiCurGetPos[iIdIndex]; i<iSize; i++) {
      S_ReplayAction stAction = _m_oActions[i];
      if((i_iId==-1 || stAction.iId==i_iId) && stAction.iMsg==i_iType) {
        if(stAction.iTimeSlot <= _m_iCurTimeSlot) { //do it as fast as possible if <
          _m_aiCurGetPos[iIdIndex] = i+1;
          return stAction;
        } else break;
      }
    }
    return null;
  }

  bool LoadFromMem(List<int> i_pMem)
  {
    bool bResult = false;
    S_ReplayAction stAction;
    int i, j;
    Reset();

    Uint8List buf = new Uint8List.fromList(i_pMem);
    ByteData bdata = new ByteData.view(buf.buffer);

    //header
    if(i_pMem.length<160) return false;
    _m_stHeader.iVersion = bdata.getInt32(0, Endianness.LITTLE_ENDIAN);
    for(i=4; i<52; i++) {
      if(i_pMem[i]==0) break;
      _m_stHeader.szMapFile += new String.fromCharCode(i_pMem[i]); // new String(i_pMem[i]);
    }
    _m_stHeader.iLevelType = i_pMem[52];
    _m_stHeader.bNoShooting = i_pMem[56]!=0;
    _m_stHeader.bFlyThrough = i_pMem[57]!=0;
    _m_stHeader.bNoBulletsHit = i_pMem[58]!=0;
    _m_stHeader.iNumPlayers = i_pMem[60];
    for(j=0; j<8; j++) {
      for(i=64+j*12; i<64+j*12+12; i++) {
        if(i_pMem[i]==0) break;
        _m_stHeader.szPlayerName[j] += new String.fromCharCode(i_pMem[i]);
      }
    }
    if(VerifyHeader()) {
      int iTemp, iPos = 160;
      //elements
      while(iPos+64<=i_pMem.length) {
        bResult = false;
        stAction = new S_ReplayAction();

        stAction.iTimeSlot = bdata.getInt32(iPos+0, Endianness.LITTLE_ENDIAN);
        stAction.iId = i_pMem[iPos+4]; if(stAction.iId==255) stAction.iId = -1;
        stAction.iMsg = i_pMem[iPos+5];
        if(stAction.iMsg==1) {
          //stAction.stMsg1
          iTemp = bdata.getInt16(iPos+6, Endianness.LITTLE_ENDIAN); //size 58
          if(iTemp!=58) break;
          iTemp = bdata.getInt16(iPos+8, Endianness.LITTLE_ENDIAN); //type 0x50
          if(iTemp!=0x50) break;
          stAction.stMsg1.iMsg = i_pMem[iPos+10];
          stAction.stMsg1.iSenderId = i_pMem[iPos+11]; if(stAction.stMsg1.iSenderId==255) stAction.stMsg1.iSenderId = -1;
          stAction.stMsg1.fPosX = bdata.getFloat32(iPos+12, Endianness.LITTLE_ENDIAN);
          stAction.stMsg1.fPosY = bdata.getFloat32(iPos+16, Endianness.LITTLE_ENDIAN);
          stAction.stMsg1.fVelX = bdata.getFloat32(iPos+20, Endianness.LITTLE_ENDIAN);
          stAction.stMsg1.fVelY = bdata.getFloat32(iPos+24, Endianness.LITTLE_ENDIAN);
          stAction.stMsg1.fAcceleration = bdata.getFloat32(iPos+28, Endianness.LITTLE_ENDIAN);
          stAction.stMsg1.fDirection = bdata.getFloat32(iPos+32, Endianness.LITTLE_ENDIAN);
          stAction.stMsg1.fGravityY = bdata.getFloat32(iPos+36, Endianness.LITTLE_ENDIAN);
          stAction.stMsg1.iCargoInHold = bdata.getInt16(iPos+40, Endianness.LITTLE_ENDIAN);
          stAction.stMsg1.bCanCollide = i_pMem[iPos+42]!=0;
          stAction.stMsg1.iKeyFlag = i_pMem[iPos+43];
          stAction.stMsg1.iNumBullets = i_pMem[iPos+44];
          stAction.stMsg1.iBulletUniqeValBase = bdata.getInt32(iPos+45, Endianness.LITTLE_ENDIAN);
          stAction.stMsg1.iReason_Hits_Door = i_pMem[iPos+49];
          stAction.stMsg1.iCauseID = bdata.getInt32(iPos+50, Endianness.LITTLE_ENDIAN);
          stAction.stMsg1.iScore = bdata.getInt32(iPos+54, Endianness.LITTLE_ENDIAN);
          stAction.stMsg1.iLap = i_pMem[iPos+58];
          stAction.stMsg1.iCP = i_pMem[iPos+59];
          stAction.stMsg1.fBestLapTime = bdata.getFloat32(iPos+60, Endianness.LITTLE_ENDIAN);
        } else if(stAction.iMsg==2) {
          //stAction.stMsg2
          iTemp = bdata.getInt16(iPos+6, Endianness.LITTLE_ENDIAN); //size 48
          if(iTemp!=48) break;
          stAction.stMsg2.iType = bdata.getInt16(iPos+8, Endianness.LITTLE_ENDIAN);
          stAction.stMsg2.iPlayerId = bdata.getInt16(iPos+10, Endianness.LITTLE_ENDIAN);
          stAction.stMsg2.iGameId = bdata.getInt16(iPos+12, Endianness.LITTLE_ENDIAN);
          stAction.stMsg2.iInt1 = bdata.getInt32(iPos+14, Endianness.LITTLE_ENDIAN);
          stAction.stMsg2.iInt2 = bdata.getInt32(iPos+18, Endianness.LITTLE_ENDIAN);
          stAction.stMsg2.iInt3 = bdata.getInt32(iPos+22, Endianness.LITTLE_ENDIAN);
          for(i=22; i<22+24; i++) {
            if(i_pMem[i]==0) break;
            stAction.stMsg2.szLevelName += new String.fromCharCode(i_pMem[i]);
          }
          for(i=22; i<22+18; i++) {
            if(i_pMem[i]==0) break;
            stAction.stMsg2.szGameName += new String.fromCharCode(i_pMem[i]);
          }
          for(i=40; i<40+12; i++) {
            if(i_pMem[i]==0) break;
            stAction.stMsg2.szPlayerName += new String.fromCharCode(i_pMem[i]);
          }
          stAction.stMsg2.iOptions = bdata.getInt16(iPos+52, Endianness.LITTLE_ENDIAN);
        }
        if(!VerifyElement(stAction)) {
          bResult = false;
          break;
        }
        _m_oActions.add(stAction);
        iPos += 64;
        bResult = true;
      }
    } else bResult = false;
    if(!bResult) Reset();
    m_bLoaded = bResult;
    return bResult;
  }

  bool VerifyString(String i_szText)
  {
    int i;
    for(i=0; i<i_szText.length; i++) {
      if(i_szText.codeUnitAt(i)<32) {
        return false;
      }
    }
    return true;
  }

  bool VerifyHeader()
  {
    int i;
    if(_m_stHeader.iVersion!=1) return false;
    if(_m_stHeader.iNumPlayers<1 || _m_stHeader.iNumPlayers>8) return false;
    if(_m_stHeader.iLevelType<1 || _m_stHeader.iLevelType>4) return false;

    if(!VerifyString(_m_stHeader.szMapFile)) return false;
    for(i=0; i<8; i++) if(!VerifyString(_m_stHeader.szPlayerName[i])) return false;

    return true;
  }

  bool VerifyElement(S_ReplayAction i_stAction)
  {
    if(i_stAction.iMsg!=1 && i_stAction.iMsg!=2) return false;
    if(i_stAction.iTimeSlot<0) return false;
    //if(i_stAction.iId<-1) return false;

    if(i_stAction.iMsg==1) {
      //TODO verify i_pstAction->stMsg1
    } else {
      int iType = i_stAction.stMsg2.iType;
      if(iType!=N_MSG2_GAMEOVER && iType!=N_MSG2_SCORE
        && iType!=N_MSG2_NUMKILLS && iType!=N_MSG2_NUMLAPS
        && iType!=N_MSG2_CARGO_R && iType!=N_MSG2_CARGORESPAWN
        && iType!=N_MSG2_LIFE_R && iType!=N_MSG2_NEXTLEVEL)
        return false;
      //TODO further verify i_pstAction->stMsg2
    }
    return true;
  }

}
