library web_replay2;

import 'package:crypto/crypto.dart';
import 'package:vector_math/vector_math.dart';
import 'dart:html';
import 'dart:collection';
import 'dart:web_gl' as webgl;
import 'dart:typed_data';
import 'dart:math';
import 'dart:web_audio';

part 'C_Graph.dart';
part 'C_Font.dart';
part 'C_Sound.dart';

part 'C_Replay.dart';
part 'C_ReplayConnection.dart';
part 'S_NetMsg.dart';

part 'C_Object.dart';
part 'C_Viewport.dart';
part 'C_TileMap.dart';
part 'C_MapObjects.dart';
part 'C_Player.dart';
part 'C_Bullet.dart';
part 'C_Collision.dart';


class C_GameHandler
{
  C_Sound _m_oSound;
  C_Graph _m_oGraph;
  C_Font _m_oFont;

  C_ReplayConnection _m_oReplayCon;
  bool _m_bAllLoaded = false;

  double _m_dLastTime = 0.0;

  C_Image _m_oShipImg;
  C_Image _m_oExplosionImg;
  C_Image _m_oBulletImg;
  C_Image _m_oLandingZoneImg;
  C_Image _m_oCpImg;
  C_Image _m_oE0Img, _m_oE1Img, _m_oE2Img, _m_oE3Img;
  C_Image _m_oE4Img, _m_oE5Img, _m_oE6Img, _m_oE7Img;
  C_Image _m_oDoorImg;
  C_Image _m_oCargoFuel;
  C_Image _m_oLapsTime;

  C_TileMap _m_oMap;
  C_Replay _m_oReplay;
  C_Viewport _m_oViewport;
  bool _m_bPause, _m_bEnd;
  double _m_dTimeLeft, _m_dTimeToEnd;
  int _m_iState = 0;

  CanvasElement m_oCanvas;
  int _m_iOrginalCanvasW, _m_iOrginalCanvasH;

  bool _m_bHasWebGL = false;

  C_GameHandler(CanvasElement canvas)
  {
    m_oCanvas = canvas;
    _m_iOrginalCanvasW = m_oCanvas.width;
    _m_iOrginalCanvasH = m_oCanvas.height;
    
    _m_oSound = new C_Sound();
    _m_oGraph = new C_Graph();
    _m_bHasWebGL = _m_oGraph.Init(canvas); //init webgl

    _m_oReplayCon = new C_ReplayConnection();
  }

  void _loadTextures()
  {
    _m_oShipImg = new C_Image();
    _m_oExplosionImg = new C_Image();
    _m_oBulletImg = new C_Image();
    _m_oLandingZoneImg = new C_Image();
    _m_oCpImg = new C_Image();
    _m_oE0Img = new C_Image();
    _m_oE1Img = new C_Image();
    _m_oE2Img = new C_Image();
    _m_oE3Img = new C_Image();
    _m_oE4Img = new C_Image();
    _m_oE5Img = new C_Image();
    _m_oE6Img = new C_Image();
    _m_oE7Img = new C_Image();
    _m_oDoorImg = new C_Image();
    _m_oCargoFuel = new C_Image();
    _m_oLapsTime = new C_Image();

    _m_oGraph.LoadTextureFromImageSrc(_m_oShipImg, "./files/ship_w.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oExplosionImg, "./files/explosion.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oBulletImg, "./files/bullet.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oLandingZoneImg, "./files/zone.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oCpImg, "./files/cpbk.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE0Img, "./files/enemy0.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE1Img, "./files/enemy1.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE2Img, "./files/enemy2.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE3Img, "./files/enemy3.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE4Img, "./files/enemy4.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE5Img, "./files/enemy5.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE6Img, "./files/enemy6.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oE7Img, "./files/enemy7.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oDoorImg, "./files/door.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oCargoFuel, "./files/cargo_fuel.png");
    _m_oGraph.LoadTextureFromImageSrc(_m_oLapsTime, "./files/laps_time.png");

    _m_oFont = new C_Font(_m_oGraph);
    _m_oFont.Load("./files/arial-12");
  }

  void ScreenSize()
  {
    double dScale = 1.0;
    if(window.innerWidth>900) dScale = 0.75;
    if(window.innerWidth>1400) dScale = 0.5;
    //double dAspect = window.innerWidth/window.innerHeight;
    m_oCanvas.width  = (window.innerWidth*dScale).toInt();
    m_oCanvas.height = (window.innerHeight*dScale).toInt();
    m_oCanvas.style.width = (window.innerWidth-0).toInt().toString()+'px';
    m_oCanvas.style.height = (window.innerHeight-0).toInt().toString()+'px';

    _m_oGraph.Init(m_oCanvas);
    if(_m_oViewport!=null) {
      _m_oViewport.SetSize(_m_oGraph.m_iViewportWidth, _m_oGraph.m_iViewportHeight);
    }
  }

  void HandleKeyEvent(KeyboardEvent event)
  {
    KeyEvent keyEvent = new KeyEvent.wrap(event);
    switch(keyEvent.keyCode) {
      case 118: //KeyCode.V
        if(_m_iState==2) _m_oViewport.ReplayFollow(true);
        break;
      case 110: //KeyCode.N
      case 115: //KeyCode.S
        C_Sound.s_oSound.ToggleOnOff();
        break;
      case 112: //KeyCode.P
        if(_m_iState == 2) {
          _m_oViewport.SoundOn(_m_bPause);
          _m_bPause = !_m_bPause;
        }
        break;
      case 114: //KeyCode.R (KeyCode.ESC not working)
        _m_bPause = false;
        if(_m_oViewport!=null) _m_oViewport.SoundOn(false);
        _m_iState = 0; //restart regardless of current state
        break;
    }
  }
  
  static Vector4 oColorDG = new Vector4(0x55/255, 0x55/255, 0x55/255, 1.0);
  static Vector4 oColorLG = new Vector4(0xaa/255, 0xaa/255, 0xaa/255, 1.0);
  static Vector4 oColorR = new Vector4(0xee/255, 0x00/255, 0x00/255, 1.0);
  void DrawLoaderStatus(int progress, int total)
  {
    //_canvas.width
    int iYPos = _m_oGraph.m_iViewportHeight~/2 - 40~/2;
    S_Rect stRect = new S_Rect(0, 0, _m_oGraph.m_iViewportWidth, _m_oGraph.m_iViewportHeight);
    _m_oGraph.DrawRectangle2D(stRect, oColorDG);
    stRect.Set(0, iYPos, _m_oGraph.m_iViewportWidth, 40);
    _m_oGraph.DrawRectangle2D(stRect, oColorLG);
    int step = _m_oGraph.m_iViewportWidth~/(total+1);
    stRect.Set(0, iYPos+2, progress*step+step, 40-4);
    _m_oGraph.DrawRectangle2D(stRect, oColorR);
  }

  bool Loop(double time)
  {
    double dCurTime = time - _m_dLastTime;
    if(dCurTime>500.0) dCurTime = 500.0; //limit frametime to protect against huge lags
    _m_dLastTime = time;
    
    _m_oGraph.Set2D();
    _m_oGraph.BeginScene();

    int iLoaderProgress = 0;

    // update and draw frame
    switch(_m_iState) {
      case 0: //wait for files to load
        if(_m_oReplayCon.IsLoaded()) iLoaderProgress++;
        if(_m_oShipImg.isLoaded) iLoaderProgress++;
        if(_m_oExplosionImg.isLoaded) iLoaderProgress++;
        if(_m_oBulletImg.isLoaded) iLoaderProgress++;
        if(_m_oLandingZoneImg.isLoaded) iLoaderProgress++;
        if(_m_oCpImg.isLoaded) iLoaderProgress++;
        if(_m_oE0Img.isLoaded) iLoaderProgress++;
        if(_m_oE1Img.isLoaded) iLoaderProgress++;
        if(_m_oE2Img.isLoaded) iLoaderProgress++;
        if(_m_oE3Img.isLoaded) iLoaderProgress++;
        if(_m_oE4Img.isLoaded) iLoaderProgress++;
        if(_m_oE5Img.isLoaded) iLoaderProgress++;
        if(_m_oE6Img.isLoaded) iLoaderProgress++;
        if(_m_oE7Img.isLoaded) iLoaderProgress++;
        if(_m_oDoorImg.isLoaded) iLoaderProgress++;
        if(_m_oCargoFuel.isLoaded) iLoaderProgress++;
        if(_m_oLapsTime.isLoaded) iLoaderProgress++;
        if(_m_oFont.IsLoaded()) iLoaderProgress++;

        for(int i=0; i<10; i++) {
          if(_m_oSound.IsLoaded(i)) iLoaderProgress++;
        }
        DrawLoaderStatus(iLoaderProgress, 30);

        //check loading
        bool bLoaded = true; //assume ok
        if(!_m_oReplayCon.IsLoaded()) bLoaded = false;

        if(!_m_oShipImg.isLoaded || !_m_oExplosionImg.isLoaded || !_m_oBulletImg.isLoaded ||
          !_m_oLandingZoneImg.isLoaded || !_m_oCpImg.isLoaded || !_m_oE0Img.isLoaded ||
          !_m_oE1Img.isLoaded || !_m_oE2Img.isLoaded || !_m_oE3Img.isLoaded ||
          !_m_oE4Img.isLoaded || !_m_oE5Img.isLoaded || !_m_oE6Img.isLoaded ||
          !_m_oE7Img.isLoaded || !_m_oDoorImg.isLoaded || !_m_oCargoFuel.isLoaded ||
          !_m_oLapsTime.isLoaded) bLoaded = false;

        if(!_m_oFont.IsLoaded()) bLoaded = false;

        for(int i=0; i<10; i++) {
          if(!_m_oSound.IsLoaded(i)) bLoaded = false;
        }

        if(bLoaded) {
          _m_oGraph.GenerateImageMask(_m_oShipImg);
          _m_oGraph.GenerateImageMask(_m_oBulletImg);
          _m_oGraph.GenerateImageMask(_m_oE0Img);
          _m_oGraph.GenerateImageMask(_m_oE1Img);
          _m_oGraph.GenerateImageMask(_m_oE2Img);
          _m_oGraph.GenerateImageMask(_m_oE3Img);
          _m_oGraph.GenerateImageMask(_m_oE4Img);
          _m_oGraph.GenerateImageMask(_m_oE5Img);
          _m_oGraph.GenerateImageMask(_m_oE6Img);
          _m_oGraph.GenerateImageMask(_m_oE7Img);
          _m_oGraph.GenerateImageMask(_m_oDoorImg);

          C_Viewport.InitImagesAndStuff(_m_oGraph, _m_oShipImg, _m_oExplosionImg, _m_oBulletImg, _m_oLandingZoneImg,
              _m_oCpImg, _m_oE0Img, _m_oE1Img, _m_oE2Img, _m_oE3Img, _m_oE4Img, _m_oE5Img, _m_oE6Img, _m_oE7Img, _m_oDoorImg,
              _m_oCargoFuel, _m_oLapsTime, _m_oFont);
          C_Object.s_oGraph = _m_oGraph;

          _m_bPause = false;
          S_Rect stDrawRect = new S_Rect(0, 0, _m_oGraph.m_iViewportWidth, _m_oGraph.m_iViewportHeight);
          _m_oReplay = _m_oReplayCon.m_oReplay;
          _m_oMap = new C_TileMap( _m_oReplay, _m_oGraph, stDrawRect);

          _m_iState++;
        }

        break;
      case 1:  //game init
        if(_m_oMap.IsLoaded()) {
          _m_oMap.SetMapWrapping(false, false);
          _m_oViewport = _m_oMap.GetExtendedInfo();
          _m_oViewport.InitPlayersAndSetMap(_m_oMap);
          _m_oViewport.SetSize(_m_oGraph.m_iViewportWidth, _m_oGraph.m_iViewportHeight);
          _m_dTimeLeft = 0.0;
          _m_dTimeToEnd = 0.0;
          _m_bEnd = false;
          _m_oReplay.ResetBeforePlay();

          _m_iState++;
        }

        break;
      case 2:  //game loop
        if(!_m_bPause) {
          _m_dTimeLeft += dCurTime;
          if(_m_bEnd) _m_dTimeToEnd += dCurTime;

          while(_m_dTimeLeft>=4.0) {
            S_ReplayAction stAction;
            do {
              stAction = _m_oReplay.Get(-1, 2);
              if(stAction!=null) {
                S_NetMsg2 stMsg2 = stAction.stMsg2;
                switch(stMsg2.iType) {
                  case S_NetMsg2.N_MSG2_NUMKILLS:
                    break;
                  case S_NetMsg2.N_MSG2_CARGO_R:
                    _m_oViewport.HandleCargoReply(stMsg2);
                    break;
                  case S_NetMsg2.N_MSG2_CARGORESPAWN:
                    _m_oViewport.HandleCargoRespawn(stMsg2);
                    break;
                  case S_NetMsg2.N_MSG2_GAMEOVER:
                    _m_bEnd = true;
                    break;
                }
              }
            } while(stAction!=null);

            if(_m_dTimeToEnd>6500) {
              _m_oViewport.SoundOn(false);
              _m_iState = 0; //restart
            }

            _m_oViewport.Update(4.0);
            _m_dTimeLeft -= 4.0;
            _m_oReplay.IncTimeSlot();
          }
        }
        _m_oViewport.Draw();
        break;
      case 4: //error
//        m_oDBImage.getGraphics().setColor(Color.RED);
//        m_oDBImage.getGraphics().drawString("Error: Bad replay", 5, this.getSize().height-32);
        break;
    }

    _m_oGraph.EndScene();

    // keep looping
    this._renderFrame();
    return true;
  }

  void Start()
  {
    if(!_m_bHasWebGL) {
      window.alert("WebGL not supported by this browser");
      return; //exit program
    }

    window.onKeyPress.listen(HandleKeyEvent);

    InputElement e1 = document.getElementById('Level');
    InputElement e2 = document.getElementById('Name');
    InputElement e3 = document.getElementById('debug_localhost');

    String szMap = e1.value;
    String szName = e2.value;
    //use this in the release version:
    String szURL = "hiscore_getreplay2.php?Level="+szMap+"&Name="+szName;
    //use this while debugging locally:
    if(e3!=null && int.parse(e3.value)==1) {
      szURL = "files/hiscore_getreplay.php_1.htm";
    }
    _m_oReplayCon.LoadReplay(szURL); //begin loading the replay data

    _loadTextures(); //begin loading all static textures

    _m_oSound.LoadSound("files/bulletfire.wav", 0);
    _m_oSound.LoadSound("files/explosion.wav", 1);
    _m_oSound.LoadSound("files/engine.wav", 2);
    _m_oSound.LoadSound("files/landonzone.wav", 3);
    _m_oSound.LoadSound("files/bullethit.wav", 4);
    _m_oSound.LoadSound("files/cp.wav", 5);
    _m_oSound.LoadSound("files/cp_end.wav", 6);
    _m_oSound.LoadSound("files/bounceonzone.wav", 7);
    _m_oSound.LoadSound("files/loadcargo.wav", 8);
    _m_oSound.LoadSound("files/unloadcargo.wav", 9);

    this._renderFrame();
  }

  void _renderFrame() {
    window.requestAnimationFrame((num time) { this.Loop(time); });
  }

}

void main()
{
  C_GameHandler oGame = new C_GameHandler(querySelector('#webreplay-canvas'));
  oGame.Start();
}
