// C_Global

//class with global helpers with OS specific impl
// - special directories handling
// - open file dialog (mac and windows, linux has no unified (same on all linuxes) way of doing this?)
// - get clipboard text (mac and windows, linux via xclip)
// - helper methods to read text files in the format used in the game

#include "global.h"
#include "fileresource.h"
#include <string.h>

#if defined WIN32
 #include <shlobj.h>
#elif defined __APPLE__
 #include <mach-o/dyld.h>
 #ifndef NOGUI
  #include "fssimplewindow.h"
 #endif
#endif

char C_Global::szCurrentPath[MAX_PATH];
char C_Global::szUserPath[MAX_PATH];
char C_Global::szExePath[MAX_PATH];
char C_Global::szDataPath[MAX_PATH];
char C_Global::szResourceFile[MAX_PATH];
char C_Global::szTemp[MAX_PATH];

char *C_Global::GetNextLineAndCommand(FILE *i_pFile, char *o_szLine, char *o_szCommand)
{
	o_szCommand[0] = 0; //reset command
	char *szRet = fgets(o_szLine, 512, i_pFile);
	if(szRet) sscanf(o_szLine, "%s", o_szCommand); //COMMAND
	return szRet;
}

bool C_Global::GetNextLineAndCommand(char *i_szSrc, char *o_szLine, char *o_szCommand, int *io_iNextLineIndex)
{
	int iStartLineIndex = 0;
	int iEndLineIndex = 0;
	bool bEnd = false;

	//get line
	do {
		char iCur = i_szSrc[iEndLineIndex];

		if(iCur=='\n' || iCur=='\r' || iCur==0) {
			if(iStartLineIndex<iEndLineIndex || iCur==0) {
				bEnd = true;
			} else {
				iStartLineIndex++;
				iEndLineIndex++;
			}
		} else iEndLineIndex++;
	} while(!bEnd);

	int iLen = iEndLineIndex-iStartLineIndex;
	if(iLen<=0) return false;

	memcpy(o_szLine, i_szSrc+iStartLineIndex, iLen);
	o_szLine[iLen] = 0; //zero terminate
	*io_iNextLineIndex += iEndLineIndex;

	//get command from line
	if(o_szCommand) {
		bEnd = false;
		iStartLineIndex = 0;
		iEndLineIndex = 0;
		do {
			char iCur = o_szLine[iEndLineIndex];

			if(iCur==' ' || iCur=='\t' || iCur==0) {
				if(iStartLineIndex<iEndLineIndex) {
					bEnd = true;
				} else {
					iStartLineIndex++;
					iEndLineIndex++;
				}
			} else iEndLineIndex++;
		} while(!bEnd);

		iLen = iEndLineIndex-iStartLineIndex;
		if(iLen<=0) iLen = 0;
		else memcpy(o_szCommand, o_szLine+iStartLineIndex, iLen);
		o_szCommand[iLen] = 0; //zero terminate
	}

	return true;
}

bool C_Global::GetWithin(char *i_szString, char i_cCh)
{
	int i = 0, iStart = 0;

	while(i_szString[i]!=i_cCh && i_szString[i]!=0) i++;
	if(i_szString[i]==0) return false;
	i++;

	while(i_szString[i]!=i_cCh && i_szString[i]!=0) i_szString[iStart++] = i_szString[i++];
	if(i_szString[i]!=i_cCh) return false;
	i_szString[iStart] = 0;
	return true;
}

char *C_Global::GetPositionAfter(char *i_szString, char i_cCh, int i_iIndex)
{
	int j, i = 0;
	for(j=0; j<i_iIndex; j++) {
		while(i_szString[i]!=i_cCh && i_szString[i]!=0) i++;
		if(i_szString[i]==0) return NULL;
		i++;
	}
	return i_szString+i;
}

//removes ../ and set the correct path, this is because in windows
// ../ does not work everywhere (in the GetOpenFileName for example)
void C_Global::FixPath(char *i_szPath)
{
	char *szPos = i_szPath;
	while(szPos) {
		szPos = strstr(i_szPath, "..");
		if(szPos) {
			char *szClipPosStart = szPos;
			char *szClipPosEnd = szPos+3;

			int iCount = 0;
			while(szClipPosStart>i_szPath) {
				szClipPosStart--;
				if(szClipPosStart[0]=='\\' || szClipPosStart[0]=='/') {
					iCount++;
					if(iCount==2) {
						szClipPosStart++;
						break;
					}
				}
			}
			if(iCount==2) {
				//do the string clipping
				strcpy(szClipPosStart, szClipPosEnd);
			}
		}
	}
}

bool C_Global::FindDataPath(char *i_szDataFile)
{
	char szPath[MAX_PATH];
	if(i_szDataFile && i_szDataFile[0]) {
		//first try exe_path, to find data when normally installed
		strcpy(szDataPath, szExePath);
		sprintf(szPath, "%s/%s", szExePath, i_szDataFile);
		FILE *pFile = fopen(szPath, "rb");
		if(!pFile) {
			//second try exe_path/../data, to find data when running from the project/exe/os/ directory
			sprintf(szDataPath, "%s/../data", szExePath);
			FixPath(szDataPath);
			sprintf(szPath, "%s/%s", szDataPath, i_szDataFile);
			pFile = fopen(szPath, "rb");
            if(!pFile) {
                //third try exe_path/data, to find data when running certain versions (mac)
                sprintf(szDataPath, "%s/data", szExePath);
                sprintf(szPath, "%s/%s", szDataPath, i_szDataFile);
                pFile = fopen(szPath, "rb");
                if(!pFile) {
                    //fourth, try current_directory/../data, to find data if running debug from visual studio
                    sprintf(szDataPath, "%s/../data", szCurrentPath);
					FixPath(szDataPath);
                    sprintf(szPath, "%s/%s", szDataPath, i_szDataFile);
                    pFile = fopen(szPath, "rb");
                    if(!pFile) {
                        szDataPath[0] = 0;
                    }
                }
            }
        }
		if(pFile) fclose(pFile);
	}
	return szDataPath[0]!=0;
}

char* C_Global::PathCombine(char *i_szFile, char *i_szPath)
{
	sprintf(szTemp, "%s/%s", i_szPath, i_szFile);
	return szTemp;
}

void C_Global::Init(char *i_szDataFile, char *i_szUserDirName)
{
	bool bCreateUserPath = false;

	srand((uint32_t)time(NULL)); //init random numbers once

	//make the best of it if finding fails
	strcpy(szCurrentPath, "./");
	strcpy(szUserPath, "./");
	strcpy(szExePath, "./");
	strcpy(szDataPath, ""); //not used/found

	//find exe path
#if defined WIN32
	char szPath[MAX_PATH];
	int iStringLen = GetModuleFileName(NULL, szExePath, sizeof(szExePath));
	bool bFound = iStringLen>0 && iStringLen<sizeof(szExePath);
	if(bFound) {
		//locate the last "\" or "/" in the filename
		char* szLastSlash = strrchr(szExePath, '\\');
		if(!szLastSlash) szLastSlash = strrchr(szExePath, '/');
		if(szLastSlash) {
			*szLastSlash = 0; //cut on last "\" or "/"
		}
	}

	//find current path
	iStringLen = GetCurrentDirectory(sizeof(szCurrentPath), szCurrentPath);
	bFound = iStringLen>0 && iStringLen<sizeof(szCurrentPath);

	//find user path
	//use "My Documents" folder to support none admin users
	if(!SHGetSpecialFolderPath(NULL, szPath, CSIDL_PERSONAL, false)) {
		//make the best of it
		if(szDataPath[0]) strcpy(szPath, szDataPath);
		else strcpy(szPath, szExePath);
	}
	sprintf(szUserPath, "%s/%s", szPath, i_szUserDirName);
	bCreateUserPath = true;
#else
	
#if defined __APPLE__
	uint32_t iSize = sizeof(szExePath);
	if(_NSGetExecutablePath(szExePath, &iSize) == 0) {
		//locate the last "/" in the filename
		char* szLastSlash = NULL, *szTmp = szExePath;
		while((szTmp = strstr(szTmp+1, ".app/"))!=NULL) szLastSlash = szTmp;
		if(szLastSlash) {
			*szLastSlash = 0; //cut on last ".app/"
			szLastSlash = strrchr(szExePath, '/');
			if(szLastSlash) {
				*szLastSlash = 0; //cut on last "/"
			}
		}
	}
#elif defined __linux__
	if(readlink("/proc/self/exe", szExePath, sizeof(szExePath))>0) {
		//locate the last "/" in the filename
		char* szLastSlash = strrchr(szExePath, '/');
		if(szLastSlash) {
			*szLastSlash = 0; //cut on last "/"
		}
	}
#endif

	//find current path
	getcwd(szCurrentPath, sizeof(szCurrentPath));

	//find user path
	char *szHome = getenv("HOME");
	if(szHome) {
		snprintf(szUserPath, sizeof(szUserPath), "%s/%s", szHome, i_szUserDirName);
		bCreateUserPath = true;
	}
#endif

	//find data path, relative to the exe path or current directory
	FindDataPath(i_szDataFile);
	sprintf(szResourceFile, "%s/resource.dat", szDataPath);

#ifdef WIN32
	//forward slashes are mostly ok on windows,
	// but some functions are not happy with them (the opendialog for example)
	char *szSlash;
	while(szSlash = strchr(szCurrentPath, '/')) *szSlash = '\\';
	while(szSlash = strchr(szUserPath, '/')) *szSlash = '\\';
	while(szSlash = strchr(szExePath, '/')) *szSlash = '\\';
	while(szSlash = strchr(szDataPath, '/')) *szSlash = '\\';
	while(szSlash = strchr(szResourceFile, '/')) *szSlash = '\\';
#endif

	//create user dirs if they don't exist
	if(bCreateUserPath) {
		C_Resource::CreateDir(szUserPath);
	}
}

#ifndef NOGUI

#define MAX_FILE_FILTERS 8
bool C_Global::OpenDialog(char *o_szFile, int i_iSize, char *i_szInitialPath, char *i_szzFilter, char *i_szzFilterDescription)
{
	bool bResult = false;
	int i, iNumFilters = 0;
	char *szFilters[MAX_FILE_FILTERS];
	char *szFilterDescs[MAX_FILE_FILTERS];

	memset(&szFilterDescs, 0, sizeof(szFilterDescs));

	if(!o_szFile || i_iSize<2) return false;

	if(i_szzFilter) {
		char *szTmp_1 = i_szzFilter;
		char *szTmp_2 = i_szzFilterDescription;
		do { //at least one
			szFilters[iNumFilters] = szTmp_1;
			if(szTmp_2 && *szTmp_2!=0) {
				szFilterDescs[iNumFilters] = szTmp_2;
				szTmp_2 = szTmp_2+strlen(szTmp_2)+1;
			}
			szTmp_1 += strlen(szTmp_1)+1; //next string
			iNumFilters++;
			if(iNumFilters>MAX_FILE_FILTERS-1) return false;
		} while(*szTmp_1!=0);
	}

#if defined WIN32
	if(i_szInitialPath && (int)(strlen(i_szInitialPath)+32)>i_iSize) return false;
	if(iNumFilters<=0) sprintf(o_szFile, "%s\\*", i_szInitialPath);
	else sprintf(o_szFile, "%s\\*.%s", i_szInitialPath, i_szzFilter);

	OPENFILENAME stOFN;
	char szFilter[1024];
	int iPos = 0;
	szFilter[0] = 0;
	if(iNumFilters<=0) iPos++;
	for(i=0; i<iNumFilters; i++) {
		int iLen = (int)strlen(szFilters[i]);
		if(szFilterDescs[i]) {
			int iLen2 = (int)strlen(szFilterDescs[i]);
			memcpy(szFilter+iPos, szFilterDescs[i], iLen2+1);
			iPos += iLen2+1;
		} else {
			memcpy(szFilter+iPos, szFilters[i], iLen+1);
			iPos += iLen+1;
		}
		sprintf(szFilter+iPos, "*.%s", szFilters[i]);
		iLen = (int)strlen(szFilter+iPos);
		iPos += iLen+1;
	}
	szFilter[iPos] = 0; //double termination

	memset(&stOFN, 0, sizeof(OPENFILENAME));
	stOFN.lStructSize = sizeof(OPENFILENAME);
	stOFN.hwndOwner   = NULL;
	stOFN.Flags       = OFN_HIDEREADONLY | OFN_FILEMUSTEXIST | OFN_LONGNAMES;
	stOFN.lpstrFile   = o_szFile;
	stOFN.nMaxFile    = i_iSize;
	if(iNumFilters>0) {
		stOFN.lpstrFilter  = szFilter;
		stOFN.nFilterIndex = 1;
	}

	bResult = GetOpenFileName(&stOFN) != 0;
#elif defined __APPLE__
	struct file_dialog_info stFI;
	stFI.szStartPath = i_szInitialPath;
	stFI.iNumFilters = iNumFilters;
	for(i=0; i<iNumFilters; i++) {
		stFI.szFilters[i] = szFilters[i];
	}
	bResult = FsFileOpenDialog(o_szFile, i_iSize, &stFI);
#elif defined __linux__
#endif

	return bResult;
}

bool C_Global::GetClipboardBuffer(char *o_szText, int i_iSize)
{
	bool bResult = false;
#if defined WIN32
	HANDLE hClip;
	if(OpenClipboard(NULL)) {
		hClip = GetClipboardData(CF_TEXT);
		if(hClip!=NULL) {
			char *szText = (char*)hClip;
			strncpy(o_szText, szText, i_iSize);
			o_szText[i_iSize-1] = 0;
			bResult = true;
		}
		CloseClipboard();
	}
#elif defined __APPLE__
	bResult = FsGetClipboardText(o_szText, i_iSize);
#elif defined __linux__
	//need: sudo apt-get install xclip
	FILE *hPipe = (FILE *)popen("xclip -o", "r");
	if(hPipe) {
		int r = fread(o_szText, 1, i_iSize, hPipe);
		if(r>0) o_szText[r] = 0;
		bResult = r>0;
		pclose(hPipe);
	}
#endif
	return bResult;
}

bool C_Global::OpenBrowser(char *i_szURL)
{
	bool bResult = false;
#if defined WIN32
	ShellExecute(NULL, "open", i_szURL, NULL, NULL, SW_SHOWNORMAL);
#elif defined __APPLE__	
	char szCommand[MAX_PATH];
	sprintf(szCommand, "open %s", i_szURL);
	bResult = system(szCommand);
#elif defined __linux__
	char szCommand[MAX_PATH];
	sprintf(szCommand, "xdg-open %s", i_szURL);
	bResult = system(szCommand);
#endif
	return bResult;
}


#endif
