// C_Control

//class to unify input handling of different OS
// - keyboard 
// - mouse
// - joystick pad
// max one device of each type

//since the class was first written for DirectInput, defines
// from it's headers lives on in the linux and mac versions.

#include "input.h"
#include <math.h> //fabs

#define KEYDOWN(name,key) (char)(name[key] & 0x80)

C_Control *C_Control::s_pclTheControl = NULL;

#ifdef WIN32
#pragma comment(lib, "dinput8.lib")
#endif

#ifdef __APPLE__
#include "fssimplewindow.h"
#endif

////////////////////////////////////////////////////////////////////////////////
//common for both windows, linux and mac

void C_Control::SetJoystickKeyMapping(S_JoystickMapping *i_pstMapping)
{
	if(!i_pstMapping && m_pstPMapping) {
		//turn off any active mapped keys
		if(m_pstPMapping->iRIGHT!=0) {
			m_acKBuf[m_pstPMapping->iRIGHT] = (char)0x00;
		}
		if(m_pstPMapping->iLEFT!=0) {
			m_acKBuf[m_pstPMapping->iLEFT] = (char)0x00;
		}
		if(m_pstPMapping->iDOWN!=0) {
			m_acKBuf[m_pstPMapping->iDOWN] = (char)0x00;
		}
		if(m_pstPMapping->iUP!=0) {
			m_acKBuf[m_pstPMapping->iUP] = (char)0x00;
		}
		for(int i=0; i<m_iNumButtons; i++) {
			if(m_pstPMapping->iB[i]!=0) m_acKBuf[m_pstPMapping->iB[i]] = (char)0x00;
		}
	}
	if(m_pstPMapping) {
		delete m_pstPMapping;
		m_pstPMapping = NULL;
	}
	if(i_pstMapping) {
		m_pstPMapping = new S_JoystickMapping;
		*m_pstPMapping = *i_pstMapping;
	}
}

//keyboard
int C_Control::IsPressed(int i_iKey)
{
	m_acPressed[i_iKey] = KEYDOWN(m_acKBuf, i_iKey);
	return m_acPressed[i_iKey];
}

int C_Control::IsPressedOnce(int i_iKey)
{
	if(!m_acPressed[i_iKey] && KEYDOWN(m_acKBuf, i_iKey)) {
		m_acPressed[i_iKey] = 1;
		m_adRepeat[i_iKey] = 0.1;
		m_aiRepeatCount[i_iKey] = 0;
		return 1;
	}
	return 0;
}

int C_Control::IsPressedRepeat(int i_iKey)
{
	uint32_t iRepeatVal = m_iRepeatRate;
	if(m_aiRepeatCount[i_iKey] == 1) iRepeatVal = m_iRepeatDelay;

	if(m_adRepeat[i_iKey] > iRepeatVal) {
		m_aiRepeatCount[i_iKey]++;
		m_acPressed[i_iKey] = 1;
		m_adRepeat[i_iKey] = 0.1;
		return 1;
	}
	return 0;
}

int C_Control::GetAscii(int i_iDiIndex)
{
	if(KEYDOWN(m_acKBuf, DIK_RSHIFT) | KEYDOWN(m_acKBuf, DIK_LSHIFT))
		return ASCIILOOKUPSHIFTED[i_iDiIndex];
	else
		return ASCIILOOKUP[i_iDiIndex];
}

void C_Control::InitAsciiLookup()
{
	int i;
	for(i=0; i<256; i++) ASCIILOOKUP[i] = -1;
	for(i=0; i<256; i++) ASCIILOOKUPSHIFTED[i] = -1; //shift has no effect
	ASCIILOOKUP[DIK_ESCAPE] = 27;
	ASCIILOOKUP[DIK_1] = '1'; ASCIILOOKUPSHIFTED[DIK_1] = '!';
	ASCIILOOKUP[DIK_2] = '2'; ASCIILOOKUPSHIFTED[DIK_2] = '@';
	ASCIILOOKUP[DIK_3] = '3'; ASCIILOOKUPSHIFTED[DIK_3] = '#';
	ASCIILOOKUP[DIK_4] = '4'; ASCIILOOKUPSHIFTED[DIK_4] = '$';
	ASCIILOOKUP[DIK_5] = '5'; //ASCIILOOKUPSHIFTED[DIK_5] = '%';
	ASCIILOOKUP[DIK_6] = '6'; ASCIILOOKUPSHIFTED[DIK_6] = '^';
	ASCIILOOKUP[DIK_7] = '7'; //ASCIILOOKUPSHIFTED[DIK_7] = '&';
	ASCIILOOKUP[DIK_8] = '8'; ASCIILOOKUPSHIFTED[DIK_8] = '*';
	ASCIILOOKUP[DIK_9] = '9'; ASCIILOOKUPSHIFTED[DIK_9] = '(';
	ASCIILOOKUP[DIK_0] = '0'; ASCIILOOKUPSHIFTED[DIK_0] = ')';
	ASCIILOOKUP[DIK_MINUS] = '-'; ASCIILOOKUPSHIFTED[DIK_MINUS] = '_';
	ASCIILOOKUP[DIK_EQUALS] = '='; ASCIILOOKUPSHIFTED[DIK_EQUALS] = '+';
	ASCIILOOKUP[DIK_BACK] = 8;
	//ASCIILOOKUP[DIK_TAB] = 9;
	ASCIILOOKUP[DIK_Q] = 'q'; ASCIILOOKUPSHIFTED[DIK_Q] = 'Q';
	ASCIILOOKUP[DIK_W] = 'w'; ASCIILOOKUPSHIFTED[DIK_W] = 'W';
	ASCIILOOKUP[DIK_E] = 'e'; ASCIILOOKUPSHIFTED[DIK_E] = 'E';
	ASCIILOOKUP[DIK_R] = 'r'; ASCIILOOKUPSHIFTED[DIK_R] = 'R';
	ASCIILOOKUP[DIK_T] = 't'; ASCIILOOKUPSHIFTED[DIK_T] = 'T';
	ASCIILOOKUP[DIK_Y] = 'y'; ASCIILOOKUPSHIFTED[DIK_Y] = 'Y';
	ASCIILOOKUP[DIK_U] = 'u'; ASCIILOOKUPSHIFTED[DIK_U] = 'U';
	ASCIILOOKUP[DIK_I] = 'i'; ASCIILOOKUPSHIFTED[DIK_I] = 'I';
	ASCIILOOKUP[DIK_O] = 'o'; ASCIILOOKUPSHIFTED[DIK_O] = 'O';
	ASCIILOOKUP[DIK_P] = 'p'; ASCIILOOKUPSHIFTED[DIK_P] = 'P';
	ASCIILOOKUP[DIK_LBRACKET] = 0xe5; ASCIILOOKUPSHIFTED[DIK_LBRACKET] = 0xc5; //åÅ
	ASCIILOOKUP[DIK_RBRACKET] = '~'; ASCIILOOKUPSHIFTED[DIK_RBRACKET] = '^';
	ASCIILOOKUP[DIK_RETURN] = 13; ASCIILOOKUPSHIFTED[DIK_RETURN] = 13;
	ASCIILOOKUP[DIK_A] = 'a'; ASCIILOOKUPSHIFTED[DIK_A] = 'A';
	ASCIILOOKUP[DIK_S] = 's'; ASCIILOOKUPSHIFTED[DIK_S] = 'S';
	ASCIILOOKUP[DIK_D] = 'd'; ASCIILOOKUPSHIFTED[DIK_D] = 'D';
	ASCIILOOKUP[DIK_F] = 'f'; ASCIILOOKUPSHIFTED[DIK_F] = 'F';
	ASCIILOOKUP[DIK_G] = 'g'; ASCIILOOKUPSHIFTED[DIK_G] = 'G';
	ASCIILOOKUP[DIK_H] = 'h'; ASCIILOOKUPSHIFTED[DIK_H] = 'H';
	ASCIILOOKUP[DIK_J] = 'j'; ASCIILOOKUPSHIFTED[DIK_J] = 'J';
	ASCIILOOKUP[DIK_K] = 'k'; ASCIILOOKUPSHIFTED[DIK_K] = 'K';
	ASCIILOOKUP[DIK_L] = 'l'; ASCIILOOKUPSHIFTED[DIK_L] = 'L';
	ASCIILOOKUP[DIK_SEMICOLON] = 0xf6; ASCIILOOKUPSHIFTED[DIK_SEMICOLON] = 0xd6; //öÖ
	ASCIILOOKUP[DIK_APOSTROPHE] = 0xe4; ASCIILOOKUPSHIFTED[DIK_APOSTROPHE] = 0xc4; //äÄ
	ASCIILOOKUP[DIK_GRAVE] = '\''; ASCIILOOKUPSHIFTED[DIK_GRAVE] = '*';
	ASCIILOOKUP[DIK_BACKSLASH] = '<'; ASCIILOOKUPSHIFTED[DIK_BACKSLASH] = '>';
	ASCIILOOKUP[DIK_Z] = 'z'; ASCIILOOKUPSHIFTED[DIK_Z] = 'Z';
	ASCIILOOKUP[DIK_X] = 'x'; ASCIILOOKUPSHIFTED[DIK_X] = 'X';
	ASCIILOOKUP[DIK_C] = 'c'; ASCIILOOKUPSHIFTED[DIK_C] = 'C';
	ASCIILOOKUP[DIK_V] = 'v'; ASCIILOOKUPSHIFTED[DIK_V] = 'V';
	ASCIILOOKUP[DIK_B] = 'b'; ASCIILOOKUPSHIFTED[DIK_B] = 'B';
	ASCIILOOKUP[DIK_N] = 'n'; ASCIILOOKUPSHIFTED[DIK_N] = 'N';
	ASCIILOOKUP[DIK_M] = 'm'; ASCIILOOKUPSHIFTED[DIK_M] = 'M';
	ASCIILOOKUP[DIK_COMMA] = ','; ASCIILOOKUPSHIFTED[DIK_COMMA] = ';';
	ASCIILOOKUP[DIK_PERIOD] = '.'; ASCIILOOKUPSHIFTED[DIK_PERIOD] = ':';
	ASCIILOOKUP[DIK_SLASH] = '-'; ASCIILOOKUPSHIFTED[DIK_SLASH] = '_';
	ASCIILOOKUP[DIK_MULTIPLY] = '*'; ASCIILOOKUPSHIFTED[DIK_MULTIPLY] = '*';
	ASCIILOOKUP[DIK_SPACE] = 32;
};

void C_Control::UpdateKeys()
{
	//set some data from the device states (m_acKBuf)
	m_dMouseTimer = m_pclTimer->GetInterval();
	for(int i=0; i<256; i++) {
		if(KEYDOWN(m_acKBuf, i)) {
			if(m_adRepeat[i]>0.0) m_adRepeat[i] += m_dMouseTimer;
			else {
				m_aiRepeatCount[i] = 0;
				m_adRepeat[i] = m_iRepeatRate+0.1;
			}
		} else {
			m_adRepeat[i] = 0;
			m_acPressed[i] = 0;
		}
	}
}

//mouse
int C_Control::GetMouseEvent(int *o_iX, int *o_iY, int *o_iExtraVal)
{
	if(m_pclMouseEvents->empty()) return 0;
	int iEvent = m_pclMouseEvents->front().iEvent;
	if(o_iX) *o_iX = m_pclMouseEvents->front().iX;
	if(o_iY) *o_iY = m_pclMouseEvents->front().iY;
	if(o_iExtraVal) *o_iExtraVal = m_pclMouseEvents->front().iExtraVal;
	m_pclMouseEvents->pop_front();
	return iEvent;
}

void C_Control::GetMousePos(int *o_iX, int *o_iY)
{
	if(o_iX) *o_iX = m_iMStateXBufLast;
	if(o_iY) *o_iY = m_iMStateYBufLast;
}

void C_Control::GetMouseButton(int *o_iLeft, int *o_iRight, int *o_iMiddle)
{
	if(o_iLeft) *o_iLeft = m_iMButtonStateBuf[0];
	if(o_iRight) *o_iRight = m_iMButtonStateBuf[1];
	if(o_iMiddle) *o_iMiddle = m_iMButtonStateBuf[2];
}

bool C_Control::MouseIsPressed(int i_iButton)
{
	if(i_iButton<0 || i_iButton>2) return false;
	return m_iMButtonStateBuf[i_iButton] != 0;
}

bool C_Control::MouseIsPressedOnce(int i_iButton)
{
	if(i_iButton<0 || i_iButton>2) return false;
	if(m_iMButtonState[0]==0 && m_iMButtonStateBuf[0]!=0) {
		m_iMButtonState[0] = 1;
		return true;
	}
	return false;
}

void C_Control::GetMouseMovement(int *o_iX, int *o_iY)
{
	if(o_iX) *o_iX = m_iMStateXBuf;
	if(o_iY) *o_iY = m_iMStateYBuf;
}

#define SPREADTIME 50.0 //ms
void C_Control::GetMouseMovementSmooth(double *o_dX, double *o_dY)
{
	double dResultX = 0, dResultY = 0;
	if(m_iMStateXBuf || m_iMStateYBuf) {
		m_dCurX += m_iMStateXBuf;
		m_dCurY += m_iMStateYBuf;

		double dStep = m_dMouseTimer/SPREADTIME; //% of movement to distribute each time called
		if(dStep > 1.0) dStep = 1.0;  //if frametime too long we must not move faster
		if(dStep <= 0.0) dStep = 1.0; //if the timer has to low resolution compared to the framerate

		m_dStepX = m_dCurX * dStep;
		m_dStepY = m_dCurY * dStep;
	}

	if(fabs(m_dCurX) > fabs(m_dStepX)) {
		m_dCurX -= m_dStepX;
		dResultX += m_dStepX;
	} else {
		dResultX += m_dCurX;
		m_dCurX = 0;
	}

	if(fabs(m_dCurY) > fabs(m_dStepY)) {
		m_dCurY -= m_dStepY;
		dResultY += m_dStepY;
	} else {
		dResultY += m_dCurY;
		m_dCurY = 0;
	}

	if(o_dX) *o_dX = dResultX;
	if(o_dY) *o_dY = dResultY;
}

void C_Control::GetJoystickState(S_JoystickState *o_pstState)
{
	*o_pstState = m_stJState;
}

void C_Control::GetJoystickStateDigital(S_JoystickState *o_pstState)
{
	o_pstState->iX = o_pstState->iY = 0;

	//limit deadzone to 300 on an analog joystick (it's still impossible to stear)
	if(m_stJState.iX > 300) {
		o_pstState->iX = 1;
	} else if(m_stJState.iX < -300) {
		o_pstState->iX = -1;
	}
	if(m_stJState.iY > 300) {
		o_pstState->iY = 1;
	} else if(m_stJState.iY < -300) {
		o_pstState->iY = -1;
	}

	o_pstState->iButton = m_stJState.iButton;
}

#ifdef WIN32
////////////////////////////////////////////////////////////////////////////////
//windows

BOOL CALLBACK DIEnumJoysticksProc(LPCDIDEVICEINSTANCE lpddi, void *pvRef)
{
	static bool bGamePadFound = false;
	BOOL bResult = DIENUM_CONTINUE;

	//best yet
	DIDEVICEINSTANCE *pstDevice = (DIDEVICEINSTANCE*)pvRef;
	*pstDevice = *lpddi;

	//take first gamepad (or joystick^ if no gamepads shows up)
	bool bDeviceGood = (((lpddi->dwDevType>>8)&0xff) == DI8DEVTYPEGAMEPAD_LIMITED);
	if(bDeviceGood) {
		bResult = DIENUM_STOP;
	}

	return bResult;
}

C_Control::C_Control(HWND i_hWnd, int i_iPadDevice, bool i_bExclusiveMouse)
{
	m_pstPMapping = NULL;
	m_iNumButtons = 4; //default
	memset(&m_iMButtonStateBuf[0], 0, sizeof(m_iMButtonStateBuf));
	memset(&m_iMButtonState[0], 0, sizeof(m_iMButtonState));
	m_iMStateXBuf = m_iMStateYBuf = 0;
	memset(&m_stJState, 0, sizeof(m_stJState));
	m_bExclusiveMouse = i_bExclusiveMouse;

	m_iMStateXBufLast = m_iMStateYBufLast = 0;
	m_pclMouseEvents = new std::deque<S_MouseEvent>;

	m_pclTimer = new C_Timer();
	m_iRepeatRate = 45;
	m_iRepeatDelay = 200;
	InitAsciiLookup();
	memset(m_acKBuf, 0, sizeof(m_acKBuf));

	m_dCurX = 0, m_dCurY = 0;
	m_dStepX = 0, m_dStepY = 0;

	s_pclTheControl = this;
	m_pclDIP = NULL;
	m_pclDIM = NULL;
	m_pclDIK = NULL;

	//note: for some reason this fails with error 5 access denied if exe is located somewhere in C:\Users\*user*\ in windows 10
	//run as admin works, but should really not be needed...
	DirectInput8Create(GetModuleHandle(NULL), DIRECTINPUT_VERSION, IID_IDirectInput8, (LPVOID*)&m_pclDI, NULL);
	if(!m_pclDI) return;

	//create a device for keyboard
	m_pclDI->CreateDevice(GUID_SysKeyboard, &m_pclDIK, NULL);
	if(m_pclDIK) {
		m_pclDIK->SetDataFormat(&c_dfDIKeyboard);
		m_pclDIK->SetCooperativeLevel(i_hWnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
		//make it possible to get input from the keyboard
		m_pclDIK->Acquire();
	}

	m_hKeyLayout = GetKeyboardLayout(0);

	//create a device for mouse
	m_pclDI->CreateDevice(GUID_SysMouse, &m_pclDIM, NULL);
	if(m_pclDIM) {
		m_pclDIM->SetDataFormat(&c_dfDIMouse);

		m_pclDIM->SetCooperativeLevel(i_hWnd, (i_bExclusiveMouse?DISCL_EXCLUSIVE:DISCL_NONEXCLUSIVE) | DISCL_FOREGROUND);
		//^must be nonexclusive in windowed mode

		m_pclDIM->Acquire();
		//use non buffered data, only get as often as we can handle it
	}

	//create a device for gamepad if one exists
	HRESULT          hRes;
	DIDEVICEINSTANCE stDevice;
	stDevice.dwSize = 0;
	hRes = m_pclDI->EnumDevices(DI8DEVTYPE_GAMEPAD, &DIEnumJoysticksProc, (void*)&stDevice, DIEDFL_ATTACHEDONLY);
	if(!stDevice.dwSize) hRes = m_pclDI->EnumDevices(DI8DEVTYPE_JOYSTICK, &DIEnumJoysticksProc, (void*)&stDevice, DIEDFL_ATTACHEDONLY);
	//hRes does not tell if we got a device, check that stDevice.dwSize is set instead

	if(stDevice.dwSize && ((stDevice.dwDevType & 0xff)==DI8DEVTYPE_GAMEPAD || (stDevice.dwDevType & 0xff)==DI8DEVTYPE_JOYSTICK)) {
		LPDIRECTINPUTDEVICE pclDIP;
		m_pclDI->CreateDevice(stDevice.guidInstance, &pclDIP, NULL);

		if(pclDIP) {
			bool bJoyFailed = true; //assume
			DIPROPDWORD  stDW;
			DIPROPRANGE  dipr;

			pclDIP->QueryInterface(IID_IDirectInputDevice8, (LPVOID *)&m_pclDIP);
			pclDIP->Release(); //done with old interface now

			if(!m_pclDIP) goto joyfailed;

			m_pclDIP->SetDataFormat(&c_dfDIJoystick);

			DIDEVCAPS didc;
			didc.dwSize = sizeof(didc);
			m_pclDIP->GetCapabilities(&didc);
			m_iNumButtons = didc.dwButtons;
			if(m_iNumButtons>MAX_BUTTONS) m_iNumButtons = MAX_BUTTONS;

			stDW.diph.dwSize       = sizeof(DIPROPDWORD);
			stDW.diph.dwHeaderSize = sizeof(DIPROPHEADER);
			stDW.diph.dwObj        = 0;
			stDW.diph.dwHow        = DIPH_DEVICE;
			stDW.dwData            = DIPROPAXISMODE_ABS;
			m_pclDIP->SetProperty(DIPROP_AXISMODE, &stDW.diph);

			//set the axis ranges for the device
			dipr.diph.dwSize        = sizeof(DIPROPRANGE);
			dipr.diph.dwHeaderSize  = sizeof(DIPROPHEADER);
			dipr.diph.dwHow         = DIPH_BYOFFSET;
			dipr.lMin               = JS_RANGE_MIN;  //negative to the left/top
			dipr.lMax               = JS_RANGE_MAX;  //positive to the right/bottom
			dipr.diph.dwObj         = DIJOFS_X;
			hRes = m_pclDIP->SetProperty(DIPROP_RANGE, &dipr.diph);
			if(FAILED(hRes)) goto joyfailed;
			dipr.diph.dwObj         = DIJOFS_Y;
			hRes = m_pclDIP->SetProperty(DIPROP_RANGE, &dipr.diph);
			if(FAILED(hRes)) goto joyfailed;

			//set the deadzone for the device
			stDW.diph.dwHow        = DIPH_BYOFFSET;
			stDW.dwData            = JS_DEADZONE;
			stDW.diph.dwObj        = DIJOFS_X;
			hRes = m_pclDIP->SetProperty(DIPROP_DEADZONE, &stDW.diph);
			if(FAILED(hRes)) goto joyfailed;
			stDW.diph.dwObj        = DIJOFS_Y;
			hRes = m_pclDIP->SetProperty(DIPROP_DEADZONE, &stDW.diph);
			if(FAILED(hRes)) goto joyfailed;

			m_pclDIP->SetCooperativeLevel(i_hWnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);

			m_pclDIP->Acquire();
			//use non buffered data, only get as often as we can handle it

			bJoyFailed = false;
joyfailed:
			if(bJoyFailed && m_pclDIP) {
				m_pclDIP->Release();
				m_pclDIP = NULL;
			}
		}
	}
}

void C_Control::ReleaseMouse()
{
	if(m_pclDIM) {
		m_pclDIM->Unacquire();
	}
}
void C_Control::AcquireMouse()
{
	if(m_pclDIM) {
		m_pclDIM->Acquire();
	}
}

C_Control::~C_Control()
{
	//pad
	if(m_pclDIP) {
		m_pclDIP->Unacquire();
		m_pclDIP->Release();
	}
	delete m_pstPMapping;

	//mouse
	if(m_pclDIM) {
		m_pclDIM->Unacquire();
		m_pclDIM->Release();
	}

	//keyboard
	if(m_pclDIK) {
		m_pclDIK->Unacquire();
		m_pclDIK->Release();
	}

	//direct input
	if(m_pclDI) m_pclDI->Release();

	delete m_pclMouseEvents;
	delete m_pclTimer;
	s_pclTheControl = NULL;
}

bool C_Control::ReadJoystickState()
{
	m_pclDIP->Poll();
	HRESULT hr = m_pclDIP->GetDeviceState(sizeof(m_stPBuf), &m_stPBuf);
	if(hr != DI_OK) {
		m_pclDIP->Acquire();
		m_pclDIP->Poll();
		hr = m_pclDIP->GetDeviceState(sizeof(m_stPBuf), &m_stPBuf);
	}
	if(hr != DI_OK) return false;

	//hat overrides analog
	if(m_stPBuf.rgdwPOV[0] != -1)
		switch(m_stPBuf.rgdwPOV[0])
	{
		case 0:     m_stPBuf.lX = 0;     m_stPBuf.lY = -1000; break;
		case 4500:  m_stPBuf.lX = 1000;  m_stPBuf.lY = -1000; break;
		case 9000:  m_stPBuf.lX = 1000;  m_stPBuf.lY = 0;     break;
		case 13500: m_stPBuf.lX = 1000;  m_stPBuf.lY = 1000;  break;
		case 18000: m_stPBuf.lX = 0;     m_stPBuf.lY = 1000;  break;
		case 22500: m_stPBuf.lX = -1000; m_stPBuf.lY = 1000;  break;
		case 27000: m_stPBuf.lX = -1000; m_stPBuf.lY = 0;     break;
		case 31500: m_stPBuf.lX = -1000; m_stPBuf.lY = -1000; break;
	}

	return true;
}

void C_Control::Update()
{
	int i;

	//mouse
	ZeroMemory(&m_stMBuf, sizeof(m_stMBuf));
	if(m_pclDIM) {
		if(m_pclDIM->GetDeviceState(sizeof(m_stMBuf), &m_stMBuf) != DI_OK) {
			m_pclTimer->Reset(); //we do not want the program to think it has gone any time when it gets active again
			m_pclDIM->Acquire();
			m_pclDIM->GetDeviceState(sizeof(m_stMBuf), &m_stMBuf);
		}
		m_iMButtonStateBuf[0] = m_stMBuf.rgbButtons[0];
		m_iMButtonStateBuf[1] = m_stMBuf.rgbButtons[1];
		m_iMButtonStateBuf[2] = m_stMBuf.rgbButtons[2];
		m_iMStateXBuf = m_stMBuf.lX;
		m_iMStateYBuf = m_stMBuf.lY;
	}
	if(m_iMButtonStateBuf[0] == 0) m_iMButtonState[0] = 0;
	if(m_iMButtonStateBuf[1] == 0) m_iMButtonState[1] = 0;
	if(m_iMButtonStateBuf[2] == 0) m_iMButtonState[2] = 0;

	//keyboard
	if(m_pclDIK) {
		if(m_pclDIK->GetDeviceState(sizeof(m_acKBuf), &m_acKBuf) != DI_OK) {
			m_pclTimer->Reset();
			m_pclDIK->Acquire();
			m_pclDIK->GetDeviceState(sizeof(m_acKBuf), &m_acKBuf);
		}
	}

	//joypad
	if(m_pclDIP) {
		//get state to m_stJState
		bool bResult = ReadJoystickState();

		if(bResult) {
			m_stJState.iX = m_stPBuf.lX;
			m_stJState.iY = m_stPBuf.lY;
			m_stJState.iButton = 0; //mask bits for buttons
			for(i=0; i<MAX_BUTTONS; i++)
				m_stJState.iButton |= (m_stPBuf.rgbButtons[i]!=0) ? 1<<i : 0;

			//joystick to keyboard mapping
			if(m_pstPMapping) {
				S_JoystickState stJState;
				GetJoystickStateDigital(&stJState);

				if(m_pstPMapping->iRIGHT!=0) {
					if(stJState.iX== 1)                                   m_acKBuf[m_pstPMapping->iRIGHT] = (char)0x80;
				}
				if(m_pstPMapping->iLEFT!=0) {
					if(stJState.iX==-1)                                   m_acKBuf[m_pstPMapping->iLEFT] = (char)0x80;
				}
				//stJState.iX==0 below is a hack that fit this game, and bad joysticks
				// where down or up may be indicated accidentally
				if(m_pstPMapping->iDOWN!=0) {
					if(stJState.iX==0 && stJState.iY== 1)                 m_acKBuf[m_pstPMapping->iDOWN] = (char)0x80;
				}
				if(m_pstPMapping->iUP!=0) {
					if(stJState.iX==0 && stJState.iY==-1)                 m_acKBuf[m_pstPMapping->iUP] = (char)0x80;
				}

				for(i=0; i<m_iNumButtons; i++) {
					if((stJState.iButton&(1<<i)) && m_pstPMapping->iB[i]!=0) m_acKBuf[m_pstPMapping->iB[i]] = (char)0x80;
				}
			}
		}
	}

	UpdateKeys();
}

int C_Control::GetAscii2(int i_iDiIndex)
{
	uint16_t iKeyAscii;
	uint8_t  acKeyState[256];
	if(GetKeyboardState(acKeyState)==false) return -1;

	uint32_t vk = MapVirtualKeyEx(i_iDiIndex, 3, m_hKeyLayout);

	int result = ToAsciiEx(vk, i_iDiIndex, acKeyState, &iKeyAscii, 0, m_hKeyLayout);
	if(result != 1) return -1;
	return iKeyAscii;
}

void C_Control::XLateMouseEvent(UINT iMessage, WPARAM wParam, LPARAM lParam)
{
	S_MouseEvent stEvent;
	stEvent.iX = LOWORD(lParam);
	stEvent.iY = HIWORD(lParam);
	switch(iMessage) {
		case WM_MOUSEMOVE:
			m_iMStateXBufLast = stEvent.iX;
			m_iMStateYBufLast = stEvent.iY;
			stEvent.iEvent = EVT_MOUSE_MOVE;
			m_pclMouseEvents->push_back(stEvent);
			break;
		case WM_LBUTTONDOWN:
			stEvent.iEvent = EVT_MOUSE_LEFT_DOWN;
			m_pclMouseEvents->push_back(stEvent);
			break;
		case WM_LBUTTONUP:
			stEvent.iEvent = EVT_MOUSE_LEFT_UP;
			m_pclMouseEvents->push_back(stEvent);
			break;
		case WM_RBUTTONDOWN:
			stEvent.iEvent = EVT_MOUSE_RIGHT_DOWN;
			m_pclMouseEvents->push_back(stEvent);
			break;
		case WM_RBUTTONUP:
			stEvent.iEvent = EVT_MOUSE_RIGHT_UP;
			m_pclMouseEvents->push_back(stEvent);
			break;
		case WM_MBUTTONDOWN:
			stEvent.iEvent = EVT_MOUSE_MID_DOWN;
			m_pclMouseEvents->push_back(stEvent);
			break;
		case WM_MBUTTONUP:
			stEvent.iEvent = EVT_MOUSE_MID_UP;
			m_pclMouseEvents->push_back(stEvent);
			break;
		case WM_MOUSEWHEEL:
			stEvent.iEvent = (((short)HIWORD(wParam))<0 ? EVT_MOUSE_WHEEL_DOWN : EVT_MOUSE_WHEEL_UP);
			stEvent.iX = m_iMStateXBufLast;
			stEvent.iY = m_iMStateYBufLast;
			stEvent.iExtraVal = (stEvent.iEvent==EVT_MOUSE_WHEEL_DOWN) ? 1 : -1;
			m_pclMouseEvents->push_back(stEvent);
			break;
	}
	if(m_pclMouseEvents->size()>12) m_pclMouseEvents->pop_front();
}

#else
////////////////////////////////////////////////////////////////////////////////
//linux, mac

C_Control::C_Control(int i_iPadDevice, bool i_bExclusiveMouse)
{
	m_pstPMapping = NULL;
	m_iNumButtons = 8; //default
	memset(&m_iMButtonStateBuf[0], 0, sizeof(m_iMButtonStateBuf));
	memset(&m_iMButtonState[0], 0, sizeof(m_iMButtonState));
	m_iMStateXBuf = m_iMStateYBuf = 0;
	memset(&m_stJState, 0, sizeof(m_stJState));
	memset(&m_stJStateLast, 0, sizeof(m_stJStateLast));
	m_bExclusiveMouse = i_bExclusiveMouse;

	m_pclTimer = new C_Timer();
	m_iRepeatRate = 45;
	m_iRepeatDelay = 200;
	InitAsciiLookup();
	memset(m_acKBuf, 0, sizeof(m_acKBuf));

	m_dCurX = 0, m_dCurY = 0;
	m_dStepX = 0, m_dStepY = 0;

#ifndef __APPLE__
	char szDevName[128];
	sprintf(szDevName, "%s%d", LINUX_JOYSTICK_DEVNAME, i_iPadDevice);
	m_hJoystick = open(szDevName, O_RDONLY | O_NONBLOCK);
#else
	FsSetMouseExclusive(m_bExclusiveMouse);
	m_bJoystickValid = (YsJoyReader::SetUpJoystick(m_clJoystick, 1)==1);
#endif
	m_iMStateXBufLast = m_iMStateYBufLast = 0;
	m_iMStateXBufLast2 = m_iMStateYBufLast2 = 0;

	m_pclMouseEvents = new std::deque<S_MouseEvent>;

	s_pclTheControl = this;
}

C_Control::~C_Control()
{
#ifndef __APPLE__
	if(m_hJoystick!=-1) close(m_hJoystick);
#endif

	delete m_pclMouseEvents;
	delete m_pclTimer;
	s_pclTheControl = NULL;
}

#ifndef __APPLE__
////////////////////////////////////////////////////////////////////////////////
//linux

void C_Control::SetDpyWnd(Display *i_pDisplay, Window i_wndRoot)
{
	//used for exclusive mouse mode
	// to emulate behaviour of that of directinput DISCL_EXCLUSIVE
	//mouse cursor is hidden and the mouse is locked in the window
	// and relative mouse movements do not stop when the cursor
	// should have reached the window border
	//this is achieved by constantly setting the cursor to the centre
	// with XWarpPointer. XWarpPointer only work in a virtual machine
	// when mouse integration is disabled

	m_pDisplay = i_pDisplay;
	m_wndRoot = i_wndRoot;
	//hide cursor
	Cursor cursor;
	Pixmap blank;
	XColor dummy;
	char data[1] = {0};
	blank = XCreateBitmapFromData(m_pDisplay, m_wndRoot, data, 1, 1);
	cursor = XCreatePixmapCursor(m_pDisplay, blank, blank, &dummy, &dummy, 0, 0);
	XDefineCursor(m_pDisplay, m_wndRoot, cursor);
	XFreePixmap(m_pDisplay, blank);
	//set initial position
	XWindowAttributes winAttrib;
	XGetWindowAttributes(m_pDisplay, m_wndRoot, &winAttrib);
	XWarpPointer(m_pDisplay, None, m_wndRoot, 0, 0, 0, 0, winAttrib.width/2, winAttrib.height/2);
	m_iMStateXBufLast2 = winAttrib.width/2;
	m_iMStateYBufLast2 = winAttrib.height/2;
}

int C_Control::ReadJoystickEvent(S_JSEvent *pstJSE)
{
	int iLen;
	if(m_hJoystick<0) return -1;
	iLen = read(m_hJoystick, pstJSE, sizeof(*pstJSE));

	if(iLen == -1) return 0;
	if(iLen == sizeof(*pstJSE)) return 1;
	return -1;
}

void C_Control::ReadJoystickState()
{
	S_JSEvent stJSE;
	if(m_hJoystick<0) return;
	//get state to m_stJState
	while(ReadJoystickEvent(&stJSE)==1) {
		stJSE.type &= ~JS_EVENT_INIT; //ignore synthetic events
		if(stJSE.type == JS_EVENT_AXIS) {
			switch(stJSE.number) {
				case 0: m_stJState.iX = stJSE.value; break;
				case 1: m_stJState.iY = stJSE.value; break;
			}
		} else if(stJSE.type == JS_EVENT_BUTTON) {
			if(stJSE.number < MAX_BUTTONS) {
				if(stJSE.value>0) m_stJState.iButton |= (1 << stJSE.number);
				else              m_stJState.iButton &= ~(1 << stJSE.number);
			}
		}
	}
}
#else
////////////////////////////////////////////////////////////////////////////////
//mac

void C_Control::ReadJoystickState()
{
	if(!m_bJoystickValid) return;
	//get state to m_stJState
	m_clJoystick[0].Read();
	m_stJState.iButton = 0;
	m_stJState.iX = (int)(1000.0*m_clJoystick[0].axis[0].GetCalibratedValue()); //X
	m_stJState.iY = (int)(1000.0*m_clJoystick[0].axis[1].GetCalibratedValue()); //Y
	//hat overrides analog
	if(m_clJoystick[0].hatSwitch[0].exist)
		switch(m_clJoystick[0].hatSwitch[0].value)
	{
		case 0: m_stJState.iX = 0;     m_stJState.iY = 0;     break;
		case 1: m_stJState.iX = 0;     m_stJState.iY = -1000;  break;
		case 2: m_stJState.iX = 1000;  m_stJState.iY = -1000;  break;
		case 3: m_stJState.iX = 1000;  m_stJState.iY = 0;     break;
		case 4: m_stJState.iX = 1000;  m_stJState.iY = 1000; break;
		case 5: m_stJState.iX = 0;     m_stJState.iY = 1000; break;
		case 6: m_stJState.iX = -1000; m_stJState.iY = 1000; break;
		case 7: m_stJState.iX = -1000; m_stJState.iY = 0;     break;
		case 8: m_stJState.iX = -1000; m_stJState.iY = -1000;  break;
	}
	for(int i=0; i<MAX_BUTTONS; i++) {
		if(m_clJoystick[0].button[i].value>0) m_stJState.iButton |= (1 << i);
		else                                  m_stJState.iButton &= ~(1 << i);
	}
}
#endif

////////////////////////////////////////////////////////////////////////////////
//linux, mac

void C_Control::Update()
{
	int i;

	//mouse
	if(m_iMButtonStateBuf[0] == 0) m_iMButtonState[0] = 0;
	if(m_iMButtonStateBuf[1] == 0) m_iMButtonState[1] = 0;
	if(m_iMButtonStateBuf[2] == 0) m_iMButtonState[2] = 0;

	//joypad
#ifndef __APPLE__
	if(m_hJoystick>=0) {
#else
	if(m_bJoystickValid) {
#endif
		//get state to m_stJState
		ReadJoystickState();
		//since key state is not read from scratch every time (for mac, linux), but based on events
		// the joystick mapping will interfere with m_acKBuf, so some more code is needed

		//joystick to keyboard mapping
		if(m_pstPMapping) {
			S_JoystickState stJState;
			GetJoystickStateDigital(&stJState);

			if(m_pstPMapping->iRIGHT!=0) {
				if(stJState.iX!=m_stJStateLast.iX && stJState.iX!= 1) m_acKBuf[m_pstPMapping->iRIGHT] = 0;
				if(stJState.iX== 1)                                   m_acKBuf[m_pstPMapping->iRIGHT] = (char)0x80;
			}
			if(m_pstPMapping->iLEFT!=0) {
				if(stJState.iX!=m_stJStateLast.iX && stJState.iX!=-1) m_acKBuf[m_pstPMapping->iLEFT] = 0;
				if(stJState.iX==-1)                                   m_acKBuf[m_pstPMapping->iLEFT] = (char)0x80;
			}
			//stJState.iX==0 below is a hack that fit this game, and bad joysticks
			// where down or up may be indicated accidentally
			if(m_pstPMapping->iDOWN!=0) {
				if(stJState.iY!=m_stJStateLast.iY && stJState.iY!= 1) m_acKBuf[m_pstPMapping->iDOWN] = 0;
				if(stJState.iX==0 && stJState.iY== 1)                 m_acKBuf[m_pstPMapping->iDOWN] = (char)0x80;
			}
			if(m_pstPMapping->iUP!=0) {
				if(stJState.iY!=m_stJStateLast.iY && stJState.iY!=-1) m_acKBuf[m_pstPMapping->iUP] = 0;
				if(stJState.iX==0 && stJState.iY==-1)                 m_acKBuf[m_pstPMapping->iUP] = (char)0x80;
			}

			for(i=0; i<m_iNumButtons; i++) {
				if(m_pstPMapping->iB[i]!=0) {
					int iCur  = stJState.iButton&(1<<i);
					int iLast = m_stJStateLast.iButton&(1<<i);
					if(iCur!=iLast && !iCur) m_acKBuf[m_pstPMapping->iB[i]] = 0;
					if(iCur) m_acKBuf[m_pstPMapping->iB[i]] = (char)0x80;
				}
			}
		}
		m_stJStateLast = m_stJState;
	}

	UpdateKeys();
}

int C_Control::GetAscii2(int i_iDiIndex)
{
	return GetAscii(i_iDiIndex);
}

#ifdef __APPLE__
////////////////////////////////////////////////////////////////////////////////
//mac

void C_Control::PollInput()
{
	S_MouseEvent stEvent;
	
	m_iMStateXBuf = 0;
	m_iMStateYBuf = 0;

	//mouse
	int lb, mb, rb, mx, my, wheel;
	int iEvent = FsGetMouseEvent(lb, mb, rb, mx, my, wheel);
	if(iEvent) {
		stEvent.iX = mx;
		stEvent.iY = my;

		m_iMStateXBuf = mx-m_iMStateXBufLast2;
		m_iMStateYBuf = my-m_iMStateYBufLast2;

		//exclusive mode handling
		if(m_bExclusiveMouse) {
			int w, h;
			FsGetWindowSize(w, h);
			m_iMStateXBufLast2 = w/2;
			m_iMStateYBufLast2 = h/2;
			m_iMStateXBufLast += m_iMStateXBuf; if(m_iMStateXBufLast<0) m_iMStateXBufLast=0; if(m_iMStateXBufLast>=w) m_iMStateXBufLast=w-1;
			m_iMStateYBufLast += m_iMStateYBuf; if(m_iMStateYBufLast<0) m_iMStateYBufLast=0; if(m_iMStateYBufLast>=h) m_iMStateYBufLast=h-1;
		} else {
			m_iMStateXBufLast2 = mx;
			m_iMStateYBufLast2 = my;
			m_iMStateXBufLast = mx;
			m_iMStateYBufLast = my;
		}
		
		switch(iEvent) {
			case FSMOUSEEVENT_MOVE:
				stEvent.iEvent = EVT_MOUSE_MOVE;       m_pclMouseEvents->push_back(stEvent);
				break;
			case FSMOUSEEVENT_LBUTTONDOWN:
				m_iMButtonStateBuf[0] = 1;
				stEvent.iEvent = EVT_MOUSE_LEFT_DOWN;  m_pclMouseEvents->push_back(stEvent);
				break;
			case FSMOUSEEVENT_LBUTTONUP:
				m_iMButtonStateBuf[0] = 0;
				stEvent.iEvent = EVT_MOUSE_LEFT_UP;    m_pclMouseEvents->push_back(stEvent);
				break;
			case FSMOUSEEVENT_MBUTTONDOWN:
				m_iMButtonStateBuf[2] = 1;
				stEvent.iEvent = EVT_MOUSE_MID_DOWN;   m_pclMouseEvents->push_back(stEvent);
				break;
			case FSMOUSEEVENT_MBUTTONUP:
				m_iMButtonStateBuf[2] = 0;
				stEvent.iEvent = EVT_MOUSE_MID_UP;     m_pclMouseEvents->push_back(stEvent);
				break;
			case FSMOUSEEVENT_RBUTTONDOWN:
				m_iMButtonStateBuf[1] = 1;
				stEvent.iEvent = EVT_MOUSE_RIGHT_DOWN; m_pclMouseEvents->push_back(stEvent);
				break;
			case FSMOUSEEVENT_RBUTTONUP:
				m_iMButtonStateBuf[1] = 0;
				stEvent.iEvent = EVT_MOUSE_RIGHT_UP;   m_pclMouseEvents->push_back(stEvent);
				break;
			case FSMOUSEEVENT_WHEEL:
				stEvent.iEvent = wheel>0 ? EVT_MOUSE_WHEEL_UP : EVT_MOUSE_WHEEL_DOWN;
				lb = -1; //sign
				if(wheel<0) lb = 1;
				//make some sort of acceleration curve
				if(abs(wheel)<10)      wheel = lb*1;
				else if(abs(wheel)<20) wheel = lb*2;
				else                   wheel = lb*5;
				stEvent.iExtraVal = wheel;
				m_pclMouseEvents->push_back(stEvent);
				break;
		}
	}

	/*if(FsGetKeyState(FSKEY_WHEELUP)!=0) {
		stEvent.iX = m_iMStateXBufLast;
		stEvent.iY = m_iMStateYBufLast;
		stEvent.iEvent = EVT_MOUSE_WHEEL_UP;
	} else if(FsGetKeyState(FSKEY_WHEELDOWN)!=0) {
		stEvent.iX = m_iMStateXBufLast;
		stEvent.iY = m_iMStateYBufLast;
		stEvent.iEvent = EVT_MOUSE_WHEEL_DOWN;
	}*/

	//keys
	m_acKBuf[DIK_SPACE] = FsGetKeyState(FSKEY_SPACE)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_0] = FsGetKeyState(FSKEY_0)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_1] = FsGetKeyState(FSKEY_1)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_2] = FsGetKeyState(FSKEY_2)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_3] = FsGetKeyState(FSKEY_3)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_4] = FsGetKeyState(FSKEY_4)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_5] = FsGetKeyState(FSKEY_5)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_6] = FsGetKeyState(FSKEY_6)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_7] = FsGetKeyState(FSKEY_7)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_8] = FsGetKeyState(FSKEY_8)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_9] = FsGetKeyState(FSKEY_9)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_A] = FsGetKeyState(FSKEY_A)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_B] = FsGetKeyState(FSKEY_B)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_C] = FsGetKeyState(FSKEY_C)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_D] = FsGetKeyState(FSKEY_D)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_E] = FsGetKeyState(FSKEY_E)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F] = FsGetKeyState(FSKEY_F)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_G] = FsGetKeyState(FSKEY_G)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_H] = FsGetKeyState(FSKEY_H)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_I] = FsGetKeyState(FSKEY_I)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_J] = FsGetKeyState(FSKEY_J)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_K] = FsGetKeyState(FSKEY_K)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_L] = FsGetKeyState(FSKEY_L)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_M] = FsGetKeyState(FSKEY_M)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_N] = FsGetKeyState(FSKEY_N)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_O] = FsGetKeyState(FSKEY_O)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_P] = FsGetKeyState(FSKEY_P)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_Q] = FsGetKeyState(FSKEY_Q)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_R] = FsGetKeyState(FSKEY_R)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_S] = FsGetKeyState(FSKEY_S)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_T] = FsGetKeyState(FSKEY_T)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_U] = FsGetKeyState(FSKEY_U)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_V] = FsGetKeyState(FSKEY_V)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_W] = FsGetKeyState(FSKEY_W)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_X] = FsGetKeyState(FSKEY_X)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_Y] = FsGetKeyState(FSKEY_Y)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_Z] = FsGetKeyState(FSKEY_Z)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_ESCAPE] = FsGetKeyState(FSKEY_ESC)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F1] = FsGetKeyState(FSKEY_F1)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F2] = FsGetKeyState(FSKEY_F2)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F3] = FsGetKeyState(FSKEY_F3)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F4] = FsGetKeyState(FSKEY_F4)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F5] = FsGetKeyState(FSKEY_F5)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F6] = FsGetKeyState(FSKEY_F6)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F7] = FsGetKeyState(FSKEY_F7)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F8] = FsGetKeyState(FSKEY_F8)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F9] = FsGetKeyState(FSKEY_F9)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F10] = FsGetKeyState(FSKEY_F10)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F11] = FsGetKeyState(FSKEY_F11)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_F12] = FsGetKeyState(FSKEY_F12)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_CAPITAL] = FsGetKeyState(FSKEY_CAPSLOCK)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_SCROLL] = FsGetKeyState(FSKEY_SCROLLLOCK)!=0 ? (char)0x80 : 0;
	//m_acKBuf[DIK_PAUSE] = FsGetKeyState(FSKEY_PAUSEBREAK)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_BACK] = FsGetKeyState(FSKEY_BS)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_TAB] = FsGetKeyState(FSKEY_TAB)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_RETURN] = FsGetKeyState(FSKEY_ENTER)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_LSHIFT] = FsGetKeyState(FSKEY_SHIFT)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_LCONTROL] = FsGetKeyState(FSKEY_CTRL)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_LMENU] = FsGetKeyState(FSKEY_ALT)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_INSERT] = FsGetKeyState(FSKEY_INS)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_DELETE] = FsGetKeyState(FSKEY_DEL)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_HOME]  = FsGetKeyState(FSKEY_HOME)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_END]   = FsGetKeyState(FSKEY_END)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_PRIOR] = FsGetKeyState(FSKEY_PAGEUP)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_NEXT]  = FsGetKeyState(FSKEY_PAGEDOWN)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_UP]    = FsGetKeyState(FSKEY_UP)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_DOWN]  = FsGetKeyState(FSKEY_DOWN)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_LEFT]  = FsGetKeyState(FSKEY_LEFT)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_RIGHT] = FsGetKeyState(FSKEY_RIGHT)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_NUMLOCK] = FsGetKeyState(FSKEY_NUMLOCK)!=0 ? (char)0x80 : 0;
	//m_acKBuf[DIK_] = FsGetKeyState(FSKEY_TILDA)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_MINUS] = FsGetKeyState(FSKEY_MINUS)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_EQUALS] = FsGetKeyState(FSKEY_PLUS)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_LBRACKET] = FsGetKeyState(FSKEY_LBRACKET)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_RBRACKET] = FsGetKeyState(FSKEY_RBRACKET)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_BACKSLASH] = FsGetKeyState(FSKEY_BACKSLASH)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_SEMICOLON] = FsGetKeyState(FSKEY_SEMICOLON)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_APOSTROPHE] = FsGetKeyState(FSKEY_SINGLEQUOTE)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_COMMA] = FsGetKeyState(FSKEY_COMMA)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_PERIOD] = FsGetKeyState(FSKEY_DOT)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_SLASH] = FsGetKeyState(FSKEY_SLASH)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_NUMPAD0] = FsGetKeyState(FSKEY_TEN0)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_NUMPAD1] = FsGetKeyState(FSKEY_TEN1)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_NUMPAD2] = FsGetKeyState(FSKEY_TEN2)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_NUMPAD3] = FsGetKeyState(FSKEY_TEN3)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_NUMPAD4] = FsGetKeyState(FSKEY_TEN4)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_NUMPAD5] = FsGetKeyState(FSKEY_TEN5)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_NUMPAD6] = FsGetKeyState(FSKEY_TEN6)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_NUMPAD7] = FsGetKeyState(FSKEY_TEN7)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_NUMPAD8] = FsGetKeyState(FSKEY_TEN8)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_NUMPAD9] = FsGetKeyState(FSKEY_TEN9)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_DECIMAL] = FsGetKeyState(FSKEY_TENDOT)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_DIVIDE] = FsGetKeyState(FSKEY_TENSLASH)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_MULTIPLY] = FsGetKeyState(FSKEY_TENSTAR)!=0 ? (char)0x80 : 0;
	//m_acKBuf[DIK_SUBTRACT] = FsGetKeyState(FSKEY_TENMINUS)!=0 ? (char)0x80 : 0;
	//m_acKBuf[DIK_ADD] = FsGetKeyState(FSKEY_TENPLUS)!=0 ? (char)0x80 : 0;
	m_acKBuf[DIK_SUBTRACT] = (FsGetKeyState(FSKEY_TENMINUS)!=0 || FsGetKeyState(FSKEY_MINUS)!=0) ? (char)0x80 : 0;
	m_acKBuf[DIK_ADD] = (FsGetKeyState(FSKEY_TENPLUS)!=0 || FsGetKeyState(FSKEY_PLUS)!=0) ? (char)0x80 : 0;
	m_acKBuf[DIK_NUMPADENTER] = FsGetKeyState(FSKEY_TENENTER)!=0 ? (char)0x80 : 0;

	if(m_pclMouseEvents->size()>12) m_pclMouseEvents->pop_front();
}

#else
////////////////////////////////////////////////////////////////////////////////
//linux

void C_Control::PollInput(XEvent *pKeyMouseEvent)
{
	XLateKeyEvent(pKeyMouseEvent);
	XLateMouseEvent(pKeyMouseEvent);
}

void C_Control::XLateKeyEvent(XEvent *pKeyEvent)
{
	if(pKeyEvent==NULL) return;

	if(pKeyEvent->type==KeyPress || pKeyEvent->type==KeyRelease) {
		int key = -1;
		KeySym keysym;

		keysym = XLookupKeysym((XKeyEvent*)pKeyEvent, 0);

		//printf("%d  pressed %d\n", keysym, pKeyEvent->type==KeyPress);
		switch(keysym) {
			case XK_KP_Page_Up:   key = DIK_NUMPAD9; break;
			case XK_Page_Up:      key = DIK_NEXT;    break;
			case XK_KP_Page_Down: key = DIK_NUMPAD3; break;
			case XK_Page_Down:    key = DIK_PRIOR;   break;
			case XK_KP_Home:      key = DIK_NUMPAD7; break;
			case XK_Home:         key = DIK_HOME;    break;
			case XK_KP_End:       key = DIK_NUMPAD1; break;
			case XK_End:          key = DIK_END;     break;
			case XK_KP_4:
			case XK_KP_Left:
			case XK_Left:         key = DIK_LEFT;    break;
			case XK_KP_6:
			case XK_KP_Right:
			case XK_Right:        key = DIK_RIGHT;   break;
			case XK_KP_2:
			case XK_KP_Down:
			case XK_Down:         key = DIK_DOWN;    break;
			case XK_KP_8:
			case XK_KP_Up:
			case XK_Up:           key = DIK_UP;      break;
			case XK_Insert:       key = DIK_INSERT;  break;
			case XK_Delete:       key = DIK_DELETE;  break;

			case XK_Escape:       key = DIK_ESCAPE;  break;
			case XK_KP_Enter:     key = DIK_RETURN;  break;
			case XK_Return:       key = DIK_RETURN;  break;

			case XK_Tab:          key = DIK_TAB;  break;
			case XK_F1:           key = DIK_F1;   break;
			case XK_F2:           key = DIK_F2;   break;
			case XK_F3:           key = DIK_F3;   break;
			case XK_F4:           key = DIK_F4;   break;
			case XK_F5:           key = DIK_F5;   break;
			case XK_F6:           key = DIK_F6;   break;
			case XK_F7:           key = DIK_F7;   break;
			case XK_F8:           key = DIK_F8;   break;
			case XK_F9:           key = DIK_F9;   break;
			case XK_F10:          key = DIK_F10;  break;
			case XK_F11:          key = DIK_F11;  break;
			case XK_F12:          key = DIK_F12;  break;

			case XK_BackSpace:    key = DIK_BACK; break;

			case XK_Shift_L:      key = DIK_LSHIFT;   break;
			case XK_Shift_R:      key = DIK_RSHIFT;   break;

			case XK_Execute:
			case XK_Control_L:
			case XK_Control_R:    key = DIK_LCONTROL; break;

			case XK_Alt_L:
			case XK_Meta_L:
			case XK_ISO_Level3_Shift:
			case XK_Alt_R:
			case XK_Meta_R:       key = DIK_LMENU;    break;

			case XK_KP_Multiply:  key = DIK_MULTIPLY; break;
			case XK_KP_Add:       key = DIK_ADD;      break;
			case XK_KP_Subtract:  key = DIK_SUBTRACT; break;
			case XK_KP_Divide:    key = DIK_DIVIDE;   break; 

			case XK_0:            key = DIK_0; break;
			case XK_1:            key = DIK_1; break;
			case XK_2:            key = DIK_2; break;
			case XK_3:            key = DIK_3; break;
			case XK_4:            key = DIK_4; break;
			case XK_5:            key = DIK_5; break;
			case XK_6:            key = DIK_6; break;
			case XK_7:            key = DIK_7; break;
			case XK_8:            key = DIK_8; break;
			case XK_9:            key = DIK_9; break;

			case XK_a:
			case XK_A:            key = DIK_A; break;
			case XK_b:
			case XK_B:            key = DIK_B; break;
			case XK_c:
			case XK_C:            key = DIK_C; break;
			case XK_d:
			case XK_D:            key = DIK_D; break;
			case XK_e:
			case XK_E:            key = DIK_E; break;
			case XK_f:
			case XK_F:            key = DIK_F; break;
			case XK_g:
			case XK_G:            key = DIK_G; break;
			case XK_h:
			case XK_H:            key = DIK_H; break;
			case XK_i:
			case XK_I:            key = DIK_I; break;
			case XK_j:
			case XK_J:            key = DIK_J; break;
			case XK_k:
			case XK_K:            key = DIK_K; break;
			case XK_l:
			case XK_L:            key = DIK_L; break;
			case XK_m:
			case XK_M:            key = DIK_M; break;
			case XK_n:
			case XK_N:            key = DIK_N; break;
			case XK_o:
			case XK_O:            key = DIK_O; break;
			case XK_p:
			case XK_P:            key = DIK_P; break;
			case XK_q:
			case XK_Q:            key = DIK_Q; break;
			case XK_r:
			case XK_R:            key = DIK_R; break;
			case XK_s:
			case XK_S:            key = DIK_S; break;
			case XK_t:
			case XK_T:            key = DIK_T; break;
			case XK_u:
			case XK_U:            key = DIK_U; break;
			case XK_v:
			case XK_V:            key = DIK_V; break;
			case XK_w:
			case XK_W:            key = DIK_W; break;
			case XK_x:
			case XK_X:            key = DIK_X; break;
			case XK_y:
			case XK_Y:            key = DIK_Y; break;
			case XK_z:
			case XK_Z:            key = DIK_Z; break;
			case XK_comma:        key = DIK_COMMA;  break;
			case XK_period:       key = DIK_PERIOD; break;
		}
		if(key>=0) m_acKBuf[key] = (pKeyEvent->type==KeyPress) ? (char)0x80 : 0;
	}
}

void C_Control::XLateMouseEvent(XEvent *pMouseEvent)
{
	m_iMStateXBuf = 0;
	m_iMStateYBuf = 0;
	if(pMouseEvent==NULL) return;

	if(pMouseEvent->type==MotionNotify || pMouseEvent->type==ButtonPress || pMouseEvent->type==ButtonRelease) {
		static int aiButtonLookup[4] = {0,0,2,1};
		if((pMouseEvent->type==ButtonPress || pMouseEvent->type==ButtonRelease)
		 && pMouseEvent->xbutton.button>=1 && pMouseEvent->xbutton.button<=3) {
			char cButton = pMouseEvent->type==ButtonPress ? 1 : 0;
			m_iMButtonStateBuf[aiButtonLookup[pMouseEvent->xbutton.button]] = cButton;
		}
		m_iMStateXBuf = pMouseEvent->xmotion.x-m_iMStateXBufLast2;
		m_iMStateYBuf = pMouseEvent->xmotion.y-m_iMStateYBufLast2;

		//exclusive mode handling
		if(m_bExclusiveMouse) {
			XWindowAttributes winAttrib;
			XGetWindowAttributes(m_pDisplay, m_wndRoot, &winAttrib);
			m_iMStateXBufLast2 = winAttrib.width/2;
			m_iMStateYBufLast2 = winAttrib.height/2;
			m_iMStateXBufLast += m_iMStateXBuf; if(m_iMStateXBufLast<0) m_iMStateXBufLast=0; if(m_iMStateXBufLast>=winAttrib.width) m_iMStateXBufLast=winAttrib.width-1;
			m_iMStateYBufLast += m_iMStateYBuf; if(m_iMStateYBufLast<0) m_iMStateYBufLast=0; if(m_iMStateYBufLast>=winAttrib.height) m_iMStateYBufLast=winAttrib.height-1;
			XWarpPointer(m_pDisplay, None, m_wndRoot, 0, 0, 0, 0, m_iMStateXBufLast2, m_iMStateYBufLast2);
			XFlush(m_pDisplay);
		} else {
			m_iMStateXBufLast2 = pMouseEvent->xmotion.x;
			m_iMStateYBufLast2 = pMouseEvent->xmotion.y;
			m_iMStateXBufLast = pMouseEvent->xmotion.x;
			m_iMStateYBufLast = pMouseEvent->xmotion.y;
		}

		S_MouseEvent stEvent;
		stEvent.iX = pMouseEvent->xmotion.x;
		stEvent.iY = pMouseEvent->xmotion.y;
		if(pMouseEvent->type==MotionNotify) {
			stEvent.iEvent = EVT_MOUSE_MOVE;
			m_pclMouseEvents->push_back(stEvent);
		} else if(pMouseEvent->type==ButtonPress) {
			switch(pMouseEvent->xbutton.button) {
				case Button1: stEvent.iEvent = EVT_MOUSE_LEFT_DOWN;  m_pclMouseEvents->push_back(stEvent); break;
				case Button3: stEvent.iEvent = EVT_MOUSE_RIGHT_DOWN; m_pclMouseEvents->push_back(stEvent); break;
				case Button2: stEvent.iEvent = EVT_MOUSE_MID_DOWN;   m_pclMouseEvents->push_back(stEvent); break;
				case Button4: stEvent.iEvent = EVT_MOUSE_WHEEL_UP;   stEvent.iExtraVal = -1; m_pclMouseEvents->push_back(stEvent); break;
				case Button5: stEvent.iEvent = EVT_MOUSE_WHEEL_DOWN; stEvent.iExtraVal = 1;  m_pclMouseEvents->push_back(stEvent); break;
			}
		} else if(pMouseEvent->type==ButtonRelease) {
			switch(pMouseEvent->xbutton.button) {
				case Button1: stEvent.iEvent = EVT_MOUSE_LEFT_UP;  m_pclMouseEvents->push_back(stEvent); break;
				case Button3: stEvent.iEvent = EVT_MOUSE_RIGHT_UP; m_pclMouseEvents->push_back(stEvent); break;
				case Button2: stEvent.iEvent = EVT_MOUSE_MID_UP;   m_pclMouseEvents->push_back(stEvent); break;
				case Button4: stEvent.iEvent = EVT_MOUSE_WHEEL_UP;   stEvent.iExtraVal = -1; m_pclMouseEvents->push_back(stEvent); break;
				case Button5: stEvent.iEvent = EVT_MOUSE_WHEEL_DOWN; stEvent.iExtraVal = 1;  m_pclMouseEvents->push_back(stEvent); break;
			}
		}
		if(m_pclMouseEvents->size()>12) m_pclMouseEvents->pop_front();
	}
}

#endif

#endif
