// C_Sound

//default
//a class to output wave and mp3/ogg sounds using the fmod library.
// supports windows linux and mac.

//ifdef _PA_SOUND_
//a class to output wave and .ogg sounds using the portaudio library
// and http://www.nothings.org/stb_vorbis/.
//supports windows linux and mac.
//useful in a game for example, where you have control over what sound files you play.
//wav - 44100Hz, 16bit mono
//ogg - 44100Hz, mono or stereo
//output - all mixed 44100Hz, stereo

#include "fileresource.h"
#include "soundfx.h"

C_Sound *C_Sound::s_pclTheSound = NULL;

#ifdef _NO_SOUND_

C_Sound::C_Sound() {}
C_Sound::~C_Sound() {}
bool C_Sound::Init(char *i_szDiskWriteFilename) { return true; }
int C_Sound::LoadWav(char *i_szFilename, char *i_szResname) { return 1; }
bool C_Sound::LoadWav(char *i_szFilename, char *i_szResname, int i_iIndex) { return true; }
void C_Sound::Unload(int i_iIndex) {}
void *C_Sound::Play(int i_iIndex, float i_fVolume, float i_fPanning, bool i_bLooping,
	int i_iLoopStartMs, int i_iLoopEndMs) { return NULL; }
void C_Sound::Stop(void *i_pChannel) {}
void C_Sound::Update() {}
void C_Sound::SetMasterVolume(float i_fVolume) {}
void C_Sound::ToggleSoundsOnOff() {}
void C_Sound::SetVolume(void *i_pChannel, float i_fVolume) {}
void C_Sound::SetPanning(void *i_pChannel, float i_fPanning) {}
float C_Sound::GetVolume(void *i_pChannel) { return 0.0f; };
float C_Sound::GetPanning(void *i_pChannel) { return 0.0f; };

#elif defined _PA_SOUND_ //portaudio

C_Sound::C_Sound()
{
	int i;
	m_pStream = NULL;
	m_fMasterVolume = 1.0f;

	m_bOutputEnabled = true;
	for(i=0; i<MAX_INDEX; i++) {
		m_pSounds[i].iType = -1;
	}
	for(i=0; i<MAX_CHANNELS; i++) {
		m_pChannels[i].bPlaying = false;
		m_pChannels[i].pVStream = NULL;
	}
	s_pclTheSound = this;
	m_pFile = NULL;

	m_pclMutex = new C_Mutex();
	m_pclUpdateThread = NULL;
}

void C_Sound::ReleasePA()
{
	m_bThreadDone = true;
	if(m_pclUpdateThread!=NULL) {
		m_pclUpdateThread->WaitForThreadExit();
		m_pclUpdateThread = NULL;
	}

	int i;
	if(m_pStream) {
		Pa_StopStream(m_pStream);
		Pa_CloseStream(m_pStream);
		Pa_Terminate();

		m_pStream = NULL;
	}

	for(i=0; i<MAX_CHANNELS; i++) {
		if(m_pChannels[i].pVStream) stb_vorbis_close(m_pChannels[i].pVStream);
		m_pChannels[i].pVStream = NULL;
		m_pChannels[i].bPlaying = false;
	}
}

C_Sound::~C_Sound()
{
	ReleasePA();

	for(int i=0; i<MAX_INDEX; i++) {
		if(m_pSounds[i].iType != -1) {
			delete[] m_pSounds[i].pSndData;
			m_pSounds[i].iType = -1;
		}
	}

	delete m_pclMutex;

	s_pclTheSound = NULL;
}

bool C_Sound::Init(char *i_szDiskWriteFilename)
{
	ReleasePA();

	//disk writer code
	if(m_pFile) {
		strncpy(m_stHeader.szRiff, "RIFF", 4);
		m_stHeader.iChunkSize = m_stHeader.iTotLength-36;
		strncpy(m_stHeader.szWaveFmt, "WAVEfmt ", 8);
		m_stHeader.iSizeFmt = 16;
		
		m_stHeader.wFormatTag = 1;
		m_stHeader.nChannels = 2;
		m_stHeader.nSamplesPerSec = 44100;
		m_stHeader.nAvgBytesPerSec = 176400;
		m_stHeader.nBlockAlign = 4;
		m_stHeader.wBitsPerSample = 16;

		strncpy(m_stHeader.szData, "data", 4);

		fseek(m_pFile, 0, SEEK_SET);
		fwrite(&m_stHeader, sizeof(m_stHeader), 1, m_pFile);
		fclose(m_pFile);
	}
	if(i_szDiskWriteFilename) {
		m_pFile = fopen(i_szDiskWriteFilename, "wb");
		memset(&m_stHeader, 0, sizeof(m_stHeader));
		fwrite(&m_stHeader, sizeof(m_stHeader), 1, m_pFile); //header rewritten later
	} else {
		m_pFile = NULL;
	}

	//init portaudio
	PaError iError = Pa_Initialize();
	if(iError != paNoError) goto error;

	PaStreamParameters stOutParameters;
	stOutParameters.device = Pa_GetDefaultOutputDevice(); // default output device
	if(stOutParameters.device == paNoDevice) {
		fprintf(stderr, "Error: No default output device.\n");
		goto error;
	}
	stOutParameters.channelCount = 2;         // stereo output
	stOutParameters.sampleFormat = paFloat32; // 32 bit floating point output
	stOutParameters.suggestedLatency = Pa_GetDeviceInfo(stOutParameters.device)->defaultLowOutputLatency;
	stOutParameters.hostApiSpecificStreamInfo = NULL;

	iError = Pa_OpenStream(
		&m_pStream,
		NULL, // no input
		&stOutParameters,
		SAMPLE_RATE,
		FRAMES_PER_BUFFER,
		paNoFlag,
		paCallback,
		NULL);
	if(iError != paNoError) goto error;

	//start output
	iError = Pa_StartStream(m_pStream);
	if(iError != paNoError) goto error;

	//start update thread
	m_bThreadDone = false;
	m_pclUpdateThread = new C_Thread(UpdateThread, this);

	return true;
error:
	ReleasePA();
	return false;
}

/*
 * This routine will be called by the PortAudio engine when audio is needed.
 * It may be called at interrupt level on some machines so don't do anything
 * that could mess up the system like calling malloc() or free().
 */
int C_Sound::paCallback(const void *inputBuffer, void *outputBuffer,
	unsigned long framesPerBuffer,
	const PaStreamCallbackTimeInfo* timeInfo,
	PaStreamCallbackFlags statusFlags,
	void *userData)
{
	float *out = (float*)outputBuffer;
	unsigned long i, j;

	(void)timeInfo; //prevent unused variable warnings
	(void)statusFlags;
	(void)inputBuffer;

	C_Sound *pSnd = C_Sound::GetSoundObject();

	//reset
	for(i=0; i<framesPerBuffer*2; i++)	{
		out[i] = 0;
	}

	for(j=0; j<MAX_CHANNELS; j++) {
		S_Channel *pCh = &pSnd->m_pChannels[j];
		if(pCh->bPlaying) {
			int idx = pCh->iSndIndex;
			float fLeft = pSnd->m_fMasterVolume*pCh->fVol*pCh->fPanL;
			float fRight = pSnd->m_fMasterVolume*pCh->fVol*pCh->fPanR;
			
			if(pSnd->m_pSounds[idx].iType == 0) {
				//sound data is wave
				for(i=0; i<framesPerBuffer; i++) {
					//fade out until 5% reached, then stop (reaches 5% of volume in 300 samples)
					if(pCh->bStopping) pCh->fFadeOut *= 0.99f;
					
					float fS = MIXER_SCALE*pSnd->m_pSounds[idx].pSndData[pCh->iInPos] * pCh->fFadeOut;

					out[i * 2 + 0] += fLeft*fS;
					out[i * 2 + 1] += fRight*fS;

					pCh->iInPos += 1;
					if(pCh->iInPos >= pSnd->m_pSounds[idx].iLength) {
						if(!pCh->bLooping) {
							pCh->bPlaying = false;
							break;
						} else pCh->iInPos = pCh->iLoopStartPos;
					}
				}
			} else {
				//sound data is ogg
				
				//consume the input queue
				uint32_t iLength = PaUtil_GetRingBufferReadAvailable(&pCh->rBufToRT);
				if(iLength >= framesPerBuffer * 2) iLength = framesPerBuffer * 2;
				PaUtil_ReadRingBuffer(&pCh->rBufToRT, &pSnd->m_aBuf1[0], iLength);

				for(i=0; i<iLength/2; i++) {
					//fade out until 5% reached, then stop (reaches 5% of volume in 300 samples)
					if(pCh->bStopping) pCh->fFadeOut *= 0.99f;

					out[i * 2 + 0] += MIXER_SCALE*fLeft*pSnd->m_aBuf1[i * 2 + 0] * pCh->fFadeOut;
					out[i * 2 + 1] += MIXER_SCALE*fRight*pSnd->m_aBuf1[i * 2 + 1] * pCh->fFadeOut;
				}
				if(pCh->bEndReached) {
					if(!pCh->bLooping) {
						pCh->bPlaying = false;
						break;
					}
				}
			}

			if(pCh->bStopping && pCh->fFadeOut<0.05f) pCh->bPlaying = false; //stop
		}
	}

	//disk writer (breaks the rule of doing as little as possible in this callback)
	if(pSnd->m_pFile) {
		pSnd->m_stHeader.iTotLength += framesPerBuffer * 2*2;

		for(i=0; i<framesPerBuffer*2; i++) {
			float s = out[i]*32767;
			if(s>32767) s = 32767;
			if(s<-32767) s = -32767;
			pSnd->m_aBuf16[i] = (int16_t)s;
		}
		fwrite(pSnd->m_aBuf16, framesPerBuffer * 2 * 2, 1, pSnd->m_pFile);
	}

	return paContinue;
}

THREAD_RET C_Sound::UpdateThread(void *i_pclThis)
{
	C_Sound *pThis = (C_Sound*)i_pclThis;

	while(!pThis->m_bThreadDone) {
		pThis->m_pclMutex->Enter();
		pThis->UpdateLL();
		pThis->m_pclMutex->Leave();

		mssleep(10);
	}
	return THREAD_RET_0;
}

void C_Sound::Update()
{
	//do nothing sice PA updates are run from callback
}

void C_Sound::UpdateLL(int i_iChannel)
{
	//ogg only, fill up the fifo consumed in the callback
	for(int i=0; i<MAX_CHANNELS; i++) {
		S_Channel *pCh = &m_pChannels[i];
		if(pCh->bPlaying || i==i_iChannel) {
			int idx = pCh->iSndIndex;
			if(m_pSounds[idx].iType == 1) {
				int iNumCh, n;
				float **outputs;
				while(PaUtil_GetRingBufferReadAvailable(&pCh->rBufToRT)<OGG_FIFO_LOW && !pCh->bEndReached) {
					int iUsed = stb_vorbis_decode_frame_pushdata(pCh->pVStream, m_pSounds[idx].pOggData + pCh->iInPos,
						m_pSounds[idx].iLength - pCh->iInPos, &iNumCh, &outputs, &n);
					if(iUsed == 0) {
						if (!pCh->bLooping) pCh->bEndReached = true;
						else {
							stb_vorbis_flush_pushdata(pCh->pVStream);
							pCh->iInPos = pCh->iLoopStartPos; //restart at the loop start
						}
					}
					pCh->iInPos += iUsed;
					if (n == 0) continue;
					pCh->iOutPos += n;
					if(pCh->iLoopStartPos == -1 && ((pCh->iOutPos * 1000) / 44100) > pCh->iLoopStartMs) {
						pCh->iLoopStartPos = pCh->iInPos;
					}

					for(i=0; i<n; i++) {
						m_aBuf2[i * 2 + 0] = outputs[0][i];
						m_aBuf2[i * 2 + 1] = (iNumCh >= 2) ? outputs[1][i] : outputs[0][i];
					}
					PaUtil_WriteRingBuffer(&pCh->rBufToRT, m_aBuf2, n * 2);
				}
			}
		}
	}
}

bool C_Sound::LoadWav(char *i_szFilename, char *i_szResname, int i_iIndex)
{
	S_WaveHdr stHDR;
	S_LoadedSound *pSnd = &m_pSounds[i_iIndex];
	uint32_t iSize;
	char     *pBuffer = NULL;
	char     *szExt = NULL;

	bool bResult = false;
	if(i_iIndex >= MAX_INDEX) return false;

	//unload previouse file (if any)
	if(pSnd->iType != -1) {
		delete[] pSnd->pSndData;
		pSnd->iType = -1;
	}
	delete m_pszFilenames[i_iIndex]; m_pszFilenames[i_iIndex] = NULL;
	delete m_pszResnames[i_iIndex]; m_pszResnames[i_iIndex] = NULL;

	C_Resource *pclResource = new C_Resource();
	if(!pclResource->SetFilename(i_szFilename, i_szResname)) goto error;
	iSize = pclResource->GetSize();
	pBuffer = new char[iSize];
	if(!pclResource->Read(pBuffer, iSize)) goto error; //read wavedata
	
	szExt = strrchr(i_szFilename, '.');
	if(szExt && _stricmp(szExt, ".ogg") == 0) {
		pSnd->iType = 1;

		pSnd->pOggData = new uint8_t[iSize];
		memcpy(pSnd->pOggData, pBuffer, iSize);
		pSnd->iLength = iSize;
	} else {
		memcpy(&stHDR, pBuffer, sizeof(S_WaveHdr));

		if(strncmp(stHDR.szRiff, "RIFF", 4) != 0) goto error;
		if(strncmp(stHDR.szWaveFmt, "WAVEfmt ", 8) != 0) goto error;

		int iNumSamples = stHDR.iTotLength / 2;

		pSnd->iType = 0;
		pSnd->iLength = iNumSamples;
		pSnd->pSndData = new float[iNumSamples];
		int16_t *pSndData16 = (int16_t*)pBuffer+sizeof(S_WaveHdr);

		for(int i=0; i<iNumSamples; i++) {
			pSnd->pSndData[i] = pSndData16[i] / 32768.0f;
		}
	}

	//save filenames
	if(i_szFilename) {
		m_pszFilenames[i_iIndex] = new char[strlen(i_szFilename)+1];
		strcpy(m_pszFilenames[i_iIndex], i_szFilename);
	}
	if(i_szResname) {
		m_pszResnames[i_iIndex] = new char[strlen(i_szResname)+1];
		strcpy(m_pszResnames[i_iIndex], i_szResname);
	}

	bResult = true;
error:
	delete pclResource;
	delete []pBuffer;
	return bResult;
}

int C_Sound::LoadWav(char *i_szFilename, char *i_szResname)
{
	bool bFound = false;
	int iFirstFreeIndex = MAX_INDEX;
	int i;
	for(i=0; i<MAX_INDEX; i++) {
		if(!m_pszFilenames[i] && !m_pszResnames[i] && iFirstFreeIndex>i) iFirstFreeIndex = i; //is first free index?
		//is same resource?
		if((!i_szResname && m_pszResnames[i]) || (i_szResname && !m_pszResnames[i])) continue;
		if(i_szResname != m_pszResnames[i]) {
			if(strcmp(i_szResname, m_pszResnames[i]) != 0) continue;
		}

		//is same file
		if((!i_szFilename && m_pszFilenames[i]) || (i_szFilename && !m_pszFilenames[i])) continue;
		if(i_szFilename != m_pszFilenames[i]) {
			if(strcmp(i_szFilename, m_pszFilenames[i]) != 0) continue;
		}

		bFound = true;
		break;
	}
	if(!bFound) { //load it
		if(LoadWav(i_szFilename, i_szResname, iFirstFreeIndex)) return iFirstFreeIndex;
		else return -1;
	}
	return i;
}


void C_Sound::Unload(int i_iIndex)
{
	S_LoadedSound *pSnd = &m_pSounds[i_iIndex];
	if(pSnd->iType != -1) {
		delete[] pSnd->pSndData;
		pSnd->iType = -1;
	}
}

void *C_Sound::Play(int i_iIndex, float i_fVolume, float i_fPanning, bool i_bLooping, int i_iLoopStartMs)
{
	if(!m_bOutputEnabled) return NULL;

	if(m_pSounds[i_iIndex].iType == -1) return NULL;

	for(int i=0; i<MAX_CHANNELS; i++) {
		if(!m_pChannels[i].bPlaying) {
			//use the first free channel found

			//free any previous stream running on this channel
			if(m_pChannels[i].pVStream) {
				stb_vorbis_close(m_pChannels[i].pVStream);
				m_pChannels[i].pVStream = NULL;
			}

			m_pChannels[i].iOutPos = 0;
			m_pChannels[i].bLooping = i_bLooping;
			m_pChannels[i].iLoopStartMs = i_iLoopStartMs;
			m_pChannels[i].iLoopStartPos = -1; //before it is known

			if(m_pSounds[i_iIndex].iType == 1) {
				//sound data is ogg
				int iUsed, iError;
				m_pChannels[i].pVStream = stb_vorbis_open_pushdata(m_pSounds[i_iIndex].pOggData,
					m_pSounds[i_iIndex].iLength, &iUsed, &iError, NULL);
//printf("PLAY %d %d\n", iError, iUsed);
				m_pChannels[i].iInPos = iUsed;
				m_pChannels[i].bEndReached = false;

				//initialize communication buffer (fifo)
				PaUtil_InitializeRingBuffer(&m_pChannels[i].rBufToRT, sizeof(float),
					32768, m_pChannels[i].rBufToRTData);
			} else {
				//sound data is wave
				m_pChannels[i].iInPos = 0;
				m_pChannels[i].iLoopStartPos = (int)((i_iLoopStartMs*SAMPLE_RATE)/1000);
			}

			m_pChannels[i].iSndIndex = i_iIndex;
			SetVolume(&m_pChannels[i], i_fVolume); ;
			SetPanning(&m_pChannels[i], i_fPanning);
			m_pChannels[i].bStopping = false;
			m_pChannels[i].fFadeOut = 1.0f;

			//fill buffer for ogg sound this first time
			m_pclMutex->Enter();
			UpdateLL(i);
			m_pclMutex->Leave();

			//start playing when all is set
			m_pChannels[i].bPlaying = true;

			return &m_pChannels[i];
		}
	}

	return NULL;
}

void C_Sound::Stop(void *i_pChannel)
{
	S_Channel *pCh = (S_Channel*)i_pChannel;
	if(pCh==NULL) return;

	//fade out will be done on this channel, then it will stop playing
	pCh->bStopping = true;
}

void C_Sound::ToggleSoundsOnOff()
{
	m_bOutputEnabled = !m_bOutputEnabled;
}

void C_Sound::SetMasterVolume(float i_fVolume)
{
	if(i_fVolume > 1.0f) i_fVolume = 1.0f; //max
	if(i_fVolume < 0.0f) i_fVolume = 0.0f; //min

	m_fMasterVolume = i_fVolume;
}

void C_Sound::SetVolume(void *i_pChannel, float i_fVolume)
{
	if(i_fVolume > 1.0f) i_fVolume = 1.0f; //max
	if(i_fVolume < 0.0f) i_fVolume = 0.0f; //min

	if(i_pChannel) {
		((S_Channel*)i_pChannel)->fVol = i_fVolume;
	}
}

float C_Sound::GetVolume(void *i_pChannel)
{
	float fVol = 0.0f;
	if (i_pChannel) {
		fVol = ((S_Channel*)i_pChannel)->fVol;
	}
	return fVol;
}

void C_Sound::SetPanning(void *i_pChannel, float i_fPanning)
{
	if(i_fPanning > 1.0f) i_fPanning = 1.0f; //max
	if(i_fPanning < -1.0f) i_fPanning = -1.0f; //min
	if(i_pChannel) {
		if(i_fPanning>0) {
			((S_Channel*)i_pChannel)->fPanL = 1.0f - i_fPanning;
			((S_Channel*)i_pChannel)->fPanR = 1.0f;
		} else {
			((S_Channel*)i_pChannel)->fPanL = 1.0f;
			((S_Channel*)i_pChannel)->fPanR = 1.0f + i_fPanning;
		}
	}
}

#else //fmod

C_Sound::C_Sound()
{
	int i;
	m_pclSoundSys = NULL;
	m_fMasterVolume = 1.0f;

	m_bOutputEnabled = true;
	for(i=0; i<MAX_INDEX; i++) {
		m_pszResnames[i]  = NULL;
		m_pszFilenames[i] = NULL;
		m_pclSounds[i]    = NULL;
		m_pSoundMem[i]    = NULL;
	}
	s_pclTheSound = this;
	m_bDiskWrite = false;
}

void C_Sound::ReleaseFMOD()
{
	int i;
	for(i=0; i<MAX_INDEX; i++) {
		if(m_pclSounds[i]) {
			FMOD_Sound_Release(m_pclSounds[i]);
			m_pclSounds[i] = NULL;
			delete[] m_pSoundMem[i]; m_pSoundMem[i] = NULL;
		}
	}

	if(m_pclSoundSys) {
		FMOD_System_Close(m_pclSoundSys);
		FMOD_System_Release(m_pclSoundSys);
		m_pclSoundSys = NULL;
	}
}

C_Sound::~C_Sound()
{
	ReleaseFMOD();

	int i;
	for(i=0; i<MAX_INDEX; i++) {
		delete m_pszResnames[i]; m_pszResnames[i] = NULL;
		delete m_pszFilenames[i]; m_pszFilenames[i] = NULL;
	}

	s_pclTheSound = NULL;
}

void C_Sound::ReloadSounds()
{
	char szResname[MAX_PATH];
	char szFilename[MAX_PATH];
	int i;
	for(i=0; i<MAX_INDEX; i++) {
		szResname[0] = 0;
		szFilename[0] = 0;
		if(m_pszResnames[i]) strcpy(szResname, m_pszResnames[i]);
		if(m_pszFilenames[i]) strcpy(szFilename, m_pszFilenames[i]);
		if(szResname[0]!=0 || szFilename[0]!=0) LoadWav(szFilename, szResname, i);
	}
}

bool C_Sound::Init(char *i_szDiskWriteFilename)
{
	ReleaseFMOD();
	if(FMOD_System_Create(&m_pclSoundSys) != FMOD_OK) goto error;
	if(i_szDiskWriteFilename) {
		FMOD_System_SetOutput(m_pclSoundSys, FMOD_OUTPUTTYPE_WAVWRITER);
//		FMOD_System_SetSoftwareFormat(m_pclSoundSys, 44100, FMOD_SOUND_FORMAT_PCM16, 2, MAX_CHANNELS, FMOD_DSP_RESAMPLER_LINEAR);
		m_bDiskWrite = true;
	} else m_bDiskWrite = false;
	if(FMOD_System_Init(m_pclSoundSys, MAX_CHANNELS, FMOD_INIT_NORMAL, i_szDiskWriteFilename) != FMOD_OK) goto error;

	ReloadSounds(); //now takes an acceptable amount of time (800ms->9ms) with the use of CreateStream

	return true;
error:
	ReleaseFMOD();
	return false;
}

void C_Sound::Update()
{
	if(m_pclSoundSys) FMOD_System_Update(m_pclSoundSys);
}

bool C_Sound::LoadWav(char *i_szFilename, char *i_szResname, int i_iIndex)
{
	uint32_t iSize;
	char     *pBuffer = NULL;
	char     *szExt = NULL;
	bool     bKeepMem = false;

	FMOD_CREATESOUNDEXINFO stInfo;

	if(i_iIndex>=MAX_INDEX || !m_pclSoundSys) return false;

	//unload previous file (if any)
	if(m_pclSounds[i_iIndex]) {
		FMOD_Sound_Release(m_pclSounds[i_iIndex]);
		m_pclSounds[i_iIndex] = NULL;
	}
	delete m_pszFilenames[i_iIndex]; m_pszFilenames[i_iIndex] = NULL;
	delete m_pszResnames[i_iIndex]; m_pszResnames[i_iIndex] = NULL;

	C_Resource *pclResource = new C_Resource();
	if(!pclResource->SetFilename(i_szFilename, i_szResname)) goto error;
	iSize = pclResource->GetSize();
	pBuffer = new char[iSize];
	if(!pclResource->Read(pBuffer, iSize)) goto error; //read wavedata

	memset(&stInfo, 0, sizeof(FMOD_CREATESOUNDEXINFO));
	stInfo.cbsize = sizeof(FMOD_CREATESOUNDEXINFO);
	stInfo.length = iSize;

	szExt = strrchr(i_szFilename, '.');
	if(szExt && (_stricmp(szExt, ".mp3")==0 || _stricmp(szExt, ".ogg")==0) && iSize>200 * 1024) {
		if(FMOD_System_CreateStream(m_pclSoundSys, (const char *)pBuffer,
			(m_bDiskWrite?FMOD_SOFTWARE:FMOD_HARDWARE) | FMOD_OPENMEMORY, &stInfo, &m_pclSounds[i_iIndex]) != FMOD_OK) goto error;
		bKeepMem = true;
	} else {
		if(FMOD_System_CreateSound(m_pclSoundSys, (const char *)pBuffer,
			(m_bDiskWrite?FMOD_SOFTWARE:FMOD_HARDWARE) | FMOD_OPENMEMORY, &stInfo, &m_pclSounds[i_iIndex]) != FMOD_OK) goto error;
	}
 
	FMOD_Sound_SetMode(m_pclSounds[i_iIndex], FMOD_LOOP_OFF);

	//save filenames
	if(i_szFilename) {
		m_pszFilenames[i_iIndex] = new char[strlen(i_szFilename)+1];
		strcpy(m_pszFilenames[i_iIndex], i_szFilename);
	}
	if(i_szResname) {
		m_pszResnames[i_iIndex] = new char[strlen(i_szResname)+1];
		strcpy(m_pszResnames[i_iIndex], i_szResname);
	}

	delete pclResource;
	m_pSoundMem[i_iIndex] = NULL;
	if(bKeepMem) m_pSoundMem[i_iIndex] = pBuffer; //pBuffer needs to be kept, playback is done from this memory
	else delete[] pBuffer;

	return true;
error:
	delete pclResource;
	delete []pBuffer;
	return false;
}

int C_Sound::LoadWav(char *i_szFilename, char *i_szResname)
{
	bool bFound = false;
	int iFirstFreeIndex = MAX_INDEX;
	int i;
	for(i=0; i<MAX_INDEX; i++) {
		if(!m_pszFilenames[i] && !m_pszResnames[i] && iFirstFreeIndex>i) iFirstFreeIndex = i; //is first free index?
		//is same resource?
		if((!i_szResname && m_pszResnames[i]) || (i_szResname && !m_pszResnames[i])) continue;
		if(i_szResname != m_pszResnames[i]) {
			if(strcmp(i_szResname, m_pszResnames[i]) != 0) continue;
		}

		//is same file
		if((!i_szFilename && m_pszFilenames[i]) || (i_szFilename && !m_pszFilenames[i])) continue;
		if(i_szFilename != m_pszFilenames[i]) {
			if(strcmp(i_szFilename, m_pszFilenames[i]) != 0) continue;
		}

		bFound = true;
		break;
	}
	if(!bFound) { //load it
		if(LoadWav(i_szFilename, i_szResname, iFirstFreeIndex)) return iFirstFreeIndex;
		else return -1;
	}
	return i;
}

void C_Sound::Unload(int i_iIndex)
{
	if(m_pclSounds[i_iIndex]) FMOD_Sound_Release(m_pclSounds[i_iIndex]);
}

void *C_Sound::Play(int i_iIndex, float i_fVolume, float i_fPanning, bool i_bLooping, int i_iLoopStartMs)
{
	if(!m_bOutputEnabled) return NULL;

	if(m_pclSounds[i_iIndex] == NULL) return NULL;

	FMOD_CHANNEL *pclChannel = NULL;
	FMOD_System_PlaySound(m_pclSoundSys, FMOD_CHANNEL_FREE, m_pclSounds[i_iIndex], 0, &pclChannel);
	if(pclChannel && i_bLooping) {
		FMOD_Channel_SetMode(pclChannel, FMOD_LOOP_NORMAL);
		int iLoopEndMs = -1;
		if(i_iLoopStartMs!=0 || iLoopEndMs!=-1) {
			unsigned int iLength;
			FMOD_Sound_GetLength(m_pclSounds[i_iIndex], &iLength, FMOD_TIMEUNIT_MS);
			if(i_iLoopStartMs<0) i_iLoopStartMs = 0;
			if(iLoopEndMs<0)   iLoopEndMs   = (int)iLength;
			FMOD_Channel_SetLoopPoints(pclChannel, i_iLoopStartMs, FMOD_TIMEUNIT_MS, iLoopEndMs, FMOD_TIMEUNIT_MS);
		}
	}

	if(pclChannel) FMOD_Channel_SetPan(pclChannel, i_fPanning);
	if(pclChannel) FMOD_Channel_SetVolume(pclChannel, m_fMasterVolume*i_fVolume);
	//^should realy be done before we start playing, don't know how yet.

	return (void*)pclChannel;
}

void C_Sound::Stop(void *i_pChannel)
{
	if(i_pChannel) FMOD_Channel_Stop((FMOD_CHANNEL*)i_pChannel);
}

void C_Sound::SetMasterVolume(float i_fVolume)
{
	m_fMasterVolume = i_fVolume;
}

void C_Sound::SetVolume(void *i_pChannel, float i_fVolume)
{
	if(i_pChannel) {
		FMOD_Channel_SetVolume((FMOD_CHANNEL*)i_pChannel, m_fMasterVolume*i_fVolume);
	}
}

float C_Sound::GetVolume(void *i_pChannel)
{
	float fVol = 0.0f;
	if (i_pChannel) {
		FMOD_Channel_GetVolume((FMOD_CHANNEL*)i_pChannel, &fVol);
		fVol /= m_fMasterVolume;
	}
	return fVol;
}

void C_Sound::ToggleSoundsOnOff()
{
	m_bOutputEnabled = !m_bOutputEnabled;
}

void C_Sound::SetPanning(void *i_pChannel, float i_fPanning)
{
	if(i_pChannel) {
		FMOD_Channel_SetPan((FMOD_CHANNEL*)i_pChannel, i_fPanning);
	}
}

float C_Sound::GetPanning(void *i_pChannel)
{
	float fPan = 0.0f;
	if (i_pChannel) {
		FMOD_Channel_GetPan((FMOD_CHANNEL*)i_pChannel, &fPan);
	}
	return fPan;
}

#endif
