// C_Sound

//define _NO_SOUND_ to skip fmod includes

#ifndef _SOUNDFX_H
#define _SOUNDFX_H

#define MAX_INDEX    32
#define MAX_CHANNELS 16

//#define _NO_SOUND_

#include "common.h"

#if defined _PA_SOUND_

#define SAMPLE_RATE        44100
#define FRAMES_PER_BUFFER  2048   //stereo 44100 -> 46ms (4096 samples)
#define OGG_FIFO_LOW       8000   //stereo 44100 -> 91ms (8000 samples)
#define MIXER_SCALE        0.85f

#include "common.h"

#include "portaudio.h"
#include "pa_ringbuffer.h"
#define STB_VORBIS_HEADER_ONLY
#include "stb_vorbis.c"

struct S_LoadedSound
{
	int iType; //-1 not used, 0 = WAV 44 16 MONO, 1 = OGG stream 44 ST
	union {
		float *pSndData;
		uint8_t *pOggData;
	};

	int iLength; //bytes
};

struct S_Channel
{
	//common
	volatile bool bPlaying;
	volatile bool bStopping;
	float fFadeOut;
	float fPanL, fPanR, fVol;
	int iSndIndex;
	int iInPos; //pos in input
	int iOutPos; //pos in output

	bool bLooping;
	int iLoopStartMs;
	int iLoopStartPos;

	//ogg stream
	stb_vorbis *pVStream;
	bool bEndReached;
	//ring buffer (FIFO) towards audio callback
	PaUtilRingBuffer rBufToRT;
	float            rBufToRTData[32768];
};

#pragma pack(1)
struct S_WaveHdr {
	char        szRiff[4];       //'RIFF'
	uint32_t    iChunkSize;      //not important
	char        szWaveFmt[8];    //'WAVEfmt '
	uint32_t    iSizeFmt;        //not important

	uint16_t    wFormatTag;      // format type
	uint16_t    nChannels;       // number of channels (i.e. mono, stereo...)
	uint32_t    nSamplesPerSec;  // sample rate
	uint32_t    nAvgBytesPerSec; // for buffer estimation
	uint16_t    nBlockAlign;     // block size of data
	uint16_t    wBitsPerSample;  // number of bits per sample of mono data
	//uint16_t    cbSize;          // the count in bytes of the size of extra information (after cbSize)

	char        szData[4];       //not important
	uint32_t    iTotLength;      //smp data length
};
#pragma pack()

#elif defined _NO_SOUND_
#else
#include "fmod.h"
#endif

class C_Sound {
public:
	C_Sound();
	~C_Sound();
	bool Init(char *i_szDiskWriteFilename = NULL);

	static C_Sound *GetSoundObject() {return s_pclTheSound;};

	int LoadWav(char *i_szFilename, char *i_szResname); //returns index of a wave loaded (-1 for error)
	bool LoadWav(char *i_szFilename, char *i_szResname, int i_iIndex);
	void Unload(int i_iIndex);

	void *Play(int i_iIndex, float i_fVolume = 1.0f, float i_fPanning = 0.0f, bool i_bLooping = false, int i_iLoopStartMs = 0); //returns channel pointer (only useful for looping sounds)
	void Stop(void *i_pChannel);
	void Update();

	void SetMasterVolume(float i_fVolume);
	void ToggleSoundsOnOff();

	void SetVolume(void *i_pChannel, float i_fVolume);
	void SetPanning(void *i_pChannel, float i_fPanning);

	float GetVolume(void *i_pChannel);
	float GetPanning(void *i_pChannel);

private:
	static C_Sound *s_pclTheSound; //there can only be one sound obj and this is it
#if defined _PA_SOUND_
	S_LoadedSound m_pSounds[MAX_INDEX];
	S_Channel m_pChannels[MAX_CHANNELS];
	float m_fMasterVolume;
	bool m_bOutputEnabled;

	static int paCallback(const void *inputBuffer, void *outputBuffer,
		unsigned long framesPerBuffer,
		const PaStreamCallbackTimeInfo* timeInfo,
		PaStreamCallbackFlags statusFlags,
		void *userData);

	void ReleasePA();

	PaStream *m_pStream; //pa output stream 44100Hz stereo

	float m_aBuf1[FRAMES_PER_BUFFER * 2]; //temp buf used in callback for ogg (stereo data), read fifo
	float m_aBuf2[FRAMES_PER_BUFFER * 2]; //temp buf used in update for ogg (stereo data), write fifo

	//disk writer
	int16_t m_aBuf16[FRAMES_PER_BUFFER * 2]; //temp buf
	S_WaveHdr m_stHeader;
	FILE *m_pFile;

	char *m_pszResnames[MAX_INDEX];
	char *m_pszFilenames[MAX_INDEX];

	//update thread
	bool m_bThreadDone;
	C_Mutex *m_pclMutex;
	C_Thread *m_pclUpdateThread;
	void UpdateLL(int i_iChannel = -1);
	static THREAD_RET UpdateThread(void *i_pclThis);
#elif defined _NO_SOUND_
#else //fmod
	void ReleaseFMOD();
	void ReloadSounds();

	FMOD_SYSTEM    *m_pclSoundSys;
	FMOD_SOUND     *m_pclSounds[MAX_INDEX];
	//FMOD_CHANNEL   *m_pclChannels[MAX_CHANNELS];

	char *m_pszResnames[MAX_INDEX];
	char *m_pszFilenames[MAX_INDEX];
	char *m_pSoundMem[MAX_INDEX];

	float m_fMasterVolume;

	bool m_bOutputEnabled;
	bool m_bDiskWrite;
#endif
};

#endif
