// C_Object (for mapeditor)

//class for all objects in the in the mapeditor.
//keeps direction and position
// - a simple drawing method (to draw an image at a given map position)
// sometimes objects need a more complex drawing method, but can still use this for basic drawing.

#include <math.h>
#include "e_object.h"

S_Rect C_Object::s_stScreenArea;

C_Object::C_Object()
{
	m_iDirStep = 10;
	m_pclGraph = C_GraphWrapperGL::GetGraphWrapperObject();
	m_pclImage = NULL;
	m_pstImageRect = NULL;

	m_iOwnerID = -1;
}

void C_Object::Init(S_Rect *i_pstScreenArea)
{
	s_stScreenArea = *i_pstScreenArea;
}

void C_Object::GetPos(int *o_iX, int *o_iY)
{
	*o_iX = (int)m_dPosX;
	*o_iY = (int)m_dPosY;
}

void C_Object::SetPos(int i_iX, int i_iY)
{
	m_dPosX = i_iX;
	m_dPosY = i_iY;
}

bool C_Object::GetSize(int *o_iSizeW, int *o_iSizeH)
{
	if(!m_pclImage) {
		*o_iSizeW = *o_iSizeH = 0;
	} else if(m_pstImageRect) {
		*o_iSizeW = m_pstImageRect->width;
		*o_iSizeH = m_pstImageRect->height;
	} else m_pclImage->GetInfo(o_iSizeW, o_iSizeH, NULL);
	return true;
}

bool C_Object::GetCenterPos(int *o_iCenterX, int *o_iCenterY)
{
	int iSizeX, iSizeY;
	if(!GetSize(&iSizeX, &iSizeY)) return false;

	*o_iCenterX = (int)(m_dPosX + iSizeX*0.5);
	*o_iCenterY = (int)(m_dPosY + iSizeY*0.5);
	return true;
}

#define OUTLINE_W 2
void C_Object::ODraw(C_Image *i_pclImg, S_Rect *i_pstRect, int i_iX, int i_iY, S_FColor *i_pstColor, bool i_bOutline)
{
	if(i_bOutline) {
		int iWidth, iHeight;
		if(i_pstRect) {
			iWidth = i_pstRect->width;
			iHeight = i_pstRect->height;
		} else {
			i_pclImg->GetInfo(&iWidth, &iHeight, NULL);
		}

		int x, y, i, j, iLineBytesSrc, iLineBytesDst;
		uint8_t *pBufSrc, *pBufDst;
		iWidth += 2*OUTLINE_W;
		iHeight += 2*OUTLINE_W;
		C_Image *pclDstImg = new C_Image(iWidth, iHeight, _RGBA8888);
		C_Image *pclSrcImg = new C_Image(iWidth, iHeight, _RGBA8888);
		pclDstImg->FillWithAbsoluteColor(NULL, 0);
		pclSrcImg->FillWithAbsoluteColor(NULL, 0);
		pclSrcImg->Blt2(i_pclImg, i_pstRect, OUTLINE_W, OUTLINE_W);
		pclDstImg->GetBufferMemory(&pBufDst);
		pclDstImg->GetLineSizeInfo(NULL, &iLineBytesDst, NULL);
		pclSrcImg->GetBufferMemory(&pBufSrc);
		pclSrcImg->GetLineSizeInfo(NULL, &iLineBytesSrc, NULL);
		for(y=0; y<iHeight; y++) {
			for(x=0; x<iWidth; x++) {
				uint32_t *srcpixel = (uint32_t*)(pBufSrc + (y)*iLineBytesSrc+4*(x));
				if(((*srcpixel)&0xff000000)==0) { //alpha is 0
					for(j=-OUTLINE_W; j<=OUTLINE_W; j++) {
						for(i=-OUTLINE_W; i<=OUTLINE_W; i++) {
							if(x+i<iWidth && y+j<iHeight && x+i>=0 && y+j>=0) {
								srcpixel = (uint32_t*)(pBufSrc + (y+j)*iLineBytesSrc+4*(x+i));
								if(((*srcpixel)&0xff000000)!=0) { //alpha is >0
									uint32_t *dstpixel = (uint32_t*)(pBufDst + (y)*iLineBytesDst+4*(x));
									*dstpixel = 0x703030ff;
									j = OUTLINE_W+1;
									break;
								}
							}
						}
					}
				}
			}
		}
		ODrawInternal(pclDstImg, NULL, i_iX+OUTLINE_W, i_iY+OUTLINE_W);
		delete pclSrcImg;
		delete pclDstImg;
	}
	m_pclGraph->SetBltColor(i_pstColor);
	ODrawInternal(i_pclImg, i_pstRect, i_iX, i_iY);
	m_pclGraph->SetBltColor(NULL);
}


void C_Object::ODrawInternal(C_Image *i_pclImg, S_Rect *i_pstRect, int i_iX, int i_iY)
{
	if(!i_pclImg) return;

	bool bDraw = true;
	int iPosX = (int)m_dPosX - i_iX;
	int iPosY = (int)m_dPosY - i_iY;
	int iWidth, iHeight;
	S_Rect stRect;

	if(i_pstRect) {
		stRect = *i_pstRect;
		iWidth = stRect.width;
		iHeight = stRect.height;
	} else {
		i_pclImg->GetInfo(&iWidth, &iHeight, NULL);
		stRect.x = stRect.y = 0;
		stRect.width = iWidth;
		stRect.height = iHeight;
	}

	//clipping x
	if(iPosX < 0) {
		stRect.x -= iPosX;
		stRect.width += iPosX;
		iPosX = 0;
	} else if(iPosX+iWidth > s_stScreenArea.width) {
		stRect.width -= (iPosX+iWidth) - s_stScreenArea.width;
	}
	if(stRect.width <= 0) bDraw = false;
	else { //clipping y
		if(iPosY < 0) {
			stRect.y -= iPosY;
			stRect.height += iPosY;
			iPosY = 0;
		} else if(iPosY+iHeight > s_stScreenArea.height) {
			stRect.height -= (iPosY+iHeight) - s_stScreenArea.height;
		}
		if(stRect.height <= 0) bDraw = false;
	}
	//draw?
	if(bDraw)
		m_pclGraph->Blt(i_pclImg, &stRect, iPosX+s_stScreenArea.x, iPosY+s_stScreenArea.y, true);
}
