// C_Map (for mapeditor)

#ifndef TILEMAP_H
#define TILEMAP_H

#include "e_object.h"
#include "common.h"
#include "image.h"
#include "glgraph.h"

typedef enum
{
	MAP_MISSION = 1,
	MAP_DOGFIGHT = 2,
	MAP_RACE = 3,
	MAP_MISSION_COOP = 4
} MapType;

struct S_MapHeader
{
	char szGridfile[128];
	char szTileset[128];
	int iTileSize;
	int iWidth, iHeight;
};

class C_Map
{
public:
	C_Map(char *i_szGridpath, S_MapHeader *i_pstMapInfo, S_Rect *i_pstDrawarea, bool i_bNew);
	~C_Map();

	//common methods
	void SetMapWrapping(bool i_bHorizontal, bool i_bVertical);
	void MoveRelative(int i_iX, int i_iY); //pixel map coordinates
	void MoveAbsolute(int i_iX, int i_iY); //pixel map coordinates
	void GetPos(int *o_iX, int *o_iY) {*o_iX = m_iLastX; *o_iY = m_iLastY;};

	int GetTileSize() {return m_iTileSize;}; //returns the size of the tiles in tileset
	void SetScreenRect(S_Rect *i_pstDrawarea);
	S_Rect *GetScreenRect() {return &m_stDrawarea;};
	void GetMapSize(int *o_iWidth, int *o_iHeight) {*o_iWidth = m_iWidth*m_iTileSize; *o_iHeight = m_iHeight*m_iTileSize;};

	bool IsValidPos(int i_iMapTileX, int i_iMapTileY);
	void SetElementAbsolute(int i_iMapTileX, int i_iMapTileY, char i_cVal);
	int GetElementAbsolute(int i_iMapTileX, int i_iMapTileY);
	bool Load(S_MapHeader *i_pstMapInfo = NULL);
	bool Save();
	void Draw();

	void SetElement(int i_iMapTileX, int i_iMapTileY, char i_cVal);
	int GetElement(int i_iMapTileX, int i_iMapTileY);

	//editor specific methods
	int GetNumTiles() {return m_iNumTiles;}; //returns number of tiles in tileset
	void Clear();
	void Crop(int iSide);
	void Extend(int iSide);
	void DrawTile(int i_iX, int i_iY, char i_cVal);

private:
	//map grid info
	char **m_pGrid;
	int m_iWidth, m_iHeight;

	//draw info
	S_Rect m_stDrawarea;
	int m_iDrawareaWidthSave, m_iDrawareaHeightSave;
	int m_iPixMaxX, m_iPixMaxY;
	int m_iScrollMaxX, m_iScrollMaxY;
	int m_iLastX, m_iLastY;
	bool m_iWrappX, m_iWrappY;
	bool WrapAroundTest(int *o_iX, int *o_iY, int i_iWidth, int i_iHeight);
	void DrawTileClipped(int i_iTileX, int i_iTileY, char i_cVal);

	//tileset info
	void LoadTileSet();
	int       m_iTileSize;
	C_Image   *m_pclTilesImg, *m_pclTilesImg_Sysmem;
	int       m_iNumTiles;
	S_Rect    *m_stTileRects;

	//graph
	C_GraphWrapperGL *m_pclGraph;

	//file info
	S_MapHeader m_stHeader;
	char m_szGridpath[MAX_PATH];
};

#endif
