// C_MapBuilder

//class to build a map from a mask image

#include "mapbuilder.h"
#include "global.h"

C_MapBuilder::C_MapBuilder(char *i_szSavePath)
{
	m_pstTileRects = NULL;
	m_pclTilesImg = NULL;
	m_pTileData = NULL;
	m_bHeaderValid = false;

	strcpy(m_szSavePath, i_szSavePath);
}

C_MapBuilder::~C_MapBuilder()
{
	delete m_pclTilesImg;
	delete[] m_pstTileRects;
}

void C_MapBuilder::InitPossibleList() //supports only built in tilesets
{
	m_clPossibleList.clear();

	//fill with all valid tiles
	int i;
	for(i=0; i<=10; i++) m_clPossibleList.push_back(i);
	for(i=14; i<=24; i++) m_clPossibleList.push_back(i);
	for(i=29; i<=32; i++) m_clPossibleList.push_back(i);
	for(i=43; i<=46; i++) m_clPossibleList.push_back(i);
	for(i=57; i<=58; i++) m_clPossibleList.push_back(i);
	for(i=65; i<=66; i++) m_clPossibleList.push_back(i);
	for(i=71; i<=72; i++) m_clPossibleList.push_back(i);
	for(i=79; i<=80; i++) m_clPossibleList.push_back(i);
}

bool C_MapBuilder::LoadTileset()
{
	int i, j;
	bool bResult;

	//hardcoded tileset support
	char szTilesetFile[MAX_PATH];
	sprintf(szTilesetFile, "ts_alien.tga");
	m_iTileSize = 32;
	InitPossibleList();

	delete m_pclTilesImg;
	delete[] m_pstTileRects;
	m_pstTileRects = NULL;
	m_pclTilesImg = new C_Image(szTilesetFile, C_Global::szResourceFile, &bResult, _RGB555); //load as 16 bits
	if(!bResult) {
		delete m_pclTilesImg; m_pclTilesImg = NULL;
		return false;
	}

	//analyse tileset (get rectangles)
	int iWidth, iHeight, iNumX, iNumY, iTileNum;
	m_pclTilesImg->GetInfo(&iWidth, &iHeight, NULL);
	m_pclTilesImg->GetLineSizeInfo(NULL, &m_iTileLineWidth, NULL);
	m_pclTilesImg->GetBufferMemory((uint8_t**)&m_pTileData);
	iNumX = iWidth  / m_iTileSize;
	iNumY = iHeight / m_iTileSize;
	m_iTileLineWidth /= 2;
	m_iNumTiles = iNumX * iNumY;
	m_pstTileRects = new S_Rect[m_iNumTiles];

	iTileNum = 0;
	for(j=0; j<iNumY; j++) {
		for(i=0; i<iNumX; i++) {
			m_pstTileRects[iTileNum].x = i * m_iTileSize;
			m_pstTileRects[iTileNum].y = j * m_iTileSize;
			m_pstTileRects[iTileNum].width = m_iTileSize;
			m_pstTileRects[iTileNum].height = m_iTileSize;
			iTileNum++;
		}
	}
	return true;
}

double C_MapBuilder::GetTileMatchScore(C_Map *i_pclMap, int x, int y, int i_iTileNum, int i_iCheckNeighbours)
{
	double dScore=0.0, dValue;
	int iMax = (m_iTileSize-1);
	int iMaskPosX = x*m_iTileSize;
	int iMaskPosY = y*m_iTileSize;
	S_Rect *pstRect = &m_pstTileRects[i_iTileNum];
	uint16_t *pMaskbuf = m_pMaskData;
	uint16_t *pTilebuf = m_pTileData;

	//calc offset in mask
	pMaskbuf += iMaskPosY*m_iMaskLineWidth + iMaskPosX;
	//calc offset in tileset
	pTilebuf += pstRect->y*m_iTileLineWidth + pstRect->x;

	//get left neighbour info
	uint16_t *pTilebufLeft = m_pTileData;
	int iTileNeighbourLeft = i_pclMap->GetElement(x-1, y);
	if(iTileNeighbourLeft>=0) {
		S_Rect *pstRectLeft = &m_pstTileRects[iTileNeighbourLeft];
		pTilebufLeft += pstRectLeft->y*m_iTileLineWidth + pstRectLeft->x+iMax;
	}
	//get up neighbour info
	uint16_t *pTilebufUp = m_pTileData;
	int iTileNeighbourUp = i_pclMap->GetElement(x, y-1);
	if(iTileNeighbourUp>=0) {
		S_Rect *pstRectUp = &m_pstTileRects[iTileNeighbourUp];
		pTilebufUp += (pstRectUp->y+iMax)*m_iTileLineWidth + pstRectUp->x;
	}
	//get left up neighbour info
	uint16_t *pTilebufUpLeft = m_pTileData;
	int iTileNeighbourUpL = i_pclMap->GetElement(x-1, y-1);
	if(iTileNeighbourUpL>=0) {
		S_Rect *pstRectUpL = &m_pstTileRects[iTileNeighbourUpL];
		pTilebufUpLeft += (pstRectUpL->y+iMax)*m_iTileLineWidth + pstRectUpL->x;
	}
	//get right up neighbour info
	uint16_t *pTilebufUpRight = m_pTileData;
	int iTileNeighbourUpR = i_pclMap->GetElement(x+1, y-1);
	if(iTileNeighbourUpR>=0) {
		S_Rect *pstRectUpR = &m_pstTileRects[iTileNeighbourUpR];
		pTilebufUpRight += (pstRectUpR->y+iMax)*m_iTileLineWidth + pstRectUpR->x;
	}

	/*uint16_t *pTilebufRight, *pTilebufDown, *pTilebufDownLeft, *pTilebufDownRight;
	int iTileNeighbourRight, iTileNeighbourDown, iTileNeighbourDownL, iTileNeighbourDownR;
	if(i_iCheckNeighbours>4) {
		//get right neighbour info
		pTilebufRight = m_pTileData;
		iTileNeighbourRight = i_pclMap->GetElement(x+1, y);
		if(iTileNeighbourRight>=0) {
			S_Rect *pstRectRight = &m_pstTileRects[iTileNeighbourRight];
			pTilebufRight += pstRectRight->y*m_iTileLineWidth + pstRectRight->x+iMax;
		}
		//get down neighbour info
		pTilebufDown = m_pTileData;
		iTileNeighbourDown = i_pclMap->GetElement(x, y+1);
		if(iTileNeighbourDown>=0) {
			S_Rect *pstRectDown = &m_pstTileRects[iTileNeighbourDown];
			pTilebufDown += (pstRectDown->y+iMax)*m_iTileLineWidth + pstRectDown->x;
		}
		//get left down neighbour info
		pTilebufDownLeft = m_pTileData;
		iTileNeighbourDownL = i_pclMap->GetElement(x-1, y+1);
		if(iTileNeighbourDownL>=0) {
			S_Rect *pstRectDownL = &m_pstTileRects[iTileNeighbourDownL];
			pTilebufDownLeft += (pstRectDownL->y+iMax)*m_iTileLineWidth + pstRectDownL->x;
		}
		//get right down neighbour info
		pTilebufDownRight = m_pTileData;
		iTileNeighbourDownR = i_pclMap->GetElement(x+1, y+1);
		if(iTileNeighbourDownR>=0) {
			S_Rect *pstRectDownR = &m_pstTileRects[iTileNeighbourDownR];
			pTilebufDownRight += (pstRectDownR->y+iMax)*m_iTileLineWidth + pstRectDownR->x;
		}
	}*/

	int i, j;
	for(j=0; j<m_iTileSize; j++) {
		for(i=0; i<m_iTileSize; i++) {
			//handle every pixel in both mask and tile
			dValue = 0.35; //default point
			if(j==0 || i==0 || j==iMax || i==iMax)
				dValue = 10.0; //border is more points
			if((j==0 && i==0) || (j==0 && i==iMax) || (j==iMax && i==0) || (j==iMax && i==iMax))
				dValue = 60.0; //corners are most points

			dScore += (pMaskbuf[i]==0)==(pTilebuf[i]==0) ? dValue : 0; //every pixel that is the same

			if(i_iCheckNeighbours>-1) {
				//analyse left neighbour border
				if(i==0 && iTileNeighbourLeft>=0) { //left border, compare with leftneighbour
					dScore += (pTilebuf[i]==0)==(pTilebufLeft[i]==0) ? 20 : 0;
				}
			}
			if(i_iCheckNeighbours>0) {
				//analyse up neighbour border
				if(j==0 && iTileNeighbourUp>=0) { //top border, compare with upneighbour
					dScore += (pTilebuf[i]==0)==(pTilebufUp[i]==0) ? 20 : 0;
				}
			}
			if(i_iCheckNeighbours>1) {
				//analyse up left neighbour border
				if(j==0 && i==0 && iTileNeighbourUpL>=0) { //top left corner, compare with upleftneighbour (corner only)
					dScore += (pTilebuf[0]==0)==(pTilebufUpLeft[iMax]==0) ? 50 : 0;
				}
			}
			if(i_iCheckNeighbours>2) {
				//analyse up right neighbour border
				if(j==0 && i==iMax && iTileNeighbourUpR>=0) { //top right corner, compare with uprightneighbour (corner only)
					dScore += (pTilebuf[iMax]==0)==(pTilebufUpRight[0]==0) ? 50 : 0;
				}
			}

			/*if(i_iCheckNeighbours>4) {
				//analyse right neighbour border
				if(i==m_iTileSize-1 && iTileNeighbourRight>=0) { //right border, compare with rightneighbour
					dScore += (pTilebuf[i]==0)==(pTilebufRight[i]==0) ? 5 : 0;
				}
			}
			if(i_iCheckNeighbours>5) {
				//analyse bottom left neighbour border
				if(j==m_iTileSize-1 && i==0 && iTileNeighbourDownL>=0) { //bottom left corner, compare with downleftneighbour (corner only)
					dScore += (pTilebuf[0]==0)==(pTilebufDownLeft[iMax]==0) ? 10 : 0;
				}
			}
			if(i_iCheckNeighbours>6) {
				//analyse bottom neighbour border
				if(j==m_iTileSize-1 && iTileNeighbourDown>=0) { //bottom border, compare with downneighbour
					dScore += (pTilebuf[i]==0)==(pTilebufDown[i]==0) ? 5 : 0;
				}
			}
			if(i_iCheckNeighbours>7) {
				//analyse bottom right neighbour border
				if(j==m_iTileSize-1 && i==iMax && iTileNeighbourDownR>=0) { //bottom right corner, compare with downrightneighbour (corner only)
					dScore += (pTilebuf[iMax]==0)==(pTilebufDownRight[0]==0) ? 10 : 0;
				}
			}*/
		}
		pMaskbuf     += m_iMaskLineWidth;
		pTilebuf     += m_iTileLineWidth;
		pTilebufLeft += m_iTileLineWidth;
	}

	return dScore;
}

C_Map *C_MapBuilder::MakeMapFromMask(char *i_szMaskFile)
{
	int i, j, k, iSize, iTileNum, iBestTile;
	double dBestScore, dScore;
	bool bResult;

	//load tileset
	if(!LoadTileset()) return NULL;

	//load mask
	m_pclMaskImg = new C_Image(i_szMaskFile, NULL, &bResult, _RGB555); //load as 16 bits
	if(!bResult) {
		delete m_pclMaskImg; m_pclMaskImg = NULL;
		return NULL;
	}
	int iWidth, iHeight; //size in tiles, not pixels
	m_pclMaskImg->GetInfo(&iWidth, &iHeight, NULL);
	m_pclMaskImg->GetLineSizeInfo(NULL, &m_iMaskLineWidth, NULL);
	m_pclMaskImg->GetBufferMemory((uint8_t**)&m_pMaskData);
	m_iMaskLineWidth /= 2;
	iWidth  /= m_iTileSize;
	iHeight /= m_iTileSize;

	//init output map
	m_stHeader.iWidth    = iWidth;
	m_stHeader.iHeight   = iHeight;
	m_stHeader.iTileSize = m_iTileSize;
	strcpy(m_stHeader.szTileset, "ts_alien.tga");
	char *szFileName = strrchr(i_szMaskFile, '\\');
	if(!szFileName) szFileName = strrchr(i_szMaskFile, '/');
	strcpy(m_stHeader.szGridfile, szFileName+1); //get the file name without path
	strcpy(strrchr(m_stHeader.szGridfile, '.'), ".map"); //replace extension
	C_Map *pclMap = new C_Map(m_szSavePath, &m_stHeader, NULL, true);

	//build map
	for(j=0; j<iHeight; j++) {
		for(i=0; i<iWidth; i++) {
			dBestScore=-1;
			//calc score for all possible tiles, and save the best tile
			iSize = (int)m_clPossibleList.size();
			/*int iTilePos = -1;
			for(int l=-1; l<=1; l++) {
				for(int m=-1; m<=1; m++) {
					iTilePos++;
					if(iTilePos<4) continue;
					if(i+m>0 && j+l>0 && i+m<iWidth && j+l<iHeight) {
						for(k=0; k<iSize; k++) {
							iTileNum = m_clPossibleList[k];
							dScore = GetTileMatchScore(pclMap, i+m, j+l, iTileNum, 4);
							if(dScore > dBestScore) {
								dBestScore = dScore;
								iBestTile = iTileNum;
							}
						}
						pclMap->SetElement(i+m, j+l, iBestTile);
					}
				}
			}*/

			dBestScore=-1;
			for(k=0; k<iSize; k++) {
				iTileNum = m_clPossibleList[k];
				dScore = GetTileMatchScore(pclMap, i, j, iTileNum, 4/*8*/);
				if(dScore > dBestScore) {
					dBestScore = dScore;
					iBestTile = iTileNum;
				}
			}

			if(dBestScore!=-1) {
				pclMap->SetElement(i, j, iBestTile);
			} else {
				//could not find any suitable tile for this position!
				pclMap->SetElement(i, j, 13); //set red (invalid) tile
			}
		}
	}
	m_bHeaderValid = true;
	return pclMap;
}

bool C_MapBuilder::GetHeader(S_MapHeader *o_stHeader)
{
	if(!m_bHeaderValid) return false;
	memcpy(o_stHeader, &m_stHeader, sizeof(m_stHeader));
	return true;
}
