
//main startup function and
//class to tie everything in the editor together.
// - init
// - input loop
// - deinit

#include <stdio.h>
#include <stdlib.h>
#include <vector>

#include "common.h"
#include "glgraph.h"
#include "image.h"
#include "font.h"
#include "e_tilemap.h"
#include "e_object.h"
#include "mapbuilder.h"
#include "global.h"
#include "input.h"

#include "resource_edt.h"

#define DEFAULT_WINDOW_W 1024
#define DEFAULT_WINDOW_H 768

#define PROGRAM_TITLE ((char*)"Galaxy MapEditor v3.02")

////////////////////////////////////////////////////////////////////////////////

//checks if a position is within a rectangle (_x1, _y1 noninclusive)
#define CHKRECT(_x, _y, _x1, _y1, _valx, _valy) ((_valx >= _x) && (_valx < _x1) && (_valy >= _y) && (_valy < _y1))

#ifdef WIN32
	#define ERROR_MSG(szMsg) \
	MessageBox(NULL, szMsg, "Error", MB_OK | MB_SETFOREGROUND | MB_TOPMOST | MB_SERVICE_NOTIFICATION);
#else
	#define ERROR_MSG(szMsg) \
	printf("%s\n", szMsg);
#endif

struct S_ZoneInfo
{
	int x, y;
	int iZoneSize;
	bool bHomeBase;
	bool bAntenna, bHouse;
	bool bExtraLife;
	std::vector<int> *pclCargoList; //array of cargo weights
};

struct S_CheckPointInfo
{
	int iX1, iY1;
	int iX2, iY2;
};

#define MAX_WAYPOINTS 10
struct S_EnemyInfo
{
	int iEnemyType;
	int iAngle;
	int iFireInterval;
	int iSpeed;
	int iNumWayPoints;
	POINT astWayPoint[MAX_WAYPOINTS];
};

struct S_DoorInfo
{
	int iSize;
	bool bHorizontal;
	int iOpenedForTime;
	int iClosedForTime;
	int iOpeningSpeed; //pixels/sec
	int iClosingSpeed; //pixels/sec
	int iNumButtons; //0 -> interval opening
	POINT stButtonPos[3];
};

struct S_ObjectInfo {
	//info kept for drawing
	int iDirection;  //0=0, 1=90, 2=180, 3=270
	int iObjectType; //0=player, 1=landingzone, 2=checkpoint, 3=door, 4..11=enemy0..7, 12..13=z-object0..1
	C_Object *pclObject;

	//info kept for loading and saving (depends on iObjectType)
	union {
		int              iCPNum;
		S_ZoneInfo       stZone;
		S_EnemyInfo      stEnemy;
		S_DoorInfo       stDoor;
	};
};
typedef std::vector<S_ObjectInfo> C_ObjectInfoList;

#define NUMOBJECTTYPES 16

static S_Rect s_stZObjRect[4][2] = {
	{{0,0,32 * 4,12},{0,0,12,32 * 4}}, {{0,0,32 * 8,12},{0,0,12,32 * 8}},
	{{0,0,32 * 12,16},{0,0,16,32 * 12}}, {{0,0,32 * 16,12},{0,0,16,32 * 12}} };

static S_Rect s_stVBaseRect[] = {{0,0,52,36},{0,36,52,36}};
static S_Rect s_stVDoorRect[] = {{23,0,6,6},{23,6,6,24},{23,30,6,6}};
static S_Rect s_stHBaseRect[] = {{52,20,36,52},{88,20,36,52}};
static S_Rect s_stHDoorRect[] = {{52,43,6,6},{58,43,24,6},{82,43,6,6}};
static S_Rect s_stButtonRect[] = {{52,0,23,19},{75,0,23,19}};

S_Rect PLAYERRECT        = {0,0,48,48};
S_Rect ZONERECT          = {0,44,96,4};
S_Rect ZONEHOUSERECT     = {70,23,24,21};
S_Rect ZONEANTENNARECT   = {0,15,4,29};
S_Rect ZONECARGORECTS[]  = {{54,38,14,6},{54,32,14,6},{54,26,14,6},{54,20,14,6}};
S_Rect ZONEEXTRALIFERECT = {20,25,14,19};

S_Rect CPRECT = {0,0,20,16};
S_Rect ENEMYRECT[8][4] = { //8 enemies (rectangles are 4 directions (0, 90, 180, 270))
	{{0,16,16,12}, {12,0,12,16},{0,28,16,12}, {0,0,12,16}}, //enemy0
	{{14,14,14,24},{0,38,24,14},{0,14,14,24}, {0,0,24,14}}, //enemy1
	{{0,16,16,12}, {12,0,12,16},{0,28,16,12}, {0,0,12,16}}, //enemy2
	{{14,14,14,24},{0,38,24,14},{0,14,14,24}, {0,0,24,14}}, //enemy3
	{{0,0,64,32},  {0,0,64,32}, {0,0,64,32},  {0,0,64,32}}, //enemy4
	{{0,26,13,22}, {0,13,22,13},{13,26,13,22},{0,0,22,13}}, //enemy5
	{{0,14,14,14}, {14,0,14,14},{14,14,14,14},{0,0,14,14}}, //enemy6
	{{0,0,64,64},  {0,0,64,64}, {0,0,64,64},  {0,0,64,64}}  //enemy7
};

S_FColor g_stShipColors[8] = {
	{0xAD/255.0f,0x16/255.0f,0x16/255.0f,1},
	{0x46/255.0f,0xB3/255.0f,0x9E/255.0f,1},
	{0x16/255.0f,0xAE/255.0f,0x16/255.0f,1},
	{0x19/255.0f,0x4F/255.0f,0xBF/255.0f,1},
	{0xC4/255.0f,0xAF/255.0f,0x46/255.0f,1},
	{0x9E/255.0f,0x16/255.0f,0x98/255.0f,1},
	{0xCD/255.0f,0xCD/255.0f,0xCD/255.0f,1},
	{0x93/255.0f,0x9F/255.0f,0x16/255.0f,1},
};

#define STARTXY     16
#define TILESPACE   2
#define TILESETW    500
#define SPACING     16
#define BORDER_SIZE 6 //must be less than SPACING/2

#ifdef WIN32
LRESULT CALLBACK WindowProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam);
#endif

class C_Editor
{
public:
	C_Editor();
	~C_Editor();

	bool IsInited() {return m_bInited;};
	bool IsLoaded() {return m_pclMap!=NULL;};
	C_Map *GetMap() {return m_pclMap;};
#ifdef WIN32
	HWND GetWindow() {return m_bInited ? m_pclGraph->GetWindow() : NULL;};
#endif

	bool Open(char *i_szFilename);

	void SetSize(int i_iWidth, int i_iHeight);
	void RedrawScreen();

	void Load();
	void Save();

	void SetDrawingToolSize(int i_iSize);
	void SetTile(int x, int y);
	void MouseClick(int x, int y);
	void RemoveSelected();
	void IncDecSelectedType(int iWay);
	void IncSelectedObj(int i_iWay);
	void RotateSelectedObj();
	void IncDecSelectedPos(int iWayX, int iWayY);
	void CropMap(int iSide);
	void ExtendMap(int iSide);
	void NewObject();
	void NewWP();

private:
	bool InitInstance();
	bool LoadObjects();
	void FreeObjects();
	void FreeObjectList();
	void FreeAll();

	bool LoadDescFile();
	void SaveDescFile();

	void AddNewObject(int i_iX, int i_iY);
	void AddNewWP(int i_iX, int i_iY);
	bool SelectObject(int i_iX, int i_iY);

	void DrawObjects();
	void DrawDoor(C_Object *i_pclDoor, S_DoorInfo *i_pstInfo, bool i_bSelected);
	void DrawHelpText(int i_iX, int i_iY);
	void DrawBackgroundRect(int i_iX, int i_iY, int i_iW, int i_iH);
	void DrawCurTile(int i_iX, int i_iY);

	char             m_szSavePath[MAX_PATH];

	//graph data
	C_GraphWrapperGL *m_pclGraph;  //the only instance of graphwrapper
	C_Image          *m_pclBKImg;  //background image
	C_Font           *m_pclFont;   //the font
	bool             m_bInited;
	bool             m_bWasDrawn;

	//tilesetdata
	char             m_iCurTile;        //what tile user has selected
	int              m_iCurTileSquare;  //size of drawing tool

	//point
	int              m_iSetPoint;

	//mapdata
	C_Map            *m_pclMap;

	//descriptionfile info
	char             m_szDescFile[MAX_PATH];
	char             m_szDescText[32];
	S_MapHeader      m_stHeader;
	MapType          m_iLevelType;
	int              m_iRaceLaps;
	int              m_iMinPlayers;
	bool             m_bSaveGravity, m_bSaveResistance;
	double           m_dGravityX, m_dGravityY;
	double           m_dResistanceX, m_dResistanceY;

	//objects
	C_ObjectInfoList *m_pclObjectList;
	C_Image          *m_apclObjectImg[NUMOBJECTTYPES];
	int              m_iSelectedObject;
	int              m_iSelectedWaypoint;
};

C_Editor::C_Editor()
{
	char szPath[MAX_PATH];
	m_szSavePath[0] = 0;

	m_pclGraph = NULL;
	m_pclBKImg = NULL;
	m_pclFont = NULL;
	m_szDescFile[0] = 0;
	m_szDescText[0] = 0;
	m_pclObjectList = NULL;

	m_bWasDrawn = false;
	m_bInited = InitInstance();

	for(int i=0; i<NUMOBJECTTYPES; i++) m_apclObjectImg[i] = NULL;
	if(m_bInited) m_bInited = LoadObjects();

	if(m_bInited) {
		bool bResult;
		sprintf(szPath, "%s/editor_bk.jpg", C_Global::szDataPath);
		m_pclBKImg = new C_Image(szPath, NULL, &bResult);
		m_bInited = bResult;
	}
	m_pclFont = new C_Font(_RGBA8888);
	if(m_bInited) m_bInited = m_pclFont->Load((char*)"arial-12", C_Global::szResourceFile);

	m_pclMap = NULL;
	m_bSaveGravity = m_bSaveResistance = false;
}

C_Editor::~C_Editor()
{
	FreeAll();
}

bool C_Editor::InitInstance()
{
#ifdef WIN32
	m_pclGraph = new C_GraphWrapperGL(WindowProc, LoadIcon(GetModuleHandle(NULL), (LPCTSTR)IDI_ICON1));
#else
	m_pclGraph = new C_GraphWrapperGL();
#endif
	bool bTest = m_pclGraph->SetMode(DEFAULT_WINDOW_W, DEFAULT_WINDOW_H, false, true);
	if(!bTest) {
		ERROR_MSG("Error initializing OpenGL.");
		return false;
	}
	m_pclGraph->SetWindowTitle(PROGRAM_TITLE);

	return true;
}

bool C_Editor::LoadObjects()
{
	char *aszImages[] = {
		(char*)"ship_w.tga", (char*)"zone.tga",   (char*)"cpbk.tga",   (char*)"door.tga",
		(char*)"enemy0.tga", (char*)"enemy1.tga", (char*)"enemy2.tga", (char*)"enemy3.tga",
		(char*)"enemy4.tga", (char*)"enemy5.tga", (char*)"enemy6.tga", (char*)"enemy7.tga"};

	int i;
	bool bResult = false;
	//load all images but the z-objects
	for(i=0; i<NUMOBJECTTYPES-4; i++) {
		m_apclObjectImg[i]  = new C_Image(aszImages[i], C_Global::szResourceFile, &bResult);
		if(!bResult) break;
	}
	//the z-objects will just be white bars for now
	S_Color stWhite = S_COLORMAKE(255, 255, 255, 255);
	m_apclObjectImg[12] = new C_Image(256, 256, _DEFAULT); m_apclObjectImg[12]->FillWithRGBColor(NULL, &stWhite);
	m_apclObjectImg[13] = new C_Image(256, 256, _DEFAULT); m_apclObjectImg[13]->FillWithRGBColor(NULL, &stWhite);
	S_Color stRed1 = S_COLORMAKE(220, 40, 40, 255);
	S_Color stRed2 = S_COLORMAKE(200, 60, 100, 255);
	m_apclObjectImg[14] = new C_Image(512, 512, _DEFAULT); m_apclObjectImg[14]->FillWithRGBColor(NULL, &stRed1);
	m_apclObjectImg[15] = new C_Image(512, 512, _DEFAULT); m_apclObjectImg[15]->FillWithRGBColor(NULL, &stRed2);
	return bResult;
}

void C_Editor::FreeObjects()
{
	for(int i=0; i<NUMOBJECTTYPES; i++) {
		delete m_apclObjectImg[i];
		m_apclObjectImg[i] = NULL;
	}
}

void C_Editor::FreeObjectList()
{
	if(m_pclObjectList) {
		int i, iSize = (int)m_pclObjectList->size();
		for(i=0; i<iSize; i++) {
			S_ObjectInfo stObjInfo = (*m_pclObjectList)[i];
			delete stObjInfo.pclObject;
			if(stObjInfo.iObjectType==1) delete stObjInfo.stZone.pclCargoList;
		}
		delete m_pclObjectList; m_pclObjectList = NULL;
	}
}

void C_Editor::FreeAll()
{
	delete m_pclMap;   m_pclMap   = NULL;
	delete m_pclFont;  m_pclFont  = NULL;
	delete m_pclBKImg; m_pclBKImg = NULL;
	delete m_pclGraph; m_pclGraph = NULL;
	FreeObjectList();
	FreeObjects();
}

bool C_Editor::Open(char *i_szFilename)
{
	m_iSelectedObject   = -1;
	m_iSelectedWaypoint = -1;
	m_iSetPoint         = 0;
	m_iCurTile          = 0;
	m_iCurTileSquare    = 1;

	delete m_pclMap;
	m_pclMap = NULL;
	memset(&m_stHeader, 0, sizeof(m_stHeader));

	strcpy(m_szSavePath, i_szFilename);
	//locate the last "\" or "/" in the filename
	char* szLastSlash = strrchr(m_szSavePath, '\\');
	if(!szLastSlash) szLastSlash = strrchr(m_szSavePath, '/');
	if(szLastSlash) {
		*szLastSlash = 0; //cut on last "\" or "/"
	}

	char szPath[MAX_PATH];
	strcpy(m_szDescFile, i_szFilename);
	char *szExt = strrchr(m_szDescFile, '.');
	if(szExt && (_stricmp(szExt, ".des")==0 || _stricmp(szExt, ".txt")==0)) {
		//load .des
		if(!LoadDescFile()) {
			ERROR_MSG("Error loading .des file!");
			return false;
		}

		//new case, if the souce des file is a .txt file, also check/load/save the .map file as .txt
		// this is to support the new unity project
		if (_stricmp(szExt, ".txt") == 0) {
			char *szExt2 = strrchr(m_stHeader.szGridfile, '.');
			//replace .map with .txt
			if(szExt2) strcpy(szExt2, ".txt");
		}

		//make map from new or existing grid?
		sprintf(szPath, "%s/%s", m_szSavePath, m_stHeader.szGridfile);

		bool bNewFile = true;
		FILE *pFile = fopen(szPath, "rb");
		if(pFile) {
			fclose(pFile);
			bNewFile = false;
		}
		m_pclMap = new C_Map(m_szSavePath, &m_stHeader, NULL, bNewFile);
	} else if(szExt && !_stricmp(szExt, ".tga")) {
		//make map from mask if maskfile is specified
		C_MapBuilder *pclMapBuilder = new C_MapBuilder(m_szSavePath);
		m_pclMap = pclMapBuilder->MakeMapFromMask(m_szDescFile);
		pclMapBuilder->GetHeader(&m_stHeader);
		delete pclMapBuilder;
		if(!m_pclMap) {
			ERROR_MSG("Error building map from .tga!");
			return false;
		}

		strcpy(strrchr(m_szDescFile, '.'), ".des"); //replace extension

		FreeObjectList();
		m_pclObjectList = new C_ObjectInfoList();
	} else {
		ERROR_MSG("No file selected!");
		return false;
	}

	RedrawScreen();
	return true;
}

void C_Editor::SetSize(int i_iWidth, int i_iHeight)
{
	m_pclGraph->SetSize(i_iWidth, i_iHeight);
}

void C_Editor::Load()
{
	LoadDescFile();

	//new case, if the souce des file is a .txt file, also check/load/save the .map file as .txt
	// this is to support the new unity project
	char *szExt = strrchr(m_szDescFile, '.');
	if (_stricmp(szExt, ".txt") == 0) {
		char *szExt2 = strrchr(m_stHeader.szGridfile, '.');
		//replace .map with .txt
		if (szExt2) strcpy(szExt2, ".txt");
	}

	m_pclMap->Load(&m_stHeader);
}

void C_Editor::Save()
{
	if(!m_pclMap->Save()) {
		ERROR_MSG("Cannot open file for writing.");
	} else {
		SaveDescFile();
	}
}

//if the coordinate is within the editwindow, the map is updated
void C_Editor::SetTile(int x, int y)
{
	if(!m_bWasDrawn) return;
	int iTileSize = m_pclMap->GetTileSize();
	int iX, iY;
	m_pclMap->GetPos(&iX, &iY);
	S_Rect *pstRect = m_pclMap->GetScreenRect();

	if(CHKRECT(pstRect->x, pstRect->y, pstRect->x+pstRect->width, pstRect->y+pstRect->height, x, y)) {
		x = x-pstRect->x+iX;
		y = y-pstRect->y+iY;

		if(!SelectObject(x, y)) {
			if(m_iSetPoint != 0) {
				if(m_iSetPoint == -1) { //add new object
					AddNewObject(x, y);
				} else if(m_iSetPoint == -2) { //add new waypoint
					AddNewWP(x, y);
				}
				m_iSetPoint = 0;
			} else {
				int i, j;
				for(i=0; i<m_iCurTileSquare; i++) {
					for(j=0; j<m_iCurTileSquare; j++) {
						m_pclMap->SetElement(x/iTileSize+i, y/iTileSize+j, m_iCurTile);
					}
				}
			}
		}
	}
}

//run when the user clicks/holds down the mouse button
// this does the right action, depending on the position of the mouse
void C_Editor::MouseClick(int x, int y)
{
	if(!m_bWasDrawn) return;
	int rx, ry, rx1, ry1, j, i;
	char iTilenum = 0;
	S_Rect *pstRect = m_pclMap->GetScreenRect();
	int iPosX = STARTXY;
	int iPosY = STARTXY+pstRect->height+SPACING;

	int iTileSize = m_pclMap->GetTileSize();
	int iNumTilesX = TILESETW / (m_pclMap->GetTileSize()+TILESPACE);
	int iNumTilesY = ((m_pclMap->GetNumTiles()/iNumTilesX)+1);

	//change current tile
	for(j=0; j<iNumTilesY; j++) {
		for(i=0; i<iNumTilesX; i++) {
			rx = iPosX+i*(iTileSize+TILESPACE);
			ry = iPosY+j*(iTileSize+TILESPACE);
			rx1 = rx+iTileSize;
			ry1 = ry+iTileSize;
			if(CHKRECT(rx, ry, rx1, ry1, x, y)) {
				m_iCurTile = iTilenum;
				j = iNumTilesY;
				break;
			}
			iTilenum++;
			if(iTilenum >= m_pclMap->GetNumTiles()) {
				j = iNumTilesY;
				break;
			}
		}
	}

	SetTile(x, y);
}

void C_Editor::SetDrawingToolSize(int i_iSize)
{
	m_iCurTileSquare = i_iSize;
	if(m_iCurTileSquare<1) m_iCurTileSquare = 1;
	if(m_iCurTileSquare>5) m_iCurTileSquare = 5;
}

bool C_Editor::SelectObject(int i_iX, int i_iY) //in mapcoordinates
{
	bool bFound = false;
	int iMapX, iMapY;
	int iObjPosX, iObjPosY;
	S_Rect stRect;
	m_pclMap->GetPos(&iMapX, &iMapY);

	m_iSelectedWaypoint = -1;

	int j, i, iSize = (int)m_pclObjectList->size();
	for(i=0; i<iSize; i++) {
		S_ObjectInfo stObjInfo = (*m_pclObjectList)[i];
		stObjInfo.pclObject->GetPos(&iObjPosX, &iObjPosY);
		int iObjType = stObjInfo.iObjectType;
		switch(iObjType) {
			case 0: stRect = PLAYERRECT; break;
			case 1:
				stRect = ZONERECT;
				stRect.width = stObjInfo.stZone.iZoneSize*32;
				break;
			case 2: stRect = CPRECT; break;
			case 3:
				stRect.x=0; stRect.y=0;
				if(stObjInfo.stDoor.bHorizontal) {
					iObjPosX += 23;
					stRect.width=6; stRect.height=stObjInfo.stDoor.iSize+36*2;
				} else {
					iObjPosY += 23;
					stRect.width=stObjInfo.stDoor.iSize+36*2; stRect.height=6;
				}
				if(m_iSelectedObject==i) {
					//maybe select a "button"
					for(j=0; j<stObjInfo.stDoor.iNumButtons; j++) {
						if(CHKRECT(stObjInfo.stDoor.stButtonPos[j].x, stObjInfo.stDoor.stButtonPos[j].y,
							stObjInfo.stDoor.stButtonPos[j].x+23, stObjInfo.stDoor.stButtonPos[j].y+19, i_iX, i_iY))
						{
							m_iSelectedWaypoint = j;
							return true;
						}
					}
				}
				break;
			default: //4..11
				if(m_iSelectedObject==i) {
					//maybe select a waypoint
					for(j=1; j<stObjInfo.stEnemy.iNumWayPoints; j++) {
						if(abs(i_iX-stObjInfo.stEnemy.astWayPoint[j].x)<5
						 && abs(i_iY-stObjInfo.stEnemy.astWayPoint[j].y)<5) {

							m_iSelectedWaypoint = j;
							return true;
						}
					}
				}
				stRect = ENEMYRECT[iObjType-4][stObjInfo.iDirection];
				break;
			case 12:
			case 13:
			case 14:
			case 15:
				stRect = s_stZObjRect[iObjType-12][stObjInfo.iDirection%2];
				break;
		}

		stRect.x=iObjPosX; stRect.y=iObjPosY; //make object rect
		if(CHKRECT(stRect.x, stRect.y, stRect.x+stRect.width, stRect.y+stRect.height, i_iX, i_iY)) {
			//make object selected
			bFound = true;
			m_iSelectedObject = i;
			break;
		}
	}

	return bFound;
}

void C_Editor::NewObject()
{
	m_iSetPoint = -1;
}

void C_Editor::NewWP()
{
	if(m_iSelectedObject != -1) {
		S_ObjectInfo *pstObjInfo = &(*m_pclObjectList)[m_iSelectedObject];
		if(pstObjInfo->iObjectType==3 && pstObjInfo->stDoor.iNumButtons<3) {
			m_iSetPoint = -2;
		} else if(pstObjInfo->iObjectType>3 && pstObjInfo->stEnemy.iNumWayPoints<MAX_WAYPOINTS) {
			m_iSetPoint = -2;
		}
	}
}

void C_Editor::AddNewObject(int i_iX, int i_iY) //in mapcoordinates
{
	S_ObjectInfo stObjInfo;
	stObjInfo.iDirection = 0;  //always 0, rotation is changed later
	stObjInfo.iObjectType = 0; //always 0, type is changed later
	stObjInfo.pclObject = new C_Object();
	stObjInfo.pclObject->SetPos(i_iX, i_iY);

	m_iSelectedObject = (int)m_pclObjectList->size(); //this new object will be the selected one
	m_pclObjectList->push_back(stObjInfo);
}

void C_Editor::AddNewWP(int i_iX, int i_iY) //in mapcoordinates
{
	if(m_iSelectedObject != -1) {
		S_ObjectInfo *pstObjInfo = &(*m_pclObjectList)[m_iSelectedObject];
		if(pstObjInfo->iObjectType==3 && pstObjInfo->stDoor.iNumButtons<3) {
			m_iSelectedWaypoint = pstObjInfo->stDoor.iNumButtons;
			POINT stP = {i_iX, i_iY};
			pstObjInfo->stDoor.stButtonPos[m_iSelectedWaypoint] = stP;
			pstObjInfo->stDoor.iNumButtons++;
		} else if(pstObjInfo->iObjectType>3 && pstObjInfo->stEnemy.iNumWayPoints<MAX_WAYPOINTS) {
			m_iSelectedWaypoint = pstObjInfo->stEnemy.iNumWayPoints;
			POINT stP = {i_iX, i_iY};
			pstObjInfo->stEnemy.astWayPoint[m_iSelectedWaypoint] = stP;
			pstObjInfo->stEnemy.iNumWayPoints++;
		}
	}
}

void C_Editor::RemoveSelected()
{
	if(m_iSelectedObject != -1) {
		S_ObjectInfo *pstObjInfo = &(*m_pclObjectList)[m_iSelectedObject];
		if(m_iSelectedWaypoint!=-1 && pstObjInfo->iObjectType==3) {
			//delete door "button"
			for(int i=m_iSelectedWaypoint; i<pstObjInfo->stDoor.iNumButtons-1; i++) {
				pstObjInfo->stDoor.stButtonPos[i] = pstObjInfo->stDoor.stButtonPos[i+1];
			}
			pstObjInfo->stDoor.iNumButtons--;
			if(m_iSelectedWaypoint>=pstObjInfo->stDoor.iNumButtons) m_iSelectedWaypoint = pstObjInfo->stDoor.iNumButtons-1;
			if(m_iSelectedWaypoint<=0) m_iSelectedWaypoint = -1;
		} else if(m_iSelectedWaypoint!=-1 && pstObjInfo->iObjectType>3) {
			//delete waypoint
			for(int i=m_iSelectedWaypoint; i<pstObjInfo->stEnemy.iNumWayPoints-1; i++) {
				pstObjInfo->stEnemy.astWayPoint[i] = pstObjInfo->stEnemy.astWayPoint[i+1];
			}
			pstObjInfo->stEnemy.iNumWayPoints--;
			if(m_iSelectedWaypoint>=pstObjInfo->stEnemy.iNumWayPoints) m_iSelectedWaypoint = pstObjInfo->stEnemy.iNumWayPoints-1;
			if(m_iSelectedWaypoint<=0) m_iSelectedWaypoint = -1;
		} else {
			//delete object
			m_pclObjectList->erase(m_pclObjectList->begin()+m_iSelectedObject);
			m_iSelectedObject = -1;
		}
	}
}

void C_Editor::IncDecSelectedType(int iWay) //-1 or +1
{
	int x, y;

	if(m_iSelectedObject != -1) {
		S_ObjectInfo *pstObjInfo = &(*m_pclObjectList)[m_iSelectedObject];
		pstObjInfo->iObjectType+=iWay;

		if(pstObjInfo->iObjectType<0) pstObjInfo->iObjectType = 0;
		if(pstObjInfo->iObjectType>(NUMOBJECTTYPES-1)) pstObjInfo->iObjectType = NUMOBJECTTYPES-1;

		switch(pstObjInfo->iObjectType) {
			case 0:  //player
				break;
			case 1:  //landing zone
				pstObjInfo->stZone.iZoneSize  = 3;
				pstObjInfo->stZone.bAntenna   = false;
				pstObjInfo->stZone.bExtraLife = false;
				pstObjInfo->stZone.bHomeBase  = false;
				pstObjInfo->stZone.bHouse     = false;
				pstObjInfo->stZone.pclCargoList = NULL;
				pstObjInfo->pclObject->GetPos(&pstObjInfo->stZone.x, &pstObjInfo->stZone.y);
				break;
			case 2:  //checkpoint
				pstObjInfo->iCPNum = 0;
				break;
			case 3:  //door
				pstObjInfo->stDoor.bHorizontal = (pstObjInfo->iDirection%2)!=0;
				pstObjInfo->stDoor.iClosedForTime = 2000;
				pstObjInfo->stDoor.iOpenedForTime = 4000;
				pstObjInfo->stDoor.iClosingSpeed  = 100;
				pstObjInfo->stDoor.iOpeningSpeed  = 100;
				pstObjInfo->stDoor.iSize          = 64;
				pstObjInfo->stDoor.iNumButtons    = 0;
				break;
			default: //enemy (4..11)
				pstObjInfo->stEnemy.iEnemyType    = pstObjInfo->iObjectType-4;
				pstObjInfo->stEnemy.iAngle        = pstObjInfo->iDirection*90;
				pstObjInfo->stEnemy.iFireInterval = 2000;
				pstObjInfo->stEnemy.iSpeed        = 0;
				if(pstObjInfo->stEnemy.iEnemyType==4) pstObjInfo->stEnemy.iSpeed = 80;
				if(pstObjInfo->stEnemy.iEnemyType==5) pstObjInfo->stEnemy.iSpeed = 30;
				if(pstObjInfo->stEnemy.iEnemyType==7) pstObjInfo->stEnemy.iSpeed = 90;
				pstObjInfo->stEnemy.iNumWayPoints = 1;
				pstObjInfo->pclObject->GetPos(&x, &y);
				pstObjInfo->stEnemy.astWayPoint[0].x = x;
				pstObjInfo->stEnemy.astWayPoint[0].y = y;
				break;
			case 12:
			case 13:
			case 14:
			case 15:
				break;
		}
	}
}

void C_Editor::IncSelectedObj(int i_iWay)
{
	if(m_iSelectedObject != -1) {
		S_ObjectInfo *pstObjInfo = &(*m_pclObjectList)[m_iSelectedObject];
		if(pstObjInfo->iObjectType==1) { //LZ
			pstObjInfo->stZone.iZoneSize += i_iWay;
			if(pstObjInfo->stZone.iZoneSize<2) pstObjInfo->stZone.iZoneSize = 2;
			if(pstObjInfo->stZone.iZoneSize>12) pstObjInfo->stZone.iZoneSize = 12;
		} else if(pstObjInfo->iObjectType==2) { //CP
			pstObjInfo->iCPNum += i_iWay;
			if(pstObjInfo->iCPNum<0) pstObjInfo->iCPNum = 0;
			if(pstObjInfo->iCPNum>39) pstObjInfo->iCPNum = 39; //cap
		} else if(pstObjInfo->iObjectType==3) { //door
			pstObjInfo->stDoor.iSize += i_iWay*2;
			if(pstObjInfo->stDoor.iSize<20) pstObjInfo->stDoor.iSize = 20;
			if(pstObjInfo->stDoor.iSize>600) pstObjInfo->stDoor.iSize = 600;
		}
	}
}

void C_Editor::RotateSelectedObj()
{
	if(m_iSelectedObject != -1) {
		S_ObjectInfo *pstObjInfo = &(*m_pclObjectList)[m_iSelectedObject];
		pstObjInfo->iDirection++;

		if(pstObjInfo->iDirection>3) pstObjInfo->iDirection = 0;

		if(pstObjInfo->iObjectType==3) pstObjInfo->stDoor.bHorizontal = (pstObjInfo->iDirection%2)!=0;
	}
}

void C_Editor::IncDecSelectedPos(int iWayX, int iWayY) //(-1 or +1, 0) or (0, -1 or +1)
{
	int iX, iY;
	if(m_iSelectedObject != -1) {
		S_ObjectInfo *pstObjInfo = &(*m_pclObjectList)[m_iSelectedObject];

		if(m_iSelectedWaypoint!=-1 && pstObjInfo->iObjectType>3) {
			//move waypoint
			pstObjInfo->stEnemy.astWayPoint[m_iSelectedWaypoint].x += iWayX;
			pstObjInfo->stEnemy.astWayPoint[m_iSelectedWaypoint].y += iWayY;
		} else if(m_iSelectedWaypoint!=-1 && pstObjInfo->iObjectType==3) {
			//move door "button"
			pstObjInfo->stDoor.stButtonPos[m_iSelectedWaypoint].x += iWayX;
			pstObjInfo->stDoor.stButtonPos[m_iSelectedWaypoint].y += iWayY;
		} else {
			//move object
			pstObjInfo->pclObject->GetPos(&iX, &iY);
			iX += iWayX;
			iY += iWayY;
			pstObjInfo->pclObject->SetPos(iX, iY);
		}
	}
}

void C_Editor::CropMap(int iSide)
{
	int iSizeX, iSizeY;
	m_pclMap->GetMapSize(&iSizeX, &iSizeY);
	int iTileSize = m_pclMap->GetTileSize();
	if(iSizeX<=2 || iSizeY<=2) return;
	//if left column or top row is removed we need to adjust all objects
	int iWayX = 0, iWayY = 0;
	int iWayMaxX = 0, iWayMaxY = 0;
	switch(iSide) {
		case 0: iWayX=-iTileSize; break;
		case 2: iWayY=-iTileSize; break;
		case 1: iWayMaxX=iTileSize; break;
		case 3: iWayMaxY=iTileSize; break;
	}
	int iX, iY;
	int j, i, iSize = (int)m_pclObjectList->size();
	//pass 1: check if objects are outside the new map
	for(i=0; i<iSize; i++) {
		S_ObjectInfo *pstObjInfo = &(*m_pclObjectList)[i];
		//check object
		pstObjInfo->pclObject->GetPos(&iX, &iY);
		if(iX + iWayX <0) return;
		if(iY + iWayY <0) return;
		if(iX + iWayMaxX >iSizeX) return;
		if(iY + iWayMaxY >iSizeY) return;
		if(pstObjInfo->iObjectType==3) {
			//check door "button"
			for(j=0; j<pstObjInfo->stDoor.iNumButtons; j++) {
				iX = pstObjInfo->stDoor.stButtonPos[j].x;
				iY = pstObjInfo->stDoor.stButtonPos[j].y;
				if(iX + iWayX <0) return;
				if(iY + iWayY <0) return;
				if(iX + iWayMaxX >iSizeX) return;
				if(iY + iWayMaxY >iSizeY) return;
			}
		}
		if(pstObjInfo->iObjectType>3 && pstObjInfo->iObjectType<12) {
			//check waypoints
			for(j=0; j<pstObjInfo->stEnemy.iNumWayPoints; j++) {
				iX = pstObjInfo->stEnemy.astWayPoint[j].x;
				iY = pstObjInfo->stEnemy.astWayPoint[j].y;
				if(iX + iWayX <0) return;
				if(iY + iWayY <0) return;
				if(iX + iWayMaxX >iSizeX) return;
				if(iY + iWayMaxY >iSizeY) return;
			}
		}
	}
	//pass 2: change positions
	if(iWayX!=0 || iWayY!=0) {
		for(i=0; i<iSize; i++) {
			S_ObjectInfo *pstObjInfo = &(*m_pclObjectList)[i];
			//move object
			pstObjInfo->pclObject->GetPos(&iX, &iY);
			iX += iWayX;
			iY += iWayY;
			pstObjInfo->pclObject->SetPos(iX, iY);
			if(pstObjInfo->iObjectType==3) {
				//move door "button"
				for(j=0; j<pstObjInfo->stDoor.iNumButtons; j++) {
					pstObjInfo->stDoor.stButtonPos[j].x += iWayX;
					pstObjInfo->stDoor.stButtonPos[j].y += iWayY;
				}
			}
			if(pstObjInfo->iObjectType>3) {
				//move waypoints
				for(j=0; j<pstObjInfo->stEnemy.iNumWayPoints; j++) {
					pstObjInfo->stEnemy.astWayPoint[j].x += iWayX;
					pstObjInfo->stEnemy.astWayPoint[j].y += iWayY;
				}
			}
		}
	}

	m_pclMap->Crop(iSide);
	//need to reacquire map size since it have changed
	m_pclMap->GetMapSize(&m_stHeader.iWidth, &m_stHeader.iHeight);
	m_stHeader.iWidth /= m_pclMap->GetTileSize();
	m_stHeader.iHeight /= m_pclMap->GetTileSize();
}

void C_Editor::ExtendMap(int iSide)
{
	int iSizeX, iSizeY;
	m_pclMap->GetMapSize(&iSizeX, &iSizeY);
	int iTileSize = m_pclMap->GetTileSize();
	//if left column or top row is extended we need to adjust all objects
	int iWayX = 0, iWayY = 0;
	switch(iSide) {
		case 0: iWayX = iTileSize; break;
		case 2: iWayY = iTileSize; break;
	}
	int iX, iY;
	int j, i, iSize = (int)m_pclObjectList->size();

	//change positions
	if(iWayX != 0 || iWayY != 0) {
		for(i = 0; i<iSize; i++) {
			S_ObjectInfo *pstObjInfo = &(*m_pclObjectList)[i];
			//move object
			pstObjInfo->pclObject->GetPos(&iX, &iY);
			iX += iWayX;
			iY += iWayY;
			pstObjInfo->pclObject->SetPos(iX, iY);
			if(pstObjInfo->iObjectType == 3) {
				//move door "button"
				for(j = 0; j<pstObjInfo->stDoor.iNumButtons; j++) {
					pstObjInfo->stDoor.stButtonPos[j].x += iWayX;
					pstObjInfo->stDoor.stButtonPos[j].y += iWayY;
				}
			}
			if(pstObjInfo->iObjectType>3 && pstObjInfo->iObjectType<12) {
				//move waypoints
				for(j = 0; j<pstObjInfo->stEnemy.iNumWayPoints; j++) {
					pstObjInfo->stEnemy.astWayPoint[j].x += iWayX;
					pstObjInfo->stEnemy.astWayPoint[j].y += iWayY;
				}
			}
		}
	}

	m_pclMap->Extend(iSide);
	//need to reacquire map size since it have changed
	m_pclMap->GetMapSize(&m_stHeader.iWidth, &m_stHeader.iHeight);
	m_stHeader.iWidth /= m_pclMap->GetTileSize();
	m_stHeader.iHeight /= m_pclMap->GetTileSize();
}

////////////////////////////////////////////////////////////////////////////////
void C_Editor::DrawHelpText(int i_iX, int i_iY)
{
	char szTemp[50];

	if(m_iSetPoint == -1) sprintf(szTemp, "Click object pos");
	else if(m_iSetPoint == -2) sprintf(szTemp, "Click WP pos");

	m_pclFont->Print(i_iX, i_iY, szTemp);
}

void C_Editor::DrawCurTile(int i_iX, int i_iY)
{
	int i, j;
	int iTileSize = m_pclMap->GetTileSize();

	for(i=0; i<m_iCurTileSquare; i++) {
		for(j=0; j<m_iCurTileSquare; j++) {
			m_pclMap->DrawTile(i_iX+i*iTileSize, i_iY+j*iTileSize, m_iCurTile);
		}
	}
}

void C_Editor::DrawDoor(C_Object *i_pclDoor, S_DoorInfo *i_pstInfo, bool i_bSelected)
{
	int x, y, i;
	m_pclMap->GetPos(&x, &y);

	int iPosX, iPosY;
	i_pclDoor->GetPos(&iPosX, &iPosY);
	S_Rect stRect;
	if(i_pstInfo->bHorizontal) {
		i_pclDoor->ODraw(m_apclObjectImg[3], &s_stVBaseRect[0], x, y, NULL, i_bSelected);
		i_pclDoor->SetPos(iPosX, iPosY+i_pstInfo->iSize+36);
		i_pclDoor->ODraw(m_apclObjectImg[3], &s_stVBaseRect[1], x, y, NULL, i_bSelected);

		stRect = s_stVDoorRect[1];
		int iSize = (i_pstInfo->iSize-0)/2;
		int iPos = -6;
		while(iPos<iSize-6) {
			if(iPos+24>=iSize-6) stRect.height=(iSize-6)-iPos;
			i_pclDoor->SetPos(iPosX+23, iPosY+36+iPos);
			i_pclDoor->ODraw(m_apclObjectImg[3], &stRect, x, y, NULL, false);
			i_pclDoor->SetPos(iPosX+23, iPosY+i_pstInfo->iSize+36-iPos-stRect.height);
			i_pclDoor->ODraw(m_apclObjectImg[3], &stRect, x, y, NULL, false);
			iPos+=stRect.height;
		}

		i_pclDoor->SetPos(iPosX+23, iPosY+36+iPos);
		i_pclDoor->ODraw(m_apclObjectImg[3], &s_stVDoorRect[2], x, y, NULL, false);
		i_pclDoor->SetPos(iPosX+23, iPosY+i_pstInfo->iSize+36-iPos-6);
		i_pclDoor->ODraw(m_apclObjectImg[3], &s_stVDoorRect[0], x, y, NULL, false);
	} else {
		i_pclDoor->ODraw(m_apclObjectImg[3], &s_stHBaseRect[0], x, y, NULL, i_bSelected);
		i_pclDoor->SetPos(iPosX+i_pstInfo->iSize+36, iPosY);
		i_pclDoor->ODraw(m_apclObjectImg[3], &s_stHBaseRect[1], x, y, NULL, i_bSelected);

		stRect = s_stHDoorRect[1];
		int iSize = (i_pstInfo->iSize-0)/2;
		int iPos = -6;
		while(iPos<iSize-6) {
			if(iPos+24>=iSize-6) stRect.width=(iSize-6)-iPos;
			i_pclDoor->SetPos(iPosX+36+iPos, iPosY+23);
			i_pclDoor->ODraw(m_apclObjectImg[3], &stRect, x, y, NULL, false);
			i_pclDoor->SetPos(iPosX+i_pstInfo->iSize+36-iPos-stRect.width, iPosY+23);
			i_pclDoor->ODraw(m_apclObjectImg[3], &stRect, x, y, NULL, false);
			iPos+=stRect.width;
		}

		i_pclDoor->SetPos(iPosX+36+iPos, iPosY+23);
		i_pclDoor->ODraw(m_apclObjectImg[3], &s_stHDoorRect[2], x, y, NULL, false);
		i_pclDoor->SetPos(iPosX+i_pstInfo->iSize+36-iPos-6, iPosY+23);
		i_pclDoor->ODraw(m_apclObjectImg[3], &s_stHDoorRect[0], x, y, NULL, false);
	}
	for(i=0; i<i_pstInfo->iNumButtons; i++) {
		i_pclDoor->SetPos(i_pstInfo->stButtonPos[i].x, i_pstInfo->stButtonPos[i].y);
		i_pclDoor->ODraw(m_apclObjectImg[3], &s_stButtonRect[0], x, y, NULL, i_bSelected && i==m_iSelectedWaypoint);
	}

	i_pclDoor->SetPos(iPosX, iPosY);
}

void C_Editor::DrawObjects()
{
	char szNum[32];
	int iX, iY;
	C_Image *pclImg;
	S_Rect *pstRect;
	m_pclMap->GetPos(&iX, &iY);

	int iPlayerNumber = -1;
	int j, i, iSize = (int)m_pclObjectList->size();
	for(i=0; i<iSize; i++) {
		S_ObjectInfo stObjInfo = (*m_pclObjectList)[i];
		int iObjType = stObjInfo.iObjectType;
		pclImg = m_apclObjectImg[iObjType];
		switch(iObjType) {
			case 0:
				pstRect = &PLAYERRECT;
				if(iPlayerNumber<7) iPlayerNumber++;
				break;
			case 1: pstRect = &ZONERECT; break;
			case 2: pstRect = &CPRECT;   break;
			case 3:
				break;
			default: //4..11
				pstRect = &ENEMYRECT[iObjType-4][stObjInfo.iDirection];
				if(m_iSelectedObject==i && stObjInfo.stEnemy.iNumWayPoints>1) {
					int x, y;
					S_Rect *pstRect = m_pclMap->GetScreenRect();
					m_pclGraph->SetClippingRect(pstRect); //in opengl the text must be clipped

					for(j=1; j<stObjInfo.stEnemy.iNumWayPoints; j++) {
						POINT *pstP = &stObjInfo.stEnemy.astWayPoint[j];

						x = pstRect->x+pstP->x-iX;
						y = pstRect->y+pstP->y-iY;
						S_FColor stRed  = S_FCOLORMAKE(1.0f,0.0f,0.0f,0.88f);
						S_FColor stGrey = S_FCOLORMAKE(0.6f,0.6f,0.6f,0.88f);
						m_pclGraph->SetColor(m_iSelectedWaypoint==j ? &stRed : &stGrey);
						m_pclGraph->Circle(x, y, 6);

						sprintf(szNum, "%d", j);
						m_pclFont->Print(x-2,y+6, szNum);

						sprintf(szNum, "[%d, %d]", pstP->x, pstP->y);
						m_pclFont->Print(x-2,y+6+12, szNum);
					}
					stObjInfo.pclObject->GetPos(&x, &y);
					sprintf(szNum, "[%d, %d]", x, y);
					x = pstRect->x+x-iX;
					y = pstRect->y+y-iY;
					m_pclFont->Print(x-2,y-20, szNum);

					m_pclGraph->SetClippingRect(NULL); //no clipping
				}
				break;
			case 12:
			case 13:
			case 14:
			case 15:
				pstRect = &s_stZObjRect[iObjType - 12][stObjInfo.iDirection%2];
				break;
		}

		if(iObjType==0) { //special draw for player
			stObjInfo.pclObject->ODraw(pclImg, pstRect, iX, iY, &g_stShipColors[iPlayerNumber], m_iSelectedObject==i);
		} else if(iObjType==1) { //special draw for landing zone
			int j;
			int iZoneSize = stObjInfo.stZone.iZoneSize;
			S_Rect stRect = *pstRect;
			stRect.width = 32;
			stObjInfo.pclObject->ODraw(pclImg, &stRect, iX, iY, NULL, m_iSelectedObject==i); //left part
			stRect.x += 32;
			for(j=0; j<iZoneSize-2; j++) {
			stObjInfo.pclObject->ODraw(pclImg, &stRect, iX-32-j*32, iY, NULL, false); //centre parts
			}
			stRect.x += 32;
			stObjInfo.pclObject->ODraw(pclImg, &stRect, iX-(iZoneSize-1)*32, iY, NULL, m_iSelectedObject==i); //right part

			int iAdjustX = -(stObjInfo.stZone.iZoneSize-3)*32;
			if(stObjInfo.stZone.bHouse) stObjInfo.pclObject->ODraw(pclImg, &ZONEHOUSERECT, iX-70+iAdjustX, iY+21, NULL, m_iSelectedObject==i);
			if(stObjInfo.stZone.bAntenna) stObjInfo.pclObject->ODraw(pclImg, &ZONEANTENNARECT, iX, iY+29, NULL, m_iSelectedObject==i);
			if(stObjInfo.stZone.pclCargoList!=NULL) {
				int j, iSize2 = (int)stObjInfo.stZone.pclCargoList->size();
				int iBaseX = stObjInfo.stZone.bHouse ? -50+iAdjustX : -78+iAdjustX;
				for(j=0; j<iSize2; j++) {
					int iCargoType = 3-((*stObjInfo.stZone.pclCargoList)[j]-1)/5; //1..5=small container, 6..10, 11..15, 16..20+ = large container
					if(iCargoType<0) iCargoType=0;
					//if(iCargoType>3) iCargoType=3;
					stObjInfo.pclObject->ODraw(pclImg, &ZONECARGORECTS[iCargoType], iX+iBaseX+((j/3)*14), iY+6+((j%3)*6), NULL, m_iSelectedObject==i);
				}
			}
			if(stObjInfo.stZone.bExtraLife) {
				int iBaseX = stObjInfo.stZone.bAntenna ? -8 : -4;
				stObjInfo.pclObject->ODraw(pclImg, &ZONEEXTRALIFERECT, iX+iBaseX, iY+ZONEEXTRALIFERECT.height, NULL, m_iSelectedObject==i);
			}

		} else if(iObjType==2) { //special draw for checkpoint
			stObjInfo.pclObject->ODraw(pclImg, pstRect, iX, iY, NULL, m_iSelectedObject==i);

			S_Rect *pstRect = m_pclMap->GetScreenRect();
			m_pclGraph->SetClippingRect(pstRect); //in opengl the text must be clipped

			char szString[8];
			sprintf(szString, "%d", stObjInfo.iCPNum);

			int iW;
			pclImg->GetInfo(&iW, NULL, NULL);

			int iPosX2, iPosY2;
			stObjInfo.pclObject->GetPos(&iPosX2, &iPosY2);

			int iWidth;
			m_pclFont->GetTextSize(szString, &iWidth);
			int iPosX = pstRect->x+ (iPosX2 - iX);
			int iPosY = pstRect->y+ (iPosY2 - iY);
			m_pclFont->Print(iPosX+((iW-iWidth)/2), iPosY, szString);

			m_pclGraph->SetClippingRect(NULL); //no clipping
		} else if(iObjType==3) { //special draw door
			DrawDoor(stObjInfo.pclObject, &stObjInfo.stDoor, m_iSelectedObject==i);
		} else { //all other objects
			stObjInfo.pclObject->ODraw(pclImg, pstRect, iX, iY, NULL, m_iSelectedObject==i);
		}
	}
}

void C_Editor::DrawBackgroundRect(int i_iX, int i_iY, int i_iW, int i_iH)
{
	int iW, iH;
	m_pclBKImg->GetInfo(&iW, &iH, NULL);

	S_Rect stBR[9];
	stBR[0] = S_RECTMAKE(0,              0,              BORDER_SIZE, BORDER_SIZE); //u left corner
	stBR[1] = S_RECTMAKE(iW-BORDER_SIZE, 0,              BORDER_SIZE, BORDER_SIZE); //u right corner
	stBR[2] = S_RECTMAKE(0,              iH-BORDER_SIZE, BORDER_SIZE, BORDER_SIZE); //l left corner
	stBR[3] = S_RECTMAKE(iW-BORDER_SIZE, iH-BORDER_SIZE, BORDER_SIZE, BORDER_SIZE); //l right corner
	stBR[4] = S_RECTMAKE(BORDER_SIZE,    0,              2,           BORDER_SIZE); //upper
	stBR[5] = S_RECTMAKE(BORDER_SIZE,    iH-BORDER_SIZE, 2,           BORDER_SIZE); //lower
	stBR[6] = S_RECTMAKE(0,              BORDER_SIZE,    BORDER_SIZE, 2);           //left
	stBR[7] = S_RECTMAKE(iW-BORDER_SIZE, BORDER_SIZE,    BORDER_SIZE, 2);           //right
	stBR[8] = S_RECTMAKE(BORDER_SIZE,    BORDER_SIZE,    4,           4);           //middle

	//dst rects
	int w2 = i_iW-BORDER_SIZE;
	int h2 = i_iH-BORDER_SIZE;

	S_Rect dstRectW1 = S_RECTMAKE(i_iX+BORDER_SIZE,      i_iY,                  i_iW-2*BORDER_SIZE, BORDER_SIZE);
	S_Rect dstRectW2 = S_RECTMAKE(i_iX+BORDER_SIZE,      i_iY+i_iH-BORDER_SIZE, i_iW-2*BORDER_SIZE, BORDER_SIZE);
	S_Rect dstRectH1 = S_RECTMAKE(i_iX,                  i_iY+BORDER_SIZE,      BORDER_SIZE, i_iH-2*BORDER_SIZE);
	S_Rect dstRectH2 = S_RECTMAKE(i_iX+i_iW-BORDER_SIZE, i_iY+BORDER_SIZE,      BORDER_SIZE, i_iH-2*BORDER_SIZE);
	S_Rect dstRectM  = S_RECTMAKE(i_iX+BORDER_SIZE,      i_iY+BORDER_SIZE,      i_iW-2*BORDER_SIZE, i_iH-2*BORDER_SIZE);

	//draw
	m_pclGraph->Blt(m_pclBKImg, &stBR[0], i_iX,    i_iY, true);
	m_pclGraph->Blt(m_pclBKImg, &stBR[1], i_iX+w2, i_iY, true);
	m_pclGraph->Blt(m_pclBKImg, &stBR[2], i_iX,    i_iY+h2, true);
	m_pclGraph->Blt(m_pclBKImg, &stBR[3], i_iX+w2, i_iY+h2, true);

	m_pclGraph->Blt(m_pclBKImg, &stBR[4], &dstRectW1, true);
	m_pclGraph->Blt(m_pclBKImg, &stBR[5], &dstRectW2, true);
	m_pclGraph->Blt(m_pclBKImg, &stBR[6], &dstRectH1, true);
	m_pclGraph->Blt(m_pclBKImg, &stBR[7], &dstRectH2, true);

	m_pclGraph->Blt(m_pclBKImg, &stBR[8], &dstRectM,  true);
}

void C_Editor::RedrawScreen()
{
	int w = m_pclGraph->GetModeWidth();
	int h = m_pclGraph->GetModeHeight();

	int iPosX, iPosY, iHelpW, iHelpH;
	m_pclBKImg->GetInfo(&iHelpW, &iHelpH, NULL);
	iPosX = w-(iHelpW+SPACING-BORDER_SIZE);
	iPosY = h-(iHelpH+SPACING-BORDER_SIZE);

	m_pclGraph->BeginScene(false);

	//background
	S_FColor stBlack = S_FCOLORMAKE(0,0,0,1);
	m_pclGraph->SetColor(&stBlack);
	m_pclGraph->Rect(NULL, false);
	//help
	if(iPosX>0 && iPosY>0) m_pclGraph->Blt(m_pclBKImg, NULL, iPosX, iPosY);

	m_bWasDrawn = false;
	if(m_pclMap) {
		int iTileSize = m_pclMap->GetTileSize();
		int iNumTilesX = TILESETW / (m_pclMap->GetTileSize()+TILESPACE);
		int iNumTilesY = m_pclMap->GetNumTiles()/iNumTilesX;
		if(m_pclMap->GetNumTiles()%iNumTilesX >0) iNumTilesY++;
		int iMapAreaW = w-((5*iTileSize)+2*SPACING+STARTXY);
		iHelpH -= 2*BORDER_SIZE;
		if(iHelpH<iNumTilesY*(TILESPACE+iTileSize)) iHelpH = iNumTilesY*(TILESPACE+iTileSize);
		int iMapAreaH = h-(STARTXY+2*SPACING+iHelpH);

		//draw?
		if(iMapAreaH>(5*iTileSize+16+SPACING) && w>(iNumTilesX*(iTileSize+TILESPACE)+SPACING+iHelpW+SPACING)) {
			//center map
			int w2, h2;
			S_Rect stRect = {STARTXY, STARTXY, iMapAreaW, iMapAreaH};
			m_pclMap->GetMapSize(&w2, &h2);
			if(iMapAreaW>w2) {
				stRect.x = STARTXY+(iMapAreaW-w2)/2;
				stRect.width = w2;
			}
			if(iMapAreaH>h2) {
				stRect.y = STARTXY+(iMapAreaH-h2)/2;
				stRect.height = h2;
			}
			m_pclMap->SetScreenRect(&stRect);
			m_pclMap->MoveRelative(0, 0);
			C_Object::Init(&stRect);

			//tileset
			iPosX = STARTXY;
			iPosY = STARTXY+iMapAreaH+SPACING;
			w = iNumTilesX*(iTileSize+TILESPACE);
			h = iNumTilesY*(iTileSize+TILESPACE);
			DrawBackgroundRect(iPosX-BORDER_SIZE, iPosY-BORDER_SIZE, w+2*BORDER_SIZE, h+2*BORDER_SIZE);
			int i, j;
			char iTilenum = 0;
			for(j=0; j<iNumTilesY; j++) {
				for(i=0; i<iNumTilesX; i++) {
					m_pclMap->DrawTile(iPosX+i*(iTileSize+TILESPACE), iPosY+j*(iTileSize+TILESPACE), iTilenum);
					iTilenum++;
					if(iTilenum >= m_pclMap->GetNumTiles()) {
						j = iNumTilesY;
						break;
					}
				}
			}

			//drawing tool
			iPosX = STARTXY+iMapAreaW+SPACING;
			iPosY = STARTXY;
			w = 5*iTileSize;
			h = 5*iTileSize;
			DrawBackgroundRect(iPosX-BORDER_SIZE, iPosY-BORDER_SIZE, w+2*BORDER_SIZE, h+2*BORDER_SIZE);
			DrawCurTile(iPosX, iPosY);

			//text
			if(m_iSetPoint!=0) {
				iPosX = STARTXY+iMapAreaW+SPACING;
				iPosY = STARTXY+5*iTileSize+SPACING;
				w = 5*iTileSize;
				h = 5*iTileSize;
				DrawBackgroundRect(iPosX-BORDER_SIZE, iPosY-BORDER_SIZE, w+2*BORDER_SIZE, 16+2*BORDER_SIZE);
				DrawHelpText(iPosX, iPosY);
			}

			//map
			DrawBackgroundRect(stRect.x-BORDER_SIZE, stRect.y-BORDER_SIZE, stRect.width+2*BORDER_SIZE, stRect.height+2*BORDER_SIZE);
			m_pclMap->Draw();
			DrawObjects();
			m_bWasDrawn = true;
		}
	}

	m_pclGraph->Flip();
}

////////////////////////////////////////////////////////////////////////////////

//loads a map header
bool C_Editor::LoadDescFile()
{
	int j;

	//free old map if exists
	m_iSelectedObject   = -1;
	m_iSelectedWaypoint = -1;
	m_iSetPoint         = 0;
	m_iRaceLaps = 3;
	m_iLevelType = MAP_MISSION;
	m_bSaveGravity = m_bSaveResistance = false;
	m_szDescText[0] = 0;

	FILE *pstFile = fopen(m_szDescFile, "rt");
	if(!pstFile) return false;

	FreeObjectList();
	m_pclObjectList = new C_ObjectInfoList();

	char szLine[512], szCommand[32];
	char szTemp[64];
	while(C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand)) {

		if(szCommand[0]!='*') continue;

		if(strcmp(szCommand, "*MAPTYPE") == 0) {
			sscanf(szLine, "%s %s %d", szCommand, szTemp, &m_iRaceLaps);
			if(strcmp(szTemp, "RACE") == 0) m_iLevelType = MAP_RACE;
			else if(strcmp(szTemp, "DOGFIGHT") == 0) m_iLevelType = MAP_DOGFIGHT;
			else if(strcmp(szTemp, "MISSION_COOP") == 0) m_iLevelType = MAP_MISSION_COOP;
			//else if(strcmp(szTemp, "MISSION") == 0) m_iLevelType = MAP_MISSION;
		} else if(strcmp(szCommand, "*MAPSIZE") == 0) {
			sscanf(szLine, "%s %d %d", szCommand, &m_stHeader.iWidth, &m_stHeader.iHeight);
		} else if(strcmp(szCommand, "*MAPFILE") == 0) {
			sscanf(szLine, "%s %s", szCommand, m_stHeader.szGridfile);
		} else if(strcmp(szCommand, "*MAPDESC") == 0) {
			char *pos = strrchr(szLine, '\"');
			if (pos) pos[0] = 0;
			pos = strchr(szLine, '\"');
			if (pos) strncpy(m_szDescText, pos + 1, sizeof(m_szDescText));
			m_szDescText[31] = 0;
		} else if(strcmp(szCommand, "*TILESIZE") == 0) {
			sscanf(szLine, "%s %d", szCommand, &m_stHeader.iTileSize);
		} else if(strcmp(szCommand, "*TILESET") == 0) {
			sscanf(szLine, "%s %s", szCommand, m_stHeader.szTileset);
		} else if(strcmp(szCommand, "*GRAVITY") == 0) {
			sscanf(szLine, "%s %lf %lf", szCommand, &m_dGravityX, &m_dGravityY);
			m_bSaveGravity = true;
		} else if(strcmp(szCommand, "*RESISTANCE") == 0) {
			sscanf(szLine, "%s %lf %lf", szCommand, &m_dResistanceX, &m_dResistanceY);
			m_bSaveResistance = true;
		} else if(strcmp(szCommand, "*LANDINGZONE") == 0) {
			int iNumCargo = 0;
			S_ZoneInfo stZone;
			stZone.pclCargoList = NULL;
			char szHomeBase[32], szExtra[32], szAntenna[32], szHouse[32];
			int iFields = sscanf(szLine, "%s %d %d %d %s %s %s %s %d", szCommand, &stZone.x, &stZone.y, &stZone.iZoneSize, szHomeBase, szExtra, szAntenna, szHouse, &iNumCargo);
			stZone.bHomeBase = (strcmp(szHomeBase, "HOMEBASE") == 0);
			stZone.bExtraLife = (strcmp(szExtra, "EXTRALIFE") == 0);
			stZone.bAntenna = (strcmp(szAntenna, "ANTENNA") == 0);
			stZone.bHouse = (strcmp(szHouse, "WAREHOUSE") == 0);

			if(iNumCargo>0) {
				int iCargos[12];
				iFields = sscanf(szLine, "%s %d %d %d %s %s %s %s %d %d %d %d %d %d %d %d %d %d %d %d %d",
					szCommand, &stZone.x, &stZone.y, &stZone.iZoneSize, szHomeBase, szExtra, szAntenna, szHouse, &iNumCargo, &iCargos[0],
					&iCargos[1], &iCargos[2], &iCargos[3], &iCargos[4], &iCargos[5], &iCargos[6],
					&iCargos[7], &iCargos[8], &iCargos[9], &iCargos[10], &iCargos[11]);
				//^make smarter (with loop)
				stZone.pclCargoList = new std::vector<int>;
				for(j=0; j<iNumCargo; j++) {
					stZone.pclCargoList->push_back(iCargos[j]);
				}
			}

			//add object
			S_ObjectInfo stObjInfo;
			stObjInfo.iDirection = 0;
			stObjInfo.iObjectType = 1;
			stObjInfo.pclObject = new C_Object();
			stObjInfo.pclObject->SetPos(stZone.x, stZone.y);
			stObjInfo.stZone = stZone;
			m_pclObjectList->push_back(stObjInfo);
		} else if(strcmp(szCommand, "*CHECKPOINTS") == 0) {
			int j, iNumCP = 0;
			sscanf(szLine, "%s %d", szCommand, &iNumCP);

			//read checkpoints
			S_CheckPointInfo stCPInfo;
			for(j=0; j<iNumCP; j++) {
				C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand);
				sscanf(szLine, "%s %d %d %d %d", szCommand, &stCPInfo.iX1, &stCPInfo.iY1, &stCPInfo.iX2, &stCPInfo.iY2);

				//add 2 objects per checkpoint
				S_ObjectInfo stObjInfo;
				stObjInfo.iCPNum = j;
				stObjInfo.iDirection = 0;
				stObjInfo.iObjectType = 2;
				stObjInfo.pclObject = new C_Object();
				stObjInfo.pclObject->SetPos(stCPInfo.iX1, stCPInfo.iY1);
				m_pclObjectList->push_back(stObjInfo);
				stObjInfo.pclObject = new C_Object();
				stObjInfo.pclObject->SetPos(stCPInfo.iX2, stCPInfo.iY2);
				m_pclObjectList->push_back(stObjInfo);
			}
		} else if(strcmp(szCommand, "*ENEMY") == 0) {
			S_EnemyInfo stParams;
			sscanf(szLine, "%s %d", szCommand, &stParams.iEnemyType);

			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*ANGLE"
			sscanf(szLine, "%s %d", szCommand, &stParams.iAngle);
			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*FIREINTERVAL"
			sscanf(szLine, "%s %d", szCommand, &stParams.iFireInterval);
			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*SPEED"
			sscanf(szLine, "%s %d", szCommand, &stParams.iSpeed);

			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*NUMWAYPOINTS"
			sscanf(szLine, "%s %d", szCommand, &stParams.iNumWayPoints);

			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*WAYPOINTSX"
			sscanf(szLine, "%s %d %d %d %d %d %d %d %d %d %d", szCommand, &stParams.astWayPoint[0].x, &stParams.astWayPoint[1].x, &stParams.astWayPoint[2].x, &stParams.astWayPoint[3].x, &stParams.astWayPoint[4].x, &stParams.astWayPoint[5].x, &stParams.astWayPoint[6].x, &stParams.astWayPoint[7].x, &stParams.astWayPoint[8].x, &stParams.astWayPoint[9].x);
			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*WAYPOINTSY"
			sscanf(szLine, "%s %d %d %d %d %d %d %d %d %d %d", szCommand, &stParams.astWayPoint[0].y, &stParams.astWayPoint[1].y, &stParams.astWayPoint[2].y, &stParams.astWayPoint[3].y, &stParams.astWayPoint[4].y, &stParams.astWayPoint[5].y, &stParams.astWayPoint[6].y, &stParams.astWayPoint[7].y, &stParams.astWayPoint[8].y, &stParams.astWayPoint[9].y);

			//add object
			S_ObjectInfo stObjInfo;
			stObjInfo.iDirection = stParams.iAngle/90;
			stObjInfo.iObjectType = 4+stParams.iEnemyType;
			stObjInfo.pclObject = new C_Object();
			stObjInfo.pclObject->SetPos(stParams.astWayPoint[0].x, stParams.astWayPoint[0].y);
			stObjInfo.stEnemy = stParams;
			m_pclObjectList->push_back(stObjInfo);
		} else if(strcmp(szCommand, "*DOOR") == 0) {
			S_DoorInfo stParams;
			int x, y, iAngle;
			sscanf(szLine, "%s %d %d %d", szCommand, &x, &y, &stParams.iSize);

			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*ANGLE"
			sscanf(szLine, "%s %d", szCommand, &iAngle);
			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*OPENTIME"
			sscanf(szLine, "%s %d", szCommand, &stParams.iOpenedForTime);
			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*CLOSETIME"
			sscanf(szLine, "%s %d", szCommand, &stParams.iClosedForTime);
			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*OPENSPEED"
			sscanf(szLine, "%s %d", szCommand, &stParams.iOpeningSpeed);
			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*CLOSESPEED"
			sscanf(szLine, "%s %d", szCommand, &stParams.iClosingSpeed);

			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*NUMBUTTONS"
			sscanf(szLine, "%s %d", szCommand, &stParams.iNumButtons);

			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*BUTTONSX"
			sscanf(szLine, "%s %d %d %d", szCommand, &stParams.stButtonPos[0].x, &stParams.stButtonPos[1].x, &stParams.stButtonPos[2].x);
			C_Global::GetNextLineAndCommand(pstFile, szLine, szCommand); //command "*BUTTONSY"
			sscanf(szLine, "%s %d %d %d", szCommand, &stParams.stButtonPos[0].y, &stParams.stButtonPos[1].y, &stParams.stButtonPos[2].y);

			//add object
			S_ObjectInfo stObjInfo;
			stObjInfo.iDirection = iAngle/90;
			stParams.bHorizontal = (stObjInfo.iDirection%2)!=0;
			stObjInfo.iObjectType = 3;
			stObjInfo.pclObject = new C_Object();
			stObjInfo.pclObject->SetPos(x, y);
			stObjInfo.stDoor = stParams;
			m_pclObjectList->push_back(stObjInfo);
		} else if(strcmp(szCommand, "*PLAYERSTARTPOS") == 0) {
			int x, y;
			sscanf(szLine, "%s %d %d", szCommand, &x, &y);

			S_ObjectInfo stObjInfo;
			stObjInfo.iDirection = 0;
			stObjInfo.iObjectType = 0;
			stObjInfo.pclObject = new C_Object();
			stObjInfo.pclObject->SetPos(x, y);
			m_pclObjectList->push_back(stObjInfo);
		} else if(strcmp(szCommand, "*MAXPLAYERS") == 0) {
			//sscanf(szLine, "%s %d", szCommand, &iMaxPlayers);
			//^not used, generated when saving
		} else if(strcmp(szCommand, "*MINPLAYERS") == 0) {
			sscanf(szLine, "%s %d", szCommand, &m_iMinPlayers);
		} else if (strcmp(szCommand, "*ZOBJECT") == 0) {
			int x, y, iType, iAngle;
			sscanf(szLine, "%s %d %d %d %d", szCommand, &iType, &x, &y, &iAngle);

			S_ObjectInfo stObjInfo;
			stObjInfo.iDirection = iAngle/90;
			stObjInfo.iObjectType = 12+iType;
			stObjInfo.pclObject = new C_Object();
			stObjInfo.pclObject->SetPos(x, y);
			m_pclObjectList->push_back(stObjInfo);
		}
	}
	fclose(pstFile);

	//header loaded
	return true;
}

void C_Editor::SaveDescFile()
{
	S_CheckPointInfo *pstCP = NULL;

	int iX, iY;
	int iObjTypeToWrite, i, j, iSize = (int)m_pclObjectList->size();
	int iMaxPlayers = 0;
	int iNumCP = 0;
	FILE *pstFile = NULL;

	//find if its a racemap and prepare/check checkpoint info
	for(i=0; i<iSize; i++) {
		S_ObjectInfo stObjInfo = (*m_pclObjectList)[i];
		if(stObjInfo.iObjectType==2) iNumCP++; //found a checkpoint
	}
	if(iNumCP%2!=0) {
		ERROR_MSG("Not an even number of checkpoints. Map not saved.");
		goto error;
	}
	iNumCP/=2;

	pstCP = new S_CheckPointInfo[iNumCP];
	memset(pstCP, 0xff, sizeof(S_CheckPointInfo)*iNumCP);
	for(i=0; i<iSize; i++) {
		S_ObjectInfo stObjInfo = (*m_pclObjectList)[i];
		if(stObjInfo.iObjectType==0) if(iMaxPlayers<8) iMaxPlayers++;
		if(stObjInfo.iObjectType==2) {
			int iNum = stObjInfo.iCPNum;
			if(iNum>=iNumCP) {
				ERROR_MSG("Checkpoints not numbered correctly. Map not saved.");
				goto error;
			}

			stObjInfo.pclObject->GetPos(&iX, &iY);
			if(pstCP[iNum].iX1==-1) {
				pstCP[iNum].iX1 = iX;
				pstCP[iNum].iY1 = iY;
			} else if(pstCP[iNum].iX2==-1) {
				pstCP[iNum].iX2 = iX;
				pstCP[iNum].iY2 = iY;
			} else {
				ERROR_MSG("More than two checkpoints have the same number. Map not saved.");
				goto error;
			}
		}
	}
	for(i=0; i<iNumCP; i++) {
		if(pstCP[i].iX1==-1 || pstCP[i].iX2==-1) {
			ERROR_MSG("Checkpoints not numbered correctly. Map not saved.");
			goto error;
		}
	}
	if(iMaxPlayers==0) {
		ERROR_MSG("No player starting pos. Map not saved.");
		goto error;
	}

	pstFile = fopen(m_szDescFile, "wt");
	if (!pstFile) {
		ERROR_MSG("Cannot open file for writing.");
		goto error;
	}

	//start writing file
	if(iNumCP!=0) fprintf(pstFile, "*MAPTYPE RACE %d\n", m_iRaceLaps);
	else if(m_iLevelType==MAP_DOGFIGHT) fprintf(pstFile, "*MAPTYPE DOGFIGHT\n");
	else if(m_iLevelType==MAP_MISSION_COOP) fprintf(pstFile, "*MAPTYPE MISSION_COOP\n");
	else /*if(m_iLevelType==MAP_MISSION)*/ fprintf(pstFile, "*MAPTYPE MISSION\n");

	if (m_szDescText[0] != 0) {
		fprintf(pstFile, "*MAPDESC  \"%s\"\n", m_szDescText);
	}

	//other required header info
	fprintf(pstFile, "*MAPSIZE  %d %d\n", m_stHeader.iWidth, m_stHeader.iHeight);
	fprintf(pstFile, "*MAPFILE  %s\n", m_stHeader.szGridfile);
	fprintf(pstFile, "*TILESIZE %d\n", m_stHeader.iTileSize);
	fprintf(pstFile, "*TILESET  %s\n\n", m_stHeader.szTileset);

	//write optional objects
	if(m_iMinPlayers>1) fprintf(pstFile, "*MINPLAYERS %d\n", m_iMinPlayers);
	if(m_bSaveGravity) {
		fprintf(pstFile, "*GRAVITY    %.2lf %.2lf\n", m_dGravityX, m_dGravityY);
	}
	if(m_bSaveResistance) {
		fprintf(pstFile, "*RESISTANCE %.2lf %.2lf\n", m_dResistanceX, m_dResistanceY);
	}

	//write the rest (mostly optional)
	fprintf(pstFile, "*MAXPLAYERS %d\n", iMaxPlayers);
	for(iObjTypeToWrite=0; iObjTypeToWrite<NUMOBJECTTYPES; iObjTypeToWrite++) {

		for(i=0; i<iSize; i++) {
			S_ObjectInfo stObjInfo = (*m_pclObjectList)[i];
			int iObjType = stObjInfo.iObjectType;
			if(iObjTypeToWrite!=iObjType) continue;

			stObjInfo.pclObject->GetPos(&iX, &iY);

			switch(iObjType) {
				case 0:
					fprintf(pstFile, "*PLAYERSTARTPOS %5d %5d\n", iX, iY);
					break;
				case 1:
					fprintf(pstFile, "*LANDINGZONE %5d %5d %2d %s %s %s %s", iX, iY,
						stObjInfo.stZone.iZoneSize,
						stObjInfo.stZone.bHomeBase  ? "HOMEBASE" : "CARGOBASE",
						stObjInfo.stZone.bExtraLife ? "EXTRALIFE": "NOEXTRALIFE",
						stObjInfo.stZone.bAntenna   ? "ANTENNA"  : "NOANTENNA",
						stObjInfo.stZone.bHouse     ? "WAREHOUSE": "NOWAREHOUSE");

					if(!stObjInfo.stZone.bHomeBase) {
						int iNumCargo = 0;
						if(stObjInfo.stZone.pclCargoList) {
							iNumCargo = (int)stObjInfo.stZone.pclCargoList->size();
						}
						fprintf(pstFile, " %d", iNumCargo);

						for(j=0; j<iNumCargo; j++) {
							int iCargo = (*stObjInfo.stZone.pclCargoList)[j];
							fprintf(pstFile, " %d", iCargo);
						}
					}
					fprintf(pstFile, "\n");
					break;
				case 2:
					//handled above and below
					break;
				case 3:
					fprintf(pstFile, "*DOOR %d %d %d\n", iX, iY, stObjInfo.stDoor.iSize);
					fprintf(pstFile, "  *ANGLE         %d\n", stObjInfo.stDoor.bHorizontal ? 270 : 0);
					fprintf(pstFile, "  *OPENTIME      %d\n", stObjInfo.stDoor.iOpenedForTime);
					fprintf(pstFile, "  *CLOSETIME     %d\n", stObjInfo.stDoor.iClosedForTime);
					fprintf(pstFile, "  *OPENSPEED     %d\n", stObjInfo.stDoor.iOpeningSpeed);
					fprintf(pstFile, "  *CLOSESPEED    %d\n", stObjInfo.stDoor.iClosingSpeed);
					fprintf(pstFile, "  *NUMBUTTONS    %d\n", stObjInfo.stDoor.iNumButtons);

					fprintf(pstFile, "  *BUTTONSX  ");
					for(j=0; j<stObjInfo.stDoor.iNumButtons; j++) {
						fprintf(pstFile, " %5d", stObjInfo.stDoor.stButtonPos[j].x);
					}
					fprintf(pstFile, "\n");
					fprintf(pstFile, "  *BUTTONSY  ");
					for(j=0; j<stObjInfo.stDoor.iNumButtons; j++) {
						fprintf(pstFile, " %5d", stObjInfo.stDoor.stButtonPos[j].y);
					}
					fprintf(pstFile, "\n");
					break;
				default: //4..11
					stObjInfo.stEnemy.astWayPoint[0].x = iX; //update waypoint 0 which is the object pos
					stObjInfo.stEnemy.astWayPoint[0].y = iY;

					fprintf(pstFile, "*ENEMY %d\n", iObjType-4);
					fprintf(pstFile, "  *ANGLE         %d\n", stObjInfo.iDirection*90);
					fprintf(pstFile, "  *FIREINTERVAL  %d //-1=random\n", stObjInfo.stEnemy.iFireInterval);
					fprintf(pstFile, "  *SPEED         %d\n", stObjInfo.stEnemy.iSpeed);
					fprintf(pstFile, "  *NUMWAYPOINTS  %d\n", stObjInfo.stEnemy.iNumWayPoints);

					fprintf(pstFile, "  *WAYPOINTSX  ");
					for(j=0; j<stObjInfo.stEnemy.iNumWayPoints; j++) {
						fprintf(pstFile, " %5d", stObjInfo.stEnemy.astWayPoint[j].x);
					}
					fprintf(pstFile, "\n");
					fprintf(pstFile, "  *WAYPOINTSY  ");
					for(j=0; j<stObjInfo.stEnemy.iNumWayPoints; j++) {
						fprintf(pstFile, " %5d", stObjInfo.stEnemy.astWayPoint[j].y);
					}
					fprintf(pstFile, "\n");
					break;
				case 12:
				case 13:
				case 14:
				case 15:
					fprintf(pstFile, "*ZOBJECT %d %5d %5d %d\n", iObjType - 12, iX, iY, stObjInfo.iDirection*90);
					break;
			}
		}
	}

	if(iNumCP!=0) {
		fprintf(pstFile, "*CHECKPOINTS %d\n", iNumCP);
		for(i=0; i<iNumCP; i++) {
			fprintf(pstFile, "  *CHECKPOINT %5d %5d %5d %5d\n", pstCP[i].iX1, pstCP[i].iY1, pstCP[i].iX2, pstCP[i].iY2);
		}
	}

error:
	delete pstCP;
	if(pstFile) fclose(pstFile);
}

////////////////////////////////////////////////////////////////////////////////
//os

C_Control *g_pclControl = NULL;
C_Editor *g_pclEditor = NULL;

bool g_bRedraw = false; //global redraw, set by message handler

int  g_iMouseX = 0, g_iMouseY = 0;
bool g_bLButton = false;

void Update()
{
	bool bRedraw = false;

	g_pclControl->Update();

	//get and handle input events
	int iX, iY;
	int iMEvent = g_pclControl->GetMouseEvent(&iX, &iY, NULL);
	bool bShiftPressed = g_pclControl->IsPressed(DIK_RSHIFT) || g_pclControl->IsPressed(DIK_LSHIFT);
	bool bControlPressed = g_pclControl->IsPressed(DIK_RCONTROL) || g_pclControl->IsPressed(DIK_LCONTROL);

	switch(iMEvent) {
		case EVT_MOUSE_MOVE:
			if(!g_pclEditor->IsLoaded()) break;
			g_iMouseX = iX;
			g_iMouseY = iY;
			if(g_bLButton) {
				g_pclEditor->SetTile(g_iMouseX, g_iMouseY);
				bRedraw = true;
			}
			break;
		case EVT_MOUSE_LEFT_DOWN:
			if(!g_pclEditor->IsLoaded()) break;
			g_pclEditor->MouseClick(iX, iY);
			g_bLButton = true;
			bRedraw = true;
			break;
		case EVT_MOUSE_LEFT_UP:
			g_bLButton = false;
			break;
	}

	if(!g_pclEditor->IsLoaded()) return;
	int iTileSize = g_pclEditor->GetMap()->GetTileSize();

	if(g_pclControl->IsPressedOnce(DIK_S)) {
		g_pclEditor->Save();
	}
	if(g_pclControl->IsPressedOnce(DIK_L)) {
		bRedraw = true;
		g_pclEditor->Load();
	}

	if(g_pclControl->IsPressedOnce(DIK_F1)) {
		bRedraw = true;
		g_pclEditor->SetDrawingToolSize(1);
	}
	if(g_pclControl->IsPressedOnce(DIK_F2)) {
		bRedraw = true;
		g_pclEditor->SetDrawingToolSize(2);
	}
	if(g_pclControl->IsPressedOnce(DIK_F3)) {
		bRedraw = true;
		g_pclEditor->SetDrawingToolSize(3);
	}
	if(g_pclControl->IsPressedOnce(DIK_F4)) {
		bRedraw = true;
		g_pclEditor->SetDrawingToolSize(4);
	}
	if(g_pclControl->IsPressedOnce(DIK_F5)) {
		bRedraw = true;
		g_pclEditor->SetDrawingToolSize(5);
	}

	if(g_pclControl->IsPressedRepeat(DIK_DOWN)) {
		bRedraw = true;
		g_pclEditor->GetMap()->MoveRelative(0, iTileSize/2);
		if(g_bLButton) g_pclEditor->SetTile(g_iMouseX, g_iMouseY);
	}
	if(g_pclControl->IsPressedRepeat(DIK_UP)) {
		bRedraw = true;
		g_pclEditor->GetMap()->MoveRelative(0, -iTileSize/2);
		if(g_bLButton) g_pclEditor->SetTile(g_iMouseX, g_iMouseY);
	}
	if(g_pclControl->IsPressedRepeat(DIK_RIGHT)) {
		bRedraw = true;
		g_pclEditor->GetMap()->MoveRelative(iTileSize/2, 0);
		if(g_bLButton) g_pclEditor->SetTile(g_iMouseX, g_iMouseY);
	}
	if(g_pclControl->IsPressedRepeat(DIK_LEFT)) {
		bRedraw = true;
		g_pclEditor->GetMap()->MoveRelative(-iTileSize/2, 0);
		if(g_bLButton) g_pclEditor->SetTile(g_iMouseX, g_iMouseY);
	}

	if(g_pclControl->IsPressedRepeat(DIK_HOME)) {
		bRedraw = true;
		if(bShiftPressed) g_pclEditor->CropMap(0);
		else if(bControlPressed) g_pclEditor->ExtendMap(0);
		else g_pclEditor->GetMap()->MoveRelative(-(g_pclEditor->GetMap()->GetScreenRect()->width/iTileSize)*iTileSize, 0);
	}
	if(g_pclControl->IsPressedRepeat(DIK_END)) {
		bRedraw = true;
		if(bShiftPressed) g_pclEditor->CropMap(1);
		else if(bControlPressed) g_pclEditor->ExtendMap(1);
		else g_pclEditor->GetMap()->MoveRelative((g_pclEditor->GetMap()->GetScreenRect()->width/iTileSize)*iTileSize, 0);
	}
	if(g_pclControl->IsPressedRepeat(DIK_PRIOR)) { //page up
		bRedraw = true;
		if(bShiftPressed) g_pclEditor->CropMap(2);
		else if(bControlPressed) g_pclEditor->ExtendMap(2);
		else g_pclEditor->GetMap()->MoveRelative(0, -(g_pclEditor->GetMap()->GetScreenRect()->height/iTileSize)*iTileSize);
	}
	if(g_pclControl->IsPressedRepeat(DIK_NEXT)) { //page down
		bRedraw = true;
		if(bShiftPressed) g_pclEditor->CropMap(3);
		else if(bControlPressed) g_pclEditor->ExtendMap(3);
		else g_pclEditor->GetMap()->MoveRelative(0, (g_pclEditor->GetMap()->GetScreenRect()->height/iTileSize)*iTileSize);
	}

	if(g_pclControl->IsPressedOnce(DIK_O)) { //add new object
		bRedraw = true;
		g_pclEditor->NewObject();
	}
	if(g_pclControl->IsPressedOnce(DIK_W)) { //add new waypoint to selected enemy
		bRedraw = true;
		g_pclEditor->NewWP();
	}

	if(g_pclControl->IsPressedOnce(DIK_DELETE)) { //delete selected object
		bRedraw = true;
		g_pclEditor->RemoveSelected();
	}
	if(g_pclControl->IsPressedOnce(DIK_ADD)) { //inc selected object type
		bRedraw = true;
		g_pclEditor->IncDecSelectedType(1);
	}
	if(g_pclControl->IsPressedOnce(DIK_SUBTRACT)) { //dec selected object type
		bRedraw = true;
		g_pclEditor->IncDecSelectedType(-1);
	}

	if(g_pclControl->IsPressedOnce(DIK_R)) { //inc selected object rotation with wrap around
		bRedraw = true;
		g_pclEditor->RotateSelectedObj();
	}

	if(g_pclControl->IsPressedRepeat(DIK_X)) { //inc/dec selected object x position
		bRedraw = true;
		if(!bShiftPressed) g_pclEditor->IncDecSelectedPos(1, 0);
		else g_pclEditor->IncDecSelectedPos(-1, 0); //shifted
	}
	if(g_pclControl->IsPressedRepeat(DIK_Y)) { //inc/dec selected object y position
		bRedraw = true;
		if(!bShiftPressed) g_pclEditor->IncDecSelectedPos(0, 1);
		else g_pclEditor->IncDecSelectedPos(0, -1); //shifted
	}

	if(g_pclControl->IsPressedRepeat(DIK_1)) { //dec selected check point number or landing zone width
		bRedraw = true;
		g_pclEditor->IncSelectedObj(-1);
	}
	if(g_pclControl->IsPressedRepeat(DIK_2)) { //inc selected check point number or landing zone width
		bRedraw = true;
		g_pclEditor->IncSelectedObj(1);
	}

	//draw if needed
	if(g_bRedraw || bRedraw) g_pclEditor->RedrawScreen();
	g_bRedraw = false;
}

#if defined WIN32

//message-handling loop
LRESULT CALLBACK WindowProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam)
{
	bool bRedraw = false;
	if(!g_pclEditor || !g_pclEditor->IsInited()) return DefWindowProc(hWnd, iMessage, wParam, lParam);

	switch(iMessage) {
		case WM_MOUSEMOVE:
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_MOUSEWHEEL:
			if(g_pclControl) g_pclControl->XLateMouseEvent(iMessage, wParam, lParam);
			break;

		case WM_SIZE:
			g_pclEditor->SetSize(LOWORD(lParam), HIWORD(lParam));
			g_bRedraw = true;
			break;
		case WM_ACTIVATEAPP:
			if(wParam != 0) g_bRedraw = true;
			break;
		case WM_PAINT:
			//g_bRedraw = true;
			g_pclEditor->RedrawScreen();
			break;
		case WM_DESTROY:
			PostQuitMessage(0);
			break;
		default:
			return DefWindowProc(hWnd, iMessage, wParam, lParam);
	}

	return 0;
}

bool HandleEvents() //returns true on exit
{
	MSG stMsg;
	if(PeekMessage(&stMsg, NULL, 0, 0, PM_NOREMOVE)) {
		if(!GetMessage(&stMsg, NULL, 0, 0)) return true;

		TranslateMessage(&stMsg);
		DispatchMessage(&stMsg);
	}
	return false;
}

#elif defined __APPLE__

bool HandleEvents() //returns true on exit
{
	static int iLastW = 0, iLastH = 0;
	int iW, iH;
	if(!FsPollDevice()) return true;
	FsGetWindowSize(iW, iH);
	if(iW!=iLastW || iH!=iLastH) {
		g_pclEditor->SetSize(iW, iH);
		iLastW = iW; iLastH = iH;
		g_bRedraw = true;
	}
	if(g_pclControl) g_pclControl->PollInput();

	return false;
}

void Redraw()
{
	int iW, iH;
	FsGetWindowSize(iW, iH);
	if(g_pclEditor) {
		g_pclEditor->SetSize(iW, iH);
		g_pclEditor->RedrawScreen();
	}
}
#else

bool HandleEvents() //returns true on exit
{
	if(!g_pclEditor || !g_pclEditor->IsInited()) return true;

	XEvent stEvent;
	XEvent *pstEvent = NULL;

	C_GraphWrapperGL *pclGraph = C_GraphWrapperGL::GetGraphWrapperObject();
	Display *pclDisplay = pclGraph->GetDisplay();

	while(XPending(pclDisplay)) {
		XNextEvent(pclDisplay, &stEvent); //blocks until an event is available
		switch(stEvent.type) {
			case KeyPress:
			case KeyRelease:
				//if(g_pclControl) g_pclControl->XLateKeyEvent(&stEvent);
				pstEvent = &stEvent;
				break;

			case MotionNotify:
			case ButtonPress:
			case ButtonRelease:
				//if(g_pclControl) g_pclControl->XLateMouseEvent(&stEvent);
				pstEvent = &stEvent;
				break;

			case Expose:
				g_bRedraw = true;
				break;

			case ConfigureNotify:
				g_pclEditor->SetSize(stEvent.xconfigure.width, stEvent.xconfigure.height);
				g_bRedraw = true;
				//g_pclEditor->RedrawScreen();
				break;
			case ClientMessage:
				//window manager messages
				if((stEvent.xclient.format == 32) && ((unsigned int)stEvent.xclient.data.l[0] == pclGraph->m_Atom_DeleteWindow)) {
					return true;
				}
				break;
			case DestroyNotify: //apparently does not happen (the above does the job)
				//window has been destroyed
				return true;
		}
	}
	if(g_pclControl) g_pclControl->PollInput(pstEvent);
	return false;
}
#endif

//main
#ifdef WIN32
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
#else
int main(int argc, char *argv[])
#endif
{
	bool bShowDialog = false;
	char szOpenName[1024];
	szOpenName[0] = 0;

	C_Global::Init((char*)"resource.dat", (char*)".GalaxyV2");
#if defined WIN32
	if(lpCmdLine[0]==0) {
		bShowDialog = true;
	} else {
		strcpy(szOpenName, lpCmdLine);
		if(szOpenName[0]=='"') {
			C_Global::GetWithin(szOpenName, '"'); //remove ""
		}
	}
#elif defined __APPLE__
	bShowDialog = true;
#else
	if(argc!=2) {
		ERROR_MSG("Usage: mapeditor [file.des]|[file.tga]");
		goto out;
	}
	strcpy(szOpenName, argv[1]);
#endif

	g_pclEditor = new C_Editor();
	if(!g_pclEditor->IsInited()) goto out;

	if(bShowDialog) {
		szOpenName[0] = 0;
		C_Global::OpenDialog(szOpenName, sizeof(szOpenName), C_Global::szDataPath,
			(char*)"des\0txt\0tga\0\0", (char*)"MapDescriptorFiles (.des)\0MapDescriptorFiles (.txt)\0MapApproximationMask (.tga)\0\0");
	}

	if(!g_pclEditor->Open(szOpenName)) goto out;

#ifdef WIN32
	g_pclControl = new C_Control(g_pclEditor->GetWindow());
#else
	g_pclControl = new C_Control();
#endif

#ifdef __APPLE__
	FsSetRedrawCallback(Redraw);
#endif

	//main loop
	while(1) {
		if(HandleEvents()) break;
		Update(); //updates input and does redraw on editor events
		mssleep(10);
	}
out:
	delete g_pclEditor;
	delete g_pclControl;
#ifdef WIN32
#ifdef _DEBUG
	_CrtDumpMemoryLeaks();
#endif
#endif

	return 0;
}
