// C_AISteering

//class to handle AI steering.
//!old AI

#include <deque> //stl

#include "AISteering.h"
#include "PathUtils.h"
#include "common.h"

int C_AISteering::s_iLevelsToGo = 5;
S_AIOutput C_AISteering::s_TheMoves[6] = {{0,0,0,0},{0,0,1,0},{1,0,0,0},{1,0,1,0},{0,1,0,0},{0,1,1,0}}; //the only 6 moves there is

void C_AISteering::GetOutput(S_ShipInfo *i_pstShipInfo, S_AIOutput *o_pstOutput)
{
	m_stShipDynamic = *i_pstShipInfo;
	GetShipOutput(o_pstOutput);
}

void C_AISteering::SetMap(C_Grid<char> *i_pclMap)
{
	delete m_pclPathFinder;

	//process the map to have a special "outline"
	//to prevent the ship from going too near a wall
	int iCount;
	char iTemp;
	int x, x2, iWidth = i_pclMap->iWidth;
	int y, y2, iHeight = i_pclMap->iHeight;
	delete m_pclMap;
	m_pclMap = new C_Grid<char>(i_pclMap);
	C_Grid<char> *pclTempMap = new C_Grid<char>(i_pclMap);

	//pass 1
	for(y=0; y<iHeight; y++) {
		for(x=0; x<iWidth; x++) {
			if(i_pclMap->aaGrid[y][x] != 0) { //source is a wall (!=0)
				pclTempMap->aaGrid[y][x] = -1;   //set a wall (==-1)
			} else {
				//count surrounding walls
				iCount = 0;
				for(y2=y-1; y2<=y+1; y2++) {
					for(x2=x-1; x2<=x+1; x2++) {
						if(i_pclMap->IsValid(x2, y2) && (i_pclMap->aaGrid[y2][x2] !=0)) iCount++;
					}
				}
				pclTempMap->aaGrid[y][x] = (char)(iCount*10); //set the cost
			}
      }
   }

	//pass 2
	for(y=0; y<iHeight; y++) {
		for(x=0; x<iWidth; x++) {
			iTemp = pclTempMap->aaGrid[y][x];
			if(iTemp != 0) { //source is a wall or next to a wall
				m_pclMap->aaGrid[y][x] = iTemp; //set cost same as source
			} else { 
				//count surrounding non 0
				iCount = 0;
				for(y2=y-1; y2<=y+1; y2++) {
					for(x2=x-1; x2<=x+1; x2++) {
						if(pclTempMap->IsValid(x2, y2) && (pclTempMap->aaGrid[y2][x2] !=0)) {
							iCount++;
							break; //no need to count to more than 1
						}
					}
					if(iCount > 0) break;
				}
				m_pclMap->aaGrid[y][x] = (char)(iCount+1); //set the cost (1 or 2)
			}
		}
	}

	delete pclTempMap;
	m_pclPathFinder = new C_AStar(m_pclMap, 0.92); //0<i_dQuant<1, bad,fast<->accurate,slow
}

//debug-functions
bool C_AISteering::GetDebug(S_DebugDraw *o_stDbgElement)
{
	return m_clDebug.Pop(o_stDbgElement);
}
////////////////////////////////////////////////////////////////////////////////

C_AISteering::C_AISteering(S_StaticShipInfo *i_pstSShipInfo)
{ 
	m_dPathTime = AI_TIMEPATH;
	m_dResultTime = AI_TIMESTEP;
	m_pclWPList = NULL;
	m_iNumWP = 0;
	m_iCurWP = 0;

	m_pclMap = NULL;
	m_pclPathFinder = NULL;

	m_stShipStatic = *i_pstSShipInfo;
	m_bLoop = false;

	m_dDebugVal = 0.0;
	memset(&m_stResult, 0, sizeof(m_stResult));
	memset(&m_stCurrent, 0, sizeof(m_stCurrent));

	m_bRunThread = false;

	m_bWPSet = false;
	m_pclInputWPList = NULL;
	m_bInputLoop = false;

	m_iDistLimit = AI_LIMIT;

	m_pclThread = NULL;
	m_pclMutex = new C_Mutex();
}

C_AISteering::~C_AISteering()
{ 
	m_bRunThread = false;
	m_pclThread->WaitForThreadExit();
	delete m_pclThread;

	delete m_pclWPList;

	delete m_pclMap;
	delete m_pclPathFinder;

	delete m_pclMutex;
}

void C_AISteering::SetWayPoints(C_PointList *i_pclWPList, bool i_bLoop)
{
	if(!m_bWPSet) {
		m_pclInputWPList = i_pclWPList;
		m_bInputLoop = i_bLoop;
		m_bWPSet = true; //indicator to thread
	} else {
		delete i_pclWPList;
	}
}

void C_AISteering::GetShipOutput(S_AIOutput *o_pstOutput)
{
	if(o_pstOutput==NULL) return;
	m_pclMutex->Enter();

	int i;
	bool bPosJump = (PathUtils::PointDistance(m_stCurrent.stPos.x, m_stCurrent.stPos.y, m_stShipDynamic.stPosition.x, m_stShipDynamic.stPosition.y) > 80.0);
	/**/if(bPosJump) m_iCurWP = 0;

	if(!m_pclWPList) { //hold still if no path to follow
		POINT p = {(int)m_stShipDynamic.stPosition.x, (int)m_stShipDynamic.stPosition.y};
		m_pclWPList = new C_PointList();
		m_pclWPList->push_back(p);
		m_pclWPList->push_back(p);
		m_iNumWP = (int)m_pclWPList->size();
		m_iCurWP = 0;
	}
	//////////////

	//debug output
	m_clDebug.Clear();
	int iSize = (int)m_pclWPList->size();
	for(i=0; i<iSize-1; i++) {
		m_clDebug.Push(DBG_LINE, 0x00ffffff, (*m_pclWPList)[i].x, (*m_pclWPList)[i].y, (*m_pclWPList)[(i+1)%m_iNumWP].x, (*m_pclWPList)[(i+1)%m_iNumWP].y, (char*)"");
	}
	char szText[16];
	sprintf(szText, "%.0lf", m_stShipDynamic.dPointing);
	m_clDebug.Push(DBG_TEXT, 0x0000ff00, (int)m_stShipDynamic.stPosition.x+32, (int)m_stShipDynamic.stPosition.y, 0, 0, szText);
	m_clDebug.Push(DBG_LINE, 0x000000ff, (int)m_stShipDynamic.stPosition.x,    (int)m_stShipDynamic.stPosition.y,
		(int)(m_stShipDynamic.stPosition.x+m_stShipDynamic.stVelocity.x),
		(int)(m_stShipDynamic.stPosition.y+m_stShipDynamic.stVelocity.y), (char*)"");
	sprintf(szText, "%.0lf", m_dDebugVal);
	m_clDebug.Push(DBG_TEXT, 0x0000ff00, (int)m_stShipDynamic.stPosition.x+64, (int)m_stShipDynamic.stPosition.y, 0, 0, szText);

	m_clDebug.Push(DBG_LINE, 0x000000ff, m_stCurrent.l0.x, m_stCurrent.l0.y, m_stCurrent.l1.x, m_stCurrent.l1.y, (char*)"");
	double dGoalAngle = PathUtils::CalcLineAngle(m_stCurrent.stPos.x, m_stCurrent.stPos.y, m_stCurrent.l1.x, m_stCurrent.l1.y);
	double dDiff = fabs(dGoalAngle - m_stCurrent.dPointing);
	if(dDiff > 180) dDiff = 360 - dDiff; //the difference of two angles can be max 180.
	sprintf(szText, "%.0lf", dDiff);
	m_clDebug.Push(DBG_TEXT, 0x0000ff00, (int)m_stShipDynamic.stPosition.x+32, (int)m_stShipDynamic.stPosition.y+16, 0, 0, szText);
	//////////////

	m_dResultTime += m_stShipDynamic.dTimeStep;
	//check if AI_TIMESTEP has passed (time for new descision)
	//so this is not done every frame
	//^now done in a thread to smooth cpu load
	//start thread first time
	if(!m_bRunThread) {
		m_bRunThread = true;
		m_pclThread = new C_Thread((THREADTYPE)WorkThread, (void*)this);
	}

	m_pclMutex->Leave();

	*o_pstOutput = m_stResult; //last result
}

bool C_AISteering::GetCurrentLineSegment(POINT *o_p1, POINT *o_p2)
{
	m_pclMutex->Enter();
	//if(m_pclWPList) {
	*o_p1 = (*m_pclWPList)[m_iCurWP];
	*o_p2 = (*m_pclWPList)[(m_iCurWP+1)%m_iNumWP];
	//}
	m_pclMutex->Leave();
	return true;
}

void C_AISteering::CheckCurrentLineSegment(int *io_iCurWP, POINTD *i_pstShipPos, POINT *o_p1, POINT *o_p2)
{
	//init waypoint facts
	int iCurWP = *io_iCurWP;
	*o_p1 = (*m_pclWPList)[iCurWP];
	*o_p2 = (*m_pclWPList)[(iCurWP+1)%m_iNumWP];
	//next waypoint?
	double dCurDist = PathUtils::PointDistance(i_pstShipPos->x, i_pstShipPos->y, o_p2->x, o_p2->y);
	if(dCurDist < m_iDistLimit) {
		iCurWP++; //switch linesegment
		iCurWP %= m_iNumWP;
		*o_p1 = (*m_pclWPList)[iCurWP];
		*o_p2 = (*m_pclWPList)[(iCurWP+1)%m_iNumWP];
	}
	*io_iCurWP = iCurWP;
}

void C_AISteering::GenerateMoves(S_CurValues *i_pstCurrent, double i_dTime, S_Move *o_stMoves)
{
	/*int i;
	for(i=0; i<6; i++) {
		S_CurValues *pstNext = &o_stMoves[i].stValue;
		*pstNext = *m_pstCurrent;

		pstNext->dPointing +=g_TheMoves[i].bRight ? m_stShipStatic.dSteerSpeed*i_dTime:0;
		pstNext->dPointing -=g_TheMoves[i].bLeft ? m_stShipStatic.dSteerSpeed*i_dTime:0;
		if(pstNext->dPointing <0) pstNext->dPointing +=360;
		if(pstNext->dPointing>360) pstNext->dPointing -=360;

		double dThrust = g_TheMoves[i].bThrottle ? m_stShipStatic.dThrust : 0;

		double ax = cos(pstNext->dPointing*(PI/180)) * dThrust;
		double ay = sin(pstNext->dPointing*(PI/180)) * dThrust;
		ax += m_stShipStatic.stGravity.x;
		ay += m_stShipStatic.stGravity.y;
		ax -= pstNext->stVel.x*m_stShipStatic.stResistance.x;
		ay -= pstNext->stVel.y*m_stShipStatic.stResistance.y;

		pstNext->stPos.x += (pstNext->stVel.x * i_dTime) + (0.5*ax*i_dTime*i_dTime); //??
		pstNext->stPos.y += (pstNext->stVel.y * i_dTime) + (0.5*ay*i_dTime*i_dTime); //??

		pstNext->stVel.x += ax * i_dTime;
		pstNext->stVel.y += ay * i_dTime;

		m_iCurWP = m_pstCurrent->iCurWP;
		CheckCurrentLineSegment(&pstNext->l0, &pstNext->l1);
		pstNext->iCurWP = m_iCurWP;

		o_stMoves[i].a = Strategy0(pstNext);
		o_stMoves[i].b = Strategy1(pstNext);
		o_stMoves[i].c = Strategy2(pstNext);
		o_stMoves[i].d = Strategy3(pstNext);
	}*/
	//optimizing of code above...
	//setup
	double dPointAhead = i_pstCurrent->dPointing;
	double dPointRight = i_pstCurrent->dPointing + m_stShipStatic.dSteerSpeed*i_dTime;
	double dPointLeft  = i_pstCurrent->dPointing - m_stShipStatic.dSteerSpeed*i_dTime;
	if(dPointRight <0) dPointRight +=360;
	if(dPointRight >360) dPointRight -=360;
	if(dPointLeft <0) dPointLeft +=360;
	if(dPointLeft >360) dPointLeft -=360;

	double dAccBaseX = m_stShipStatic.stGravity.x - (i_pstCurrent->stVel.x*m_stShipStatic.stResistance.x);
	double dAccBaseY = m_stShipStatic.stGravity.y - (i_pstCurrent->stVel.y*m_stShipStatic.stResistance.y);
	double dAccAheadX = dAccBaseX + cos(dPointAhead*(PI/180)) * m_stShipStatic.dThrust;
	double dAccAheadY = dAccBaseY + sin(dPointAhead*(PI/180)) * m_stShipStatic.dThrust;
	double dAccRightX = dAccBaseX + cos(dPointRight*(PI/180)) * m_stShipStatic.dThrust;
	double dAccRightY = dAccBaseY + sin(dPointRight*(PI/180)) * m_stShipStatic.dThrust;
	double dAccLeftX = dAccBaseX + cos(dPointLeft*(PI/180)) * m_stShipStatic.dThrust;
	double dAccLeftY = dAccBaseY + sin(dPointLeft*(PI/180)) * m_stShipStatic.dThrust;

	S_CurValues *pstNext;
	//move 0, no thust, no turn
	pstNext = &o_stMoves[0].stValue;
	*pstNext = *i_pstCurrent;
	pstNext->stPos.x += (pstNext->stVel.x * i_dTime) + (0.5*dAccBaseX*i_dTime*i_dTime); //??
	pstNext->stPos.y += (pstNext->stVel.y * i_dTime) + (0.5*dAccBaseY*i_dTime*i_dTime); //??
	pstNext->stVel.x += dAccBaseX * i_dTime;
	pstNext->stVel.y += dAccBaseY * i_dTime; //after pos really...
	CheckCurrentLineSegment(&pstNext->iCurWP, &pstNext->stPos, &pstNext->l0, &pstNext->l1);
	//move 2, no thust, left
	pstNext = &o_stMoves[2].stValue;
	*pstNext = o_stMoves[0].stValue;
	pstNext->dPointing = dPointLeft;
	//move 4, no thust, right
	pstNext = &o_stMoves[4].stValue;
	*pstNext = o_stMoves[0].stValue;
	pstNext->dPointing = dPointRight;

	//move 1, thust, no turn
	pstNext = &o_stMoves[1].stValue;
	*pstNext = *i_pstCurrent;
	pstNext->stPos.x += (pstNext->stVel.x * i_dTime) + (0.5*dAccAheadX*i_dTime*i_dTime); //??
	pstNext->stPos.y += (pstNext->stVel.y * i_dTime) + (0.5*dAccAheadY*i_dTime*i_dTime); //??
	pstNext->stVel.x += dAccAheadX * i_dTime;
	pstNext->stVel.y += dAccAheadY * i_dTime; //after pos really...
	CheckCurrentLineSegment(&pstNext->iCurWP, &pstNext->stPos, &pstNext->l0, &pstNext->l1);

	//move 3, thust, left
	pstNext = &o_stMoves[3].stValue;
	*pstNext = o_stMoves[0].stValue;
	pstNext->dPointing = dPointLeft;
	pstNext->stPos.x += (pstNext->stVel.x * i_dTime) + (0.5*dAccLeftX*i_dTime*i_dTime); //??
	pstNext->stPos.y += (pstNext->stVel.y * i_dTime) + (0.5*dAccLeftY*i_dTime*i_dTime); //??
	pstNext->stVel.x += dAccLeftX * i_dTime;
	pstNext->stVel.y += dAccLeftY * i_dTime; //after pos really...
	CheckCurrentLineSegment(&pstNext->iCurWP, &pstNext->stPos, &pstNext->l0, &pstNext->l1);

	//move 5, thust, right
	pstNext = &o_stMoves[5].stValue;
	*pstNext = o_stMoves[0].stValue;
	pstNext->dPointing = dPointRight;
	pstNext->stPos.x += (pstNext->stVel.x * i_dTime) + (0.5*dAccRightX*i_dTime*i_dTime); //??
	pstNext->stPos.y += (pstNext->stVel.y * i_dTime) + (0.5*dAccRightY*i_dTime*i_dTime); //??
	pstNext->stVel.x += dAccRightX * i_dTime;
	pstNext->stVel.y += dAccRightY * i_dTime; //after pos really...
	CheckCurrentLineSegment(&pstNext->iCurWP, &pstNext->stPos, &pstNext->l0, &pstNext->l1);

	//calc points
	int i;
	for(i=0; i<6; i++) {
		o_stMoves[i].a = Strategy0(&o_stMoves[i].stValue);
		o_stMoves[i].b = Strategy1(&o_stMoves[i].stValue);
		o_stMoves[i].c = Strategy2(&o_stMoves[i].stValue);
		o_stMoves[i].d = Strategy3(&o_stMoves[i].stValue);
	}
}

double TIME_FACTOR[10] = {1.0, 2.0, 2.0, 3.0, 3.0,  5.0, 5.0, 5.0, 5.0, 5.0};
int C_AISteering::EvaluateMoves(S_CurValues *i_pstCurrent, int i_iLevel, bool i_bShallow)
{
	int    i, iMove = 0;
	double dBest = 1000000000, dWorst = -1000000000;
	double dTest[6], a, b, c, d, e;
	S_Move stMoves[6];
	int iWorst;
	//double dShallowTest[6];

	S_CurValues stCurrent = *i_pstCurrent;
	/**/stCurrent.dTimestep = AI_TIMESTEP*TIME_FACTOR[i_iLevel];
	GenerateMoves(&stCurrent, AI_TIMESTEP*TIME_FACTOR[i_iLevel], &stMoves[0]);

	for(i=0; i<6; i++) {
		a = stMoves[i].a;
		b = stMoves[i].b;
		c = stMoves[i].c;
		d = stMoves[i].d;
		e = 0;
		if(a<10) a = 10; //limit the effect of a so that "stall" never happens
		if(i>=3) e = -1500; //prio for thrust
		//sometimes it can otherwise be worth more to stay than go towards the goal
		//dTest[i] = (a*a)/1.5 + b*2 + c*2.5; //orginal
		dTest[i] = (a*a)*1.6 + b*14.7 + c*3.5 + d*0.0; //maybe better
		//dTest[i] = (a*a)*1.6 + b*80 + c*3.5 + d*0.0 + e; //maybe better
		//dTest[i] = (a*a)*1.6 + (b*b)*((b<0)?-0.5:0.5) + c*3.5 + e;
		//^experiment with the weights...

		if(dTest[i] > dWorst) {
			dWorst = dTest[i];
			iWorst = i;
		}
	}

	//do shallow search
	/*if(i_iLevel>0 && i_iLevel<(m_iLevelsToGo-2) && i_bShallow) {
		for(i=0; i<6; i++) {
			stCurrent = stMoves[i].stValue;
			dShallowTest[i] = dTest[i] + EvaluateMoves(&stCurrent, i_iLevel+2, false);

			if(dShallowTest[i] < dBest) {
				dBest = dShallowTest[i];
				iMove = i;
			}
		}
		stCurrent = stMoves[iMove].stValue;
		dBest = dTest[iMove] + EvaluateMoves(&stCurrent, i_iLevel+1, true);
	} else*/ { //full search
		for(i=0; i<6; i++) {
			//check if the move shall recurse or if it is bad even now...
			//1. exclude right or left. both cannot be good.
			//if(i==2 && dTest[i]>dTest[4]) continue;
			//if(i==4 && dTest[i]>dTest[2]) continue;
			//if(i==3 && dTest[i]>dTest[5]) continue;
			//if(i==5 && dTest[i]>dTest[3]) continue;
			//2. exclude the worst
			//if(i==iWorst) continue;
			//^does not work well

			//recurse...
			stCurrent = stMoves[i].stValue;
			if(i_iLevel < s_iLevelsToGo) {
				dTest[i] += EvaluateMoves(&stCurrent, i_iLevel+1, true);
			}

			if(dTest[i] < dBest) {
				dBest = dTest[i];
				/**/m_dDebugVal = stMoves[i].b;
				iMove = i;
			}
		}
	}

	if(i_iLevel>0) return (int)dBest;
	else return iMove;
}

//score for how far from the line
double C_AISteering::Strategy0(S_CurValues *pstValues)
{
	VECTOR p = PathUtils::S_VECTORMAKE2((double)pstValues->stPos.x, (double)pstValues->stPos.y);
	VECTOR l0 = PathUtils::S_VECTORMAKE2((double)pstValues->l0.x,   (double)pstValues->l0.y);
	VECTOR l1 = PathUtils::S_VECTORMAKE2((double)pstValues->l1.x,   (double)pstValues->l1.y);

	double dScore = PathUtils::PointDistanceToLineSeg(p, l0, l1);
	return dScore;
}

//score for how much closer to goal
double C_AISteering::Strategy1(S_CurValues *pstValues)
{
	POINTD *p0 = &pstValues->stPos;
	POINT  *p1 = &pstValues->l1;
	VECTOR *vel = &pstValues->stVel;
	double dDistNow    = PathUtils::PointDistance(p0->x, p0->y, p1->x, p1->y);
	double dDistBefore = PathUtils::PointDistance(p0->x - vel->x*pstValues->dTimestep, p0->y - vel->y*pstValues->dTimestep, p1->x, p1->y);

	double dScore = (dDistNow - dDistBefore); //negative value is closer
	return dScore;
}

//score for pointing toward goal
double C_AISteering::Strategy2(S_CurValues *pstValues)
{
	double dGoalAngle = PathUtils::CalcLineAngle(pstValues->stPos.x, pstValues->stPos.y, pstValues->l1.x, pstValues->l1.y);
	double dDiff = dGoalAngle - pstValues->dPointing;
	if(dDiff < 0) dDiff += 360;
	if(dDiff > 360) dDiff -= 360;
	if(dDiff > 180) dDiff = 360 - dDiff; //the difference of two angles can be max 180.

	double dScore = dDiff;
	return dScore;
}

//score for speed
double C_AISteering::Strategy3(S_CurValues *pstValues)
{
	double dVel = PathUtils::PointDistance(pstValues->stVel.x, pstValues->stVel.y);

	double dScore = -dVel;
	return dScore;
}

THREAD_RET C_AISteering::WorkThread(void *i_pclAI)
{
	C_AISteering *pclAI = (C_AISteering *)i_pclAI;

	while(pclAI->m_bRunThread) {
		mssleep(5);
		if(!pclAI->m_bRunThread) break; //another chance

		//calc new path if set by user
		if(pclAI->m_bWPSet) {
			bool bDelete = true;
			int iNumWP = 0;

			C_PointList *pclWPListToCreate = NULL;
			if(pclAI->m_pclInputWPList) iNumWP = (int)pclAI->m_pclInputWPList->size();

			if(iNumWP>1) {
				pclAI->m_iCurWP = 0;
				pclAI->m_bLoop = pclAI->m_bInputLoop;

				int i, j;
				for(i=0; i<iNumWP-(pclAI->m_bLoop?0:1); i++) {
					POINT aP[2];
					aP[0] = (*pclAI->m_pclInputWPList)[i];
					aP[0].x /= 32; aP[0].y /= 32;
					aP[1] = (*pclAI->m_pclInputWPList)[(i+1)%iNumWP];
					aP[1].x /= 32; aP[1].y /= 32;
					C_PointList *pclWPList = NULL;
					pclWPList = pclAI->m_pclPathFinder->FindPath(aP[0], aP[1]);
					//^TODO: use ship enter direction as a "preferred direction" param, so it will cost more to turn back.
					if(pclWPList) {
						pclAI->m_pclPathFinder->SmothPath(pclWPList);

						int iSize = (int)pclWPList->size();
						for(j=0; j<iSize; j++) {
							if(j==iSize-1) {
								(*pclWPList)[j] = (*pclAI->m_pclInputWPList)[(i+1)%iNumWP]; //make endpoints the same as inputted
							} else {
								(*pclWPList)[j].x = (*pclWPList)[j].x*32 +16;
								(*pclWPList)[j].y = (*pclWPList)[j].y*32 +16;
							}
						}

						if(pclWPListToCreate==NULL) pclWPListToCreate = pclWPList; //the entire list the first time
						else {
							pclWPListToCreate->insert(pclWPListToCreate->end(), pclWPList->begin()+1, pclWPList->end());
							delete pclWPList;  //^begin+1 to avoid getting the same point twice (last endpoint is already there)
						}
					}
				}
				if(pclWPListToCreate) {
					//pclAI->m_pclPathFinder->SmothPath(pclWPListToCreate);
					//^if done here start and endpoints of path input segments might be optimized away
					/*pclAI->m_iNumWP = pclWPListToCreate->size();
					for(i=0; i<pclAI->m_iNumWP; i++) {
						(*pclWPListToCreate)[i].x = (*pclWPListToCreate)[i].x*32 +16;
						(*pclWPListToCreate)[i].y = (*pclWPListToCreate)[i].y*32 +16;
					}*/
				} else {
					pclWPListToCreate = pclAI->m_pclInputWPList;
					bDelete = false;
				}
				pclAI->m_iNumWP = (int)pclWPListToCreate->size();

				pclAI->m_pclMutex->Enter();
				delete pclAI->m_pclWPList;
				pclAI->m_pclWPList = pclWPListToCreate;
				pclAI->m_pclMutex->Leave();
			}
			if(bDelete) delete pclAI->m_pclInputWPList;
			pclAI->m_pclInputWPList = NULL;
			pclAI->m_bWPSet = false;
		}

		//calc steering
		if(pclAI->m_dResultTime >= AI_TIMESTEP) {
			pclAI->m_dResultTime -= AI_TIMESTEP;

			//init known facts
			pclAI->m_stCurrent.stPos = pclAI->m_stShipDynamic.stPosition;
			pclAI->m_stCurrent.stVel = pclAI->m_stShipDynamic.stVelocity;
			pclAI->m_stCurrent.dPointing = pclAI->m_stShipDynamic.dPointing;
			pclAI->m_stCurrent.iCurWP = pclAI->m_iCurWP;
			pclAI->CheckCurrentLineSegment(&pclAI->m_stCurrent.iCurWP, &pclAI->m_stCurrent.stPos, &pclAI->m_stCurrent.l0, &pclAI->m_stCurrent.l1);
			pclAI->m_iCurWP = pclAI->m_stCurrent.iCurWP;

			/**/pclAI->m_stCurrent.l2 = (*pclAI->m_pclWPList)[(pclAI->m_iCurWP+2)%pclAI->m_iNumWP];

			//get best move
			int iBestMove = pclAI->EvaluateMoves(&pclAI->m_stCurrent, 0, true);

			pclAI->m_stResult = s_TheMoves[iBestMove];
		}
	}
	return THREAD_RET_0;
}

//TODO: fix code and clean from temporaty code (/**/)
