// C_AISteering

#ifndef _AISTEERING_H_
#define _AISTEERING_H_

#include <deque>
#include "common.h"
#include "pathfind.h"

#include "DataStructsAI.h"

#include "DebugOutputProvider.h"

////////////////////////////////////////////////////////////////////////////////
#define AI_TIMEPATH 0.80 //sec
#define AI_TIMESTEP 0.10 //sec
#define AI_LIMIT 40     //max pixels from line

typedef struct
{
	POINTD stPos;
	VECTOR stVel;
	double dPointing;

	POINT  l0, l1, l2; //?
	int iCurWP;

	double dTimestep;
} S_CurValues;

typedef struct
{
	S_CurValues stValue;
	double a;
	double b;
	double c;
	double d;
} S_Move;


class C_AISteering
{
public:
	C_AISteering(S_StaticShipInfo *i_pstSShipInfo);
	~C_AISteering();

	void SetMap(C_Grid<char> *i_pclMap);
	void SetWayPoints(C_PointList *i_pclWPList, bool i_bLoop);
	void GetOutput(S_ShipInfo *i_pstShipInfo, S_AIOutput *o_pstOutput);

	//debugfunction that is called every timestep.
	bool GetDebug(S_DebugDraw *o_stDbgElement);
	//^returns true if the result is valid (shall be shown), false if not.
	//the function is called until false is recieved

	bool GetCurrentLineSegment(POINT *o_p1, POINT *o_p2);

private:
	void GetShipOutput(S_AIOutput *o_pstOutput);

	//advances current checkpoint
	void CheckCurrentLineSegment(int *io_iCurWP, POINTD *i_pstShipPos, POINT *o_p1, POINT *o_p2);

	//path following evaluation
	S_CurValues m_stCurrent;
	void GenerateMoves(S_CurValues *i_pstCurrent, double i_dTime, S_Move *o_stMoves);
	int EvaluateMoves(S_CurValues *i_pstCurrent, int i_iLevel, bool i_bShallow); //returns best move
	double Strategy0(S_CurValues *pstValues); //score for how far from the line
	double Strategy1(S_CurValues *pstValues); //score for how far from the next wp
	double Strategy2(S_CurValues *pstValues); //score for the ship pointing to the next wp
	double Strategy3(S_CurValues *pstValues); //score for speed
	double m_dDebugVal;

	double m_dPathTime;
	//result
	double m_dResultTime;
	S_AIOutput m_stResult;


	C_DebugOutputProvider m_clDebug;
	C_AStar *m_pclPathFinder;
	C_Grid<char> *m_pclMap;

	static S_AIOutput s_TheMoves[6];
	static int s_iLevelsToGo;

	S_StaticShipInfo m_stShipStatic;
	S_ShipInfo       m_stShipDynamic;

	C_PointList *m_pclWPList;
	int   m_iCurWP;
	int   m_iNumWP;
	bool  m_bLoop;

	static THREAD_RET WorkThread(void *i_pclAI);
	volatile bool m_bRunThread;

	volatile bool m_bWPSet;
	C_PointList *m_pclInputWPList;
	bool m_bInputLoop;

	/**/int m_iDistLimit;

	C_Mutex *m_pclMutex;
	C_Thread *m_pclThread;
};
////////////////////////////////////////////////////////////////////////////////

#endif
