#include "AIWorkerThread.h"


AIWorkerThread::AIWorkerThread()
	: _terminate(false), _thread(NULL)
{
	_thread = new C_Thread(WorkThread, this);
}

AIWorkerThread::~AIWorkerThread()
{
	if (_thread)
	{
		terminate();
		join();
		delete _thread;
		_thread = NULL;
	}
}

void AIWorkerThread::join()
{
	_thread->WaitForThreadExit();
}

THREAD_RET AIWorkerThread::WorkThread(void *pParam)
{
	if (pParam)
	{
		AIWorkerThread* threadWrapper = (AIWorkerThread*)pParam;

		threadWrapper->execute();
	}

	return THREAD_RET_0;
}

void AIWorkerThread::addTask(AIWorkerTask* task)
{
	bool needToSignal = false;

	_lock.Enter();

	if (_tasks.size() == 0)
	{
		needToSignal = true;
	}

	_tasks.push(task);

	_lock.Leave();

	if (needToSignal)
	{
		_newTaskEvt.Signal();
	}
}

void AIWorkerThread::execute()
{
	while (!_terminate)
	{
		AIWorkerTask* task = getNextTask();

		if (task)
		{
			task->execute();
			task->done();
		}
	}
}

AIWorkerTask* AIWorkerThread::getNextTask()
{
	AIWorkerTask* task = NULL;

	do 
	{
		_lock.Enter();
		if (_tasks.size() > 0)
		{
			task = _tasks.front();
			_tasks.pop();
		}
		_lock.Leave();

		if (task == NULL)
		{
			_newTaskEvt.Wait(100);
		}
	}
	while (!_terminate && task == NULL);

	return task;
}
