#pragma once

#include "common.h"

#include <queue>
#include <vector>

using std::queue;
using std::vector;

class AIWorkerTask
{
public:
	AIWorkerTask() : _finished(false) {}
	virtual ~AIWorkerTask() {}
	virtual void execute() = 0;
	virtual void stop() = 0;

	void wait(int timeoutMs = INFINITE) { _finishedEvt.Wait(timeoutMs); }

	bool isFinished() const { return _finished; }
	
protected:
	void done() { _finished = true; _finishedEvt.Signal(); }

	volatile bool _finished;
	C_Event _finishedEvt;

	friend class AIWorkerThread;
};

typedef queue<AIWorkerTask*> AITaskQueue;

class AIWorkerThread
{
public:
	AIWorkerThread();
	~AIWorkerThread();

	void addTask(AIWorkerTask* task);
	void terminate() { _terminate = true; }
	void join();
	int getPendingTaskCount() const { return (int)_tasks.size(); }

private:
	static THREAD_RET WorkThread(void *pParam);

	void execute();
	AIWorkerTask* getNextTask();

	volatile bool _terminate;

	C_Thread* _thread;
	C_Event _newTaskEvt;
	C_Mutex _lock;

	AITaskQueue _tasks;
};

typedef vector<AIWorkerThread*> AIWorkerThreadVector;
