#pragma once

#include "common.h"

////////////////////////////////////////////////////////////////////////////////

typedef struct
{
	int iID; //the id of this ship, to make it possible to identify bullets and other ships in S_EnemyInfo
	double dThrust;     //pixel/sec^2 (limited by resistance and enhanced by gravity)
	double dSteerSpeed; //degrees/sec

	VECTOR stGravity;   //(pixel/sec^2)
	VECTOR stResistance;//velocity dependent factor. stVelocity*stResistance (pixel/sec^2)
	//acceleration  a[xy] = acceleration_from_thrust[xy] + gravity[xy] - (velocity[xy]*resistancefactor[xy])
	//velocity      v[xy] += a[xy] * timestep;
	//position      p[xy] += v[xy] * timestep;
} S_StaticShipInfo;

typedef struct
{
	double dTimeStep;    //time since last frame (seconds)

	POINTD stPosition;   //current position (pixels)

	VECTOR stVelocity;   //current velocity (x and y components) (pixels/sec)
	double dVelocity;    //current combined velocity (from stVelocity) in the heading direction (pixels/sec)
	double dHeading;     //the direction the ship is heading (deg, -> = 0, clockwise). if dVelocity is 0 this is invalid

	double dPointing;    //the direction the nose of the ship is pointing (deg, -> = 0, clockwise)
} S_ShipInfo;

//the only output from the AI-system:
//how the ship is controlled. the buttons to press
typedef struct
{
	bool bLeft, bRight;
	bool bThrottle;
	bool bFire;
} S_AIOutput;

typedef struct
{
	POINTD stPos;
	VECTOR stVel;
	double dPointing;
} S_ShipState;

typedef struct
{
	S_AIOutput Command;
	S_ShipState State;
	double TimeStep;
} S_ExtendedAIOutput;
