#pragma once

#include "common.h"

//debug
typedef enum
{
	DBG_TEXT = 1,
	DBG_LINE = 2,
	DBG_CIRCLE = 3,
} DebugOutputShapeType;

typedef struct
{ //all coordinates are map coords (not screen coords)
	POINT p1; //used for DBG_TEXT, DBG_LINE, DBG_CIRCLE
	union {
		POINT p2;        //used for DBG_LINE
		int   iRadius;   //used for DBG_CIRCLE
		char  szText[32];//used for DBG_TEXT
	};
	unsigned int iColor; //used for DBG_TEXT, DBG_LINE, DBG_CIRCLE
	DebugOutputShapeType iType; //is either DBG_TEXT, DBG_LINE or DBG_CIRCLE
} S_DebugDraw;


class C_DebugOutputProvider
{
public:
	virtual ~C_DebugOutputProvider() {}

	void Clear()
	{
		m_stDebugList.clear();
	}

	void Push(DebugOutputShapeType iType, int iColor, int x1, int y1, int x2_radius, int y2, char *szText)
	{
		S_DebugDraw stTestDebug;
		stTestDebug.iType = iType;
		stTestDebug.iColor = iColor;
		stTestDebug.p1.x = x1;
		stTestDebug.p1.y = y1;
		switch(iType) {
			case DBG_TEXT:
				strcpy(stTestDebug.szText, szText);
				break;
			case DBG_LINE:
				stTestDebug.p2.x = x2_radius;
				stTestDebug.p2.y = y2;
				break;
			case DBG_CIRCLE:
				stTestDebug.iRadius = x2_radius;
				break;
		}
		m_stDebugList.push_back(stTestDebug);
	}

	bool Pop(S_DebugDraw *o_stDbgElement)
	{
		if(m_stDebugList.empty()) return false;
		*o_stDbgElement = m_stDebugList.front();
		m_stDebugList.pop_front();
		return true;
	}

protected:
	std::deque<S_DebugDraw> m_stDebugList;
};
