#include "GameAI.h"

const double C_GameAI::DEFAULT_POLL_INTERVAL = 0.005;

C_GameAI::C_GameAI(const S_StaticShipInfo& staticShipInfo) : _map(NULL), _staticShipInfo(staticShipInfo),
		_pollIntervalSum(0), _shipImg(NULL), _shipRectangles(NULL), _numActualDirections(0)
{
}

C_GameAI::~C_GameAI()
{
	for (AIWorkerThreadVector::iterator w = _workerThreads.begin(); w != _workerThreads.end(); ++w)
	{
		(*w)->terminate();
	}
	for (AIWorkerThreadVector::iterator w = _workerThreads.begin(); w != _workerThreads.end(); ++w)
	{
		(*w)->join();
		delete *w;
	}
	_workerThreads.clear();
}

void C_GameAI::setShipImage(const C_Image* shipImg, S_Rect shipRectangles[], int numActualDirections)
{
	_shipImg = shipImg;
	_shipRectangles = shipRectangles;
	_numActualDirections = numActualDirections;
}

void C_GameAI::commitPollInterval(double interval)
{
	if ((int)_pollIntervals.size() >= POLL_INTERVAL_QUEUE_LENGTH)
	{
		_pollIntervalSum -= _pollIntervals.front();
		_pollIntervals.pop();
	}
	_pollIntervals.push(interval);
	_pollIntervalSum += interval;
}

double C_GameAI::getAveragePollInterval() const
{
	return _pollIntervals.size() > 0 ? _pollIntervalSum / _pollIntervals.size() : DEFAULT_POLL_INTERVAL;
}

void C_GameAI::getOutput(const S_ShipInfo& shipInfo, S_AIOutput& output)
{
	commitPollInterval(shipInfo.dTimeStep);
	_lastKnownDynamicShipInfo = shipInfo;
}

int C_GameAI::getNumWorkerThreadsToSpawn()
{
	if (C_Thread::NumberOfCPUs() > 1)
	{
		return C_Thread::NumberOfCPUs() - 1;
	}

	return 1;
}

void C_GameAI::spawnWorkerThreads(int numThreads)
{
	for (int k = 0; k < numThreads; ++k)
	{
		_workerThreads.push_back(new AIWorkerThread());
	}
}

void C_GameAI::setUp()
{
	spawnWorkerThreads(getNumWorkerThreadsToSpawn());
}

void C_GameAI::enqueueTask(AIWorkerTask* task)
{
	AIWorkerThread* _leastEncumberedThread = NULL;

	for (AIWorkerThreadVector::iterator w = _workerThreads.begin(); w != _workerThreads.end(); ++w)
	{
		if (_leastEncumberedThread == NULL || _leastEncumberedThread->getPendingTaskCount() > (*w)->getPendingTaskCount())
		{
			_leastEncumberedThread = *w;
		}
	}

	if (_leastEncumberedThread == NULL)
	{
		std::runtime_error("Out of threads to perform task!");
	}

	_leastEncumberedThread->addTask(task);
}
