#pragma once

#include "tilemap.h"

#include "DataStructsAI.h"
#include "AIWorkerThread.h"

#include <stdexcept>
#include <queue>

using std::queue;

class C_CheckPoint;
typedef std::vector<C_CheckPoint*> C_CheckPointVector;

//define general contract for an AI
class C_GameAI
{
public:
	C_GameAI(const S_StaticShipInfo& staticShipInfo);
	virtual ~C_GameAI();

	virtual void setShipImage(const C_Image* shipImg, S_Rect shipRectangles[], int numActualDirections);
	virtual void setMap(C_Map* pMap) { _map = pMap; }
	virtual bool supportsMapType(MapType type) const { return false; }

	virtual void setUp();
	virtual void getOutput(const S_ShipInfo& shipInfo, S_AIOutput& output);

protected:
	void commitPollInterval(double interval);
	double getAveragePollInterval() const;
	virtual int getNumWorkerThreadsToSpawn();
	void spawnWorkerThreads(int numThreads);
	void enqueueTask(AIWorkerTask* task);

	C_Map* _map;
	S_ShipInfo _lastKnownDynamicShipInfo;
	S_StaticShipInfo _staticShipInfo;
	queue<double> _pollIntervals;
	double _pollIntervalSum;

	AIWorkerThreadVector _workerThreads;

	const C_Image* _shipImg;
	S_Rect* _shipRectangles;
	int _numActualDirections;

	static const int POLL_INTERVAL_QUEUE_LENGTH = 100;
	static const double DEFAULT_POLL_INTERVAL;
};

//AI specific to race maps
class C_RaceMapAI : public C_GameAI
{
public:
	C_RaceMapAI(const S_StaticShipInfo& staticShipInfo) : 
	  C_GameAI(staticShipInfo), _checkpoints(NULL) {}

	virtual bool supportsMapType(MapType type) const { return type == MAP_RACE; }
	virtual ~C_RaceMapAI() {}

	void setChecpoints(const C_CheckPointVector* pCheckptList) { _checkpoints = pCheckptList; }
	virtual void setCurrentCheckpoint(const C_CheckPoint* pCheckpt) = 0;

protected:
	const C_CheckPointVector* _checkpoints;
};

//AI specific to dogfight maps 
class C_DogfightMapAI : public C_GameAI
{
public:
	C_DogfightMapAI(const S_StaticShipInfo& staticShipInfo) : C_GameAI(staticShipInfo) {}
	virtual ~C_DogfightMapAI() {}

	virtual bool supportsMapType(MapType type) const { return type == MAP_DOGFIGHT; }

	//TODO add params
	virtual void setEnemies();

protected:
	//TODO add storage for enemy info
};
