#pragma once

#include "common.h"
#include "environment_defaults.h"

struct PathFinderSettings
{
	PathFinderSettings(int numSubtilesPerAxis, int timeStepMs, bool useHeuristicDistance, int bufferZoneWidth, int maxDistanceToTarget, int numDirections, bool timeFlowsForward) :
		NumSubtilesPerAxis(numSubtilesPerAxis), TimeStepMs(timeStepMs), UseHeuristicDistance(useHeuristicDistance), 
		BufferZoneWidth(bufferZoneWidth), MaxDistanceToTarget(maxDistanceToTarget), NumDirectionCategories(numDirections), TimeFlowsForward(timeFlowsForward) {}

	PathFinderSettings() : NumSubtilesPerAxis(DEFAULT_NUM_TILE_SUBPARTS), TimeStepMs(DEFAULT_STEP_INTERVAL_MS), UseHeuristicDistance(true),
		BufferZoneWidth(DEFAULT_BUFFER_ZONE_WIDTH), MaxDistanceToTarget(MAX_ANTICIPATED_DISTANCE), NumDirectionCategories(DEFAULT_NUM_DIRECTION_CATEGORIES), 
		TimeFlowsForward(DEFAULT_TIME_FLOW_FORWARD), UseSimplifiedMoveValidation(DEFAULT_USE_SIMPLIFIED_MOVE_VALIDATION) {}

	bool SettingsValid() const;

	int NumSubtilesPerAxis;
	int TimeStepMs;
	bool UseHeuristicDistance;
	int BufferZoneWidth;
	int MaxDistanceToTarget;
	int NumDirectionCategories;
	bool TimeFlowsForward;
	bool UseSimplifiedMoveValidation;

	static const int DEFAULT_NUM_TILE_SUBPARTS = 8;
	static const int DEFAULT_STEP_INTERVAL_MS = 90;
	static const int DEFAULT_BUFFER_ZONE_WIDTH = 5;
	static const int MAX_ANTICIPATED_DISTANCE = 200000;
	static const int DEFAULT_NUM_DIRECTION_CATEGORIES = 360 / OBJECT_HEADING_STEP;
	static const bool DEFAULT_TIME_FLOW_FORWARD = true;
	static const bool DEFAULT_USE_SIMPLIFIED_MOVE_VALIDATION = false;
};
