//PathUtils

//helper functions for the AI

#pragma once

#include "common.h"
#include <math.h>

#define Dot(v1, v2) ((v1).x*(v2).x + (v1).y*(v2).y)

namespace PathUtils
{
	double PointDistance(double i_iX1, double i_iY1, double i_iX2, double i_iY2);
	double PointDistance(double i_iX1, double i_iY1);

	double PointDistance2(double i_iX1, double i_iY1, double i_iX2, double i_iY2);
	double PointDistance2(double i_iX1, double i_iY1);

	double CalcLineAngle(double x1, double y1, double x2, double y2);

	void MidPoint(double i_iX1, double i_iY1, double i_iX2, double i_iY2, double *o_iX, double *o_iY);
	POINT MidPoint(double i_iX1, double i_iY1, double i_iX2, double i_iY2);

	int Round(double x);

	double EuclideanDistance(const POINT& p, const POINT& q);

	bool IsUnitDistance(const POINT& p, const POINT& q);

	double Signum(const double& x);

	double PointDistanceToLineSeg(POINTD& p, VECTOR& l0, VECTOR& l1, POINTD *o_pstClosest = NULL);
	bool PointWithinSegment(POINTD& p, VECTOR& l0, VECTOR& l1);

	void Normalize(VECTOR& v);

	VECTOR S_VECTORMAKE2(double x, double y);
	VECTOR operator + (const VECTOR& i_stV, const VECTOR& i_stV2);
	VECTOR operator - (const VECTOR& i_stV, const VECTOR& i_stV2);
	VECTOR operator * (const VECTOR& i_stV, double val);
	VECTOR operator / (const VECTOR& i_stV, double val);
	VECTOR operator * (const VECTOR& i_stV, const VECTOR& i_stV2);
	VECTOR operator / (const VECTOR& i_stV, const VECTOR& i_stV2);
}

inline double PathUtils::PointDistance(double i_iX1, double i_iY1, double i_iX2, double i_iY2)
{
	double iDistX = i_iX1 - i_iX2;
	double iDistY = i_iY1 - i_iY2;
	return sqrt(iDistX*iDistX + iDistY*iDistY);
}

inline double PathUtils::PointDistance(double i_iX1, double i_iY1)
{
	double iDistX = i_iX1;
	double iDistY = i_iY1;
	return sqrt(iDistX*iDistX + iDistY*iDistY);
}

inline double PathUtils::PointDistance2(double i_iX1, double i_iY1, double i_iX2, double i_iY2)
{
	double iDistX = i_iX1 - i_iX2;
	double iDistY = i_iY1 - i_iY2;
	return (iDistX*iDistX + iDistY*iDistY);
}

inline double PathUtils::PointDistance2(double i_iX1, double i_iY1)
{
	double iDistX = i_iX1;
	double iDistY = i_iY1;
	return (iDistX*iDistX + iDistY*iDistY);
}

inline double PathUtils::EuclideanDistance(const POINT& p, const POINT& q)
{
	double dx = (p.x - q.x);
	double dy = (p.y - q.y);

	return sqrt(dx*dx + dy*dy);
}

inline bool PathUtils::IsUnitDistance(const POINT& p, const POINT& q)
{
	return (p.x == q.x || p.y == q.y);
}

inline double PathUtils::CalcLineAngle(double x1, double y1, double x2, double y2)
{
	double difx = x2 - x1;
	double dify = y2 - y1;

	//alt1
	//double l = PointDistance(x1, y1, x2, y2);
	//if(l < 0.0001) l = 0.0001;
	//double dResult = acos(difx / l) * (180.0/PI);
	//if(dify < 0.0) dResult = 360.0 - dResult;

	//alt2
	double dResult = 360.0 - (atan2(-dify, difx) * (180.0/PI));
	if (dResult >= 360.0) dResult -= 360.0;
	else if (dResult < 0.0) dResult += 360.0;

	return dResult;
}

inline void PathUtils::MidPoint(double i_iX1, double i_iY1, double i_iX2, double i_iY2, double *o_iX, double *o_iY)
{
	*o_iX = (i_iX1+i_iX2)/2;
	*o_iY = (i_iY1+i_iY2)/2;
}

inline POINT PathUtils::MidPoint(double i_iX1, double i_iY1, double i_iX2, double i_iY2)
{
	POINT p = {(int)((i_iX1+i_iX2)/2), (int)((i_iY1+i_iY2)/2)};

	return p;
}

inline int PathUtils::Round(double x)
{
	return (int)(x + 0.5);
}

inline double PathUtils::Signum(const double& x)
{
	if(x < 0) {
		return -1.0;
	}
	if(x > 0) {
		return 1.0;
	}

	return 0.0;
}

inline void PathUtils::Normalize(VECTOR& v)
{
	double len = (double)sqrt(v.x*v.x + v.y*v.y);
	v.x /= len;
	v.y /= len;
}

inline bool PathUtils::PointWithinSegment(POINTD& p, VECTOR& l0, VECTOR& l1)
{
	double dDistTot = PointDistance(l0.x, l0.y, l1.x, l1.y);
	double dDistP = PointDistance(l0.x, l0.y, p.x, p.y) + PointDistance(l1.x, l1.y, p.x, p.y);
	return (dDistP <= dDistTot+0.0002);
}

inline double PathUtils::PointDistanceToLineSeg(POINTD& p, VECTOR& l0, VECTOR& l1, POINTD *o_pstClosest)
{
	VECTOR l0_l1;
	double dist;

	l0_l1 = l1 - l0;

	double l2 = PointDistance2(l0_l1.x, l0_l1.y); // i.e. |l0-l1|^2 -  avoid a sqrt

	if(l2 < 0.1) {
		//just take l0, the line is too short
		dist = PointDistance(l0.x, l0.y, p.x, p.y);
		if(o_pstClosest) *o_pstClosest = l0;
	} else {
		// Consider the line extending the segment, parameterized as l0 + t (l1 - l0).
		// We find projection of point p onto the line. 
		// It falls where t = [(p-l0) . (l1-l0)] / |l1-l0|^2
		double t = Dot(p - l0, l0_l1) / l2;
		if(t < 0.0) {
			// Beyond the 'l0' end of the segment
			dist = PointDistance(l0.x, l0.y, p.x, p.y);
			if(o_pstClosest) *o_pstClosest = l0;
		} else if(t > 1.0) {
			// Beyond the 'l0' end of the segment
			dist = PointDistance(l1.x, l1.y, p.x, p.y);
			if(o_pstClosest) *o_pstClosest = l1;
		} else {
			// Projection falls on the segment
			VECTOR projection = l0 + (l0_l1 * t);
			dist = PointDistance(p.x, p.y, projection.x, projection.y);
			if(o_pstClosest) *o_pstClosest = projection;
		}
	}

	return dist;
}

//"constructor" for VECTOR
inline VECTOR PathUtils::S_VECTORMAKE2(double x, double y)
{
	VECTOR result;
	result.x = x;
	result.y = y;
	return result;
}

inline VECTOR PathUtils::operator + (const VECTOR& i_stV, const VECTOR& i_stV2)
{
	return S_VECTORMAKE2(i_stV.x+i_stV2.x, i_stV.y+i_stV2.y);
}

inline VECTOR PathUtils::operator - (const VECTOR& i_stV, const VECTOR& i_stV2)
{
	return S_VECTORMAKE2(i_stV.x-i_stV2.x, i_stV.y-i_stV2.y);
}

inline VECTOR PathUtils::operator * (const VECTOR& i_stV, double val)
{
	return S_VECTORMAKE2(i_stV.x*val, i_stV.y*val);
}

inline VECTOR PathUtils::operator / (const VECTOR& i_stV, double val)
{
	return S_VECTORMAKE2(i_stV.x/val, i_stV.y/val);
}

inline VECTOR PathUtils::operator * (const VECTOR& i_stV, const VECTOR& i_stV2)
{
	return S_VECTORMAKE2(i_stV.x*i_stV2.x, i_stV.y*i_stV2.y);
}

inline VECTOR PathUtils::operator / (const VECTOR& i_stV, const VECTOR& i_stV2)
{
	return S_VECTORMAKE2(i_stV.x/i_stV2.x, i_stV.y/i_stV2.y);
}
