#include "StandardRaceAI.h"

#include "RaceAITasks.h"
#include "mapobjects.h"

StandardRaceAI::StandardRaceAI(const S_StaticShipInfo& staticShipInfo) :
	  C_RaceMapAI(staticShipInfo), _pathFinder(NULL), _checkpoint(NULL), 
		  _initPathfinderTask(NULL), _findPathTask(NULL) {}

StandardRaceAI::~StandardRaceAI(void)
{
	if (_pathFinder)
	{
		delete _pathFinder;
	}
}

void StandardRaceAI::getOutput(const S_ShipInfo& shipInfo, S_AIOutput& output)
{
	C_RaceMapAI::getOutput(shipInfo, output);

	if (_findPathTask != NULL && _findPathTask->isFinished())
	{
		output.bFire = false;
	}
	else
	{
		output.bFire = output.bLeft = output.bRight = output.bThrottle = false;
	}
}

void StandardRaceAI::setUp()
{
	C_RaceMapAI::setUp();
	PathFinderSettings settings;  //standard settings
	settings.BufferZoneWidth = 0;
	settings.NumSubtilesPerAxis = 4;
	settings.TimeStepMs = 150;
	
	_pathFinder = new FineGrainedPathFinder(*_map, _shipImg, _shipRectangles, _numActualDirections, settings);

	_initPathfinderTask = new SetUpPathfinderTask(*_pathFinder);
	enqueueTask(_initPathfinderTask);
}

void StandardRaceAI::setCurrentCheckpoint(const C_CheckPoint* pCheckpt)
{
	if (!_initPathfinderTask->isFinished())
	{
		return;
	}

	if (pCheckpt != _checkpoint)
	{
		_checkpoint = const_cast<C_CheckPoint*>(pCheckpt);
		if (_findPathTask != NULL)
		{
			if (!_findPathTask->isFinished())
			{
				_findPathTask->stop();
				_findPathTask->wait();
			}
			delete _findPathTask;
			_findPathTask = NULL;
		}

		S_CheckPointInfo stCP;
		_checkpoint->GetCheckPointInfo(&stCP);

        POINT dest = PathUtils::MidPoint(stCP.iX1, stCP.iY1, stCP.iX2, stCP.iY2);
        
		_findPathTask = new FindPathTask(*_pathFinder, _lastKnownDynamicShipInfo, dest, _moves);
		enqueueTask(_findPathTask);
	}
}

S_ExtendedAIOutput* StandardRaceAI::getClosestState()
{
	return NULL;
}
