// C_Bullet, C_BulletList

//classes to handle bullets. a list class that holds all existing
// bullets in a game.
// - updates bullets position
// - draw bullets
// - collision detections against the map and objects.

#include "bullet.h"
#include "global.h"

C_Image *C_Bullet::s_pclBulletImg = NULL;
int C_Bullet::s_iObjCount = 0;

C_Bullet::C_Bullet(C_Map *i_pclMap, S_BulletInfo *i_pstBulletInfo, int i_iOwnerID) : C_Object()
{
	m_iOwnerID = i_iOwnerID;

	m_pclMap = i_pclMap;
	m_stScreenArea = *(m_pclMap->GetScreenRect());

	m_dPosX = i_pstBulletInfo->dPosX;
	m_dPosY = i_pstBulletInfo->dPosY;
	m_dDirection = i_pstBulletInfo->dDirection;
	m_dVelX = i_pstBulletInfo->dVelX;
	m_dVelY = i_pstBulletInfo->dVelY;

	if(!s_iObjCount) {
		s_pclBulletImg = new C_Image((char*)"bullet.tga", C_Global::szResourceFile);
		C_GraphWrapperGL::GetGraphWrapperObject()->LoadTextureAbs(s_pclBulletImg);
	}
	s_iObjCount++;
	m_pclImage = s_pclBulletImg;
	m_pstImageRect = NULL;
	m_dTotAlive = 0.0;
}

C_Bullet::~C_Bullet()
{
	//last object alive?
	s_iObjCount--;
	if(!s_iObjCount) {
		delete s_pclBulletImg;
		s_pclBulletImg = NULL;
	}
}

bool C_Bullet::Update(double i_dTime)
{
	m_dTotAlive += i_dTime;
	if(m_dTotAlive > BULLETLIFETIME) return false;
	OUpdate(i_dTime);

	return !m_pclMap->CheckCollision(this);
}

void C_Bullet::Draw()
{
	int iX, iY;
	m_pclMap->GetPos(&iX, &iY);
	m_stScreenArea = *(m_pclMap->GetScreenRect());
	ODraw(s_pclBulletImg, m_pstImageRect, iX, iY);
}

////////////////////////////////////////////////////////////////////////////////
C_BulletList::C_BulletList()
{
	m_iGetPos = 0;
}

C_BulletList::~C_BulletList()
{
	C_Bullet *pclBullet;
	int i, iCount = (int)m_clBullets.size();

	for(i=0; i<iCount; i++) {
		pclBullet = m_clBullets[i];
		delete pclBullet;
	}
}

C_Bullet *C_BulletList::GetNext(bool i_bReset)
{
	if(i_bReset) m_iGetPos = 0;
	if(m_iGetPos<(int)m_clBullets.size()) return m_clBullets[m_iGetPos++];
	else return NULL;
}

void C_BulletList::Add(C_Bullet *i_pclBullet)
{
	m_clBullets.push_back(i_pclBullet);
}

void C_BulletList::Remove(int i_iUID)
{
	C_Bullet *pclBullet;
	int i, iCount = (int)m_clBullets.size();

	for(i=0; i<iCount; i++) {
		pclBullet = m_clBullets[i];

		if(pclBullet->GetUId() == i_iUID) {
			if(i<m_iGetPos) m_iGetPos--;
			m_clBullets.erase(m_clBullets.begin()+i);
			delete pclBullet;
			break;
		}
	}
}

void C_BulletList::Update(double i_dTime)
{
	C_Bullet *pclBullet;
	int i, iCount = (int)m_clBullets.size();

	for(i=0; i<iCount; i++) {
		pclBullet = m_clBullets[i];

		if(!pclBullet->Update(i_dTime)) {
			//collision with map
			m_clBullets.erase(m_clBullets.begin()+i);
			iCount--; i--; //compensate list info for removal
			delete pclBullet;
		}
	}
}

void C_BulletList::Draw()
{
	C_Bullet *pclBullet;
	int i, iCount = (int)m_clBullets.size();

	for(i=0; i<iCount; i++) {
		pclBullet = m_clBullets[i];
		pclBullet->Draw();
	}
}

bool C_BulletList::CheckCollision(C_Object *i_pclObject, int *o_iUID, bool i_bPlayerBulletsCollision)
{
	C_Bullet *pclBullet;
	int i, iCount = (int)m_clBullets.size();
	bool bResult = false;
	int iBulletOwner;
	int iObjId = i_pclObject->GetOwnerID();

	if(iObjId > COMPUTER_ENEMY_BASEID) iObjId = COMPUTER_ENEMY_BASEID; //special case for computer enemies

	for(i=0; i<iCount; i++) {
		pclBullet = m_clBullets[i];

		iBulletOwner = pclBullet->GetOwnerID();
		if(iBulletOwner > COMPUTER_ENEMY_BASEID) iBulletOwner = COMPUTER_ENEMY_BASEID; //computer enemy

		//skip player bullet if !i_bPlayerBulletsCollision
		if(!i_bPlayerBulletsCollision && iBulletOwner<COMPUTER_ENEMY_BASEID) continue;

		//free from own bullets
		if(iBulletOwner != iObjId) {
			if(i_pclObject->OCheckCollision(pclBullet)) {
				m_clBullets.erase(m_clBullets.begin()+i);
				*o_iUID = pclBullet->GetUId();
				bResult = true;
				delete pclBullet;
				break; //first bullet hit wins
			}
		}
	}
	return bResult;
}
