// C_GameView

//class to control the game gui.
// - handles input for the 4+1 dialogs
// - draw the active dialog
// - delivers set gui options to the game handler on request

#include "gameview2.h"

#include "common.h"
#include "fileresource.h"
#include "global.h"

#define LASTCONNECTED_FILENAME "last_connect.txt"

#define GW_DIALOG0    0x06 //member login
#define GW_DIALOG1    0x01 //main connect gui
#define GW_DIALOG2    0x02 //main game gui
#define GW_DIALOG3    0x03 //replay gui
#define GW_DIALOG5    0x05 //achievements "popup" (displayed when GW_DIALOG2 is displayed)


//gui controls

//GW_DIALOG0
#define ID_BLOGIN        1
#define ID_BCANCEL       2
#define ID_EMEMBER       3
#define ID_EPASS         4
#define ID_ELOGEDIN      5
#define ID_BSTARTWEB     6

//GW_DIALOG1
#define ID_BCONNECT      1
#define ID_BJOIN         2
#define ID_BNEW          3
#define ID_EIP           4
#define ID_EPLAYER       5
#define ID_EGAME         6
#define ID_ECONNECTED    7
#define ID_LGAMES        8
#define ID_BREPLAY       9
#define ID_BWEBSCORE     10

//GW_DIALOG2
#define ID_LPLAYERS      1
#define ID_BSTART        2
#define ID_BBACK         3
#define ID_BSEND         4
#define ID_LCHAT_D       5
#define ID_ECHAT         6
#define ID_CFLYTHROUGH   8
#define ID_CNOFIRE       10
#define ID_CNOBULLETSHIT 16
#define ID_LMAPS         13
#define ID_EMAXPLAYERS   14
#define ID_BSAVEREPLAY   18
#define ID_BADDAI        19
#define ID_BSUBAI        20

//GW_DIALOG3
#define ID_LRFILES       1
#define ID_BRPLAY        2
#define ID_BRBACK        3

//GW_DIALOG5
#define ID_SSCORE        1
#define ID_LLIMITS       2

//constructor
C_GameView::C_GameView(C_GraphWrapperGL *i_pclGraph, C_MemberAchievements *i_pclMemberAch)
{
	m_pclGraph  = i_pclGraph;

	m_pclMemberAch = i_pclMemberAch;

	m_pclImgGame        = new C_Image((char*)"icongame16.tga", C_Global::szResourceFile);
	m_pclImgPlayer      = new C_Image((char*)"iconship16.tga", C_Global::szResourceFile);
	m_pclImgAchievement = new C_Image((char*)"iconachieve16.tga", C_Global::szResourceFile);

	m_pclMiniMapImg = NULL;

	m_szGame[0] = 0;
	LoadLastConnected();

	m_pclMessageTimer = new C_Timer();
	m_szTextMessage[0] = 0;

	strcpy(m_szExternalIP, "Your server IP: unknown");

	strcpy(m_szLastMapName, "race00_3pl");

	m_pclDlg0 = NULL;
	m_pclDlg1 = NULL;
	m_pclDlg2 = NULL;
	m_pclDlg3 = NULL;
	m_pclDlg5 = NULL;
	ReInitDialogs();

	m_iState = GW_DIALOG0;
	m_iLogedIn = 0;
}

//destructor
C_GameView::~C_GameView()
{
	delete m_pclDlg0;
	delete m_pclDlg1;
	delete m_pclDlg2;
	delete m_pclDlg3;
	delete m_pclDlg5;

	delete m_pclMiniMapImg;
	delete m_pclImgGame;
	delete m_pclImgPlayer;
	delete m_pclImgAchievement;

	delete m_pclMessageTimer;
}

//private method that reinits dialogs (called only from contructor)
bool C_GameView::ReInitDialogs()
{
	delete m_pclDlg0;
	delete m_pclDlg1;
	delete m_pclDlg2;
	delete m_pclDlg3;
	delete m_pclDlg5;

	if(m_pclMiniMapImg) delete m_pclMiniMapImg;
	m_pclMiniMapImg = NULL;

	m_iPollMsg = 0;
	m_bMaster  = false;
	m_iState = GW_DIALOG1;
	m_bHidden = true;

	m_iConnected   = 0;
	m_szConnect[0] = 0;
	m_iGameIndex   = 0;
	m_bNoShootingForced = false;
	m_bFlyThroughForced = false;
	m_bNoBulletHitForced = false;

	m_szChatMsg[0] = 0;

	//create all dialoges
	m_pclDlg0 = new C_GUI();
	m_pclDlg1 = new C_GUI();
	m_pclDlg2 = new C_GUI();
	m_pclDlg3 = new C_GUI();
	m_pclDlg5 = new C_GUI();
	
	//correct position is set before draw
	m_pclDlg0->LoadTemplate((char*)"gui_desc0.txt", C_Global::szResourceFile, 0, 0);
	m_pclDlg1->LoadTemplate((char*)"gui_desc1.txt", C_Global::szResourceFile, 0, 0);
	m_pclDlg2->LoadTemplate((char*)"gui_desc2.txt", C_Global::szResourceFile, 0, 0);
	m_pclDlg3->LoadTemplate((char*)"gui_desc3.txt", C_Global::szResourceFile, 0, 0);
	m_pclDlg5->LoadTemplate((char*)"gui_desc5.txt", C_Global::szResourceFile, 0, 0);

	//init dlg0
	m_pclDlg0->SetText(ID_EMEMBER, m_szMember);
	m_pclDlg0->SetText(ID_EPASS,   m_szPassword);
	UpdateDlg0ButtonStatus();

	//init dlg1
	m_pclDlg1->SetText(ID_EIP,     m_szLastConnected);
	m_pclDlg1->SetText(ID_EPLAYER, m_szPlayer);
	//m_pclDlg1->SetText(ID_EGAME,   m_szGame);

	//init dlg2
	S_GUIListEl stEl;
	memset(&stEl, 0, sizeof(stEl));
	stEl.iID = 0; //not used, identified by the strings
	stEl.pclIcon = m_pclImgAchievement;
	stEl.iIconRect = 3; //this is set later after the player name is entered

	//all dogfight levels
	strcpy(stEl.szText[0], "dogfight0"); m_pclDlg2->ListAddElement(ID_LMAPS, &stEl);
	strcpy(stEl.szText[0], "dogfight1"); m_pclDlg2->ListAddElement(ID_LMAPS, &stEl);
	strcpy(stEl.szText[0], "dogfight2"); m_pclDlg2->ListAddElement(ID_LMAPS, &stEl);
	strcpy(stEl.szText[0], "dogfight3"); m_pclDlg2->ListAddElement(ID_LMAPS, &stEl);

	//all score levels (mission, race)
	S_LevelElement *pclLevel;
	while((pclLevel = m_pclMemberAch->GetNextLevelInfo())!=NULL) {
		strcpy(stEl.szText[0], pclLevel->szLevel);
		m_pclDlg2->ListAddElement(ID_LMAPS, &stEl);
	}

	//add all external map files to the list
	char szExternalMapPath[MAX_PATH+32];
	sprintf(szExternalMapPath, "%s/external_maps", C_Global::szDataPath);
	void *pSearch = C_Resource::FileSearchOpen(szExternalMapPath, (char*)"*.des");
	while(C_Resource::FileSearchNext(pSearch, stEl.szText[0], sizeof(stEl.szText[0]))) {
		char *szDot = strrchr(stEl.szText[0], '.');
		if(szDot) *szDot = 0;
		m_pclDlg2->ListAddElement(ID_LMAPS, &stEl);
	}
	C_Resource::FileSearchClose(pSearch);

	return true;
}

//set dialogs to show or not (considered in the draw method)
void C_GameView::Show(bool i_bShow)
{
	m_bHidden = !i_bShow;
}

//set data to gui (dlg0)
void C_GameView::SetLogedInState(int i_iMemberLogedIn) //0 not loged in, 1 loged in, -1 waiting
{
	m_iLogedIn = i_iMemberLogedIn;

	if(m_iLogedIn!=1) {
		if(m_iLogedIn==-1) {
			m_pclDlg0->SetText(ID_ELOGEDIN, (char*)"Waiting...");
			m_pclDlg0->SetColor(ID_ELOGEDIN, COLOR_DISABLED, 0.8f, 0.7f, 0.1f, 0.8f);
		} else {
			m_pclDlg0->SetText(ID_ELOGEDIN, (char*)"");
			m_pclDlg0->SetColor(ID_ELOGEDIN, COLOR_DISABLED, 0.8f, 0.0f, 0.0f, 0.8f);
		}
	} else {
		m_pclDlg0->SetText(ID_ELOGEDIN, (char*)"");
		m_pclDlg0->SetColor(ID_ELOGEDIN, COLOR_DISABLED, 0.0f, 0.8f, 0.0f, 0.8f);
		/**/SaveLastConnected();
		m_iState = GW_DIALOG1;

		//force player name to member name
		strcpy(m_szPlayer, m_szMember);
		m_pclDlg1->SetText(ID_EPLAYER, m_szPlayer);
		m_pclDlg1->SetActive(ID_EPLAYER, false);
	}
}

//get data entered in gui (dlg0)
void C_GameView::GetUserPass(char *o_szUser, char *o_szPass)
{
	strcpy(o_szUser, m_szMember);
	strcpy(o_szPass, m_szPasswordMD5);
}

//get data entered in gui (dlg1)
void C_GameView::GetConnectString(char *o_szConnect)
{
	if(o_szConnect) strcpy(o_szConnect, m_szConnect);
}

//get data entered in gui (dlg1)
void C_GameView::GetGameName(char *o_szGame)
{
	if(o_szGame) strcpy(o_szGame, m_szGame);
}

//get data entered in gui (dlg1)
void C_GameView::GetPlayerName(char *o_szPlayer)
{
	if(o_szPlayer) strcpy(o_szPlayer, m_szPlayer);
}

//get data entered in gui (dlg2)
void C_GameView::GetChatString(char *o_szChatString)
{
	sprintf(o_szChatString, "%s: %s", m_szPlayer, m_szChatMsg);
}

//get data entered in gui (dlg2)
bool C_GameView::GetGameOpt(char *o_szFilename, int *o_iNumLaps, bool *o_bNoShooting, bool *o_bFlyThrough, bool *o_bNoBulletHit, int *o_iLevelType, int *o_iNumAI)
{
	if(m_iLevelType!=MAP_RACE) *o_iNumLaps = 0;
	else                       *o_iNumLaps = m_iNumLaps;

	*o_iLevelType   = m_iLevelType;
	*o_bNoShooting  = m_bNoShooting;
	*o_bFlyThrough  = m_bFlyThrough;
	*o_bNoBulletHit = m_bNoBulletHit;
	*o_iNumAI       = m_iNumAI;

	strcpy(o_szFilename, m_szFilename);
	return true;
}

//get given player name (dlg2)
void C_GameView::GetPlayerName(int i_iIndex, char *o_szPlayer)
{
	S_GUIListEl *pclPlayer = m_pclDlg2->ListGetElementById(ID_LPLAYERS, i_iIndex);
	if(pclPlayer) {
		strcpy(o_szPlayer, pclPlayer->szText[0]);
	} else o_szPlayer[0] = 0;
}

//set data to gui (dlg1)
void C_GameView::SetExternalIP(char *i_szExternalIP)
{
	sprintf(m_szExternalIP, "Your server IP: %s", i_szExternalIP);
}

//set data to gui (dlg1)
void C_GameView::AddGame(char *i_szName, bool i_bPlaying)
{
	S_GUIListEl stListEl;
	memset(&stListEl, 0, sizeof(stListEl));
	stListEl.iID = m_iGameIndex;
	strcpy(stListEl.szText[0], i_szName);
	stListEl.pclIcon = m_pclImgGame;
	stListEl.iIconRect = i_bPlaying ? 0 : 1;
	m_pclDlg1->ListAddElement(ID_LGAMES, &stListEl);
	UpdateDlg1ButtonStatus();

	m_iGameIndex++;
}

//set data to gui (dlg1)
void C_GameView::RemoveGame(char *i_szName)
{
	m_pclDlg1->ListRemoveElement(ID_LGAMES, i_szName);
	UpdateDlg1ButtonStatus();
}

//set data to gui (dlg1)
void C_GameView::SetGameStatus(char *i_szName, bool i_bPlaying)
{
	S_GUIListEl *pclGame = m_pclDlg1->ListGetElement(ID_LGAMES, i_szName);
	if(pclGame) {
		pclGame->iIconRect = i_bPlaying ? 0 : 1;
		UpdateDlg1ButtonStatus(); //needed if the changed game is the selected game
	}
}

//set data to gui (dlg2)
void C_GameView::AddPlayer(char *i_szName, int i_iIndex)
{
	//no player must have same name!
	S_GUIListEl stListEl;
	memset(&stListEl, 0, sizeof(stListEl));
	stListEl.iID = i_iIndex;
	strcpy(stListEl.szText[0], i_szName);
	stListEl.pclIcon = m_pclImgPlayer;
	stListEl.iIconRect = i_iIndex;
	m_pclDlg2->ListAddElement(ID_LPLAYERS, &stListEl);

	UpdateStartState();
}

//set data to gui (dlg2)
void C_GameView::RemovePlayer(int i_iIndex)
{
	m_pclDlg2->ListRemoveElementById(ID_LPLAYERS, i_iIndex);

	UpdateStartState();
}

//set data to gui (dlg2)
void C_GameView::SetPlayerScore(int i_iIndex, int i_iScore, bool i_bAI)
{
	char szScore[12];
	sprintf(szScore, "%.2f", i_iScore/1000.0f);

	if(!i_bAI) {
		S_GUIListEl *pclPlayer = m_pclDlg2->ListGetElementById(ID_LPLAYERS, i_iIndex);
		if(pclPlayer) {
			strcpy(pclPlayer->szText[1], szScore);
		}
		if(i_iIndex==m_iIndex) {
			//save achievements - now done globally to webserver
/*			char szLevel[40];
			strcpy(szLevel, m_szFilename);
			char *szDot = strrchr(szLevel, '.');
			if(szDot) szDot[0] = 0;
			m_pclMemberAch->SetPlayerScore(szLevel, i_iScore);
			m_pclLevels->SetPlayerScore(szLevel, i_iScore);
			SetMapName(NULL); //get achievements gui to update
			SetAchievements(); //get achievements to update in level list
*/
		}
	} else {
		strcpy(m_szAIScore[i_iIndex], szScore);
	}
}

void C_GameView::UpdateAchievements()
{
	SetMapName(NULL); //get achievements gui to update
	SetAchievements(); //get achievements to update in level list
}

//set data to gui (dlg2)
void C_GameView::AddChatString(char *i_szChatString)
{
	S_GUIListEl stListEl;
	memset(&stListEl, 0, sizeof(stListEl));
	strcpy(stListEl.szText[0], i_szChatString);
	m_pclDlg2->ListAddElement(ID_LCHAT_D, &stListEl);
	m_pclDlg2->ListSetSelectedElement(ID_LCHAT_D, NULL); //selects last element
}

//set dlg1 as active dialog
bool m_bMemberLoginVisited = false;
void C_GameView::SetDlg1State()
{
	SetConnected(0);

	if (!m_bMemberLoginVisited) {
		m_bMemberLoginVisited = true;
		m_iState = GW_DIALOG0;
	} else {
		m_iState = GW_DIALOG1;
	}
	m_bHidden  = false;
	m_iPollMsg = 0; //reset
}

//handle server messages

void C_GameView::SetConnected(int i_iConnected, bool i_bClearLists) //0 not connected, 1 connected, -1 waiting
{
	m_iConnected = i_iConnected;
	if(m_iConnected!=1) {
		m_iGameIndex = 0;
		m_bMaster = false;

		i_bClearLists = true;
		if(i_iConnected==-1) {
			m_pclDlg1->SetText(ID_ECONNECTED, (char*)"Waiting...");
			m_pclDlg1->SetColor(ID_ECONNECTED, COLOR_DISABLED, 0.8f, 0.7f, 0.1f, 0.8f);
		} else {
			m_pclDlg1->SetText(ID_ECONNECTED, (char*)"");
			m_pclDlg1->SetColor(ID_ECONNECTED, COLOR_DISABLED, 0.8f, 0.0f, 0.0f, 0.8f);
		}
	} else {
		m_pclDlg1->SetText(ID_ECONNECTED, (char*)"");
		m_pclDlg1->SetColor(ID_ECONNECTED, COLOR_DISABLED, 0.0f, 0.8f, 0.0f, 0.8f);
		strcpy(m_szLastConnected, m_szConnect);
		//SaveLastConnected();
	}
	m_iState = GW_DIALOG1;
	if(i_bClearLists) {
		m_pclDlg1->ListRemoveAll(ID_LGAMES);
		m_pclDlg2->ListRemoveAll(ID_LPLAYERS);
	}
	UpdateDlg1ButtonStatus();
}

void C_GameView::SetJoinNewOK(int i_iIndex)
{
	m_iIndex = i_iIndex; //player index
	if(m_iState==GW_DIALOG1) {
		SaveLastConnected(); //save player name and connected ip

		ClearDlg2();
		SetMapName(m_szLastMapName);

		m_iState = GW_DIALOG2;
	}
}

void C_GameView::NewMaster()
{
	if(m_iState==GW_DIALOG2) {
		m_bMaster = true;
		m_pclDlg2->ListSetHasSelection(ID_LPLAYERS, true);
		m_pclDlg2->SetActive(ID_LPLAYERS, true);

		SetMapName(NULL);
		m_pclDlg2->SetActive(ID_LMAPS, true);
		UpdateDlg2CheckStatus();
		UpdateStartState();
	}
}

void C_GameView::ClearDlg2()
{
	m_pclDlg2->SetActive(ID_LMAPS, m_bMaster);
	m_pclDlg2->SetText(ID_ECHAT, (char*)"");
	m_pclDlg2->ListRemoveAll(ID_LCHAT_D);

	m_iNumLaps = 0;

	m_pclDlg2->SetActive(ID_BSTART,  false);

	m_bNoShootingForced  = false;
	m_bFlyThroughForced  = false;
	m_bNoBulletHitForced = false;
	m_bNoShooting  = false;
	m_bFlyThrough  = false;
	m_bNoBulletHit = false;
	m_pclDlg2->SetChecked(ID_CFLYTHROUGH,   false);
	m_pclDlg2->SetChecked(ID_CNOFIRE,       false);
	m_pclDlg2->SetChecked(ID_CNOBULLETSHIT, false);
	m_pclDlg2->SetActive(ID_CFLYTHROUGH,    m_bMaster);
	m_pclDlg2->SetActive(ID_CNOFIRE,        m_bMaster);
	m_pclDlg2->SetActive(ID_CNOBULLETSHIT,  m_bMaster);

	m_iNumAI = 0;
	for(int i=0; i<8; i++) m_szAIScore[i][0] = 0;
	m_pclDlg2->SetActive(ID_BADDAI,         m_bMaster);
	m_pclDlg2->SetActive(ID_BSUBAI,         false);

	m_pclDlg2->ListSetHasSelection(ID_LPLAYERS, m_bMaster);
	m_pclDlg2->SetActive(ID_LPLAYERS, m_bMaster);

	SetAchievements();
}

void C_GameView::SetAchievements()
{
	//set achievements in GUI for all levels for the current player name
	S_LevelElement *pclLevel;
	m_pclMemberAch->RestartLevelInfo();
	while((pclLevel = m_pclMemberAch->GetNextLevelInfo())!=NULL) {
		S_GUIListEl *pclEl = m_pclDlg2->ListGetElement(ID_LMAPS, pclLevel->szLevel);
		int iPlace = m_pclMemberAch->GetPlayerAchievements(pclLevel->szLevel, NULL);
		pclEl->iIconRect = iPlace;
	}
}

void C_GameView::SetGameOpt(char *i_szFilename, int i_iNumLaps, bool i_bNoShooting, bool i_bFlyThrough, bool i_bNoBulletHit, int i_iLevelType, int i_iNumAI)
{
	char *pDot = strrchr(i_szFilename, '.');
	if(pDot) *pDot = 0; //remove .des

	SetMapName(i_szFilename); //changes m_iLevelType

	m_bNoShooting  = i_bNoShooting;
	m_bFlyThrough  = i_bFlyThrough;
	m_bNoBulletHit = i_bNoBulletHit;
	if(i_iLevelType==MAP_RACE) { //numlaps
		m_iNumLaps   = i_iNumLaps;
	}
	m_iNumAI = i_iNumAI;

	if(m_iLevelType != i_iLevelType) {
		DisplayShortMessage((char*)"Selected map not found");
	}

	m_pclDlg2->SetChecked(ID_CNOFIRE,       m_bNoShooting);
	m_pclDlg2->SetChecked(ID_CFLYTHROUGH,   m_bFlyThrough);
	m_pclDlg2->SetChecked(ID_CNOBULLETSHIT, m_bNoBulletHit);
}

//handle gui actions

void C_GameView::HandleMemberLogin(bool i_bLogin)
{
	if(!i_bLogin) {
		m_iState = GW_DIALOG1;
	} else {
		if(m_iLogedIn==-1) return; //already waiting
		SetLogedInState(-1); //waiting for login
		m_pclDlg0->GetText(ID_EMEMBER, m_szMember);
		m_pclDlg0->GetText(ID_EPASS, m_szPassword);
		if(strcmp(m_szPassword, "Last Password")==0 ) {
			//use last m_szPasswordMD5
		} else {
			//get MD5 sum
			C_HttpUtil::md5chksum_encode(m_szPassword, m_szPasswordMD5, sizeof(m_szPasswordMD5));
		}
		
		m_iPollMsg = D0_MLOGIN;
	}

}

void C_GameView::HandleStartBrowser()
{
	m_iPollMsg = D0_MCREATE;
}

void C_GameView::HandleConnect()
{
	if(m_iConnected==-1) return; //already waiting
	SetConnected(-1); //waiting for connection
	m_pclDlg1->GetText(ID_EIP, m_szConnect);
	m_iPollMsg = D1_CONNECT;
}

void C_GameView::HandleExit()
{
	m_iPollMsg = -1;
}

void C_GameView::HandleNew()
{
	m_bMaster = true;

	m_pclDlg1->GetText(ID_EPLAYER, m_szPlayer);
	m_pclDlg1->GetText(ID_EGAME, m_szGame);

	m_iPollMsg = D1_NEW;
}

void C_GameView::HandleJoin()
{
	m_bMaster = false;

	m_pclDlg1->GetText(ID_EPLAYER, m_szPlayer);

	S_GUIListEl *pclGame = m_pclDlg1->ListGetSelectedElement(ID_LGAMES);
	if((m_szPlayer[0]!=0) && (m_iConnected==1) && (pclGame!=NULL)) {
		strcpy(m_szGame, pclGame->szText[0]);
		m_iPollMsg = D1_JOIN;
	}
}

void C_GameView::HandleStartBrowser2()
{
	m_iPollMsg = D1_MWEB;
}

void C_GameView::HandleStart()
{
	m_iPollMsg = D2_START;
}

void C_GameView::HandleBack()
{
	m_iPollMsg = D2_LEAVE;
	SetConnected(1, true);
	m_pclDlg1->ListRemoveAll(ID_LGAMES);
	m_pclDlg2->ListRemoveAll(ID_LPLAYERS);

	m_iState = GW_DIALOG1;
}

void C_GameView::HandleChatSend()
{
	m_pclDlg2->GetText(ID_ECHAT, m_szChatMsg);
	m_pclDlg2->SetText(ID_ECHAT, (char*)"");
	m_pclDlg2->SetFocused(ID_ECHAT);
	if(m_szChatMsg[0]!=0) {
		m_iPollMsg = D2_CHATSEND;
	}
}

void C_GameView::HandleSelectMap()
{
	S_GUIListEl *pclMap = m_pclDlg2->ListGetSelectedElement(ID_LMAPS);
	m_szFilename[0] = 0; m_iLevelType = -1;
	if(pclMap) SetMapName(pclMap->szText[0]);
	if(m_iLevelType==-1) DisplayShortMessage((char*)"Selected map is broken!");
	UpdateStartState();

	m_iPollMsg = D2_OPTCHANGE;
}

void C_GameView::HandleNumAI(int i_iAdd)
{
	m_iNumAI += i_iAdd;
	if(m_iNumAI<0) m_iNumAI = 0;
	if(m_iNumAI>7) m_iNumAI = 7;
	SetMapName(NULL);
	UpdateStartState();

	m_iPollMsg = D2_OPTCHANGE;
}

void C_GameView::HandleKickPlayer()
{
	if(m_bMaster) {
		//get marked item
		S_GUIListEl *pclPlayer = m_pclDlg2->ListGetSelectedElement(ID_LPLAYERS);

		if(pclPlayer && (pclPlayer->iID!=m_iIndex)) {
			//kick player if not self
			m_iKickIndex = pclPlayer->iID;
			m_iPollMsg = D2_KICK;
		}
	}
}

void C_GameView::HandleSave()
{
	m_iPollMsg = D2_SAVEREPLAY;
	m_pclDlg2->SetActive(ID_BSAVEREPLAY, false);
}

void C_GameView::ActivateSaveReplay(bool i_bActivate)
{
	m_pclDlg2->SetActive(ID_BSAVEREPLAY, i_bActivate);
}

void C_GameView::HandleReplay()
{
	m_iState = GW_DIALOG3;
	m_iPollMsg = D3_BEGIN;

	bool bReplayActive = false;
	S_GUIListEl stEl;
	memset(&stEl, 0, sizeof(stEl));
	stEl.iID = 0; //not used, identified by the strings
	stEl.szText[4][0] = -1;

	m_pclDlg3->ListRemoveAll(ID_LRFILES);
	if(m_pclDlg2->IsActive(ID_BSAVEREPLAY)) {
		strcpy(stEl.szText[0], "Last game");
		m_pclDlg3->ListAddElement(ID_LRFILES, &stEl);
		bReplayActive = true;
	}

	char szReplayPath[MAX_PATH+32];
	sprintf(szReplayPath, "%s/replays", C_Global::szUserPath);
	void *pSearch = C_Resource::FileSearchOpen(szReplayPath, (char*)"*.rpl");
	while(C_Resource::FileSearchNext(pSearch, stEl.szText[5],
		sizeof(stEl.szText[5]))) //store full name in string nr 5
	{
		char *pChr1, *pChr2;
		pChr2 = strrchr(stEl.szText[5], '.'); //at ".rpl"
		pChr1 = strchr(stEl.szText[5], '_'); //at _ #1
		if(pChr1 && pChr2 && pChr1<pChr2) {
			if(pChr1) pChr1 = strchr(pChr1+1, '_'); //at map name or NULL on old naming
			memset(stEl.szText[0], 0, sizeof(stEl.szText[0]));
			memset(stEl.szText[1], 0, sizeof(stEl.szText[0]));
			if(!pChr1) { //old naming before v1.78
				int iLen1 = (int)(pChr2-stEl.szText[5]);
				strncpy(stEl.szText[0], stEl.szText[5], iLen1);
			} else {
				int iLen1 = (int)(pChr1-stEl.szText[5]);
				int iLen2 = (int)(pChr2-pChr1)-1;
				strncpy(stEl.szText[1], stEl.szText[5], iLen1);
				strncpy(stEl.szText[0], pChr1+1, iLen2);
			}
			m_pclDlg3->ListAddElement(ID_LRFILES, &stEl);
			bReplayActive = true;
		}
	}
	C_Resource::FileSearchClose(pSearch);

	m_pclDlg3->SetActive(ID_BRPLAY, bReplayActive);
}

bool C_GameView::GetHiscoreMapName(char *o_szLevel)
{
	if(m_bIsHiscoreMap) {
		strcpy(o_szLevel, m_szFilename);
		char *szDot = strrchr(o_szLevel, '.');
		if(szDot) szDot[0] = 0;
	}
	return m_bIsHiscoreMap;
}

void C_GameView::HandleRBack()
{
	m_iState = GW_DIALOG1;
}

void C_GameView::HandleRPlay()
{
	m_iPollMsg = D3_REPLAY;
}

int C_GameView::GetReplayFile(char *o_szFilename, char *o_szName)
{
	int iResult = -2;
	S_GUIListEl *pstEl = m_pclDlg3->ListGetSelectedElement(ID_LRFILES);
	if(pstEl) {
		iResult = -1;
		//note:
		//place is stored in first char in string nr 4 for webserver files
		//filename is stored in string nr 5 for local files
		if(strcmp(pstEl->szText[0], "Last game")==0) {
			//last game
			strcpy(o_szFilename, pstEl->szText[0]);
		} else if(pstEl->szText[4][0]==-1) {
			//local file
			sprintf(o_szFilename, "%s/replays/%s", C_Global::szUserPath, pstEl->szText[5]);
		} else {
			//webserver replay
			iResult = pstEl->szText[4][0];
			strcpy(o_szFilename, pstEl->szText[0]);
			strcpy(o_szName, pstEl->szText[1]);
		}
	}
	return iResult;
}

void C_GameView::AddExternalReplay(char *i_szLevel, char *i_szName, int i_iPlace, int i_iScoreMs)
{
	S_GUIListEl stEl;
	memset(&stEl, 0, sizeof(stEl));
	stEl.iID = 0; //not used, identified by the strings
	S_LevelElement *pstInfo = m_pclMemberAch->GetLevelInfo(i_szLevel);
	sprintf(stEl.szText[0], "%s", i_szLevel);
	sprintf(stEl.szText[1], "%s", i_szName);
	sprintf(stEl.szText[2], "%d", i_iPlace);
	if(pstInfo) sprintf(stEl.szText[3], "%.2f%s", i_iScoreMs/1000.0f, pstInfo->bTime ? "s":"");
	stEl.szText[4][0] = (char)i_iPlace; //store place here for easy access later (column not visible)

	m_pclDlg3->ListAddElement(ID_LRFILES, &stEl);
	m_pclDlg3->SetActive(ID_BRPLAY, true);
}

int C_GameView::Poll()
{
	int iTmpMsg = m_iPollMsg;
	m_iPollMsg = 0;
	return iTmpMsg;
}

void C_GameView::UpdateDlg0ButtonStatus()
{
	bool bEnableLogin = true;

	char szName[48];
	char szPass[48];

	m_pclDlg0->GetText(ID_EMEMBER, szName);
	bool bPlayerNameTooLong = (m_pclDlg1->GetTextWidth(szName) > 88);
	if(szName[0]==0 || bPlayerNameTooLong) {
		bEnableLogin = false;
	}

	m_pclDlg0->GetText(ID_EPASS, szPass);
	if(szPass[0]==0) bEnableLogin = false;

	m_pclDlg0->SetActive(ID_BLOGIN, bEnableLogin);
}

void C_GameView::UpdateDlg1ButtonStatus()
{
	bool bEnableConnect = false;
	bool bEnableJoin    = false;
	bool bEnableNew     = false;

	char szName[48];

	m_pclDlg1->GetText(ID_EPLAYER, szName);
	bool bPlayerNameTooLong = (m_pclDlg1->GetTextWidth(szName) > 88);
	if(szName[0]!=0) {
		if(m_iConnected==1 && !bPlayerNameTooLong) {
			S_GUIListEl *pclGame = m_pclDlg1->ListGetSelectedElement(ID_LGAMES);
			if(pclGame && pclGame->iIconRect==1) bEnableJoin = true;
		}
		m_pclDlg1->GetText(ID_EGAME, szName);
	}
	if(szName[0]!=0 && m_iConnected==1 && !bPlayerNameTooLong) bEnableNew = true;

	m_pclDlg1->GetText(ID_EIP, szName);
	if(szName[0]!=0 && m_iConnected!=-1) bEnableConnect = true;

	m_pclDlg1->SetActive(ID_BCONNECT, bEnableConnect);
	m_pclDlg1->SetActive(ID_BJOIN,    bEnableJoin);
	m_pclDlg1->SetActive(ID_BNEW,     bEnableNew);
	m_pclDlg1->SetActive(ID_BREPLAY,  m_iConnected==1);
}

void C_GameView::UpdateDlg2ChatStatus()
{
	bool bEnableSend = false;

	char szText[128];
	m_pclDlg2->GetText(ID_ECHAT, szText);
	if(szText[0]!=0) bEnableSend = true;

	m_pclDlg2->SetActive(ID_BSEND, bEnableSend);
}

void C_GameView::UpdateDlg2CheckStatus()
{
	//end conditions
	//all end conditions now depends on map file

	//game options
	m_bNoShooting  = m_pclDlg2->IsChecked(ID_CNOFIRE);
	m_bFlyThrough  = m_pclDlg2->IsChecked(ID_CFLYTHROUGH);
	m_bNoBulletHit = m_pclDlg2->IsChecked(ID_CNOBULLETSHIT);

	UpdateStartState();

	m_iPollMsg = D2_OPTCHANGE;
}

int C_GameView::NumPlayers(bool i_bWithAI)
{
	S_GUIObject *pclList = m_pclDlg2->GetObject(ID_LPLAYERS);
	int iNumPlayers = (int)pclList->pclListEl->size();

	if(i_bWithAI) return iNumPlayers + m_iNumAI;
	else return iNumPlayers;
}

void C_GameView::UpdateStartState()
{
	bool bEnable = true; //assume we can start

	//must be master and ready
	if(!m_bMaster) bEnable = false;

	//a valid map must be selected
	S_GUIListEl *pclMap = m_pclDlg2->ListGetSelectedElement(ID_LMAPS);
	if(!pclMap) bEnable = false;
	if(m_iLevelType==-1) bEnable = false;

	//there is no ready state for players now
	//all players are always ready
	S_GUIObject *pclList = m_pclDlg2->GetObject(ID_LPLAYERS);
	int iNumPlayers = (int)pclList->pclListEl->size();

	//must not exceed player limit for the level
	if((iNumPlayers+m_iNumAI) > m_iMaxPlayers) bEnable = false;
	if((iNumPlayers+m_iNumAI) < m_iMinPlayers) bEnable = false;

	//cannot run dogfight with only one player
	if((iNumPlayers+m_iNumAI)==1 && m_iLevelType==MAP_DOGFIGHT) bEnable = false;

	//currently cannot have AI players in mission
	if(m_iNumAI>0 && (m_iLevelType==MAP_MISSION || m_iLevelType==MAP_MISSION_COOP)) bEnable = false;

	m_pclDlg2->SetActive(ID_BSTART, bEnable);
}

void C_GameView::SetMapName(char *i_szMapName)
{
	bool bHasDoors = false;
	if(i_szMapName) {
		m_bIsHiscoreMap = false;

		m_pclDlg2->ListSetSelectedElement(ID_LMAPS, i_szMapName);
		S_GUIListEl *pstMap = m_pclDlg2->ListGetSelectedElement(ID_LMAPS);
		if(pstMap) {
			C_MemberAchievements::GetLevelFile(pstMap->szText[0], m_szFilename); //add ".des"
			strcpy(m_szLastMapName, pstMap->szText[0]);
		}

		if(m_pclMiniMapImg) {
			delete m_pclMiniMapImg;
			m_pclMiniMapImg = NULL;
		}
		m_iMinPlayers = 1;
		m_iMaxPlayers = 1;
		m_iNumLaps = 0;
		m_iLevelType  = -1; //assume error
		bool bLoaded = false;
		C_Map *pclMap = new C_Map(m_szFilename, NULL, NULL, &bLoaded);
		if(!bLoaded) {
			delete pclMap;
			return;
		}
		m_iNumLaps = pclMap->GetRaceLaps();
		m_pclMiniMapImg = pclMap->GetMiniMap(1);
		m_iMaxPlayers = pclMap->GetMaxPlayers();
		m_iMinPlayers = pclMap->GetMinPlayers();
		bHasDoors = pclMap->HasDoors();
		m_iLevelType = pclMap->GetType();
		delete pclMap;

		S_LevelElement *pclLevel = m_pclMemberAch->GetLevelInfo(i_szMapName);
		m_bIsHiscoreMap = (pclLevel!=NULL);

		char szString[12];
		if(m_iMinPlayers==m_iMaxPlayers) sprintf(szString, "%d %s", m_iMaxPlayers, m_iMaxPlayers>1 ? "Players" : "Player");
		else sprintf(szString, "%d-%d Players", m_iMinPlayers, m_iMaxPlayers);
		m_pclDlg2->SetText(ID_EMAXPLAYERS, szString);
	}

	{
		//update achievements gui
		char szLevel[48];
		char szScore[48];
		char szExtra[2];
		S_LevelElement *pclLevel = NULL;
		S_GUIListEl stEl;
		float fScore;
		int iPlace, iScore = 0;
		memset(&stEl, 0, sizeof(stEl));
		stEl.pclIcon = m_pclImgAchievement;
		m_pclDlg5->ListRemoveAll(ID_LLIMITS);
		m_pclDlg5->SetText(ID_SSCORE, (char*)"Not a hiscore level");

		strcpy(szLevel, m_szFilename);
		char *szDot = strrchr(szLevel, '.');
		if(szDot) *szDot = 0;
		pclLevel = m_pclMemberAch->GetLevelInfo(szLevel);

		if(pclLevel) {
			iPlace = m_pclMemberAch->GetPlayerAchievements(szLevel, &iScore);
			fScore = iScore/1000.0f;
			memset(szExtra, 0, sizeof(szExtra));
			if(pclLevel->bTime) szExtra[0] = 's';
			if(iPlace!=3) sprintf(szScore, "Your best is %.2f%s", fScore, szExtra);
			else sprintf(szScore, (char*)"This level is not cleared");
			m_pclDlg5->SetText(ID_SSCORE, szScore);

			for(int i=2; i>=0; i--) {
				stEl.iIconRect = i;
				fScore = pclLevel->iLimit[i]/1000.0f;
				sprintf(stEl.szText[0], "%.2f%s", fScore, szExtra);
				m_pclDlg5->ListAddElement(ID_LLIMITS, &stEl);
			}
		}
	}

	if(m_bMaster) {
		//set default to non forced
		if(m_bNoShootingForced) m_bNoShooting = !m_bNoShooting;
		if(m_bFlyThroughForced) m_bFlyThrough = !m_bFlyThrough;
		if(m_bNoBulletHitForced) m_bNoBulletHit = !m_bNoBulletHit;
		m_bNoShootingForced = false;
		m_bFlyThroughForced = false;
		m_bNoBulletHitForced = false;
		//enable options to set
		bool bDogfight = m_iLevelType==MAP_DOGFIGHT;
		bool bRace = m_iLevelType==MAP_RACE;
		if(bDogfight) {
			m_bNoShootingForced = (m_bNoShooting);
			m_bFlyThroughForced = (m_bFlyThrough);
			m_bNoBulletHitForced = (m_bNoBulletHit);
			m_bNoShooting  = false;
			m_bFlyThrough  = false;
			m_bNoBulletHit = false;
		}
		if(bRace) {
			m_bNoBulletHitForced = (!m_bNoBulletHit);
			m_bNoBulletHit = true;
		}
		bool bAIAllowed = (m_iLevelType!=MAP_MISSION) && (m_iLevelType!=MAP_MISSION_COOP) && !bHasDoors;
		if(!bAIAllowed) m_iNumAI = 0;
		m_pclDlg2->SetActive(ID_CNOFIRE,        !bDogfight);
		m_pclDlg2->SetActive(ID_CFLYTHROUGH,    !bDogfight);
		m_pclDlg2->SetActive(ID_CNOBULLETSHIT,  !bDogfight && !bRace);
		m_pclDlg2->SetActive(ID_BADDAI,         bAIAllowed && (m_iNumAI<7));
		m_pclDlg2->SetActive(ID_BSUBAI,         bAIAllowed && (m_iNumAI>0));
		m_pclDlg2->SetChecked(ID_CNOFIRE,       m_bNoShooting);
		m_pclDlg2->SetChecked(ID_CFLYTHROUGH,   m_bFlyThrough);
		m_pclDlg2->SetChecked(ID_CNOBULLETSHIT, m_bNoBulletHit);
	}
}

void C_GameView::SaveLastConnected()
{
	char szFile[MAX_PATH];
	sprintf(szFile, "%s/%s", C_Global::szUserPath, LASTCONNECTED_FILENAME);

	FILE *pFile = fopen(szFile, "wt");
	if(pFile) {
		fprintf(pFile, "*SERVER \"%s\"\n", m_szLastConnected);
		fprintf(pFile, "*PLAYER \"%s\"\n", m_szPlayer);
		//fprintf(pFile, "*GAME   \"%s\"\n", m_szGame);

		fprintf(pFile, "*MEMBER \"%s\"\n", m_szMember);
		fprintf(pFile, "*PW \"%s\"\n", m_szPasswordMD5);

		fclose(pFile);
	}
}

void C_GameView::LoadLastConnected()
{
	char szFile[MAX_PATH];
	char szLine[512], szCommand[128];
	sprintf(szFile, "%s/%s", C_Global::szUserPath, LASTCONNECTED_FILENAME);

	//set default
	//strcpy(m_szLastConnected, "galaxyv2.mooo.com");
	strcpy(m_szLastConnected, "127.0.0.1");
	strcpy(m_szPlayer,        "Incognito");
	//strcpy(m_szGame,          "The game");
	strcpy(m_szMember,        "Member");
	strcpy(m_szPassword,      "Password");
	m_szPasswordMD5[0] = 0;

	FILE *pFile = fopen(szFile, "rt");
	if(pFile) {
		while(C_Global::GetNextLineAndCommand(pFile, szLine, szCommand)) {
			if(strcmp(szCommand, "*SERVER") == 0) {
				C_Global::GetWithin(szLine, '"');
				strcpy(m_szLastConnected, szLine);
				continue;
			}
			if(strcmp(szCommand, "*PLAYER") == 0) {
				C_Global::GetWithin(szLine, '"');
				strcpy(m_szPlayer, szLine);
				continue;
			}
			if(strcmp(szCommand, "*GAME") == 0) {
				C_Global::GetWithin(szLine, '"');
				strcpy(m_szGame, szLine);
				continue;
			}
			if(strcmp(szCommand, "*MEMBER") == 0) {
				C_Global::GetWithin(szLine, '"');
				strcpy(m_szMember, szLine);
				continue;
			}
			if(strcmp(szCommand, "*PW") == 0) {
				C_Global::GetWithin(szLine, '"');
				strcpy(m_szPasswordMD5, szLine);
				strcpy(m_szPassword, "Last Password");
				continue;
			}
		}

		fclose(pFile);
	}
}

int C_GameView::GetIndexForAI(int i_iAI)
{
	int i, iFreePos = 0;
	S_GUIListEl *pstEl;
	for(i=0; i<8; i++) {
		pstEl = m_pclDlg2->ListGetElementById(ID_LPLAYERS, i);
		if(!pstEl) {
			if(iFreePos==i_iAI) return i;
			iFreePos++;
		}
	}
	return -1;
}

void C_GameView::Draw()
{
	C_Font *pclFont = NULL;
	if(!m_bHidden) {
		m_pclDlg0->SetPos(m_pclGraph->GetModeWidth()/2-324/2, m_pclGraph->GetModeHeight()/2-310/2 -15);
		m_pclDlg1->SetPos(m_pclGraph->GetModeWidth()/2-324/2, m_pclGraph->GetModeHeight()/2-310/2 -15);
		m_pclDlg3->SetPos(m_pclGraph->GetModeWidth()/2-492/2, m_pclGraph->GetModeHeight()/2-246/2 -15);
		int iDlg2X = m_pclGraph->GetModeWidth()/2-520/2;
		int iDlg2Y = m_pclGraph->GetModeHeight()/2-396/2 +12;
		m_pclDlg2->SetPos(iDlg2X, iDlg2Y);
		m_pclDlg5->SetPos(iDlg2X+520-180, iDlg2Y-104);

		if(m_iState==GW_DIALOG0 && m_pclDlg0) {
			pclFont = m_pclDlg0->GetFont();
			m_pclDlg0->Draw();
		}
		if(m_iState==GW_DIALOG1 && m_pclDlg1) {
			pclFont = m_pclDlg1->GetFont();
			m_pclDlg1->Draw();
		}
		if(m_iState==GW_DIALOG2 && m_pclDlg2) {
			pclFont = m_pclDlg2->GetFont();
			//temporary add AI players
			int aiIdx[7];
			if(m_iNumAI) {
				S_GUIListEl stEl;
				memset(&stEl, 0, sizeof(stEl));
				stEl.pclIcon = m_pclImgPlayer;

				for(int i=0; i<m_iNumAI; i++) {
					aiIdx[i] = GetIndexForAI(0); //0 since we add a new every time
					if(aiIdx[i]!=-1) {
						stEl.iID = aiIdx[i];
						stEl.iIconRect = aiIdx[i]+8; //AI always ready
						sprintf(stEl.szText[0], "AI %d", i);
						strcpy(stEl.szText[1], m_szAIScore[i]);
						m_pclDlg2->ListAddElement(ID_LPLAYERS, &stEl);
					}
				}
			}
			//draw
			m_pclDlg2->Draw();
			if(m_pclDlg5) {
				char szText[48];
				szText[0] = 0;
				m_pclDlg5->GetText(ID_SSCORE, szText);
				if(strcmp(szText, "Not a hiscore level")!=0) m_pclDlg5->Draw();
			}
			//remove AI players
			if(m_iNumAI) {
				for(int i=0; i<m_iNumAI; i++) {
					if(aiIdx[i]!=-1) m_pclDlg2->ListRemoveElementById(ID_LPLAYERS, aiIdx[i]);
				}
			}

			if(m_pclMiniMapImg) {
				int iW, iH;
				S_Rect stRect = {0,0,100,128}; //maximum size to show
				m_pclMiniMapImg->GetInfo(&iW, &iH, NULL);
				if(iW<stRect.width)  stRect.width  = iW;
				if(iH<stRect.height) stRect.height = iH;

				m_pclGraph->SetBltColor(NULL);
				m_pclGraph->Blt(m_pclMiniMapImg, &stRect, iDlg2X+418, iDlg2Y+144, true);
			}
		}
		if(m_iState==GW_DIALOG3 && m_pclDlg3) {
			pclFont = m_pclDlg3->GetFont();
			m_pclDlg3->Draw();
		}

		if(pclFont) {
			static S_FColor stGrey = {0.6f, 0.6f, 0.6f, 1};
			pclFont->SetColor(stGrey);
			pclFont->Print(8, 6, m_szExternalIP);
		}
		if(pclFont && m_pclMessageTimer->PeekInterval()<7000) {
			static S_FColor stRed = {1, 0, 0, 1};
			pclFont->SetColor(stRed);
			pclFont->Print(8, iDlg2Y+24, m_szTextMessage);
		}
	}
}

void C_GameView::UpdateInput()
{
	if(!m_bHidden) {
		if(m_iState == GW_DIALOG0 && m_pclDlg0) {
			m_pclDlg0->UpdateInput();

			if(m_pclDlg0->WasChanged(ID_EMEMBER))            UpdateDlg0ButtonStatus();
			else if(m_pclDlg0->WasChanged(ID_EPASS))         UpdateDlg0ButtonStatus();

			else if(m_pclDlg0->WasClicked(ID_BLOGIN))        HandleMemberLogin(true);
			else if(m_pclDlg0->WasClicked(ID_BCANCEL))       HandleMemberLogin(false);

			else if(m_pclDlg0->WasClicked(ID_BSTARTWEB))     HandleStartBrowser();
		}
		if(m_iState == GW_DIALOG1 && m_pclDlg1) {
			m_pclDlg1->UpdateInput();

			if(m_pclDlg1->WasChanged(ID_EIP))                UpdateDlg1ButtonStatus();
			else if(m_pclDlg1->WasChanged(ID_EPLAYER))       UpdateDlg1ButtonStatus();
			else if(m_pclDlg1->WasChanged(ID_EGAME))         UpdateDlg1ButtonStatus();
			else if(m_pclDlg1->WasChanged(ID_LGAMES))        UpdateDlg1ButtonStatus();

			else if(m_pclDlg1->WasClicked(ID_BCONNECT))      HandleConnect();
			else if(m_pclDlg1->WasClicked(ID_EIP))           HandleConnect();

			else if(m_pclDlg1->WasClicked(ID_BJOIN))         HandleJoin();
			else if(m_pclDlg1->WasClicked(ID_BNEW))          HandleNew();

			else if(m_pclDlg1->WasClicked(ID_BREPLAY))       HandleReplay();

			else if(m_pclDlg1->WasClicked(ID_BWEBSCORE))     HandleStartBrowser2();
		}
		if(m_iState==GW_DIALOG2 && m_pclDlg2) {
			m_pclDlg2->UpdateInput();

			if(m_pclDlg2->WasChanged(ID_LMAPS))              HandleSelectMap();

			else if(m_pclDlg2->WasClicked(ID_BBACK))         HandleBack();
			else if(m_pclDlg2->WasClicked(ID_BSTART))        HandleStart();

			else if(m_pclDlg2->WasChanged(ID_ECHAT))         UpdateDlg2ChatStatus();
			else if(m_pclDlg2->WasClicked(ID_BSEND))         HandleChatSend();
			else if(m_pclDlg2->WasClicked(ID_ECHAT))         HandleChatSend();

			else if(m_pclDlg2->WasChanged(ID_CNOFIRE))       UpdateDlg2CheckStatus();
			else if(m_pclDlg2->WasChanged(ID_CFLYTHROUGH))   UpdateDlg2CheckStatus();
			else if(m_pclDlg2->WasChanged(ID_CNOBULLETSHIT)) UpdateDlg2CheckStatus();

			else if(m_pclDlg2->ListWasDeletePressed(ID_LPLAYERS)) HandleKickPlayer();

			else if(m_pclDlg2->WasClicked(ID_BSAVEREPLAY))   HandleSave();

			else if(m_pclDlg2->WasClicked(ID_BADDAI))        HandleNumAI(1);
			else if(m_pclDlg2->WasClicked(ID_BSUBAI))        HandleNumAI(-1);
		}
		if(m_iState==GW_DIALOG3 && m_pclDlg3) {
			m_pclDlg3->UpdateInput();

			if(m_pclDlg3->WasClicked(ID_BRBACK))             HandleRBack();
			else if(m_pclDlg3->WasClicked(ID_BRPLAY))        HandleRPlay();
		}
	}
}

void C_GameView::MouseMove(int x, int y)
{
	if(!m_bHidden) {
		if(m_iState==GW_DIALOG0 && m_pclDlg0) m_pclDlg0->MouseMove(x, y);
		if(m_iState==GW_DIALOG1 && m_pclDlg1) m_pclDlg1->MouseMove(x, y);
		if(m_iState==GW_DIALOG2 && m_pclDlg2) m_pclDlg2->MouseMove(x, y);
		if(m_iState==GW_DIALOG3 && m_pclDlg3) m_pclDlg3->MouseMove(x, y);
	}
}

void C_GameView::MouseDown(int x, int y)
{
	if(!m_bHidden) {
		if(m_iState==GW_DIALOG0 && m_pclDlg0) m_pclDlg0->MouseDown(x, y);
		if(m_iState==GW_DIALOG1 && m_pclDlg1) m_pclDlg1->MouseDown(x, y);
		if(m_iState==GW_DIALOG2 && m_pclDlg2) m_pclDlg2->MouseDown(x, y);
		if(m_iState==GW_DIALOG3 && m_pclDlg3) m_pclDlg3->MouseDown(x, y);
	}
}

void C_GameView::MouseUp(int x, int y)
{
	if(!m_bHidden) {
		if(m_iState==GW_DIALOG0 && m_pclDlg0) m_pclDlg0->MouseUp(x, y);
		if(m_iState==GW_DIALOG1 && m_pclDlg1) m_pclDlg1->MouseUp(x, y);
		if(m_iState==GW_DIALOG2 && m_pclDlg2) m_pclDlg2->MouseUp(x, y);
		if(m_iState==GW_DIALOG3 && m_pclDlg3) m_pclDlg3->MouseUp(x, y);
	}
}

void C_GameView::MouseWheel(int way)
{
	if(!m_bHidden) {
		if(m_iState==GW_DIALOG0 && m_pclDlg0) m_pclDlg0->MouseWheel(way);
		if(m_iState==GW_DIALOG1 && m_pclDlg1) m_pclDlg1->MouseWheel(way);
		if(m_iState==GW_DIALOG2 && m_pclDlg2) m_pclDlg2->MouseWheel(way);
		if(m_iState==GW_DIALOG3 && m_pclDlg3) m_pclDlg3->MouseWheel(way);
	}
}

bool C_GameView::IsAnyEditFocused()
{
	if(!m_bHidden) {
		if(m_iState==GW_DIALOG0 && m_pclDlg0) return m_pclDlg0->IsAnyEditFocused();
		if(m_iState==GW_DIALOG1 && m_pclDlg1) return m_pclDlg1->IsAnyEditFocused();
		if(m_iState==GW_DIALOG2 && m_pclDlg2) return m_pclDlg2->IsAnyEditFocused();
		if(m_iState==GW_DIALOG3 && m_pclDlg3) return m_pclDlg3->IsAnyEditFocused();
	}
	return false;
}

void C_GameView::DisplayShortMessage(char *i_szMessage)
{
	strncpy(m_szTextMessage, i_szMessage, sizeof(m_szTextMessage));
	m_szTextMessage[sizeof(m_szTextMessage)-1] = 0;
	m_pclMessageTimer->Reset();
}
