// C_GameView

#ifndef GAMEVIEW_H
#define GAMEVIEW_H

#include "global_hiscore.h"
#include "common.h"
#include "s_hiscore.h"
#include "timer.h"
#include "tilemap.h"
#include "gui.h"


#define D0_MLOGIN     0x4000
#define D0_MCREATE    0x8000
#define D1_MWEB       0x2000
#define D1_CONNECT    0x0001
#define D1_NEW        0x0002
#define D1_JOIN       0x0004
#define D2_LEAVE      0x0008
#define D2_OPTCHANGE  0x0010
//#define D2_READY      0x0020
#define D2_START      0x0040
#define D2_CHATSEND   0x0080
#define D2_KICK       0x0100
#define D2_SAVEREPLAY 0x0200
#define D3_BEGIN      0x0400
#define D3_REPLAY     0x0800

class C_GameView
{
public:
	C_GameView(C_GraphWrapperGL *i_pclGraph, C_MemberAchievements *i_pclMemberAch);
	~C_GameView();

	void Show(bool i_bShow);

	int Poll();
	void Draw();
	void UpdateInput();
	void MouseMove(int x, int y);
	void MouseDown(int x, int y);
	void MouseUp(int x, int y);
	void MouseWheel(int way);

	//dialog0 data
	//input
	void SetLogedInState(int i_iMemberLogedIn);
	//output (valid after poll() msgs)
	void GetUserPass(char *o_szUser, char *o_szPass);

	//dialog1 data
	//input
	void SetExternalIP(char *i_szExternalIP);
	void AddGame(char *i_szName, bool i_bPlaying);
	void RemoveGame(char *i_szName);
	void SetGameStatus(char *i_szName, bool i_bPlaying);
	//output (valid after poll() msgs)
	void GetConnectString(char *o_szConnect);
	void GetPlayerName(char *o_szPlayer);               //when in dialog 1 (and 2)
	void GetPlayerName(int i_iIndex, char *o_szPlayer); //when in dialog 2
	void GetGameName(char *o_szGame);

	//dialog2 data
	//input
	void SetGameOpt(char *i_szFilename, int i_iNumLaps, bool i_bNoShooting, bool i_bFlyThrough, bool i_bNoBulletHit, int i_iLevelType, int i_iNumAI);
	void AddChatString(char *i_szChatString);
	void AddPlayer(char *i_szName, int i_iIndex);
	void RemovePlayer(int i_iIndex);
	void SetPlayerScore(int i_iIndex, int i_iScore, bool i_bAI);
	void UpdateAchievements();
	//output (valid after poll() msgs)
	bool GetGameOpt(char *o_szFilename, int *o_iNumLaps, bool *o_bNoShooting, bool *o_bFlyThrough, bool *o_bNoBulletHit, int *o_iLevelType, int *o_iNumAI);
	bool GetHiscoreMapName(char *o_szLevel);
	void GetChatString(char *o_szChatString);
	int GetPlayerToKick() {return m_iKickIndex;};
	void ActivateSaveReplay(bool i_bActivate);
	int GetIndexForAI(int i_iAI);

	//dialog3 data
	//input
	void AddExternalReplay(char *i_szLevel, char *i_szName, int i_iPlace, int i_iScoreMs);
	//output (valid after poll() msgs)
	int GetReplayFile(char *o_szFilename, char *o_szName); //0 file, 1..3 place on external on web server

	//info
	bool IsAnyEditFocused();
	
	//update state
	void SetDlg1State();                 //-> dialog1
	void SetConnected(int i_iConnected, bool i_bClearLists = false); //0 not connected, 1 connected, -1 waiting
	int  GetConnected() {return m_iConnected;};
	void SetJoinNewOK(int i_iIndex);     //dialog1 -> dialog2

	void NewMaster();

	void DisplayShortMessage(char *i_szMessage);

	int NumPlayers(bool i_bWithAI);
private:
	bool ReInitDialogs();

	//handle gui actions
	void HandleConnect();
	void HandleExit();
	void HandleNew();
	void HandleJoin();
	void HandleStart();
	void HandleBack();
	void HandleChatSend();
	void HandleSelectMap();
	void HandleKickPlayer();
	void HandleReplay();
	void HandleSave();
	void HandleRBack();
	void HandleRPlay();
	void HandleNumAI(int i_iAdd);
	void HandleMemberLogin(bool i_bLogin);
	void HandleStartBrowser();
	void HandleStartBrowser2();

	//gui state
	void UpdateDlg0ButtonStatus();
	void UpdateDlg1ButtonStatus();
	void UpdateDlg2CheckStatus();
	void UpdateDlg2ChatStatus();
	void UpdateStartState();

	void SetMapName(char *i_szMapName);
	void ClearDlg2();
	void SetAchievements();

	int  m_iState;
	bool m_bHidden;
	int  m_iPollMsg;

	bool m_bMaster;
	int  m_iIndex; //this player id

	char m_szMember[MAXPLAYERNAMELENGTH];
	char m_szPassword[MAXPASSWORDLENGTH];
	char m_szPasswordMD5[MAXPASSWORDLENGTH];
	int m_iLogedIn;

	char m_szExternalIP[40];
	char m_szConnect[48];
	char m_szPlayer[MAXPLAYERNAMELENGTH];
	char m_szGame[MAXGAMENAMELENGTH];
	int  m_iConnected;
	int  m_iGameIndex;
	char m_szAIScore[8][12];

	char m_szChatMsg[2048];

	int  m_iNumLaps;
	int  m_iLevelType;

	bool m_bNoShootingForced;
	bool m_bFlyThroughForced;
	bool m_bNoBulletHitForced;
	bool m_bNoShooting;
	bool m_bFlyThrough;
	bool m_bNoBulletHit;
	int  m_iNumAI;

	char m_szFilename[128];
	int m_iMaxPlayers, m_iMinPlayers;

	bool m_bIsHiscoreMap;

	C_Image *m_pclImgGame;
	C_Image *m_pclImgPlayer;
	C_Image *m_pclImgAchievement;

	int m_iKickIndex;

	//last connected
	char m_szLastConnected[32];
	void SaveLastConnected();
	void LoadLastConnected();

	//last map name
	char m_szLastMapName[128];

	//minimap
	C_Image *m_pclMiniMapImg;
	C_GraphWrapperGL *m_pclGraph;

	C_GUI *m_pclDlg0;
	C_GUI *m_pclDlg1;
	C_GUI *m_pclDlg2;
	C_GUI *m_pclDlg3;
	C_GUI *m_pclDlg5;

	C_MemberAchievements *m_pclMemberAch;

	C_Timer *m_pclMessageTimer;
	char m_szTextMessage[256];
};

#endif
