
// C_MemberAchievements

//class to deal with the web server.
// - handles member achievements
//  member verify, test/send score for hiscore levels
//  and so on
// - handles hiscore and replay functionality
// - replays are encoded to base64 on the web server
// - gets info such as client external ip and latest
//  game version from the web server
// 

#include "global_hiscore.h"
#include "global.h"
#include "timer.h"

#include <ctype.h>

#define BUFFER_SIZE 16384
//#define WEB_HOST "galaxyv2.sourceforge.net"
#define WEB_HOST "galaxy-forces.com"


#define NUM_LEVELS (/*4+*/22+23+4)
S_LevelElement g_stLevels[NUM_LEVELS] = {
//	{"dogfight0",false,false,{0,0,0}},
//	{"dogfight1",false,false,{0,0,0}},
//	{"dogfight2",false,false,{0,0,0}},
//	{"dogfight3",false,false,{0,0,0}},

	{"race00_3pl",true,-1,{35000,41000,300000}},
	{"race01_3pl",true,-1,{61000,72000,300000}},
	{"race02_4pl",true,-1,{47000,57000,300000}},
	{"race03_2pl",true,-1,{52000,61000,300000}},
	{"race04_2pl",true,-1,{45000,56000,300000}},
	{"race05_4pl",true,-1,{53000,64000,300000}},
	{"race06_4pl",true,-1,{52000,63000,300000}},
	{"race07_2pl",true,-1,{92000,103000,300000}},
	{"race08_8pl",true,-1,{64000,72000,300000}},
	{"race09_4pl",true,-1,{90000,112000,300000}},
	{"race10_4pl",true,-1,{99000,113000,300000}},
	{"race11_4pl",true,-1,{51000,58000,300000}},
	{"race12_4pl",true,-1,{88000,110000,300000}},
	{"race13_3pl",true,-1,{58000,67000,300000}},
	{"race14_2pl",true,-1,{69000,82000,300000}},
	{"race15_4pl",true,-1,{145000,158000,300000}},
	{"race16_3pl",true,-1,{85000,97000,300000}},
	{"race17_3pl",true,-1,{89000,100000,300000}},
	{"race18_4pl",true,-1,{27000,32000,300000}},
	{"race00_zero",true,-1,{39000,52000,300000}},
	{"race04_jupiter",true,-1,{57500,69000,300000}},
	{"race10_mod",true,-1,{135000,150000,300000}},

	{"mission00",false,-1,{282000,230000,1000}},
	{"mission01",false,-1,{302000,250000,1000}},
	{"mission02",false,-1,{345000,300000,1000}},
	{"mission03",false,-1,{355000,318000,1000}},
	{"mission04",false,-1,{388000,330000,1000}},
	{"mission05",false,-1,{394000,340000,1000}},
	{"mission06",false,-1,{371000,325000,1000}},
	{"mission07",false,-1,{364000,330000,1000}},
	{"mission08",false,-1,{300000,270000,1000}},
	{"mission09",false,-1,{347000,320000,1000}},
	{"mission10",false,-1,{283000,260000,1000}},
	{"mission11",false,-1,{462000,390000,1000}},
	{"mission12",false,-1,{510000,380000,1000}},
	{"mission13",false,-1,{520000,400000,1000}},
	{"mission14",false,-1,{380000,290000,1000}},
	{"mission15",false,-1,{500000,380000,1000}},
	{"mission16",false,-1,{270000,150000,1000}},
	{"mission17",false,-1,{330000,250000,1000}},
	{"mission01_mod",false,-1,{310000,160000,1000}},
	{"mission04_mod",false,-1,{385000,360000,1000}},
	{"mission06_mod",false,-1,{360000,260000,1000}},
	{"mission_narrow",false,-1,{150000,90000,1000}},
	{"mission_hard",false,-1,{220000,80000,1000}},

	{"mission_coop0",false,-1,{300000,150000,1000}},
	{"mission_coop1",false,-1,{330000,160000,1000}},
	{"mission_coop2",false,-1,{350000,180000,1000}},
	{"mission_coop3",false,-1,{280000,140000,1000}},
};

C_MemberAchievements::C_MemberAchievements()
{
	m_szBuffer = new char[BUFFER_SIZE];

	m_pclAchievements = NULL;
	m_iCurrent = 0;
	m_bLoginOk = false;
}

C_MemberAchievements::~C_MemberAchievements()
{
	delete[] m_szBuffer;

	delete m_pclAchievements;
}

int C_MemberAchievements::ReceiveAllAndCheckOK(C_Socket *i_pclSocket, int i_iTimeout)
{
	int iBufSize = 15;
	C_Timer clTimer;
	while(i_pclSocket->ConnectedToIP() && clTimer.PeekInterval()<i_iTimeout) i_pclSocket->WaitForReadEvent(10000); //connection is terminated by server when all sent
	i_pclSocket->GetRecvBuffer(m_szBuffer, &iBufSize, 0);
	m_szBuffer[iBufSize]=0;
	if(strncmp(m_szBuffer, "HTTP/1.1 200 OK", 15)==0) {
		if(!i_pclSocket->ConnectedToIP()) return i_pclSocket->RecvBufferSize();
	}
	return 0;
}

bool C_MemberAchievements::GetInfo(char *o_szExternalIP, int *o_iLatestVersionMajor, int *o_iLatestVersionMinor, bool *o_bPortOk)
{
	if(!o_szExternalIP || !o_iLatestVersionMajor || !o_iLatestVersionMinor || !o_bPortOk) return false;
	bool bResult = false;
	C_Socket *pclSocket = new C_Socket();
	if(pclSocket->Connect((char*)WEB_HOST, 80, 5000)) {
		//send get command
		sprintf(m_szBuffer, "GET /getinfo.php HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n", (char*)WEB_HOST);
		pclSocket->SendQueued(m_szBuffer, (int)strlen(m_szBuffer));
		//get result
		int iBufSize = ReceiveAllAndCheckOK(pclSocket, 3000);
		if(iBufSize>BUFFER_SIZE) iBufSize = BUFFER_SIZE;
		pclSocket->GetRecvBuffer(m_szBuffer, &iBufSize, 0);
		m_szBuffer[iBufSize]=0;
		if(iBufSize>0) {
			//parse result
			char *szMsg;
			if((szMsg = strstr(m_szBuffer, "\r\n\r\n")) != NULL) {
				szMsg+=4;
				int iPortOk = 0;
				int iResult = sscanf(szMsg, "%15s %d.%d %d", o_szExternalIP, o_iLatestVersionMajor, o_iLatestVersionMinor, &iPortOk);
				if(iResult==4) bResult = true;
				*o_bPortOk = iPortOk!=0;
			}
		}
	}
	delete pclSocket;
	return bResult;
}

bool C_MemberAchievements::LoginCheck(char *i_szUser, char *i_szPass)
{
	m_bLoginOk = false;

	if(!i_szUser || !i_szPass) return false;
	strcpy(m_szUser, i_szUser);
	strcpy(m_szPassHash, i_szPass); //i_szPass is now MD5-Hash set to avoid storing the password as plain text.
//	if( !C_HttpUtil::md5chksum_encode(i_szPass, m_szPassHash, 33)) return false;

	C_Socket *pclSocket = new C_Socket();
	if(pclSocket->Connect((char*)WEB_HOST, 80, 5000)) {
		//send get command
		sprintf(m_szBuffer, "GET /member_qualify.php?User=%s&Password=%s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n", m_szUser, m_szPassHash, (char*)WEB_HOST);
		pclSocket->SendQueued(m_szBuffer, (int)strlen(m_szBuffer));
		//get result
		int iBufSize = ReceiveAllAndCheckOK(pclSocket, 3000);
		if(iBufSize>BUFFER_SIZE) iBufSize = BUFFER_SIZE;
		pclSocket->GetRecvBuffer(m_szBuffer, &iBufSize, 0);
		m_szBuffer[iBufSize]=0;
		if(iBufSize>0) {
			//parse result
			char *szMsg;
			if((szMsg = strstr(m_szBuffer, "\r\n\r\n")) != NULL) {
				szMsg+=4;
				int iOk = 0;
				int iResult = sscanf(szMsg, "%d", &iOk);
				if(iResult==1 && iOk!=0) m_bLoginOk = true;
			}
		}
	}
	delete pclSocket;
	return m_bLoginOk;
}

C_LevelInfoList *C_MemberAchievements::FetchLevelInfoList()
{
	char szLine[512], szTmp[512];
	C_LevelInfoList *pclLevelInfo = NULL;
	C_Socket *pclSocket = new C_Socket();
	if(pclSocket->Connect((char*)WEB_HOST, 80, 5000)) {
		//send get command
		sprintf(m_szBuffer, "GET /achievements_get.php?User=%s&Password=%s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n", m_szUser, m_szPassHash, (char*)WEB_HOST);
		pclSocket->SendQueued(m_szBuffer, (int)strlen(m_szBuffer));
		//get result
		int iBufSize = ReceiveAllAndCheckOK(pclSocket, 3000);
		if(iBufSize>BUFFER_SIZE) iBufSize = BUFFER_SIZE;
		pclSocket->GetRecvBuffer(m_szBuffer, &iBufSize, 0);
		m_szBuffer[iBufSize]=0;
		if(iBufSize>1) {
			//parse result
			char *szMsg;
			if((szMsg = strstr(m_szBuffer, "\r\n\r\n")) != NULL) {
				int iResult, iLineCharIndex = 0;
				S_LevelElement stEl;
				pclLevelInfo = new C_LevelInfoList();
				while(szMsg[1]=='\r' || szMsg[1]=='\n') szMsg++;
				do {
					iResult = 0;
					if(C_Global::GetNextLineAndCommand(szMsg+1, szLine, NULL, &iLineCharIndex)) {
						strcpy(szTmp, szLine);
						if(C_Global::GetWithin(szTmp, '"')) {
							strcpy(stEl.szLevel, szTmp);
							iResult++;
							char *szAfter = C_Global::GetPositionAfter(szLine, '"', 2);
							int iTime;
							iResult += sscanf(szAfter, "%d %d %d %d %d", &iTime, &stEl.iLimit[0], &stEl.iLimit[1], &stEl.iLimit[2], &stEl.iScoreMs);
							stEl.bTime = iTime != 0;
						}
					}
					if(iResult==6) pclLevelInfo->push_back(stEl);
				} while((szMsg = strstr(szMsg+1, "\n")) && iResult==6);
				if(pclLevelInfo->empty()) {
					delete pclLevelInfo;
					pclLevelInfo = NULL;
				}
			}
		}
	}
	delete pclSocket;
	delete m_pclAchievements;
	m_pclAchievements = pclLevelInfo;
	return pclLevelInfo;
}

void C_MemberAchievements::RestartLevelInfo()
{
	m_iCurrent = 0;
}

S_LevelElement *C_MemberAchievements::GetNextLevelInfo()
{
	if(m_pclAchievements && m_iCurrent < (int)m_pclAchievements->size()) {
		return &m_pclAchievements->at(m_iCurrent++);
	} else if(!m_pclAchievements && m_iCurrent<NUM_LEVELS) {
		return &g_stLevels[m_iCurrent++];
	} else {
		m_iCurrent = 0;
		return NULL;
	}
}

S_LevelElement *C_MemberAchievements::GetLevelInfo(char *i_szLevel)
{
	if(!m_pclAchievements) return NULL;

	int i;
	for(i=0; i<(int)m_pclAchievements->size(); i++) {
		S_LevelElement *pclLevel = &m_pclAchievements->at(i);
		if(strcmp(i_szLevel, pclLevel->szLevel)==0)
			return pclLevel;
	}
	return NULL;
}

void C_MemberAchievements::GetLevelFile(char *i_szLevel, char *o_szLevel)
{
	strcpy(o_szLevel, i_szLevel);
	strcat(o_szLevel, ".des");
}

int C_MemberAchievements::GetPlayerAchievements(char *i_szLevel, int *o_iScoreMs)
{
	S_LevelElement *pstLevelInfo = GetLevelInfo(i_szLevel);

	if(pstLevelInfo) {
		//convert pstAchieve->iScore to gold/silver/bronze
		if(o_iScoreMs) *o_iScoreMs = pstLevelInfo->iScoreMs;
		for(int j=0; j<3; j++) {
			if((!pstLevelInfo->bTime && pstLevelInfo->iScoreMs > pstLevelInfo->iLimit[j])
				|| (pstLevelInfo->bTime && pstLevelInfo->iScoreMs < pstLevelInfo->iLimit[j] && pstLevelInfo->iScoreMs>0)) //no negative score on race
			{
				return j;
			}
		}
	}
	return 3; //no_score
}

bool C_MemberAchievements::CheckHiscore(char *i_szLevel, int i_iScoreMs)
{
	bool bHiscore = false;
	S_LevelElement *pstLevelInfo = GetLevelInfo(i_szLevel);

	if(pstLevelInfo) {
		//check current score against best score (pstAchieve->iScore)
		//and it must be better than bronze score to qualify
		if (!pstLevelInfo->bTime && i_iScoreMs > pstLevelInfo->iLimit[2]) {
			if (i_iScoreMs>pstLevelInfo->iScoreMs) bHiscore = true;
		} else if(pstLevelInfo->bTime && i_iScoreMs < pstLevelInfo->iLimit[2] && i_iScoreMs>0) { //no negative score on race
			if (pstLevelInfo->iScoreMs==-1 || i_iScoreMs<pstLevelInfo->iScoreMs) bHiscore = true;
		}
		if (bHiscore) {
			pstLevelInfo->iScoreMs = i_iScoreMs;
		}
	}
	return bHiscore;
}

bool C_MemberAchievements::SendHiscore(char *i_szLevel, int i_iScoreMs, C_Replay *i_pclReplay)
{
	bool bResult = false;
	char szSmallBuffer[1024];
	int iReplaySize = 512*1024;
	char *pReplay;//, szNameEnc[128];
	C_Socket *pclSocket = new C_Socket();
	i_pclReplay->SaveToMem(NULL, &iReplaySize);
//	C_HttpUtil::url_encode(i_szName, szNameEnc, sizeof(szNameEnc));
	pReplay = new char[iReplaySize];
	i_pclReplay->SaveToMem(pReplay, &iReplaySize);
	char *pReplayBase64 = (char*)C_HttpUtil::base64_encode((unsigned char*)pReplay, iReplaySize, &iReplaySize);
	delete[] pReplay;
	if(pclSocket->Connect((char*)WEB_HOST, 80, 5000)) {
		//send post command
		sprintf(m_szBuffer, "POST /achievements_post.php HTTP/1.1\r\nHost: %s\r\nConnection: close\r\nContent-Length: ", (char*)WEB_HOST);
		sprintf(szSmallBuffer, "LEVEL=%s&NAME=%s&PASSWORD=%s&SCORE=%d&REPLAY=", i_szLevel, m_szUser, m_szPassHash, i_iScoreMs);
		pclSocket->SendQueued(m_szBuffer, (int)strlen(m_szBuffer));
		sprintf(m_szBuffer, "%d\r\nContent-Type: application/x-www-form-urlencoded\r\n\r\n", (int)strlen(szSmallBuffer)+iReplaySize);
		pclSocket->SendQueued(m_szBuffer, (int)strlen(m_szBuffer));
		pclSocket->SendQueued(szSmallBuffer, (int)strlen(szSmallBuffer));
		pclSocket->SendQueued(pReplayBase64, iReplaySize);
		//basic result check
		int iBufSize = ReceiveAllAndCheckOK(pclSocket, 6000);
		bResult = (iBufSize>0);
	}
	delete[] pReplayBase64;
	delete pclSocket;
	return bResult;
}

/////////////////////////////////////////////////////////////
C_Hiscores *C_MemberAchievements::FetchHiscores()
{
	char szLine[512], szTmp[512];
	C_Hiscores *pclHiscores = NULL;
	C_Socket *pclSocket = new C_Socket();
	if(pclSocket->Connect((char*)WEB_HOST, 80, 5000)) {
		//send get command
		if (m_bLoginOk) sprintf(m_szBuffer, "GET /hiscore_get2.php?Name=%s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n", m_szUser, (char*)WEB_HOST);
		else sprintf(m_szBuffer, "GET /hiscore_get2.php HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n", (char*)WEB_HOST);
		pclSocket->SendQueued(m_szBuffer, (int)strlen(m_szBuffer));
		//get result
		int iBufSize = ReceiveAllAndCheckOK(pclSocket, 15000);
		if(iBufSize>BUFFER_SIZE) iBufSize = BUFFER_SIZE;
		pclSocket->GetRecvBuffer(m_szBuffer, &iBufSize, 0);
		m_szBuffer[iBufSize]=0;
		if(iBufSize>1) {
			//parse result
			char *szMsg;
			if((szMsg = strstr(m_szBuffer, "\r\n\r\n")) != NULL) {
				int iResult, iLineCharIndex = 0;
				S_HiscoreElement stEl;
				pclHiscores = new C_Hiscores();
				while(szMsg[1]=='\r' || szMsg[1]=='\n') szMsg++;
				do {
					iResult = 0;
					if(C_Global::GetNextLineAndCommand(szMsg+1, szLine, NULL, &iLineCharIndex)) {
						strcpy(szTmp, szLine);
						if(C_Global::GetWithin(szTmp, '"')) {
							strcpy(stEl.szLevel, szTmp);
							iResult++;
							char *szAfter = C_Global::GetPositionAfter(szLine, '"', 2);
							iResult += sscanf(szAfter, "%d", &stEl.iPlace);

							strcpy(szTmp, szAfter);
							if(C_Global::GetWithin(szTmp, '"')) {
								strcpy(stEl.szName, szTmp);
								iResult++;
								szAfter = C_Global::GetPositionAfter(szLine, '"', 4);
								iResult += sscanf(szAfter, "%d", &stEl.iScoreMs);
							}
						}
					}
					if(iResult==4) pclHiscores->push_back(stEl);
				} while((szMsg = strstr(szMsg+1, "\n")) && iResult==4);
				if(pclHiscores->empty()) {
					delete pclHiscores;
					pclHiscores = NULL;
				}
			}
		}
	}
	delete pclSocket;
	return pclHiscores;
}

C_Replay *C_MemberAchievements::FetchReplay(char *i_szLevel, char *i_szName)
{
	C_Replay *pclReplay = NULL;
	C_Socket *pclSocket = new C_Socket();
	if(pclSocket->Connect((char*)WEB_HOST, 80, 5000)) {
		//send get command
		sprintf(m_szBuffer, "GET /hiscore_getreplay2.php?Level=%s&Name=%s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n\r\n", i_szLevel, i_szName, (char*)WEB_HOST);
		pclSocket->SendQueued(m_szBuffer, (int)strlen(m_szBuffer));
		//get result
		int iBufSize = ReceiveAllAndCheckOK(pclSocket, 15000);
		char *pBuffer = new char[iBufSize+1];
		pclSocket->GetRecvBuffer(pBuffer, &iBufSize, 0);
		pBuffer[iBufSize]=0;
		if(iBufSize>0) {
			//parse result
			char *szMsg, *szMsgR;
			//find end
			while (iBufSize>2 && pBuffer[iBufSize - 1] != '=') {
				iBufSize--;
				pBuffer[iBufSize] = 0;
			}
			//find start
			szMsg = strrchr(pBuffer, '\n');
			szMsgR = strrchr(pBuffer, '\r');
			if(szMsgR > szMsg) szMsg = szMsgR;
			if(szMsg != NULL) {
				szMsg++; //next char is our start
				int iDecSize, iSize = iBufSize-(int)(szMsg-pBuffer);
				if(iSize>0) {
					char *pDecData = (char*)C_HttpUtil::base64_decode((unsigned char*)szMsg, iSize, &iDecSize);
					if(pDecData) {
						pclReplay = new C_Replay();
						if(!pclReplay->LoadFromMem(pDecData, iDecSize)) {
							delete pclReplay;
							pclReplay = NULL;
						}
						delete[] pDecData;
					}
				}
			}
		}
		delete[] pBuffer;
	}
	delete pclSocket;
	return pclReplay;
}
