// C_GlobalHiscore, C_MemberAchievements

#ifndef GLOBAL_HISCORE_H
#define GLOBAL_HISCORE_H

#include "socket_2.h"
#include "replay.h"

#include "http_util.h"

struct S_HiscoreElement
{
	char szLevel[32];
	char szName[20];
	int iScoreMs; //score times 1000 or ms for time score
	int iPlace;
};
typedef std::vector<S_HiscoreElement> C_Hiscores;

struct S_LevelElement
{
	char szLevel[32];
	bool bTime;
	//score times 1000 or ms for time score
	int iScoreMs; //best score for this member
	int iLimit[3]; //score limits for gold, silver, bronz
};
typedef std::vector<S_LevelElement> C_LevelInfoList;

class C_MemberAchievements
{
public:
	C_MemberAchievements();
	~C_MemberAchievements();

	//keeps user/pass
	bool LoginCheck(char *i_szUser, char *i_szPass);

	//demands m_bLoginOk
	//level info
	C_LevelInfoList *FetchLevelInfoList(); //for all hiscore levels: score limits, and best score for this member
	void RestartLevelInfo();
	S_LevelElement *GetNextLevelInfo();
	S_LevelElement *GetLevelInfo(char *i_szLevel);
	static void GetLevelFile(char *i_szLevel, char *o_szLevel); //adds ".des" at the end
	//player achievements
	int GetPlayerAchievements(char *i_szLevel, int *o_iScoreMs);
	bool CheckHiscore(char *i_szLevel, int i_iScoreMs);
	bool SendHiscore(char *i_szLevel, int i_iScoreMs, C_Replay *i_pclReplay);
	//get replay (best for this player for the given level)
	C_Replay *FetchReplay(char *i_szLevel, char *i_szName=NULL); //set name=NULL for best place on level
	C_Hiscores *FetchHiscores();
	

	//get global info (has nothing to do with logged in player)
	bool GetInfo(char *o_szExternalIP, int *o_iLatestVersionMajor, int *o_iLatestVersionMinor, bool *o_bPortOk);

private:
	char *m_szBuffer;
	
	int ReceiveAllAndCheckOK(C_Socket *i_pclSocket, int i_iTimeout); //returns amount after HTTP/OK, or 0

	char m_szUser[24], m_szPassHash[33];
	bool m_bLoginOk;

	C_LevelInfoList *m_pclAchievements;
	int m_iCurrent;
};

#endif
