// C_HttpUtil

#include "http_util.h"
#include "md5.h"
#include <ctype.h>

static const char base64_chars[] = 
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static inline bool is_base64(unsigned char c)
{
	return (isalnum(c) || (c == '+') || (c == '/'));
}

unsigned char *C_HttpUtil::base64_encode(unsigned char *i_pData, int i_iLen, int *o_iLen)
{
	int i = 0;
	int j = 0;
	unsigned char char_array_3[3];
	unsigned char char_array_4[4];
	int out_len = 0;
	unsigned char *ret = new unsigned char[(i_iLen/3)*4+4];

	while(i_iLen--) {
		char_array_3[i++] = *(i_pData++);
		if(i == 3) {
			char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
			char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
			char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
			char_array_4[3] = char_array_3[2] & 0x3f;

			for(i = 0; (i <4) ; i++) {
				ret[out_len++] = base64_chars[char_array_4[i]];
			}
			i = 0;
		}
	}

	if(i) {
		for(j = i; j < 3; j++) {
			char_array_3[j] = '\0';
		}

		char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
		char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
		char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
		char_array_4[3] = char_array_3[2] & 0x3f;

		for(j = 0; (j < i + 1); j++) {
			ret[out_len++] = base64_chars[char_array_4[j]];
		}

		while((i++ < 3)) {
			ret[out_len++] = '=';
		}
	}

	*o_iLen = out_len;
	return ret;
}

unsigned char *C_HttpUtil::base64_decode(unsigned char *i_pData, int i_iLen, int *o_iLen)
{
	int i = 0;
	int j = 0;
	int in_ = 0;
	unsigned char char_array_4[4], char_array_3[3];
	int out_len = 0;
	unsigned char *ret = new unsigned char[(i_iLen/4)*3+3];

	while(i_iLen-- && (i_pData[in_] != '=') && is_base64(i_pData[in_])) {
		char_array_4[i++] = i_pData[in_]; in_++;
		if(i ==4) {
			for(i = 0; i <4; i++) {
				char_array_4[i] = (unsigned char)(strchr(base64_chars, char_array_4[i]) - base64_chars);
			}

			char_array_3[0] = (char_array_4[0] << 2) + ((char_array_4[1] & 0x30) >> 4);
			char_array_3[1] = ((char_array_4[1] & 0xf) << 4) + ((char_array_4[2] & 0x3c) >> 2);
			char_array_3[2] = ((char_array_4[2] & 0x3) << 6) + char_array_4[3];

			for(i = 0; (i < 3); i++) {
				ret[out_len++] = char_array_3[i];
			}
			i = 0;
		}
	}

	if(i) {
		for(j = i; j <4; j++) {
			char_array_4[j] = 0;
		}

		for(j = 0; j <4; j++) {
			char_array_4[j] = (unsigned char)(strchr(base64_chars, char_array_4[j]) - base64_chars);
		}

		char_array_3[0] = (char_array_4[0] << 2) + ((char_array_4[1] & 0x30) >> 4);
		char_array_3[1] = ((char_array_4[1] & 0xf) << 4) + ((char_array_4[2] & 0x3c) >> 2);
		char_array_3[2] = ((char_array_4[2] & 0x3) << 6) + char_array_4[3];

		for(j = 0; (j < i - 1); j++) ret[out_len++] = char_array_3[j];
	}

	if(i_iLen>3) {
		delete[] ret;
		ret = NULL;
	} else *o_iLen = out_len;
	return ret;
}

bool C_HttpUtil::url_encode(char *i_szString, char *o_szString, int i_iMaxOutLen)
{
	if(!i_szString || !o_szString || !i_iMaxOutLen) return false;
	int iPos = 0;
	o_szString[0] = 0;
	while(*i_szString) {
		if(*i_szString=='@') {
			strcpy(o_szString+iPos, "%40");
			iPos += 3;
		} else if(*i_szString=='.') {
			strcpy(o_szString+iPos, "%2E");
			iPos += 3;
		} else if(*i_szString=='-') {
			strcpy(o_szString+iPos, "%2D");
			iPos += 3;
		} else if(*i_szString=='_') {
			strcpy(o_szString+iPos, "%5F");
			iPos += 3;
		} else if(*i_szString==' ') {
			strcpy(o_szString+iPos, "+"); //%20
			iPos += 1;
		} else {
			o_szString[iPos] = *i_szString;
			iPos++;
		}
		o_szString[iPos] = 0;
		if(iPos>i_iMaxOutLen-4) return false;
		i_szString++;
	}
	return true;
}

//not complete, support ascii for now
// used by md5chksum_encode, should match web/utf8_encode.js, web/md5.js
bool C_HttpUtil::utf8_encode(char *i_szString, char *o_szString, int i_iMaxOutLen)
{
	if(i_szString==NULL || o_szString==NULL) {
		return false;
	}

	int len = (int)strlen(i_szString);
	char *utf8text = new char[len*4]; //alloc enough memory
	int pos = 0;
	utf8text[0] = 0;
	for(int i=0; i < len; i++) {
		int c1 = i_szString[i]; //only support ascii for now

		if (c1 < 128) {
			utf8text[pos++] = c1;
			utf8text[pos] = 0;
		} else if (c1 > 127 && c1 < 2048) {
			utf8text[pos++] = (c1 >> 6) | 192;
			utf8text[pos++] = (c1 & 63) | 128;
			utf8text[pos] = 0;
		} /*else if ((c1 & 0xF800) != 0xD800) {
			utf8text[pos++] = (c1 >> 12) | 224;
			utf8text[pos++] = ((c1 >> 6) & 63) | 128;
			utf8text[pos++] = (c1 & 63) | 128;
			utf8text[pos] = 0;
		} else {
			// surrogate pairs
			if ((c1 & 0xFC00) != 0xD800) {
				throw new RangeError('Unmatched trail surrogate at ' + n);
			}
			var c2 = string.charCodeAt(++n);
			if ((c2 & 0xFC00) != 0xDC00) {
				throw new RangeError('Unmatched lead surrogate at ' + (n - 1));
			}
			c1 = ((c1 & 0x3FF) << 10) + (c2 & 0x3FF) + 0x10000;

			utf8text[pos++] = (c1 >> 18) | 240;
			utf8text[pos++] = ((c1 >> 12) & 63) | 128;
			utf8text[pos++] = ((c1 >> 6) & 63) | 128;
			utf8text[pos++] = (c1 & 63) | 128;
			utf8text[pos] = 0;
		}*/
	}

	bool bResult = true;
	if(i_iMaxOutLen<(int)strlen(utf8text)+1) bResult = false;
	else strcpy(o_szString, utf8text);
	delete[] utf8text;
	return bResult;
}

bool C_HttpUtil::md5chksum_encode(char *i_szString, char *o_szString, int i_iMaxOutLen)
{
	MD5 md5;
	if (i_iMaxOutLen < 33) return false;
	int iUtf8Len = (int)strlen(i_szString) * 2;
	char *szUtf8text = new char[iUtf8Len];
	C_HttpUtil::utf8_encode(i_szString, szUtf8text, iUtf8Len);
	strcpy(o_szString, md5.digestString(szUtf8text));
	delete[] szUtf8text;
	return true;
}
