// C_Chat

//class to handle the text entering and drawing for the in game chat.

#include "ingamechat.h"
#include "global.h"

#define ENTER_POSX    (4+4)
#define ENTER_POSY    (160)


C_Chat::C_Chat()
{
	m_pclInput = C_Control::GetControlObject();
	m_pclGraph = C_GraphWrapperGL::GetGraphWrapperObject();

	m_pclFont = new C_Font(_RGBA8888);
	m_pclFont->Load((char*)"arial16", C_Global::szResourceFile);
	S_FColor stColor = S_FCOLORMAKE(1.0, 1.0, 1.0, 1.0);
	m_pclFont->SetColor(stColor);

	m_bEnteringText = false;
	m_szEnteredText[0] = 0;

	int i;
	for(i=0; i<MAX_INGAME_CHATMSG; i++) {
		m_szChatText[i][0] = 0;
		m_dTimer[i] = 0;
	}
}

C_Chat::~C_Chat()
{
	delete m_pclFont;
}

bool C_Chat::Update(double i_dTime)
{
	int i;
	for(i=0; i<MAX_INGAME_CHATMSG; i++) {
		m_dTimer[i] -= i_dTime;
		if(m_dTimer[i]<=0) {
			m_dTimer[i] = 0;
			m_szChatText[i][0] = 0;
		}
	}

	if(m_bEnteringText) {
		int c = GetInput();
		if(c!=13) {
			if(c>31 || c==8) AddChar((char)c);
		} else {
			m_bEnteringText = false;
			return (strlen(m_szEnteredText)>0);
		}
	} else {
		//if user pressed tab go to entering text mode
		m_bEnteringText = m_pclInput->IsPressedOnce(DIK_TAB)!=0;
		m_szEnteredText[0] = 0; //reset enter string
	}
	return false;
}

void C_Chat::Draw(int x, int y)
{
	if(m_bEnteringText) {
		//draw entering string
		char szText[MAX_INGAME_CHATLENGTH+2];
		strcpy(szText, m_szEnteredText);
		strcat(szText, "_");
		m_pclFont->Print(x+ENTER_POSX, y+ENTER_POSY, szText);
	}

	//draw chat strings
	int i;
	for(i=0; i<MAX_INGAME_CHATMSG; i++) {
		if(m_szChatText[i][0]!=0) {
			m_pclFont->Print(x+ENTER_POSX, y+ENTER_POSY-(24+i*16), m_szChatText[i]);
		}
	}
}

void C_Chat::AddChar(char i_cC)
{
	int i = (int)strlen(m_szEnteredText);

	//c=8 -> backspace
	if((i_cC == 8) && i) m_szEnteredText[--i] = 0;   //backspace
	if((i_cC != 8) && i<(MAX_INGAME_CHATLENGTH-1)) { //other chars
		m_szEnteredText[i++] = i_cC;
		m_szEnteredText[i] = 0;                      //always terminate
	}
}

int C_Chat::GetInput()
{
	for(uint8_t i=1; i<220; i++) {
		if(m_pclInput->IsPressedRepeat(i)) {
			int iAscii = m_pclInput->GetAscii(i);
			if(iAscii!=-1) {
				return (unsigned char)iAscii;
			}
		}
	}
	return -1;
}

void C_Chat::AddChatString(char *i_szText)
{
	int i;
	for(i=MAX_INGAME_CHATMSG-1; i>=1; i--) {
		strcpy(m_szChatText[i], m_szChatText[i-1]);
		m_dTimer[i] = m_dTimer[i-1];
	}

	m_dTimer[0] = 8000; //8sec
	strncpy(m_szChatText[0], i_szText, sizeof(m_szChatText[0]));
	m_szChatText[0][sizeof(m_szChatText[0])-1] = 0; //be sure to null terminate
}
