// C_NetMsg

//class to send and receive object data packets (S_NetMsg2) between clients.
//the class can be run in both TCP and UDP mode (see below).

#include "netmsg.h"


//UDP and TCP mode: create object and init UDP (use -1, 0 as params for TCP mode)
C_NetMsg::C_NetMsg(int i_iClientPortBase, int i_iClientPortSeekRange)
{
	m_pclServerComm = NULL;
	Init(i_iClientPortBase, i_iClientPortSeekRange);
}

//UDP mode: init receive socket
void C_NetMsg::Init(int i_iClientPortBase, int i_iClientPortSeekRange)
{
	m_pclMsg1List = new C_Msg1List();

	int i;
	for(i=0; i<8; i++) m_pclUDPSocket[i] = NULL;
	for(i=0; i<256; i++) m_aiNumInList[i] = 0;

	m_pclRecvSocket = new C_Socket();
	for(i=0; i<i_iClientPortSeekRange; i++) {
		m_iRecvPort = m_pclRecvSocket->InitRecvUDP(i_iClientPortBase+i);
		if(m_iRecvPort!=-1) break; //found a free port
	}
}

//UDP mode: must be run to use client<-UDP->clients communication
void C_NetMsg::SetIPList(int i_iIndex, int i_iIPNum, int i_iPort)
{
	m_pclUDPSocket[i_iIndex] = new C_Socket();
	m_pclUDPSocket[i_iIndex]->InitSendUDP(i_iIPNum, i_iPort);
	//^the port must be sent to the server to all players
}

//TCP mode: must be run to use client<-TCP->server<-TCP->clients communication
void C_NetMsg::SetServerComm(C_NetMsgServer *i_pclServerComm)
{
	delete m_pclRecvSocket; m_pclRecvSocket = NULL;
	m_pclServerComm = i_pclServerComm;
}

C_NetMsg::~C_NetMsg()
{
	delete m_pclMsg1List;

	delete m_pclRecvSocket;

	int i;
	for(i=0; i<8; i++) delete m_pclUDPSocket[i];
}

void C_NetMsg::Reset()
{
	m_pclMsg1List->clear();

	int i;
	for(i=0; i<256; i++) m_aiNumInList[i] = 0;
}

//UDP and TCP mode: run to send data
void C_NetMsg::SendMsg1(S_NetMsg1 *i_pstMsg1)
{
	int i;

	i_pstMsg1->iSize = sizeof(S_NetMsg1);
	i_pstMsg1->iType = N_MSG1;
	if(!m_pclServerComm) {
		for(i=0; i<8; i++) {
			if(m_pclUDPSocket[i]) {
				m_pclUDPSocket[i]->SendQueued((char*)i_pstMsg1, sizeof(S_NetMsg1));
			}
		}
	} else {
		m_pclServerComm->SendMsg1(i_pstMsg1);
	}
}

//TCP mode: must be run externally to make data available for objects
void C_NetMsg::AddMsg1(S_NetMsg1 *i_pstMsg1)
{
	//add struct to queue
	m_pclMsg1List->push_back(*i_pstMsg1);
	m_aiNumInList[i_pstMsg1->iSenderId]++;
}

//UDP mode: internal method to receive all data on a given socket
void C_NetMsg::DrainSocket(C_Socket *i_pclSocket)
{
	S_NetMsg1 stMsg1;

	int iLength = sizeof(S_NetMsg1);
	bool bResult;
	do {
		bResult = i_pclSocket->GetRecvBuffer((char*)&stMsg1, &iLength);
		if(bResult) {
			if(m_aiNumInList[stMsg1.iSenderId] < 3) {
				//add struct to queue
				m_pclMsg1List->push_back(stMsg1);
				m_aiNumInList[stMsg1.iSenderId]++;
			}
		}
		iLength = sizeof(S_NetMsg1);
	} while(bResult);
}

//UDP mode: polling method to receive all available data
void C_NetMsg::Recieve()
{
	if(m_pclRecvSocket) DrainSocket(m_pclRecvSocket);
}

//UDP and TCP mode: request the first message for an object ID
bool C_NetMsg::GetMsg1(int i_iId, S_NetMsg1 *o_pstMsg1)
{
	S_NetMsg1 *pstMsg1;

	Recieve();

	//check all recieved Msgs SenderId == i_iId
	int i, iSize = (int)m_pclMsg1List->size();
	for(i=0; i<iSize; i++) {
		pstMsg1 = &((*m_pclMsg1List)[i]);
		if(pstMsg1->iSenderId == i_iId) {
			memcpy(o_pstMsg1, pstMsg1, sizeof(S_NetMsg1));
			m_pclMsg1List->erase(m_pclMsg1List->begin()+i);
			m_aiNumInList[i_iId]--;
			return true;
		}
	}
	return false;
}
