// C_NetMsgServer

#ifndef _NETMSGSERVER_H
#define _NETMSGSERVER_H

#include "socket_2.h"
#include "s_hiscore.h"
#include "common.h"
#include <vector>

//simulate server network lag to be able to see problems due to network speed while testing
//#define DEBUG_SERVER_DELAY_MS 0 //delay milliseconds (0 should be used when not testing)
//^not implemented/used yet

class C_Replay;

//SERVER MSG
//sent to inform the server of certain events

//game setup msg                   valid fields in struct (iType always valid)
#define N_MSG2_ADDGAME      0x01 //iGameStatus, szGameName
#define N_MSG2_GAMESTATUS   0x02 //iGameStatus, szGameName
#define N_MSG2_REMOVEGAME   0x03 //szGameName
#define N_MSG2_DISCONNECT   0x04 //iPlayerId, 
#define N_MSG2_JOIN         0x05 //szPlayerName, szGameName
#define N_MSG2_NEW          0x06 //szPlayerName, szGameName
#define N_MSG2_JOIN_NEW_R   0x07 //iPlayerId, iGameId
#define N_MSG2_ADDPLAYER    0x09 //iPlayerId, szPlayerName
#define N_MSG2_OPTIONS      0x0B //szLevelName, iLevelType, iNumLaps, iOptions, iNumAI
#define N_MSG2_START        0x0C //
#define N_MSG2_START_R      0x0D //
#define N_MSG2_NEWMASTER    0x30 //
#define N_MSG2_LEAVE        0x31 //
//ingame msg
#define N_MSG2_GAMEOVER     0x0E //
#define N_MSG2_SCORE        0x0F //iScore
#define N_MSG2_NUMKILLS     0x10 //iNumKills(server) or iKilledBy(client)
#define N_MSG2_NUMLAPS      0x11 //iNumLaps
#define N_MSG2_NEXTLEVEL    0x22 //
//for ingame decisions (mutex)
#define N_MSG2_CARGO        0x20 //iZoneId, iCargoNum, iCargoWeight
#define N_MSG2_CARGO_R      0x21 //iZoneId, iCargoNum, iCargoReqAccepted
#define N_MSG2_CARGORESPAWN 0x23 //iZoneId, iCargoNum, iCargoWeight
#define N_MSG2_LIFE         0x24 //iZoneId
#define N_MSG2_LIFE_R       0x25 //iZoneId, iLifeReqAccepted

//uses separate datastructure from S_NetMsg2 because thay can be very long
#define N_MSG2_CHAT         0x08 // [strlength+4,N_MSG2_CHAT,nullterminated string]
#define N_MSG2_HISCORELIST  0x40 // [sizeof(S_HiScore)+4,N_MSG2_HISCORELIST,S_HiScore]

//player to player messages 
#define N_MSG1              0x50 // [S_NetMsg1]

//used instead of bitfields for cross platform compatibility
#define OPT_NOSHOOTING  0x01
#define OPT_FLYTHROUGH  0x02
#define OPT_NOBULLETHIT 0x04

//this struct could be packed harder, but there is little winnings since this
// message is not used that often and in replays it is shared with S_NetMsg1
// which is bigger.
//it would need a new version of replays, so it is kept as it
// is for now.
#pragma pack(1)
struct S_NetMsg2 //size 48 bytes
{
	uint16_t iSize;
	uint16_t iType; //the msg (one of the above)

	int16_t iPlayerId; //tag "id"
	int16_t iGameId;   //tag "gid"
	union { //tag "int1"
		int32_t iKilledBy;
		int32_t iNumKills;
		int32_t iNumLaps;
		int32_t iScore;
		int32_t iZoneId;
	};
	union { //tag "int2"
		int32_t iGameStatus;
		int32_t iCargoNum;
		uint32_t iLevelType;
	};
	union {
		union {
			int32_t iCargoWeight;
			int32_t iCargoReqAccepted;
			int32_t iLifeReqAccepted;
			struct {
				char szLevelName[24];
				uint16_t iNumAI;
			};
		};
		struct {
			char szGameName[MAXGAMENAMELENGTH];
			char szPlayerName[MAXPLAYERNAMELENGTH];
		};
	};
	union {
		uint16_t iOptions; //OPT_ defines above
	};
};
#pragma pack()


//PLAYER MSG
//sent for all C_PlayerBase objects whenever an event happens

//MSG
#define N_MSG1_MOVEMENT     0x01 //pos, vel, acc, dir valid
#define N_MSG1_KEYCHANGE    0x02 //user pressed or released a steering key
#define N_MSG1_NEWBULLET    0x04 //bullet info valid
#define N_MSG1_KILLED       0x08 //killed info valid
#define N_MSG1_LANDED       0x10 //landed on LandingZone
#define N_MSG1_SCORE        0x20 //score info valid
#define N_MSG1_LAPINFO      0x40 //lap info valid

//KEYSTATE
#define GF_KEY_ACCELERATE 0x01
#define GF_KEY_LEFT       0x02
#define GF_KEY_RIGHT      0x04
#define GF_KEY_DOWN       0x08

//this struct should use int32_t instead of float to be more
// compatible/deterministic (16bit integer, and 16bit fractional is ok).
//but it would need a new version of replays, so it is kept as it
// is for now. floats seems ok on both windows, linux and mac and java,
// but only on intel cpus ofcourse...
#pragma pack(1)
struct S_NetMsg1 //size 58 bytes
{
	//is always sent
	uint16_t iSize;         //sizeof(S_NetMsg1)
	uint16_t iType;         //N_MSG1 (0x50)
	uint8_t iMsg;           //the msg (from the above)
	int8_t iSenderId;       //player id (or -1 for server)
	//int32_t iSeqNr;         //not used

	//N_MSG1_MOVEMENT, sent every time for movable objects
	float fPosX, fPosY;   //player position at sendtime
	float fVelX, fVelY;   //player velocity at sendtime
	float fAcceleration;  //player acceleration at sendtime
	float fDirection;     //player direction in degrees at sendtime

	float    fGravityY;     //fGravityX is a waste to send for now since it never changes
	uint16_t iCargoInHold;  //weight of cargo
	bool     bCanCollide;

	//N_MSG1_KEYCHANGE, sent if changed (for correction)
	uint8_t  iKeyFlag;       //player keyboardstate at sendtime

	//N_MSG1_NEWBULLET, new bullet (is sent if a bullet has been fired)
	//N_MSG1_KILLED, remove bullet (is sent if a bullet has collided with the player)
	uint8_t  iNumBullets;
	int32_t  iBulletUniqeValBase; //the uniqe value that identifies this bullet (and the base if iNumBullets>1)
	//bullet direction is same as fDirection, velocity is same+BULLET_BASESPEED (use cos,sin(realdir))

	//N_MSG1_KILLED, player hit (computerenemy, AI or human player)
	union {
		uint8_t iDoorAction;   //only for door object
		uint8_t iKilledReason; //by bullet, player or self
		uint8_t iHitsLeft;     //computer enemy numhits left
	};
	int32_t iCauseID;

	//N_MSG1_SCORE
	int32_t iScore;
	//N_MSG1_LAPINFO
	uint8_t iLap;
	uint8_t iCP;
	float fBestLapTime;
};
#pragma pack()
typedef std::vector<S_NetMsg1> C_Msg1List;


class C_NetMsgServer
{
public:
	C_NetMsgServer(C_Socket *i_pclSocket);
	~C_NetMsgServer();

	void SetReplayObject(C_Replay *i_pclReplay) {m_pclReplay = i_pclReplay;};
	void SetReplayMode(bool bRunReplay);

	void SendMsg1(S_NetMsg1 *i_pstMsg1);
	void SendMsg2(S_NetMsg2 *i_pstMsg2);
	void SendMsg2WithId(S_NetMsg2 *i_pstMsg2);
	void SendChatMsg(char *i_szChatString);
	void SendHiScoreMsg(S_HiScore *i_pstHiScore);

	int GetMsg(S_NetMsg1 *o_pstMsg1, S_NetMsg2 *o_pstMsg2, S_HiScore *o_pstScore, char *o_szChatString, int i_iMaxChatLen);
	//returns 0 if fail, 1 if S_NetMsg2, 2 if S_HiScore, 3 if chat string, 4 if S_NetMsg1 is recieved

	int m_iPlayerId, m_iGameId;

private:
	C_Socket *m_pclSocket;
	C_Replay *m_pclReplay;

	bool m_bRunReplay;
	void AddReplayMsg1(S_NetMsg1 *i_pstMsg1);
	void AddReplayMsg2(S_NetMsg2 *i_pstMsg2);
};

#endif
