// C_Object

#ifndef OBJECT_H
#define OBJECT_H

#include "common.h"
#include "glgraph.h"
#include "image.h"

#include "environment_defaults.h"

#define COMPUTER_ENEMY_BASEID 8
#define DOOR_BASEID           112 //allows for 16 doors for now

class C_Object
{
public: 
	friend class C_LandingZone;
	friend class C_Door;
	friend class C_PlayerBase;

	C_Object();

	int GetOwnerID() {return m_iOwnerID;};
	void SetOwnerID(int i_iOwnerID) {m_iOwnerID = i_iOwnerID;};
	void GetPos(int *o_iX, int *o_iY);
	void SetPos(int i_iX, int i_iY);
	bool GetSize(int *o_iSizeW, int *o_iSizeH);
	bool GetCenterPos(int *o_iCenterX, int *o_iCenterY);
	bool GetCenterPos(POINT *o_pstCenter);
	double GetDistanceTo(C_Object *i_pclObj);

	/* calculates new position and speed, depending on the elapsed time
	i_dTime    time in milliseconds since last call
	uses m_dDirection, m_dAcceleration as input; they must be set to the newest value */
	void OUpdate(double i_dTime);

	/* transparent draw of the provided image onto screen if visible
	(clipping preformed if partly visible)
	i_Img      image
	i_pstRect  object rectangle within i_Img, NULL = entire img
	i_iX       x-cordinate of upper left corner of screen in the global space
	i_iY       y-cordinate of upper left corner of screen in the global space
	screenposition is (m_dPosX-i_x, m_dPosY-i_Y)  */
	void ODraw(C_Image *i_pclImg, S_Rect *i_pstRect, int i_iX, int i_iY);

	/* tests if two images have collided, first check if they overlap,
	and if they do, check their common overlapping rectangles pixel by pixel,
	and if two pixels on the same position is != 0 there is a collision
	i_OtherImg other object image
	i_iOtherX  x-cordinate of upper left corner of OtherImg in the global space
	i_iOtherY  y-cordinate of upper left corner of OtherImg in the global space */
	bool OCheckCollision(C_Image *i_pclOtherImg, int i_iOtherX, int i_iOtherY);
	bool OCheckCollision(C_Object *i_pclObject);

	static double PointDistance(int i_iX1, int i_iY1, int i_iX2, int i_iY2);
protected:
	double m_dTime;           //time
	double m_dPosX;           //object global position in x-direction
	double m_dPosY;           //object global position in y-direction
	double m_dVelX;           //object speed in x-direction
	double m_dVelY;           //object speed in y-direction
	double m_dAcceleration;   //object acceleration in the heading direction
	double m_dDirection;      //object direction in degrees

	double m_dResistanceX, m_dResistanceY; //object air resistance (0 for no resistance)
	double m_dGravityX, m_dGravityY;       //the gravity the object will feel (can be negative) (0 for no gravity) (pixel/sec2)

	int    m_iDirStep;        //number of degrees between each step (10=default)
	int    m_iRealDir;        //m_dDirection/m_iDirStep (the direction that the object can have)

	int    m_iOwnerID;

	S_Rect m_stScreenArea;    //screen area 
	C_GraphWrapperGL *m_pclGraph;//graphics interface

	bool m_bCanCollide;       //this must be false when the object is displayed but it cannot collide. ex. during explosion maybe

	S_Rect *m_pstImageRect;
	C_Image *m_pclImage;
};

#endif
