// C_AStar

//class that implement the A* algorithm
//!old AI

#include "pathfind.h"

#include <algorithm> //reverse()
#include <math.h> //sqrt()

////////////////////////////////////////////////////////////////////////////////

inline bool operator == (const POINT& a, const POINT& b)
{
	return (a.x == b.x) && (a.y == b.y);
}

inline bool operator != (const POINT& a, const POINT& b)
{
	return (a.x != b.x) || (a.y != b.y);
}

////////////////////////////////////////////////////////////////////////////////
#define COST_MOVE 100
#define COST_MOVEDIAG 42 //extra penelty for diagonal movement (+41% is normal)
#define MAXIMUM_PATH_LENGTH 1000000

POINT C_AStar::s_stNoPoint = {-1,-1};
POINT C_AStar::s_stDirections[8] = {{-1,0},{0,-1},{1,0},{0,1},{-1,-1},{1,-1},{1,1},{-1,1}};

C_AStar::C_AStar(C_Grid<char> *i_pclMap, double i_dQuant)
{
	m_dPathDiv = i_dQuant;

	m_pclMap = new C_Grid<char>(i_pclMap);
	m_aaMap = m_pclMap->aaGrid;
	m_iMapWidth = m_pclMap->iWidth;
	m_iMapHeight = m_pclMap->iHeight;

	m_pclNodes = new C_Grid<S_Node>(m_iMapWidth, m_iMapHeight);
	m_aaNodes = m_pclNodes->aaGrid;

	//this takes some time, but is only done once
	m_iNumOpen = 0;
	int j, i;
	for(j=0; j<m_iMapHeight; j++) {
		for(i=0; i<m_iMapWidth; i++) {
			//init all nodes
			m_aaNodes[j][i].iState = UNVISITED;
		}
	}
	//m_stOpenFirst = s_stNoPoint;
	//m_stOpenLast = s_stNoPoint;
	m_aiClosed = new int*[m_iMapHeight*m_iMapWidth]; //maximum size of closed list
	m_iNumClosed = 0;
}

C_AStar::~C_AStar()
{
	delete m_pclMap;
	delete m_pclNodes;
	delete[] m_aiClosed;
}

inline int C_AStar::ManhattanDist(POINT *start, POINT *a, POINT *b)
{
	// The Manhattan distance is what should be used in A*'s heuristic
	// distance estimate, *not* the straight-line distance.  This is because
	// A* wants to know the estimated distance for its paths, which involve
	// steps along the grid.  (Of course, if you assign 1.4 to the cost of
	// a diagonal, then you should use a distance function close to the
	// real distance.)
	double dx1 = a->x - b->x;
	double dy1 = a->y - b->y;
	double dx2 = start->x - b->x;
	double dy2 = start->y - b->y;
	double cross = dx1*dy2 - dx2*dy1;
	if(cross < 0) cross = -cross;

	// Scaling is nice because you can adjust the accuracy/speed tradeoff.
	// If you want a faster search, you can get an approximate answer.
	//return dist(a, b) + int(cross/20000);
	int dx = abs(a->x - b->x);
	int dy = abs(a->y - b->y);

	double dResult = sqrt((double)(dx*dx + dy*dy)) * (double)COST_MOVE;
	return (int)(dResult / m_dPathDiv);
	//return (COST_MOVE*(dx+dy) / m_dPathDiv) + (int)(cross * 0.001); //0.01
	//return ((dx>dy ? COST_MOVE*dx : COST_MOVE*dy) / m_dPathDiv) + (int)(cross * 0.001); //0.01
}

inline int C_AStar::Cost(POINT *a, POINT *b)
{
	//This is the cost of moving one step. To get completely accurate
	//paths, this must be greater than or equal to the change in the
	//distance function when you take a step.
	int iTerrain = m_aaMap[b->y][b->x];

	//no walking through walls!
	if(iTerrain == -1) return MAXIMUM_PATH_LENGTH;

	//calc cost (1=100, 2=110 ..., 10..80 *= 100)
	if(iTerrain >= 10) iTerrain *= 100;
	else iTerrain = (iTerrain+9)*10;

	if(a->x!=b->x && a->y!=b->y) { //diagonal
		return (int)(iTerrain * 1.42);
		//check for diagonal valid (so the below does not happen)
		//if(m_aaMap[a->y][b->x] == 0)  //01B0
		// if(m_aaMap[b->y][a->x] == 0) //0A10
		//  return iTerrain * 1.6;
		//return MAXIMUM_PATH_LENGTH;
	}
	return (int)(iTerrain * 1.0);
}

//the algorithm: take map, two points (A and B) output the path as result.
//TODO: consider previous heading, and make it a greater cost to turn around. this needs a point
// 'preA' (note: we don't know the velocity here since it's calculated in the steering,
// but we can assume previous path heading is the ship heading)
C_PointList *C_AStar::FindPath(POINT A, POINT B)
{
	C_PointList *pclResult = NULL;
	S_Node      *n, *n2;
	POINT       stPoint, stNeighbor;
	int         i;

	//valid positions?
	if(B.y<0 || B.x<0 || A.y<0 || A.x<0 || B.y>=m_iMapHeight
		|| B.x>=m_iMapWidth || A.y>=m_iMapHeight || A.x>=m_iMapWidth) return NULL;

	//fast precheck to see if the goal itself is invalid
	if(m_aaMap[B.y][B.x]==-1) return NULL;

	{
		// insert the first position (A)
		n = &m_aaNodes[A.y][A.x];
		n->gval = 0;                         //how far we travelled
		n->hval = ManhattanDist(&A, &A, &B); //how far we think it is left
		n->stParent = s_stNoPoint;

		OpenAdd(&A);
	}

	int iNumVisited = 0, iNumSearch = 0, iNumOpened = 1;

	//while there are still nodes to visit, visit them!
	while(m_iNumOpen) {
		iNumSearch++;
		GetBestOpen(&stPoint);
		//if there is no good open point and we have
		//not reached point B, we never will
		if(stPoint == s_stNoPoint) break;

		//the current point to work with
		n = &m_aaNodes[stPoint.y][stPoint.x];

		//remove from open
		OpenRemove(&stPoint);

		//if we're at the goal, then exit
		if(stPoint == B) break;

		//look at the neighbors
		for(i=0; i<8; i++) {
			stNeighbor = stPoint;
			stNeighbor.x += s_stDirections[i].x;
			stNeighbor.y += s_stDirections[i].y;

			//if it's off the end of the map, then don't keep scanning
			if(!m_pclMap->IsValid(stNeighbor.x, stNeighbor.y)) continue;
			iNumVisited++;
			int k = Cost(&stPoint, &stNeighbor);
			/**/if(k >= MAXIMUM_PATH_LENGTH) continue;
			//^is this ok?

			n2 = &m_aaNodes[stNeighbor.y][stNeighbor.x];
			int gval = n->gval + k;

			if(n2->iState == UNVISITED) {
				n2->gval = gval;
				n2->hval = ManhattanDist(&A, &stNeighbor, &B); //heuristic
				n2->stParent = stPoint;
				OpenAdd(&stNeighbor);
				iNumOpened++;
			} else if(n2->iState == OPEN) {
				//it's in OPEN, so figure out if this is better
				if(gval < n2->gval) {
					n2->gval = gval;
					//n2->hval is correct already
					n2->stParent = stPoint;
				}
			} //else CLOSED => skip
		}
	}

	if(stPoint==B && n->gval<MAXIMUM_PATH_LENGTH) {
		pclResult = new C_PointList();

		//found a path, copy it into result
		POINT h = B;
		while(h != A) {
			n2 = &m_aaNodes[h.y][h.x];
			pclResult->push_back(h);
			h = n2->stParent;
		}
		pclResult->push_back(h); //add A

		//path now contains the points of the path... backwards
		//swap the order
		std::reverse(pclResult->begin(), pclResult->end());
	} else {
		//no path
	}

	ClearAll();

	/*FILE *f = fopen("stat.txt", "wt");
	fprintf(f, "numopened: %d\n", iNumOpened);
	fprintf(f, "numsearches for the best: %d\n", iNumSearch);
	fprintf(f, "numvisited: %d\n", iNumVisited);
	fclose(f);

	FILE *f = fopen("map.txt", "wt");
	for(int j=0; j<m_iMapHeight; j++) {
		for(int i=0; i<m_iMapWidth; i++) {
			fprintf(f, "%5d ", m_aaNodes[j][i].gval);
		}
		fprintf(f, "\n");
	}
	fclose(f);*/

	return pclResult;
}

void C_AStar::GetBestOpen(POINT *o_pstBest)
{
	int iCost, iBest = MAXIMUM_PATH_LENGTH; //set high
	S_Node *n;
	POINT  *p;

	*o_pstBest = s_stNoPoint;
	p = &m_stOpenFirst;
	while(p->x != -1) { //not end of list
		n = &m_aaNodes[p->y][p->x];
		//assert(n->iState == OPEN);
		iCost = n->gval + n->hval;
		if(iCost < iBest) {
			iBest = iCost;
			o_pstBest->x = p->x;
			o_pstBest->y = p->y;
		}
		p = &n->stNext;
	}
}

void C_AStar::OpenAdd(POINT *i_pstPoint)
{
	S_Node *n = &m_aaNodes[i_pstPoint->y][i_pstPoint->x];
	if(m_iNumOpen == 0) {
		m_stOpenFirst = *i_pstPoint;
		m_stOpenLast  = *i_pstPoint;
		n->stPrev = s_stNoPoint;
		n->stNext = s_stNoPoint;
		//assert(!(n->iState == OPEN)); //must not be in open already!
	} else {
		S_Node *last = &m_aaNodes[m_stOpenLast.y][m_stOpenLast.x];
		last->stNext = *i_pstPoint;
		n->stPrev = m_stOpenLast;
		n->stNext = s_stNoPoint;
		m_stOpenLast = *i_pstPoint;
	}
	n->iState = OPEN;
	m_iNumOpen++;
}

void C_AStar::OpenRemove(POINT *i_pstPoint)
{
	//assert(m_iNumOpen > 0); //list must not be empty
	S_Node *np, *nn, *n = &m_aaNodes[i_pstPoint->y][i_pstPoint->x];
	//assert(n->iState == OPEN); //must be in open!

	bool bMiddle = true;
	if(n->stPrev == s_stNoPoint) { //remove first
		m_stOpenFirst = n->stNext;
		if(n->stNext != s_stNoPoint) { //relink next
			nn = &m_aaNodes[n->stNext.y][n->stNext.x];
			nn->stPrev = s_stNoPoint;
			m_stOpenFirst = n->stNext;
		}
		bMiddle = false;
	}
	if(n->stNext == s_stNoPoint) { //remove last
		m_stOpenLast  = n->stPrev;
		if(n->stPrev != s_stNoPoint) { //relink prev
			np = &m_aaNodes[n->stPrev.y][n->stPrev.x];
			np->stNext = s_stNoPoint;
			m_stOpenLast = n->stPrev;
		}
		bMiddle = false;
	}
	if(bMiddle) { //relink both
		nn = &m_aaNodes[n->stNext.y][n->stNext.x];
		np = &m_aaNodes[n->stPrev.y][n->stPrev.x];
		np->stNext = n->stNext;
		nn->stPrev = n->stPrev;
	}

	m_aiClosed[m_iNumClosed++] = &n->iState; //add to closed list
	n->iState = CLOSED;
	m_iNumOpen--;
}

void C_AStar::ClearAll()
{
	S_Node *n;
	POINT  *p;
	int    i;

	//clear open
	if(m_iNumOpen>0) {
		p = &m_stOpenFirst;
		while(p->x != -1) { //not end of list
			n = &m_aaNodes[p->y][p->x];
			//assert(n->iState == OPEN);
			n->iState = UNVISITED;
			p = &n->stNext;
		}
		m_stOpenFirst = s_stNoPoint;
		m_stOpenLast = s_stNoPoint;
		m_iNumOpen = 0;
	}

	//clear closed
	for(i=0; i<m_iNumClosed; i++) {
		*m_aiClosed[i] = UNVISITED;
	}
	m_iNumClosed = 0;
}

bool C_AStar::Walkable(POINT *a, POINT *b)
{
	bool bFound = false;
	int iNumTimes, iTemp[2];
	int idx = b->x - a->x;
	int idy = b->y - a->y;
	double x = a->x, y = a->y;
	double dx = idx, dy = idy;
	idx = abs(idx);
	idy = abs(idy);
	if(idy > idx) {
		iNumTimes = idy;
		dy = dy/idy;
		dx = dx/idy;
	} else {
		iNumTimes = idx;
		dx = dx/idx;
		dy = dy/idx;
	}

	//test two offset lines
	double x1 = x + dy*0.3;
	double y1 = y - dx*0.3;
	double x2 = x - dy*0.3;
	double y2 = y + dx*0.3;
	//5 steps per tile distance
	dx /= 5;
	dy /= 5;
	iNumTimes *= 5;
	//test along the two offset lines
	for(int i=0; i<iNumTimes; i++) {
		x1 += dx; y1 += dy;
		x2 += dx; y2 += dy;
		iTemp[0] = m_aaMap[(int)y1][(int)x1];
		iTemp[1] = m_aaMap[(int)y2][(int)x2];
		if(iTemp[0]==-1 || iTemp[0]>=10 || iTemp[1]==-1 || iTemp[1]>=10) { //a wall or next to a wall
			bFound = true;
			break;
		}
	}

	return !bFound;
}

void C_AStar::SmothPath(C_PointList *io_pclPoints)
{
	POINT *p0, *p2;
	int i, iSize = (int)io_pclPoints->size();

	if(iSize<3) return; //can't do anything

	for(i=0; i<iSize-2; i++) {
		p0 = &(*io_pclPoints)[i];
		p2 = &(*io_pclPoints)[i+2];
		if(Walkable(p0, p2)) {
			io_pclPoints->erase(io_pclPoints->begin()+(i+1));
			iSize--; i--; //compensate for erase
		}
	}

	ReduceSegments(io_pclPoints);
}

void C_AStar::ReduceSegments(C_PointList *io_pclPoints)
{
	POINT *p0, *p1, *p2;
	int i, iSize = (int)io_pclPoints->size();

	if(iSize<3) return; //can't do anything

	for(i=0; i<iSize-2; i++) {
		p0 = &(*io_pclPoints)[i];
		p1 = &(*io_pclPoints)[i+1];
		p2 = &(*io_pclPoints)[i+2];
		double dx1 = p1->x - p0->x;
		double dy1 = p1->y - p0->y;
		double dx2 = p2->x - p1->x;
		double dy2 = p2->y - p1->y;
		double dLen1 = sqrt(dx1*dx1+dy1*dy1);
		double dLen2 = sqrt(dx2*dx2+dy2*dy2);
		dx1 /= dLen1; dy1 /= dLen1; //normalize
		dx2 /= dLen2; dy2 /= dLen2; //normalize
		dx1 -= dx2; dy1 -= dy2;

		if(dx1<0.001 && dx1>-0.001 && dy1<0.001 && dy1>-0.001) { //never trust floats :/
			io_pclPoints->erase(io_pclPoints->begin()+(i+1));
			iSize--; i--; //compemsate for erase
		}
	}
}

//test program
/*#include <stdio.h>
void main()
{
	C_Grid<char> clGrid(10, 10);

	//make map with border
	int x, y;
	for(y=0; y<10; y++) {
		for(x=0; x<10; x++) {
			if(x==0 || y==0 || x==9 || y==9) clGrid.aaGrid[y][x] = -1;
			else clGrid.aaGrid[y][x] = 1;
		}
	}
	//additional obstacles
	clGrid.aaGrid[1][4] = -1;
	clGrid.aaGrid[2][4] = -1;
	clGrid.aaGrid[3][4] = -1;
	clGrid.aaGrid[4][4] = -1;
	clGrid.aaGrid[5][4] = -1;
	clGrid.aaGrid[6][4] = -1;
	clGrid.aaGrid[7][4] = -1;
	//clGrid.aaGrid[8][4] = -1;

	POINT A = {1,1}; //start
	POINT B = {8,8}; //destination

	C_AStar clAStar(&clGrid, 1.0);
	C_PointList *pclPointList = clAStar.FindPath(A, B);

	if(pclPointList) {
		int i, iSize = pclPointList->size();
		for(i=0; i<iSize; i++) {
			printf("[%d %d]\n", (*pclPointList)[i].x, (*pclPointList)[i].y);
		}
	} else {
		printf("impossible!\n");
	}
}*/
