// C_AStar

#ifndef PATHFIND_H
#define PATHFIND_H

#include <vector> //stl

#include "common.h"

#include "Grid.h"

#define UNVISITED 0
#define OPEN      1
#define CLOSED    2
struct S_Node
{
	int   gval;
	int   hval;
	POINT stParent;
	POINT stPrev, stNext;
	int   iState;
};

typedef std::vector<POINT> C_PointList;

class C_AStar
{
public:
	C_AStar(C_Grid<char> *i_pclMap, double i_dQuant); //1<i_dQuant<+MAX, 1 => best quality, slowest
	~C_AStar();

	//returns the shortest path from A to B
	C_PointList *FindPath(POINT i_stPointA, POINT i_stPointB);

	//removes some points in the list to get rid of the zigzaging
	void SmothPath(C_PointList *io_pclPoints);
private:
	double m_dPathDiv;

	C_Grid<char> *m_pclMap;
	int m_iMapWidth, m_iMapHeight;
	char   **m_aaMap;   //from m_pclMap

	C_Grid<S_Node> *m_pclNodes;
	S_Node **m_aaNodes; //from m_pclNodes

	//helpers
	int ManhattanDist(POINT *start, POINT *a, POINT *b);
	int Cost(POINT *a, POINT *b);

	//list functions
	void GetBestOpen(POINT *o_pstBest);
	void OpenAdd(POINT *i_pstPoint);
	void OpenRemove(POINT *i_pstPoint);
	void ClearAll(); //clear to ALL nodes to UNVISITED
	POINT m_stOpenFirst, m_stOpenLast;
	int   m_iNumOpen;

	static POINT s_stNoPoint;
	static POINT s_stDirections[8];

	int **m_aiClosed;
	int m_iNumClosed;

	//smothing functions
	bool Walkable(POINT *a, POINT *b);
	void ReduceSegments(C_PointList *io_pclPoints); //joins all segments next to each other that have the same direction
};

#endif
