// C_PlayerBase, C_ShipPlayer, C_Player, C_Enemy

//classes and base classes for all player objects.
//code added for ai player in C_Player, this is ongoing work

#include "player.h"
#include "global.h"
#include <math.h>

#include "soundfx.h"

//AI////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
#include "PathUtils.h"
#include "AISteering.h"
#include "mapobjects.h"
#include "viewport.h"
#define AI_DEBUG_DRAW
using namespace PathUtils;
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////

C_Image *C_PlayerBase::s_pclExplosionImg = NULL;
S_Rect C_PlayerBase::s_stExplosionRects[16];
int C_PlayerBase::s_iObjCount = 0;
int C_PlayerBase::s_iBulletCounter = 0;
C_Object *C_PlayerBase::s_pclListner = NULL;

C_PlayerBase::C_PlayerBase()
	: C_Object()
{
	int i, j;

	if(!s_iObjCount) {
		s_pclExplosionImg = new C_Image((char*)"explosion.tga", C_Global::szResourceFile);
		int iExposionSizeX, iExposionSizeY;
		s_pclExplosionImg->GetInfo(&iExposionSizeX, &iExposionSizeY, NULL);
		iExposionSizeX /= 8; iExposionSizeY /= 2;
		for(j=0; j<2; j++) {
			for(i=0; i<8; i++) {
				s_stExplosionRects[j*8+i].x = i*iExposionSizeX;
				s_stExplosionRects[j*8+i].y = j*iExposionSizeY;
				s_stExplosionRects[j*8+i].width = iExposionSizeX;
				s_stExplosionRects[j*8+i].height = iExposionSizeY;
			}
		}
	}
	s_iObjCount++;

	m_pclExpTimer = new C_Timer();
	m_dFrameNr = 0.0;
	m_iRealDir = 0;
	m_fVol = 0.5f; m_fPan = 0.0f;

	m_bAlive = true;
}

C_PlayerBase::~C_PlayerBase()
{
	//last object left?
	s_iObjCount--;
	if(!s_iObjCount) {
		delete s_pclExplosionImg;
		s_pclExplosionImg = NULL;
		s_iBulletCounter = 0;
	}

	delete m_pclExpTimer;
}

void C_PlayerBase::InitFromObject(C_Object *i_pclSrcObject)
{
	m_dPosX = i_pclSrcObject->m_dPosX;
	m_dPosY = i_pclSrcObject->m_dPosY;
	m_dVelX = i_pclSrcObject->m_dVelX;
	m_dVelY = i_pclSrcObject->m_dVelY;
	m_iDirStep = i_pclSrcObject->m_iDirStep;
	m_dDirection = i_pclSrcObject->m_dDirection;
	m_iRealDir = i_pclSrcObject->m_iRealDir;
}

bool C_PlayerBase::UpdateExplosion(double i_dTime)
{
	if(i_dTime==0.0) m_dFrameNr = 0; //reset
	m_dFrameNr += i_dTime*0.018; //18 fps anim

	bool bReady = (m_dFrameNr > 15);
	return bReady;
}

void C_PlayerBase::DrawExplosion(int i_iX, int i_iY)
{
	if(m_dFrameNr < 16) { //only draw until finished
		S_Rect *pstRect = &s_stExplosionRects[(int)m_dFrameNr];

		//center explosion
		int iSizeX, iSizeY;
		GetSize(&iSizeX, &iSizeY);
		i_iX -= (iSizeX/2) - (pstRect->width/2);
		i_iY -= (iSizeY/2) - (pstRect->height/2);

		ODraw(s_pclExplosionImg, pstRect, i_iX, i_iY);
	}
}

#define HEARING_RANGE 2200.0 //maxdistance sound is played at all
#define PANNING_RANGE 500.0  //maxdistance sound is played in both speakers
void C_PlayerBase::CalcSoundVolPan(C_Object *i_pclListner, float *o_fVol, float *o_fPan)
{
	if(!i_pclListner) {
		*o_fVol = 0.5f; //half vol
		*o_fPan = 0.0f; //centered
		return;
	}

	int iX, iY, iSizeX, iSizeY, iOwnSizeX, iOwnSizeY;
	i_pclListner->GetPos(&iX, &iY);
	i_pclListner->GetSize(&iSizeX, &iSizeY);
	GetSize(&iOwnSizeX, &iOwnSizeY);

	double dDistX = (m_dPosX+iOwnSizeX*0.5) - (iX+iSizeX*0.5);
	double dDistY = (m_dPosY+iOwnSizeY*0.5) - (iY+iSizeX*0.5);
	double dDist = sqrt(dDistX*dDistX + dDistY*dDistY);

	//volume
	*o_fVol = (float)(1.0 - dDist / HEARING_RANGE);
	//panning
	*o_fPan = (float)(dDistX / PANNING_RANGE);
	if(*o_fPan < -1.0f) *o_fPan = -1.0f;
	else if(*o_fPan > 1.0f) *o_fPan = 1.0f;
}

void C_PlayerBase::SetCommonInfo(S_NetMsg1 *o_pstMsg1)
{
	o_pstMsg1->iSenderId = (int8_t)m_iOwnerID;
	o_pstMsg1->fAcceleration = (float)m_dAcceleration;
	o_pstMsg1->fDirection = (float)m_dDirection;
	o_pstMsg1->fPosX = (float)m_dPosX;
	o_pstMsg1->fPosY = (float)m_dPosY;
	o_pstMsg1->fVelX = (float)m_dVelX;
	o_pstMsg1->fVelY = (float)m_dVelY;

	o_pstMsg1->fGravityY = (float)m_dGravityY;
	o_pstMsg1->bCanCollide = m_bCanCollide;
}

void C_PlayerBase::GetCommonInfo(S_NetMsg1 *i_pstMsg1)
{
	m_dAcceleration = i_pstMsg1->fAcceleration;
	m_dDirection = i_pstMsg1->fDirection;
	m_dPosX = i_pstMsg1->fPosX;
	m_dPosY = i_pstMsg1->fPosY;
	m_dVelX = i_pstMsg1->fVelX;
	m_dVelY = i_pstMsg1->fVelY;

	m_dGravityY = i_pstMsg1->fGravityY;
	m_bCanCollide = i_pstMsg1->bCanCollide;
}

////////////////////////////////////////////////////////////////////////////////

int C_ShipPlayer::s_iObjCount = 0;
C_Image *C_ShipPlayer::s_pclShipImg = NULL;
S_Rect C_ShipPlayer::s_stShipRects[72];
int C_ShipPlayer::s_iShipSize;

S_FColor C_ShipPlayer::s_stShipColors[8] = {
	{0xAD/255.0f,0x16/255.0f,0x16/255.0f,1},
	{0x46/255.0f,0xB3/255.0f,0x9E/255.0f,1},
	{0x16/255.0f,0xAE/255.0f,0x16/255.0f,1},
	{0x19/255.0f,0x4F/255.0f,0xBF/255.0f,1},
	{0xC4/255.0f,0xAF/255.0f,0x46/255.0f,1},
	{0x9E/255.0f,0x16/255.0f,0x98/255.0f,1},
	{0xCD/255.0f,0xCD/255.0f,0xCD/255.0f,1},
	{0x93/255.0f,0x9F/255.0f,0x16/255.0f,1},
};

//code to remove clicks/pops when stopping looping audio
//reduce volume until samples are close to 0 before stopping
std::vector<void*> C_ShipPlayer::s_clFadeOutList;
void C_ShipPlayer::UpdateFadeOut(double i_dTime)
{
	int i;
	for(i=0; i<(int)s_clFadeOutList.size(); i++) {
		float fVol = C_Sound::GetSoundObject()->GetVolume(C_ShipPlayer::s_clFadeOutList[i]);
		if(fVol>0.1f) {
			//done in 30 steps (each step ~4ms)
			C_Sound::GetSoundObject()->SetVolume(C_ShipPlayer::s_clFadeOutList[i], fVol*0.92f);
		} else {
			C_Sound::GetSoundObject()->Stop(C_ShipPlayer::s_clFadeOutList[i]);
			C_ShipPlayer::s_clFadeOutList.erase(C_ShipPlayer::s_clFadeOutList.begin()+i);
			i--;
		}
	}
}

C_ShipPlayer::C_ShipPlayer(C_Map *i_pclMap, char *i_szName, int i_iPlayerID, int i_iStartX, int i_iStartY, C_NetMsg *i_pclNetMsg)
	: C_PlayerBase()
{
	int i, j;
	m_iOwnerID = i_iPlayerID;
	strcpy(m_szName, i_szName);

	m_pclNetMsg = i_pclNetMsg;

	m_pclMap = i_pclMap;
	m_stScreenArea = *(m_pclMap->GetScreenRect());
	m_pclMap->GetMapSize(&m_iMapWidth, &m_iMapHeight);

	m_iStartPosX = i_iStartX;
	m_iStartPosY = i_iStartY;
	m_dPosX = m_iStartPosX; 
	m_dPosY = m_iStartPosY;
	m_dVelX = m_dVelY = m_dAcceleration = 0.0;
	m_dDirection = 270.0;
	m_pEngineSoundCh = NULL;

	m_dResistanceX = C_Environment::GetShipResistanceX();
	m_dResistanceY = C_Environment::GetShipResistanceY();

	//cargo
	m_iSpaceUsed = 0;
	m_iNumUsed = 0;
	ResetGravity();

	if(!s_iObjCount) {
		s_pclShipImg = new C_Image((char*)"ship_w.tga", C_Global::szResourceFile);
		s_pclShipImg->GetInfo(&s_iShipSize, NULL, NULL);
		s_iShipSize /= 9;

		for(j=0; j<4; j++) {       //ship
			for(i=0; i<9; i++) {
				s_stShipRects[(j*9+i)].x      = i*s_iShipSize;
				s_stShipRects[(j*9+i)].y      = ((j+1)%4)*s_iShipSize;
				s_stShipRects[(j*9+i)].width  = s_iShipSize;
				s_stShipRects[(j*9+i)].height = s_iShipSize;
			}
		}
		for(j=4; j<8; j++) {       //thrust
			for(i=0; i<9; i++) {
				s_stShipRects[(j*9+i)].x      = i*s_iShipSize;
				s_stShipRects[(j*9+i)].y      = (4+((j+1)%4))*s_iShipSize;
				s_stShipRects[(j*9+i)].width  = s_iShipSize;
				s_stShipRects[(j*9+i)].height = s_iShipSize;
			}
		}
	}
	s_iObjCount++;

	m_bAlive = true;
	m_bFreeFromBullets = false;
	m_pclImage = s_pclShipImg;
	m_iScore = 0;

	memset(&m_stMsg1, 0, sizeof(m_stMsg1));
	m_bLanded = false;

	m_iCurCP = 0;
	m_iCurLap = 0;
	m_dBestLapTime = 10000000; //init to a high value

	m_iNumLifes = -1;

	m_dTimeWithoutInfo = 0;
	m_bPlayerLost = false;

	m_dMaxFuelInTank = 60.0;
	m_dFuel = m_dMaxFuelInTank;

//AI////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
	m_pclFont = NULL;
	m_pclAI   = NULL;
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
}

C_ShipPlayer::~C_ShipPlayer()
{
	if(m_dAcceleration) {
		C_ShipPlayer::s_clFadeOutList.push_back(m_pEngineSoundCh);
		//if (C_Sound::GetSoundObject()) C_Sound::GetSoundObject()->Stop(m_pEngineSoundCh); //stop engine sound
	}

	s_iObjCount--;
	if(!s_iObjCount) {
		delete s_pclShipImg;
		s_pclShipImg = NULL;
	}
}

void C_ShipPlayer::Draw()
{
	int iMapX, iMapY;
	m_pclMap->GetPos(&iMapX, &iMapY);

	m_stScreenArea = *(m_pclMap->GetScreenRect());

//AI////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
#ifdef AI_DEBUG_DRAW
	if(m_pclAI) {
		//draw debug
		S_DebugDraw stDbgElement;

		m_pclGraph->SetClippingRect(&m_stScreenArea);
		while(m_pclAI->GetDebug(&stDbgElement)) {
			S_FColor stColor = {(stDbgElement.iColor&0xff)/255.0f, 
				((stDbgElement.iColor&0xff00)>>8)/255.0f,
				((stDbgElement.iColor&0xff0000)>>16)/255.0f,1};
			m_pclGraph->SetColor(&stColor);

			switch(stDbgElement.iType) {
				case DBG_TEXT:
					m_pclFont->Print(m_stScreenArea.x+stDbgElement.p1.x-iMapX,
						m_stScreenArea.y+stDbgElement.p1.y-iMapY, stDbgElement.szText);
					break;
				case DBG_LINE:
					m_pclGraph->Line(m_stScreenArea.x+stDbgElement.p1.x-iMapX, m_stScreenArea.y+stDbgElement.p1.y-iMapY,
						m_stScreenArea.x+stDbgElement.p2.x-iMapX, m_stScreenArea.y+stDbgElement.p2.y-iMapY);
					break;
				case DBG_CIRCLE:
					m_pclGraph->Circle(m_stScreenArea.x+stDbgElement.p1.x, m_stScreenArea.y+stDbgElement.p1.y, stDbgElement.iRadius);
					break;
			}
		}
		m_pclGraph->SetClippingRect(NULL);
	}
#endif
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

	if(m_bAlive) {
		S_FColor stColor = s_stShipColors[m_iOwnerID];
		if(m_bFreeFromBullets && (((int)m_dFreeFromBulletsTimer) % 300 > 200)) stColor.a = 0.5f; //transparent

		//draw thrust
		if(m_dAcceleration>0) ODraw(s_pclShipImg, &s_stShipRects[m_iRealDir + 36], iMapX, iMapY);
		//draw ship
		m_pclGraph->SetBltColor(&stColor);
		ODraw(s_pclShipImg, &s_stShipRects[m_iRealDir], iMapX, iMapY);

		m_pclGraph->SetBltColor(NULL); //restore
	} else {
		DrawExplosion(iMapX, iMapY);
	}
}

void C_ShipPlayer::DrawMiniShip(int x, int y, int dim)
{
	S_Rect stDstRect = {x,y,dim,dim};
	S_FColor stColor = s_stShipColors[m_iOwnerID];
	//draw ship
	m_pclGraph->SetBltColor(&stColor);
	m_pclGraph->Blt(s_pclShipImg, &s_stShipRects[27], &stDstRect, true);

	m_pclGraph->SetBltColor(NULL); //restore
}

void C_ShipPlayer::ResetGravity()
{
	m_dGravityX = C_Environment::GetShipGravityBaseX();
	m_dGravityY = C_Environment::GetShipGravityBaseY() + (m_iSpaceUsed*2);
}

void C_ShipPlayer::Stop()
{
	m_dVelX = 0;
	m_dVelY = 0;
	if(m_dAcceleration && C_Sound::GetSoundObject()) {
		C_ShipPlayer::s_clFadeOutList.push_back(m_pEngineSoundCh);
		//C_Sound::GetSoundObject()->Stop(m_pEngineSoundCh); //stop engine sound
	}
	m_dAcceleration = 0.0;
}

C_Bullet *C_ShipPlayer::CreateBullet()
{
	S_BulletInfo stBulletInfo;
	double dTemp = m_iRealDir*(m_iDirStep*(PI/180));
	double dSin = sin(dTemp);
	double dCos = cos(dTemp);
	stBulletInfo.dPosX = m_dPosX + s_iShipSize/2 + dCos * (s_iShipSize/2-2);
	stBulletInfo.dPosY = m_dPosY + s_iShipSize/2 + dSin * (s_iShipSize/2-2);
	stBulletInfo.dVelX = m_dVelX + dCos * BULLETBASEVEL;
	stBulletInfo.dVelY = m_dVelY + dSin * BULLETBASEVEL;
	stBulletInfo.dDirection = m_iRealDir * m_iDirStep;

	return new C_Bullet(m_pclMap, &stBulletInfo, m_iOwnerID);
}

void C_ShipPlayer::ConsumeFuel(double i_dTime)
{
	if(m_pclMap->GetType()!=MAP_RACE) {
		//consume fuel
		if(m_dAcceleration!=0) m_dFuel-=i_dTime/1000.0;
		if(m_dFuel<0) {
			m_dFuel=0;
			//Kill(); //no thrust possible instead... it usually leads to death
		}
	}
}

void C_ShipPlayer::AddFuel(double dFuel)
{
	if(m_dFuel<m_dMaxFuelInTank) {
		m_dFuel += dFuel; 
		if(m_dFuel>m_dMaxFuelInTank) m_dFuel=m_dMaxFuelInTank;
	}
}

void C_ShipPlayer::AddLife()
{
	if(m_iNumLifes!=-1) m_iNumLifes++;
}

////////////////////////////////////////////////////////////////////////////////

C_Player::C_Player(C_Map *i_pclMap, C_BulletList *i_pclBulletList, char *i_szName, int i_iPlayerID, int i_iStartX, int i_iStartY, C_NetMsg *i_pclNetMsg, C_NetMsgServer *i_pclNetMsgServer, /*AI->*/C_Viewport *i_pclParent)
	: C_ShipPlayer(i_pclMap, i_szName, i_iPlayerID, i_iStartX, i_iStartY, i_pclNetMsg)
{
	m_pclNetMsgServer = i_pclNetMsgServer;
	m_pclBulletList = i_pclBulletList;
	m_pclControl = C_Control::GetControlObject();

	m_dLandTime = 0.0;

	m_stMsg1.iMsg = 0; //no message is to be sent yet
	m_dLastSendDataTime = 0.0;

	m_bFireAllowed = true;
	m_iKilledReason = -1;

//AI////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
	m_pclParent = i_pclParent;
	if(m_pclParent) {
		m_dAIUpdateSTime = 0;
		//init AI
		S_StaticShipInfo stSShipInfo;
		stSShipInfo.iID = m_iOwnerID;
		stSShipInfo.dSteerSpeed  = C_Environment::GetShipSteerSpeed();
		stSShipInfo.dThrust      = C_Environment::GetShipThrust();
		stSShipInfo.stGravity.x  = C_Environment::GetShipGravityBaseX();
		stSShipInfo.stGravity.y  = C_Environment::GetShipGravityBaseY();
		stSShipInfo.stResistance.x = C_Environment::GetShipResistanceX();
		stSShipInfo.stResistance.y = C_Environment::GetShipResistanceY();
		m_pclAI = new C_AISteering(&stSShipInfo);
		m_pclFont = new C_Font(_RGBA8888);
		m_pclFont->Load((char*)"arial12", C_Global::szResourceFile);

		//map init
		int x, y, w, h, i;
		m_pclMap->GetMapSize(&w, &h);
		i = m_pclMap->GetTileSize();
		w/=i; h/=i; 
		C_Grid<char> *pclGrid = new C_Grid<char>(w, h);
		for(y=0; y<h; y++) {
			for(x=0; x<w; x++) {
				pclGrid->aaGrid[y][x] = m_pclMap->GetElementAbsolute(x, y);
			}
		}
		m_pclAI->SetMap(pclGrid);
		delete pclGrid;
	}
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
}

C_Player::~C_Player()
{
//AI////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
	if(m_pclParent) {
		delete m_pclAI;
		delete m_pclFont;
	}
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
}

void C_Player::Draw()
{
	((C_ShipPlayer*)this)->Draw();
}

bool C_Player::GetInput(double i_dSTime)
{ 
	bool bThrottle = false, bLeft = false, bRight = false, bFire = false, bAdjust = false;
	if(!m_pclAI) {
		bThrottle = m_pclControl->IsPressed(DIK_UP)!=0 || m_pclControl->IsPressed(DIK_W)!=0;
		bLeft     = m_pclControl->IsPressed(DIK_LEFT)!=0 || m_pclControl->IsPressed(DIK_A)!=0;
		bRight    = m_pclControl->IsPressed(DIK_RIGHT)!=0 || m_pclControl->IsPressed(DIK_D)!=0;
		bFire     = m_pclControl->IsPressedOnce(DIK_RETURN)!=0 || m_pclControl->IsPressedOnce(DIK_SPACE)!=0;
		bAdjust   = m_pclControl->IsPressed(DIK_DOWN)!=0 || m_pclControl->IsPressed(DIK_S)!=0;
	} else {
//AI////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
		//move
		m_dAIUpdateSTime += i_dSTime;
		if(m_dAIUpdateSTime>0.30) {
			m_dAIUpdateSTime -= 0.30;

			if(m_pclMap->GetType()==MAP_RACE) { //seek checkpoints
				POINTD p;
				POINT pi;
				C_PointList *pclPoints = new C_PointList();
				C_CheckPointVector *pclCheckpointList = m_pclParent->GetCheckPoints();
				int i, iSize = (int)pclCheckpointList->size();

				S_CheckPointInfo stCP;
				//ship position
				GetCenterPos(&pi); //default
				//nearest point on current wp line
				{
					POINT l0, l1;
					POINTD l0d, l1d, posd;
					if(m_pclAI->GetCurrentLineSegment(&l0, &l1)) {
						l0d.x = l0.x; l0d.y = l0.y;
						l1d.x = l1.x; l1d.y = l1.y;
						posd.x = pi.x; posd.y = pi.y;
						double dDist = PathUtils::PointDistanceToLineSeg(posd, l0d, l1d, &posd);
						if(dDist<60) { //take ship pos when ship has died and been moved
							pi.x = (int)posd.x; pi.y = (int)posd.y;
						}
					}
				}
				pclPoints->push_back(pi);

				for(i=0; i<2; i++) {
					(*pclCheckpointList)[(m_iCurCP+i)%iSize]->GetCheckPointInfo(&stCP);
					//PathUtils::MidPoint(stCP.iX1, stCP.iY1, stCP.iX2, stCP.iY2, &p.x, &p.y);
					VECTOR l0; l0.x = stCP.iX1; l0.y = stCP.iY1;
					VECTOR l1; l1.x = stCP.iX2; l1.y = stCP.iY2;
					POINTD pstart; pstart.x = pi.x; pstart.y = pi.y;
					PathUtils::PointDistanceToLineSeg(pstart, l0, l1, &p);

					VECTOR delta = l0 - l1;
					PathUtils::Normalize(delta);

					int loop = 0;
					int count_p = 0;
					int count_n = 0;
					int num_stuck;
					bool break_found = false;
#define TCHECK 12
#define TCLEAR 7
					while(1) {
						num_stuck = 0;

						POINTD pos_p = p + delta*(TCHECK*loop);
						POINTD pos_n = p - delta*(TCHECK*loop);
						loop++;

						double dDist = PathUtils::PointDistanceToLineSeg(pos_p, l0, l1, NULL);
						if(dDist>1.0) {
							num_stuck++;
						} else {
							int tile = m_pclMap->GetElementAbsolute((int)pos_p.x/32, (int)pos_p.y/32);
							if(tile == 0) {
								count_p++;
							} else {
								break_found = true;
								count_p = 0;
							}
						}
						
						dDist = PathUtils::PointDistanceToLineSeg(pos_n, l0, l1, NULL);
						if(dDist>1.0) {
							num_stuck++;
						} else {
							int tile = m_pclMap->GetElementAbsolute((int)pos_n.x/32, (int)pos_n.y/32);
							if(tile == 0) {
								count_n++;
							} else {
								break_found = true;
								count_n = 0;
							}
						}

						if(count_p-1 >= TCLEAR) {
							p = pos_p - delta*((TCHECK*(count_p-1))/2.0);
							break;
						}
						if(count_n-1 >= TCLEAR) {
							p = pos_n + delta*((TCHECK*(count_n-1))/2.0);
							break;
						}
#if 1
						if(!break_found) {
							if((count_p+count_n-2) >= TCLEAR) {
								p = pos_p - delta*(TCHECK*(count_p+count_n-2)/2.0);
								break;
							}
						}
#endif

						if(num_stuck==2) {
							break; //should never happen, something is wrong with the race map
						}

					}
					pi.x = (int)p.x;
					pi.y = (int)p.y;
					pclPoints->push_back(pi);
				}
				m_pclAI->SetWayPoints(pclPoints, false);
				//^can fail, but always deletes point list
				//delete pclPoints; //shall not be done here
			} else { //dogfight or mission, seek enemies
				double dDist, dBestDist = 1000000;
				POINT pi;
				C_PointList *pclPoints = new C_PointList();
				GetCenterPos(&pi);
				pclPoints->push_back(pi);
				for(int i=0; i<8; i++) {
					if(i==m_iOwnerID) continue;
					C_Object *pclEnemy = m_pclParent->GetShipPlayerObject(i);
					if(pclEnemy) {
						int x, y;
						pclEnemy->GetPos(&x, &y);
						dDist = PathUtils::PointDistance(m_dPosX, m_dPosY, x, y);
						if(dDist<dBestDist) {
							dBestDist = dDist;
							pclEnemy->GetCenterPos(&pi);
						}
					}
				}
				pclPoints->push_back(pi);
				m_pclAI->SetWayPoints(pclPoints, false);
				//^can fail, but always deletes point list
				//delete pclPoints; //shall not be done here
			}
		}

		S_AIOutput stMove;
		S_ShipInfo stShipInfo;
		stShipInfo.dTimeStep = i_dSTime;
		stShipInfo.stPosition.x = m_dPosX+s_iShipSize/2;
		stShipInfo.stPosition.y = m_dPosY+s_iShipSize/2;
		stShipInfo.stVelocity.x = m_dVelX;
		stShipInfo.stVelocity.y = m_dVelY;
		stShipInfo.dVelocity = sqrt(m_dVelX*m_dVelX+m_dVelY*m_dVelY); //verify!
		stShipInfo.dHeading = PathUtils::CalcLineAngle(0, 0, m_dVelX, m_dVelY); //verify!
		stShipInfo.dPointing = m_dDirection;
		m_pclAI->GetOutput(&stShipInfo, &stMove);
		bThrottle = stMove.bThrottle;
		bLeft     = stMove.bLeft;
		bRight    = stMove.bRight;
		bFire     = stMove.bFire;
		bAdjust   = false;
	}
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

	bool bResult = false; //no bullet fired

	double dTemp = (m_dFuel!=0 && bThrottle) ? C_Environment::GetShipThrust():0;
	if((dTemp!=m_dAcceleration) && C_Sound::GetSoundObject()) {
		if(dTemp!=0) m_pEngineSoundCh = C_Sound::GetSoundObject()->Play(2, m_fVol, m_fPan, true);
		else {
			C_ShipPlayer::s_clFadeOutList.push_back(m_pEngineSoundCh);
			//C_Sound::GetSoundObject()->Stop(m_pEngineSoundCh);
		}
	}
	m_dAcceleration = dTemp;

	//lifted from landingzone?
	if(m_dAcceleration && m_bLanded) {
		m_bLanded = false;
		ResetGravity();
	}

	if(!m_bLanded) { //cannot stear ship if landed
		if(bAdjust) {
			if(m_dDirection!=270.0) {
				//get way to turn
				int iSign = 1, iSignAfter = 1;
				if(m_dDirection<90 || m_dDirection>270) iSign = -1;
				m_dDirection += iSign*(C_Environment::GetShipSteerSpeed()/3*i_dSTime);
				if(m_dDirection<90 || m_dDirection>270) iSignAfter = -1;
				if(iSign!=iSignAfter) m_dDirection = 270.0;
			}
		} else { //normal steering
			m_dDirection +=bRight ? C_Environment::GetShipSteerSpeed()*i_dSTime:0;
			m_dDirection -=bLeft  ? C_Environment::GetShipSteerSpeed()*i_dSTime:0;
		}
		//clipping
		int iNumCounts = 0;
		if(m_dDirection<0) iNumCounts = (int)(m_dDirection/360.0)-1;
		else iNumCounts = (int)(m_dDirection/360.0);
		m_dDirection-=iNumCounts*360;
	}

	if(m_bFireAllowed && bFire)
		bResult = true;

	int iKeyFlag =
		(bRight ? GF_KEY_RIGHT:0) | 
		(bLeft  ? GF_KEY_LEFT:0) | 
		((m_dFuel!=0 && bThrottle) ? GF_KEY_ACCELERATE:0) |
		(bAdjust ? GF_KEY_DOWN:0);

	if(m_stMsg1.iKeyFlag != iKeyFlag) m_stMsg1.iMsg |= N_MSG1_KEYCHANGE;
	m_stMsg1.iKeyFlag = (uint8_t)iKeyFlag;

	return bResult;
}

void C_Player::LandOnZone(double i_dAdjustedPosY)
{
	m_dPosY = i_dAdjustedPosY;
	m_dDirection = 270.0;

	m_dLandTime = 0.0;
	m_bLanded = true;
	m_dGravityX = m_dGravityY = 0;

	if(C_Sound::GetSoundObject()) C_Sound::GetSoundObject()->Play(3, m_fVol, m_fPan); //play landing sound

	//msg to network
	m_stMsg1.iMsg |= N_MSG1_LANDED;

	Stop();
}

void C_Player::BounceOnZone(double i_dAdjustedPosY)
{
	m_dPosY = i_dAdjustedPosY;
	m_dDirection = 270.0;

	if(this==s_pclListner) if(C_Sound::GetSoundObject()) C_Sound::GetSoundObject()->Play(7, m_fVol, m_fPan); //play bouncing sound

	m_dVelY = -m_dVelY*0.65; //invert ymovement and damp it
	m_dVelX *= 0.85; //damp xmovement

	m_stMsg1.iMsg |= N_MSG1_MOVEMENT;
}

void C_Player::AddScore(int i_iPoints)
{
	m_iScore += i_iPoints;

	//msg to network
	m_stMsg1.iMsg |= N_MSG1_SCORE;
	m_stMsg1.iScore = m_iScore;
}

void C_Player::Update(double i_dTime)
{
	bool   bBullet = false;
	double dSTime = (i_dTime/1000);

	m_dLandTime += i_dTime;

	if(m_bFreeFromBullets) {
		m_dFreeFromBulletsTimer += i_dTime;
		if(m_dFreeFromBulletsTimer > BULLETFREETIME) {
			m_bFreeFromBullets = false;
		}
	}

	//maybe come to life again
	if(!m_bAlive) {
		if(UpdateExplosion(i_dTime)) {
			if(m_iNumLifes != 0) {
				//set alive again
				m_bAlive = true;
				m_bCanCollide = true;
				m_bFreeFromBullets = true;
				m_dFreeFromBulletsTimer = 0.0;
				m_dPosX = m_iStartPosX;
				m_dPosY = m_iStartPosY;
				m_dDirection = 270.0;
				m_dGravityX  = C_Environment::GetShipGravityBaseX();
				m_dGravityY  = C_Environment::GetShipGravityBaseY();
				m_bLanded = false;
				m_dFuel = m_dMaxFuelInTank;
				m_stMsg1.iMsg |= N_MSG1_MOVEMENT; //must send message when come to life
			}
		}
	}

	if(m_bAlive) bBullet = GetInput(dSTime);
	ConsumeFuel(i_dTime);
	OUpdate(i_dTime);

	CalcSoundVolPan(s_pclListner, &m_fVol, &m_fPan);

	if(bBullet) {
		C_Bullet *pclBullet = CreateBullet();

		s_iBulletCounter++;
		int iUId = m_iOwnerID | (s_iBulletCounter << 8);
		pclBullet->SetUId(iUId);
		m_stMsg1.iMsg |= N_MSG1_NEWBULLET;
		m_stMsg1.iNumBullets = 1;
		m_stMsg1.iBulletUniqeValBase = iUId;
		m_pclBulletList->Add(pclBullet);
		if(C_Sound::GetSoundObject()) C_Sound::GetSoundObject()->Play(0, m_fVol, m_fPan); //play bullet sound
	}

	m_pstImageRect = &s_stShipRects[m_iRealDir];

	//set new map position
	if(s_pclListner==this) {
		int iPosX = (int)m_dPosX - (m_stScreenArea.width/2 - s_iShipSize/2);
		int iPosY = (int)m_dPosY - (m_stScreenArea.height/2 - s_iShipSize/2);
		m_pclMap->MoveAbsolute(iPosX, iPosY);
	}
}

bool C_Player::CheckMapCollision()
{
	if(!m_bAlive) return false;

	bool bResult = m_pclMap->CheckCollision(this);
	if(bResult) Kill(KILLEDBY_SELF, 0);

	return bResult;
}

void C_Player::Kill(int i_iReason, int i_iID)
{
	if(!m_bAlive) return;

	bool bSend = true;
	if(i_iID == -1) bSend = false;

	if(i_iReason == KILLEDBY_BULLET) m_iKilledReason = i_iID;
	else m_iKilledReason = -3;

	m_bAlive = false;
	m_bCanCollide = false;
	UpdateExplosion(0.0); //reset explosion

	//to prevent movement during explosion
	m_dGravityX = 0;
	m_dGravityY = 0;

	Stop();
	if(i_iReason == KILLEDBY_SELF) {
		AddScore(-10);
	}
	if(C_Sound::GetSoundObject()) C_Sound::GetSoundObject()->Play(1, m_fVol, m_fPan); //play explosion sound

	//respawn cargo on landingzone taken from
	int i;
	for(i=0; i<m_iNumUsed; i++) {
		//send to server, respawns on reply
		S_NetMsg2 stMsg2;
		
		stMsg2.iPlayerId = m_iOwnerID;
		stMsg2.iType = N_MSG2_CARGORESPAWN;
		stMsg2.iCargoWeight = m_aiHold[i];
		//stMsg2.iCargoNum = ;
		stMsg2.iZoneId = m_aiHoldZoneId[i];
		m_pclNetMsgServer->SendMsg2(&stMsg2);
	}
	//reset cargo
	m_iNumUsed = m_iSpaceUsed = 0;
	//take a life if not unlimited
	if(m_iNumLifes!=-1) m_iNumLifes--;

	if(bSend) {
		m_stMsg1.iMsg |= N_MSG1_KILLED;
		m_stMsg1.iCauseID = i_iID;
		m_stMsg1.iKilledReason = (uint8_t)i_iReason;
	}
}

int C_Player::WasKilled()
{
	int iRet = m_iKilledReason;
	m_iKilledReason = -1;
	return iRet;
}

bool C_Player::SpaceInHold(int iCargo)
{
	return (m_iNumUsed<MAXSPACEINHOLDUNITS && (m_iSpaceUsed+iCargo)<=MAXSPACEINHOLDWEIGHT);
}
bool C_Player::LoadCargo(int iCargo, int iFromZone)
{
	bool bCargoLoaded = false;

	if(SpaceInHold(iCargo)) {
		//space in hold, add cargo
		m_aiHoldZoneId[m_iNumUsed] = iFromZone;
		m_aiHold[m_iNumUsed++]     = iCargo;
		m_iSpaceUsed += iCargo;
		bCargoLoaded = true;

		//calc new gravity is done when lifted from the zone, not here

		//play load cargo sound
		if(this==s_pclListner) if(C_Sound::GetSoundObject()) C_Sound::GetSoundObject()->Play(8, m_fVol, m_fPan);
	}

	return bCargoLoaded;
}
void C_Player::UnloadCargo()
{
	if(m_iNumUsed>0) {
		m_iNumUsed--;
		m_iSpaceUsed -= m_aiHold[m_iNumUsed];

		//calc new gravity is done when lifted from the zone, not here

		//add score for the cargo moved
		AddScore(m_aiHold[m_iNumUsed]);

		//play unload cargo sound
		if(this==s_pclListner) if(C_Sound::GetSoundObject()) C_Sound::GetSoundObject()->Play(9, m_fVol, m_fPan);

		//reset landtime just to make the delay before a level ends the same every time
		m_dLandTime = 0;
	}
}

void C_Player::SetShootingMode(bool i_bFireAllowed)
{
	m_bFireAllowed = i_bFireAllowed;
}

//position(etc), if hit by bullet(whos, what), if collided with other player,
//if collided with wall, or new shot fired
void C_Player::SendData(double i_dTime)
{
	SetCommonInfo(&m_stMsg1);
	m_stMsg1.iCargoInHold = (uint16_t)m_iSpaceUsed;

	m_dLastSendDataTime += i_dTime;

	if(m_stMsg1.iMsg || m_dLastSendDataTime>500) {
		m_stMsg1.iMsg |= N_MSG1_MOVEMENT; //always send movement info
		m_stMsg1.iMsg |= N_MSG1_KEYCHANGE;//always send key info

		m_stMsg1.iMsg |= N_MSG1_LAPINFO;  //always send lap info since the player does not know when it is changed
		m_stMsg1.iLap = (uint8_t)m_iCurLap;
		m_stMsg1.iCP  = (uint8_t)m_iCurCP;
		m_stMsg1.fBestLapTime = (float)m_dBestLapTime;

		m_pclNetMsg->SendMsg1(&m_stMsg1);
		m_stMsg1.iMsg = 0; //reset
		m_dLastSendDataTime = 0;
	}
}

////////////////////////////////////////////////////////////////////////////////

C_Enemy::C_Enemy(C_Map *i_pclMap, C_BulletList *i_pclBulletList, char *i_szName, int i_iPlayerID, int i_iStartX, int i_iStartY, C_NetMsg *i_pclNetMsg)
	: C_ShipPlayer(i_pclMap, i_szName, i_iPlayerID, i_iStartX, i_iStartY, i_pclNetMsg)
{
	m_pclBulletList = i_pclBulletList;

	m_iKeyFlag = 0;
}

C_Enemy::~C_Enemy()
{
}

void C_Enemy::LandOnGround()
{
	m_bLanded = true;
	m_dGravityX = 0;
	m_dGravityY = 0;
	if(C_Sound::GetSoundObject()) C_Sound::GetSoundObject()->Play(3, m_fVol, m_fPan); //play landing sound
}

int C_Enemy::Update(double i_dTime)
{
	double dSTime = (i_dTime/1000);
	int    iResult = -1;

	m_dLandTime += i_dTime;

	if(m_bFreeFromBullets) {
		m_dFreeFromBulletsTimer += i_dTime;
		if(m_dFreeFromBulletsTimer > BULLETFREETIME) {
			m_bFreeFromBullets = false;
		}
	}

	//maybe come to life again
	if(!m_bAlive) {
		if(UpdateExplosion(i_dTime)) { //set alive again
			//^because this can happen before it happens on the other players side
			//the ship can die again if there is something on the explosion site.
			if(m_iNumLifes != 0) {
				m_bAlive = true;
				//m_bCanCollide = true;
				//^fixes the bug above, (but we still need to set m_bCanCollide=false in Kill())
				m_bFreeFromBullets = true;
				m_dFreeFromBulletsTimer = 0.0;
				m_dPosX = m_iStartPosX;
				m_dPosY = m_iStartPosY;
				m_dDirection = 270.0;
				m_dGravityX  = C_Environment::GetShipGravityBaseX();
				m_dGravityY  = C_Environment::GetShipGravityBaseY();
				m_dFuel = m_dMaxFuelInTank;
				m_bLanded = false;
			}
		}
	}

	//always
	bool bGotOne = false;
	while(m_pclNetMsg->GetMsg1(m_iOwnerID, &m_stMsg1)) {
		//movement info
		if(m_stMsg1.iMsg & N_MSG1_MOVEMENT) {
			if(m_stMsg1.fAcceleration!=m_dAcceleration && C_Sound::GetSoundObject()) {
				if(m_dAcceleration>0) {
					C_ShipPlayer::s_clFadeOutList.push_back(m_pEngineSoundCh);
					//C_Sound::GetSoundObject()->Stop(m_pEngineSoundCh);
				}
				else m_pEngineSoundCh = C_Sound::GetSoundObject()->Play(2, m_fVol, m_fPan, true);
			}

			GetCommonInfo(&m_stMsg1);
			m_iSpaceUsed = m_stMsg1.iCargoInHold;
		}
		//keystate change
		if(m_stMsg1.iMsg & N_MSG1_KEYCHANGE) m_iKeyFlag = m_stMsg1.iKeyFlag;

		//bullet fired
		if(m_stMsg1.iMsg & N_MSG1_NEWBULLET) {
			C_Bullet *pclBullet = CreateBullet();

			pclBullet->SetUId(m_stMsg1.iBulletUniqeValBase);
			m_pclBulletList->Add(pclBullet);
			if(C_Sound::GetSoundObject()) C_Sound::GetSoundObject()->Play(0, m_fVol, m_fPan); //play bullet sound
		}

		//player killed
		if(m_stMsg1.iMsg & N_MSG1_KILLED) {
			switch(m_stMsg1.iKilledReason) {
				case KILLEDBY_SELF:
					Kill(KILLEDBY_SELF);
					break;
				case KILLEDBY_BULLET:
					Kill(KILLEDBY_BULLET);
					m_pclBulletList->Remove(m_stMsg1.iCauseID);
					iResult = (m_stMsg1.iCauseID & 0xff);
					break;
				case KILLEDBY_PLAYER:
					iResult = m_stMsg1.iCauseID + 256;
					Kill(KILLEDBY_PLAYER);
					break;
			}
		}

		//score change
		if(m_stMsg1.iMsg & N_MSG1_SCORE) m_iScore = m_stMsg1.iScore;

		//player landed
		if(m_stMsg1.iMsg & N_MSG1_LANDED) LandOnGround();

		if(m_stMsg1.iMsg & N_MSG1_LAPINFO) {
			m_iCurLap = m_stMsg1.iLap;
			m_iCurCP  = m_stMsg1.iCP;
			m_dBestLapTime = m_stMsg1.fBestLapTime;
		}
      
		m_dTimeWithoutInfo = 0;
		m_bPlayerLost = false;
		bGotOne = true;
	}
	if(!bGotOne) {
		m_dTimeWithoutInfo += dSTime;
		if(m_dTimeWithoutInfo > 10) {
			//players info is not received for 10 sec, firewall problem?
			m_bPlayerLost = true;
		}
	}

	//make the explosion be still even if messages lags behind
	if(!m_bAlive) {
		m_dVelX = 0.0;
		m_dVelY = 0.0;
	}

	//lifted from landingzone?
	if(m_dAcceleration && m_bLanded) {
		m_bLanded = false;
		ResetGravity();
	}
 
	//steering
	{
		if(m_iKeyFlag & GF_KEY_DOWN) {
			if(m_dDirection!=270.0) {
				//get way to turn
				int iSign = 1, iSignAfter = 1;
				if(m_dDirection<90 || m_dDirection>270) iSign = -1;
				m_dDirection += iSign*(C_Environment::GetShipSteerSpeed()/3*dSTime);
				if(m_dDirection<90 || m_dDirection>270) iSignAfter = -1;
				if(iSign!=iSignAfter) m_dDirection = 270.0;
			}
		} else { //normal steering
			m_dDirection +=(m_iKeyFlag & GF_KEY_RIGHT) ? C_Environment::GetShipSteerSpeed()*dSTime:0;
			m_dDirection -=(m_iKeyFlag & GF_KEY_LEFT) ? C_Environment::GetShipSteerSpeed()*dSTime:0;
		}
		//clipping
		int iNumCounts = 0;
		if(m_dDirection<0) iNumCounts = (int)(m_dDirection/360.0)-1;
		else iNumCounts = (int)(m_dDirection/360.0);
		m_dDirection-=iNumCounts*360;
	}

	ConsumeFuel(i_dTime);
	OUpdate(i_dTime);

	CalcSoundVolPan(s_pclListner, &m_fVol, &m_fPan);

	if(m_dAcceleration!=0 && C_Sound::GetSoundObject()) { //ship is moving, set current engine volume
		C_Sound::GetSoundObject()->SetVolume(m_pEngineSoundCh, m_fVol);
		C_Sound::GetSoundObject()->SetPanning(m_pEngineSoundCh, m_fPan);
	}

	m_pstImageRect = &s_stShipRects[m_iRealDir];
	return iResult;
}

void C_Enemy::Kill(int i_iReason)
{
	if(!m_bAlive) return;

	m_bAlive = false;
	m_bCanCollide = false;
	UpdateExplosion(0.0); //reset explosion

	//to prevent movement during explosion
	m_dGravityX = 0; 
	m_dGravityY = 0;

	//take a life if not unlimited
	if(m_iNumLifes!=-1) m_iNumLifes--;

	Stop();
	//if(i_iReason == KILLEDBY_SELF) m_iScore--;
	if(C_Sound::GetSoundObject()) C_Sound::GetSoundObject()->Play(1, m_fVol, m_fPan); //play explosion sound
}
