// C_PlayerBase, C_ShipPlayer, C_Player, C_Enemy

//temporary code added for ai player in C_Player, this must be rewritten if it shall be used

#ifndef _PLAYER_H
#define _PLAYER_H

#include "netmsg.h"
#include "netmsgserver.h"

#include "common.h"
#include "tilemap.h"
#include "object.h"
#include "bullet.h"
#include "input.h"
#include "glgraph.h"
#include "font.h"
#include "environment_defaults.h"

#define KILLEDBY_SELF   0x01 //dec score
#define KILLEDBY_BULLET 0x02 //inc other player's score
#define KILLEDBY_PLAYER 0x03 //do nothing to any score

#define MAXSPACEINHOLDWEIGHT 50
#define MAXSPACEINHOLDUNITS  3

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
class C_AISteering;
class C_Viewport;
class C_CheckPoint;
typedef std::vector<C_CheckPoint*> C_CheckPointVector;
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class C_PlayerBase : public C_Object
{
public:
	C_PlayerBase();
	~C_PlayerBase();

	void CalcSoundVolPan(C_Object *i_pclListner, float *o_fVol, float *o_fPan);
	bool UpdateExplosion(double i_dTime);
	void DrawExplosion(int i_iX, int i_iY);
	static void SetListener(C_Object *i_pclListner) {s_pclListner = i_pclListner;};
	void InitFromObject(C_Object *i_pclSrcObject);
protected:
	void SetCommonInfo(S_NetMsg1 *o_pstMsg1);
	void GetCommonInfo(S_NetMsg1 *i_pstMsg1);

	static int s_iBulletCounter;
	static C_Object *s_pclListner;
	float m_fVol, m_fPan;

	bool m_bAlive;
private:
	//explosion gfx
	static C_Image *s_pclExplosionImg;
	static S_Rect s_stExplosionRects[16];
	double m_dFrameNr;
	C_Timer *m_pclExpTimer;

	static int s_iObjCount;
};

class C_ShipPlayer :public C_PlayerBase
{
public:
	C_ShipPlayer(C_Map *i_pclMap, char *i_szName, int i_iPlayerID, int i_iStartX, int i_iStartY, C_NetMsg *i_pclNetMsg);
	~C_ShipPlayer();
	void Draw();
	void DrawMiniShip(int x, int y, int dim); //screen coords

	int GetScore() {return m_iScore;};
	void SetScore(int i_iScore) {m_iScore = i_iScore;};

	bool FreeFromBullets() {return m_bFreeFromBullets;};

	void Stop();
	bool LandedOnGround() {return m_bLanded;};

	bool IsPlayerLost() {return m_bPlayerLost;};

	//race variables
	int m_iCurCP;
	int m_iCurLap;
	double m_dBestLapTime;

	//dogfight and mission
	int m_iNumLifes;

	//player name
	char m_szName[MAXPLAYERNAMELENGTH];

	int CargoWeightInHold() {return m_iSpaceUsed;};
	int CargoNumInHold() {return m_iNumUsed;};

	void AddFuel(double dFuel);
	double FuelPercentLeft() {return m_dFuel/m_dMaxFuelInTank;};

	void AddLife();

	static void UpdateFadeOut(double i_dTime);

protected:
	//create bullet method
	C_Bullet *CreateBullet();

	//player startposition
	int m_iStartPosX, m_iStartPosY;
	//screen position calculation
	C_Map *m_pclMap;
	int   m_iMapWidth, m_iMapHeight; //total map limit

	//ship gfx, and blink control
	static int s_iShipSize;
	static C_Image *s_pclShipImg;
	static S_Rect s_stShipRects[72];
	static S_FColor s_stShipColors[8];
	double m_dFreeFromBulletsTimer;
	bool m_bFreeFromBullets;

	//network
	C_NetMsg *m_pclNetMsg;
	S_NetMsg1 m_stMsg1;

	//score variables
	int m_iScore;

	//fuel
	void ConsumeFuel(double i_dTime);
	double m_dMaxFuelInTank;
	double m_dFuel; //number of seconds one can use the thruster

	//other
	bool   m_bLanded;
	double m_dLandTime;
	//sound
	void *m_pEngineSoundCh;

	//cargo variables
	void ResetGravity();
	int m_iSpaceUsed;
	int m_iNumUsed;
	int m_aiHold[MAXSPACEINHOLDUNITS];
	int m_aiHoldZoneId[MAXSPACEINHOLDUNITS];

	//detect network trouble (C_Enemy only)
	bool   m_bPlayerLost;
	double m_dTimeWithoutInfo;

	static std::vector<void*> s_clFadeOutList;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
	C_Font *m_pclFont; //NULL for normal player
	C_AISteering *m_pclAI; //NULL for normal player
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

private:
	static int s_iObjCount;
};


class C_Player :public C_ShipPlayer
{
public: 
	C_Player(C_Map *i_pclMap, C_BulletList *i_pclBulletList, char *i_szName, int i_iPlayerID,
		int i_iStartX, int i_iStartY, C_NetMsg *i_pclNetMsg, C_NetMsgServer *i_pclNetMsgServer,
		/*AI->*/C_Viewport *i_pclParent);

	~C_Player();
	bool CheckMapCollision();

	void Draw();

	void Update(double i_dTime);
	void Kill(int i_iReason, int i_iID);
	int WasKilled(); //returns the reason for the any kill that has happened or -1, must be checked every frame
	void AddScore(int i_iPoints);
	void LandOnZone(double i_dAdjustedPosY);
	void BounceOnZone(double i_dAdjustedPosY);

	bool SpaceInHold(int iCargo);
	bool LoadCargo(int iCargo, int iFromZone);
	void UnloadCargo();

	void SendData(double i_dTime);

	void SetShootingMode(bool i_bFireAllowed);

	bool IsAlive() {return m_bAlive;};
private:
	//bullets
	C_BulletList *m_pclBulletList;

	double m_dLastSendDataTime;

	//to get inputs
	bool GetInput(double i_dSTime); //true if bullet shall fire
	C_Control  *m_pclControl;       //input from keyboard

	bool m_bFireAllowed;
	int m_iKilledReason;

	C_NetMsgServer *m_pclNetMsgServer;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
	C_Viewport *m_pclParent; //NULL for normal player
	double m_dAIUpdateSTime;
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
};

class C_Enemy :public C_ShipPlayer
{
public: 
	C_Enemy(C_Map *i_pclMap, C_BulletList *i_pclBulletList, char *i_szName, int i_iPlayerID, int i_iStartX, int i_iStartY, C_NetMsg *i_pclNetMsg);
	~C_Enemy();

	int Update(double i_dTime);
	void Kill(int i_iReason);
	void LandOnGround();

private:
	//to get inputs
	int m_iKeyFlag;

	//bullets
	C_BulletList *m_pclBulletList;
};

#endif
