// C_Replay

#ifndef REPLAY_H
#define REPLAY_H

#include "netmsgserver.h"
#include "common.h"
#include <vector>

#pragma pack(1)
struct S_ReplayHeader
{
	int32_t iVersion; //1
	char szMapFile[48];
	int32_t iLevelType;
	bool bNoShooting, bFlyThrough, bNoBulletsHit;
	bool bReserved;

	int32_t iNumPlayers;
	char szPlayerName[8][MAXPLAYERNAMELENGTH];
};

struct S_ReplayAction
{
	int32_t iTimeSlot;

	int8_t  iId;  //player id (-1 for server), also in the messages below
	uint8_t iMsg; //1 or 2

	union {
		S_NetMsg1 stMsg1;
		S_NetMsg2 stMsg2;
	};
};
#pragma pack()

typedef std::vector<S_ReplayAction> C_ActionVector;

class C_Replay
{
public: 
	C_Replay();
	C_Replay(C_Replay *pclReplay);
	~C_Replay();
	void Reset(); //clears all but replay header
	void ResetBeforePlay();
	void SetGameInfo(S_ReplayHeader *i_pstHeader);
	void GetGameInfo(S_ReplayHeader *o_pstHeader);
	void IncTimeSlot();                    //used both in save and replay
	bool Add(S_ReplayAction *i_pstAction); //must be added in timeslot order (uses m_iCurTimeSlot as timeslot), save only
	bool Get(S_ReplayAction *o_pstAction); //returns the first action not yet returned, if any (based on m_iCurTimeSlot), replay only
	bool LoadFromMem(char *i_pMem, int i_iSize);
	bool SaveToMem(char *o_pMem, int *io_iSize);
	bool Load(char *i_szFile);
	bool Save(char *i_szFile);

private:
	static bool VerifyString(char *i_szText, int i_iSize);
	bool VerifyHeader();
	bool VerifyElement(S_ReplayAction *i_pstAction);
	int m_iCurTimeSlot;
	int m_iCurGetPos;
	C_ActionVector *m_pclActions;
	S_ReplayHeader m_stHeader;
};

#endif
