// C_Settings

//static class to load/keep some settings
//used both by the game and server

#include "settings.h"
#include "global.h"

//default if no cfg file is found
#define SERVER_PORT     1999 //tcp connect

S_JoystickMapping C_Settings::stPadMapping = {
	DIK_LEFT,     DIK_RIGHT,  DIK_UP,   DIK_DOWN,
	{ DIK_RETURN, DIK_RETURN, DIK_UP,   DIK_DOWN,
	DIK_UP,       DIK_RETURN, DIK_DOWN, DIK_RETURN,
	DIK_UP,       DIK_UP }
};

int  C_Settings::iServerPort;
int  C_Settings::iFullscreen;
int  C_Settings::iPadEnabled;
int  C_Settings::iPadDeviceNr;
double C_Settings::dMasterVolume;

//set the given control (index) to the given game effect
void C_Settings::FillPadMapping(int i_iIndex, char *i_szInGame)
{
	int iDIKCode = 0; //default no map
	if(strcmp(i_szInGame, "RIGHT")==0)       iDIKCode = DIK_RIGHT;
	else if(strcmp(i_szInGame, "LEFT")==0)   iDIKCode = DIK_LEFT;
	else if(strcmp(i_szInGame, "THRUST")==0) iDIKCode = DIK_UP;
	else if(strcmp(i_szInGame, "DOWN")==0)   iDIKCode = DIK_DOWN;
	else if(strcmp(i_szInGame, "FIRE")==0)   iDIKCode = DIK_RETURN;

	switch(i_iIndex) {
		case 0: stPadMapping.iRIGHT = iDIKCode; break;
		case 1: stPadMapping.iLEFT  = iDIKCode; break;
		case 2: stPadMapping.iUP    = iDIKCode; break;
		case 3: stPadMapping.iDOWN  = iDIKCode; break;
		default:
			i_iIndex-=4;
			if(i_iIndex>=0 && i_iIndex<MAX_BUTTONS) {
				stPadMapping.iB[i_iIndex] = iDIKCode;
			}
			break;
	}
}

//reads cfg file, or sets default if not found
void C_Settings::Init(char *i_szSettingsFile)
{
	FILE   *pFile;
	char   szPath[MAX_PATH];
	char   szLine[512], *pszStr;
	char   szCommand[128];
	char   szTemp[128];

	//default settings
	dMasterVolume  = 1.0;
	iServerPort    = SERVER_PORT;
	iFullscreen    = 0;
	iPadEnabled    = 1;
	iPadDeviceNr   = 0;
	//load settings from disk
	sprintf(szPath, "%s/%s", C_Global::szDataPath, i_szSettingsFile);
	pFile = fopen(szPath, "rt");
	if(!pFile) return;
	do {
		pszStr = C_Global::GetNextLineAndCommand(pFile, szLine, szCommand);

		if(strcmp(szCommand, "*SERVERPORT") == 0) {
			sscanf(szLine, "%s %d", szCommand, &iServerPort);
			continue;
		}

		if(strcmp(szCommand, "*FULLSCREEN") == 0) {
			sscanf(szLine, "%s %d", szCommand, &iFullscreen);
			continue;
		}

		if(strcmp(szCommand, "*MASTERVOLUME") == 0) {
			sscanf(szLine, "%s %lf", szCommand, &dMasterVolume);
			continue;
		}

		if(strcmp(szCommand, "*PAD_ENABLE") == 0) {
			sscanf(szLine, "%s %d", szCommand, &iPadEnabled);
			continue;
		}

		if(strcmp(szCommand, "*PAD_DEVICE") == 0) {
			//only used in linux to be able to set the pad device path
			sscanf(szLine, "%s %d", szCommand, &iPadDeviceNr);
			continue;
		}

		//game pad mappings
		if(strcmp(szCommand, "*PAD_RIGHT") == 0) {
			sscanf(szLine, "%s %s", szCommand, szTemp);
			FillPadMapping(0, szTemp);
			continue;
		}
		if(strcmp(szCommand, "*PAD_LEFT") == 0) {
			sscanf(szLine, "%s %s", szCommand, szTemp);
			FillPadMapping(1, szTemp);
			continue;
		}
		if(strcmp(szCommand, "*PAD_UP") == 0) {
			sscanf(szLine, "%s %s", szCommand, szTemp);
			FillPadMapping(2, szTemp);
			continue;
		}
		if(strcmp(szCommand, "*PAD_DOWN") == 0) {
			sscanf(szLine, "%s %s", szCommand, szTemp);
			FillPadMapping(3, szTemp);
			continue;
		}
		if(strncmp(szCommand, "*PAD_BUTTON", 11) == 0) {
			int iRet, iBNum;
			iRet = sscanf(&szCommand[11], "%d", &iBNum);
			iRet += sscanf(szLine, "%s %s", szCommand, szTemp);
			if(iRet==3) {
				FillPadMapping(3+iBNum, szTemp);
			}
			continue;
		}
	} while(pszStr);

	fclose(pFile);
}
