// C_Map

#ifndef TILEMAP_H
#define TILEMAP_H

#include "object.h"
#include "common.h"
#include "image.h"
#include "glgraph.h"
#include <vector>

//fwd decl to avoid recursive things if including "mapobjects.h" that would be required otherwise
struct S_ExtraObjects;

using std::vector;

typedef enum
{
	MAP_MISSION = 1,
	MAP_DOGFIGHT = 2,
	MAP_RACE = 3,
	MAP_MISSION_COOP = 4
} MapType;

struct S_MapHeader
{
	char szGridfile[128];
	char szTileset[128];
	int iTileSize;
	int iWidth, iHeight;
};

class C_Map
{
public:
	C_Map(char *i_szFilename, S_Rect *i_pstDrawarea, S_ExtraObjects *io_pclExtra, bool *o_bResult); //create a map from .txt and .map file and loads extra objects
	~C_Map();

	//common methods
	void SetMapWrapping(bool i_bHorizontal, bool i_bVertical);
	void MoveRelative(int i_iX, int i_iY); //pixel map coordinates
	void MoveAbsolute(int i_iX, int i_iY); //pixel map coordinates
	void GetPos(int *o_iX, int *o_iY) {*o_iX = m_iLastX; *o_iY = m_iLastY;};

	int GetTileSize() {return m_iTileSize;}; //returns the size of the tiles in tileset
	void SetScreenRect(S_Rect *i_pstDrawarea);
	S_Rect *GetScreenRect() {return &m_stDrawarea;};
	void GetMapSize(int *o_iWidth, int *o_iHeight) {*o_iWidth = m_iWidth*m_iTileSize; *o_iHeight = m_iHeight*m_iTileSize;};

	bool IsValidPos(int i_iMapTileX, int i_iMapTileY);
	void SetElementAbsolute(int i_iMapTileX, int i_iMapTileY, char i_cVal);
	int GetElementAbsolute(int i_iMapTileX, int i_iMapTileY);
	bool Load(char *i_szMapHeader, S_ExtraObjects *io_pclExtra);
	void Draw();           //uses draw area passed in constructor as position and max size

	C_Image *GetMiniMap(int i_iMiniTileSize); //renders a mini image of the map

	C_Image* GetTilesetImage() { return m_pclTilesImg; }

	MapType GetType() const {return m_iLevelType;};
	int NumRaceLaps() const {return m_iRaceLaps;};
	int GetMaxPlayers() const {return m_iMaxPlayers;};
	int GetMinPlayers() const {return m_iMinPlayers;};
	int GetRaceLaps() const {return m_iRaceLaps;};
	bool HasDoors() const {return m_iNumDoors!=0;};

	//game specific methods
	bool CheckCollision(C_Object *i_pclObject);
	//returns true if the object has collided with nonblack pixels,
	// or if the object is outside the map

	bool GetPlayerStartingPos(int i_iPlayer, POINT *o_stPos); //0..7 player map coordinates

	void Partition(int numPartsPerAxis, vector<vector<bool> >& subtiles);
	void DrawOn(C_Image* image, const S_Rect* range);

private:
	//map grid info
	bool m_bMapLoaded;
	void FreeGrid();
	char **m_pGrid;
	int m_iWidth, m_iHeight;

	//draw info
	S_Rect m_stDrawarea;
	int m_iDrawareaWidthSave, m_iDrawareaHeightSave;
	int m_iPixMaxX, m_iPixMaxY;
	int m_iScrollMaxX, m_iScrollMaxY;
	int m_iLastX, m_iLastY;
	bool m_iWrappX, m_iWrappY;
	bool WrapAroundTest(int *o_iX, int *o_iY, int i_iWidth, int i_iHeight);

	void ProcessTile(C_Image* tileImg, int tileNumber, int tileSize, int xStart, int yStart, int nParts, vector<vector<bool> >& subtiles);

	//tileset info
	bool LoadTileSet(char *i_szTileset);
	int       m_iTileSize;
	C_Image   *m_pclTilesImg, *m_pclTilesImg_Sysmem;
	int       m_iNumTiles;
	S_Rect    *m_stTileRects;

	//graph
	C_GraphWrapperGL *m_pclGraph;

	//file info
	S_MapHeader m_stHeader;
	MapType     m_iLevelType;
	int         m_iRaceLaps;
	int         m_iMaxPlayers, m_iMinPlayers;
	int         m_iNumDoors;
	POINT       m_stPlayerStartPos[8];
};

#endif
