// C_Viewport

//class to handle an active game or replay.
//keeps all map objects and the map itself and is responsible
// for drawing everything visible when a game is active.
//depending on map type, different paths are taken to control the game.

#include "viewport.h"
#include "fileresource.h"
#include "global.h"
#include <math.h>

//keeps text positions and color
struct S_StatusPos
{
	int iX, iY;
	S_FColor stColor1, stColor2;
};

//coordinates set here are not used, they are recalculated in constructor to make it work with any resolution
S_Rect      MAPRECT             = {0, 0, 864, 632};
S_StatusPos MAINPLAYERSTATUSPOS = {4, 632-50, {0x09/255.0f, 0x6f/255.0f, 0xcc/255.0f, 1}, {0xcc/255.0f, 0x0a/255.0f, 0x2e/255.0f, 1}};
S_StatusPos STATUSPOS[8] = { //based on !race
	{120+2,     632-84, {0xAD/255.0f,0x16/255.0f,0x16/255.0f,1}, {0x4A/255.0f,0x0E/255.0f,0x0E/255.0f,1}},
	{120+2+228, 632-84, {0x46/255.0f,0xB3/255.0f,0x9E/255.0f,1}, {0x20/255.0f,0x69/255.0f,0x60/255.0f,1}},
	{120+2,     632-64, {0x16/255.0f,0xAE/255.0f,0x16/255.0f,1}, {0x0E/255.0f,0x4B/255.0f,0x0E/255.0f,1}},
	{120+2+228, 632-64, {0x19/255.0f,0x4F/255.0f,0xBF/255.0f,1}, {0x0F/255.0f,0x26/255.0f,0x4C/255.0f,1}},
	{120+2,     632-44, {0xC4/255.0f,0xAF/255.0f,0x46/255.0f,1}, {0x65/255.0f,0x63/255.0f,0x22/255.0f,1}},
	{120+2+228, 632-44, {0x9E/255.0f,0x16/255.0f,0x98/255.0f,1}, {0x4B/255.0f,0x0E/255.0f,0x3C/255.0f,1}},
	{120+2,     632-24, {0xCD/255.0f,0xCD/255.0f,0xCD/255.0f,1}, {0x79/255.0f,0x79/255.0f,0x79/255.0f,1}},
	{120+2+228, 632-24, {0x93/255.0f,0x9F/255.0f,0x16/255.0f,1}, {0x3C/255.0f,0x4B/255.0f,0x0E/255.0f,1}}
};
S_FColor g_stColorWhite = S_FCOLORMAKE(1, 1, 1, 1);

//initializes a game
C_Viewport::C_Viewport(char *i_szMapFile, int i_iPlayerIndex, char *i_szPlayerName, bool i_bMaster, C_NetMsg *i_pclNetMsg, C_NetMsgServer *i_pclNetMsgServer)
{
	int i;
	m_pclControl = C_Control::GetControlObject();
	m_pclGraph = C_GraphWrapperGL::GetGraphWrapperObject();
	m_pclSoundFX = C_Sound::GetSoundObject();

	m_bMaster = i_bMaster;
	m_pclNetMsgServer = i_pclNetMsgServer;
	m_pclNetMsg = i_pclNetMsg;
	m_pclNetMsg->Reset();

	m_dCargoEndTimer = 0;

	//fps counter init
	m_bShowFPS = false;
	m_pclFont = new C_Font(_RGBA8888);
	m_pclFont->Load((char*)"arial14", C_Global::szResourceFile);
	m_pclFont->SetColor(g_stColorWhite);

	m_dFPSTime = 0;
	m_iFPSCount = m_iFPS = 0;

	//end condition init
	m_iKillsLeft = -1;

	//bullet init
	m_pclBulletList = new C_BulletList();

	//map init
	S_ExtraObjects stExtra;
	stExtra.pclBulletList        = m_pclBulletList;
	stExtra.pclLandingZoneList   = &m_clLandingZoneList;
	stExtra.pclCheckPointList    = &m_clCheckPointList;
	stExtra.pclComputerEnemyList = &m_clComputerEnemyList;
	stExtra.pclDoorList          = &m_clDoorList;
	stExtra.bMaster              = m_bMaster;
	stExtra.pclNetMsg            = m_pclNetMsg;
	stExtra.pclNetMsgServer      = m_pclNetMsgServer;
	bool bLoaded;
	m_pclMap = new C_Map(i_szMapFile, &MAPRECT, &stExtra, &bLoaded);

	//player init
	POINT stStart;
	m_pclPlayer = NULL;
	m_iPlayerIndex = i_iPlayerIndex;
	C_PlayerBase::SetListener(NULL);
	if(m_iPlayerIndex!=-1) {
		m_pclMap->GetPlayerStartingPos(m_iPlayerIndex, &stStart);
		m_pclPlayer = new C_Player(m_pclMap, m_pclBulletList, i_szPlayerName, m_iPlayerIndex, stStart.x, stStart.y, m_pclNetMsg, m_pclNetMsgServer, NULL);
		C_ComputerEnemy::AddFireRangeObject((C_Object**)&m_pclPlayer);
		C_PlayerBase::SetListener((C_Object*)m_pclPlayer);
	}
	m_iReplayFollowPlayer = 0;
	m_iReplayOldCP = 0;

	//other players init
	for(i=0; i<8; i++) {
		m_pclEnemy[i] = NULL;
		m_abIsAI[i] = false;
		m_pclAIPlayer[i] = NULL;
		//race time
		m_bTimeCounting[i] = false;
		m_dLastNewLapTime[i] = 0.0;
		m_dTotalTime[i] = 0.0;
	}

	m_iTotalLaps = -1;
	if(IsRace()) {
		m_iTotalLaps = m_pclMap->NumRaceLaps();
		m_pclCargoFuelBar = new C_Image((char*)"laps_time.tga", C_Global::szResourceFile);
	} else {
		m_pclCargoFuelBar = new C_Image((char*)"cargo_fuel.tga", C_Global::szResourceFile);
	}

	m_bGameFinished = false;

	m_pclChat = new C_Chat();

	//minimap
	m_bShowMiniMap = true;
	m_pclMiniMap = m_pclMap->GetMiniMap(1);

	m_bPlayerPlayerCollision = true;
	m_bPlayerBulletsCollision = true;

	m_dTimePlayed = 0.0;
	m_bNextLevelSent = false;

	//set positions in viewport
	int w = m_pclGraph->GetModeWidth();
	int h = m_pclGraph->GetModeHeight();
	MAINPLAYERSTATUSPOS.iX = 4;
	MAINPLAYERSTATUSPOS.iY = h-50;
	for(i=0; i<8; i++) {
		STATUSPOS[i].iX = 120+2+((i%2)?228:0);
		STATUSPOS[i].iY = h-(84-(i/2)*20);
	}
	MAPRECT.x = MAPRECT.y = 0;
	MAPRECT.width = w; MAPRECT.height = h;
	SetSize(w, h);
}

//destroys a game
C_Viewport::~C_Viewport()
{
	int i, iSize;

	delete m_pclMiniMap;
	delete m_pclChat;

	//fps counter
	delete m_pclFont;

	//map
	delete m_pclMap;
	//enemy
	for(i=0; i<8; i++) {
		delete m_pclEnemy[i];
		delete m_pclAIPlayer[i];
	}
	//player
	delete m_pclPlayer;
	//bullet
	delete m_pclBulletList;

	iSize = (int)m_clLandingZoneList.size();
	for(i=0; i<iSize; i++) {
		delete m_clLandingZoneList[i];
	}
	iSize = (int)m_clCheckPointList.size();
	for(i=0; i<iSize; i++) {
		delete m_clCheckPointList[i];
	}
	iSize = (int)m_clComputerEnemyList.size();
	for(i=0; i<iSize; i++) {
		delete m_clComputerEnemyList[i];
	}
	iSize = (int)m_clDoorList.size();
	for(i=0; i<iSize; i++) {
		delete m_clDoorList[i];
	}

	delete m_pclCargoFuelBar;
}

//sets the player to follow during a replay
// also detects when to play some sounds not otherwise played
void C_Viewport::ReplayFollow(bool i_bNext)
{
	int i, iSize;
	int iOldPlayer = m_iReplayFollowPlayer;
	if(!i_bNext && !m_pclEnemy[m_iReplayFollowPlayer]) i_bNext = true;
	if(i_bNext) {
		for(i=0; i<8; i++) {
			m_iReplayFollowPlayer++;
			if(m_iReplayFollowPlayer>7) m_iReplayFollowPlayer = 0;
			if(m_pclEnemy[m_iReplayFollowPlayer]) break;
		}
	}

	C_ShipPlayer *pclEnemy = m_pclEnemy[m_iReplayFollowPlayer];

	C_PlayerBase::SetListener((C_Object*)pclEnemy);
	if(pclEnemy) {
		int iX, iY, iW, iH;
		pclEnemy->GetPos(&iX, &iY);
		pclEnemy->GetSize(&iW, &iH);
		int iPosX = iX - (MAPRECT.width/2 - iW/2);
		int iPosY = iY - (MAPRECT.height/2 - iH/2);
		m_pclMap->MoveAbsolute(iPosX, iPosY);

		iSize = (int)m_clCheckPointList.size();
		if(iSize) {
			for(i=0; i<iSize; i++) {
				if(i!=pclEnemy->m_iCurCP) m_clCheckPointList[i]->SetBlinkState(false);
			}
			if(pclEnemy->m_iCurLap < m_iTotalLaps)
				m_clCheckPointList[pclEnemy->m_iCurCP]->SetBlinkState(true);
			if(m_pclSoundFX && iOldPlayer == m_iReplayFollowPlayer) {
				if(m_iReplayOldCP!=pclEnemy->m_iCurCP) {
					if(pclEnemy->m_iCurLap < m_iTotalLaps) m_pclSoundFX->Play(5);
					else {
						m_pclSoundFX->Play(6);
						m_bTimeCounting[m_iReplayFollowPlayer] = false; //stop time
					}
				}
			}
			m_iReplayOldCP = pclEnemy->m_iCurCP;
		}
	}
}

//changes the draw area
// the whole window size is now used for drawing
void C_Viewport::SetSize(int i_iWidth, int i_iHeight)
{
	int i, w, h;
	S_Rect stNewRect = {0, 0, i_iWidth, i_iHeight};
	m_pclMap->GetMapSize(&w, &h);
	if(i_iWidth>w) {
		stNewRect.x = (i_iWidth-w)/2;
		stNewRect.width = w;
	}
	if(i_iHeight>h) {
		stNewRect.y = (i_iHeight-h)/2;
		stNewRect.height = h;
	}
	MAINPLAYERSTATUSPOS.iX = stNewRect.x+4;
	MAINPLAYERSTATUSPOS.iY = stNewRect.y+stNewRect.height-50;
	for(i=0; i<8; i++) {
		STATUSPOS[i].iX = stNewRect.x+(STATUSPOS[i].iX-MAPRECT.x);
		STATUSPOS[i].iY = stNewRect.y+(STATUSPOS[i].iY-MAPRECT.height-MAPRECT.y)+stNewRect.height;
	}
	MAPRECT = stNewRect;
	m_pclMap->SetScreenRect(&MAPRECT);
}

//add an opponent network ship player (should be done before the game is started)
void C_Viewport::AddEnemy(int i_iPlayerIndex, char *i_szPlayerName, bool i_bIsAI)
{
	delete m_pclEnemy[i_iPlayerIndex];
	POINT stStart;
	m_abIsAI[i_iPlayerIndex] = i_bIsAI;
	m_pclMap->GetPlayerStartingPos(i_iPlayerIndex, &stStart);
	m_pclEnemy[i_iPlayerIndex] = new C_Enemy(m_pclMap, m_pclBulletList, i_szPlayerName, i_iPlayerIndex, stStart.x, stStart.y, m_pclNetMsg);

	C_ComputerEnemy::AddFireRangeObject((C_Object**)&m_pclEnemy[i_iPlayerIndex]);
}

//add an opponent ai ship player to the game (should be done before the game is started)
void C_Viewport::AddAI(int i_iPlayerIndex, char *i_szPlayerName)
{
	delete m_pclAIPlayer[i_iPlayerIndex];
	POINT stStart;
	m_pclMap->GetPlayerStartingPos(i_iPlayerIndex, &stStart);
	m_pclAIPlayer[i_iPlayerIndex] = new C_Player(m_pclMap, m_pclBulletList, i_szPlayerName, i_iPlayerIndex, stStart.x, stStart.y, m_pclNetMsg, m_pclNetMsgServer, this);

	C_ComputerEnemy::AddFireRangeObject((C_Object**)&m_pclAIPlayer[i_iPlayerIndex]);
}

//deletes an opponent enemy from the game (if a playes is lost/leaves the game)
void C_Viewport::RemoveEnemy(int i_iPlayerIndex)
{
	delete m_pclEnemy[i_iPlayerIndex];
	m_pclEnemy[i_iPlayerIndex] = NULL;
	m_abIsAI[i_iPlayerIndex] = false;
}

//returns a playes race time (in ms)
// - if it is a race map
// - if the map is finished
int C_Viewport::GetRaceTime(int i_iPlayerIndex)
{
	if(m_clCheckPointList.size()>0 && !m_bTimeCounting[i_iPlayerIndex]) {
		return (int)m_dTotalTime[i_iPlayerIndex];
	}

	return -1;
}

//returns a players score (*1000)
// - if it is not a race map
bool C_Viewport::GetScore(int i_iPlayerIndex, int *o_iScore)
{
	//returned score is *1000 to represent decimals
	C_ShipPlayer *pclPlayer = GetShipPlayerObject(i_iPlayerIndex);
	if(pclPlayer) {
		int iLifes;
		GetNumLifes(i_iPlayerIndex, &iLifes);
		if(iLifes<0) iLifes = 0;
		*o_iScore = (pclPlayer->GetScore()+iLifes*50)*1000;
		if(IsMission()) {
			if(MapType()==MAP_MISSION_COOP) {
				int iMaxMapPlayers = m_pclMap->GetMaxPlayers();
				int iLifesLost = 0;
				*o_iScore = 0;
				for(int i=0; i<8; i++) {
					C_ShipPlayer *pclPlayer2 = GetShipPlayerObject(i);
					if(pclPlayer2) {
						GetNumLifes(i, &iLifes);
						if(iLifes<0) iLifes = 0;
						iLifesLost += (5-iLifes);
						*o_iScore += pclPlayer2->GetScore();
					}
				}
				*o_iScore += (iMaxMapPlayers-iLifesLost)*50*1000;
			}
			*o_iScore -= (int)(m_dTimePlayed/2); //long time is bad in mission
		}
		return true;
	}
	return false;
}

//set a players score
// intended for if scores are kept between levels (not used)
void C_Viewport::SetScore(int i_iPlayerIndex, int i_iScore)
{
	//score is *1000 to represent decimals, but all decimals are currently lost
	//this method is not used any more, it was used earlier when levels were played in sequence
	C_ShipPlayer *pclPlayer = GetShipPlayerObject(i_iPlayerIndex);
	if(pclPlayer) pclPlayer->SetScore(i_iScore/1000);
}

//returns the number of lives for a player (will be -1 for race)
bool C_Viewport::GetNumLifes(int i_iPlayerIndex, int *o_iLifes)
{
	C_ShipPlayer *pclPlayer = GetShipPlayerObject(i_iPlayerIndex);
	if(pclPlayer) {
		*o_iLifes = pclPlayer->m_iNumLifes;
		return true;
	}
	return false;
}

//set the number of lifes for all players
// intended to be used before a level starts
void C_Viewport::SetNumLifes(int i_iLifes)
{
	for(int i=0; i<8; i++) {
		C_ShipPlayer *pclPlayer = GetShipPlayerObject(i);
		if(pclPlayer) pclPlayer->m_iNumLifes = i_iLifes;
	}
}

//set the number of lifes for one player
// intended for if lifes are kept between levels (not used)
void C_Viewport::SetNumLifes(int i_iPlayerIndex, int i_iLifes)
{
	C_ShipPlayer *pclPlayer = GetShipPlayerObject(i_iPlayerIndex);
	if(pclPlayer) pclPlayer->m_iNumLifes = i_iLifes;
}

//checks if all cargo is delivered
bool C_Viewport::CargoEndCondition(double dTime)
{
	//check all zones
	int iCargoInGame = 0;
	int i, iSize = (int)m_clLandingZoneList.size();
	for(i=0; i<iSize; i++) {
		iCargoInGame += m_clLandingZoneList[i]->GetTotalCargo();
	}
	if(iCargoInGame>0) return false;

	//check all players
	for(i=0; i<8; i++) {
		C_ShipPlayer *pclPlayer = GetShipPlayerObject(i);
		if(pclPlayer && pclPlayer->CargoWeightInHold()) return false;
	}

	m_dCargoEndTimer += dTime;

	if(m_dCargoEndTimer>2000) {
		m_dCargoEndTimer = 0;
		return true;
	}
	return false;
}

//the main update loop
// - updates all time based movements for all objects
// - updates all time based actions for all objects (checkpoints visited for example)
// - checks collissions for all objects
// - send resulted network messages to the server
// - update the fps counter
// - updates the replay follow player
int C_Viewport::Update(double dTime)
{
	S_NetMsg2 stMsg2;
	stMsg2.iPlayerId = -1;
	C_ComputerEnemy *pclCompEnemy; //a temp var
	int j, i, iSize, iKillerID, iBulletUID, iResult = -1;

	m_dTimePlayed += dTime;

	//update fade out (removes click/pop in looping sounds that are stopped)
	C_ShipPlayer::UpdateFadeOut(dTime);

	//move, update objects

	//doors
	iSize = (int)m_clDoorList.size();
	for(i=0; i<iSize; i++) {
		m_clDoorList[i]->Update(dTime);
	}

	//player and AI
	for(i=0; i<8; i++) {
		C_Player *pclPlayer = GetPlayerObject(i);
		if(pclPlayer) {
			pclPlayer->Update(dTime);
			for(j=0; j<iSize; j++) {
				if(m_clDoorList[j]->CheckCollission(pclPlayer, false)) {
					pclPlayer->Kill(KILLEDBY_SELF, 0);
				}
			}
		}
	}

	//other players (info from network)
	for(i=0; i<8; i++) {
		if(m_pclEnemy[i]) {
			iKillerID = m_pclEnemy[i]->Update(dTime); 
			if(iKillerID != -1) {
				//killed, need to update killer
				if(iKillerID >= 256) { //killed by collision with other player
					iKillerID -= 256;

					//killer is also killed, no network messages are sent (the other games do this also)
					C_Player *pclKiller = GetPlayerObject(iKillerID);
					if(pclKiller) pclKiller->Kill(KILLEDBY_PLAYER, -1);
					else if(iKillerID<COMPUTER_ENEMY_BASEID) m_pclEnemy[iKillerID]->Kill(KILLEDBY_PLAYER);
				} else { //killed by bullet from other player
					if(!IsMission()) {
						C_Player *pclKiller = GetPlayerObject(iKillerID);
						if(pclKiller) pclKiller->AddScore(25);
						//score for enemy players is handled in their game and sent
					}
					//no score added for killing other players in mission
				}
			}
		}
	}

	//landing zones
	iSize = (int)m_clLandingZoneList.size();
	for(i=0; i<iSize; i++) {
		m_clLandingZoneList[i]->UpdateLanding(m_pclPlayer, dTime);
		for(j=0; j<8; j++) {
			if(m_pclAIPlayer[j]) m_clLandingZoneList[i]->UpdateLanding(m_pclAIPlayer[j], dTime);
		}
		for(j=0; j<8; j++) {
			C_ShipPlayer *pclShipPlayer = GetShipPlayerObject(j, M_ENEMY);
			if(pclShipPlayer) m_clLandingZoneList[i]->UpdateFuel(pclShipPlayer, dTime);
		}
	}

	//send updated info as soon as possible
	for(int i=0; i<8; i++) {
		C_Player *pclPlayer = GetPlayerObject(i);
		if(pclPlayer) pclPlayer->SendData(dTime);
	}

	//computer enemies
	iSize = (int)m_clComputerEnemyList.size();
	for(i=0; i<iSize; i++) {
		pclCompEnemy = m_clComputerEnemyList[i];
		iKillerID = pclCompEnemy->Update(dTime);
		C_Player *pclKiller = GetPlayerObject(iKillerID);
		if(pclKiller) pclKiller->AddScore(pclCompEnemy->GetScoreValue());
		else if(iKillerID == -2) { //explosion ready after death; delete enemy
			delete pclCompEnemy;
			m_clComputerEnemyList.erase(m_clComputerEnemyList.begin()+i);
			iSize--; i--;
		}
	}

	//bullets
	m_pclBulletList->Update(dTime);
	iSize = (int)m_clDoorList.size();
	C_Bullet *pclBullet = m_pclBulletList->GetNext(true);
	while(pclBullet) {
		for(j=0; j<iSize; j++) {
			if(m_clDoorList[j]->CheckCollission(pclBullet, pclBullet->GetOwnerID()<8)) { //only player bullets can open doors for now
				//delete bullet, no need to send over network since this is run on their side too
				m_pclBulletList->Remove(pclBullet->GetUId());
				break;
			}
		}
		pclBullet = m_pclBulletList->GetNext(false);
	}

	//checkpoints
	iSize = (int)m_clCheckPointList.size();
	if(iSize>0) { //there are checkpoints (racemap)
		for(i=0; i<iSize; i++) {
			m_clCheckPointList[i]->Update(dTime);
		}

		for(j=0; j<8; j++) {
			C_ShipPlayer *pclShipPlayer = GetShipPlayerObject(j);
			if(pclShipPlayer) {
				for(i=0; i<iSize; i++) {
					//start counting if any checkpoint is passed to prevent clearing the map from enemies by going backwards
					if(m_dTotalTime[j] == 0) {
						if(m_clCheckPointList[i]->PlayerAtCP(pclShipPlayer)) m_bTimeCounting[j] = true;
					}
				}

				if(m_bTimeCounting[j]) m_dTotalTime[j] += dTime;
			}

			C_Player *pclPlayer = GetPlayerObject(j);
			if(pclPlayer) {
				i = pclPlayer->m_iCurCP;
				if(m_clCheckPointList[i]->PlayerAtCP(pclPlayer)) {
					if(j==m_iPlayerIndex) m_clCheckPointList[i]->SetBlinkState(false);
					if(m_bTimeCounting[j]) {
						if(++i == iSize) { //new lap
							i = 0;
							pclPlayer->m_iCurLap++;
							//send
							stMsg2.iType = N_MSG2_NUMLAPS;
							stMsg2.iPlayerId = (int16_t)pclPlayer->GetOwnerID();
							stMsg2.iNumLaps = pclPlayer->m_iCurLap;
							m_pclNetMsgServer->SendMsg2(&stMsg2);

							//finished?
							if(pclPlayer->m_iCurLap == m_iTotalLaps) {
								pclPlayer->m_iCurCP = 0; //reset
								m_bTimeCounting[j] = false;
								if(j==m_iPlayerIndex && m_pclSoundFX) m_pclSoundFX->Play(6);
							}

							//best lap time calculation
							double dThisLapTime = m_dTotalTime[j]-m_dLastNewLapTime[j];
							if(dThisLapTime<pclPlayer->m_dBestLapTime) {
								pclPlayer->m_dBestLapTime = dThisLapTime;
							}
							m_dLastNewLapTime[j] = m_dTotalTime[j];
						}
						if(m_bTimeCounting[j]) {
							if(j==m_iPlayerIndex && m_pclSoundFX) m_pclSoundFX->Play(5);
							pclPlayer->m_iCurCP = i;
							if(j==m_iPlayerIndex) m_clCheckPointList[i]->SetBlinkState(true);
						}
					}
				}
			}
		}
	}

	//check for player collisions (AI and normal player)
	for(j=0; j<8; j++) {
		C_Player *pclPlayer = GetPlayerObject(j);

		if(pclPlayer) {
			//player -> map collision
			if(pclPlayer->CheckMapCollision()) {
				//player KILLEDBY_SELF
			}
			//player -> bullets collision
			if(!pclPlayer->FreeFromBullets()) {
				
				if(m_pclBulletList->CheckCollision(pclPlayer, &iBulletUID, m_bPlayerBulletsCollision)) {
					//player KILLEDBY_BULLET (iBulletUID)
					iKillerID = iBulletUID & 255;
					pclPlayer->Kill(KILLEDBY_BULLET, iBulletUID);
					if(iKillerID >= COMPUTER_ENEMY_BASEID) {
						pclPlayer->AddScore(-10);
					}
					//add score to killer?
					if(iKillerID<8) {
						C_Player *pclKiller = GetPlayerObject(iKillerID);
						if(pclKiller) pclKiller->AddScore(25);
					}
				}
				//player -> computer enemy collision
				iSize = (int)m_clComputerEnemyList.size();
				for(i=0; i<iSize; i++) {
					pclCompEnemy = m_clComputerEnemyList[i];
					if(pclCompEnemy->OCheckCollision(pclPlayer)) {
						pclPlayer->Kill(KILLEDBY_SELF, 0);
					}
				}
			}
			//player <-> player collision
			if(m_bPlayerPlayerCollision) {
				for(i=0; i<8; i++) {
					if(i==j) continue;
					C_ShipPlayer *pclOtherPlayer = GetShipPlayerObject(i);
					if(pclOtherPlayer) {
						C_Player *pclOtherPlayerOrAI = GetPlayerObject(i);
						//player KILLEDBY_PLAYER (i) (collision)
						if(pclOtherPlayer->OCheckCollision(pclPlayer)) {
							if(pclOtherPlayerOrAI) pclOtherPlayerOrAI->Kill(KILLEDBY_PLAYER, -1);
							else if(m_pclEnemy[i]) m_pclEnemy[i]->Kill(KILLEDBY_PLAYER);
							pclPlayer->Kill(KILLEDBY_PLAYER, i);
						}
					}
				}
			}
		}
	}

	//send resulted network messages to the server
	//mission update (for server)
	if(IsMission() && CargoEndCondition(dTime)) {
		iResult = -2;
		if(!m_bNextLevelSent) {
			stMsg2.iType = N_MSG2_NEXTLEVEL;
			m_pclNetMsgServer->SendMsg2(&stMsg2);
			m_bNextLevelSent = true;
		}
	}

	//player killed update (for server)
	if(!IsRace()) {
		stMsg2.iType = N_MSG2_NUMKILLS;
		for(i=0; i<8; i++) {
			C_Player *pclPlayer = GetPlayerObject(i);
			if(pclPlayer) {
				int iKilled = pclPlayer->WasKilled();
				if(iKilled!=-1) {
					stMsg2.iPlayerId = (int16_t)pclPlayer->GetOwnerID();
					stMsg2.iKilledBy = iKilled;
					m_pclNetMsgServer->SendMsg2(&stMsg2);
				}
			}
		}
	}

	if(m_pclControl->IsPressedOnce(DIK_I) && !InEnteringTextMode()) m_bShowMiniMap = !m_bShowMiniMap;

	//update fps counter
	//toggle fps counter on/off
	if(m_pclControl->IsPressedOnce(DIK_F) && !InEnteringTextMode()) m_bShowFPS = !m_bShowFPS;
	if(m_bShowFPS) {
		m_dFPSTime += dTime;
		//m_iFPSCount is increased in Draw()
		if(m_dFPSTime >= 500) {
			m_iFPS = (int)((m_iFPSCount*1000) / m_dFPSTime);
			m_iFPSCount = 0;
			m_dFPSTime = 0;
		}
		sprintf(m_szFPSString, "%d", m_iFPS);
	}

	bool bChat = m_pclChat->Update(dTime);
	if(bChat) {
		char szText[64];
		sprintf(szText, "%s: %s", m_pclPlayer->m_szName, m_pclChat->GetEnteredText());
		//send to server
		m_pclNetMsgServer->SendChatMsg(szText);
	}

	//replay
	if(!m_pclPlayer) {
		ReplayFollow(m_pclControl->IsPressedOnce(DIK_V) && !InEnteringTextMode());
	}

	return iResult;
}

//draws all objects and status info
void C_Viewport::Draw()
{
	int i, iSize;

	SetSize(m_pclGraph->GetModeWidth(), m_pclGraph->GetModeHeight());

	if(m_bShowFPS) {
		m_iFPSCount++;
	}

	//draw map
	m_pclMap->Draw();

	//draw doors
	iSize = (int)m_clDoorList.size();
	for(i=0; i<iSize; i++) {
		m_clDoorList[i]->Draw();
	}

	//draw landing zones
	iSize = (int)m_clLandingZoneList.size();
	for(i=0; i<iSize; i++) {
		m_clLandingZoneList[i]->Draw();
	}

	//draw checkpoints
	iSize = (int)m_clCheckPointList.size();
	if(iSize>0) {
		m_pclGraph->SetClippingRect(&MAPRECT); //now in opengl the text must be clipped
		for(i=0; i<iSize; i++) {
			m_clCheckPointList[i]->Draw();
		}
		m_pclGraph->SetClippingRect(NULL); //now in opengl the text must be clipped
	}

	//draw computer enemies
	iSize = (int)m_clComputerEnemyList.size();
	for(i=0; i<iSize; i++) {
		m_clComputerEnemyList[i]->Draw();
	}

	//draw other players
	for(i=0; i<8; i++) {
		if(m_pclEnemy[i]) {
			m_pclEnemy[i]->Draw();
		}
		if(m_pclAIPlayer[i]) {
			m_pclAIPlayer[i]->Draw();
		}
	}

	//draw bullets
	m_pclBulletList->Draw();

	//draw player
	if(m_pclPlayer) m_pclPlayer->Draw();

	//draw fps counter?
	if(m_bShowFPS) {
		m_pclFont->SetColor(g_stColorWhite);
		m_pclFont->Print(MAPRECT.x+4, MAPRECT.y+8, m_szFPSString);
	}

	DrawAllPlayersStatus();
	if(m_pclPlayer) DrawAdditionalPlayerStatus(m_iPlayerIndex);
	else DrawAdditionalPlayerStatus(m_iReplayFollowPlayer);
	if(m_bShowMiniMap) DrawMinimap(1.5f);

	m_pclChat->Draw(MAPRECT.x, MAPRECT.y);
}

int PLAYERSPOS[8][8] = { //[numplayers][player_index] used in STATUSPOS
	{6,0,},
	{6,7,0,},
	{4,6,7,0},
	{4,5,6,7,0},
	{2,4,5,6,7,0,},
	{2,3,4,5,6,7,0,},
	{0,2,3,4,5,6,7,0},
	{0,1,2,3,4,5,6,7},
};
//draw common info for all players
void C_Viewport::DrawAllPlayersStatus()
{
	//draw status for all players.
	S_Rect r1 = {0, 0, 224, 16};
	int i, j, k = 0;

	int iNumPlayers = 0;
	for(i=0; i<8; i++) {
		if(GetShipPlayerObject(i)) iNumPlayers++;
	}

	int iCurPos = 0;
	for(i=0; i<8; i++) {
		C_ShipPlayer *pclShip = GetShipPlayerObject(i, M_AIPLAYER|M_ENEMY);
		if(pclShip && !(!m_pclPlayer && i==m_iReplayFollowPlayer)) {
			int iNumShips = pclShip->m_iNumLifes;
			if(iNumShips>0) {
				for(j=0; j<iNumShips; j++) {
					pclShip->DrawMiniShip(MAPRECT.x+2+j*12, MAPRECT.y+0+k*16, 16);
				}
			}
			k++;
		}
		if(i==m_iPlayerIndex) pclShip = m_pclPlayer;

		if(pclShip) {
			int iPos = PLAYERSPOS[iNumPlayers-1][iCurPos];
			S_FColor stCol = {0.50f, 0.50f, 0.50f, 0.36f};
			m_pclGraph->SetColor(&stCol);
			r1.x = STATUSPOS[iPos].iX;
			r1.y = STATUSPOS[iPos].iY;
			if(IsRace()) {
				r1.x -= 48;
			}
			m_pclGraph->Rect(&r1, true);

			if(m_bGameFinished || pclShip->IsPlayerLost()) {
				S_FColor stGrey = {0.75f,0.75f,0.75f, 1};
				m_pclFont->SetColor(stGrey);
			} else {
				m_pclFont->SetColor(STATUSPOS[i].stColor1);
			}

			if(IsRace()) {
				//race:             name, best lap time, current lap and cp
				double dBestLap = pclShip->m_dBestLapTime>999000 ? 999 : pclShip->m_dBestLapTime/1000.0;
				m_pclFont->Print(r1.x+2,   r1.y, pclShip->m_szName);
				m_pclFont->Print(r1.x+102, r1.y, "%.2lf", dBestLap);
				m_pclFont->Print(r1.x+164, r1.y, "Lap %d[%d]", pclShip->m_iCurLap, pclShip->m_iCurCP);
			} else {
				int iScore = 0;
				GetScore(i, &iScore);
				//mission/dogfight: name, score
				m_pclFont->Print(r1.x+2,   r1.y, pclShip->m_szName);
				m_pclFont->Print(r1.x+102, r1.y, "%d", iScore/1000);
			}
			iCurPos++;
		}
	}

	//restore font color
	m_pclFont->SetColor(g_stColorWhite);
}

//draw specific info for this player
void C_Viewport::DrawAdditionalPlayerStatus(int i_iActivePlayer)
{
	C_ShipPlayer *pclPlayer = GetShipPlayerObject(i_iActivePlayer);
	if(!pclPlayer) return;

	if(IsRace()) {
		m_pclGraph->Blt(m_pclCargoFuelBar, NULL, MAINPLAYERSTATUSPOS.iX, MAINPLAYERSTATUSPOS.iY, true);
		//draw race status (Lap)
		char szInfoString[32];
		sprintf(szInfoString, "%d/%d", pclPlayer->m_iCurLap, m_iTotalLaps);

		m_pclFont->Print(MAINPLAYERSTATUSPOS.iX+30, MAINPLAYERSTATUSPOS.iY+4, szInfoString);

		//draw race time
		double dSec = m_dTotalTime[i_iActivePlayer]/1000;
		int iDecimals = 2;
		if(dSec >= 100) iDecimals = 1;
		if(dSec >= 1000) iDecimals = 0;
		sprintf(szInfoString, "%.*f", iDecimals, dSec);

		m_pclFont->Print(MAINPLAYERSTATUSPOS.iX+30, MAINPLAYERSTATUSPOS.iY+25, szInfoString);
	} else {
		S_FColor *pstColor;
		double dPercentUsed;
		int iWidth;
		if(IsDogfight()) {
			S_Rect stRect = {0,22, 112, 20};
			m_pclGraph->Blt(m_pclCargoFuelBar, &stRect, MAINPLAYERSTATUSPOS.iX, MAINPLAYERSTATUSPOS.iY+22, true);
		} else {
			m_pclGraph->Blt(m_pclCargoFuelBar, NULL, MAINPLAYERSTATUSPOS.iX, MAINPLAYERSTATUSPOS.iY, true);
			//draw cargostatus
			pstColor = ((pclPlayer->CargoNumInHold()>=MAXSPACEINHOLDUNITS) ||
				(pclPlayer->CargoWeightInHold()>=MAXSPACEINHOLDWEIGHT)) ? &MAINPLAYERSTATUSPOS.stColor2 : &MAINPLAYERSTATUSPOS.stColor1;
			dPercentUsed = (double)pclPlayer->CargoWeightInHold() / MAXSPACEINHOLDWEIGHT;
			iWidth = (int)(110*dPercentUsed);
			if(iWidth>0) {
				S_Rect stRect = {MAINPLAYERSTATUSPOS.iX+1, MAINPLAYERSTATUSPOS.iY+1, iWidth, 18};
				m_pclGraph->SetColor(pstColor);
				m_pclGraph->Rect(&stRect);
			}
		}
		//draw fuelstatus
		pstColor = &MAINPLAYERSTATUSPOS.stColor1;
		dPercentUsed = pclPlayer->FuelPercentLeft();
		iWidth = (int)(110*dPercentUsed);
		if(iWidth>0) {
			S_Rect stRect = {MAINPLAYERSTATUSPOS.iX+1, MAINPLAYERSTATUSPOS.iY+23, iWidth, 18};
			m_pclGraph->SetColor(pstColor);
			m_pclGraph->Rect(&stRect);
		}

		//draw num lifes
		int i, iNumShips = pclPlayer->m_iNumLifes;
		if(iNumShips>0) {
			int y = (MAINPLAYERSTATUSPOS.iY-24)+(IsDogfight()?22:0);
			for(i=0; i<iNumShips; i++) {
				int x = MAINPLAYERSTATUSPOS.iX+i*12;
				pclPlayer->DrawMiniShip(x, y, 16);
			}
		}
	}
}

//draw the minimap
// ships are always drawn, but there are some specific drawing per map type
// - dogfight: no extra info
// - mission: bases with cargo left (red), homebases (blue)
// - race: current checkpoint (blue line)
void C_Viewport::DrawMinimap(float i_fScaling)
{
	//minimap
	int i, iTileSize = m_pclMap->GetTileSize();
	int iW, iH;
	m_pclMiniMap->GetInfo(&iW, &iH, NULL);
	S_FColor stColor = S_FCOLORMAKE(0.9f, 0.9f, 0.9f, 0.8f);
	m_pclGraph->SetBltColor(&stColor);
	S_Rect stDstRect = {MAPRECT.x+MAPRECT.width-(int)(iW*i_fScaling+12), MAPRECT.y+MAPRECT.height-(int)(iH*i_fScaling+12),
		(int)(iW*i_fScaling), (int)(iH*i_fScaling)};
	m_pclGraph->Blt(m_pclMiniMap, NULL, &stDstRect, true);
	m_pclGraph->SetBltColor(NULL);

	//always draw players on minimap
	for(i=0; i<8; i++) {
		C_ShipPlayer *pclShip = GetShipPlayerObject(i);
		if(pclShip) {
			int x, y;
			pclShip->GetCenterPos(&x, &y);
			S_Rect stRect = {(int)(x*i_fScaling/iTileSize)+stDstRect.x, (int)(y*i_fScaling/iTileSize)+stDstRect.y,
				(int)(2*i_fScaling), (int)(2*i_fScaling)};
			m_pclGraph->SetColor(&STATUSPOS[i].stColor1);
			m_pclGraph->Rect(&stRect);
		}
	}

	if(IsRace() && m_pclPlayer) {
		//draw current checkpoint
		S_FColor stColorCP = S_FCOLORMAKE(0.2f, 0.3f, 1.0f, 1.0f);
		m_pclGraph->SetColor(&stColorCP);

		S_CheckPointInfo stCP;
		m_clCheckPointList[m_pclPlayer->m_iCurCP]->GetCheckPointInfo(&stCP);
		stCP.iX1 = (int)(stCP.iX1*i_fScaling/iTileSize)+stDstRect.x; stCP.iY1 = (int)(stCP.iY1*i_fScaling/iTileSize)+stDstRect.y;
		stCP.iX2 = (int)(stCP.iX2*i_fScaling/iTileSize)+stDstRect.x; stCP.iY2 = (int)(stCP.iY2*i_fScaling/iTileSize)+stDstRect.y;

		m_pclGraph->Line(stCP.iX1, stCP.iY1, stCP.iX2, stCP.iY2, 1.5f*i_fScaling);
	} else if(IsMission()) {
		S_FColor stColorHome      = S_FCOLORMAKE(0.2f, 0.3f, 1.0f, 1.0f);
		S_FColor stColorWithCargo = S_FCOLORMAKE(1.0f, 0.3f, 0.2f, 1.0f);

		//draw cargo and home base positions
		int x, y, iSize = (int)m_clLandingZoneList.size();
		for(i=0; i<iSize; i++) {
			bool bDraw = false;

			if(m_clLandingZoneList[i]->IsHomeBase()) {
				m_pclGraph->SetColor(&stColorHome);
				bDraw = true;
			} else if(m_clLandingZoneList[i]->GetTotalCargo()>0) {
				m_pclGraph->SetColor(&stColorWithCargo);
				bDraw = true;
			}

			if(bDraw) {
				int iZoneSize = m_clLandingZoneList[i]->GetZoneSize();
				m_clLandingZoneList[i]->GetPos(&x, &y);
				S_Rect stRect = {(int)(x*i_fScaling/iTileSize)+stDstRect.x, (int)(y*i_fScaling/iTileSize)+stDstRect.y,
					(int)(iZoneSize*i_fScaling), (int)(2*i_fScaling)};
				m_pclGraph->Rect(&stRect);
			}
		}
	}
}

//set laps left (info from server on race maps)
// previously "num kills" was variable for dogfigt,
// but now each player has a fixed number of lifes
void C_Viewport::SetKillsOrLapsLeft(int i_iLeft)
{
	m_iKillsLeft = i_iLeft;
	if(m_iTotalLaps == -1) m_iTotalLaps = i_iLeft; //1st time
}

//set player fireing option
void C_Viewport::SetShootingMode(bool i_bFireAllowed)
{
	for(int i=0; i<8; i++) {
		C_Player *pclPlayer = GetPlayerObject(i);
		if(pclPlayer) pclPlayer->SetShootingMode(i_bFireAllowed);
	}
}

//set player<->player collission option
void C_Viewport::SetPlayerPlayerCollision(bool i_bCollision, bool i_bBulletsHit)
{
	m_bPlayerPlayerCollision  = i_bCollision;
	m_bPlayerBulletsCollision = i_bBulletsHit;
}

//take actions on reply from server
void C_Viewport::HandleCargoReply(S_NetMsg2 *i_pstMsg2)
{
	int i, iSize = (int)m_clLandingZoneList.size();
	for(i=0; i<iSize; i++) {
		C_LandingZone *pclZone = m_clLandingZoneList[i];
		if(i_pstMsg2->iZoneId == pclZone->GetOwnerID()) {
			int iCargo = pclZone->PopCargo();
			if(i_pstMsg2->iCargoReqAccepted) {
				int iId = i_pstMsg2->iPlayerId;
				C_Player *pclPlayer = GetPlayerObject(iId);
				if(pclPlayer) pclPlayer->LoadCargo(iCargo, pclZone->GetOwnerID());
			}
		}
	}
	//i_pstMsg2->iCargoNum //should not need to be used...
}

//take actions on reply from server
void C_Viewport::HandleCargoRespawn(S_NetMsg2 *i_pstMsg2)
{
	int i, iSize = (int)m_clLandingZoneList.size();
	for(i=0; i<iSize; i++) {
		C_LandingZone *pclZone = m_clLandingZoneList[i];
		if(i_pstMsg2->iZoneId == pclZone->GetOwnerID()) {
			pclZone->PushCargo(i_pstMsg2->iCargoWeight);
		}
	}
	//i_pstMsg2->iCargoNum //should not need to be used...
}

//take actions on reply from server
void C_Viewport::HandleLifeReply(S_NetMsg2 *i_pstMsg2)
{
	int i, iSize = (int)m_clLandingZoneList.size();
	for(i=0; i<iSize; i++) {
		C_LandingZone *pclZone = m_clLandingZoneList[i];
		if(i_pstMsg2->iZoneId == pclZone->GetOwnerID()) {
			pclZone->TakeExtraLife();
			if(i_pstMsg2->iLifeReqAccepted) {
				C_ShipPlayer *pclPlayer = GetShipPlayerObject(i_pstMsg2->iPlayerId);
				if(pclPlayer) pclPlayer->AddLife();
			}
		}
	}
}

//puts a chat message to be shown
void C_Viewport::AddChatString(char *i_szChatString)
{
	m_pclChat->AddChatString(i_szChatString);
}

//sets this user to be the new master of the game since
// the current master has left the game
void C_Viewport::NewMaster()
{
	m_bMaster = true;

	//computer enemies
	int i, iSize = (int)m_clComputerEnemyList.size();
	for(i=0; i<iSize; i++) {
		m_clComputerEnemyList[i]->SetMaster(true);
	}

	//AI players
	for(i=0; i<8; i++) {
		if(m_abIsAI[i]) {
			AddAI(i, m_pclEnemy[i]->m_szName);
			m_pclAIPlayer[i]->InitFromObject(m_pclEnemy[i]);
			m_pclAIPlayer[i]->SetScore(m_pclEnemy[i]->GetScore());
			m_pclAIPlayer[i]->m_iNumLifes = m_pclEnemy[i]->m_iNumLifes;
			RemoveEnemy(i);
		}
	}

	//doors
	iSize = (int)m_clDoorList.size();
	for(i=0; i<iSize; i++) {
		m_clDoorList[i]->SetMaster(true);
	}
}

//private method to get a given ship player object (this player, AI, or network player)
C_ShipPlayer *C_Viewport::GetShipPlayerObject(int i_iIndex, int i_iMask)
{
	if(i_iIndex<0 || i_iIndex>7) return NULL;
	if((i_iMask&M_PLAYER)   && i_iIndex==m_iPlayerIndex && m_pclPlayer) return m_pclPlayer;
	if((i_iMask&M_AIPLAYER) && m_pclAIPlayer[i_iIndex]) return m_pclAIPlayer[i_iIndex];
	if((i_iMask&M_ENEMY)    && m_pclEnemy[i_iIndex]) return m_pclEnemy[i_iIndex];
	return NULL;
}

//private method to get a given ship player object (this player or AI)
C_Player *C_Viewport::GetPlayerObject(int i_iIndex, int i_iMask)
{
	if(i_iIndex<0 || i_iIndex>7) return NULL;
	if((i_iMask&M_PLAYER)   && m_iPlayerIndex==i_iIndex && m_pclPlayer) return m_pclPlayer;
	if((i_iMask&M_AIPLAYER) && m_pclAIPlayer[i_iIndex]) return m_pclAIPlayer[i_iIndex];
	return NULL;
}
