// C_Viewport

#ifndef _VIEWPORT_H
#define _VIEWPORT_H

#include "netmsg.h"
#include "netmsgserver.h"

#include "tilemap.h"
#include "soundfx.h"
#include "input.h"
#include "glgraph.h"
#include "font.h"
#include "player.h"
#include "bullet.h"
#include "mapobjects.h"
#include "ingamechat.h"

#include "common.h"

#define M_PLAYER   0x01
#define M_AIPLAYER 0x02
#define M_ENEMY    0x04

class C_Viewport
{
public:
	C_Viewport(char *i_szMapFile, int i_iPlayerIndex, char *i_szPlayerName, bool i_bMaster, C_NetMsg *i_pclNetMsg, C_NetMsgServer *i_pclNetMsgServer);
	~C_Viewport();

	void AddEnemy(int i_iPlayerIndex, char *i_szPlayerName, bool i_bIsAI = false);
	void RemoveEnemy(int i_iPlayerIndex);
	void AddAI(int i_iPlayerIndex, char *i_szPlayerName);

	bool GetScore(int i_iPlayerIndex, int *o_iScore);
	void SetScore(int i_iPlayerIndex, int i_iScore);

	bool GetNumLifes(int i_iPlayerIndex, int *o_iLifes);
	void SetNumLifes(int i_iPlayerIndex, int i_iLifes);

	int MapType() {return m_pclMap->GetType();};
	bool IsRace() {return m_pclMap->GetType()==MAP_RACE;};
	bool IsMission() {return (m_pclMap->GetType()==MAP_MISSION) || (m_pclMap->GetType()==MAP_MISSION_COOP);};
	bool IsDogfight() {return m_pclMap->GetType()==MAP_DOGFIGHT;};
	int GetRaceTime(int i_iIndex);

	void SetKillsOrLapsLeft(int i_iLeft);
	int GetKillsOrLapsLeft() {return m_iKillsLeft;};

	void SetShootingMode(bool i_bFireAllowed);
	void HandleCargoReply(S_NetMsg2 *i_pstMsg2);
	void HandleCargoRespawn(S_NetMsg2 *i_pstMsg2);
	void HandleLifeReply(S_NetMsg2 *i_pstMsg2);

	void SetGameFinished() {m_bGameFinished=true;};
	void AddChatString(char *i_szChatString);
	bool InEnteringTextMode() {return m_pclChat->InEnteringTextMode();};

	void Draw();
	int Update(double i_dTime); //if you are killed by a human player, id of the killer is returned, else -1

	void NewMaster();
	void SetNumLifes(int i_iLifes);
	void SetPlayerPlayerCollision(bool i_bCollision, bool i_bBulletsHit);

	C_ShipPlayer *GetShipPlayerObject(int i_iIndex, int i_iMask = (M_PLAYER | M_AIPLAYER | M_ENEMY));
	C_Player *GetPlayerObject(int i_iIndex, int i_iMask = (M_PLAYER | M_AIPLAYER));

	//used by ai player
	C_CheckPointVector *GetCheckPoints() {return &m_clCheckPointList;};

	bool IsSinglePlayer();
private:
	bool CargoEndCondition(double dTime);
	double m_dCargoEndTimer;

	void SetSize(int i_iWidth, int i_iHeight);
	void DrawAllPlayersStatus();
	void DrawAdditionalPlayerStatus(int i_iActivePlayer);
	void DrawMinimap(float i_fScaling);
	C_Image          *m_pclCargoFuelBar;

	C_Map            *m_pclMap;
	C_Control        *m_pclControl;
	C_GraphWrapperGL *m_pclGraph;
	C_Sound          *m_pclSoundFX;

	//fps counter variables
	C_Font           *m_pclFont;
	bool             m_bShowFPS;
	double           m_dFPSTime;
	int              m_iFPSCount;
	int              m_iFPS;
	char             m_szFPSString[16];

	//minimap
	bool             m_bShowMiniMap;
	C_Image          *m_pclMiniMap;

	//end conditions
	int              m_iKillsLeft; //also laps left
	int              m_iTotalLaps;

	//network
	bool             m_bMaster;
	C_NetMsg         *m_pclNetMsg;
	C_NetMsgServer   *m_pclNetMsgServer;
	bool             m_bNextLevelSent;

	//world objects
	C_Player         *m_pclPlayer;
	int              m_iPlayerIndex;
	C_Enemy          *m_pclEnemy[8]; //only 7 can be occupied
	bool             m_abIsAI[8]; //used for setting new master
	C_Player         *m_pclAIPlayer[8]; //only 7 can be occupied, only used for master
	C_BulletList     *m_pclBulletList;
	C_LandingZoneVector   m_clLandingZoneList;
	C_ComputerEnemyVector m_clComputerEnemyList;
	C_CheckPointVector    m_clCheckPointList;
	C_DoorVector          m_clDoorList;

	//race time
	bool    m_bTimeCounting[8];
	double  m_dTotalTime[8];
	double  m_dLastNewLapTime[8];

	C_Chat  *m_pclChat;

	bool    m_bPlayerPlayerCollision;
	bool    m_bPlayerBulletsCollision;

	double  m_dTimePlayed;
	bool    m_bGameFinished;

	//replay
	void ReplayFollow(bool i_bNext);
	int  m_iReplayFollowPlayer;
	int  m_iReplayOldCP;
};

#endif
