// C_BmpImg

//class to load .bmp images

#include "bmp.h"
#include "fileresource.h"

#pragma pack(1)
struct S_BmpHeader
{
	uint8_t   TypeB;         // signature - 'BM'
	uint8_t   TypeM;         // ..
	uint32_t  Size;          // file size in bytes
	uint16_t  Reserved1;     // 0
	uint16_t  Reserved2;     // 0
	uint32_t  OffBits;       // offset to bitmap
	uint32_t  StructSize;    // size of this struct (40)
	uint32_t  Width;         // bmap width in pixels
	uint32_t  Height;        // bmap height in pixels
	uint16_t  Planes;        // num planes - always 1
	uint16_t  BitCount;      // bits per pixel
	uint32_t  Compression;   // compression flag
	uint32_t  SizeImage;     // image size in bytes
	int32_t   XPelsPerMeter; // horz resolution
	int32_t   YPelsPerMeter; // vert resolution
	uint32_t  ClrUsed;       // 0 -> color table size
	uint32_t  ClrImportant;  // important color count
};
 
#pragma pack()

C_Image *C_BmpImg::Load(char *i_szFilename, char *i_szResname)
{
	if(!i_szFilename && !i_szResname) return NULL;

	S_BmpHeader   stHeader;              //the header
	uint8_t       aPal[1024];            //pal in bmp
	C_Image       *pclImg = NULL;        //the image to fill

	C_Resource    *pclResource;

	uint32_t      iBytesPerPixel;        //number of bytes per pixel (1,2,3,4)
	int           iPixelformat;
	int y, iWidth, iHeight, iWidthBytes, iLineBytes, iBmpPadBytes;
	uint8_t *pBuf;

	pclResource = new C_Resource();
	pclResource->SetFilename(i_szFilename, i_szResname); //open file
	if(!pclResource->Read(&stHeader, sizeof(S_BmpHeader))) goto error; //read header

	//check that the bmp is valid
	iBytesPerPixel = stHeader.BitCount/8;
	if(stHeader.TypeB!='B' || stHeader.TypeM!='M') goto error;
	if(stHeader.StructSize!=40) goto error;
	if(iBytesPerPixel<1 || iBytesPerPixel>4) goto error;
	if(stHeader.Compression!=0) goto error;   

	//create the destination image
	iPixelformat = iBytesPerPixel;
	if(iBytesPerPixel == 1) iPixelformat = _PAL8;
	iWidth  = stHeader.Width;
	iHeight = stHeader.Height;
	iBmpPadBytes = (((iWidth*iBytesPerPixel+3)&(~0x3)))-(iWidth*iBytesPerPixel);
	pclImg = new C_Image(iWidth, iHeight, iPixelformat);
	pclImg->GetBufferMemory(&pBuf);
	pclImg->GetLineSizeInfo(&iWidthBytes, &iLineBytes, NULL);

	//get palette
	if(stHeader.ClrUsed) {
		if(stHeader.ClrUsed==256) {
			if(!pclResource->Read(aPal, stHeader.ClrUsed*4)) goto error; //read pal
			pclImg->SetPal(aPal);
		} else pclResource->Seek(stHeader.ClrUsed*4, SEEK_CUR); //skip
	}

	//get image
	pBuf += iHeight*iLineBytes;
	for(y=0; y<iHeight; y++) {
		pBuf -= iLineBytes;
		pclResource->Read(pBuf, iWidthBytes);
		if(iBmpPadBytes) pclResource->Seek(iBmpPadBytes, SEEK_CUR);
	}

	delete pclResource;
	return pclImg;
error:
	delete pclResource;
	delete pclImg;
	return NULL;
}
