// C_Font

//bitmap font class (use fnt_comp to create a font).
//this is a step back from glfont which uses true type fonts, but it was necessary to be able to
// use opengl with the same result on different platforms (no OS specific methods used).
// as a bonus tabs can be set and used, as well as multiline strings.

#include "fileresource.h"
#include "font.h"

//set default data
C_Font::C_Font(int i_iPixelFmt)
{
	m_pclImgFont = NULL; //no font loaded yet
	m_iSpaceY = 0;       //extra pixels between rows
	m_iSpaceX = 0;       //extra pixels between chars
	m_iFontHeight = 0;
	m_iNumChars = 0;
	m_bLoaded = false;

	m_iDstPixelFmt = i_iPixelFmt;
	m_pTabList     = NULL;

#ifdef _IMAGE_USE_OPENGL
	S_FColor stCol = S_FCOLORMAKE(1,1,1,1); //white
	SetColor(stCol);
#endif
}

//release font from memory if loaded
C_Font::~C_Font()
{
	if(m_pTabList) delete[] m_pTabList;
	if(m_pclImgFont) delete m_pclImgFont;
}

//loads a fontfile of the format fnt
bool C_Font::Load(char *i_szFontNameBase, char *i_szResname)
{
	if(m_pclImgFont) {
		delete m_pclImgFont;
		m_pclImgFont = NULL;
	}

	//read files (info about font and chars, and font image)
	S_FontHeader stHeader;
	S_CharHeader stChar;
	memset(&stHeader, 0, sizeof(stHeader));
	char szFilenameTGA[MAX_PATH];
	char szFilenameFNT[MAX_PATH];
	sprintf(szFilenameTGA, "%s.tga", i_szFontNameBase);
	sprintf(szFilenameFNT, "%s.fnt", i_szFontNameBase);
	C_Resource *pclResource = new C_Resource();
	pclResource->SetFilename(szFilenameFNT, i_szResname);
	pclResource->Read(&stHeader, sizeof(stHeader));
	if((stHeader.iMagicF != 'F') || (stHeader.iMagicN != 'N')) {
		delete pclResource;
		return false;
	}
	m_iFontHeight = stHeader.iHeight;
	m_iNumChars   = stHeader.iChars;
	m_iSpaceY     = stHeader.iSpaceY;
	for(int i=0; i<m_iNumChars; i++) {
		pclResource->Read(&stChar, sizeof(S_CharHeader));
		m_astChars[stChar.iIndex] = stChar;
	}
	delete pclResource;
	m_pclImgFont = new C_Image(szFilenameTGA, i_szResname, &m_bLoaded, m_iDstPixelFmt);
	if(!m_bLoaded) {
		delete m_pclImgFont;
		m_pclImgFont = NULL;
	}
	return m_bLoaded;
}

//set tab positions, 1st int is the total list length
void C_Font::SetTablist(int *i_pTabList)
{
	if(m_pTabList) {
		delete[] m_pTabList;
		m_pTabList = NULL;
	}
	if(!i_pTabList) return;
	int iSize = i_pTabList[0];

	m_pTabList = new int[iSize];
	memcpy(m_pTabList, i_pTabList, iSize*sizeof(int));
}

bool C_Font::GetCharYExt(int i_iIndex, int *o_iBeginY, int *o_iEndY)
{
	int x, y, iLineBytes, iBytesPerPixel;
	uint8_t *pBuf;
	S_Rect *pstRect = &m_astChars[i_iIndex].stRect;
	m_pclImgFont->GetInfo(NULL, NULL, &iBytesPerPixel);
	m_pclImgFont->GetLineSizeInfo(NULL, &iLineBytes, NULL);
	m_pclImgFont->GetBufferMemory(&pBuf);
	iBytesPerPixel = iBytesPerPixel&0xf;
	bool bFound = false;

	//scan for start line
	if(o_iBeginY) {
		int iStartPos = pstRect->y*iLineBytes+pstRect->x*iBytesPerPixel;
		int iSize = pstRect->width*iBytesPerPixel;
		for(y=0; y<pstRect->height; y++) {
			for(x=iStartPos; x<iStartPos+iSize; x++) {
				if(pBuf[x]!=0) {
					*o_iBeginY = y;
					bFound = true;
					break;
				}
			}
			iStartPos += iLineBytes;
			if(bFound) break;
		}
	}

	//scan for end line
	if(o_iEndY) {
		bFound = false;
		int iStartPos = (pstRect->y+pstRect->height-1)*iLineBytes+pstRect->x*iBytesPerPixel;
		int iSize = pstRect->width*iBytesPerPixel;
		for(y=pstRect->height-1; y>=0; y--) {
			for(x=iStartPos; x<iStartPos+iSize; x++) {
				if(pBuf[x]!=0) {
					*o_iEndY = y;
					bFound = true;
					break;
				}
			}
			iStartPos -= iLineBytes;
			if(bFound) break;
		}
	}

	return bFound;
}

//get text dimensions
bool C_Font::GetTextSize(char *i_szText, int *o_iWidth, int *o_iHeight, int *o_iActiveY, int *o_iActiveHeight, int *o_iNumLines, S_LineInfo **o_pstLineInfo)
{
	int iNumLines = 1, iLineLength;
	int i, iW = 0, iH = 0;

	if(!m_bLoaded || !i_szText) return false;

	//scan text for numlines
	char *szTemp = i_szText;
	while(*szTemp != 0) {
		if(*szTemp == '\t') szTemp++;
		else if(*szTemp == '\n') iNumLines++;
		szTemp++;
	}
	int iBeginY = -1, iEndY = -1;

	//get numcharsinline[] textlinewidth[] and maxtextwidth from text
	S_LineInfo *pstLine = new S_LineInfo[iNumLines];
	memset(pstLine, 0, iNumLines * sizeof(S_LineInfo));
	for(i=0; i<iNumLines; i++) {
		iLineLength = 0;
		while(*i_szText!='\n' && *i_szText!='\0') {
			if(*i_szText == '\t') {
				i_szText++;
				if(m_pTabList) pstLine[i].iWidth = m_pTabList[(int)(*i_szText)];
			} else {
				int iIdx = *(uint8_t*)i_szText;
				if(o_iActiveY && o_iActiveHeight) {
					bool bFetchBegin = iBeginY==-1;
					if(GetCharYExt(iIdx, bFetchBegin ? &iBeginY : NULL, &iEndY)) {
						if(i>0) {
							iEndY += i*(m_iFontHeight + m_iSpaceY);
							if(bFetchBegin) {
								iBeginY += i*(m_iFontHeight + m_iSpaceY);
							}
						}
					}
				}
				if(pstLine[i].iWidth+m_astChars[iIdx].iPreX>=0) pstLine[i].iWidth += m_astChars[iIdx].iPreX;
				pstLine[i].iWidth += m_astChars[iIdx].iPostX+m_astChars[iIdx].stRect.width + m_iSpaceX;
			}
			if(pstLine[i].iWidth > iLineLength) iLineLength = pstLine[i].iWidth;
			pstLine[i].iNumChars++;
			i_szText++;
		}
		pstLine[i].iWidth = iLineLength;
		if(pstLine[i].iWidth != 0) pstLine[i].iWidth -= m_iSpaceX;
		if(pstLine[i].iWidth > iW) iW = pstLine[i].iWidth;
		i_szText++; //skip end of line (13) (or 0 on the last line)
	}
	iH = iNumLines * (m_iFontHeight + m_iSpaceY) - m_iSpaceY;
	if(m_iSpaceY<0) iH -= m_iSpaceY;

	if(o_iActiveY && o_iActiveHeight) {
		*o_iActiveY = *o_iActiveHeight = 0;
		if(iBeginY!=-1 && iEndY!=-1) {
			*o_iActiveY = iBeginY;
			*o_iActiveHeight = (iEndY-iBeginY)+1;
		}
	}

	if(o_iWidth)  *o_iWidth  = iW;
	if(o_iHeight) *o_iHeight = iH;
	if(o_iNumLines) *o_iNumLines = iNumLines;
	if(o_pstLineInfo) *o_pstLineInfo = pstLine;
	else delete[] pstLine;
	return true;
}

//prints a formatted text with the loaded font to the given image
// within given area, centred if wanted
int C_Font::Print(C_Image **io_pclImg, S_Rect *i_pstArea, uint32_t i_iFlags, const char *i_szFmt, ...)
{
	va_list   stPList;
	char      szText[1024];

	if(!m_bLoaded || !i_szFmt) return 0;

	va_start(stPList, i_szFmt);
	vsprintf(szText, i_szFmt, stPList);
	va_end(stPList);

	return PutString(io_pclImg, i_pstArea, i_iFlags, szText);
}

#ifdef _IMAGE_USE_OPENGL
//set font color (only used when printing to screen)
void C_Font::SetColor(const S_FColor &i_stColor)
{
	m_stColor = i_stColor;
}

//prints a formatted text with the loaded font to the screen
// within given area, centred if wanted
int C_Font::Print(S_Rect *i_pstArea, uint32_t i_iFlags, const char *i_szFmt, ...)
{
	va_list   stPList;
	char      szText[1024];

	if(!m_bLoaded || !i_szFmt || !i_pstArea) return 0;
	S_Rect stArea = *i_pstArea;
	int x = stArea.x;
	int y = stArea.y;
	stArea.x = 0;
	stArea.y = 0;

	va_start(stPList, i_szFmt);
	vsprintf(szText, i_szFmt, stPList);
	va_end(stPList);

	C_Image *pclImg = NULL;
	i_iFlags |= FONT_CREATESURFACE;
	int iW = PutString(&pclImg, &stArea, i_iFlags, szText);

	if(pclImg) {
		C_GraphWrapperGL *pclGraph = C_GraphWrapperGL::GetGraphWrapperObject();
		pclGraph->SetBltColor(&m_stColor);
		pclGraph->Blt(pclImg, NULL, x, y, true);
		pclGraph->SetBltColor(NULL);
		delete pclImg;
	}

	return iW;
}

//prints a formatted text with the loaded font to the screen
int C_Font::Print(int x, int y, const char *i_szFmt, ...)
{
	va_list   stPList;
	char      szText[1024];

	if(!m_bLoaded || !i_szFmt) return 0;

	va_start(stPList, i_szFmt);
	vsprintf(szText, i_szFmt, stPList);
	va_end(stPList);

	C_Image *pclImg = NULL;
	int iFlags = FONT_CREATESURFACE;
	int iW = PutString(&pclImg, NULL, iFlags, szText);

	if(pclImg) {
		C_GraphWrapperGL *pclGraph = C_GraphWrapperGL::GetGraphWrapperObject();
		pclGraph->SetBltColor(&m_stColor);
		pclGraph->Blt(pclImg, NULL, x, y, true);
		pclGraph->SetBltColor(NULL);
		delete pclImg;
	}

	return iW;
}
#endif

//prints out a text with the loaded font to the given image
// within given area, centred if wanted
int C_Font::PutString(C_Image **io_pclImg, S_Rect *i_pstArea, uint32_t i_iFlags, char *i_szText)
{
	int         i, iNumLines, iTextW, iTextH;

	S_Rect      stTmpArea;
	int         iPixelFmt;

	S_LineInfo  *pstLine = NULL;

	//check if parameters are valid
	if(!m_bLoaded || !i_szText || !io_pclImg) return 0;
	GetTextSize(i_szText, &iTextW, &iTextH, NULL, NULL, &iNumLines, &pstLine);

	iPixelFmt = m_iDstPixelFmt;
	if(!(i_iFlags & FONT_CREATESURFACE)) (*io_pclImg)->GetInfo(&stTmpArea.width, &stTmpArea.height, &iPixelFmt);
	if(!i_pstArea) {
		i_pstArea = &stTmpArea;
		stTmpArea.x = stTmpArea.y = 0;
		if(i_iFlags & FONT_CREATESURFACE) {
			stTmpArea.width = iTextW;
			stTmpArea.height = iTextH;
		}
	}
	if(iPixelFmt != m_iDstPixelFmt || iTextH > i_pstArea->height || iTextW > i_pstArea->width)
		return 0;

	if(i_iFlags & FONT_CREATESURFACE) { //new surface
		*io_pclImg = new C_Image(i_pstArea->width, i_pstArea->height, m_iDstPixelFmt);
		(*io_pclImg)->FillWithAbsoluteColor(NULL, 0);
	}

	//put lines on image
	uint32_t iTransparentCol = 0x00000000;
	int iPos = 0;
	int iYStart = i_pstArea->y, iXStart = i_pstArea->x;
	if(i_iFlags & FONT_CENTERED) iYStart += (i_pstArea->height - iTextH)/2;
	for(i=0; i<iNumLines; i++) {
		if(i_iFlags & FONT_CENTERED) iXStart += (i_pstArea->width - pstLine[i].iWidth)/2;

		//put line on image
		while(i_szText[iPos]!='\n' && i_szText[iPos]!='\0') {
			int iIndex = (unsigned char)i_szText[iPos];
			iXStart += m_astChars[iIndex].iPreX; //can be negative
			if(iXStart<i_pstArea->x) iXStart = i_pstArea->x; //fix if negative
			(*io_pclImg)->Blt2(m_pclImgFont, &m_astChars[iIndex].stRect, iXStart, iYStart, &iTransparentCol);
			iXStart += m_astChars[iIndex].stRect.width+m_astChars[iIndex].iPostX+m_iSpaceX;
			iPos++;
		}
		iPos++;
		iXStart = i_pstArea->x;
		iYStart += m_iFontHeight+m_iSpaceY;
	}
	delete[] pstLine;

	return iTextW;
}
