// C_ImageDiskOP

//loads and saves images of different formats depending on extension
// .tga load/save (save is done as rle compressed)
// .bmp load only
// .jpg load only

#include "imagediskop.h"

#include "targa.h"
#include "jpeg.h"
#include "bmp.h"

struct S_FormatExt
{
	char const *szExt;
	int        iFormat;
};

#define NUMEXTENSIONS 7
S_FormatExt TBL_FORMATEXTENSION[NUMEXTENSIONS] = {
	{"tga", _TARGA}, {"targa", _TARGA}, {"jpg", _JPEG},
	{"jpeg", _JPEG}, {"jif", _JPEG}, {"jfif", _JPEG}, {"bmp", _BMP}
};

C_ImageDiskOP::C_ImageDiskOP()
{
}

C_ImageDiskOP::~C_ImageDiskOP()
{
}

C_ImageDiskOP::C_ImageDiskOP(char *i_szFilename, char *i_szRsname, C_Image **o_ppclDstImg)
{
	if(!o_ppclDstImg) return;
	Load(i_szFilename, i_szRsname, o_ppclDstImg);
	delete this;
}

bool C_ImageDiskOP::Load(char *i_szFilename, char *i_szRsname, C_Image **o_ppclDstImg)
{
	bool bResult = false;
	C_JpegImg  *pclJpg;

	if(!i_szFilename) return false;

	int iFormat = DetectFormat(i_szFilename); //normal resource
	switch(iFormat) {
		case _TARGA:
			*o_ppclDstImg = C_TargaImg::Load(i_szFilename, i_szRsname);
			if(*o_ppclDstImg) bResult = 1;
			break;
		case _JPEG:
			pclJpg = new C_JpegImg();
			*o_ppclDstImg = pclJpg->Load(i_szFilename, i_szRsname);
			if(*o_ppclDstImg) bResult = 1;
			delete pclJpg;
			break;
		case _BMP:
			*o_ppclDstImg = C_BmpImg::Load(i_szFilename, i_szRsname);
			if(*o_ppclDstImg) bResult = 1;
			break;
	}
	return bResult;
}

bool C_ImageDiskOP::Save(char *i_szFilename, C_Image *i_pclSrcImg)
{
	bool bResult = false;

	if(!i_szFilename) return false;

	int iFormat = DetectFormat(i_szFilename);
	switch(iFormat) {
		case _TARGA:
			bResult = C_TargaImg::SaveCompressed(i_szFilename, i_pclSrcImg);
			break;
		//case _JPEG:  //will not happen
	}
	return bResult;
}

int C_ImageDiskOP::DetectFormat(char *i_szFilename)
{
	int i = (int)strlen(i_szFilename) - 1;
	while(i>=0 && i_szFilename[i]!='.') i--;    //find last .
	if(i<0) return _ERROR;

	char szExt[16];
	strncpy(szExt, &i_szFilename[i+1], sizeof(szExt));
	szExt[sizeof(szExt)-1] = 0;

	int iFormat = _ERROR;
	for(i=0; i<NUMEXTENSIONS; i++) {
		if(_stricmp(TBL_FORMATEXTENSION[i].szExt, szExt) == 0) {
			iFormat = TBL_FORMATEXTENSION[i].iFormat;
			break;
		}
	}
	return iFormat;
}
