// C_Component

#ifndef _JPEG_INCLUDE_H
#define _JPEG_INCLUDE_H

#include "common.h"
#include "image.h"

//private to object
#define HUFF_LOOKAHEAD 16 //must be 16 right now
struct S_HuffTable
{
	int8_t  pcHT[256];    //all symbols in a huffman table
	uint8_t piBits[17];   //#codes with 1..16 bits
	uint8_t *piLookNBits; //#bits, or 0 if too long
	uint8_t *piLookSym;   //symbol, or unused
};

//color component in an image
class C_Component
{
public:
	C_Component();
	~C_Component();
	void AllocBuffer(int i_iWidthMCUs, int i_iHeightMCUs);
	void PutDCT(int16_t *i_piSrc);

	void SetNumDcts(int i_iNumDCTs, int i_iDctDimX, int i_iDctDimY) {
		m_iNumDCTs = i_iNumDCTs; m_iDimMCUx = i_iDctDimX; m_iDimMCUy = i_iDctDimY; };

	int     m_iLastDCVal;
	int     m_iACTab, m_iDCTab, m_iQTNum;

	uint8_t *m_piBuffer;

	int     m_iDimMCUx, m_iDimMCUy;
private:
	int     m_iNumMCUsX, m_iCurMCUX;
	int     m_iNumDCTs, m_iCurDCT;
	uint8_t *m_piBufPos;

	static uint8_t *s_piYUVRangeLimit;
	static int     s_iObjCount;
};

//handles bit input from stream
class C_JpegBitStream
{
public:
	C_JpegBitStream(uint8_t *i_piBitStream);
	uint32_t GetBits(int i_iNumBits); //works with maximum 24 bits
	void     IncPos(int i_iNumBits);    //works with many bits
	void     *GetStreamPosPtr() {return m_piBitPtr;};
private:
	uint8_t *m_piBitPtr;
	int     m_iCurBit;
};

//color conversion (YUV -> RGB)
class C_ColorYUV
{
public:
	C_ColorYUV();
	~C_ColorYUV();
	void Convert(C_Image *i_pclImage, C_Component *i_pclComponent, int dimx, int dimy);
private:
	static uint8_t *s_piRGBRangeLimit;
	static int     *s_CrRTab;
	static int     *s_CbBTab;
	static int     *s_CrGTab;
	static int     *s_CbGTab;
	static int     s_iObjCount;
};

#endif
