// C_HiScoreList

//class to handle the server hiscore list
// - save to disk
// - load from disk
// - add/replace entries depending on score

#include "common.h"
#include "hiscorelist.h"


//copy filename to local, and load file
C_HiScoreList::C_HiScoreList(char *i_szFileName, bool i_bAccending, bool *o_bFileExisted)
{
	m_szFileName = new char[strlen(i_szFileName)+1];
	strcpy(m_szFileName, i_szFileName);

	m_bAccending = i_bAccending;

	if(o_bFileExisted) *o_bFileExisted = Load();
	else Load();
}

//release mem
C_HiScoreList::~C_HiScoreList()
{
	delete[] m_szFileName;
}

//sets the levelnamefield of the list, need only be done in a new list
void C_HiScoreList::SetLevelName(char *i_szLevelName)
{
	strcpy(m_stList.szLevelName, i_szLevelName);
}

//puts a valid hiscore to the list
// also saves the file
bool C_HiScoreList::Add(char *i_szName, int i_iScore)
{
	bool bResult = false;

	//list is sorted in order, keep it that way
	int i, j;
	for(i=0; i<m_iNumInList; i++) { //search list for place of new score
		if((m_bAccending && m_stList.iScore[i]<i_iScore) || (!m_bAccending && i_iScore<m_stList.iScore[i])) {
			bResult = true;
			break;
		}
	}
	//value of i is the pos

	if(m_iNumInList < MAXSCORESINLIST) {
		bResult = true;
		m_iNumInList++; //resize list if not full
	}

	if(bResult) {
		//insert score to list
		for(j=m_iNumInList-2; j>=i; j--) {
			//reorder list from place of insertion
			strcpy(m_stList.szName[j+1], m_stList.szName[j]);
			m_stList.iScore[j+1] = m_stList.iScore[j];
		}

		//insert new score on position
		strcpy(m_stList.szName[i], i_szName);
		m_stList.iScore[i] = i_iScore;
		//save on disk
		Save();
	}

	return bResult;
}

//load list from file if exists
bool C_HiScoreList::Load()
{
	//assume empty list
	m_iNumInList = 0;

	FILE *pstFile = fopen(m_szFileName, "rb");
	if(pstFile) {
		size_t iNum = fread(&m_stList, sizeof(S_HiScore), 1, pstFile);
		fclose(pstFile);
		if(iNum==1 && m_stList.szMagic[0]=='G' && m_stList.szMagic[1]=='F' && m_stList.szMagic[2]=='H' && m_stList.szMagic[3]=='I') {
			m_iNumInList = m_stList.iNumInList;
			Invert(&m_stList);
			return true;
		}
	}
	//file did not exist, or is bad (or of a previous version)
	//begin again:
	m_stList.iNumInList = 0;
	m_stList.szMagic[0] = 'G';
	m_stList.szMagic[1] = 'F';
	m_stList.szMagic[2] = 'H';
	m_stList.szMagic[3] = 'I';

	return false;
}

//save list to file if list is not empty
void C_HiScoreList::Save()
{
	//if list empty, nothing to save
	if(!m_iNumInList) return;

	m_stList.iNumInList = m_iNumInList;

	Invert(&m_stList);

	FILE *pstFile = fopen(m_szFileName, "wb");
	if(pstFile) {
		fwrite(&m_stList, sizeof(S_HiScore), 1, pstFile);
		fclose(pstFile);
	}

	Invert(&m_stList);
}

//a very simple crypto to make the score data file
// unreadable; just inverts all bits
void C_HiScoreList::Invert(S_HiScore *io_pstHiscore)
{
	int i, j;

	for(i=0; i<MAXSCORESINLIST; i++) {
		//io_pstHiscore->iScore[i] = ~io_pstHiscore->iScore[i];
		for(j=0; j<MAXPLAYERNAMELENGTH; j++) {
			io_pstHiscore->szName[i][j] = (char)~io_pstHiscore->szName[i][j];
		}
	}
}

//return the hiscore list contents
bool C_HiScoreList::GetAsData(S_HiScore *o_pstHiScore)
{
	if(m_iNumInList) {
		m_stList.iNumInList = m_iNumInList;
		memcpy(o_pstHiScore, &m_stList, sizeof(S_HiScore));
		return true;
	}
	return false;
}
