// C_Game

#ifndef _CGAME_H
#define _CGAME_H

#include <vector>
#include "../galaxyv2/netmsgserver.h"
#include "timer.h"
#include "logger.h"
#include "s_player.h"

#define NUM_LIFES 5     //in galaxy.cpp also

#define MAP_MISSION  1  //in tilemap.h also
#define MAP_DOGFIGHT 2  //..
#define MAP_RACE     3  //..
#define MAP_MISSION_COOP 4

struct S_ZoneCargo
{
	int iZoneId;
	int iCargoNum;
	//int iCargoWeight;
	bool bLifeTaken;
	//int iTakenBy;
};
typedef std::vector<S_ZoneCargo> C_ZoneTakenList;

class C_Game
{
public:
	C_Game(int i_iGameId, char *i_szGameName, C_Player *i_pclPlayer, C_Logger *i_pclLog);
	~C_Game();

	//adds a player to the game, sends msg to players in the game
	void JoinToGame(C_Player *i_pclPlayer);

	//resends a received chat msg to all players in the game
	void SendChat(char *i_szChatString);
	//resends a received msg to all players in the game
	void SendMsg(S_NetMsg2 *i_pstMsg);
	//resend a received msg to all players in the game but the sending player
	void SendMsg(S_NetMsg2 *i_pstMsg, C_Player *i_pclPlayer);

	//resend a received msg to all players in the game but the sending player
	void SendMsg1(S_NetMsg1 *i_pstMsg, C_Player *i_pclPlayer);

	//check for start condition met, send START_R to all players in game
	void SendStart();

	//handles a N_MSG2_CARGO msg (sends a N_MSG2_CARGO_R msg to all players in game)
	void HandleCargoRequest(S_NetMsg2 *i_pstMsg, int i_iPlayerIndex, C_Player *i_pclSender);
	//handles a N_MSG2_CARGORESPAWN msg (and resends it to all players in game)
	void HandleCargoRespawn(S_NetMsg2 *i_pstMsg, int i_iPlayerIndex);
	//handles a N_MSG2_LIFE msg (sends a N_MSG2_LIFE_R msg to all players in game)
	void HandleLifeRequest(S_NetMsg2 *i_pstMsg, int i_iPlayerIndex, C_Player *i_pclSender);
	//handles a N_MSG2_NEXTLEVEL msg (resends the msg, or a N_MSG2_GAMEOVER msg to all players in game)
	void HandleNextLevel();

	void RemovePlayer(C_Player *i_pclPlayer);

	//helperfunctions
	//returns index of next free playerId
	// if no free index exists or if the game has started -1 is returned
	int CheckPreState();

	int  GetNumPlayers();
	bool GetPlaying() {return m_bPlaying;};
	bool GetStatus();    //red or green in gameview
	bool StatusChanged();//has it changed?
	void SetLevelType(int i_iLevelType) {m_iLevelType = i_iLevelType;};
	int GetLevelType() {return m_iLevelType;};
	char *GetLevelName() {return m_szLevelName;};

	void AddKill(int i_iPlayerIndex, int i_iKilledBy);
	void AddLap(int i_iPlayerIndex);

	bool CompareGameName(char *i_szName);
	char *GetGameName();

	C_Player *GetPlayer(int i_iIndex);

	bool CheckGameOver(bool i_bForceGameOver = false);
private:
	int m_iId;

	void ChangeGameStatus();
	void SaveImportantInfo(S_NetMsg2 *i_pstMsg);

	//sends to options of this game (if set) to the player
	void SendGameOptions(C_Player *i_pclPlayer);

	C_PlayerList *m_pclPlayerList;
	C_Timer *m_pclTimer;

	//game properties
	char m_szName[MAXGAMENAMELENGTH]; //gamename

	//gameover conditions
	int m_iNumLaps;
	//when m_iKillCount reaches 0 game is over

	//gameover status
	int    m_iKillCountAI[8];
	int    m_iLapsCountAI[8];
	int    m_iLifesAI[8];
	bool   m_aiFreeIndexForAI[8];
	int    m_iKillCount;
	int    m_iLapsCount;

	//level info
	char m_szLevelName[24];
	bool m_bNoShooting;
	bool m_bFlyThrough;
	bool m_bNoBulletHit;
	int  m_iNumAI;

	bool m_bPlaying;
	bool m_bGameStatus;     //red=1, green=0
	bool m_bGameStatusChanged;
	bool m_bOptionsSet;

	int m_iLevelType; //1 mission 2 dogfight 3 race

	C_ZoneTakenList m_clCargo;
	C_Logger *m_pclLog;
};
typedef std::vector<C_Game *> C_GameList;

#endif
