
//main startup function and
//class to tie everything in the server together.

//network
#include "socket_2.h"
#include "global.h"
#include "../galaxyv2/netmsgserver.h"
#include "../galaxyv2/settings.h"

//system
#include "common.h"
#include "resource_srv.h"

//game
#include "s_manager.h"
#include "s_game.h"
#include "s_player.h"

#define SERVER_TITLE "Galaxy V2 Server v1.85"

#define IP_TEXT      "IP addresses to this server:"
#define SERVER_TEXT  "Galaxy Forces V2 Game Server"
#define GAMES_TEXT   "Games on this server:"
#define PLAYERS_TEXT "Connected players on this server:"

class C_ServerHandler {
public:
	C_ServerHandler();
	~C_ServerHandler();

	void PollSocket();
	C_Manager *GetManager() {return m_pclManager;};

	void Update();
private:
	//network
	C_Socket  *m_pclListenSocket;
	//server
	C_Manager *m_pclManager;
};

C_ServerHandler::C_ServerHandler()
{
	C_Global::Init((char*)"config.txt", (char*)".GalaxyV2");
	C_Settings::Init((char*)"config.txt");

	m_pclListenSocket = new C_Socket(C_Settings::iServerPort);
	m_pclManager = new C_Manager();
}

C_ServerHandler::~C_ServerHandler()
{
	delete m_pclListenSocket; m_pclListenSocket = NULL;
	delete m_pclManager; m_pclManager = NULL;
}

void C_ServerHandler::PollSocket()
{
	//poll socket
	m_pclListenSocket->WaitForReadEvent(0);
	m_pclListenSocket->WaitForWriteEvent(0);

	if(m_pclListenSocket->m_bAcceptWillSucceed) {
		//put new connection to a player object
		m_pclManager->NewPlayer(m_pclListenSocket->AcceptConnection());
	}

	//it is bad to have this in a timer (polling).
	//it would be better if it runs directly when there is data to process.
	m_pclManager->ProcessNetMsg();
	//in previous code WSA_ASYNC was sent from the network class when an event
	// happened, but since the network code was changed to use socket code common
	// for both linux and windows that is not the case any more.
}

void C_ServerHandler::Update()
{
	m_pclManager->Update();
}

#ifdef WIN32

HINSTANCE   g_hInstance;
HWND        g_hWnd;
C_Manager   *g_pclManager;

void DrawAllServerInfo(HDC i_hDC, HWND i_hWnd)
{
	RECT rt;
	char *pszText;
	int i, iNum = 16;
	S_InterfaceInfo astList[16];

	//server info
	GetClientRect(i_hWnd, &rt);
	DrawText(i_hDC, SERVER_TEXT, (int)strlen(SERVER_TEXT), &rt, DT_CENTER);
	rt.top += 32;

	DrawText(i_hDC, IP_TEXT, (int)strlen(IP_TEXT), &rt, DT_LEFT);
	if(C_Socket::GetLocalAddressInfo(&astList[0], &iNum)) {
		for(i=0; i<iNum; i++) {
			rt.top += 18;
			DrawText(i_hDC, astList[i].szIP, (int)strlen(astList[i].szIP), &rt, DT_LEFT);
		}
	}
	rt.top += 32;

	DrawText(i_hDC, GAMES_TEXT, (int)strlen(GAMES_TEXT), &rt, DT_LEFT);
	if(g_pclManager) {
		//all games and players in them
		g_pclManager->ResetCurGame();
		C_Game *pclGame;
		while((pclGame = g_pclManager->GetNextGame()) != NULL) {
			pszText = g_pclManager->GetGameString(pclGame);
			rt.top += 18;
			DrawText(i_hDC, pszText, (int)strlen(pszText), &rt, DT_LEFT);
			rt.left += 32;

			pszText = g_pclManager->GetPlayerString(pclGame);
			if(pszText) {
				rt.top += 18;
				DrawText(i_hDC, pszText, (int)strlen(pszText), &rt, DT_LEFT);
				if(strchr(pszText, '\n')) rt.top += 18;
			}
			rt.left -= 32;
		}

		//all connected players not in a game
		rt.top += 32;
		DrawText(i_hDC, PLAYERS_TEXT, (int)strlen(PLAYERS_TEXT), &rt, DT_LEFT);
		g_pclManager->ResetCurUnnamedPlayer();
		while((pszText = g_pclManager->GetNextUnnamedPlayerIPString()) != NULL) {
			rt.top += 18;
			rt.left += 32;

			if(pszText) {
				DrawText(i_hDC, pszText, (int)strlen(pszText), &rt, DT_LEFT);
			}
			rt.left += 320;
			pszText = g_pclManager->GetNextUnnamedPlayerIPString();
			if(pszText) {
				DrawText(i_hDC, pszText, (int)strlen(pszText), &rt, DT_LEFT);
			}

			rt.left -= 32+320;
		}
	}
}

//message-handler for about box
LRESULT CALLBACK About(HWND hDlg, UINT iMessage, WPARAM wParam, LPARAM lParam)
{
	switch(iMessage) {
		case WM_INITDIALOG:
			{
				char szVersion[256];
				sprintf_s(szVersion, sizeof(szVersion), "%s (%s)\r\nBuilt %s %s", SERVER_TITLE, sizeof(void*)==8 ? "x64" : "x86", __DATE__, __TIME__);
				SetDlgItemText(hDlg, IDC_S_VERSION, szVersion);
			}
			return 1;
		case WM_COMMAND:
			if(LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) {
				EndDialog(hDlg, LOWORD(wParam));
				return 1;
			}
			break;
	}
	return 0;
}

//message-handler for main window
LRESULT CALLBACK WndProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;
	PAINTSTRUCT ps;
	HDC hdc;

	switch(iMessage) {
		case WM_COMMAND:
			wmId    = LOWORD(wParam); 
			wmEvent = HIWORD(wParam); 
			//parse the menu selections
			switch(wmId) {
				case IDM_ABOUT:
					DialogBox(g_hInstance, (LPCTSTR)IDD_ABOUTBOX, hWnd, (DLGPROC)About);
					break;
				case IDM_EXIT:
					DestroyWindow(hWnd);
					break;
				default:
					return DefWindowProc(hWnd, iMessage, wParam, lParam);
			}
			break;
		case WM_PAINT:
			hdc = BeginPaint(hWnd, &ps);
			DrawAllServerInfo(hdc, hWnd);
			EndPaint(hWnd, &ps);
			break;
		case WM_DESTROY:
			PostQuitMessage(0);
			break;
		case WM_TIMER:
			if(wParam==1) {
				RedrawWindow(g_hWnd, NULL, NULL, RDW_ERASE | RDW_INVALIDATE);
			}
			break;
		default:
			return DefWindowProc(hWnd, iMessage, wParam, lParam);
	}
	return 0;
}

bool InitInstance(HINSTANCE hInstance, int nCmdShow)
{
	g_hInstance = hInstance;

	WNDCLASSEX wcex;
	wcex.cbSize        = sizeof(WNDCLASSEX); 
	wcex.style         = CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc   = (WNDPROC)WndProc;
	wcex.cbClsExtra    = 0;
	wcex.cbWndExtra    = 0;
	wcex.hInstance     = hInstance;
	wcex.hIcon         = LoadIcon(hInstance, (LPCTSTR)IDI_ICON1);
	wcex.hCursor       = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName  = (LPCSTR)IDC_SERVER;
	wcex.lpszClassName = "SERVER_WND";
	wcex.hIconSm       = LoadIcon(hInstance, (LPCTSTR)IDI_ICON1);
	RegisterClassEx(&wcex);

	g_hWnd = CreateWindow("SERVER_WND", SERVER_TITLE, WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, CW_USEDEFAULT, 800, 600, NULL, NULL, hInstance, NULL);

	if(!g_hWnd) return false;

	ShowWindow(g_hWnd, nCmdShow);
	UpdateWindow(g_hWnd);

	return true;
}
#endif

#ifdef WIN32
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG    stMsg;
	HACCEL hAccelTable;
	bool bSilent = false;

	if(strstr(lpCmdLine, "-silent")!=NULL) bSilent = true;
#else
#include <signal.h>
volatile int g_iShutdown = 0;
static void sig_int(int signo)
{
	if(signo == SIGINT) {
		//make the application shutdown properly on ctrl+c
		g_iShutdown = 1;
	}
	return;
}
	
int main(int argc, char *argv[])
{
	bool bSilent = false;

	if(argc>=2 && strcmp(argv[1], "-silent")==0) bSilent = true;

	//this mechanism is here to give the pid to the game (the game uses parameter '-silent').
	//the games keeps the pid to be able to close the server when it is exited itself.
	//the process is forked here to get a pid, the new servers pid is printed on stdout by
	// the server started by the game, the server started by the game is then exited.
	//stdout is parsed by the game to get the pid of the 2nd server.
	//it was not possible to get the pid by the game when starting the server, fork had to be used.
	int iPID = 0;
	if(bSilent) iPID = fork();
	if(iPID) { //iPid is 0 for the new server, and ==pid for the process running fork
		printf("%d\n", iPID);
		return 0;
	}
	
	//create a mask holding only SIGINT - ctrl+c interrupt
	//install SIGINT handler
	if(signal(SIGINT, sig_int) == SIG_ERR) {
		printf("install SIGINT handler error\n");
	}
#endif

	C_ServerHandler *pclServer = new C_ServerHandler();

	if(!bSilent) {
#ifdef WIN32
		//windows initialization
		if(!InitInstance(hInstance, nCmdShow)) {
			return 0;
		}
		hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)IDC_SERVER);
		SetTimer(g_hWnd, 1, 5000, NULL); //redraw game info every 5 sec
#else
		//print server addresses
		int i, iNum = 16;
		S_InterfaceInfo astList[16];
		printf("%s\n\n", SERVER_TITLE);
		printf("%s\n", IP_TEXT);
		if(C_Socket::GetLocalAddressInfo(&astList[0], &iNum)) {
			for(i=0; i<iNum; i++) {
				printf("%s\n", astList[i].szIP);
			}
		}
#endif
	}

	//message loop
	int iLoopCount = 0;
	while(1) {
#ifdef WIN32
		if(!bSilent) {
			g_pclManager = pclServer->GetManager();
			if(PeekMessage(&stMsg, NULL, 0, 0, PM_NOREMOVE)) {
				if(!GetMessage(&stMsg, NULL, 0, 0)) break;

				if(!TranslateAccelerator(stMsg.hwnd, hAccelTable, &stMsg)) {
					TranslateMessage(&stMsg);
					DispatchMessage(&stMsg);
				}
				continue;
			}
		}
#endif
		pclServer->PollSocket();
		if((iLoopCount++%500)==0) pclServer->Update();
		mssleep(1); //1 to get lowest latency
#ifndef WIN32
		if(g_iShutdown!=0) break;
#endif
	}

	delete pclServer;
#ifdef WIN32
#ifdef _DEBUG
	_CrtDumpMemoryLeaks();
#endif
#endif

	return 0;
}
